(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory() :
    typeof define === 'function' && define.amd ? define(factory) :
    (factory());
}(this, (function () { 'use strict';

/**
 * @author alteredq / http://alteredqualia.com/
 * @author mrdoob / http://mrdoob.com/
 */

var _Math = {

	DEG2RAD: Math.PI / 180,
	RAD2DEG: 180 / Math.PI,

	generateUUID: function () {

		// http://www.broofa.com/Tools/Math.uuid.htm

		var chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'.split('');
		var uuid = new Array(36);
		var rnd = 0,
		    r;

		return function generateUUID() {

			for (var i = 0; i < 36; i++) {

				if (i === 8 || i === 13 || i === 18 || i === 23) {

					uuid[i] = '-';
				} else if (i === 14) {

					uuid[i] = '4';
				} else {

					if (rnd <= 0x02) rnd = 0x2000000 + Math.random() * 0x1000000 | 0;
					r = rnd & 0xf;
					rnd = rnd >> 4;
					uuid[i] = chars[i === 19 ? r & 0x3 | 0x8 : r];
				}
			}

			return uuid.join('');
		};
	}(),

	clamp: function clamp(value, min, max) {

		return Math.max(min, Math.min(max, value));
	},

	// compute euclidian modulo of m % n
	// https://en.wikipedia.org/wiki/Modulo_operation

	euclideanModulo: function euclideanModulo(n, m) {

		return (n % m + m) % m;
	},

	// Linear mapping from range <a1, a2> to range <b1, b2>

	mapLinear: function mapLinear(x, a1, a2, b1, b2) {

		return b1 + (x - a1) * (b2 - b1) / (a2 - a1);
	},

	// https://en.wikipedia.org/wiki/Linear_interpolation

	lerp: function lerp(x, y, t) {

		return (1 - t) * x + t * y;
	},

	// http://en.wikipedia.org/wiki/Smoothstep

	smoothstep: function smoothstep(x, min, max) {

		if (x <= min) return 0;
		if (x >= max) return 1;

		x = (x - min) / (max - min);

		return x * x * (3 - 2 * x);
	},

	smootherstep: function smootherstep(x, min, max) {

		if (x <= min) return 0;
		if (x >= max) return 1;

		x = (x - min) / (max - min);

		return x * x * x * (x * (x * 6 - 15) + 10);
	},

	random16: function random16() {

		console.warn('THREE.Math.random16() has been deprecated. Use Math.random() instead.');
		return Math.random();
	},

	// Random integer from <low, high> interval

	randInt: function randInt(low, high) {

		return low + Math.floor(Math.random() * (high - low + 1));
	},

	// Random float from <low, high> interval

	randFloat: function randFloat(low, high) {

		return low + Math.random() * (high - low);
	},

	// Random float from <-range/2, range/2> interval

	randFloatSpread: function randFloatSpread(range) {

		return range * (0.5 - Math.random());
	},

	degToRad: function degToRad(degrees) {

		return degrees * _Math.DEG2RAD;
	},

	radToDeg: function radToDeg(radians) {

		return radians * _Math.RAD2DEG;
	},

	isPowerOfTwo: function isPowerOfTwo(value) {

		return (value & value - 1) === 0 && value !== 0;
	},

	nearestPowerOfTwo: function nearestPowerOfTwo(value) {

		return Math.pow(2, Math.round(Math.log(value) / Math.LN2));
	},

	nextPowerOfTwo: function nextPowerOfTwo(value) {

		value--;
		value |= value >> 1;
		value |= value >> 2;
		value |= value >> 4;
		value |= value >> 8;
		value |= value >> 16;
		value++;

		return value;
	}

};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author supereggbert / http://www.paulbrunt.co.uk/
 * @author philogb / http://blog.thejit.org/
 * @author jordi_ros / http://plattsoft.com
 * @author D1plo1d / http://github.com/D1plo1d
 * @author alteredq / http://alteredqualia.com/
 * @author mikael emtinger / http://gomo.se/
 * @author timknip / http://www.floorplanner.com/
 * @author bhouston / http://clara.io
 * @author WestLangley / http://github.com/WestLangley
 */

function Matrix4() {

	this.elements = new Float32Array([1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]);

	if (arguments.length > 0) {

		console.error('THREE.Matrix4: the constructor no longer reads arguments. use .set() instead.');
	}
}

Matrix4.prototype = {

	constructor: Matrix4,

	isMatrix4: true,

	set: function set(n11, n12, n13, n14, n21, n22, n23, n24, n31, n32, n33, n34, n41, n42, n43, n44) {

		var te = this.elements;

		te[0] = n11;te[4] = n12;te[8] = n13;te[12] = n14;
		te[1] = n21;te[5] = n22;te[9] = n23;te[13] = n24;
		te[2] = n31;te[6] = n32;te[10] = n33;te[14] = n34;
		te[3] = n41;te[7] = n42;te[11] = n43;te[15] = n44;

		return this;
	},

	identity: function identity() {

		this.set(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);

		return this;
	},

	clone: function clone() {

		return new Matrix4().fromArray(this.elements);
	},

	copy: function copy(m) {

		this.elements.set(m.elements);

		return this;
	},

	copyPosition: function copyPosition(m) {

		var te = this.elements;
		var me = m.elements;

		te[12] = me[12];
		te[13] = me[13];
		te[14] = me[14];

		return this;
	},

	extractBasis: function extractBasis(xAxis, yAxis, zAxis) {

		xAxis.setFromMatrixColumn(this, 0);
		yAxis.setFromMatrixColumn(this, 1);
		zAxis.setFromMatrixColumn(this, 2);

		return this;
	},

	makeBasis: function makeBasis(xAxis, yAxis, zAxis) {

		this.set(xAxis.x, yAxis.x, zAxis.x, 0, xAxis.y, yAxis.y, zAxis.y, 0, xAxis.z, yAxis.z, zAxis.z, 0, 0, 0, 0, 1);

		return this;
	},

	extractRotation: function () {

		var v1;

		return function extractRotation(m) {

			if (v1 === undefined) v1 = new Vector3();

			var te = this.elements;
			var me = m.elements;

			var scaleX = 1 / v1.setFromMatrixColumn(m, 0).length();
			var scaleY = 1 / v1.setFromMatrixColumn(m, 1).length();
			var scaleZ = 1 / v1.setFromMatrixColumn(m, 2).length();

			te[0] = me[0] * scaleX;
			te[1] = me[1] * scaleX;
			te[2] = me[2] * scaleX;

			te[4] = me[4] * scaleY;
			te[5] = me[5] * scaleY;
			te[6] = me[6] * scaleY;

			te[8] = me[8] * scaleZ;
			te[9] = me[9] * scaleZ;
			te[10] = me[10] * scaleZ;

			return this;
		};
	}(),

	makeRotationFromEuler: function makeRotationFromEuler(euler) {

		if ((euler && euler.isEuler) === false) {

			console.error('THREE.Matrix: .makeRotationFromEuler() now expects a Euler rotation rather than a Vector3 and order.');
		}

		var te = this.elements;

		var x = euler.x,
		    y = euler.y,
		    z = euler.z;
		var a = Math.cos(x),
		    b = Math.sin(x);
		var c = Math.cos(y),
		    d = Math.sin(y);
		var e = Math.cos(z),
		    f = Math.sin(z);

		if (euler.order === 'XYZ') {

			var ae = a * e,
			    af = a * f,
			    be = b * e,
			    bf = b * f;

			te[0] = c * e;
			te[4] = -c * f;
			te[8] = d;

			te[1] = af + be * d;
			te[5] = ae - bf * d;
			te[9] = -b * c;

			te[2] = bf - ae * d;
			te[6] = be + af * d;
			te[10] = a * c;
		} else if (euler.order === 'YXZ') {

			var ce = c * e,
			    cf = c * f,
			    de = d * e,
			    df = d * f;

			te[0] = ce + df * b;
			te[4] = de * b - cf;
			te[8] = a * d;

			te[1] = a * f;
			te[5] = a * e;
			te[9] = -b;

			te[2] = cf * b - de;
			te[6] = df + ce * b;
			te[10] = a * c;
		} else if (euler.order === 'ZXY') {

			var ce = c * e,
			    cf = c * f,
			    de = d * e,
			    df = d * f;

			te[0] = ce - df * b;
			te[4] = -a * f;
			te[8] = de + cf * b;

			te[1] = cf + de * b;
			te[5] = a * e;
			te[9] = df - ce * b;

			te[2] = -a * d;
			te[6] = b;
			te[10] = a * c;
		} else if (euler.order === 'ZYX') {

			var ae = a * e,
			    af = a * f,
			    be = b * e,
			    bf = b * f;

			te[0] = c * e;
			te[4] = be * d - af;
			te[8] = ae * d + bf;

			te[1] = c * f;
			te[5] = bf * d + ae;
			te[9] = af * d - be;

			te[2] = -d;
			te[6] = b * c;
			te[10] = a * c;
		} else if (euler.order === 'YZX') {

			var ac = a * c,
			    ad = a * d,
			    bc = b * c,
			    bd = b * d;

			te[0] = c * e;
			te[4] = bd - ac * f;
			te[8] = bc * f + ad;

			te[1] = f;
			te[5] = a * e;
			te[9] = -b * e;

			te[2] = -d * e;
			te[6] = ad * f + bc;
			te[10] = ac - bd * f;
		} else if (euler.order === 'XZY') {

			var ac = a * c,
			    ad = a * d,
			    bc = b * c,
			    bd = b * d;

			te[0] = c * e;
			te[4] = -f;
			te[8] = d * e;

			te[1] = ac * f + bd;
			te[5] = a * e;
			te[9] = ad * f - bc;

			te[2] = bc * f - ad;
			te[6] = b * e;
			te[10] = bd * f + ac;
		}

		// last column
		te[3] = 0;
		te[7] = 0;
		te[11] = 0;

		// bottom row
		te[12] = 0;
		te[13] = 0;
		te[14] = 0;
		te[15] = 1;

		return this;
	},

	makeRotationFromQuaternion: function makeRotationFromQuaternion(q) {

		var te = this.elements;

		var x = q.x,
		    y = q.y,
		    z = q.z,
		    w = q.w;
		var x2 = x + x,
		    y2 = y + y,
		    z2 = z + z;
		var xx = x * x2,
		    xy = x * y2,
		    xz = x * z2;
		var yy = y * y2,
		    yz = y * z2,
		    zz = z * z2;
		var wx = w * x2,
		    wy = w * y2,
		    wz = w * z2;

		te[0] = 1 - (yy + zz);
		te[4] = xy - wz;
		te[8] = xz + wy;

		te[1] = xy + wz;
		te[5] = 1 - (xx + zz);
		te[9] = yz - wx;

		te[2] = xz - wy;
		te[6] = yz + wx;
		te[10] = 1 - (xx + yy);

		// last column
		te[3] = 0;
		te[7] = 0;
		te[11] = 0;

		// bottom row
		te[12] = 0;
		te[13] = 0;
		te[14] = 0;
		te[15] = 1;

		return this;
	},

	lookAt: function () {

		var x, y, z;

		return function lookAt(eye, target, up) {

			if (x === undefined) {

				x = new Vector3();
				y = new Vector3();
				z = new Vector3();
			}

			var te = this.elements;

			z.subVectors(eye, target).normalize();

			if (z.lengthSq() === 0) {

				z.z = 1;
			}

			x.crossVectors(up, z).normalize();

			if (x.lengthSq() === 0) {

				z.z += 0.0001;
				x.crossVectors(up, z).normalize();
			}

			y.crossVectors(z, x);

			te[0] = x.x;te[4] = y.x;te[8] = z.x;
			te[1] = x.y;te[5] = y.y;te[9] = z.y;
			te[2] = x.z;te[6] = y.z;te[10] = z.z;

			return this;
		};
	}(),

	multiply: function multiply(m, n) {

		if (n !== undefined) {

			console.warn('THREE.Matrix4: .multiply() now only accepts one argument. Use .multiplyMatrices( a, b ) instead.');
			return this.multiplyMatrices(m, n);
		}

		return this.multiplyMatrices(this, m);
	},

	premultiply: function premultiply(m) {

		return this.multiplyMatrices(m, this);
	},

	multiplyMatrices: function multiplyMatrices(a, b) {

		var ae = a.elements;
		var be = b.elements;
		var te = this.elements;

		var a11 = ae[0],
		    a12 = ae[4],
		    a13 = ae[8],
		    a14 = ae[12];
		var a21 = ae[1],
		    a22 = ae[5],
		    a23 = ae[9],
		    a24 = ae[13];
		var a31 = ae[2],
		    a32 = ae[6],
		    a33 = ae[10],
		    a34 = ae[14];
		var a41 = ae[3],
		    a42 = ae[7],
		    a43 = ae[11],
		    a44 = ae[15];

		var b11 = be[0],
		    b12 = be[4],
		    b13 = be[8],
		    b14 = be[12];
		var b21 = be[1],
		    b22 = be[5],
		    b23 = be[9],
		    b24 = be[13];
		var b31 = be[2],
		    b32 = be[6],
		    b33 = be[10],
		    b34 = be[14];
		var b41 = be[3],
		    b42 = be[7],
		    b43 = be[11],
		    b44 = be[15];

		te[0] = a11 * b11 + a12 * b21 + a13 * b31 + a14 * b41;
		te[4] = a11 * b12 + a12 * b22 + a13 * b32 + a14 * b42;
		te[8] = a11 * b13 + a12 * b23 + a13 * b33 + a14 * b43;
		te[12] = a11 * b14 + a12 * b24 + a13 * b34 + a14 * b44;

		te[1] = a21 * b11 + a22 * b21 + a23 * b31 + a24 * b41;
		te[5] = a21 * b12 + a22 * b22 + a23 * b32 + a24 * b42;
		te[9] = a21 * b13 + a22 * b23 + a23 * b33 + a24 * b43;
		te[13] = a21 * b14 + a22 * b24 + a23 * b34 + a24 * b44;

		te[2] = a31 * b11 + a32 * b21 + a33 * b31 + a34 * b41;
		te[6] = a31 * b12 + a32 * b22 + a33 * b32 + a34 * b42;
		te[10] = a31 * b13 + a32 * b23 + a33 * b33 + a34 * b43;
		te[14] = a31 * b14 + a32 * b24 + a33 * b34 + a34 * b44;

		te[3] = a41 * b11 + a42 * b21 + a43 * b31 + a44 * b41;
		te[7] = a41 * b12 + a42 * b22 + a43 * b32 + a44 * b42;
		te[11] = a41 * b13 + a42 * b23 + a43 * b33 + a44 * b43;
		te[15] = a41 * b14 + a42 * b24 + a43 * b34 + a44 * b44;

		return this;
	},

	multiplyToArray: function multiplyToArray(a, b, r) {

		var te = this.elements;

		this.multiplyMatrices(a, b);

		r[0] = te[0];r[1] = te[1];r[2] = te[2];r[3] = te[3];
		r[4] = te[4];r[5] = te[5];r[6] = te[6];r[7] = te[7];
		r[8] = te[8];r[9] = te[9];r[10] = te[10];r[11] = te[11];
		r[12] = te[12];r[13] = te[13];r[14] = te[14];r[15] = te[15];

		return this;
	},

	multiplyScalar: function multiplyScalar(s) {

		var te = this.elements;

		te[0] *= s;te[4] *= s;te[8] *= s;te[12] *= s;
		te[1] *= s;te[5] *= s;te[9] *= s;te[13] *= s;
		te[2] *= s;te[6] *= s;te[10] *= s;te[14] *= s;
		te[3] *= s;te[7] *= s;te[11] *= s;te[15] *= s;

		return this;
	},

	applyToVector3Array: function () {

		var v1;

		return function applyToVector3Array(array, offset, length) {

			if (v1 === undefined) v1 = new Vector3();
			if (offset === undefined) offset = 0;
			if (length === undefined) length = array.length;

			for (var i = 0, j = offset; i < length; i += 3, j += 3) {

				v1.fromArray(array, j);
				v1.applyMatrix4(this);
				v1.toArray(array, j);
			}

			return array;
		};
	}(),

	applyToBuffer: function () {

		var v1;

		return function applyToBuffer(buffer, offset, length) {

			if (v1 === undefined) v1 = new Vector3();
			if (offset === undefined) offset = 0;
			if (length === undefined) length = buffer.length / buffer.itemSize;

			for (var i = 0, j = offset; i < length; i++, j++) {

				v1.x = buffer.getX(j);
				v1.y = buffer.getY(j);
				v1.z = buffer.getZ(j);

				v1.applyMatrix4(this);

				buffer.setXYZ(j, v1.x, v1.y, v1.z);
			}

			return buffer;
		};
	}(),

	determinant: function determinant() {

		var te = this.elements;

		var n11 = te[0],
		    n12 = te[4],
		    n13 = te[8],
		    n14 = te[12];
		var n21 = te[1],
		    n22 = te[5],
		    n23 = te[9],
		    n24 = te[13];
		var n31 = te[2],
		    n32 = te[6],
		    n33 = te[10],
		    n34 = te[14];
		var n41 = te[3],
		    n42 = te[7],
		    n43 = te[11],
		    n44 = te[15];

		//TODO: make this more efficient
		//( based on http://www.euclideanspace.com/maths/algebra/matrix/functions/inverse/fourD/index.htm )

		return n41 * (+n14 * n23 * n32 - n13 * n24 * n32 - n14 * n22 * n33 + n12 * n24 * n33 + n13 * n22 * n34 - n12 * n23 * n34) + n42 * (+n11 * n23 * n34 - n11 * n24 * n33 + n14 * n21 * n33 - n13 * n21 * n34 + n13 * n24 * n31 - n14 * n23 * n31) + n43 * (+n11 * n24 * n32 - n11 * n22 * n34 - n14 * n21 * n32 + n12 * n21 * n34 + n14 * n22 * n31 - n12 * n24 * n31) + n44 * (-n13 * n22 * n31 - n11 * n23 * n32 + n11 * n22 * n33 + n13 * n21 * n32 - n12 * n21 * n33 + n12 * n23 * n31);
	},

	transpose: function transpose() {

		var te = this.elements;
		var tmp;

		tmp = te[1];te[1] = te[4];te[4] = tmp;
		tmp = te[2];te[2] = te[8];te[8] = tmp;
		tmp = te[6];te[6] = te[9];te[9] = tmp;

		tmp = te[3];te[3] = te[12];te[12] = tmp;
		tmp = te[7];te[7] = te[13];te[13] = tmp;
		tmp = te[11];te[11] = te[14];te[14] = tmp;

		return this;
	},

	flattenToArrayOffset: function flattenToArrayOffset(array, offset) {

		console.warn("THREE.Matrix3: .flattenToArrayOffset is deprecated " + "- just use .toArray instead.");

		return this.toArray(array, offset);
	},

	getPosition: function () {

		var v1;

		return function getPosition() {

			if (v1 === undefined) v1 = new Vector3();
			console.warn('THREE.Matrix4: .getPosition() has been removed. Use Vector3.setFromMatrixPosition( matrix ) instead.');

			return v1.setFromMatrixColumn(this, 3);
		};
	}(),

	setPosition: function setPosition(v) {

		var te = this.elements;

		te[12] = v.x;
		te[13] = v.y;
		te[14] = v.z;

		return this;
	},

	getInverse: function getInverse(m, throwOnDegenerate) {

		// based on http://www.euclideanspace.com/maths/algebra/matrix/functions/inverse/fourD/index.htm
		var te = this.elements,
		    me = m.elements,
		    n11 = me[0],
		    n21 = me[1],
		    n31 = me[2],
		    n41 = me[3],
		    n12 = me[4],
		    n22 = me[5],
		    n32 = me[6],
		    n42 = me[7],
		    n13 = me[8],
		    n23 = me[9],
		    n33 = me[10],
		    n43 = me[11],
		    n14 = me[12],
		    n24 = me[13],
		    n34 = me[14],
		    n44 = me[15],
		    t11 = n23 * n34 * n42 - n24 * n33 * n42 + n24 * n32 * n43 - n22 * n34 * n43 - n23 * n32 * n44 + n22 * n33 * n44,
		    t12 = n14 * n33 * n42 - n13 * n34 * n42 - n14 * n32 * n43 + n12 * n34 * n43 + n13 * n32 * n44 - n12 * n33 * n44,
		    t13 = n13 * n24 * n42 - n14 * n23 * n42 + n14 * n22 * n43 - n12 * n24 * n43 - n13 * n22 * n44 + n12 * n23 * n44,
		    t14 = n14 * n23 * n32 - n13 * n24 * n32 - n14 * n22 * n33 + n12 * n24 * n33 + n13 * n22 * n34 - n12 * n23 * n34;

		var det = n11 * t11 + n21 * t12 + n31 * t13 + n41 * t14;

		if (det === 0) {

			var msg = "THREE.Matrix4.getInverse(): can't invert matrix, determinant is 0";

			if (throwOnDegenerate === true) {

				throw new Error(msg);
			} else {

				console.warn(msg);
			}

			return this.identity();
		}

		var detInv = 1 / det;

		te[0] = t11 * detInv;
		te[1] = (n24 * n33 * n41 - n23 * n34 * n41 - n24 * n31 * n43 + n21 * n34 * n43 + n23 * n31 * n44 - n21 * n33 * n44) * detInv;
		te[2] = (n22 * n34 * n41 - n24 * n32 * n41 + n24 * n31 * n42 - n21 * n34 * n42 - n22 * n31 * n44 + n21 * n32 * n44) * detInv;
		te[3] = (n23 * n32 * n41 - n22 * n33 * n41 - n23 * n31 * n42 + n21 * n33 * n42 + n22 * n31 * n43 - n21 * n32 * n43) * detInv;

		te[4] = t12 * detInv;
		te[5] = (n13 * n34 * n41 - n14 * n33 * n41 + n14 * n31 * n43 - n11 * n34 * n43 - n13 * n31 * n44 + n11 * n33 * n44) * detInv;
		te[6] = (n14 * n32 * n41 - n12 * n34 * n41 - n14 * n31 * n42 + n11 * n34 * n42 + n12 * n31 * n44 - n11 * n32 * n44) * detInv;
		te[7] = (n12 * n33 * n41 - n13 * n32 * n41 + n13 * n31 * n42 - n11 * n33 * n42 - n12 * n31 * n43 + n11 * n32 * n43) * detInv;

		te[8] = t13 * detInv;
		te[9] = (n14 * n23 * n41 - n13 * n24 * n41 - n14 * n21 * n43 + n11 * n24 * n43 + n13 * n21 * n44 - n11 * n23 * n44) * detInv;
		te[10] = (n12 * n24 * n41 - n14 * n22 * n41 + n14 * n21 * n42 - n11 * n24 * n42 - n12 * n21 * n44 + n11 * n22 * n44) * detInv;
		te[11] = (n13 * n22 * n41 - n12 * n23 * n41 - n13 * n21 * n42 + n11 * n23 * n42 + n12 * n21 * n43 - n11 * n22 * n43) * detInv;

		te[12] = t14 * detInv;
		te[13] = (n13 * n24 * n31 - n14 * n23 * n31 + n14 * n21 * n33 - n11 * n24 * n33 - n13 * n21 * n34 + n11 * n23 * n34) * detInv;
		te[14] = (n14 * n22 * n31 - n12 * n24 * n31 - n14 * n21 * n32 + n11 * n24 * n32 + n12 * n21 * n34 - n11 * n22 * n34) * detInv;
		te[15] = (n12 * n23 * n31 - n13 * n22 * n31 + n13 * n21 * n32 - n11 * n23 * n32 - n12 * n21 * n33 + n11 * n22 * n33) * detInv;

		return this;
	},

	scale: function scale(v) {

		var te = this.elements;
		var x = v.x,
		    y = v.y,
		    z = v.z;

		te[0] *= x;te[4] *= y;te[8] *= z;
		te[1] *= x;te[5] *= y;te[9] *= z;
		te[2] *= x;te[6] *= y;te[10] *= z;
		te[3] *= x;te[7] *= y;te[11] *= z;

		return this;
	},

	getMaxScaleOnAxis: function getMaxScaleOnAxis() {

		var te = this.elements;

		var scaleXSq = te[0] * te[0] + te[1] * te[1] + te[2] * te[2];
		var scaleYSq = te[4] * te[4] + te[5] * te[5] + te[6] * te[6];
		var scaleZSq = te[8] * te[8] + te[9] * te[9] + te[10] * te[10];

		return Math.sqrt(Math.max(scaleXSq, scaleYSq, scaleZSq));
	},

	makeTranslation: function makeTranslation(x, y, z) {

		this.set(1, 0, 0, x, 0, 1, 0, y, 0, 0, 1, z, 0, 0, 0, 1);

		return this;
	},

	makeRotationX: function makeRotationX(theta) {

		var c = Math.cos(theta),
		    s = Math.sin(theta);

		this.set(1, 0, 0, 0, 0, c, -s, 0, 0, s, c, 0, 0, 0, 0, 1);

		return this;
	},

	makeRotationY: function makeRotationY(theta) {

		var c = Math.cos(theta),
		    s = Math.sin(theta);

		this.set(c, 0, s, 0, 0, 1, 0, 0, -s, 0, c, 0, 0, 0, 0, 1);

		return this;
	},

	makeRotationZ: function makeRotationZ(theta) {

		var c = Math.cos(theta),
		    s = Math.sin(theta);

		this.set(c, -s, 0, 0, s, c, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);

		return this;
	},

	makeRotationAxis: function makeRotationAxis(axis, angle) {

		// Based on http://www.gamedev.net/reference/articles/article1199.asp

		var c = Math.cos(angle);
		var s = Math.sin(angle);
		var t = 1 - c;
		var x = axis.x,
		    y = axis.y,
		    z = axis.z;
		var tx = t * x,
		    ty = t * y;

		this.set(tx * x + c, tx * y - s * z, tx * z + s * y, 0, tx * y + s * z, ty * y + c, ty * z - s * x, 0, tx * z - s * y, ty * z + s * x, t * z * z + c, 0, 0, 0, 0, 1);

		return this;
	},

	makeScale: function makeScale(x, y, z) {

		this.set(x, 0, 0, 0, 0, y, 0, 0, 0, 0, z, 0, 0, 0, 0, 1);

		return this;
	},

	compose: function compose(position, quaternion, scale) {

		this.makeRotationFromQuaternion(quaternion);
		this.scale(scale);
		this.setPosition(position);

		return this;
	},

	decompose: function () {

		var vector, matrix;

		return function decompose(position, quaternion, scale) {

			if (vector === undefined) {

				vector = new Vector3();
				matrix = new Matrix4();
			}

			var te = this.elements;

			var sx = vector.set(te[0], te[1], te[2]).length();
			var sy = vector.set(te[4], te[5], te[6]).length();
			var sz = vector.set(te[8], te[9], te[10]).length();

			// if determine is negative, we need to invert one scale
			var det = this.determinant();
			if (det < 0) {

				sx = -sx;
			}

			position.x = te[12];
			position.y = te[13];
			position.z = te[14];

			// scale the rotation part

			matrix.elements.set(this.elements); // at this point matrix is incomplete so we can't use .copy()

			var invSX = 1 / sx;
			var invSY = 1 / sy;
			var invSZ = 1 / sz;

			matrix.elements[0] *= invSX;
			matrix.elements[1] *= invSX;
			matrix.elements[2] *= invSX;

			matrix.elements[4] *= invSY;
			matrix.elements[5] *= invSY;
			matrix.elements[6] *= invSY;

			matrix.elements[8] *= invSZ;
			matrix.elements[9] *= invSZ;
			matrix.elements[10] *= invSZ;

			quaternion.setFromRotationMatrix(matrix);

			scale.x = sx;
			scale.y = sy;
			scale.z = sz;

			return this;
		};
	}(),

	makeFrustum: function makeFrustum(left, right, bottom, top, near, far) {

		var te = this.elements;
		var x = 2 * near / (right - left);
		var y = 2 * near / (top - bottom);

		var a = (right + left) / (right - left);
		var b = (top + bottom) / (top - bottom);
		var c = -(far + near) / (far - near);
		var d = -2 * far * near / (far - near);

		te[0] = x;te[4] = 0;te[8] = a;te[12] = 0;
		te[1] = 0;te[5] = y;te[9] = b;te[13] = 0;
		te[2] = 0;te[6] = 0;te[10] = c;te[14] = d;
		te[3] = 0;te[7] = 0;te[11] = -1;te[15] = 0;

		return this;
	},

	makePerspective: function makePerspective(fov, aspect, near, far) {

		var ymax = near * Math.tan(_Math.DEG2RAD * fov * 0.5);
		var ymin = -ymax;
		var xmin = ymin * aspect;
		var xmax = ymax * aspect;

		return this.makeFrustum(xmin, xmax, ymin, ymax, near, far);
	},

	makeOrthographic: function makeOrthographic(left, right, top, bottom, near, far) {

		var te = this.elements;
		var w = 1.0 / (right - left);
		var h = 1.0 / (top - bottom);
		var p = 1.0 / (far - near);

		var x = (right + left) * w;
		var y = (top + bottom) * h;
		var z = (far + near) * p;

		te[0] = 2 * w;te[4] = 0;te[8] = 0;te[12] = -x;
		te[1] = 0;te[5] = 2 * h;te[9] = 0;te[13] = -y;
		te[2] = 0;te[6] = 0;te[10] = -2 * p;te[14] = -z;
		te[3] = 0;te[7] = 0;te[11] = 0;te[15] = 1;

		return this;
	},

	equals: function equals(matrix) {

		var te = this.elements;
		var me = matrix.elements;

		for (var i = 0; i < 16; i++) {

			if (te[i] !== me[i]) return false;
		}

		return true;
	},

	fromArray: function fromArray(array, offset) {

		if (offset === undefined) offset = 0;

		for (var i = 0; i < 16; i++) {

			this.elements[i] = array[i + offset];
		}

		return this;
	},

	toArray: function toArray(array, offset) {

		if (array === undefined) array = [];
		if (offset === undefined) offset = 0;

		var te = this.elements;

		array[offset] = te[0];
		array[offset + 1] = te[1];
		array[offset + 2] = te[2];
		array[offset + 3] = te[3];

		array[offset + 4] = te[4];
		array[offset + 5] = te[5];
		array[offset + 6] = te[6];
		array[offset + 7] = te[7];

		array[offset + 8] = te[8];
		array[offset + 9] = te[9];
		array[offset + 10] = te[10];
		array[offset + 11] = te[11];

		array[offset + 12] = te[12];
		array[offset + 13] = te[13];
		array[offset + 14] = te[14];
		array[offset + 15] = te[15];

		return array;
	}

};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author *kile / http://kile.stravaganza.org/
 * @author philogb / http://blog.thejit.org/
 * @author mikael emtinger / http://gomo.se/
 * @author egraether / http://egraether.com/
 * @author WestLangley / http://github.com/WestLangley
 */

function Vector3(x, y, z) {

	this.x = x || 0;
	this.y = y || 0;
	this.z = z || 0;
}

Vector3.prototype = {

	constructor: Vector3,

	isVector3: true,

	set: function set(x, y, z) {

		this.x = x;
		this.y = y;
		this.z = z;

		return this;
	},

	setScalar: function setScalar(scalar) {

		this.x = scalar;
		this.y = scalar;
		this.z = scalar;

		return this;
	},

	setX: function setX(x) {

		this.x = x;

		return this;
	},

	setY: function setY(y) {

		this.y = y;

		return this;
	},

	setZ: function setZ(z) {

		this.z = z;

		return this;
	},

	setComponent: function setComponent(index, value) {

		switch (index) {

			case 0:
				this.x = value;break;
			case 1:
				this.y = value;break;
			case 2:
				this.z = value;break;
			default:
				throw new Error('index is out of range: ' + index);

		}

		return this;
	},

	getComponent: function getComponent(index) {

		switch (index) {

			case 0:
				return this.x;
			case 1:
				return this.y;
			case 2:
				return this.z;
			default:
				throw new Error('index is out of range: ' + index);

		}
	},

	clone: function clone() {

		return new this.constructor(this.x, this.y, this.z);
	},

	copy: function copy(v) {

		this.x = v.x;
		this.y = v.y;
		this.z = v.z;

		return this;
	},

	add: function add(v, w) {

		if (w !== undefined) {

			console.warn('THREE.Vector3: .add() now only accepts one argument. Use .addVectors( a, b ) instead.');
			return this.addVectors(v, w);
		}

		this.x += v.x;
		this.y += v.y;
		this.z += v.z;

		return this;
	},

	addScalar: function addScalar(s) {

		this.x += s;
		this.y += s;
		this.z += s;

		return this;
	},

	addVectors: function addVectors(a, b) {

		this.x = a.x + b.x;
		this.y = a.y + b.y;
		this.z = a.z + b.z;

		return this;
	},

	addScaledVector: function addScaledVector(v, s) {

		this.x += v.x * s;
		this.y += v.y * s;
		this.z += v.z * s;

		return this;
	},

	sub: function sub(v, w) {

		if (w !== undefined) {

			console.warn('THREE.Vector3: .sub() now only accepts one argument. Use .subVectors( a, b ) instead.');
			return this.subVectors(v, w);
		}

		this.x -= v.x;
		this.y -= v.y;
		this.z -= v.z;

		return this;
	},

	subScalar: function subScalar(s) {

		this.x -= s;
		this.y -= s;
		this.z -= s;

		return this;
	},

	subVectors: function subVectors(a, b) {

		this.x = a.x - b.x;
		this.y = a.y - b.y;
		this.z = a.z - b.z;

		return this;
	},

	multiply: function multiply(v, w) {

		if (w !== undefined) {

			console.warn('THREE.Vector3: .multiply() now only accepts one argument. Use .multiplyVectors( a, b ) instead.');
			return this.multiplyVectors(v, w);
		}

		this.x *= v.x;
		this.y *= v.y;
		this.z *= v.z;

		return this;
	},

	multiplyScalar: function multiplyScalar(scalar) {

		if (isFinite(scalar)) {

			this.x *= scalar;
			this.y *= scalar;
			this.z *= scalar;
		} else {

			this.x = 0;
			this.y = 0;
			this.z = 0;
		}

		return this;
	},

	multiplyVectors: function multiplyVectors(a, b) {

		this.x = a.x * b.x;
		this.y = a.y * b.y;
		this.z = a.z * b.z;

		return this;
	},

	applyEuler: function () {

		var quaternion;

		return function applyEuler(euler) {

			if ((euler && euler.isEuler) === false) {

				console.error('THREE.Vector3: .applyEuler() now expects an Euler rotation rather than a Vector3 and order.');
			}

			if (quaternion === undefined) quaternion = new Quaternion();

			return this.applyQuaternion(quaternion.setFromEuler(euler));
		};
	}(),

	applyAxisAngle: function () {

		var quaternion;

		return function applyAxisAngle(axis, angle) {

			if (quaternion === undefined) quaternion = new Quaternion();

			return this.applyQuaternion(quaternion.setFromAxisAngle(axis, angle));
		};
	}(),

	applyMatrix3: function applyMatrix3(m) {

		var x = this.x,
		    y = this.y,
		    z = this.z;
		var e = m.elements;

		this.x = e[0] * x + e[3] * y + e[6] * z;
		this.y = e[1] * x + e[4] * y + e[7] * z;
		this.z = e[2] * x + e[5] * y + e[8] * z;

		return this;
	},

	applyMatrix4: function applyMatrix4(m) {

		// input: THREE.Matrix4 affine matrix

		var x = this.x,
		    y = this.y,
		    z = this.z;
		var e = m.elements;

		this.x = e[0] * x + e[4] * y + e[8] * z + e[12];
		this.y = e[1] * x + e[5] * y + e[9] * z + e[13];
		this.z = e[2] * x + e[6] * y + e[10] * z + e[14];

		return this;
	},

	applyProjection: function applyProjection(m) {

		// input: THREE.Matrix4 projection matrix

		var x = this.x,
		    y = this.y,
		    z = this.z;
		var e = m.elements;
		var d = 1 / (e[3] * x + e[7] * y + e[11] * z + e[15]); // perspective divide

		this.x = (e[0] * x + e[4] * y + e[8] * z + e[12]) * d;
		this.y = (e[1] * x + e[5] * y + e[9] * z + e[13]) * d;
		this.z = (e[2] * x + e[6] * y + e[10] * z + e[14]) * d;

		return this;
	},

	applyQuaternion: function applyQuaternion(q) {

		var x = this.x,
		    y = this.y,
		    z = this.z;
		var qx = q.x,
		    qy = q.y,
		    qz = q.z,
		    qw = q.w;

		// calculate quat * vector

		var ix = qw * x + qy * z - qz * y;
		var iy = qw * y + qz * x - qx * z;
		var iz = qw * z + qx * y - qy * x;
		var iw = -qx * x - qy * y - qz * z;

		// calculate result * inverse quat

		this.x = ix * qw + iw * -qx + iy * -qz - iz * -qy;
		this.y = iy * qw + iw * -qy + iz * -qx - ix * -qz;
		this.z = iz * qw + iw * -qz + ix * -qy - iy * -qx;

		return this;
	},

	project: function () {

		var matrix;

		return function project(camera) {

			if (matrix === undefined) matrix = new Matrix4();

			matrix.multiplyMatrices(camera.projectionMatrix, matrix.getInverse(camera.matrixWorld));
			return this.applyProjection(matrix);
		};
	}(),

	unproject: function () {

		var matrix;

		return function unproject(camera) {

			if (matrix === undefined) matrix = new Matrix4();

			matrix.multiplyMatrices(camera.matrixWorld, matrix.getInverse(camera.projectionMatrix));
			return this.applyProjection(matrix);
		};
	}(),

	transformDirection: function transformDirection(m) {

		// input: THREE.Matrix4 affine matrix
		// vector interpreted as a direction

		var x = this.x,
		    y = this.y,
		    z = this.z;
		var e = m.elements;

		this.x = e[0] * x + e[4] * y + e[8] * z;
		this.y = e[1] * x + e[5] * y + e[9] * z;
		this.z = e[2] * x + e[6] * y + e[10] * z;

		return this.normalize();
	},

	divide: function divide(v) {

		this.x /= v.x;
		this.y /= v.y;
		this.z /= v.z;

		return this;
	},

	divideScalar: function divideScalar(scalar) {

		return this.multiplyScalar(1 / scalar);
	},

	min: function min(v) {

		this.x = Math.min(this.x, v.x);
		this.y = Math.min(this.y, v.y);
		this.z = Math.min(this.z, v.z);

		return this;
	},

	max: function max(v) {

		this.x = Math.max(this.x, v.x);
		this.y = Math.max(this.y, v.y);
		this.z = Math.max(this.z, v.z);

		return this;
	},

	clamp: function clamp(min, max) {

		// This function assumes min < max, if this assumption isn't true it will not operate correctly

		this.x = Math.max(min.x, Math.min(max.x, this.x));
		this.y = Math.max(min.y, Math.min(max.y, this.y));
		this.z = Math.max(min.z, Math.min(max.z, this.z));

		return this;
	},

	clampScalar: function () {

		var min, max;

		return function clampScalar(minVal, maxVal) {

			if (min === undefined) {

				min = new Vector3();
				max = new Vector3();
			}

			min.set(minVal, minVal, minVal);
			max.set(maxVal, maxVal, maxVal);

			return this.clamp(min, max);
		};
	}(),

	clampLength: function clampLength(min, max) {

		var length = this.length();

		return this.multiplyScalar(Math.max(min, Math.min(max, length)) / length);
	},

	floor: function floor() {

		this.x = Math.floor(this.x);
		this.y = Math.floor(this.y);
		this.z = Math.floor(this.z);

		return this;
	},

	ceil: function ceil() {

		this.x = Math.ceil(this.x);
		this.y = Math.ceil(this.y);
		this.z = Math.ceil(this.z);

		return this;
	},

	round: function round() {

		this.x = Math.round(this.x);
		this.y = Math.round(this.y);
		this.z = Math.round(this.z);

		return this;
	},

	roundToZero: function roundToZero() {

		this.x = this.x < 0 ? Math.ceil(this.x) : Math.floor(this.x);
		this.y = this.y < 0 ? Math.ceil(this.y) : Math.floor(this.y);
		this.z = this.z < 0 ? Math.ceil(this.z) : Math.floor(this.z);

		return this;
	},

	negate: function negate() {

		this.x = -this.x;
		this.y = -this.y;
		this.z = -this.z;

		return this;
	},

	dot: function dot(v) {

		return this.x * v.x + this.y * v.y + this.z * v.z;
	},

	lengthSq: function lengthSq() {

		return this.x * this.x + this.y * this.y + this.z * this.z;
	},

	length: function length() {

		return Math.sqrt(this.x * this.x + this.y * this.y + this.z * this.z);
	},

	lengthManhattan: function lengthManhattan() {

		return Math.abs(this.x) + Math.abs(this.y) + Math.abs(this.z);
	},

	normalize: function normalize() {

		return this.divideScalar(this.length());
	},

	setLength: function setLength(length) {

		return this.multiplyScalar(length / this.length());
	},

	lerp: function lerp(v, alpha) {

		this.x += (v.x - this.x) * alpha;
		this.y += (v.y - this.y) * alpha;
		this.z += (v.z - this.z) * alpha;

		return this;
	},

	lerpVectors: function lerpVectors(v1, v2, alpha) {

		return this.subVectors(v2, v1).multiplyScalar(alpha).add(v1);
	},

	cross: function cross(v, w) {

		if (w !== undefined) {

			console.warn('THREE.Vector3: .cross() now only accepts one argument. Use .crossVectors( a, b ) instead.');
			return this.crossVectors(v, w);
		}

		var x = this.x,
		    y = this.y,
		    z = this.z;

		this.x = y * v.z - z * v.y;
		this.y = z * v.x - x * v.z;
		this.z = x * v.y - y * v.x;

		return this;
	},

	crossVectors: function crossVectors(a, b) {

		var ax = a.x,
		    ay = a.y,
		    az = a.z;
		var bx = b.x,
		    by = b.y,
		    bz = b.z;

		this.x = ay * bz - az * by;
		this.y = az * bx - ax * bz;
		this.z = ax * by - ay * bx;

		return this;
	},

	projectOnVector: function projectOnVector(vector) {

		var scalar = vector.dot(this) / vector.lengthSq();

		return this.copy(vector).multiplyScalar(scalar);
	},

	projectOnPlane: function () {

		var v1;

		return function projectOnPlane(planeNormal) {

			if (v1 === undefined) v1 = new Vector3();

			v1.copy(this).projectOnVector(planeNormal);

			return this.sub(v1);
		};
	}(),

	reflect: function () {

		// reflect incident vector off plane orthogonal to normal
		// normal is assumed to have unit length

		var v1;

		return function reflect(normal) {

			if (v1 === undefined) v1 = new Vector3();

			return this.sub(v1.copy(normal).multiplyScalar(2 * this.dot(normal)));
		};
	}(),

	angleTo: function angleTo(v) {

		var theta = this.dot(v) / Math.sqrt(this.lengthSq() * v.lengthSq());

		// clamp, to handle numerical problems

		return Math.acos(_Math.clamp(theta, -1, 1));
	},

	distanceTo: function distanceTo(v) {

		return Math.sqrt(this.distanceToSquared(v));
	},

	distanceToSquared: function distanceToSquared(v) {

		var dx = this.x - v.x,
		    dy = this.y - v.y,
		    dz = this.z - v.z;

		return dx * dx + dy * dy + dz * dz;
	},

	distanceToManhattan: function distanceToManhattan(v) {

		return Math.abs(this.x - v.x) + Math.abs(this.y - v.y) + Math.abs(this.z - v.z);
	},

	setFromSpherical: function setFromSpherical(s) {

		var sinPhiRadius = Math.sin(s.phi) * s.radius;

		this.x = sinPhiRadius * Math.sin(s.theta);
		this.y = Math.cos(s.phi) * s.radius;
		this.z = sinPhiRadius * Math.cos(s.theta);

		return this;
	},

	setFromMatrixPosition: function setFromMatrixPosition(m) {

		return this.setFromMatrixColumn(m, 3);
	},

	setFromMatrixScale: function setFromMatrixScale(m) {

		var sx = this.setFromMatrixColumn(m, 0).length();
		var sy = this.setFromMatrixColumn(m, 1).length();
		var sz = this.setFromMatrixColumn(m, 2).length();

		this.x = sx;
		this.y = sy;
		this.z = sz;

		return this;
	},

	setFromMatrixColumn: function setFromMatrixColumn(m, index) {

		if (typeof m === 'number') {

			console.warn('THREE.Vector3: setFromMatrixColumn now expects ( matrix, index ).');
			var temp = m;
			m = index;
			index = temp;
		}

		return this.fromArray(m.elements, index * 4);
	},

	equals: function equals(v) {

		return v.x === this.x && v.y === this.y && v.z === this.z;
	},

	fromArray: function fromArray(array, offset) {

		if (offset === undefined) offset = 0;

		this.x = array[offset];
		this.y = array[offset + 1];
		this.z = array[offset + 2];

		return this;
	},

	toArray: function toArray(array, offset) {

		if (array === undefined) array = [];
		if (offset === undefined) offset = 0;

		array[offset] = this.x;
		array[offset + 1] = this.y;
		array[offset + 2] = this.z;

		return array;
	},

	fromAttribute: function fromAttribute(attribute, index, offset) {

		if (offset === undefined) offset = 0;

		index = index * attribute.itemSize + offset;

		this.x = attribute.array[index];
		this.y = attribute.array[index + 1];
		this.z = attribute.array[index + 2];

		return this;
	}

};

/**
 * @author mikael emtinger / http://gomo.se/
 * @author alteredq / http://alteredqualia.com/
 * @author WestLangley / http://github.com/WestLangley
 * @author bhouston / http://clara.io
 */

function Quaternion(x, y, z, w) {

	this._x = x || 0;
	this._y = y || 0;
	this._z = z || 0;
	this._w = w !== undefined ? w : 1;
}

Quaternion.prototype = {

	constructor: Quaternion,

	get x() {

		return this._x;
	},

	set x(value) {

		this._x = value;
		this.onChangeCallback();
	},

	get y() {

		return this._y;
	},

	set y(value) {

		this._y = value;
		this.onChangeCallback();
	},

	get z() {

		return this._z;
	},

	set z(value) {

		this._z = value;
		this.onChangeCallback();
	},

	get w() {

		return this._w;
	},

	set w(value) {

		this._w = value;
		this.onChangeCallback();
	},

	set: function set(x, y, z, w) {

		this._x = x;
		this._y = y;
		this._z = z;
		this._w = w;

		this.onChangeCallback();

		return this;
	},

	clone: function clone() {

		return new this.constructor(this._x, this._y, this._z, this._w);
	},

	copy: function copy(quaternion) {

		this._x = quaternion.x;
		this._y = quaternion.y;
		this._z = quaternion.z;
		this._w = quaternion.w;

		this.onChangeCallback();

		return this;
	},

	setFromEuler: function setFromEuler(euler, update) {

		if ((euler && euler.isEuler) === false) {

			throw new Error('THREE.Quaternion: .setFromEuler() now expects an Euler rotation rather than a Vector3 and order.');
		}

		// http://www.mathworks.com/matlabcentral/fileexchange/
		// 	20696-function-to-convert-between-dcm-euler-angles-quaternions-and-euler-vectors/
		//	content/SpinCalc.m

		var c1 = Math.cos(euler._x / 2);
		var c2 = Math.cos(euler._y / 2);
		var c3 = Math.cos(euler._z / 2);
		var s1 = Math.sin(euler._x / 2);
		var s2 = Math.sin(euler._y / 2);
		var s3 = Math.sin(euler._z / 2);

		var order = euler.order;

		if (order === 'XYZ') {

			this._x = s1 * c2 * c3 + c1 * s2 * s3;
			this._y = c1 * s2 * c3 - s1 * c2 * s3;
			this._z = c1 * c2 * s3 + s1 * s2 * c3;
			this._w = c1 * c2 * c3 - s1 * s2 * s3;
		} else if (order === 'YXZ') {

			this._x = s1 * c2 * c3 + c1 * s2 * s3;
			this._y = c1 * s2 * c3 - s1 * c2 * s3;
			this._z = c1 * c2 * s3 - s1 * s2 * c3;
			this._w = c1 * c2 * c3 + s1 * s2 * s3;
		} else if (order === 'ZXY') {

			this._x = s1 * c2 * c3 - c1 * s2 * s3;
			this._y = c1 * s2 * c3 + s1 * c2 * s3;
			this._z = c1 * c2 * s3 + s1 * s2 * c3;
			this._w = c1 * c2 * c3 - s1 * s2 * s3;
		} else if (order === 'ZYX') {

			this._x = s1 * c2 * c3 - c1 * s2 * s3;
			this._y = c1 * s2 * c3 + s1 * c2 * s3;
			this._z = c1 * c2 * s3 - s1 * s2 * c3;
			this._w = c1 * c2 * c3 + s1 * s2 * s3;
		} else if (order === 'YZX') {

			this._x = s1 * c2 * c3 + c1 * s2 * s3;
			this._y = c1 * s2 * c3 + s1 * c2 * s3;
			this._z = c1 * c2 * s3 - s1 * s2 * c3;
			this._w = c1 * c2 * c3 - s1 * s2 * s3;
		} else if (order === 'XZY') {

			this._x = s1 * c2 * c3 - c1 * s2 * s3;
			this._y = c1 * s2 * c3 - s1 * c2 * s3;
			this._z = c1 * c2 * s3 + s1 * s2 * c3;
			this._w = c1 * c2 * c3 + s1 * s2 * s3;
		}

		if (update !== false) this.onChangeCallback();

		return this;
	},

	setFromAxisAngle: function setFromAxisAngle(axis, angle) {

		// http://www.euclideanspace.com/maths/geometry/rotations/conversions/angleToQuaternion/index.htm

		// assumes axis is normalized

		var halfAngle = angle / 2,
		    s = Math.sin(halfAngle);

		this._x = axis.x * s;
		this._y = axis.y * s;
		this._z = axis.z * s;
		this._w = Math.cos(halfAngle);

		this.onChangeCallback();

		return this;
	},

	setFromRotationMatrix: function setFromRotationMatrix(m) {

		// http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm

		// assumes the upper 3x3 of m is a pure rotation matrix (i.e, unscaled)

		var te = m.elements,
		    m11 = te[0],
		    m12 = te[4],
		    m13 = te[8],
		    m21 = te[1],
		    m22 = te[5],
		    m23 = te[9],
		    m31 = te[2],
		    m32 = te[6],
		    m33 = te[10],
		    trace = m11 + m22 + m33,
		    s;

		if (trace > 0) {

			s = 0.5 / Math.sqrt(trace + 1.0);

			this._w = 0.25 / s;
			this._x = (m32 - m23) * s;
			this._y = (m13 - m31) * s;
			this._z = (m21 - m12) * s;
		} else if (m11 > m22 && m11 > m33) {

			s = 2.0 * Math.sqrt(1.0 + m11 - m22 - m33);

			this._w = (m32 - m23) / s;
			this._x = 0.25 * s;
			this._y = (m12 + m21) / s;
			this._z = (m13 + m31) / s;
		} else if (m22 > m33) {

			s = 2.0 * Math.sqrt(1.0 + m22 - m11 - m33);

			this._w = (m13 - m31) / s;
			this._x = (m12 + m21) / s;
			this._y = 0.25 * s;
			this._z = (m23 + m32) / s;
		} else {

			s = 2.0 * Math.sqrt(1.0 + m33 - m11 - m22);

			this._w = (m21 - m12) / s;
			this._x = (m13 + m31) / s;
			this._y = (m23 + m32) / s;
			this._z = 0.25 * s;
		}

		this.onChangeCallback();

		return this;
	},

	setFromUnitVectors: function () {

		// http://lolengine.net/blog/2014/02/24/quaternion-from-two-vectors-final

		// assumes direction vectors vFrom and vTo are normalized

		var v1, r;

		var EPS = 0.000001;

		return function setFromUnitVectors(vFrom, vTo) {

			if (v1 === undefined) v1 = new Vector3();

			r = vFrom.dot(vTo) + 1;

			if (r < EPS) {

				r = 0;

				if (Math.abs(vFrom.x) > Math.abs(vFrom.z)) {

					v1.set(-vFrom.y, vFrom.x, 0);
				} else {

					v1.set(0, -vFrom.z, vFrom.y);
				}
			} else {

				v1.crossVectors(vFrom, vTo);
			}

			this._x = v1.x;
			this._y = v1.y;
			this._z = v1.z;
			this._w = r;

			return this.normalize();
		};
	}(),

	inverse: function inverse() {

		return this.conjugate().normalize();
	},

	conjugate: function conjugate() {

		this._x *= -1;
		this._y *= -1;
		this._z *= -1;

		this.onChangeCallback();

		return this;
	},

	dot: function dot(v) {

		return this._x * v._x + this._y * v._y + this._z * v._z + this._w * v._w;
	},

	lengthSq: function lengthSq() {

		return this._x * this._x + this._y * this._y + this._z * this._z + this._w * this._w;
	},

	length: function length() {

		return Math.sqrt(this._x * this._x + this._y * this._y + this._z * this._z + this._w * this._w);
	},

	normalize: function normalize() {

		var l = this.length();

		if (l === 0) {

			this._x = 0;
			this._y = 0;
			this._z = 0;
			this._w = 1;
		} else {

			l = 1 / l;

			this._x = this._x * l;
			this._y = this._y * l;
			this._z = this._z * l;
			this._w = this._w * l;
		}

		this.onChangeCallback();

		return this;
	},

	multiply: function multiply(q, p) {

		if (p !== undefined) {

			console.warn('THREE.Quaternion: .multiply() now only accepts one argument. Use .multiplyQuaternions( a, b ) instead.');
			return this.multiplyQuaternions(q, p);
		}

		return this.multiplyQuaternions(this, q);
	},

	premultiply: function premultiply(q) {

		return this.multiplyQuaternions(q, this);
	},

	multiplyQuaternions: function multiplyQuaternions(a, b) {

		// from http://www.euclideanspace.com/maths/algebra/realNormedAlgebra/quaternions/code/index.htm

		var qax = a._x,
		    qay = a._y,
		    qaz = a._z,
		    qaw = a._w;
		var qbx = b._x,
		    qby = b._y,
		    qbz = b._z,
		    qbw = b._w;

		this._x = qax * qbw + qaw * qbx + qay * qbz - qaz * qby;
		this._y = qay * qbw + qaw * qby + qaz * qbx - qax * qbz;
		this._z = qaz * qbw + qaw * qbz + qax * qby - qay * qbx;
		this._w = qaw * qbw - qax * qbx - qay * qby - qaz * qbz;

		this.onChangeCallback();

		return this;
	},

	slerp: function slerp(qb, t) {

		if (t === 0) return this;
		if (t === 1) return this.copy(qb);

		var x = this._x,
		    y = this._y,
		    z = this._z,
		    w = this._w;

		// http://www.euclideanspace.com/maths/algebra/realNormedAlgebra/quaternions/slerp/

		var cosHalfTheta = w * qb._w + x * qb._x + y * qb._y + z * qb._z;

		if (cosHalfTheta < 0) {

			this._w = -qb._w;
			this._x = -qb._x;
			this._y = -qb._y;
			this._z = -qb._z;

			cosHalfTheta = -cosHalfTheta;
		} else {

			this.copy(qb);
		}

		if (cosHalfTheta >= 1.0) {

			this._w = w;
			this._x = x;
			this._y = y;
			this._z = z;

			return this;
		}

		var sinHalfTheta = Math.sqrt(1.0 - cosHalfTheta * cosHalfTheta);

		if (Math.abs(sinHalfTheta) < 0.001) {

			this._w = 0.5 * (w + this._w);
			this._x = 0.5 * (x + this._x);
			this._y = 0.5 * (y + this._y);
			this._z = 0.5 * (z + this._z);

			return this;
		}

		var halfTheta = Math.atan2(sinHalfTheta, cosHalfTheta);
		var ratioA = Math.sin((1 - t) * halfTheta) / sinHalfTheta,
		    ratioB = Math.sin(t * halfTheta) / sinHalfTheta;

		this._w = w * ratioA + this._w * ratioB;
		this._x = x * ratioA + this._x * ratioB;
		this._y = y * ratioA + this._y * ratioB;
		this._z = z * ratioA + this._z * ratioB;

		this.onChangeCallback();

		return this;
	},

	equals: function equals(quaternion) {

		return quaternion._x === this._x && quaternion._y === this._y && quaternion._z === this._z && quaternion._w === this._w;
	},

	fromArray: function fromArray(array, offset) {

		if (offset === undefined) offset = 0;

		this._x = array[offset];
		this._y = array[offset + 1];
		this._z = array[offset + 2];
		this._w = array[offset + 3];

		this.onChangeCallback();

		return this;
	},

	toArray: function toArray(array, offset) {

		if (array === undefined) array = [];
		if (offset === undefined) offset = 0;

		array[offset] = this._x;
		array[offset + 1] = this._y;
		array[offset + 2] = this._z;
		array[offset + 3] = this._w;

		return array;
	},

	onChange: function onChange(callback) {

		this.onChangeCallback = callback;

		return this;
	},

	onChangeCallback: function onChangeCallback() {}

};

Object.assign(Quaternion, {

	slerp: function slerp(qa, qb, qm, t) {

		return qm.copy(qa).slerp(qb, t);
	},

	slerpFlat: function slerpFlat(dst, dstOffset, src0, srcOffset0, src1, srcOffset1, t) {

		// fuzz-free, array-based Quaternion SLERP operation

		var x0 = src0[srcOffset0 + 0],
		    y0 = src0[srcOffset0 + 1],
		    z0 = src0[srcOffset0 + 2],
		    w0 = src0[srcOffset0 + 3],
		    x1 = src1[srcOffset1 + 0],
		    y1 = src1[srcOffset1 + 1],
		    z1 = src1[srcOffset1 + 2],
		    w1 = src1[srcOffset1 + 3];

		if (w0 !== w1 || x0 !== x1 || y0 !== y1 || z0 !== z1) {

			var s = 1 - t,
			    cos = x0 * x1 + y0 * y1 + z0 * z1 + w0 * w1,
			    dir = cos >= 0 ? 1 : -1,
			    sqrSin = 1 - cos * cos;

			// Skip the Slerp for tiny steps to avoid numeric problems:
			if (sqrSin > Number.EPSILON) {

				var sin = Math.sqrt(sqrSin),
				    len = Math.atan2(sin, cos * dir);

				s = Math.sin(s * len) / sin;
				t = Math.sin(t * len) / sin;
			}

			var tDir = t * dir;

			x0 = x0 * s + x1 * tDir;
			y0 = y0 * s + y1 * tDir;
			z0 = z0 * s + z1 * tDir;
			w0 = w0 * s + w1 * tDir;

			// Normalize in case we just did a lerp:
			if (s === 1 - t) {

				var f = 1 / Math.sqrt(x0 * x0 + y0 * y0 + z0 * z0 + w0 * w0);

				x0 *= f;
				y0 *= f;
				z0 *= f;
				w0 *= f;
			}
		}

		dst[dstOffset] = x0;
		dst[dstOffset + 1] = y0;
		dst[dstOffset + 2] = z0;
		dst[dstOffset + 3] = w0;
	}

});

/**
 * https://github.com/mrdoob/eventdispatcher.js/
 */

function EventDispatcher() {}

Object.assign(EventDispatcher.prototype, {

	addEventListener: function addEventListener(type, listener) {

		if (this._listeners === undefined) this._listeners = {};

		var listeners = this._listeners;

		if (listeners[type] === undefined) {

			listeners[type] = [];
		}

		if (listeners[type].indexOf(listener) === -1) {

			listeners[type].push(listener);
		}
	},

	hasEventListener: function hasEventListener(type, listener) {

		if (this._listeners === undefined) return false;

		var listeners = this._listeners;

		if (listeners[type] !== undefined && listeners[type].indexOf(listener) !== -1) {

			return true;
		}

		return false;
	},

	removeEventListener: function removeEventListener(type, listener) {

		if (this._listeners === undefined) return;

		var listeners = this._listeners;
		var listenerArray = listeners[type];

		if (listenerArray !== undefined) {

			var index = listenerArray.indexOf(listener);

			if (index !== -1) {

				listenerArray.splice(index, 1);
			}
		}
	},

	dispatchEvent: function dispatchEvent(event) {

		if (this._listeners === undefined) return;

		var listeners = this._listeners;
		var listenerArray = listeners[event.type];

		if (listenerArray !== undefined) {

			event.target = this;

			var array = [],
			    i = 0;
			var length = listenerArray.length;

			for (i = 0; i < length; i++) {

				array[i] = listenerArray[i];
			}

			for (i = 0; i < length; i++) {

				array[i].call(this, event);
			}
		}
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 * @author WestLangley / http://github.com/WestLangley
 * @author bhouston / http://clara.io
 */

function Euler(x, y, z, order) {

	this._x = x || 0;
	this._y = y || 0;
	this._z = z || 0;
	this._order = order || Euler.DefaultOrder;
}

Euler.RotationOrders = ['XYZ', 'YZX', 'ZXY', 'XZY', 'YXZ', 'ZYX'];

Euler.DefaultOrder = 'XYZ';

Euler.prototype = {

	constructor: Euler,

	isEuler: true,

	get x() {

		return this._x;
	},

	set x(value) {

		this._x = value;
		this.onChangeCallback();
	},

	get y() {

		return this._y;
	},

	set y(value) {

		this._y = value;
		this.onChangeCallback();
	},

	get z() {

		return this._z;
	},

	set z(value) {

		this._z = value;
		this.onChangeCallback();
	},

	get order() {

		return this._order;
	},

	set order(value) {

		this._order = value;
		this.onChangeCallback();
	},

	set: function set(x, y, z, order) {

		this._x = x;
		this._y = y;
		this._z = z;
		this._order = order || this._order;

		this.onChangeCallback();

		return this;
	},

	clone: function clone() {

		return new this.constructor(this._x, this._y, this._z, this._order);
	},

	copy: function copy(euler) {

		this._x = euler._x;
		this._y = euler._y;
		this._z = euler._z;
		this._order = euler._order;

		this.onChangeCallback();

		return this;
	},

	setFromRotationMatrix: function setFromRotationMatrix(m, order, update) {

		var clamp = _Math.clamp;

		// assumes the upper 3x3 of m is a pure rotation matrix (i.e, unscaled)

		var te = m.elements;
		var m11 = te[0],
		    m12 = te[4],
		    m13 = te[8];
		var m21 = te[1],
		    m22 = te[5],
		    m23 = te[9];
		var m31 = te[2],
		    m32 = te[6],
		    m33 = te[10];

		order = order || this._order;

		if (order === 'XYZ') {

			this._y = Math.asin(clamp(m13, -1, 1));

			if (Math.abs(m13) < 0.99999) {

				this._x = Math.atan2(-m23, m33);
				this._z = Math.atan2(-m12, m11);
			} else {

				this._x = Math.atan2(m32, m22);
				this._z = 0;
			}
		} else if (order === 'YXZ') {

			this._x = Math.asin(-clamp(m23, -1, 1));

			if (Math.abs(m23) < 0.99999) {

				this._y = Math.atan2(m13, m33);
				this._z = Math.atan2(m21, m22);
			} else {

				this._y = Math.atan2(-m31, m11);
				this._z = 0;
			}
		} else if (order === 'ZXY') {

			this._x = Math.asin(clamp(m32, -1, 1));

			if (Math.abs(m32) < 0.99999) {

				this._y = Math.atan2(-m31, m33);
				this._z = Math.atan2(-m12, m22);
			} else {

				this._y = 0;
				this._z = Math.atan2(m21, m11);
			}
		} else if (order === 'ZYX') {

			this._y = Math.asin(-clamp(m31, -1, 1));

			if (Math.abs(m31) < 0.99999) {

				this._x = Math.atan2(m32, m33);
				this._z = Math.atan2(m21, m11);
			} else {

				this._x = 0;
				this._z = Math.atan2(-m12, m22);
			}
		} else if (order === 'YZX') {

			this._z = Math.asin(clamp(m21, -1, 1));

			if (Math.abs(m21) < 0.99999) {

				this._x = Math.atan2(-m23, m22);
				this._y = Math.atan2(-m31, m11);
			} else {

				this._x = 0;
				this._y = Math.atan2(m13, m33);
			}
		} else if (order === 'XZY') {

			this._z = Math.asin(-clamp(m12, -1, 1));

			if (Math.abs(m12) < 0.99999) {

				this._x = Math.atan2(m32, m22);
				this._y = Math.atan2(m13, m11);
			} else {

				this._x = Math.atan2(-m23, m33);
				this._y = 0;
			}
		} else {

			console.warn('THREE.Euler: .setFromRotationMatrix() given unsupported order: ' + order);
		}

		this._order = order;

		if (update !== false) this.onChangeCallback();

		return this;
	},

	setFromQuaternion: function () {

		var matrix;

		return function setFromQuaternion(q, order, update) {

			if (matrix === undefined) matrix = new Matrix4();

			matrix.makeRotationFromQuaternion(q);

			return this.setFromRotationMatrix(matrix, order, update);
		};
	}(),

	setFromVector3: function setFromVector3(v, order) {

		return this.set(v.x, v.y, v.z, order || this._order);
	},

	reorder: function () {

		// WARNING: this discards revolution information -bhouston

		var q = new Quaternion();

		return function reorder(newOrder) {

			q.setFromEuler(this);

			return this.setFromQuaternion(q, newOrder);
		};
	}(),

	equals: function equals(euler) {

		return euler._x === this._x && euler._y === this._y && euler._z === this._z && euler._order === this._order;
	},

	fromArray: function fromArray(array) {

		this._x = array[0];
		this._y = array[1];
		this._z = array[2];
		if (array[3] !== undefined) this._order = array[3];

		this.onChangeCallback();

		return this;
	},

	toArray: function toArray(array, offset) {

		if (array === undefined) array = [];
		if (offset === undefined) offset = 0;

		array[offset] = this._x;
		array[offset + 1] = this._y;
		array[offset + 2] = this._z;
		array[offset + 3] = this._order;

		return array;
	},

	toVector3: function toVector3(optionalResult) {

		if (optionalResult) {

			return optionalResult.set(this._x, this._y, this._z);
		} else {

			return new Vector3(this._x, this._y, this._z);
		}
	},

	onChange: function onChange(callback) {

		this.onChangeCallback = callback;

		return this;
	},

	onChangeCallback: function onChangeCallback() {}

};

/**
 * @author mrdoob / http://mrdoob.com/
 */

function Layers() {

	this.mask = 1;
}

Layers.prototype = {

	constructor: Layers,

	set: function set(channel) {

		this.mask = 1 << channel;
	},

	enable: function enable(channel) {

		this.mask |= 1 << channel;
	},

	toggle: function toggle(channel) {

		this.mask ^= 1 << channel;
	},

	disable: function disable(channel) {

		this.mask &= ~(1 << channel);
	},

	test: function test(layers) {

		return (this.mask & layers.mask) !== 0;
	}

};

/**
 * @author alteredq / http://alteredqualia.com/
 * @author WestLangley / http://github.com/WestLangley
 * @author bhouston / http://clara.io
 * @author tschw
 */

function Matrix3() {

	this.elements = new Float32Array([1, 0, 0, 0, 1, 0, 0, 0, 1]);

	if (arguments.length > 0) {

		console.error('THREE.Matrix3: the constructor no longer reads arguments. use .set() instead.');
	}
}

Matrix3.prototype = {

	constructor: Matrix3,

	isMatrix3: true,

	set: function set(n11, n12, n13, n21, n22, n23, n31, n32, n33) {

		var te = this.elements;

		te[0] = n11;te[1] = n21;te[2] = n31;
		te[3] = n12;te[4] = n22;te[5] = n32;
		te[6] = n13;te[7] = n23;te[8] = n33;

		return this;
	},

	identity: function identity() {

		this.set(1, 0, 0, 0, 1, 0, 0, 0, 1);

		return this;
	},

	clone: function clone() {

		return new this.constructor().fromArray(this.elements);
	},

	copy: function copy(m) {

		var me = m.elements;

		this.set(me[0], me[3], me[6], me[1], me[4], me[7], me[2], me[5], me[8]);

		return this;
	},

	setFromMatrix4: function setFromMatrix4(m) {

		var me = m.elements;

		this.set(me[0], me[4], me[8], me[1], me[5], me[9], me[2], me[6], me[10]);

		return this;
	},

	applyToVector3Array: function () {

		var v1;

		return function applyToVector3Array(array, offset, length) {

			if (v1 === undefined) v1 = new Vector3();
			if (offset === undefined) offset = 0;
			if (length === undefined) length = array.length;

			for (var i = 0, j = offset; i < length; i += 3, j += 3) {

				v1.fromArray(array, j);
				v1.applyMatrix3(this);
				v1.toArray(array, j);
			}

			return array;
		};
	}(),

	applyToBuffer: function () {

		var v1;

		return function applyToBuffer(buffer, offset, length) {

			if (v1 === undefined) v1 = new Vector3();
			if (offset === undefined) offset = 0;
			if (length === undefined) length = buffer.length / buffer.itemSize;

			for (var i = 0, j = offset; i < length; i++, j++) {

				v1.x = buffer.getX(j);
				v1.y = buffer.getY(j);
				v1.z = buffer.getZ(j);

				v1.applyMatrix3(this);

				buffer.setXYZ(j, v1.x, v1.y, v1.z);
			}

			return buffer;
		};
	}(),

	multiplyScalar: function multiplyScalar(s) {

		var te = this.elements;

		te[0] *= s;te[3] *= s;te[6] *= s;
		te[1] *= s;te[4] *= s;te[7] *= s;
		te[2] *= s;te[5] *= s;te[8] *= s;

		return this;
	},

	determinant: function determinant() {

		var te = this.elements;

		var a = te[0],
		    b = te[1],
		    c = te[2],
		    d = te[3],
		    e = te[4],
		    f = te[5],
		    g = te[6],
		    h = te[7],
		    i = te[8];

		return a * e * i - a * f * h - b * d * i + b * f * g + c * d * h - c * e * g;
	},

	getInverse: function getInverse(matrix, throwOnDegenerate) {

		if (matrix && matrix.isMatrix4) {

			console.error("THREE.Matrix3.getInverse no longer takes a Matrix4 argument.");
		}

		var me = matrix.elements,
		    te = this.elements,
		    n11 = me[0],
		    n21 = me[1],
		    n31 = me[2],
		    n12 = me[3],
		    n22 = me[4],
		    n32 = me[5],
		    n13 = me[6],
		    n23 = me[7],
		    n33 = me[8],
		    t11 = n33 * n22 - n32 * n23,
		    t12 = n32 * n13 - n33 * n12,
		    t13 = n23 * n12 - n22 * n13,
		    det = n11 * t11 + n21 * t12 + n31 * t13;

		if (det === 0) {

			var msg = "THREE.Matrix3.getInverse(): can't invert matrix, determinant is 0";

			if (throwOnDegenerate === true) {

				throw new Error(msg);
			} else {

				console.warn(msg);
			}

			return this.identity();
		}

		var detInv = 1 / det;

		te[0] = t11 * detInv;
		te[1] = (n31 * n23 - n33 * n21) * detInv;
		te[2] = (n32 * n21 - n31 * n22) * detInv;

		te[3] = t12 * detInv;
		te[4] = (n33 * n11 - n31 * n13) * detInv;
		te[5] = (n31 * n12 - n32 * n11) * detInv;

		te[6] = t13 * detInv;
		te[7] = (n21 * n13 - n23 * n11) * detInv;
		te[8] = (n22 * n11 - n21 * n12) * detInv;

		return this;
	},

	transpose: function transpose() {

		var tmp,
		    m = this.elements;

		tmp = m[1];m[1] = m[3];m[3] = tmp;
		tmp = m[2];m[2] = m[6];m[6] = tmp;
		tmp = m[5];m[5] = m[7];m[7] = tmp;

		return this;
	},

	flattenToArrayOffset: function flattenToArrayOffset(array, offset) {

		console.warn("THREE.Matrix3: .flattenToArrayOffset is deprecated " + "- just use .toArray instead.");

		return this.toArray(array, offset);
	},

	getNormalMatrix: function getNormalMatrix(matrix4) {

		return this.setFromMatrix4(matrix4).getInverse(this).transpose();
	},

	transposeIntoArray: function transposeIntoArray(r) {

		var m = this.elements;

		r[0] = m[0];
		r[1] = m[3];
		r[2] = m[6];
		r[3] = m[1];
		r[4] = m[4];
		r[5] = m[7];
		r[6] = m[2];
		r[7] = m[5];
		r[8] = m[8];

		return this;
	},

	fromArray: function fromArray(array, offset) {

		if (offset === undefined) offset = 0;

		for (var i = 0; i < 9; i++) {

			this.elements[i] = array[i + offset];
		}

		return this;
	},

	toArray: function toArray(array, offset) {

		if (array === undefined) array = [];
		if (offset === undefined) offset = 0;

		var te = this.elements;

		array[offset] = te[0];
		array[offset + 1] = te[1];
		array[offset + 2] = te[2];

		array[offset + 3] = te[3];
		array[offset + 4] = te[4];
		array[offset + 5] = te[5];

		array[offset + 6] = te[6];
		array[offset + 7] = te[7];
		array[offset + 8] = te[8];

		return array;
	}

};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author mikael emtinger / http://gomo.se/
 * @author alteredq / http://alteredqualia.com/
 * @author WestLangley / http://github.com/WestLangley
 * @author elephantatwork / www.elephantatwork.ch
 */

function Object3D() {

	Object.defineProperty(this, 'id', { value: Object3DIdCount() });

	this.uuid = _Math.generateUUID();

	this.name = '';
	this.type = 'Object3D';

	this.parent = null;
	this.children = [];

	this.up = Object3D.DefaultUp.clone();

	var position = new Vector3();
	var rotation = new Euler();
	var quaternion = new Quaternion();
	var scale = new Vector3(1, 1, 1);

	function onRotationChange() {

		quaternion.setFromEuler(rotation, false);
	}

	function onQuaternionChange() {

		rotation.setFromQuaternion(quaternion, undefined, false);
	}

	rotation.onChange(onRotationChange);
	quaternion.onChange(onQuaternionChange);

	Object.defineProperties(this, {
		position: {
			enumerable: true,
			value: position
		},
		rotation: {
			enumerable: true,
			value: rotation
		},
		quaternion: {
			enumerable: true,
			value: quaternion
		},
		scale: {
			enumerable: true,
			value: scale
		},
		modelViewMatrix: {
			value: new Matrix4()
		},
		normalMatrix: {
			value: new Matrix3()
		}
	});

	this.matrix = new Matrix4();
	this.matrixWorld = new Matrix4();

	this.matrixAutoUpdate = Object3D.DefaultMatrixAutoUpdate;
	this.matrixWorldNeedsUpdate = false;

	this.layers = new Layers();
	this.visible = true;

	this.castShadow = false;
	this.receiveShadow = false;

	this.frustumCulled = true;
	this.renderOrder = 0;

	this.userData = {};

	this.onBeforeRender = function () {};
	this.onAfterRender = function () {};
}

Object3D.DefaultUp = new Vector3(0, 1, 0);
Object3D.DefaultMatrixAutoUpdate = true;

Object.assign(Object3D.prototype, EventDispatcher.prototype, {

	isObject3D: true,

	applyMatrix: function applyMatrix(matrix) {

		this.matrix.multiplyMatrices(matrix, this.matrix);

		this.matrix.decompose(this.position, this.quaternion, this.scale);
	},

	setRotationFromAxisAngle: function setRotationFromAxisAngle(axis, angle) {

		// assumes axis is normalized

		this.quaternion.setFromAxisAngle(axis, angle);
	},

	setRotationFromEuler: function setRotationFromEuler(euler) {

		this.quaternion.setFromEuler(euler, true);
	},

	setRotationFromMatrix: function setRotationFromMatrix(m) {

		// assumes the upper 3x3 of m is a pure rotation matrix (i.e, unscaled)

		this.quaternion.setFromRotationMatrix(m);
	},

	setRotationFromQuaternion: function setRotationFromQuaternion(q) {

		// assumes q is normalized

		this.quaternion.copy(q);
	},

	rotateOnAxis: function () {

		// rotate object on axis in object space
		// axis is assumed to be normalized

		var q1 = new Quaternion();

		return function rotateOnAxis(axis, angle) {

			q1.setFromAxisAngle(axis, angle);

			this.quaternion.multiply(q1);

			return this;
		};
	}(),

	rotateX: function () {

		var v1 = new Vector3(1, 0, 0);

		return function rotateX(angle) {

			return this.rotateOnAxis(v1, angle);
		};
	}(),

	rotateY: function () {

		var v1 = new Vector3(0, 1, 0);

		return function rotateY(angle) {

			return this.rotateOnAxis(v1, angle);
		};
	}(),

	rotateZ: function () {

		var v1 = new Vector3(0, 0, 1);

		return function rotateZ(angle) {

			return this.rotateOnAxis(v1, angle);
		};
	}(),

	translateOnAxis: function () {

		// translate object by distance along axis in object space
		// axis is assumed to be normalized

		var v1 = new Vector3();

		return function translateOnAxis(axis, distance) {

			v1.copy(axis).applyQuaternion(this.quaternion);

			this.position.add(v1.multiplyScalar(distance));

			return this;
		};
	}(),

	translateX: function () {

		var v1 = new Vector3(1, 0, 0);

		return function translateX(distance) {

			return this.translateOnAxis(v1, distance);
		};
	}(),

	translateY: function () {

		var v1 = new Vector3(0, 1, 0);

		return function translateY(distance) {

			return this.translateOnAxis(v1, distance);
		};
	}(),

	translateZ: function () {

		var v1 = new Vector3(0, 0, 1);

		return function translateZ(distance) {

			return this.translateOnAxis(v1, distance);
		};
	}(),

	localToWorld: function localToWorld(vector) {

		return vector.applyMatrix4(this.matrixWorld);
	},

	worldToLocal: function () {

		var m1 = new Matrix4();

		return function worldToLocal(vector) {

			return vector.applyMatrix4(m1.getInverse(this.matrixWorld));
		};
	}(),

	lookAt: function () {

		// This routine does not support objects with rotated and/or translated parent(s)

		var m1 = new Matrix4();

		return function lookAt(vector) {

			m1.lookAt(vector, this.position, this.up);

			this.quaternion.setFromRotationMatrix(m1);
		};
	}(),

	add: function add(object) {

		if (arguments.length > 1) {

			for (var i = 0; i < arguments.length; i++) {

				this.add(arguments[i]);
			}

			return this;
		}

		if (object === this) {

			console.error("THREE.Object3D.add: object can't be added as a child of itself.", object);
			return this;
		}

		if (object && object.isObject3D) {

			if (object.parent !== null) {

				object.parent.remove(object);
			}

			object.parent = this;
			object.dispatchEvent({ type: 'added' });

			this.children.push(object);
		} else {

			console.error("THREE.Object3D.add: object not an instance of THREE.Object3D.", object);
		}

		return this;
	},

	remove: function remove(object) {

		if (arguments.length > 1) {

			for (var i = 0; i < arguments.length; i++) {

				this.remove(arguments[i]);
			}
		}

		var index = this.children.indexOf(object);

		if (index !== -1) {

			object.parent = null;

			object.dispatchEvent({ type: 'removed' });

			this.children.splice(index, 1);
		}
	},

	getObjectById: function getObjectById(id) {

		return this.getObjectByProperty('id', id);
	},

	getObjectByName: function getObjectByName(name) {

		return this.getObjectByProperty('name', name);
	},

	getObjectByProperty: function getObjectByProperty(name, value) {

		if (this[name] === value) return this;

		for (var i = 0, l = this.children.length; i < l; i++) {

			var child = this.children[i];
			var object = child.getObjectByProperty(name, value);

			if (object !== undefined) {

				return object;
			}
		}

		return undefined;
	},

	getWorldPosition: function getWorldPosition(optionalTarget) {

		var result = optionalTarget || new Vector3();

		this.updateMatrixWorld(true);

		return result.setFromMatrixPosition(this.matrixWorld);
	},

	getWorldQuaternion: function () {

		var position = new Vector3();
		var scale = new Vector3();

		return function getWorldQuaternion(optionalTarget) {

			var result = optionalTarget || new Quaternion();

			this.updateMatrixWorld(true);

			this.matrixWorld.decompose(position, result, scale);

			return result;
		};
	}(),

	getWorldRotation: function () {

		var quaternion = new Quaternion();

		return function getWorldRotation(optionalTarget) {

			var result = optionalTarget || new Euler();

			this.getWorldQuaternion(quaternion);

			return result.setFromQuaternion(quaternion, this.rotation.order, false);
		};
	}(),

	getWorldScale: function () {

		var position = new Vector3();
		var quaternion = new Quaternion();

		return function getWorldScale(optionalTarget) {

			var result = optionalTarget || new Vector3();

			this.updateMatrixWorld(true);

			this.matrixWorld.decompose(position, quaternion, result);

			return result;
		};
	}(),

	getWorldDirection: function () {

		var quaternion = new Quaternion();

		return function getWorldDirection(optionalTarget) {

			var result = optionalTarget || new Vector3();

			this.getWorldQuaternion(quaternion);

			return result.set(0, 0, 1).applyQuaternion(quaternion);
		};
	}(),

	raycast: function raycast() {},

	traverse: function traverse(callback) {

		callback(this);

		var children = this.children;

		for (var i = 0, l = children.length; i < l; i++) {

			children[i].traverse(callback);
		}
	},

	traverseVisible: function traverseVisible(callback) {

		if (this.visible === false) return;

		callback(this);

		var children = this.children;

		for (var i = 0, l = children.length; i < l; i++) {

			children[i].traverseVisible(callback);
		}
	},

	traverseAncestors: function traverseAncestors(callback) {

		var parent = this.parent;

		if (parent !== null) {

			callback(parent);

			parent.traverseAncestors(callback);
		}
	},

	updateMatrix: function updateMatrix() {

		this.matrix.compose(this.position, this.quaternion, this.scale);

		this.matrixWorldNeedsUpdate = true;
	},

	updateMatrixWorld: function updateMatrixWorld(force) {

		if (this.matrixAutoUpdate === true) this.updateMatrix();

		if (this.matrixWorldNeedsUpdate === true || force === true) {

			if (this.parent === null) {

				this.matrixWorld.copy(this.matrix);
			} else {

				this.matrixWorld.multiplyMatrices(this.parent.matrixWorld, this.matrix);
			}

			this.matrixWorldNeedsUpdate = false;

			force = true;
		}

		// update children

		var children = this.children;

		for (var i = 0, l = children.length; i < l; i++) {

			children[i].updateMatrixWorld(force);
		}
	},

	toJSON: function toJSON(meta) {

		// meta is '' when called from JSON.stringify
		var isRootObject = meta === undefined || meta === '';

		var output = {};

		// meta is a hash used to collect geometries, materials.
		// not providing it implies that this is the root object
		// being serialized.
		if (isRootObject) {

			// initialize meta obj
			meta = {
				geometries: {},
				materials: {},
				textures: {},
				images: {}
			};

			output.metadata = {
				version: 4.4,
				type: 'Object',
				generator: 'Object3D.toJSON'
			};
		}

		// standard Object3D serialization

		var object = {};

		object.uuid = this.uuid;
		object.type = this.type;

		if (this.name !== '') object.name = this.name;
		if (JSON.stringify(this.userData) !== '{}') object.userData = this.userData;
		if (this.castShadow === true) object.castShadow = true;
		if (this.receiveShadow === true) object.receiveShadow = true;
		if (this.visible === false) object.visible = false;

		object.matrix = this.matrix.toArray();

		//

		if (this.geometry !== undefined) {

			if (meta.geometries[this.geometry.uuid] === undefined) {

				meta.geometries[this.geometry.uuid] = this.geometry.toJSON(meta);
			}

			object.geometry = this.geometry.uuid;
		}

		if (this.material !== undefined) {

			if (meta.materials[this.material.uuid] === undefined) {

				meta.materials[this.material.uuid] = this.material.toJSON(meta);
			}

			object.material = this.material.uuid;
		}

		//

		if (this.children.length > 0) {

			object.children = [];

			for (var i = 0; i < this.children.length; i++) {

				object.children.push(this.children[i].toJSON(meta).object);
			}
		}

		if (isRootObject) {

			var geometries = extractFromCache(meta.geometries);
			var materials = extractFromCache(meta.materials);
			var textures = extractFromCache(meta.textures);
			var images = extractFromCache(meta.images);

			if (geometries.length > 0) output.geometries = geometries;
			if (materials.length > 0) output.materials = materials;
			if (textures.length > 0) output.textures = textures;
			if (images.length > 0) output.images = images;
		}

		output.object = object;

		return output;

		// extract data from the cache hash
		// remove metadata on each item
		// and return as array
		function extractFromCache(cache) {

			var values = [];
			for (var key in cache) {

				var data = cache[key];
				delete data.metadata;
				values.push(data);
			}
			return values;
		}
	},

	clone: function clone(recursive) {

		return new this.constructor().copy(this, recursive);
	},

	copy: function copy(source, recursive) {

		if (recursive === undefined) recursive = true;

		this.name = source.name;

		this.up.copy(source.up);

		this.position.copy(source.position);
		this.quaternion.copy(source.quaternion);
		this.scale.copy(source.scale);

		this.matrix.copy(source.matrix);
		this.matrixWorld.copy(source.matrixWorld);

		this.matrixAutoUpdate = source.matrixAutoUpdate;
		this.matrixWorldNeedsUpdate = source.matrixWorldNeedsUpdate;

		this.visible = source.visible;

		this.castShadow = source.castShadow;
		this.receiveShadow = source.receiveShadow;

		this.frustumCulled = source.frustumCulled;
		this.renderOrder = source.renderOrder;

		this.userData = JSON.parse(JSON.stringify(source.userData));

		if (recursive === true) {

			for (var i = 0; i < source.children.length; i++) {

				var child = source.children[i];
				this.add(child.clone());
			}
		}

		return this;
	}

});

var count = 0;
function Object3DIdCount() {
	return count++;
}

function hub$1() {
  return new Object3D();
}

function put(object) {
  var box = {
    on: null,
    at: null,
    rot: null,
    scale: null,
    obj: function obj() {
      return object;
    }
  },
      on = function on(scene) {
    if (scene.appendChild) {
      scene.appendChild(object);
    } else {
      scene.add(object);
    }
    box.on = null;
    if (box.at || box.rot || box.scale) {
      return box;
    } else {
      return object;
    }
  },
      at = function at(x, y, z) {
    object.position.set(x, y, z);
    box.at = null;
    if (box.on || box.rot || box.scale) {
      return box;
    } else {
      return object;
    }
  },
      rot = function rot(x, y, z) {
    object.rotation.set(x, y, z);
    box.rot = null;
    if (box.on || box.at || box.scale) {
      return box;
    } else {
      return object;
    }
  },
      scale = function scale(x, y, z) {
    object.scale.set(x, y, z);
    box.scale = null;
    if (box.on || box.at || box.rot) {
      return box;
    } else {
      return object;
    }
  };

  box.on = on;
  box.at = at;
  box.rot = rot;
  box.scale = scale;

  return box;
}

var _cache = {};
function cache(hash, makeObject, onCacheHit) {
  if (!_cache[hash]) {
    _cache[hash] = makeObject();
  } else if (onCacheHit) {
    console.log(hash);
    onCacheHit(_cache[hash]);
  }
  return _cache[hash];
}

var REVISION = '82';

var CullFaceNone = 0;
var CullFaceBack = 1;
var CullFaceFront = 2;

var FrontFaceDirectionCW = 0;


var PCFShadowMap = 1;
var PCFSoftShadowMap = 2;
var FrontSide = 0;
var BackSide = 1;
var DoubleSide = 2;
var FlatShading = 1;
var SmoothShading = 2;
var NoColors = 0;
var FaceColors = 1;
var VertexColors = 2;
var NoBlending = 0;
var NormalBlending = 1;
var AdditiveBlending = 2;
var SubtractiveBlending = 3;
var MultiplyBlending = 4;
var CustomBlending = 5;
var BlendingMode = {
	NoBlending: NoBlending,
	NormalBlending: NormalBlending,
	AdditiveBlending: AdditiveBlending,
	SubtractiveBlending: SubtractiveBlending,
	MultiplyBlending: MultiplyBlending,
	CustomBlending: CustomBlending
};
var AddEquation = 100;
var SubtractEquation = 101;
var ReverseSubtractEquation = 102;
var MinEquation = 103;
var MaxEquation = 104;
var ZeroFactor = 200;
var OneFactor = 201;
var SrcColorFactor = 202;
var OneMinusSrcColorFactor = 203;
var SrcAlphaFactor = 204;
var OneMinusSrcAlphaFactor = 205;
var DstAlphaFactor = 206;
var OneMinusDstAlphaFactor = 207;
var DstColorFactor = 208;
var OneMinusDstColorFactor = 209;
var SrcAlphaSaturateFactor = 210;
var NeverDepth = 0;
var AlwaysDepth = 1;
var LessDepth = 2;
var LessEqualDepth = 3;
var EqualDepth = 4;
var GreaterEqualDepth = 5;
var GreaterDepth = 6;
var NotEqualDepth = 7;
var MultiplyOperation = 0;
var MixOperation = 1;
var AddOperation = 2;
var NoToneMapping = 0;
var LinearToneMapping = 1;
var ReinhardToneMapping = 2;
var Uncharted2ToneMapping = 3;
var CineonToneMapping = 4;
var UVMapping = 300;
var CubeReflectionMapping = 301;
var CubeRefractionMapping = 302;
var EquirectangularReflectionMapping = 303;
var EquirectangularRefractionMapping = 304;
var SphericalReflectionMapping = 305;
var CubeUVReflectionMapping = 306;
var CubeUVRefractionMapping = 307;
var TextureMapping = {
	UVMapping: UVMapping,
	CubeReflectionMapping: CubeReflectionMapping,
	CubeRefractionMapping: CubeRefractionMapping,
	EquirectangularReflectionMapping: EquirectangularReflectionMapping,
	EquirectangularRefractionMapping: EquirectangularRefractionMapping,
	SphericalReflectionMapping: SphericalReflectionMapping,
	CubeUVReflectionMapping: CubeUVReflectionMapping,
	CubeUVRefractionMapping: CubeUVRefractionMapping
};
var RepeatWrapping = 1000;
var ClampToEdgeWrapping = 1001;
var MirroredRepeatWrapping = 1002;
var TextureWrapping = {
	RepeatWrapping: RepeatWrapping,
	ClampToEdgeWrapping: ClampToEdgeWrapping,
	MirroredRepeatWrapping: MirroredRepeatWrapping
};
var NearestFilter = 1003;
var NearestMipMapNearestFilter = 1004;
var NearestMipMapLinearFilter = 1005;
var LinearFilter = 1006;
var LinearMipMapNearestFilter = 1007;
var LinearMipMapLinearFilter = 1008;
var TextureFilter = {
	NearestFilter: NearestFilter,
	NearestMipMapNearestFilter: NearestMipMapNearestFilter,
	NearestMipMapLinearFilter: NearestMipMapLinearFilter,
	LinearFilter: LinearFilter,
	LinearMipMapNearestFilter: LinearMipMapNearestFilter,
	LinearMipMapLinearFilter: LinearMipMapLinearFilter
};
var UnsignedByteType = 1009;
var ByteType = 1010;
var ShortType = 1011;
var UnsignedShortType = 1012;
var IntType = 1013;
var UnsignedIntType = 1014;
var FloatType = 1015;
var HalfFloatType = 1016;
var UnsignedShort4444Type = 1017;
var UnsignedShort5551Type = 1018;
var UnsignedShort565Type = 1019;
var UnsignedInt248Type = 1020;
var AlphaFormat = 1021;
var RGBFormat = 1022;
var RGBAFormat = 1023;
var LuminanceFormat = 1024;
var LuminanceAlphaFormat = 1025;

var DepthFormat = 1026;
var DepthStencilFormat = 1027;
var RGB_S3TC_DXT1_Format = 2001;
var RGBA_S3TC_DXT1_Format = 2002;
var RGBA_S3TC_DXT3_Format = 2003;
var RGBA_S3TC_DXT5_Format = 2004;
var RGB_PVRTC_4BPPV1_Format = 2100;
var RGB_PVRTC_2BPPV1_Format = 2101;
var RGBA_PVRTC_4BPPV1_Format = 2102;
var RGBA_PVRTC_2BPPV1_Format = 2103;
var RGB_ETC1_Format = 2151;



var InterpolateDiscrete = 2300;
var InterpolateLinear = 2301;
var InterpolateSmooth = 2302;
var ZeroCurvatureEnding = 2400;
var ZeroSlopeEnding = 2401;
var WrapAroundEnding = 2402;
var TrianglesDrawMode = 0;
var TriangleStripDrawMode = 1;
var TriangleFanDrawMode = 2;
var LinearEncoding = 3000;
var sRGBEncoding = 3001;
var GammaEncoding = 3007;
var RGBEEncoding = 3002;

var RGBM7Encoding = 3004;
var RGBM16Encoding = 3005;
var RGBDEncoding = 3006;
var BasicDepthPacking = 3200;
var RGBADepthPacking = 3201;

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 */

function Material() {

	Object.defineProperty(this, 'id', { value: MaterialIdCount() });

	this.uuid = _Math.generateUUID();

	this.name = '';
	this.type = 'Material';

	this.fog = true;
	this.lights = true;

	this.blending = NormalBlending;
	this.side = FrontSide;
	this.shading = SmoothShading; // THREE.FlatShading, THREE.SmoothShading
	this.vertexColors = NoColors; // THREE.NoColors, THREE.VertexColors, THREE.FaceColors

	this.opacity = 1;
	this.transparent = false;

	this.blendSrc = SrcAlphaFactor;
	this.blendDst = OneMinusSrcAlphaFactor;
	this.blendEquation = AddEquation;
	this.blendSrcAlpha = null;
	this.blendDstAlpha = null;
	this.blendEquationAlpha = null;

	this.depthFunc = LessEqualDepth;
	this.depthTest = true;
	this.depthWrite = true;

	this.clippingPlanes = null;
	this.clipIntersection = false;
	this.clipShadows = false;

	this.colorWrite = true;

	this.precision = null; // override the renderer's default precision for this material

	this.polygonOffset = false;
	this.polygonOffsetFactor = 0;
	this.polygonOffsetUnits = 0;

	this.alphaTest = 0;
	this.premultipliedAlpha = false;

	this.overdraw = 0; // Overdrawn pixels (typically between 0 and 1) for fixing antialiasing gaps in CanvasRenderer

	this.visible = true;

	this._needsUpdate = true;
}

Material.prototype = {

	constructor: Material,

	isMaterial: true,

	get needsUpdate() {

		return this._needsUpdate;
	},

	set needsUpdate(value) {

		if (value === true) this.update();
		this._needsUpdate = value;
	},

	setValues: function setValues(values) {

		if (values === undefined) return;

		for (var key in values) {

			var newValue = values[key];

			if (newValue === undefined) {

				console.warn("THREE.Material: '" + key + "' parameter is undefined.");
				continue;
			}

			var currentValue = this[key];

			if (currentValue === undefined) {

				console.warn("THREE." + this.type + ": '" + key + "' is not a property of this material.");
				continue;
			}

			if (currentValue && currentValue.isColor) {

				currentValue.set(newValue);
			} else if (currentValue && currentValue.isVector3 && newValue && newValue.isVector3) {

				currentValue.copy(newValue);
			} else if (key === 'overdraw') {

				// ensure overdraw is backwards-compatible with legacy boolean type
				this[key] = Number(newValue);
			} else {

				this[key] = newValue;
			}
		}
	},

	toJSON: function toJSON(meta) {

		var isRoot = meta === undefined;

		if (isRoot) {

			meta = {
				textures: {},
				images: {}
			};
		}

		var data = {
			metadata: {
				version: 4.4,
				type: 'Material',
				generator: 'Material.toJSON'
			}
		};

		// standard Material serialization
		data.uuid = this.uuid;
		data.type = this.type;

		if (this.name !== '') data.name = this.name;

		if (this.color && this.color.isColor) data.color = this.color.getHex();

		if (this.roughness !== undefined) data.roughness = this.roughness;
		if (this.metalness !== undefined) data.metalness = this.metalness;

		if (this.emissive && this.emissive.isColor) data.emissive = this.emissive.getHex();
		if (this.specular && this.specular.isColor) data.specular = this.specular.getHex();
		if (this.shininess !== undefined) data.shininess = this.shininess;

		if (this.map && this.map.isTexture) data.map = this.map.toJSON(meta).uuid;
		if (this.alphaMap && this.alphaMap.isTexture) data.alphaMap = this.alphaMap.toJSON(meta).uuid;
		if (this.lightMap && this.lightMap.isTexture) data.lightMap = this.lightMap.toJSON(meta).uuid;
		if (this.bumpMap && this.bumpMap.isTexture) {

			data.bumpMap = this.bumpMap.toJSON(meta).uuid;
			data.bumpScale = this.bumpScale;
		}
		if (this.normalMap && this.normalMap.isTexture) {

			data.normalMap = this.normalMap.toJSON(meta).uuid;
			data.normalScale = this.normalScale.toArray();
		}
		if (this.displacementMap && this.displacementMap.isTexture) {

			data.displacementMap = this.displacementMap.toJSON(meta).uuid;
			data.displacementScale = this.displacementScale;
			data.displacementBias = this.displacementBias;
		}
		if (this.roughnessMap && this.roughnessMap.isTexture) data.roughnessMap = this.roughnessMap.toJSON(meta).uuid;
		if (this.metalnessMap && this.metalnessMap.isTexture) data.metalnessMap = this.metalnessMap.toJSON(meta).uuid;

		if (this.emissiveMap && this.emissiveMap.isTexture) data.emissiveMap = this.emissiveMap.toJSON(meta).uuid;
		if (this.specularMap && this.specularMap.isTexture) data.specularMap = this.specularMap.toJSON(meta).uuid;

		if (this.envMap && this.envMap.isTexture) {

			data.envMap = this.envMap.toJSON(meta).uuid;
			data.reflectivity = this.reflectivity; // Scale behind envMap
		}

		if (this.size !== undefined) data.size = this.size;
		if (this.sizeAttenuation !== undefined) data.sizeAttenuation = this.sizeAttenuation;

		if (this.blending !== NormalBlending) data.blending = this.blending;
		if (this.shading !== SmoothShading) data.shading = this.shading;
		if (this.side !== FrontSide) data.side = this.side;
		if (this.vertexColors !== NoColors) data.vertexColors = this.vertexColors;

		if (this.opacity < 1) data.opacity = this.opacity;
		if (this.transparent === true) data.transparent = this.transparent;

		data.depthFunc = this.depthFunc;
		data.depthTest = this.depthTest;
		data.depthWrite = this.depthWrite;

		if (this.alphaTest > 0) data.alphaTest = this.alphaTest;
		if (this.premultipliedAlpha === true) data.premultipliedAlpha = this.premultipliedAlpha;
		if (this.wireframe === true) data.wireframe = this.wireframe;
		if (this.wireframeLinewidth > 1) data.wireframeLinewidth = this.wireframeLinewidth;
		if (this.wireframeLinecap !== 'round') data.wireframeLinecap = this.wireframeLinecap;
		if (this.wireframeLinejoin !== 'round') data.wireframeLinejoin = this.wireframeLinejoin;

		data.skinning = this.skinning;
		data.morphTargets = this.morphTargets;

		// TODO: Copied from Object3D.toJSON

		function extractFromCache(cache) {

			var values = [];

			for (var key in cache) {

				var data = cache[key];
				delete data.metadata;
				values.push(data);
			}

			return values;
		}

		if (isRoot) {

			var textures = extractFromCache(meta.textures);
			var images = extractFromCache(meta.images);

			if (textures.length > 0) data.textures = textures;
			if (images.length > 0) data.images = images;
		}

		return data;
	},

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(source) {

		this.name = source.name;

		this.fog = source.fog;
		this.lights = source.lights;

		this.blending = source.blending;
		this.side = source.side;
		this.shading = source.shading;
		this.vertexColors = source.vertexColors;

		this.opacity = source.opacity;
		this.transparent = source.transparent;

		this.blendSrc = source.blendSrc;
		this.blendDst = source.blendDst;
		this.blendEquation = source.blendEquation;
		this.blendSrcAlpha = source.blendSrcAlpha;
		this.blendDstAlpha = source.blendDstAlpha;
		this.blendEquationAlpha = source.blendEquationAlpha;

		this.depthFunc = source.depthFunc;
		this.depthTest = source.depthTest;
		this.depthWrite = source.depthWrite;

		this.colorWrite = source.colorWrite;

		this.precision = source.precision;

		this.polygonOffset = source.polygonOffset;
		this.polygonOffsetFactor = source.polygonOffsetFactor;
		this.polygonOffsetUnits = source.polygonOffsetUnits;

		this.alphaTest = source.alphaTest;

		this.premultipliedAlpha = source.premultipliedAlpha;

		this.overdraw = source.overdraw;

		this.visible = source.visible;
		this.clipShadows = source.clipShadows;
		this.clipIntersection = source.clipIntersection;

		var srcPlanes = source.clippingPlanes,
		    dstPlanes = null;

		if (srcPlanes !== null) {

			var n = srcPlanes.length;
			dstPlanes = new Array(n);

			for (var i = 0; i !== n; ++i) {
				dstPlanes[i] = srcPlanes[i].clone();
			}
		}

		this.clippingPlanes = dstPlanes;

		return this;
	},

	update: function update() {

		this.dispatchEvent({ type: 'update' });
	},

	dispose: function dispose() {

		this.dispatchEvent({ type: 'dispose' });
	}

};

Object.assign(Material.prototype, EventDispatcher.prototype);

var count$1 = 0;
function MaterialIdCount() {
	return count$1++;
}

/**
 * @author mrdoob / http://mrdoob.com/
 * @author philogb / http://blog.thejit.org/
 * @author egraether / http://egraether.com/
 * @author zz85 / http://www.lab4games.net/zz85/blog
 */

function Vector2(x, y) {

	this.x = x || 0;
	this.y = y || 0;
}

Vector2.prototype = {

	constructor: Vector2,

	isVector2: true,

	get width() {

		return this.x;
	},

	set width(value) {

		this.x = value;
	},

	get height() {

		return this.y;
	},

	set height(value) {

		this.y = value;
	},

	//

	set: function set(x, y) {

		this.x = x;
		this.y = y;

		return this;
	},

	setScalar: function setScalar(scalar) {

		this.x = scalar;
		this.y = scalar;

		return this;
	},

	setX: function setX(x) {

		this.x = x;

		return this;
	},

	setY: function setY(y) {

		this.y = y;

		return this;
	},

	setComponent: function setComponent(index, value) {

		switch (index) {

			case 0:
				this.x = value;break;
			case 1:
				this.y = value;break;
			default:
				throw new Error('index is out of range: ' + index);

		}

		return this;
	},

	getComponent: function getComponent(index) {

		switch (index) {

			case 0:
				return this.x;
			case 1:
				return this.y;
			default:
				throw new Error('index is out of range: ' + index);

		}
	},

	clone: function clone() {

		return new this.constructor(this.x, this.y);
	},

	copy: function copy(v) {

		this.x = v.x;
		this.y = v.y;

		return this;
	},

	add: function add(v, w) {

		if (w !== undefined) {

			console.warn('THREE.Vector2: .add() now only accepts one argument. Use .addVectors( a, b ) instead.');
			return this.addVectors(v, w);
		}

		this.x += v.x;
		this.y += v.y;

		return this;
	},

	addScalar: function addScalar(s) {

		this.x += s;
		this.y += s;

		return this;
	},

	addVectors: function addVectors(a, b) {

		this.x = a.x + b.x;
		this.y = a.y + b.y;

		return this;
	},

	addScaledVector: function addScaledVector(v, s) {

		this.x += v.x * s;
		this.y += v.y * s;

		return this;
	},

	sub: function sub(v, w) {

		if (w !== undefined) {

			console.warn('THREE.Vector2: .sub() now only accepts one argument. Use .subVectors( a, b ) instead.');
			return this.subVectors(v, w);
		}

		this.x -= v.x;
		this.y -= v.y;

		return this;
	},

	subScalar: function subScalar(s) {

		this.x -= s;
		this.y -= s;

		return this;
	},

	subVectors: function subVectors(a, b) {

		this.x = a.x - b.x;
		this.y = a.y - b.y;

		return this;
	},

	multiply: function multiply(v) {

		this.x *= v.x;
		this.y *= v.y;

		return this;
	},

	multiplyScalar: function multiplyScalar(scalar) {

		if (isFinite(scalar)) {

			this.x *= scalar;
			this.y *= scalar;
		} else {

			this.x = 0;
			this.y = 0;
		}

		return this;
	},

	divide: function divide(v) {

		this.x /= v.x;
		this.y /= v.y;

		return this;
	},

	divideScalar: function divideScalar(scalar) {

		return this.multiplyScalar(1 / scalar);
	},

	min: function min(v) {

		this.x = Math.min(this.x, v.x);
		this.y = Math.min(this.y, v.y);

		return this;
	},

	max: function max(v) {

		this.x = Math.max(this.x, v.x);
		this.y = Math.max(this.y, v.y);

		return this;
	},

	clamp: function clamp(min, max) {

		// This function assumes min < max, if this assumption isn't true it will not operate correctly

		this.x = Math.max(min.x, Math.min(max.x, this.x));
		this.y = Math.max(min.y, Math.min(max.y, this.y));

		return this;
	},

	clampScalar: function () {

		var min, max;

		return function clampScalar(minVal, maxVal) {

			if (min === undefined) {

				min = new Vector2();
				max = new Vector2();
			}

			min.set(minVal, minVal);
			max.set(maxVal, maxVal);

			return this.clamp(min, max);
		};
	}(),

	clampLength: function clampLength(min, max) {

		var length = this.length();

		return this.multiplyScalar(Math.max(min, Math.min(max, length)) / length);
	},

	floor: function floor() {

		this.x = Math.floor(this.x);
		this.y = Math.floor(this.y);

		return this;
	},

	ceil: function ceil() {

		this.x = Math.ceil(this.x);
		this.y = Math.ceil(this.y);

		return this;
	},

	round: function round() {

		this.x = Math.round(this.x);
		this.y = Math.round(this.y);

		return this;
	},

	roundToZero: function roundToZero() {

		this.x = this.x < 0 ? Math.ceil(this.x) : Math.floor(this.x);
		this.y = this.y < 0 ? Math.ceil(this.y) : Math.floor(this.y);

		return this;
	},

	negate: function negate() {

		this.x = -this.x;
		this.y = -this.y;

		return this;
	},

	dot: function dot(v) {

		return this.x * v.x + this.y * v.y;
	},

	lengthSq: function lengthSq() {

		return this.x * this.x + this.y * this.y;
	},

	length: function length() {

		return Math.sqrt(this.x * this.x + this.y * this.y);
	},

	lengthManhattan: function lengthManhattan() {

		return Math.abs(this.x) + Math.abs(this.y);
	},

	normalize: function normalize() {

		return this.divideScalar(this.length());
	},

	angle: function angle() {

		// computes the angle in radians with respect to the positive x-axis

		var angle = Math.atan2(this.y, this.x);

		if (angle < 0) angle += 2 * Math.PI;

		return angle;
	},

	distanceTo: function distanceTo(v) {

		return Math.sqrt(this.distanceToSquared(v));
	},

	distanceToSquared: function distanceToSquared(v) {

		var dx = this.x - v.x,
		    dy = this.y - v.y;
		return dx * dx + dy * dy;
	},

	distanceToManhattan: function distanceToManhattan(v) {

		return Math.abs(this.x - v.x) + Math.abs(this.y - v.y);
	},

	setLength: function setLength(length) {

		return this.multiplyScalar(length / this.length());
	},

	lerp: function lerp(v, alpha) {

		this.x += (v.x - this.x) * alpha;
		this.y += (v.y - this.y) * alpha;

		return this;
	},

	lerpVectors: function lerpVectors(v1, v2, alpha) {

		return this.subVectors(v2, v1).multiplyScalar(alpha).add(v1);
	},

	equals: function equals(v) {

		return v.x === this.x && v.y === this.y;
	},

	fromArray: function fromArray(array, offset) {

		if (offset === undefined) offset = 0;

		this.x = array[offset];
		this.y = array[offset + 1];

		return this;
	},

	toArray: function toArray(array, offset) {

		if (array === undefined) array = [];
		if (offset === undefined) offset = 0;

		array[offset] = this.x;
		array[offset + 1] = this.y;

		return array;
	},

	fromAttribute: function fromAttribute(attribute, index, offset) {

		if (offset === undefined) offset = 0;

		index = index * attribute.itemSize + offset;

		this.x = attribute.array[index];
		this.y = attribute.array[index + 1];

		return this;
	},

	rotateAround: function rotateAround(center, angle) {

		var c = Math.cos(angle),
		    s = Math.sin(angle);

		var x = this.x - center.x;
		var y = this.y - center.y;

		this.x = x * c - y * s + center.x;
		this.y = x * s + y * c + center.y;

		return this;
	}

};

/**
 * @author mrdoob / http://mrdoob.com/
 */

function Color(r, g, b) {

	if (g === undefined && b === undefined) {

		// r is THREE.Color, hex or string
		return this.set(r);
	}

	return this.setRGB(r, g, b);
}

Color.prototype = {

	constructor: Color,

	isColor: true,

	r: 1, g: 1, b: 1,

	set: function set(value) {

		if (value && value.isColor) {

			this.copy(value);
		} else if (typeof value === 'number') {

			this.setHex(value);
		} else if (typeof value === 'string') {

			this.setStyle(value);
		}

		return this;
	},

	setScalar: function setScalar(scalar) {

		this.r = scalar;
		this.g = scalar;
		this.b = scalar;

		return this;
	},

	setHex: function setHex(hex) {

		hex = Math.floor(hex);

		this.r = (hex >> 16 & 255) / 255;
		this.g = (hex >> 8 & 255) / 255;
		this.b = (hex & 255) / 255;

		return this;
	},

	setRGB: function setRGB(r, g, b) {

		this.r = r;
		this.g = g;
		this.b = b;

		return this;
	},

	setHSL: function () {

		function hue2rgb(p, q, t) {

			if (t < 0) t += 1;
			if (t > 1) t -= 1;
			if (t < 1 / 6) return p + (q - p) * 6 * t;
			if (t < 1 / 2) return q;
			if (t < 2 / 3) return p + (q - p) * 6 * (2 / 3 - t);
			return p;
		}

		return function setHSL(h, s, l) {

			// h,s,l ranges are in 0.0 - 1.0
			h = _Math.euclideanModulo(h, 1);
			s = _Math.clamp(s, 0, 1);
			l = _Math.clamp(l, 0, 1);

			if (s === 0) {

				this.r = this.g = this.b = l;
			} else {

				var p = l <= 0.5 ? l * (1 + s) : l + s - l * s;
				var q = 2 * l - p;

				this.r = hue2rgb(q, p, h + 1 / 3);
				this.g = hue2rgb(q, p, h);
				this.b = hue2rgb(q, p, h - 1 / 3);
			}

			return this;
		};
	}(),

	setStyle: function setStyle(style) {

		function handleAlpha(string) {

			if (string === undefined) return;

			if (parseFloat(string) < 1) {

				console.warn('THREE.Color: Alpha component of ' + style + ' will be ignored.');
			}
		}

		var m;

		if (m = /^((?:rgb|hsl)a?)\(\s*([^\)]*)\)/.exec(style)) {

			// rgb / hsl

			var color;
			var name = m[1];
			var components = m[2];

			switch (name) {

				case 'rgb':
				case 'rgba':

					if (color = /^(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*(,\s*([0-9]*\.?[0-9]+)\s*)?$/.exec(components)) {

						// rgb(255,0,0) rgba(255,0,0,0.5)
						this.r = Math.min(255, parseInt(color[1], 10)) / 255;
						this.g = Math.min(255, parseInt(color[2], 10)) / 255;
						this.b = Math.min(255, parseInt(color[3], 10)) / 255;

						handleAlpha(color[5]);

						return this;
					}

					if (color = /^(\d+)\%\s*,\s*(\d+)\%\s*,\s*(\d+)\%\s*(,\s*([0-9]*\.?[0-9]+)\s*)?$/.exec(components)) {

						// rgb(100%,0%,0%) rgba(100%,0%,0%,0.5)
						this.r = Math.min(100, parseInt(color[1], 10)) / 100;
						this.g = Math.min(100, parseInt(color[2], 10)) / 100;
						this.b = Math.min(100, parseInt(color[3], 10)) / 100;

						handleAlpha(color[5]);

						return this;
					}

					break;

				case 'hsl':
				case 'hsla':

					if (color = /^([0-9]*\.?[0-9]+)\s*,\s*(\d+)\%\s*,\s*(\d+)\%\s*(,\s*([0-9]*\.?[0-9]+)\s*)?$/.exec(components)) {

						// hsl(120,50%,50%) hsla(120,50%,50%,0.5)
						var h = parseFloat(color[1]) / 360;
						var s = parseInt(color[2], 10) / 100;
						var l = parseInt(color[3], 10) / 100;

						handleAlpha(color[5]);

						return this.setHSL(h, s, l);
					}

					break;

			}
		} else if (m = /^\#([A-Fa-f0-9]+)$/.exec(style)) {

			// hex color

			var hex = m[1];
			var size = hex.length;

			if (size === 3) {

				// #ff0
				this.r = parseInt(hex.charAt(0) + hex.charAt(0), 16) / 255;
				this.g = parseInt(hex.charAt(1) + hex.charAt(1), 16) / 255;
				this.b = parseInt(hex.charAt(2) + hex.charAt(2), 16) / 255;

				return this;
			} else if (size === 6) {

				// #ff0000
				this.r = parseInt(hex.charAt(0) + hex.charAt(1), 16) / 255;
				this.g = parseInt(hex.charAt(2) + hex.charAt(3), 16) / 255;
				this.b = parseInt(hex.charAt(4) + hex.charAt(5), 16) / 255;

				return this;
			}
		}

		if (style && style.length > 0) {

			// color keywords
			var hex = ColorKeywords[style];

			if (hex !== undefined) {

				// red
				this.setHex(hex);
			} else {

				// unknown color
				console.warn('THREE.Color: Unknown color ' + style);
			}
		}

		return this;
	},

	clone: function clone() {

		return new this.constructor(this.r, this.g, this.b);
	},

	copy: function copy(color) {

		this.r = color.r;
		this.g = color.g;
		this.b = color.b;

		return this;
	},

	copyGammaToLinear: function copyGammaToLinear(color, gammaFactor) {

		if (gammaFactor === undefined) gammaFactor = 2.0;

		this.r = Math.pow(color.r, gammaFactor);
		this.g = Math.pow(color.g, gammaFactor);
		this.b = Math.pow(color.b, gammaFactor);

		return this;
	},

	copyLinearToGamma: function copyLinearToGamma(color, gammaFactor) {

		if (gammaFactor === undefined) gammaFactor = 2.0;

		var safeInverse = gammaFactor > 0 ? 1.0 / gammaFactor : 1.0;

		this.r = Math.pow(color.r, safeInverse);
		this.g = Math.pow(color.g, safeInverse);
		this.b = Math.pow(color.b, safeInverse);

		return this;
	},

	convertGammaToLinear: function convertGammaToLinear() {

		var r = this.r,
		    g = this.g,
		    b = this.b;

		this.r = r * r;
		this.g = g * g;
		this.b = b * b;

		return this;
	},

	convertLinearToGamma: function convertLinearToGamma() {

		this.r = Math.sqrt(this.r);
		this.g = Math.sqrt(this.g);
		this.b = Math.sqrt(this.b);

		return this;
	},

	getHex: function getHex() {

		return this.r * 255 << 16 ^ this.g * 255 << 8 ^ this.b * 255 << 0;
	},

	getHexString: function getHexString() {

		return ('000000' + this.getHex().toString(16)).slice(-6);
	},

	getHSL: function getHSL(optionalTarget) {

		// h,s,l ranges are in 0.0 - 1.0

		var hsl = optionalTarget || { h: 0, s: 0, l: 0 };

		var r = this.r,
		    g = this.g,
		    b = this.b;

		var max = Math.max(r, g, b);
		var min = Math.min(r, g, b);

		var hue, saturation;
		var lightness = (min + max) / 2.0;

		if (min === max) {

			hue = 0;
			saturation = 0;
		} else {

			var delta = max - min;

			saturation = lightness <= 0.5 ? delta / (max + min) : delta / (2 - max - min);

			switch (max) {

				case r:
					hue = (g - b) / delta + (g < b ? 6 : 0);break;
				case g:
					hue = (b - r) / delta + 2;break;
				case b:
					hue = (r - g) / delta + 4;break;

			}

			hue /= 6;
		}

		hsl.h = hue;
		hsl.s = saturation;
		hsl.l = lightness;

		return hsl;
	},

	getStyle: function getStyle() {

		return 'rgb(' + (this.r * 255 | 0) + ',' + (this.g * 255 | 0) + ',' + (this.b * 255 | 0) + ')';
	},

	offsetHSL: function offsetHSL(h, s, l) {

		var hsl = this.getHSL();

		hsl.h += h;hsl.s += s;hsl.l += l;

		this.setHSL(hsl.h, hsl.s, hsl.l);

		return this;
	},

	add: function add(color) {

		this.r += color.r;
		this.g += color.g;
		this.b += color.b;

		return this;
	},

	addColors: function addColors(color1, color2) {

		this.r = color1.r + color2.r;
		this.g = color1.g + color2.g;
		this.b = color1.b + color2.b;

		return this;
	},

	addScalar: function addScalar(s) {

		this.r += s;
		this.g += s;
		this.b += s;

		return this;
	},

	sub: function sub(color) {

		this.r = Math.max(0, this.r - color.r);
		this.g = Math.max(0, this.g - color.g);
		this.b = Math.max(0, this.b - color.b);

		return this;
	},

	multiply: function multiply(color) {

		this.r *= color.r;
		this.g *= color.g;
		this.b *= color.b;

		return this;
	},

	multiplyScalar: function multiplyScalar(s) {

		this.r *= s;
		this.g *= s;
		this.b *= s;

		return this;
	},

	lerp: function lerp(color, alpha) {

		this.r += (color.r - this.r) * alpha;
		this.g += (color.g - this.g) * alpha;
		this.b += (color.b - this.b) * alpha;

		return this;
	},

	equals: function equals(c) {

		return c.r === this.r && c.g === this.g && c.b === this.b;
	},

	fromArray: function fromArray(array, offset) {

		if (offset === undefined) offset = 0;

		this.r = array[offset];
		this.g = array[offset + 1];
		this.b = array[offset + 2];

		return this;
	},

	toArray: function toArray(array, offset) {

		if (array === undefined) array = [];
		if (offset === undefined) offset = 0;

		array[offset] = this.r;
		array[offset + 1] = this.g;
		array[offset + 2] = this.b;

		return array;
	},

	toJSON: function toJSON() {

		return this.getHex();
	}

};

var ColorKeywords = { 'aliceblue': 0xF0F8FF, 'antiquewhite': 0xFAEBD7, 'aqua': 0x00FFFF, 'aquamarine': 0x7FFFD4, 'azure': 0xF0FFFF,
	'beige': 0xF5F5DC, 'bisque': 0xFFE4C4, 'black': 0x000000, 'blanchedalmond': 0xFFEBCD, 'blue': 0x0000FF, 'blueviolet': 0x8A2BE2,
	'brown': 0xA52A2A, 'burlywood': 0xDEB887, 'cadetblue': 0x5F9EA0, 'chartreuse': 0x7FFF00, 'chocolate': 0xD2691E, 'coral': 0xFF7F50,
	'cornflowerblue': 0x6495ED, 'cornsilk': 0xFFF8DC, 'crimson': 0xDC143C, 'cyan': 0x00FFFF, 'darkblue': 0x00008B, 'darkcyan': 0x008B8B,
	'darkgoldenrod': 0xB8860B, 'darkgray': 0xA9A9A9, 'darkgreen': 0x006400, 'darkgrey': 0xA9A9A9, 'darkkhaki': 0xBDB76B, 'darkmagenta': 0x8B008B,
	'darkolivegreen': 0x556B2F, 'darkorange': 0xFF8C00, 'darkorchid': 0x9932CC, 'darkred': 0x8B0000, 'darksalmon': 0xE9967A, 'darkseagreen': 0x8FBC8F,
	'darkslateblue': 0x483D8B, 'darkslategray': 0x2F4F4F, 'darkslategrey': 0x2F4F4F, 'darkturquoise': 0x00CED1, 'darkviolet': 0x9400D3,
	'deeppink': 0xFF1493, 'deepskyblue': 0x00BFFF, 'dimgray': 0x696969, 'dimgrey': 0x696969, 'dodgerblue': 0x1E90FF, 'firebrick': 0xB22222,
	'floralwhite': 0xFFFAF0, 'forestgreen': 0x228B22, 'fuchsia': 0xFF00FF, 'gainsboro': 0xDCDCDC, 'ghostwhite': 0xF8F8FF, 'gold': 0xFFD700,
	'goldenrod': 0xDAA520, 'gray': 0x808080, 'green': 0x008000, 'greenyellow': 0xADFF2F, 'grey': 0x808080, 'honeydew': 0xF0FFF0, 'hotpink': 0xFF69B4,
	'indianred': 0xCD5C5C, 'indigo': 0x4B0082, 'ivory': 0xFFFFF0, 'khaki': 0xF0E68C, 'lavender': 0xE6E6FA, 'lavenderblush': 0xFFF0F5, 'lawngreen': 0x7CFC00,
	'lemonchiffon': 0xFFFACD, 'lightblue': 0xADD8E6, 'lightcoral': 0xF08080, 'lightcyan': 0xE0FFFF, 'lightgoldenrodyellow': 0xFAFAD2, 'lightgray': 0xD3D3D3,
	'lightgreen': 0x90EE90, 'lightgrey': 0xD3D3D3, 'lightpink': 0xFFB6C1, 'lightsalmon': 0xFFA07A, 'lightseagreen': 0x20B2AA, 'lightskyblue': 0x87CEFA,
	'lightslategray': 0x778899, 'lightslategrey': 0x778899, 'lightsteelblue': 0xB0C4DE, 'lightyellow': 0xFFFFE0, 'lime': 0x00FF00, 'limegreen': 0x32CD32,
	'linen': 0xFAF0E6, 'magenta': 0xFF00FF, 'maroon': 0x800000, 'mediumaquamarine': 0x66CDAA, 'mediumblue': 0x0000CD, 'mediumorchid': 0xBA55D3,
	'mediumpurple': 0x9370DB, 'mediumseagreen': 0x3CB371, 'mediumslateblue': 0x7B68EE, 'mediumspringgreen': 0x00FA9A, 'mediumturquoise': 0x48D1CC,
	'mediumvioletred': 0xC71585, 'midnightblue': 0x191970, 'mintcream': 0xF5FFFA, 'mistyrose': 0xFFE4E1, 'moccasin': 0xFFE4B5, 'navajowhite': 0xFFDEAD,
	'navy': 0x000080, 'oldlace': 0xFDF5E6, 'olive': 0x808000, 'olivedrab': 0x6B8E23, 'orange': 0xFFA500, 'orangered': 0xFF4500, 'orchid': 0xDA70D6,
	'palegoldenrod': 0xEEE8AA, 'palegreen': 0x98FB98, 'paleturquoise': 0xAFEEEE, 'palevioletred': 0xDB7093, 'papayawhip': 0xFFEFD5, 'peachpuff': 0xFFDAB9,
	'peru': 0xCD853F, 'pink': 0xFFC0CB, 'plum': 0xDDA0DD, 'powderblue': 0xB0E0E6, 'purple': 0x800080, 'red': 0xFF0000, 'rosybrown': 0xBC8F8F,
	'royalblue': 0x4169E1, 'saddlebrown': 0x8B4513, 'salmon': 0xFA8072, 'sandybrown': 0xF4A460, 'seagreen': 0x2E8B57, 'seashell': 0xFFF5EE,
	'sienna': 0xA0522D, 'silver': 0xC0C0C0, 'skyblue': 0x87CEEB, 'slateblue': 0x6A5ACD, 'slategray': 0x708090, 'slategrey': 0x708090, 'snow': 0xFFFAFA,
	'springgreen': 0x00FF7F, 'steelblue': 0x4682B4, 'tan': 0xD2B48C, 'teal': 0x008080, 'thistle': 0xD8BFD8, 'tomato': 0xFF6347, 'turquoise': 0x40E0D0,
	'violet': 0xEE82EE, 'wheat': 0xF5DEB3, 'white': 0xFFFFFF, 'whitesmoke': 0xF5F5F5, 'yellow': 0xFFFF00, 'yellowgreen': 0x9ACD32 };

/**
 * @author WestLangley / http://github.com/WestLangley
 *
 * parameters = {
 *  color: <hex>,
 *  roughness: <float>,
 *  metalness: <float>,
 *  opacity: <float>,
 *
 *  map: new THREE.Texture( <Image> ),
 *
 *  lightMap: new THREE.Texture( <Image> ),
 *  lightMapIntensity: <float>
 *
 *  aoMap: new THREE.Texture( <Image> ),
 *  aoMapIntensity: <float>
 *
 *  emissive: <hex>,
 *  emissiveIntensity: <float>
 *  emissiveMap: new THREE.Texture( <Image> ),
 *
 *  bumpMap: new THREE.Texture( <Image> ),
 *  bumpScale: <float>,
 *
 *  normalMap: new THREE.Texture( <Image> ),
 *  normalScale: <Vector2>,
 *
 *  displacementMap: new THREE.Texture( <Image> ),
 *  displacementScale: <float>,
 *  displacementBias: <float>,
 *
 *  roughnessMap: new THREE.Texture( <Image> ),
 *
 *  metalnessMap: new THREE.Texture( <Image> ),
 *
 *  alphaMap: new THREE.Texture( <Image> ),
 *
 *  envMap: new THREE.CubeTexture( [posx, negx, posy, negy, posz, negz] ),
 *  envMapIntensity: <float>
 *
 *  refractionRatio: <float>,
 *
 *  wireframe: <boolean>,
 *  wireframeLinewidth: <float>,
 *
 *  skinning: <bool>,
 *  morphTargets: <bool>,
 *  morphNormals: <bool>
 * }
 */

function MeshStandardMaterial(parameters) {

	Material.call(this);

	this.defines = { 'STANDARD': '' };

	this.type = 'MeshStandardMaterial';

	this.color = new Color(0xffffff); // diffuse
	this.roughness = 0.5;
	this.metalness = 0.5;

	this.map = null;

	this.lightMap = null;
	this.lightMapIntensity = 1.0;

	this.aoMap = null;
	this.aoMapIntensity = 1.0;

	this.emissive = new Color(0x000000);
	this.emissiveIntensity = 1.0;
	this.emissiveMap = null;

	this.bumpMap = null;
	this.bumpScale = 1;

	this.normalMap = null;
	this.normalScale = new Vector2(1, 1);

	this.displacementMap = null;
	this.displacementScale = 1;
	this.displacementBias = 0;

	this.roughnessMap = null;

	this.metalnessMap = null;

	this.alphaMap = null;

	this.envMap = null;
	this.envMapIntensity = 1.0;

	this.refractionRatio = 0.98;

	this.wireframe = false;
	this.wireframeLinewidth = 1;
	this.wireframeLinecap = 'round';
	this.wireframeLinejoin = 'round';

	this.skinning = false;
	this.morphTargets = false;
	this.morphNormals = false;

	this.setValues(parameters);
}

MeshStandardMaterial.prototype = Object.create(Material.prototype);
MeshStandardMaterial.prototype.constructor = MeshStandardMaterial;

MeshStandardMaterial.prototype.isMeshStandardMaterial = true;

MeshStandardMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.defines = { 'STANDARD': '' };

	this.color.copy(source.color);
	this.roughness = source.roughness;
	this.metalness = source.metalness;

	this.map = source.map;

	this.lightMap = source.lightMap;
	this.lightMapIntensity = source.lightMapIntensity;

	this.aoMap = source.aoMap;
	this.aoMapIntensity = source.aoMapIntensity;

	this.emissive.copy(source.emissive);
	this.emissiveMap = source.emissiveMap;
	this.emissiveIntensity = source.emissiveIntensity;

	this.bumpMap = source.bumpMap;
	this.bumpScale = source.bumpScale;

	this.normalMap = source.normalMap;
	this.normalScale.copy(source.normalScale);

	this.displacementMap = source.displacementMap;
	this.displacementScale = source.displacementScale;
	this.displacementBias = source.displacementBias;

	this.roughnessMap = source.roughnessMap;

	this.metalnessMap = source.metalnessMap;

	this.alphaMap = source.alphaMap;

	this.envMap = source.envMap;
	this.envMapIntensity = source.envMapIntensity;

	this.refractionRatio = source.refractionRatio;

	this.wireframe = source.wireframe;
	this.wireframeLinewidth = source.wireframeLinewidth;
	this.wireframeLinecap = source.wireframeLinecap;
	this.wireframeLinejoin = source.wireframeLinejoin;

	this.skinning = source.skinning;
	this.morphTargets = source.morphTargets;
	this.morphNormals = source.morphNormals;

	return this;
};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 *
 * parameters = {
 *  color: <hex>,
 *  opacity: <float>,
 *  map: new THREE.Texture( <Image> ),
 *
 *  aoMap: new THREE.Texture( <Image> ),
 *  aoMapIntensity: <float>
 *
 *  specularMap: new THREE.Texture( <Image> ),
 *
 *  alphaMap: new THREE.Texture( <Image> ),
 *
 *  envMap: new THREE.TextureCube( [posx, negx, posy, negy, posz, negz] ),
 *  combine: THREE.Multiply,
 *  reflectivity: <float>,
 *  refractionRatio: <float>,
 *
 *  shading: THREE.SmoothShading,
 *  depthTest: <bool>,
 *  depthWrite: <bool>,
 *
 *  wireframe: <boolean>,
 *  wireframeLinewidth: <float>,
 *
 *  skinning: <bool>,
 *  morphTargets: <bool>
 * }
 */

function MeshBasicMaterial(parameters) {

	Material.call(this);

	this.type = 'MeshBasicMaterial';

	this.color = new Color(0xffffff); // emissive

	this.map = null;

	this.aoMap = null;
	this.aoMapIntensity = 1.0;

	this.specularMap = null;

	this.alphaMap = null;

	this.envMap = null;
	this.combine = MultiplyOperation;
	this.reflectivity = 1;
	this.refractionRatio = 0.98;

	this.wireframe = false;
	this.wireframeLinewidth = 1;
	this.wireframeLinecap = 'round';
	this.wireframeLinejoin = 'round';

	this.skinning = false;
	this.morphTargets = false;

	this.lights = false;

	this.setValues(parameters);
}

MeshBasicMaterial.prototype = Object.create(Material.prototype);
MeshBasicMaterial.prototype.constructor = MeshBasicMaterial;

MeshBasicMaterial.prototype.isMeshBasicMaterial = true;

MeshBasicMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.color.copy(source.color);

	this.map = source.map;

	this.aoMap = source.aoMap;
	this.aoMapIntensity = source.aoMapIntensity;

	this.specularMap = source.specularMap;

	this.alphaMap = source.alphaMap;

	this.envMap = source.envMap;
	this.combine = source.combine;
	this.reflectivity = source.reflectivity;
	this.refractionRatio = source.refractionRatio;

	this.wireframe = source.wireframe;
	this.wireframeLinewidth = source.wireframeLinewidth;
	this.wireframeLinecap = source.wireframeLinecap;
	this.wireframeLinejoin = source.wireframeLinejoin;

	this.skinning = source.skinning;
	this.morphTargets = source.morphTargets;

	return this;
};

function material$1(textureDescription, options) {
  if (options === undefined && typeof textureDescription !== "string") {
    options = textureDescription;
    textureDescription = "none";
  }
  options = Object.assign({}, {
    opacity: 1,
    roughness: 0.5,
    metalness: 0,
    color: 0xffffff,
    fog: true,
    unshaded: false,
    wireframe: false,
    side: FrontSide
  }, options);

  var materialDescription = "Primrose.material(" + textureDescription + ", " + options.color + ", " + options.unshaded + ", " + options.side + ", " + options.opacity + ", " + options.roughness + ", " + options.metalness + ", " + options.color + ", " + options.emissive + ", " + options.wireframe + ", " + options.fog + ")";

  return cache(materialDescription, function () {
    var materialOptions = {
      fog: options.fog,
      transparent: options.transparent || options.opacity !== undefined && options.opacity < 1,
      opacity: options.opacity,
      side: options.side || FrontSide
    },
        MaterialType = MeshStandardMaterial;

    if (options.unshaded) {
      materialOptions.shading = FlatShading;
      MaterialType = MeshBasicMaterial;
    } else {
      materialOptions.roughness = options.roughness;
      materialOptions.metalness = options.metalness;

      if (options.emissive !== undefined) {
        materialOptions.emissive = options.emissive;
      }
    }

    var mat = new MaterialType(materialOptions);
    if (typeof options.color === "number" || options.color instanceof Number) {
      mat.color.set(options.color);
    }
    mat.wireframe = options.wireframe;
    return mat;
  });
}

/**
 * @author bhouston / http://clara.io
 * @author WestLangley / http://github.com/WestLangley
 */

function Box3(min, max) {

	this.min = min !== undefined ? min : new Vector3(+Infinity, +Infinity, +Infinity);
	this.max = max !== undefined ? max : new Vector3(-Infinity, -Infinity, -Infinity);
}

Box3.prototype = {

	constructor: Box3,

	isBox3: true,

	set: function set(min, max) {

		this.min.copy(min);
		this.max.copy(max);

		return this;
	},

	setFromArray: function setFromArray(array) {

		var minX = +Infinity;
		var minY = +Infinity;
		var minZ = +Infinity;

		var maxX = -Infinity;
		var maxY = -Infinity;
		var maxZ = -Infinity;

		for (var i = 0, l = array.length; i < l; i += 3) {

			var x = array[i];
			var y = array[i + 1];
			var z = array[i + 2];

			if (x < minX) minX = x;
			if (y < minY) minY = y;
			if (z < minZ) minZ = z;

			if (x > maxX) maxX = x;
			if (y > maxY) maxY = y;
			if (z > maxZ) maxZ = z;
		}

		this.min.set(minX, minY, minZ);
		this.max.set(maxX, maxY, maxZ);
	},

	setFromPoints: function setFromPoints(points) {

		this.makeEmpty();

		for (var i = 0, il = points.length; i < il; i++) {

			this.expandByPoint(points[i]);
		}

		return this;
	},

	setFromCenterAndSize: function () {

		var v1 = new Vector3();

		return function setFromCenterAndSize(center, size) {

			var halfSize = v1.copy(size).multiplyScalar(0.5);

			this.min.copy(center).sub(halfSize);
			this.max.copy(center).add(halfSize);

			return this;
		};
	}(),

	setFromObject: function () {

		// Computes the world-axis-aligned bounding box of an object (including its children),
		// accounting for both the object's, and children's, world transforms

		var v1 = new Vector3();

		return function setFromObject(object) {

			var scope = this;

			object.updateMatrixWorld(true);

			this.makeEmpty();

			object.traverse(function (node) {

				var geometry = node.geometry;

				if (geometry !== undefined) {

					if (geometry && geometry.isGeometry) {

						var vertices = geometry.vertices;

						for (var i = 0, il = vertices.length; i < il; i++) {

							v1.copy(vertices[i]);
							v1.applyMatrix4(node.matrixWorld);

							scope.expandByPoint(v1);
						}
					} else if (geometry && geometry.isBufferGeometry) {

						var attribute = geometry.attributes.position;

						if (attribute !== undefined) {

							var array, offset, stride;

							if (attribute && attribute.isInterleavedBufferAttribute) {

								array = attribute.data.array;
								offset = attribute.offset;
								stride = attribute.data.stride;
							} else {

								array = attribute.array;
								offset = 0;
								stride = 3;
							}

							for (var i = offset, il = array.length; i < il; i += stride) {

								v1.fromArray(array, i);
								v1.applyMatrix4(node.matrixWorld);

								scope.expandByPoint(v1);
							}
						}
					}
				}
			});

			return this;
		};
	}(),

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(box) {

		this.min.copy(box.min);
		this.max.copy(box.max);

		return this;
	},

	makeEmpty: function makeEmpty() {

		this.min.x = this.min.y = this.min.z = +Infinity;
		this.max.x = this.max.y = this.max.z = -Infinity;

		return this;
	},

	isEmpty: function isEmpty() {

		// this is a more robust check for empty than ( volume <= 0 ) because volume can get positive with two negative axes

		return this.max.x < this.min.x || this.max.y < this.min.y || this.max.z < this.min.z;
	},

	getCenter: function getCenter(optionalTarget) {

		var result = optionalTarget || new Vector3();
		return this.isEmpty() ? result.set(0, 0, 0) : result.addVectors(this.min, this.max).multiplyScalar(0.5);
	},

	getSize: function getSize(optionalTarget) {

		var result = optionalTarget || new Vector3();
		return this.isEmpty() ? result.set(0, 0, 0) : result.subVectors(this.max, this.min);
	},

	expandByPoint: function expandByPoint(point) {

		this.min.min(point);
		this.max.max(point);

		return this;
	},

	expandByVector: function expandByVector(vector) {

		this.min.sub(vector);
		this.max.add(vector);

		return this;
	},

	expandByScalar: function expandByScalar(scalar) {

		this.min.addScalar(-scalar);
		this.max.addScalar(scalar);

		return this;
	},

	containsPoint: function containsPoint(point) {

		if (point.x < this.min.x || point.x > this.max.x || point.y < this.min.y || point.y > this.max.y || point.z < this.min.z || point.z > this.max.z) {

			return false;
		}

		return true;
	},

	containsBox: function containsBox(box) {

		if (this.min.x <= box.min.x && box.max.x <= this.max.x && this.min.y <= box.min.y && box.max.y <= this.max.y && this.min.z <= box.min.z && box.max.z <= this.max.z) {

			return true;
		}

		return false;
	},

	getParameter: function getParameter(point, optionalTarget) {

		// This can potentially have a divide by zero if the box
		// has a size dimension of 0.

		var result = optionalTarget || new Vector3();

		return result.set((point.x - this.min.x) / (this.max.x - this.min.x), (point.y - this.min.y) / (this.max.y - this.min.y), (point.z - this.min.z) / (this.max.z - this.min.z));
	},

	intersectsBox: function intersectsBox(box) {

		// using 6 splitting planes to rule out intersections.

		if (box.max.x < this.min.x || box.min.x > this.max.x || box.max.y < this.min.y || box.min.y > this.max.y || box.max.z < this.min.z || box.min.z > this.max.z) {

			return false;
		}

		return true;
	},

	intersectsSphere: function () {

		var closestPoint;

		return function intersectsSphere(sphere) {

			if (closestPoint === undefined) closestPoint = new Vector3();

			// Find the point on the AABB closest to the sphere center.
			this.clampPoint(sphere.center, closestPoint);

			// If that point is inside the sphere, the AABB and sphere intersect.
			return closestPoint.distanceToSquared(sphere.center) <= sphere.radius * sphere.radius;
		};
	}(),

	intersectsPlane: function intersectsPlane(plane) {

		// We compute the minimum and maximum dot product values. If those values
		// are on the same side (back or front) of the plane, then there is no intersection.

		var min, max;

		if (plane.normal.x > 0) {

			min = plane.normal.x * this.min.x;
			max = plane.normal.x * this.max.x;
		} else {

			min = plane.normal.x * this.max.x;
			max = plane.normal.x * this.min.x;
		}

		if (plane.normal.y > 0) {

			min += plane.normal.y * this.min.y;
			max += plane.normal.y * this.max.y;
		} else {

			min += plane.normal.y * this.max.y;
			max += plane.normal.y * this.min.y;
		}

		if (plane.normal.z > 0) {

			min += plane.normal.z * this.min.z;
			max += plane.normal.z * this.max.z;
		} else {

			min += plane.normal.z * this.max.z;
			max += plane.normal.z * this.min.z;
		}

		return min <= plane.constant && max >= plane.constant;
	},

	clampPoint: function clampPoint(point, optionalTarget) {

		var result = optionalTarget || new Vector3();
		return result.copy(point).clamp(this.min, this.max);
	},

	distanceToPoint: function () {

		var v1 = new Vector3();

		return function distanceToPoint(point) {

			var clampedPoint = v1.copy(point).clamp(this.min, this.max);
			return clampedPoint.sub(point).length();
		};
	}(),

	getBoundingSphere: function () {

		var v1 = new Vector3();

		return function getBoundingSphere(optionalTarget) {

			var result = optionalTarget || new Sphere();

			this.getCenter(result.center);

			result.radius = this.getSize(v1).length() * 0.5;

			return result;
		};
	}(),

	intersect: function intersect(box) {

		this.min.max(box.min);
		this.max.min(box.max);

		// ensure that if there is no overlap, the result is fully empty, not slightly empty with non-inf/+inf values that will cause subsequence intersects to erroneously return valid values.
		if (this.isEmpty()) this.makeEmpty();

		return this;
	},

	union: function union(box) {

		this.min.min(box.min);
		this.max.max(box.max);

		return this;
	},

	applyMatrix4: function () {

		var points = [new Vector3(), new Vector3(), new Vector3(), new Vector3(), new Vector3(), new Vector3(), new Vector3(), new Vector3()];

		return function applyMatrix4(matrix) {

			// transform of empty box is an empty box.
			if (this.isEmpty()) return this;

			// NOTE: I am using a binary pattern to specify all 2^3 combinations below
			points[0].set(this.min.x, this.min.y, this.min.z).applyMatrix4(matrix); // 000
			points[1].set(this.min.x, this.min.y, this.max.z).applyMatrix4(matrix); // 001
			points[2].set(this.min.x, this.max.y, this.min.z).applyMatrix4(matrix); // 010
			points[3].set(this.min.x, this.max.y, this.max.z).applyMatrix4(matrix); // 011
			points[4].set(this.max.x, this.min.y, this.min.z).applyMatrix4(matrix); // 100
			points[5].set(this.max.x, this.min.y, this.max.z).applyMatrix4(matrix); // 101
			points[6].set(this.max.x, this.max.y, this.min.z).applyMatrix4(matrix); // 110
			points[7].set(this.max.x, this.max.y, this.max.z).applyMatrix4(matrix); // 111

			this.setFromPoints(points);

			return this;
		};
	}(),

	translate: function translate(offset) {

		this.min.add(offset);
		this.max.add(offset);

		return this;
	},

	equals: function equals(box) {

		return box.min.equals(this.min) && box.max.equals(this.max);
	}

};

/**
 * @author bhouston / http://clara.io
 * @author mrdoob / http://mrdoob.com/
 */

function Sphere(center, radius) {

	this.center = center !== undefined ? center : new Vector3();
	this.radius = radius !== undefined ? radius : 0;
}

Sphere.prototype = {

	constructor: Sphere,

	set: function set(center, radius) {

		this.center.copy(center);
		this.radius = radius;

		return this;
	},

	setFromPoints: function () {

		var box = new Box3();

		return function setFromPoints(points, optionalCenter) {

			var center = this.center;

			if (optionalCenter !== undefined) {

				center.copy(optionalCenter);
			} else {

				box.setFromPoints(points).getCenter(center);
			}

			var maxRadiusSq = 0;

			for (var i = 0, il = points.length; i < il; i++) {

				maxRadiusSq = Math.max(maxRadiusSq, center.distanceToSquared(points[i]));
			}

			this.radius = Math.sqrt(maxRadiusSq);

			return this;
		};
	}(),

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(sphere) {

		this.center.copy(sphere.center);
		this.radius = sphere.radius;

		return this;
	},

	empty: function empty() {

		return this.radius <= 0;
	},

	containsPoint: function containsPoint(point) {

		return point.distanceToSquared(this.center) <= this.radius * this.radius;
	},

	distanceToPoint: function distanceToPoint(point) {

		return point.distanceTo(this.center) - this.radius;
	},

	intersectsSphere: function intersectsSphere(sphere) {

		var radiusSum = this.radius + sphere.radius;

		return sphere.center.distanceToSquared(this.center) <= radiusSum * radiusSum;
	},

	intersectsBox: function intersectsBox(box) {

		return box.intersectsSphere(this);
	},

	intersectsPlane: function intersectsPlane(plane) {

		// We use the following equation to compute the signed distance from
		// the center of the sphere to the plane.
		//
		// distance = q * n - d
		//
		// If this distance is greater than the radius of the sphere,
		// then there is no intersection.

		return Math.abs(this.center.dot(plane.normal) - plane.constant) <= this.radius;
	},

	clampPoint: function clampPoint(point, optionalTarget) {

		var deltaLengthSq = this.center.distanceToSquared(point);

		var result = optionalTarget || new Vector3();

		result.copy(point);

		if (deltaLengthSq > this.radius * this.radius) {

			result.sub(this.center).normalize();
			result.multiplyScalar(this.radius).add(this.center);
		}

		return result;
	},

	getBoundingBox: function getBoundingBox(optionalTarget) {

		var box = optionalTarget || new Box3();

		box.set(this.center, this.center);
		box.expandByScalar(this.radius);

		return box;
	},

	applyMatrix4: function applyMatrix4(matrix) {

		this.center.applyMatrix4(matrix);
		this.radius = this.radius * matrix.getMaxScaleOnAxis();

		return this;
	},

	translate: function translate(offset) {

		this.center.add(offset);

		return this;
	},

	equals: function equals(sphere) {

		return sphere.center.equals(this.center) && sphere.radius === this.radius;
	}

};

/**
 * @author bhouston / http://clara.io
 */

function Ray(origin, direction) {

	this.origin = origin !== undefined ? origin : new Vector3();
	this.direction = direction !== undefined ? direction : new Vector3();
}

Ray.prototype = {

	constructor: Ray,

	set: function set(origin, direction) {

		this.origin.copy(origin);
		this.direction.copy(direction);

		return this;
	},

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(ray) {

		this.origin.copy(ray.origin);
		this.direction.copy(ray.direction);

		return this;
	},

	at: function at(t, optionalTarget) {

		var result = optionalTarget || new Vector3();

		return result.copy(this.direction).multiplyScalar(t).add(this.origin);
	},

	lookAt: function lookAt(v) {

		this.direction.copy(v).sub(this.origin).normalize();

		return this;
	},

	recast: function () {

		var v1 = new Vector3();

		return function recast(t) {

			this.origin.copy(this.at(t, v1));

			return this;
		};
	}(),

	closestPointToPoint: function closestPointToPoint(point, optionalTarget) {

		var result = optionalTarget || new Vector3();
		result.subVectors(point, this.origin);
		var directionDistance = result.dot(this.direction);

		if (directionDistance < 0) {

			return result.copy(this.origin);
		}

		return result.copy(this.direction).multiplyScalar(directionDistance).add(this.origin);
	},

	distanceToPoint: function distanceToPoint(point) {

		return Math.sqrt(this.distanceSqToPoint(point));
	},

	distanceSqToPoint: function () {

		var v1 = new Vector3();

		return function distanceSqToPoint(point) {

			var directionDistance = v1.subVectors(point, this.origin).dot(this.direction);

			// point behind the ray

			if (directionDistance < 0) {

				return this.origin.distanceToSquared(point);
			}

			v1.copy(this.direction).multiplyScalar(directionDistance).add(this.origin);

			return v1.distanceToSquared(point);
		};
	}(),

	distanceSqToSegment: function () {

		var segCenter = new Vector3();
		var segDir = new Vector3();
		var diff = new Vector3();

		return function distanceSqToSegment(v0, v1, optionalPointOnRay, optionalPointOnSegment) {

			// from http://www.geometrictools.com/GTEngine/Include/Mathematics/GteDistRaySegment.h
			// It returns the min distance between the ray and the segment
			// defined by v0 and v1
			// It can also set two optional targets :
			// - The closest point on the ray
			// - The closest point on the segment

			segCenter.copy(v0).add(v1).multiplyScalar(0.5);
			segDir.copy(v1).sub(v0).normalize();
			diff.copy(this.origin).sub(segCenter);

			var segExtent = v0.distanceTo(v1) * 0.5;
			var a01 = -this.direction.dot(segDir);
			var b0 = diff.dot(this.direction);
			var b1 = -diff.dot(segDir);
			var c = diff.lengthSq();
			var det = Math.abs(1 - a01 * a01);
			var s0, s1, sqrDist, extDet;

			if (det > 0) {

				// The ray and segment are not parallel.

				s0 = a01 * b1 - b0;
				s1 = a01 * b0 - b1;
				extDet = segExtent * det;

				if (s0 >= 0) {

					if (s1 >= -extDet) {

						if (s1 <= extDet) {

							// region 0
							// Minimum at interior points of ray and segment.

							var invDet = 1 / det;
							s0 *= invDet;
							s1 *= invDet;
							sqrDist = s0 * (s0 + a01 * s1 + 2 * b0) + s1 * (a01 * s0 + s1 + 2 * b1) + c;
						} else {

							// region 1

							s1 = segExtent;
							s0 = Math.max(0, -(a01 * s1 + b0));
							sqrDist = -s0 * s0 + s1 * (s1 + 2 * b1) + c;
						}
					} else {

						// region 5

						s1 = -segExtent;
						s0 = Math.max(0, -(a01 * s1 + b0));
						sqrDist = -s0 * s0 + s1 * (s1 + 2 * b1) + c;
					}
				} else {

					if (s1 <= -extDet) {

						// region 4

						s0 = Math.max(0, -(-a01 * segExtent + b0));
						s1 = s0 > 0 ? -segExtent : Math.min(Math.max(-segExtent, -b1), segExtent);
						sqrDist = -s0 * s0 + s1 * (s1 + 2 * b1) + c;
					} else if (s1 <= extDet) {

						// region 3

						s0 = 0;
						s1 = Math.min(Math.max(-segExtent, -b1), segExtent);
						sqrDist = s1 * (s1 + 2 * b1) + c;
					} else {

						// region 2

						s0 = Math.max(0, -(a01 * segExtent + b0));
						s1 = s0 > 0 ? segExtent : Math.min(Math.max(-segExtent, -b1), segExtent);
						sqrDist = -s0 * s0 + s1 * (s1 + 2 * b1) + c;
					}
				}
			} else {

				// Ray and segment are parallel.

				s1 = a01 > 0 ? -segExtent : segExtent;
				s0 = Math.max(0, -(a01 * s1 + b0));
				sqrDist = -s0 * s0 + s1 * (s1 + 2 * b1) + c;
			}

			if (optionalPointOnRay) {

				optionalPointOnRay.copy(this.direction).multiplyScalar(s0).add(this.origin);
			}

			if (optionalPointOnSegment) {

				optionalPointOnSegment.copy(segDir).multiplyScalar(s1).add(segCenter);
			}

			return sqrDist;
		};
	}(),

	intersectSphere: function () {

		var v1 = new Vector3();

		return function intersectSphere(sphere, optionalTarget) {

			v1.subVectors(sphere.center, this.origin);
			var tca = v1.dot(this.direction);
			var d2 = v1.dot(v1) - tca * tca;
			var radius2 = sphere.radius * sphere.radius;

			if (d2 > radius2) return null;

			var thc = Math.sqrt(radius2 - d2);

			// t0 = first intersect point - entrance on front of sphere
			var t0 = tca - thc;

			// t1 = second intersect point - exit point on back of sphere
			var t1 = tca + thc;

			// test to see if both t0 and t1 are behind the ray - if so, return null
			if (t0 < 0 && t1 < 0) return null;

			// test to see if t0 is behind the ray:
			// if it is, the ray is inside the sphere, so return the second exit point scaled by t1,
			// in order to always return an intersect point that is in front of the ray.
			if (t0 < 0) return this.at(t1, optionalTarget);

			// else t0 is in front of the ray, so return the first collision point scaled by t0
			return this.at(t0, optionalTarget);
		};
	}(),

	intersectsSphere: function intersectsSphere(sphere) {

		return this.distanceToPoint(sphere.center) <= sphere.radius;
	},

	distanceToPlane: function distanceToPlane(plane) {

		var denominator = plane.normal.dot(this.direction);

		if (denominator === 0) {

			// line is coplanar, return origin
			if (plane.distanceToPoint(this.origin) === 0) {

				return 0;
			}

			// Null is preferable to undefined since undefined means.... it is undefined

			return null;
		}

		var t = -(this.origin.dot(plane.normal) + plane.constant) / denominator;

		// Return if the ray never intersects the plane

		return t >= 0 ? t : null;
	},

	intersectPlane: function intersectPlane(plane, optionalTarget) {

		var t = this.distanceToPlane(plane);

		if (t === null) {

			return null;
		}

		return this.at(t, optionalTarget);
	},

	intersectsPlane: function intersectsPlane(plane) {

		// check if the ray lies on the plane first

		var distToPoint = plane.distanceToPoint(this.origin);

		if (distToPoint === 0) {

			return true;
		}

		var denominator = plane.normal.dot(this.direction);

		if (denominator * distToPoint < 0) {

			return true;
		}

		// ray origin is behind the plane (and is pointing behind it)

		return false;
	},

	intersectBox: function intersectBox(box, optionalTarget) {

		var tmin, tmax, tymin, tymax, tzmin, tzmax;

		var invdirx = 1 / this.direction.x,
		    invdiry = 1 / this.direction.y,
		    invdirz = 1 / this.direction.z;

		var origin = this.origin;

		if (invdirx >= 0) {

			tmin = (box.min.x - origin.x) * invdirx;
			tmax = (box.max.x - origin.x) * invdirx;
		} else {

			tmin = (box.max.x - origin.x) * invdirx;
			tmax = (box.min.x - origin.x) * invdirx;
		}

		if (invdiry >= 0) {

			tymin = (box.min.y - origin.y) * invdiry;
			tymax = (box.max.y - origin.y) * invdiry;
		} else {

			tymin = (box.max.y - origin.y) * invdiry;
			tymax = (box.min.y - origin.y) * invdiry;
		}

		if (tmin > tymax || tymin > tmax) return null;

		// These lines also handle the case where tmin or tmax is NaN
		// (result of 0 * Infinity). x !== x returns true if x is NaN

		if (tymin > tmin || tmin !== tmin) tmin = tymin;

		if (tymax < tmax || tmax !== tmax) tmax = tymax;

		if (invdirz >= 0) {

			tzmin = (box.min.z - origin.z) * invdirz;
			tzmax = (box.max.z - origin.z) * invdirz;
		} else {

			tzmin = (box.max.z - origin.z) * invdirz;
			tzmax = (box.min.z - origin.z) * invdirz;
		}

		if (tmin > tzmax || tzmin > tmax) return null;

		if (tzmin > tmin || tmin !== tmin) tmin = tzmin;

		if (tzmax < tmax || tmax !== tmax) tmax = tzmax;

		//return point closest to the ray (positive side)

		if (tmax < 0) return null;

		return this.at(tmin >= 0 ? tmin : tmax, optionalTarget);
	},

	intersectsBox: function () {

		var v = new Vector3();

		return function intersectsBox(box) {

			return this.intersectBox(box, v) !== null;
		};
	}(),

	intersectTriangle: function () {

		// Compute the offset origin, edges, and normal.
		var diff = new Vector3();
		var edge1 = new Vector3();
		var edge2 = new Vector3();
		var normal = new Vector3();

		return function intersectTriangle(a, b, c, backfaceCulling, optionalTarget) {

			// from http://www.geometrictools.com/GTEngine/Include/Mathematics/GteIntrRay3Triangle3.h

			edge1.subVectors(b, a);
			edge2.subVectors(c, a);
			normal.crossVectors(edge1, edge2);

			// Solve Q + t*D = b1*E1 + b2*E2 (Q = kDiff, D = ray direction,
			// E1 = kEdge1, E2 = kEdge2, N = Cross(E1,E2)) by
			//   |Dot(D,N)|*b1 = sign(Dot(D,N))*Dot(D,Cross(Q,E2))
			//   |Dot(D,N)|*b2 = sign(Dot(D,N))*Dot(D,Cross(E1,Q))
			//   |Dot(D,N)|*t = -sign(Dot(D,N))*Dot(Q,N)
			var DdN = this.direction.dot(normal);
			var sign;

			if (DdN > 0) {

				if (backfaceCulling) return null;
				sign = 1;
			} else if (DdN < 0) {

				sign = -1;
				DdN = -DdN;
			} else {

				return null;
			}

			diff.subVectors(this.origin, a);
			var DdQxE2 = sign * this.direction.dot(edge2.crossVectors(diff, edge2));

			// b1 < 0, no intersection
			if (DdQxE2 < 0) {

				return null;
			}

			var DdE1xQ = sign * this.direction.dot(edge1.cross(diff));

			// b2 < 0, no intersection
			if (DdE1xQ < 0) {

				return null;
			}

			// b1+b2 > 1, no intersection
			if (DdQxE2 + DdE1xQ > DdN) {

				return null;
			}

			// Line intersects triangle, check if ray does.
			var QdN = -sign * diff.dot(normal);

			// t < 0, no intersection
			if (QdN < 0) {

				return null;
			}

			// Ray intersects triangle.
			return this.at(QdN / DdN, optionalTarget);
		};
	}(),

	applyMatrix4: function applyMatrix4(matrix4) {

		this.direction.add(this.origin).applyMatrix4(matrix4);
		this.origin.applyMatrix4(matrix4);
		this.direction.sub(this.origin);
		this.direction.normalize();

		return this;
	},

	equals: function equals(ray) {

		return ray.origin.equals(this.origin) && ray.direction.equals(this.direction);
	}

};

/**
 * @author bhouston / http://clara.io
 */

function Line3(start, end) {

	this.start = start !== undefined ? start : new Vector3();
	this.end = end !== undefined ? end : new Vector3();
}

Line3.prototype = {

	constructor: Line3,

	set: function set(start, end) {

		this.start.copy(start);
		this.end.copy(end);

		return this;
	},

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(line) {

		this.start.copy(line.start);
		this.end.copy(line.end);

		return this;
	},

	getCenter: function getCenter(optionalTarget) {

		var result = optionalTarget || new Vector3();
		return result.addVectors(this.start, this.end).multiplyScalar(0.5);
	},

	delta: function delta(optionalTarget) {

		var result = optionalTarget || new Vector3();
		return result.subVectors(this.end, this.start);
	},

	distanceSq: function distanceSq() {

		return this.start.distanceToSquared(this.end);
	},

	distance: function distance() {

		return this.start.distanceTo(this.end);
	},

	at: function at(t, optionalTarget) {

		var result = optionalTarget || new Vector3();

		return this.delta(result).multiplyScalar(t).add(this.start);
	},

	closestPointToPointParameter: function () {

		var startP = new Vector3();
		var startEnd = new Vector3();

		return function closestPointToPointParameter(point, clampToLine) {

			startP.subVectors(point, this.start);
			startEnd.subVectors(this.end, this.start);

			var startEnd2 = startEnd.dot(startEnd);
			var startEnd_startP = startEnd.dot(startP);

			var t = startEnd_startP / startEnd2;

			if (clampToLine) {

				t = _Math.clamp(t, 0, 1);
			}

			return t;
		};
	}(),

	closestPointToPoint: function closestPointToPoint(point, clampToLine, optionalTarget) {

		var t = this.closestPointToPointParameter(point, clampToLine);

		var result = optionalTarget || new Vector3();

		return this.delta(result).multiplyScalar(t).add(this.start);
	},

	applyMatrix4: function applyMatrix4(matrix) {

		this.start.applyMatrix4(matrix);
		this.end.applyMatrix4(matrix);

		return this;
	},

	equals: function equals(line) {

		return line.start.equals(this.start) && line.end.equals(this.end);
	}

};

/**
 * @author bhouston / http://clara.io
 */

function Plane(normal, constant) {

	this.normal = normal !== undefined ? normal : new Vector3(1, 0, 0);
	this.constant = constant !== undefined ? constant : 0;
}

Plane.prototype = {

	constructor: Plane,

	set: function set(normal, constant) {

		this.normal.copy(normal);
		this.constant = constant;

		return this;
	},

	setComponents: function setComponents(x, y, z, w) {

		this.normal.set(x, y, z);
		this.constant = w;

		return this;
	},

	setFromNormalAndCoplanarPoint: function setFromNormalAndCoplanarPoint(normal, point) {

		this.normal.copy(normal);
		this.constant = -point.dot(this.normal); // must be this.normal, not normal, as this.normal is normalized

		return this;
	},

	setFromCoplanarPoints: function () {

		var v1 = new Vector3();
		var v2 = new Vector3();

		return function setFromCoplanarPoints(a, b, c) {

			var normal = v1.subVectors(c, b).cross(v2.subVectors(a, b)).normalize();

			// Q: should an error be thrown if normal is zero (e.g. degenerate plane)?

			this.setFromNormalAndCoplanarPoint(normal, a);

			return this;
		};
	}(),

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(plane) {

		this.normal.copy(plane.normal);
		this.constant = plane.constant;

		return this;
	},

	normalize: function normalize() {

		// Note: will lead to a divide by zero if the plane is invalid.

		var inverseNormalLength = 1.0 / this.normal.length();
		this.normal.multiplyScalar(inverseNormalLength);
		this.constant *= inverseNormalLength;

		return this;
	},

	negate: function negate() {

		this.constant *= -1;
		this.normal.negate();

		return this;
	},

	distanceToPoint: function distanceToPoint(point) {

		return this.normal.dot(point) + this.constant;
	},

	distanceToSphere: function distanceToSphere(sphere) {

		return this.distanceToPoint(sphere.center) - sphere.radius;
	},

	projectPoint: function projectPoint(point, optionalTarget) {

		return this.orthoPoint(point, optionalTarget).sub(point).negate();
	},

	orthoPoint: function orthoPoint(point, optionalTarget) {

		var perpendicularMagnitude = this.distanceToPoint(point);

		var result = optionalTarget || new Vector3();
		return result.copy(this.normal).multiplyScalar(perpendicularMagnitude);
	},

	intersectLine: function () {

		var v1 = new Vector3();

		return function intersectLine(line, optionalTarget) {

			var result = optionalTarget || new Vector3();

			var direction = line.delta(v1);

			var denominator = this.normal.dot(direction);

			if (denominator === 0) {

				// line is coplanar, return origin
				if (this.distanceToPoint(line.start) === 0) {

					return result.copy(line.start);
				}

				// Unsure if this is the correct method to handle this case.
				return undefined;
			}

			var t = -(line.start.dot(this.normal) + this.constant) / denominator;

			if (t < 0 || t > 1) {

				return undefined;
			}

			return result.copy(direction).multiplyScalar(t).add(line.start);
		};
	}(),

	intersectsLine: function intersectsLine(line) {

		// Note: this tests if a line intersects the plane, not whether it (or its end-points) are coplanar with it.

		var startSign = this.distanceToPoint(line.start);
		var endSign = this.distanceToPoint(line.end);

		return startSign < 0 && endSign > 0 || endSign < 0 && startSign > 0;
	},

	intersectsBox: function intersectsBox(box) {

		return box.intersectsPlane(this);
	},

	intersectsSphere: function intersectsSphere(sphere) {

		return sphere.intersectsPlane(this);
	},

	coplanarPoint: function coplanarPoint(optionalTarget) {

		var result = optionalTarget || new Vector3();
		return result.copy(this.normal).multiplyScalar(-this.constant);
	},

	applyMatrix4: function () {

		var v1 = new Vector3();
		var m1 = new Matrix3();

		return function applyMatrix4(matrix, optionalNormalMatrix) {

			var referencePoint = this.coplanarPoint(v1).applyMatrix4(matrix);

			// transform normal based on theory here:
			// http://www.songho.ca/opengl/gl_normaltransform.html
			var normalMatrix = optionalNormalMatrix || m1.getNormalMatrix(matrix);
			var normal = this.normal.applyMatrix3(normalMatrix).normalize();

			// recalculate constant (like in setFromNormalAndCoplanarPoint)
			this.constant = -referencePoint.dot(normal);

			return this;
		};
	}(),

	translate: function translate(offset) {

		this.constant = this.constant - offset.dot(this.normal);

		return this;
	},

	equals: function equals(plane) {

		return plane.normal.equals(this.normal) && plane.constant === this.constant;
	}

};

/**
 * @author bhouston / http://clara.io
 * @author mrdoob / http://mrdoob.com/
 */

function Triangle(a, b, c) {

	this.a = a !== undefined ? a : new Vector3();
	this.b = b !== undefined ? b : new Vector3();
	this.c = c !== undefined ? c : new Vector3();
}

Triangle.normal = function () {

	var v0 = new Vector3();

	return function normal(a, b, c, optionalTarget) {

		var result = optionalTarget || new Vector3();

		result.subVectors(c, b);
		v0.subVectors(a, b);
		result.cross(v0);

		var resultLengthSq = result.lengthSq();
		if (resultLengthSq > 0) {

			return result.multiplyScalar(1 / Math.sqrt(resultLengthSq));
		}

		return result.set(0, 0, 0);
	};
}();

// static/instance method to calculate barycentric coordinates
// based on: http://www.blackpawn.com/texts/pointinpoly/default.html
Triangle.barycoordFromPoint = function () {

	var v0 = new Vector3();
	var v1 = new Vector3();
	var v2 = new Vector3();

	return function barycoordFromPoint(point, a, b, c, optionalTarget) {

		v0.subVectors(c, a);
		v1.subVectors(b, a);
		v2.subVectors(point, a);

		var dot00 = v0.dot(v0);
		var dot01 = v0.dot(v1);
		var dot02 = v0.dot(v2);
		var dot11 = v1.dot(v1);
		var dot12 = v1.dot(v2);

		var denom = dot00 * dot11 - dot01 * dot01;

		var result = optionalTarget || new Vector3();

		// collinear or singular triangle
		if (denom === 0) {

			// arbitrary location outside of triangle?
			// not sure if this is the best idea, maybe should be returning undefined
			return result.set(-2, -1, -1);
		}

		var invDenom = 1 / denom;
		var u = (dot11 * dot02 - dot01 * dot12) * invDenom;
		var v = (dot00 * dot12 - dot01 * dot02) * invDenom;

		// barycentric coordinates must always sum to 1
		return result.set(1 - u - v, v, u);
	};
}();

Triangle.containsPoint = function () {

	var v1 = new Vector3();

	return function containsPoint(point, a, b, c) {

		var result = Triangle.barycoordFromPoint(point, a, b, c, v1);

		return result.x >= 0 && result.y >= 0 && result.x + result.y <= 1;
	};
}();

Triangle.prototype = {

	constructor: Triangle,

	set: function set(a, b, c) {

		this.a.copy(a);
		this.b.copy(b);
		this.c.copy(c);

		return this;
	},

	setFromPointsAndIndices: function setFromPointsAndIndices(points, i0, i1, i2) {

		this.a.copy(points[i0]);
		this.b.copy(points[i1]);
		this.c.copy(points[i2]);

		return this;
	},

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(triangle) {

		this.a.copy(triangle.a);
		this.b.copy(triangle.b);
		this.c.copy(triangle.c);

		return this;
	},

	area: function () {

		var v0 = new Vector3();
		var v1 = new Vector3();

		return function area() {

			v0.subVectors(this.c, this.b);
			v1.subVectors(this.a, this.b);

			return v0.cross(v1).length() * 0.5;
		};
	}(),

	midpoint: function midpoint(optionalTarget) {

		var result = optionalTarget || new Vector3();
		return result.addVectors(this.a, this.b).add(this.c).multiplyScalar(1 / 3);
	},

	normal: function normal(optionalTarget) {

		return Triangle.normal(this.a, this.b, this.c, optionalTarget);
	},

	plane: function plane(optionalTarget) {

		var result = optionalTarget || new Plane();

		return result.setFromCoplanarPoints(this.a, this.b, this.c);
	},

	barycoordFromPoint: function barycoordFromPoint(point, optionalTarget) {

		return Triangle.barycoordFromPoint(point, this.a, this.b, this.c, optionalTarget);
	},

	containsPoint: function containsPoint(point) {

		return Triangle.containsPoint(point, this.a, this.b, this.c);
	},

	closestPointToPoint: function () {

		var plane, edgeList, projectedPoint, closestPoint;

		return function closestPointToPoint(point, optionalTarget) {

			if (plane === undefined) {

				plane = new Plane();
				edgeList = [new Line3(), new Line3(), new Line3()];
				projectedPoint = new Vector3();
				closestPoint = new Vector3();
			}

			var result = optionalTarget || new Vector3();
			var minDistance = Infinity;

			// project the point onto the plane of the triangle

			plane.setFromCoplanarPoints(this.a, this.b, this.c);
			plane.projectPoint(point, projectedPoint);

			// check if the projection lies within the triangle

			if (this.containsPoint(projectedPoint) === true) {

				// if so, this is the closest point

				result.copy(projectedPoint);
			} else {

				// if not, the point falls outside the triangle. the result is the closest point to the triangle's edges or vertices

				edgeList[0].set(this.a, this.b);
				edgeList[1].set(this.b, this.c);
				edgeList[2].set(this.c, this.a);

				for (var i = 0; i < edgeList.length; i++) {

					edgeList[i].closestPointToPoint(projectedPoint, true, closestPoint);

					var distance = projectedPoint.distanceToSquared(closestPoint);

					if (distance < minDistance) {

						minDistance = distance;

						result.copy(closestPoint);
					}
				}
			}

			return result;
		};
	}(),

	equals: function equals(triangle) {

		return triangle.a.equals(this.a) && triangle.b.equals(this.b) && triangle.c.equals(this.c);
	}

};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 */

function Face3(a, b, c, normal, color, materialIndex) {

	this.a = a;
	this.b = b;
	this.c = c;

	this.normal = normal && normal.isVector3 ? normal : new Vector3();
	this.vertexNormals = Array.isArray(normal) ? normal : [];

	this.color = color && color.isColor ? color : new Color();
	this.vertexColors = Array.isArray(color) ? color : [];

	this.materialIndex = materialIndex !== undefined ? materialIndex : 0;
}

Face3.prototype = {

	constructor: Face3,

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(source) {

		this.a = source.a;
		this.b = source.b;
		this.c = source.c;

		this.normal.copy(source.normal);
		this.color.copy(source.color);

		this.materialIndex = source.materialIndex;

		for (var i = 0, il = source.vertexNormals.length; i < il; i++) {

			this.vertexNormals[i] = source.vertexNormals[i].clone();
		}

		for (var i = 0, il = source.vertexColors.length; i < il; i++) {

			this.vertexColors[i] = source.vertexColors[i].clone();
		}

		return this;
	}

};

/**
 * @author supereggbert / http://www.paulbrunt.co.uk/
 * @author philogb / http://blog.thejit.org/
 * @author mikael emtinger / http://gomo.se/
 * @author egraether / http://egraether.com/
 * @author WestLangley / http://github.com/WestLangley
 */

function Vector4(x, y, z, w) {

	this.x = x || 0;
	this.y = y || 0;
	this.z = z || 0;
	this.w = w !== undefined ? w : 1;
}

Vector4.prototype = {

	constructor: Vector4,

	isVector4: true,

	set: function set(x, y, z, w) {

		this.x = x;
		this.y = y;
		this.z = z;
		this.w = w;

		return this;
	},

	setScalar: function setScalar(scalar) {

		this.x = scalar;
		this.y = scalar;
		this.z = scalar;
		this.w = scalar;

		return this;
	},

	setX: function setX(x) {

		this.x = x;

		return this;
	},

	setY: function setY(y) {

		this.y = y;

		return this;
	},

	setZ: function setZ(z) {

		this.z = z;

		return this;
	},

	setW: function setW(w) {

		this.w = w;

		return this;
	},

	setComponent: function setComponent(index, value) {

		switch (index) {

			case 0:
				this.x = value;break;
			case 1:
				this.y = value;break;
			case 2:
				this.z = value;break;
			case 3:
				this.w = value;break;
			default:
				throw new Error('index is out of range: ' + index);

		}

		return this;
	},

	getComponent: function getComponent(index) {

		switch (index) {

			case 0:
				return this.x;
			case 1:
				return this.y;
			case 2:
				return this.z;
			case 3:
				return this.w;
			default:
				throw new Error('index is out of range: ' + index);

		}
	},

	clone: function clone() {

		return new this.constructor(this.x, this.y, this.z, this.w);
	},

	copy: function copy(v) {

		this.x = v.x;
		this.y = v.y;
		this.z = v.z;
		this.w = v.w !== undefined ? v.w : 1;

		return this;
	},

	add: function add(v, w) {

		if (w !== undefined) {

			console.warn('THREE.Vector4: .add() now only accepts one argument. Use .addVectors( a, b ) instead.');
			return this.addVectors(v, w);
		}

		this.x += v.x;
		this.y += v.y;
		this.z += v.z;
		this.w += v.w;

		return this;
	},

	addScalar: function addScalar(s) {

		this.x += s;
		this.y += s;
		this.z += s;
		this.w += s;

		return this;
	},

	addVectors: function addVectors(a, b) {

		this.x = a.x + b.x;
		this.y = a.y + b.y;
		this.z = a.z + b.z;
		this.w = a.w + b.w;

		return this;
	},

	addScaledVector: function addScaledVector(v, s) {

		this.x += v.x * s;
		this.y += v.y * s;
		this.z += v.z * s;
		this.w += v.w * s;

		return this;
	},

	sub: function sub(v, w) {

		if (w !== undefined) {

			console.warn('THREE.Vector4: .sub() now only accepts one argument. Use .subVectors( a, b ) instead.');
			return this.subVectors(v, w);
		}

		this.x -= v.x;
		this.y -= v.y;
		this.z -= v.z;
		this.w -= v.w;

		return this;
	},

	subScalar: function subScalar(s) {

		this.x -= s;
		this.y -= s;
		this.z -= s;
		this.w -= s;

		return this;
	},

	subVectors: function subVectors(a, b) {

		this.x = a.x - b.x;
		this.y = a.y - b.y;
		this.z = a.z - b.z;
		this.w = a.w - b.w;

		return this;
	},

	multiplyScalar: function multiplyScalar(scalar) {

		if (isFinite(scalar)) {

			this.x *= scalar;
			this.y *= scalar;
			this.z *= scalar;
			this.w *= scalar;
		} else {

			this.x = 0;
			this.y = 0;
			this.z = 0;
			this.w = 0;
		}

		return this;
	},

	applyMatrix4: function applyMatrix4(m) {

		var x = this.x,
		    y = this.y,
		    z = this.z,
		    w = this.w;
		var e = m.elements;

		this.x = e[0] * x + e[4] * y + e[8] * z + e[12] * w;
		this.y = e[1] * x + e[5] * y + e[9] * z + e[13] * w;
		this.z = e[2] * x + e[6] * y + e[10] * z + e[14] * w;
		this.w = e[3] * x + e[7] * y + e[11] * z + e[15] * w;

		return this;
	},

	divideScalar: function divideScalar(scalar) {

		return this.multiplyScalar(1 / scalar);
	},

	setAxisAngleFromQuaternion: function setAxisAngleFromQuaternion(q) {

		// http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToAngle/index.htm

		// q is assumed to be normalized

		this.w = 2 * Math.acos(q.w);

		var s = Math.sqrt(1 - q.w * q.w);

		if (s < 0.0001) {

			this.x = 1;
			this.y = 0;
			this.z = 0;
		} else {

			this.x = q.x / s;
			this.y = q.y / s;
			this.z = q.z / s;
		}

		return this;
	},

	setAxisAngleFromRotationMatrix: function setAxisAngleFromRotationMatrix(m) {

		// http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToAngle/index.htm

		// assumes the upper 3x3 of m is a pure rotation matrix (i.e, unscaled)

		var angle,
		    x,
		    y,
		    z,
		    // variables for result
		epsilon = 0.01,
		    // margin to allow for rounding errors
		epsilon2 = 0.1,
		    // margin to distinguish between 0 and 180 degrees

		te = m.elements,
		    m11 = te[0],
		    m12 = te[4],
		    m13 = te[8],
		    m21 = te[1],
		    m22 = te[5],
		    m23 = te[9],
		    m31 = te[2],
		    m32 = te[6],
		    m33 = te[10];

		if (Math.abs(m12 - m21) < epsilon && Math.abs(m13 - m31) < epsilon && Math.abs(m23 - m32) < epsilon) {

			// singularity found
			// first check for identity matrix which must have +1 for all terms
			// in leading diagonal and zero in other terms

			if (Math.abs(m12 + m21) < epsilon2 && Math.abs(m13 + m31) < epsilon2 && Math.abs(m23 + m32) < epsilon2 && Math.abs(m11 + m22 + m33 - 3) < epsilon2) {

				// this singularity is identity matrix so angle = 0

				this.set(1, 0, 0, 0);

				return this; // zero angle, arbitrary axis
			}

			// otherwise this singularity is angle = 180

			angle = Math.PI;

			var xx = (m11 + 1) / 2;
			var yy = (m22 + 1) / 2;
			var zz = (m33 + 1) / 2;
			var xy = (m12 + m21) / 4;
			var xz = (m13 + m31) / 4;
			var yz = (m23 + m32) / 4;

			if (xx > yy && xx > zz) {

				// m11 is the largest diagonal term

				if (xx < epsilon) {

					x = 0;
					y = 0.707106781;
					z = 0.707106781;
				} else {

					x = Math.sqrt(xx);
					y = xy / x;
					z = xz / x;
				}
			} else if (yy > zz) {

				// m22 is the largest diagonal term

				if (yy < epsilon) {

					x = 0.707106781;
					y = 0;
					z = 0.707106781;
				} else {

					y = Math.sqrt(yy);
					x = xy / y;
					z = yz / y;
				}
			} else {

				// m33 is the largest diagonal term so base result on this

				if (zz < epsilon) {

					x = 0.707106781;
					y = 0.707106781;
					z = 0;
				} else {

					z = Math.sqrt(zz);
					x = xz / z;
					y = yz / z;
				}
			}

			this.set(x, y, z, angle);

			return this; // return 180 deg rotation
		}

		// as we have reached here there are no singularities so we can handle normally

		var s = Math.sqrt((m32 - m23) * (m32 - m23) + (m13 - m31) * (m13 - m31) + (m21 - m12) * (m21 - m12)); // used to normalize

		if (Math.abs(s) < 0.001) s = 1;

		// prevent divide by zero, should not happen if matrix is orthogonal and should be
		// caught by singularity test above, but I've left it in just in case

		this.x = (m32 - m23) / s;
		this.y = (m13 - m31) / s;
		this.z = (m21 - m12) / s;
		this.w = Math.acos((m11 + m22 + m33 - 1) / 2);

		return this;
	},

	min: function min(v) {

		this.x = Math.min(this.x, v.x);
		this.y = Math.min(this.y, v.y);
		this.z = Math.min(this.z, v.z);
		this.w = Math.min(this.w, v.w);

		return this;
	},

	max: function max(v) {

		this.x = Math.max(this.x, v.x);
		this.y = Math.max(this.y, v.y);
		this.z = Math.max(this.z, v.z);
		this.w = Math.max(this.w, v.w);

		return this;
	},

	clamp: function clamp(min, max) {

		// This function assumes min < max, if this assumption isn't true it will not operate correctly

		this.x = Math.max(min.x, Math.min(max.x, this.x));
		this.y = Math.max(min.y, Math.min(max.y, this.y));
		this.z = Math.max(min.z, Math.min(max.z, this.z));
		this.w = Math.max(min.w, Math.min(max.w, this.w));

		return this;
	},

	clampScalar: function () {

		var min, max;

		return function clampScalar(minVal, maxVal) {

			if (min === undefined) {

				min = new Vector4();
				max = new Vector4();
			}

			min.set(minVal, minVal, minVal, minVal);
			max.set(maxVal, maxVal, maxVal, maxVal);

			return this.clamp(min, max);
		};
	}(),

	floor: function floor() {

		this.x = Math.floor(this.x);
		this.y = Math.floor(this.y);
		this.z = Math.floor(this.z);
		this.w = Math.floor(this.w);

		return this;
	},

	ceil: function ceil() {

		this.x = Math.ceil(this.x);
		this.y = Math.ceil(this.y);
		this.z = Math.ceil(this.z);
		this.w = Math.ceil(this.w);

		return this;
	},

	round: function round() {

		this.x = Math.round(this.x);
		this.y = Math.round(this.y);
		this.z = Math.round(this.z);
		this.w = Math.round(this.w);

		return this;
	},

	roundToZero: function roundToZero() {

		this.x = this.x < 0 ? Math.ceil(this.x) : Math.floor(this.x);
		this.y = this.y < 0 ? Math.ceil(this.y) : Math.floor(this.y);
		this.z = this.z < 0 ? Math.ceil(this.z) : Math.floor(this.z);
		this.w = this.w < 0 ? Math.ceil(this.w) : Math.floor(this.w);

		return this;
	},

	negate: function negate() {

		this.x = -this.x;
		this.y = -this.y;
		this.z = -this.z;
		this.w = -this.w;

		return this;
	},

	dot: function dot(v) {

		return this.x * v.x + this.y * v.y + this.z * v.z + this.w * v.w;
	},

	lengthSq: function lengthSq() {

		return this.x * this.x + this.y * this.y + this.z * this.z + this.w * this.w;
	},

	length: function length() {

		return Math.sqrt(this.x * this.x + this.y * this.y + this.z * this.z + this.w * this.w);
	},

	lengthManhattan: function lengthManhattan() {

		return Math.abs(this.x) + Math.abs(this.y) + Math.abs(this.z) + Math.abs(this.w);
	},

	normalize: function normalize() {

		return this.divideScalar(this.length());
	},

	setLength: function setLength(length) {

		return this.multiplyScalar(length / this.length());
	},

	lerp: function lerp(v, alpha) {

		this.x += (v.x - this.x) * alpha;
		this.y += (v.y - this.y) * alpha;
		this.z += (v.z - this.z) * alpha;
		this.w += (v.w - this.w) * alpha;

		return this;
	},

	lerpVectors: function lerpVectors(v1, v2, alpha) {

		return this.subVectors(v2, v1).multiplyScalar(alpha).add(v1);
	},

	equals: function equals(v) {

		return v.x === this.x && v.y === this.y && v.z === this.z && v.w === this.w;
	},

	fromArray: function fromArray(array, offset) {

		if (offset === undefined) offset = 0;

		this.x = array[offset];
		this.y = array[offset + 1];
		this.z = array[offset + 2];
		this.w = array[offset + 3];

		return this;
	},

	toArray: function toArray(array, offset) {

		if (array === undefined) array = [];
		if (offset === undefined) offset = 0;

		array[offset] = this.x;
		array[offset + 1] = this.y;
		array[offset + 2] = this.z;
		array[offset + 3] = this.w;

		return array;
	},

	fromAttribute: function fromAttribute(attribute, index, offset) {

		if (offset === undefined) offset = 0;

		index = index * attribute.itemSize + offset;

		this.x = attribute.array[index];
		this.y = attribute.array[index + 1];
		this.z = attribute.array[index + 2];
		this.w = attribute.array[index + 3];

		return this;
	}

};

/**
 * @author mrdoob / http://mrdoob.com/
 */

function BufferAttribute(array, itemSize, normalized) {

	if (Array.isArray(array)) {

		throw new TypeError('THREE.BufferAttribute: array should be a Typed Array.');
	}

	this.uuid = _Math.generateUUID();

	this.array = array;
	this.itemSize = itemSize;
	this.count = array !== undefined ? array.length / itemSize : 0;
	this.normalized = normalized === true;

	this.dynamic = false;
	this.updateRange = { offset: 0, count: -1 };

	this.version = 0;
}

BufferAttribute.prototype = {

	constructor: BufferAttribute,

	isBufferAttribute: true,

	set needsUpdate(value) {

		if (value === true) this.version++;
	},

	setArray: function setArray(array) {

		if (Array.isArray(array)) {

			throw new TypeError('THREE.BufferAttribute: array should be a Typed Array.');
		}

		this.count = array !== undefined ? array.length / this.itemSize : 0;
		this.array = array;
	},

	setDynamic: function setDynamic(value) {

		this.dynamic = value;

		return this;
	},

	copy: function copy(source) {

		this.array = new source.array.constructor(source.array);
		this.itemSize = source.itemSize;
		this.count = source.count;
		this.normalized = source.normalized;

		this.dynamic = source.dynamic;

		return this;
	},

	copyAt: function copyAt(index1, attribute, index2) {

		index1 *= this.itemSize;
		index2 *= attribute.itemSize;

		for (var i = 0, l = this.itemSize; i < l; i++) {

			this.array[index1 + i] = attribute.array[index2 + i];
		}

		return this;
	},

	copyArray: function copyArray(array) {

		this.array.set(array);

		return this;
	},

	copyColorsArray: function copyColorsArray(colors) {

		var array = this.array,
		    offset = 0;

		for (var i = 0, l = colors.length; i < l; i++) {

			var color = colors[i];

			if (color === undefined) {

				console.warn('THREE.BufferAttribute.copyColorsArray(): color is undefined', i);
				color = new Color();
			}

			array[offset++] = color.r;
			array[offset++] = color.g;
			array[offset++] = color.b;
		}

		return this;
	},

	copyIndicesArray: function copyIndicesArray(indices) {

		var array = this.array,
		    offset = 0;

		for (var i = 0, l = indices.length; i < l; i++) {

			var index = indices[i];

			array[offset++] = index.a;
			array[offset++] = index.b;
			array[offset++] = index.c;
		}

		return this;
	},

	copyVector2sArray: function copyVector2sArray(vectors) {

		var array = this.array,
		    offset = 0;

		for (var i = 0, l = vectors.length; i < l; i++) {

			var vector = vectors[i];

			if (vector === undefined) {

				console.warn('THREE.BufferAttribute.copyVector2sArray(): vector is undefined', i);
				vector = new Vector2();
			}

			array[offset++] = vector.x;
			array[offset++] = vector.y;
		}

		return this;
	},

	copyVector3sArray: function copyVector3sArray(vectors) {

		var array = this.array,
		    offset = 0;

		for (var i = 0, l = vectors.length; i < l; i++) {

			var vector = vectors[i];

			if (vector === undefined) {

				console.warn('THREE.BufferAttribute.copyVector3sArray(): vector is undefined', i);
				vector = new Vector3();
			}

			array[offset++] = vector.x;
			array[offset++] = vector.y;
			array[offset++] = vector.z;
		}

		return this;
	},

	copyVector4sArray: function copyVector4sArray(vectors) {

		var array = this.array,
		    offset = 0;

		for (var i = 0, l = vectors.length; i < l; i++) {

			var vector = vectors[i];

			if (vector === undefined) {

				console.warn('THREE.BufferAttribute.copyVector4sArray(): vector is undefined', i);
				vector = new Vector4();
			}

			array[offset++] = vector.x;
			array[offset++] = vector.y;
			array[offset++] = vector.z;
			array[offset++] = vector.w;
		}

		return this;
	},

	set: function set(value, offset) {

		if (offset === undefined) offset = 0;

		this.array.set(value, offset);

		return this;
	},

	getX: function getX(index) {

		return this.array[index * this.itemSize];
	},

	setX: function setX(index, x) {

		this.array[index * this.itemSize] = x;

		return this;
	},

	getY: function getY(index) {

		return this.array[index * this.itemSize + 1];
	},

	setY: function setY(index, y) {

		this.array[index * this.itemSize + 1] = y;

		return this;
	},

	getZ: function getZ(index) {

		return this.array[index * this.itemSize + 2];
	},

	setZ: function setZ(index, z) {

		this.array[index * this.itemSize + 2] = z;

		return this;
	},

	getW: function getW(index) {

		return this.array[index * this.itemSize + 3];
	},

	setW: function setW(index, w) {

		this.array[index * this.itemSize + 3] = w;

		return this;
	},

	setXY: function setXY(index, x, y) {

		index *= this.itemSize;

		this.array[index + 0] = x;
		this.array[index + 1] = y;

		return this;
	},

	setXYZ: function setXYZ(index, x, y, z) {

		index *= this.itemSize;

		this.array[index + 0] = x;
		this.array[index + 1] = y;
		this.array[index + 2] = z;

		return this;
	},

	setXYZW: function setXYZW(index, x, y, z, w) {

		index *= this.itemSize;

		this.array[index + 0] = x;
		this.array[index + 1] = y;
		this.array[index + 2] = z;
		this.array[index + 3] = w;

		return this;
	},

	clone: function clone() {

		return new this.constructor().copy(this);
	}

};

function Uint16Attribute(array, itemSize) {

	return new BufferAttribute(new Uint16Array(array), itemSize);
}

function Uint32Attribute(array, itemSize) {

	return new BufferAttribute(new Uint32Array(array), itemSize);
}

function Float32Attribute(array, itemSize) {

	return new BufferAttribute(new Float32Array(array), itemSize);
}

/**
 * @author mrdoob / http://mrdoob.com/
 * @author kile / http://kile.stravaganza.org/
 * @author alteredq / http://alteredqualia.com/
 * @author mikael emtinger / http://gomo.se/
 * @author zz85 / http://www.lab4games.net/zz85/blog
 * @author bhouston / http://clara.io
 */

function Geometry() {

	Object.defineProperty(this, 'id', { value: GeometryIdCount() });

	this.uuid = _Math.generateUUID();

	this.name = '';
	this.type = 'Geometry';

	this.vertices = [];
	this.colors = [];
	this.faces = [];
	this.faceVertexUvs = [[]];

	this.morphTargets = [];
	this.morphNormals = [];

	this.skinWeights = [];
	this.skinIndices = [];

	this.lineDistances = [];

	this.boundingBox = null;
	this.boundingSphere = null;

	// update flags

	this.elementsNeedUpdate = false;
	this.verticesNeedUpdate = false;
	this.uvsNeedUpdate = false;
	this.normalsNeedUpdate = false;
	this.colorsNeedUpdate = false;
	this.lineDistancesNeedUpdate = false;
	this.groupsNeedUpdate = false;
}

Object.assign(Geometry.prototype, EventDispatcher.prototype, {

	isGeometry: true,

	applyMatrix: function applyMatrix(matrix) {

		var normalMatrix = new Matrix3().getNormalMatrix(matrix);

		for (var i = 0, il = this.vertices.length; i < il; i++) {

			var vertex = this.vertices[i];
			vertex.applyMatrix4(matrix);
		}

		for (var i = 0, il = this.faces.length; i < il; i++) {

			var face = this.faces[i];
			face.normal.applyMatrix3(normalMatrix).normalize();

			for (var j = 0, jl = face.vertexNormals.length; j < jl; j++) {

				face.vertexNormals[j].applyMatrix3(normalMatrix).normalize();
			}
		}

		if (this.boundingBox !== null) {

			this.computeBoundingBox();
		}

		if (this.boundingSphere !== null) {

			this.computeBoundingSphere();
		}

		this.verticesNeedUpdate = true;
		this.normalsNeedUpdate = true;

		return this;
	},

	rotateX: function () {

		// rotate geometry around world x-axis

		var m1;

		return function rotateX(angle) {

			if (m1 === undefined) m1 = new Matrix4();

			m1.makeRotationX(angle);

			this.applyMatrix(m1);

			return this;
		};
	}(),

	rotateY: function () {

		// rotate geometry around world y-axis

		var m1;

		return function rotateY(angle) {

			if (m1 === undefined) m1 = new Matrix4();

			m1.makeRotationY(angle);

			this.applyMatrix(m1);

			return this;
		};
	}(),

	rotateZ: function () {

		// rotate geometry around world z-axis

		var m1;

		return function rotateZ(angle) {

			if (m1 === undefined) m1 = new Matrix4();

			m1.makeRotationZ(angle);

			this.applyMatrix(m1);

			return this;
		};
	}(),

	translate: function () {

		// translate geometry

		var m1;

		return function translate(x, y, z) {

			if (m1 === undefined) m1 = new Matrix4();

			m1.makeTranslation(x, y, z);

			this.applyMatrix(m1);

			return this;
		};
	}(),

	scale: function () {

		// scale geometry

		var m1;

		return function scale(x, y, z) {

			if (m1 === undefined) m1 = new Matrix4();

			m1.makeScale(x, y, z);

			this.applyMatrix(m1);

			return this;
		};
	}(),

	lookAt: function () {

		var obj;

		return function lookAt(vector) {

			if (obj === undefined) obj = new Object3D();

			obj.lookAt(vector);

			obj.updateMatrix();

			this.applyMatrix(obj.matrix);
		};
	}(),

	fromBufferGeometry: function fromBufferGeometry(geometry) {

		var scope = this;

		var indices = geometry.index !== null ? geometry.index.array : undefined;
		var attributes = geometry.attributes;

		var positions = attributes.position.array;
		var normals = attributes.normal !== undefined ? attributes.normal.array : undefined;
		var colors = attributes.color !== undefined ? attributes.color.array : undefined;
		var uvs = attributes.uv !== undefined ? attributes.uv.array : undefined;
		var uvs2 = attributes.uv2 !== undefined ? attributes.uv2.array : undefined;

		if (uvs2 !== undefined) this.faceVertexUvs[1] = [];

		var tempNormals = [];
		var tempUVs = [];
		var tempUVs2 = [];

		for (var i = 0, j = 0; i < positions.length; i += 3, j += 2) {

			scope.vertices.push(new Vector3(positions[i], positions[i + 1], positions[i + 2]));

			if (normals !== undefined) {

				tempNormals.push(new Vector3(normals[i], normals[i + 1], normals[i + 2]));
			}

			if (colors !== undefined) {

				scope.colors.push(new Color(colors[i], colors[i + 1], colors[i + 2]));
			}

			if (uvs !== undefined) {

				tempUVs.push(new Vector2(uvs[j], uvs[j + 1]));
			}

			if (uvs2 !== undefined) {

				tempUVs2.push(new Vector2(uvs2[j], uvs2[j + 1]));
			}
		}

		function addFace(a, b, c, materialIndex) {

			var vertexNormals = normals !== undefined ? [tempNormals[a].clone(), tempNormals[b].clone(), tempNormals[c].clone()] : [];
			var vertexColors = colors !== undefined ? [scope.colors[a].clone(), scope.colors[b].clone(), scope.colors[c].clone()] : [];

			var face = new Face3(a, b, c, vertexNormals, vertexColors, materialIndex);

			scope.faces.push(face);

			if (uvs !== undefined) {

				scope.faceVertexUvs[0].push([tempUVs[a].clone(), tempUVs[b].clone(), tempUVs[c].clone()]);
			}

			if (uvs2 !== undefined) {

				scope.faceVertexUvs[1].push([tempUVs2[a].clone(), tempUVs2[b].clone(), tempUVs2[c].clone()]);
			}
		}

		if (indices !== undefined) {

			var groups = geometry.groups;

			if (groups.length > 0) {

				for (var i = 0; i < groups.length; i++) {

					var group = groups[i];

					var start = group.start;
					var count = group.count;

					for (var j = start, jl = start + count; j < jl; j += 3) {

						addFace(indices[j], indices[j + 1], indices[j + 2], group.materialIndex);
					}
				}
			} else {

				for (var i = 0; i < indices.length; i += 3) {

					addFace(indices[i], indices[i + 1], indices[i + 2]);
				}
			}
		} else {

			for (var i = 0; i < positions.length / 3; i += 3) {

				addFace(i, i + 1, i + 2);
			}
		}

		this.computeFaceNormals();

		if (geometry.boundingBox !== null) {

			this.boundingBox = geometry.boundingBox.clone();
		}

		if (geometry.boundingSphere !== null) {

			this.boundingSphere = geometry.boundingSphere.clone();
		}

		return this;
	},

	center: function center() {

		this.computeBoundingBox();

		var offset = this.boundingBox.getCenter().negate();

		this.translate(offset.x, offset.y, offset.z);

		return offset;
	},

	normalize: function normalize() {

		this.computeBoundingSphere();

		var center = this.boundingSphere.center;
		var radius = this.boundingSphere.radius;

		var s = radius === 0 ? 1 : 1.0 / radius;

		var matrix = new Matrix4();
		matrix.set(s, 0, 0, -s * center.x, 0, s, 0, -s * center.y, 0, 0, s, -s * center.z, 0, 0, 0, 1);

		this.applyMatrix(matrix);

		return this;
	},

	computeFaceNormals: function computeFaceNormals() {

		var cb = new Vector3(),
		    ab = new Vector3();

		for (var f = 0, fl = this.faces.length; f < fl; f++) {

			var face = this.faces[f];

			var vA = this.vertices[face.a];
			var vB = this.vertices[face.b];
			var vC = this.vertices[face.c];

			cb.subVectors(vC, vB);
			ab.subVectors(vA, vB);
			cb.cross(ab);

			cb.normalize();

			face.normal.copy(cb);
		}
	},

	computeVertexNormals: function computeVertexNormals(areaWeighted) {

		if (areaWeighted === undefined) areaWeighted = true;

		var v, vl, f, fl, face, vertices;

		vertices = new Array(this.vertices.length);

		for (v = 0, vl = this.vertices.length; v < vl; v++) {

			vertices[v] = new Vector3();
		}

		if (areaWeighted) {

			// vertex normals weighted by triangle areas
			// http://www.iquilezles.org/www/articles/normals/normals.htm

			var vA, vB, vC;
			var cb = new Vector3(),
			    ab = new Vector3();

			for (f = 0, fl = this.faces.length; f < fl; f++) {

				face = this.faces[f];

				vA = this.vertices[face.a];
				vB = this.vertices[face.b];
				vC = this.vertices[face.c];

				cb.subVectors(vC, vB);
				ab.subVectors(vA, vB);
				cb.cross(ab);

				vertices[face.a].add(cb);
				vertices[face.b].add(cb);
				vertices[face.c].add(cb);
			}
		} else {

			this.computeFaceNormals();

			for (f = 0, fl = this.faces.length; f < fl; f++) {

				face = this.faces[f];

				vertices[face.a].add(face.normal);
				vertices[face.b].add(face.normal);
				vertices[face.c].add(face.normal);
			}
		}

		for (v = 0, vl = this.vertices.length; v < vl; v++) {

			vertices[v].normalize();
		}

		for (f = 0, fl = this.faces.length; f < fl; f++) {

			face = this.faces[f];

			var vertexNormals = face.vertexNormals;

			if (vertexNormals.length === 3) {

				vertexNormals[0].copy(vertices[face.a]);
				vertexNormals[1].copy(vertices[face.b]);
				vertexNormals[2].copy(vertices[face.c]);
			} else {

				vertexNormals[0] = vertices[face.a].clone();
				vertexNormals[1] = vertices[face.b].clone();
				vertexNormals[2] = vertices[face.c].clone();
			}
		}

		if (this.faces.length > 0) {

			this.normalsNeedUpdate = true;
		}
	},

	computeFlatVertexNormals: function computeFlatVertexNormals() {

		var f, fl, face;

		this.computeFaceNormals();

		for (f = 0, fl = this.faces.length; f < fl; f++) {

			face = this.faces[f];

			var vertexNormals = face.vertexNormals;

			if (vertexNormals.length === 3) {

				vertexNormals[0].copy(face.normal);
				vertexNormals[1].copy(face.normal);
				vertexNormals[2].copy(face.normal);
			} else {

				vertexNormals[0] = face.normal.clone();
				vertexNormals[1] = face.normal.clone();
				vertexNormals[2] = face.normal.clone();
			}
		}

		if (this.faces.length > 0) {

			this.normalsNeedUpdate = true;
		}
	},

	computeMorphNormals: function computeMorphNormals() {

		var i, il, f, fl, face;

		// save original normals
		// - create temp variables on first access
		//   otherwise just copy (for faster repeated calls)

		for (f = 0, fl = this.faces.length; f < fl; f++) {

			face = this.faces[f];

			if (!face.__originalFaceNormal) {

				face.__originalFaceNormal = face.normal.clone();
			} else {

				face.__originalFaceNormal.copy(face.normal);
			}

			if (!face.__originalVertexNormals) face.__originalVertexNormals = [];

			for (i = 0, il = face.vertexNormals.length; i < il; i++) {

				if (!face.__originalVertexNormals[i]) {

					face.__originalVertexNormals[i] = face.vertexNormals[i].clone();
				} else {

					face.__originalVertexNormals[i].copy(face.vertexNormals[i]);
				}
			}
		}

		// use temp geometry to compute face and vertex normals for each morph

		var tmpGeo = new Geometry();
		tmpGeo.faces = this.faces;

		for (i = 0, il = this.morphTargets.length; i < il; i++) {

			// create on first access

			if (!this.morphNormals[i]) {

				this.morphNormals[i] = {};
				this.morphNormals[i].faceNormals = [];
				this.morphNormals[i].vertexNormals = [];

				var dstNormalsFace = this.morphNormals[i].faceNormals;
				var dstNormalsVertex = this.morphNormals[i].vertexNormals;

				var faceNormal, vertexNormals;

				for (f = 0, fl = this.faces.length; f < fl; f++) {

					faceNormal = new Vector3();
					vertexNormals = { a: new Vector3(), b: new Vector3(), c: new Vector3() };

					dstNormalsFace.push(faceNormal);
					dstNormalsVertex.push(vertexNormals);
				}
			}

			var morphNormals = this.morphNormals[i];

			// set vertices to morph target

			tmpGeo.vertices = this.morphTargets[i].vertices;

			// compute morph normals

			tmpGeo.computeFaceNormals();
			tmpGeo.computeVertexNormals();

			// store morph normals

			var faceNormal, vertexNormals;

			for (f = 0, fl = this.faces.length; f < fl; f++) {

				face = this.faces[f];

				faceNormal = morphNormals.faceNormals[f];
				vertexNormals = morphNormals.vertexNormals[f];

				faceNormal.copy(face.normal);

				vertexNormals.a.copy(face.vertexNormals[0]);
				vertexNormals.b.copy(face.vertexNormals[1]);
				vertexNormals.c.copy(face.vertexNormals[2]);
			}
		}

		// restore original normals

		for (f = 0, fl = this.faces.length; f < fl; f++) {

			face = this.faces[f];

			face.normal = face.__originalFaceNormal;
			face.vertexNormals = face.__originalVertexNormals;
		}
	},

	computeTangents: function computeTangents() {

		console.warn('THREE.Geometry: .computeTangents() has been removed.');
	},

	computeLineDistances: function computeLineDistances() {

		var d = 0;
		var vertices = this.vertices;

		for (var i = 0, il = vertices.length; i < il; i++) {

			if (i > 0) {

				d += vertices[i].distanceTo(vertices[i - 1]);
			}

			this.lineDistances[i] = d;
		}
	},

	computeBoundingBox: function computeBoundingBox() {

		if (this.boundingBox === null) {

			this.boundingBox = new Box3();
		}

		this.boundingBox.setFromPoints(this.vertices);
	},

	computeBoundingSphere: function computeBoundingSphere() {

		if (this.boundingSphere === null) {

			this.boundingSphere = new Sphere();
		}

		this.boundingSphere.setFromPoints(this.vertices);
	},

	merge: function merge(geometry, matrix, materialIndexOffset) {

		if ((geometry && geometry.isGeometry) === false) {

			console.error('THREE.Geometry.merge(): geometry not an instance of THREE.Geometry.', geometry);
			return;
		}

		var normalMatrix,
		    vertexOffset = this.vertices.length,
		    vertices1 = this.vertices,
		    vertices2 = geometry.vertices,
		    faces1 = this.faces,
		    faces2 = geometry.faces,
		    uvs1 = this.faceVertexUvs[0],
		    uvs2 = geometry.faceVertexUvs[0],
		    colors1 = this.colors,
		    colors2 = geometry.colors;

		if (materialIndexOffset === undefined) materialIndexOffset = 0;

		if (matrix !== undefined) {

			normalMatrix = new Matrix3().getNormalMatrix(matrix);
		}

		// vertices

		for (var i = 0, il = vertices2.length; i < il; i++) {

			var vertex = vertices2[i];

			var vertexCopy = vertex.clone();

			if (matrix !== undefined) vertexCopy.applyMatrix4(matrix);

			vertices1.push(vertexCopy);
		}

		// colors

		for (var i = 0, il = colors2.length; i < il; i++) {

			colors1.push(colors2[i].clone());
		}

		// faces

		for (i = 0, il = faces2.length; i < il; i++) {

			var face = faces2[i],
			    faceCopy,
			    normal,
			    color,
			    faceVertexNormals = face.vertexNormals,
			    faceVertexColors = face.vertexColors;

			faceCopy = new Face3(face.a + vertexOffset, face.b + vertexOffset, face.c + vertexOffset);
			faceCopy.normal.copy(face.normal);

			if (normalMatrix !== undefined) {

				faceCopy.normal.applyMatrix3(normalMatrix).normalize();
			}

			for (var j = 0, jl = faceVertexNormals.length; j < jl; j++) {

				normal = faceVertexNormals[j].clone();

				if (normalMatrix !== undefined) {

					normal.applyMatrix3(normalMatrix).normalize();
				}

				faceCopy.vertexNormals.push(normal);
			}

			faceCopy.color.copy(face.color);

			for (var j = 0, jl = faceVertexColors.length; j < jl; j++) {

				color = faceVertexColors[j];
				faceCopy.vertexColors.push(color.clone());
			}

			faceCopy.materialIndex = face.materialIndex + materialIndexOffset;

			faces1.push(faceCopy);
		}

		// uvs

		for (i = 0, il = uvs2.length; i < il; i++) {

			var uv = uvs2[i],
			    uvCopy = [];

			if (uv === undefined) {

				continue;
			}

			for (var j = 0, jl = uv.length; j < jl; j++) {

				uvCopy.push(uv[j].clone());
			}

			uvs1.push(uvCopy);
		}
	},

	mergeMesh: function mergeMesh(mesh) {

		if ((mesh && mesh.isMesh) === false) {

			console.error('THREE.Geometry.mergeMesh(): mesh not an instance of THREE.Mesh.', mesh);
			return;
		}

		mesh.matrixAutoUpdate && mesh.updateMatrix();

		this.merge(mesh.geometry, mesh.matrix);
	},

	/*
  * Checks for duplicate vertices with hashmap.
  * Duplicated vertices are removed
  * and faces' vertices are updated.
  */

	mergeVertices: function mergeVertices() {

		var verticesMap = {}; // Hashmap for looking up vertices by position coordinates (and making sure they are unique)
		var unique = [],
		    changes = [];

		var v, key;
		var precisionPoints = 4; // number of decimal points, e.g. 4 for epsilon of 0.0001
		var precision = Math.pow(10, precisionPoints);
		var i, il, face;
		var indices, j, jl;

		for (i = 0, il = this.vertices.length; i < il; i++) {

			v = this.vertices[i];
			key = Math.round(v.x * precision) + '_' + Math.round(v.y * precision) + '_' + Math.round(v.z * precision);

			if (verticesMap[key] === undefined) {

				verticesMap[key] = i;
				unique.push(this.vertices[i]);
				changes[i] = unique.length - 1;
			} else {

				//console.log('Duplicate vertex found. ', i, ' could be using ', verticesMap[key]);
				changes[i] = changes[verticesMap[key]];
			}
		}

		// if faces are completely degenerate after merging vertices, we
		// have to remove them from the geometry.
		var faceIndicesToRemove = [];

		for (i = 0, il = this.faces.length; i < il; i++) {

			face = this.faces[i];

			face.a = changes[face.a];
			face.b = changes[face.b];
			face.c = changes[face.c];

			indices = [face.a, face.b, face.c];

			var dupIndex = -1;

			// if any duplicate vertices are found in a Face3
			// we have to remove the face as nothing can be saved
			for (var n = 0; n < 3; n++) {

				if (indices[n] === indices[(n + 1) % 3]) {

					dupIndex = n;
					faceIndicesToRemove.push(i);
					break;
				}
			}
		}

		for (i = faceIndicesToRemove.length - 1; i >= 0; i--) {

			var idx = faceIndicesToRemove[i];

			this.faces.splice(idx, 1);

			for (j = 0, jl = this.faceVertexUvs.length; j < jl; j++) {

				this.faceVertexUvs[j].splice(idx, 1);
			}
		}

		// Use unique set of vertices

		var diff = this.vertices.length - unique.length;
		this.vertices = unique;
		return diff;
	},

	sortFacesByMaterialIndex: function sortFacesByMaterialIndex() {

		var faces = this.faces;
		var length = faces.length;

		// tag faces

		for (var i = 0; i < length; i++) {

			faces[i]._id = i;
		}

		// sort faces

		function materialIndexSort(a, b) {

			return a.materialIndex - b.materialIndex;
		}

		faces.sort(materialIndexSort);

		// sort uvs

		var uvs1 = this.faceVertexUvs[0];
		var uvs2 = this.faceVertexUvs[1];

		var newUvs1, newUvs2;

		if (uvs1 && uvs1.length === length) newUvs1 = [];
		if (uvs2 && uvs2.length === length) newUvs2 = [];

		for (var i = 0; i < length; i++) {

			var id = faces[i]._id;

			if (newUvs1) newUvs1.push(uvs1[id]);
			if (newUvs2) newUvs2.push(uvs2[id]);
		}

		if (newUvs1) this.faceVertexUvs[0] = newUvs1;
		if (newUvs2) this.faceVertexUvs[1] = newUvs2;
	},

	toJSON: function toJSON() {

		var data = {
			metadata: {
				version: 4.4,
				type: 'Geometry',
				generator: 'Geometry.toJSON'
			}
		};

		// standard Geometry serialization

		data.uuid = this.uuid;
		data.type = this.type;
		if (this.name !== '') data.name = this.name;

		if (this.parameters !== undefined) {

			var parameters = this.parameters;

			for (var key in parameters) {

				if (parameters[key] !== undefined) data[key] = parameters[key];
			}

			return data;
		}

		var vertices = [];

		for (var i = 0; i < this.vertices.length; i++) {

			var vertex = this.vertices[i];
			vertices.push(vertex.x, vertex.y, vertex.z);
		}

		var faces = [];
		var normals = [];
		var normalsHash = {};
		var colors = [];
		var colorsHash = {};
		var uvs = [];
		var uvsHash = {};

		for (var i = 0; i < this.faces.length; i++) {

			var face = this.faces[i];

			var hasMaterial = true;
			var hasFaceUv = false; // deprecated
			var hasFaceVertexUv = this.faceVertexUvs[0][i] !== undefined;
			var hasFaceNormal = face.normal.length() > 0;
			var hasFaceVertexNormal = face.vertexNormals.length > 0;
			var hasFaceColor = face.color.r !== 1 || face.color.g !== 1 || face.color.b !== 1;
			var hasFaceVertexColor = face.vertexColors.length > 0;

			var faceType = 0;

			faceType = setBit(faceType, 0, 0); // isQuad
			faceType = setBit(faceType, 1, hasMaterial);
			faceType = setBit(faceType, 2, hasFaceUv);
			faceType = setBit(faceType, 3, hasFaceVertexUv);
			faceType = setBit(faceType, 4, hasFaceNormal);
			faceType = setBit(faceType, 5, hasFaceVertexNormal);
			faceType = setBit(faceType, 6, hasFaceColor);
			faceType = setBit(faceType, 7, hasFaceVertexColor);

			faces.push(faceType);
			faces.push(face.a, face.b, face.c);
			faces.push(face.materialIndex);

			if (hasFaceVertexUv) {

				var faceVertexUvs = this.faceVertexUvs[0][i];

				faces.push(getUvIndex(faceVertexUvs[0]), getUvIndex(faceVertexUvs[1]), getUvIndex(faceVertexUvs[2]));
			}

			if (hasFaceNormal) {

				faces.push(getNormalIndex(face.normal));
			}

			if (hasFaceVertexNormal) {

				var vertexNormals = face.vertexNormals;

				faces.push(getNormalIndex(vertexNormals[0]), getNormalIndex(vertexNormals[1]), getNormalIndex(vertexNormals[2]));
			}

			if (hasFaceColor) {

				faces.push(getColorIndex(face.color));
			}

			if (hasFaceVertexColor) {

				var vertexColors = face.vertexColors;

				faces.push(getColorIndex(vertexColors[0]), getColorIndex(vertexColors[1]), getColorIndex(vertexColors[2]));
			}
		}

		function setBit(value, position, enabled) {

			return enabled ? value | 1 << position : value & ~(1 << position);
		}

		function getNormalIndex(normal) {

			var hash = normal.x.toString() + normal.y.toString() + normal.z.toString();

			if (normalsHash[hash] !== undefined) {

				return normalsHash[hash];
			}

			normalsHash[hash] = normals.length / 3;
			normals.push(normal.x, normal.y, normal.z);

			return normalsHash[hash];
		}

		function getColorIndex(color) {

			var hash = color.r.toString() + color.g.toString() + color.b.toString();

			if (colorsHash[hash] !== undefined) {

				return colorsHash[hash];
			}

			colorsHash[hash] = colors.length;
			colors.push(color.getHex());

			return colorsHash[hash];
		}

		function getUvIndex(uv) {

			var hash = uv.x.toString() + uv.y.toString();

			if (uvsHash[hash] !== undefined) {

				return uvsHash[hash];
			}

			uvsHash[hash] = uvs.length / 2;
			uvs.push(uv.x, uv.y);

			return uvsHash[hash];
		}

		data.data = {};

		data.data.vertices = vertices;
		data.data.normals = normals;
		if (colors.length > 0) data.data.colors = colors;
		if (uvs.length > 0) data.data.uvs = [uvs]; // temporal backward compatibility
		data.data.faces = faces;

		return data;
	},

	clone: function clone() {

		/*
  // Handle primitives
  	var parameters = this.parameters;
  	if ( parameters !== undefined ) {
  		var values = [];
  		for ( var key in parameters ) {
  			values.push( parameters[ key ] );
  		}
  		var geometry = Object.create( this.constructor.prototype );
  	this.constructor.apply( geometry, values );
  	return geometry;
  	}
  	return new this.constructor().copy( this );
  */

		return new Geometry().copy(this);
	},

	copy: function copy(source) {

		this.vertices = [];
		this.faces = [];
		this.faceVertexUvs = [[]];
		this.colors = [];

		var vertices = source.vertices;

		for (var i = 0, il = vertices.length; i < il; i++) {

			this.vertices.push(vertices[i].clone());
		}

		var colors = source.colors;

		for (var i = 0, il = colors.length; i < il; i++) {

			this.colors.push(colors[i].clone());
		}

		var faces = source.faces;

		for (var i = 0, il = faces.length; i < il; i++) {

			this.faces.push(faces[i].clone());
		}

		for (var i = 0, il = source.faceVertexUvs.length; i < il; i++) {

			var faceVertexUvs = source.faceVertexUvs[i];

			if (this.faceVertexUvs[i] === undefined) {

				this.faceVertexUvs[i] = [];
			}

			for (var j = 0, jl = faceVertexUvs.length; j < jl; j++) {

				var uvs = faceVertexUvs[j],
				    uvsCopy = [];

				for (var k = 0, kl = uvs.length; k < kl; k++) {

					var uv = uvs[k];

					uvsCopy.push(uv.clone());
				}

				this.faceVertexUvs[i].push(uvsCopy);
			}
		}

		return this;
	},

	dispose: function dispose() {

		this.dispatchEvent({ type: 'dispose' });
	}

});

var count$2 = 0;
function GeometryIdCount() {
	return count$2++;
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function DirectGeometry() {

	Object.defineProperty(this, 'id', { value: GeometryIdCount() });

	this.uuid = _Math.generateUUID();

	this.name = '';
	this.type = 'DirectGeometry';

	this.indices = [];
	this.vertices = [];
	this.normals = [];
	this.colors = [];
	this.uvs = [];
	this.uvs2 = [];

	this.groups = [];

	this.morphTargets = {};

	this.skinWeights = [];
	this.skinIndices = [];

	// this.lineDistances = [];

	this.boundingBox = null;
	this.boundingSphere = null;

	// update flags

	this.verticesNeedUpdate = false;
	this.normalsNeedUpdate = false;
	this.colorsNeedUpdate = false;
	this.uvsNeedUpdate = false;
	this.groupsNeedUpdate = false;
}

Object.assign(DirectGeometry.prototype, EventDispatcher.prototype, {

	computeBoundingBox: Geometry.prototype.computeBoundingBox,
	computeBoundingSphere: Geometry.prototype.computeBoundingSphere,

	computeFaceNormals: function computeFaceNormals() {

		console.warn('THREE.DirectGeometry: computeFaceNormals() is not a method of this type of geometry.');
	},

	computeVertexNormals: function computeVertexNormals() {

		console.warn('THREE.DirectGeometry: computeVertexNormals() is not a method of this type of geometry.');
	},

	computeGroups: function computeGroups(geometry) {

		var group;
		var groups = [];
		var materialIndex;

		var faces = geometry.faces;

		for (var i = 0; i < faces.length; i++) {

			var face = faces[i];

			// materials

			if (face.materialIndex !== materialIndex) {

				materialIndex = face.materialIndex;

				if (group !== undefined) {

					group.count = i * 3 - group.start;
					groups.push(group);
				}

				group = {
					start: i * 3,
					materialIndex: materialIndex
				};
			}
		}

		if (group !== undefined) {

			group.count = i * 3 - group.start;
			groups.push(group);
		}

		this.groups = groups;
	},

	fromGeometry: function fromGeometry(geometry) {

		var faces = geometry.faces;
		var vertices = geometry.vertices;
		var faceVertexUvs = geometry.faceVertexUvs;

		var hasFaceVertexUv = faceVertexUvs[0] && faceVertexUvs[0].length > 0;
		var hasFaceVertexUv2 = faceVertexUvs[1] && faceVertexUvs[1].length > 0;

		// morphs

		var morphTargets = geometry.morphTargets;
		var morphTargetsLength = morphTargets.length;

		var morphTargetsPosition;

		if (morphTargetsLength > 0) {

			morphTargetsPosition = [];

			for (var i = 0; i < morphTargetsLength; i++) {

				morphTargetsPosition[i] = [];
			}

			this.morphTargets.position = morphTargetsPosition;
		}

		var morphNormals = geometry.morphNormals;
		var morphNormalsLength = morphNormals.length;

		var morphTargetsNormal;

		if (morphNormalsLength > 0) {

			morphTargetsNormal = [];

			for (var i = 0; i < morphNormalsLength; i++) {

				morphTargetsNormal[i] = [];
			}

			this.morphTargets.normal = morphTargetsNormal;
		}

		// skins

		var skinIndices = geometry.skinIndices;
		var skinWeights = geometry.skinWeights;

		var hasSkinIndices = skinIndices.length === vertices.length;
		var hasSkinWeights = skinWeights.length === vertices.length;

		//

		for (var i = 0; i < faces.length; i++) {

			var face = faces[i];

			this.vertices.push(vertices[face.a], vertices[face.b], vertices[face.c]);

			var vertexNormals = face.vertexNormals;

			if (vertexNormals.length === 3) {

				this.normals.push(vertexNormals[0], vertexNormals[1], vertexNormals[2]);
			} else {

				var normal = face.normal;

				this.normals.push(normal, normal, normal);
			}

			var vertexColors = face.vertexColors;

			if (vertexColors.length === 3) {

				this.colors.push(vertexColors[0], vertexColors[1], vertexColors[2]);
			} else {

				var color = face.color;

				this.colors.push(color, color, color);
			}

			if (hasFaceVertexUv === true) {

				var vertexUvs = faceVertexUvs[0][i];

				if (vertexUvs !== undefined) {

					this.uvs.push(vertexUvs[0], vertexUvs[1], vertexUvs[2]);
				} else {

					console.warn('THREE.DirectGeometry.fromGeometry(): Undefined vertexUv ', i);

					this.uvs.push(new Vector2(), new Vector2(), new Vector2());
				}
			}

			if (hasFaceVertexUv2 === true) {

				var vertexUvs = faceVertexUvs[1][i];

				if (vertexUvs !== undefined) {

					this.uvs2.push(vertexUvs[0], vertexUvs[1], vertexUvs[2]);
				} else {

					console.warn('THREE.DirectGeometry.fromGeometry(): Undefined vertexUv2 ', i);

					this.uvs2.push(new Vector2(), new Vector2(), new Vector2());
				}
			}

			// morphs

			for (var j = 0; j < morphTargetsLength; j++) {

				var morphTarget = morphTargets[j].vertices;

				morphTargetsPosition[j].push(morphTarget[face.a], morphTarget[face.b], morphTarget[face.c]);
			}

			for (var j = 0; j < morphNormalsLength; j++) {

				var morphNormal = morphNormals[j].vertexNormals[i];

				morphTargetsNormal[j].push(morphNormal.a, morphNormal.b, morphNormal.c);
			}

			// skins

			if (hasSkinIndices) {

				this.skinIndices.push(skinIndices[face.a], skinIndices[face.b], skinIndices[face.c]);
			}

			if (hasSkinWeights) {

				this.skinWeights.push(skinWeights[face.a], skinWeights[face.b], skinWeights[face.c]);
			}
		}

		this.computeGroups(geometry);

		this.verticesNeedUpdate = geometry.verticesNeedUpdate;
		this.normalsNeedUpdate = geometry.normalsNeedUpdate;
		this.colorsNeedUpdate = geometry.colorsNeedUpdate;
		this.uvsNeedUpdate = geometry.uvsNeedUpdate;
		this.groupsNeedUpdate = geometry.groupsNeedUpdate;

		return this;
	},

	dispose: function dispose() {

		this.dispatchEvent({ type: 'dispose' });
	}

});

/**
 * @author alteredq / http://alteredqualia.com/
 * @author mrdoob / http://mrdoob.com/
 */

function BufferGeometry() {

	Object.defineProperty(this, 'id', { value: GeometryIdCount() });

	this.uuid = _Math.generateUUID();

	this.name = '';
	this.type = 'BufferGeometry';

	this.index = null;
	this.attributes = {};

	this.morphAttributes = {};

	this.groups = [];

	this.boundingBox = null;
	this.boundingSphere = null;

	this.drawRange = { start: 0, count: Infinity };
}

Object.assign(BufferGeometry.prototype, EventDispatcher.prototype, {

	isBufferGeometry: true,

	getIndex: function getIndex() {

		return this.index;
	},

	setIndex: function setIndex(index) {

		this.index = index;
	},

	addAttribute: function addAttribute(name, attribute) {

		if ((attribute && attribute.isBufferAttribute) === false && (attribute && attribute.isInterleavedBufferAttribute) === false) {

			console.warn('THREE.BufferGeometry: .addAttribute() now expects ( name, attribute ).');

			this.addAttribute(name, new BufferAttribute(arguments[1], arguments[2]));

			return;
		}

		if (name === 'index') {

			console.warn('THREE.BufferGeometry.addAttribute: Use .setIndex() for index attribute.');
			this.setIndex(attribute);

			return;
		}

		this.attributes[name] = attribute;

		return this;
	},

	getAttribute: function getAttribute(name) {

		return this.attributes[name];
	},

	removeAttribute: function removeAttribute(name) {

		delete this.attributes[name];

		return this;
	},

	addGroup: function addGroup(start, count, materialIndex) {

		this.groups.push({

			start: start,
			count: count,
			materialIndex: materialIndex !== undefined ? materialIndex : 0

		});
	},

	clearGroups: function clearGroups() {

		this.groups = [];
	},

	setDrawRange: function setDrawRange(start, count) {

		this.drawRange.start = start;
		this.drawRange.count = count;
	},

	applyMatrix: function applyMatrix(matrix) {

		var position = this.attributes.position;

		if (position !== undefined) {

			matrix.applyToVector3Array(position.array);
			position.needsUpdate = true;
		}

		var normal = this.attributes.normal;

		if (normal !== undefined) {

			var normalMatrix = new Matrix3().getNormalMatrix(matrix);

			normalMatrix.applyToVector3Array(normal.array);
			normal.needsUpdate = true;
		}

		if (this.boundingBox !== null) {

			this.computeBoundingBox();
		}

		if (this.boundingSphere !== null) {

			this.computeBoundingSphere();
		}

		return this;
	},

	rotateX: function () {

		// rotate geometry around world x-axis

		var m1;

		return function rotateX(angle) {

			if (m1 === undefined) m1 = new Matrix4();

			m1.makeRotationX(angle);

			this.applyMatrix(m1);

			return this;
		};
	}(),

	rotateY: function () {

		// rotate geometry around world y-axis

		var m1;

		return function rotateY(angle) {

			if (m1 === undefined) m1 = new Matrix4();

			m1.makeRotationY(angle);

			this.applyMatrix(m1);

			return this;
		};
	}(),

	rotateZ: function () {

		// rotate geometry around world z-axis

		var m1;

		return function rotateZ(angle) {

			if (m1 === undefined) m1 = new Matrix4();

			m1.makeRotationZ(angle);

			this.applyMatrix(m1);

			return this;
		};
	}(),

	translate: function () {

		// translate geometry

		var m1;

		return function translate(x, y, z) {

			if (m1 === undefined) m1 = new Matrix4();

			m1.makeTranslation(x, y, z);

			this.applyMatrix(m1);

			return this;
		};
	}(),

	scale: function () {

		// scale geometry

		var m1;

		return function scale(x, y, z) {

			if (m1 === undefined) m1 = new Matrix4();

			m1.makeScale(x, y, z);

			this.applyMatrix(m1);

			return this;
		};
	}(),

	lookAt: function () {

		var obj;

		return function lookAt(vector) {

			if (obj === undefined) obj = new Object3D();

			obj.lookAt(vector);

			obj.updateMatrix();

			this.applyMatrix(obj.matrix);
		};
	}(),

	center: function center() {

		this.computeBoundingBox();

		var offset = this.boundingBox.getCenter().negate();

		this.translate(offset.x, offset.y, offset.z);

		return offset;
	},

	setFromObject: function setFromObject(object) {

		// console.log( 'THREE.BufferGeometry.setFromObject(). Converting', object, this );

		var geometry = object.geometry;

		if (object && object.isPoints || object && object.isLine) {

			var positions = new Float32Attribute(geometry.vertices.length * 3, 3);
			var colors = new Float32Attribute(geometry.colors.length * 3, 3);

			this.addAttribute('position', positions.copyVector3sArray(geometry.vertices));
			this.addAttribute('color', colors.copyColorsArray(geometry.colors));

			if (geometry.lineDistances && geometry.lineDistances.length === geometry.vertices.length) {

				var lineDistances = new Float32Attribute(geometry.lineDistances.length, 1);

				this.addAttribute('lineDistance', lineDistances.copyArray(geometry.lineDistances));
			}

			if (geometry.boundingSphere !== null) {

				this.boundingSphere = geometry.boundingSphere.clone();
			}

			if (geometry.boundingBox !== null) {

				this.boundingBox = geometry.boundingBox.clone();
			}
		} else if (object && object.isMesh) {

			if (geometry && geometry.isGeometry) {

				this.fromGeometry(geometry);
			}
		}

		return this;
	},

	updateFromObject: function updateFromObject(object) {

		var geometry = object.geometry;

		if (object && object.isMesh) {

			var direct = geometry.__directGeometry;

			if (geometry.elementsNeedUpdate === true) {

				direct = undefined;
				geometry.elementsNeedUpdate = false;
			}

			if (direct === undefined) {

				return this.fromGeometry(geometry);
			}

			direct.verticesNeedUpdate = geometry.verticesNeedUpdate;
			direct.normalsNeedUpdate = geometry.normalsNeedUpdate;
			direct.colorsNeedUpdate = geometry.colorsNeedUpdate;
			direct.uvsNeedUpdate = geometry.uvsNeedUpdate;
			direct.groupsNeedUpdate = geometry.groupsNeedUpdate;

			geometry.verticesNeedUpdate = false;
			geometry.normalsNeedUpdate = false;
			geometry.colorsNeedUpdate = false;
			geometry.uvsNeedUpdate = false;
			geometry.groupsNeedUpdate = false;

			geometry = direct;
		}

		var attribute;

		if (geometry.verticesNeedUpdate === true) {

			attribute = this.attributes.position;

			if (attribute !== undefined) {

				attribute.copyVector3sArray(geometry.vertices);
				attribute.needsUpdate = true;
			}

			geometry.verticesNeedUpdate = false;
		}

		if (geometry.normalsNeedUpdate === true) {

			attribute = this.attributes.normal;

			if (attribute !== undefined) {

				attribute.copyVector3sArray(geometry.normals);
				attribute.needsUpdate = true;
			}

			geometry.normalsNeedUpdate = false;
		}

		if (geometry.colorsNeedUpdate === true) {

			attribute = this.attributes.color;

			if (attribute !== undefined) {

				attribute.copyColorsArray(geometry.colors);
				attribute.needsUpdate = true;
			}

			geometry.colorsNeedUpdate = false;
		}

		if (geometry.uvsNeedUpdate) {

			attribute = this.attributes.uv;

			if (attribute !== undefined) {

				attribute.copyVector2sArray(geometry.uvs);
				attribute.needsUpdate = true;
			}

			geometry.uvsNeedUpdate = false;
		}

		if (geometry.lineDistancesNeedUpdate) {

			attribute = this.attributes.lineDistance;

			if (attribute !== undefined) {

				attribute.copyArray(geometry.lineDistances);
				attribute.needsUpdate = true;
			}

			geometry.lineDistancesNeedUpdate = false;
		}

		if (geometry.groupsNeedUpdate) {

			geometry.computeGroups(object.geometry);
			this.groups = geometry.groups;

			geometry.groupsNeedUpdate = false;
		}

		return this;
	},

	fromGeometry: function fromGeometry(geometry) {

		geometry.__directGeometry = new DirectGeometry().fromGeometry(geometry);

		return this.fromDirectGeometry(geometry.__directGeometry);
	},

	fromDirectGeometry: function fromDirectGeometry(geometry) {

		var positions = new Float32Array(geometry.vertices.length * 3);
		this.addAttribute('position', new BufferAttribute(positions, 3).copyVector3sArray(geometry.vertices));

		if (geometry.normals.length > 0) {

			var normals = new Float32Array(geometry.normals.length * 3);
			this.addAttribute('normal', new BufferAttribute(normals, 3).copyVector3sArray(geometry.normals));
		}

		if (geometry.colors.length > 0) {

			var colors = new Float32Array(geometry.colors.length * 3);
			this.addAttribute('color', new BufferAttribute(colors, 3).copyColorsArray(geometry.colors));
		}

		if (geometry.uvs.length > 0) {

			var uvs = new Float32Array(geometry.uvs.length * 2);
			this.addAttribute('uv', new BufferAttribute(uvs, 2).copyVector2sArray(geometry.uvs));
		}

		if (geometry.uvs2.length > 0) {

			var uvs2 = new Float32Array(geometry.uvs2.length * 2);
			this.addAttribute('uv2', new BufferAttribute(uvs2, 2).copyVector2sArray(geometry.uvs2));
		}

		if (geometry.indices.length > 0) {

			var TypeArray = geometry.vertices.length > 65535 ? Uint32Array : Uint16Array;
			var indices = new TypeArray(geometry.indices.length * 3);
			this.setIndex(new BufferAttribute(indices, 1).copyIndicesArray(geometry.indices));
		}

		// groups

		this.groups = geometry.groups;

		// morphs

		for (var name in geometry.morphTargets) {

			var array = [];
			var morphTargets = geometry.morphTargets[name];

			for (var i = 0, l = morphTargets.length; i < l; i++) {

				var morphTarget = morphTargets[i];

				var attribute = new Float32Attribute(morphTarget.length * 3, 3);

				array.push(attribute.copyVector3sArray(morphTarget));
			}

			this.morphAttributes[name] = array;
		}

		// skinning

		if (geometry.skinIndices.length > 0) {

			var skinIndices = new Float32Attribute(geometry.skinIndices.length * 4, 4);
			this.addAttribute('skinIndex', skinIndices.copyVector4sArray(geometry.skinIndices));
		}

		if (geometry.skinWeights.length > 0) {

			var skinWeights = new Float32Attribute(geometry.skinWeights.length * 4, 4);
			this.addAttribute('skinWeight', skinWeights.copyVector4sArray(geometry.skinWeights));
		}

		//

		if (geometry.boundingSphere !== null) {

			this.boundingSphere = geometry.boundingSphere.clone();
		}

		if (geometry.boundingBox !== null) {

			this.boundingBox = geometry.boundingBox.clone();
		}

		return this;
	},

	computeBoundingBox: function computeBoundingBox() {

		if (this.boundingBox === null) {

			this.boundingBox = new Box3();
		}

		var positions = this.attributes.position.array;

		if (positions !== undefined) {

			this.boundingBox.setFromArray(positions);
		} else {

			this.boundingBox.makeEmpty();
		}

		if (isNaN(this.boundingBox.min.x) || isNaN(this.boundingBox.min.y) || isNaN(this.boundingBox.min.z)) {

			console.error('THREE.BufferGeometry.computeBoundingBox: Computed min/max have NaN values. The "position" attribute is likely to have NaN values.', this);
		}
	},

	computeBoundingSphere: function () {

		var box = new Box3();
		var vector = new Vector3();

		return function computeBoundingSphere() {

			if (this.boundingSphere === null) {

				this.boundingSphere = new Sphere();
			}

			var positions = this.attributes.position;

			if (positions) {

				var array = positions.array;
				var center = this.boundingSphere.center;

				box.setFromArray(array);
				box.getCenter(center);

				// hoping to find a boundingSphere with a radius smaller than the
				// boundingSphere of the boundingBox: sqrt(3) smaller in the best case

				var maxRadiusSq = 0;

				for (var i = 0, il = array.length; i < il; i += 3) {

					vector.fromArray(array, i);
					maxRadiusSq = Math.max(maxRadiusSq, center.distanceToSquared(vector));
				}

				this.boundingSphere.radius = Math.sqrt(maxRadiusSq);

				if (isNaN(this.boundingSphere.radius)) {

					console.error('THREE.BufferGeometry.computeBoundingSphere(): Computed radius is NaN. The "position" attribute is likely to have NaN values.', this);
				}
			}
		};
	}(),

	computeFaceNormals: function computeFaceNormals() {

		// backwards compatibility

	},

	computeVertexNormals: function computeVertexNormals() {

		var index = this.index;
		var attributes = this.attributes;
		var groups = this.groups;

		if (attributes.position) {

			var positions = attributes.position.array;

			if (attributes.normal === undefined) {

				this.addAttribute('normal', new BufferAttribute(new Float32Array(positions.length), 3));
			} else {

				// reset existing normals to zero

				var array = attributes.normal.array;

				for (var i = 0, il = array.length; i < il; i++) {

					array[i] = 0;
				}
			}

			var normals = attributes.normal.array;

			var vA,
			    vB,
			    vC,
			    pA = new Vector3(),
			    pB = new Vector3(),
			    pC = new Vector3(),
			    cb = new Vector3(),
			    ab = new Vector3();

			// indexed elements

			if (index) {

				var indices = index.array;

				if (groups.length === 0) {

					this.addGroup(0, indices.length);
				}

				for (var j = 0, jl = groups.length; j < jl; ++j) {

					var group = groups[j];

					var start = group.start;
					var count = group.count;

					for (var i = start, il = start + count; i < il; i += 3) {

						vA = indices[i + 0] * 3;
						vB = indices[i + 1] * 3;
						vC = indices[i + 2] * 3;

						pA.fromArray(positions, vA);
						pB.fromArray(positions, vB);
						pC.fromArray(positions, vC);

						cb.subVectors(pC, pB);
						ab.subVectors(pA, pB);
						cb.cross(ab);

						normals[vA] += cb.x;
						normals[vA + 1] += cb.y;
						normals[vA + 2] += cb.z;

						normals[vB] += cb.x;
						normals[vB + 1] += cb.y;
						normals[vB + 2] += cb.z;

						normals[vC] += cb.x;
						normals[vC + 1] += cb.y;
						normals[vC + 2] += cb.z;
					}
				}
			} else {

				// non-indexed elements (unconnected triangle soup)

				for (var i = 0, il = positions.length; i < il; i += 9) {

					pA.fromArray(positions, i);
					pB.fromArray(positions, i + 3);
					pC.fromArray(positions, i + 6);

					cb.subVectors(pC, pB);
					ab.subVectors(pA, pB);
					cb.cross(ab);

					normals[i] = cb.x;
					normals[i + 1] = cb.y;
					normals[i + 2] = cb.z;

					normals[i + 3] = cb.x;
					normals[i + 4] = cb.y;
					normals[i + 5] = cb.z;

					normals[i + 6] = cb.x;
					normals[i + 7] = cb.y;
					normals[i + 8] = cb.z;
				}
			}

			this.normalizeNormals();

			attributes.normal.needsUpdate = true;
		}
	},

	merge: function merge(geometry, offset) {

		if ((geometry && geometry.isBufferGeometry) === false) {

			console.error('THREE.BufferGeometry.merge(): geometry not an instance of THREE.BufferGeometry.', geometry);
			return;
		}

		if (offset === undefined) offset = 0;

		var attributes = this.attributes;

		for (var key in attributes) {

			if (geometry.attributes[key] === undefined) continue;

			var attribute1 = attributes[key];
			var attributeArray1 = attribute1.array;

			var attribute2 = geometry.attributes[key];
			var attributeArray2 = attribute2.array;

			var attributeSize = attribute2.itemSize;

			for (var i = 0, j = attributeSize * offset; i < attributeArray2.length; i++, j++) {

				attributeArray1[j] = attributeArray2[i];
			}
		}

		return this;
	},

	normalizeNormals: function normalizeNormals() {

		var normals = this.attributes.normal.array;

		var x, y, z, n;

		for (var i = 0, il = normals.length; i < il; i += 3) {

			x = normals[i];
			y = normals[i + 1];
			z = normals[i + 2];

			n = 1.0 / Math.sqrt(x * x + y * y + z * z);

			normals[i] *= n;
			normals[i + 1] *= n;
			normals[i + 2] *= n;
		}
	},

	toNonIndexed: function toNonIndexed() {

		if (this.index === null) {

			console.warn('THREE.BufferGeometry.toNonIndexed(): Geometry is already non-indexed.');
			return this;
		}

		var geometry2 = new BufferGeometry();

		var indices = this.index.array;
		var attributes = this.attributes;

		for (var name in attributes) {

			var attribute = attributes[name];

			var array = attribute.array;
			var itemSize = attribute.itemSize;

			var array2 = new array.constructor(indices.length * itemSize);

			var index = 0,
			    index2 = 0;

			for (var i = 0, l = indices.length; i < l; i++) {

				index = indices[i] * itemSize;

				for (var j = 0; j < itemSize; j++) {

					array2[index2++] = array[index++];
				}
			}

			geometry2.addAttribute(name, new BufferAttribute(array2, itemSize));
		}

		return geometry2;
	},

	toJSON: function toJSON() {

		var data = {
			metadata: {
				version: 4.4,
				type: 'BufferGeometry',
				generator: 'BufferGeometry.toJSON'
			}
		};

		// standard BufferGeometry serialization

		data.uuid = this.uuid;
		data.type = this.type;
		if (this.name !== '') data.name = this.name;

		if (this.parameters !== undefined) {

			var parameters = this.parameters;

			for (var key in parameters) {

				if (parameters[key] !== undefined) data[key] = parameters[key];
			}

			return data;
		}

		data.data = { attributes: {} };

		var index = this.index;

		if (index !== null) {

			var array = Array.prototype.slice.call(index.array);

			data.data.index = {
				type: index.array.constructor.name,
				array: array
			};
		}

		var attributes = this.attributes;

		for (var key in attributes) {

			var attribute = attributes[key];

			var array = Array.prototype.slice.call(attribute.array);

			data.data.attributes[key] = {
				itemSize: attribute.itemSize,
				type: attribute.array.constructor.name,
				array: array,
				normalized: attribute.normalized
			};
		}

		var groups = this.groups;

		if (groups.length > 0) {

			data.data.groups = JSON.parse(JSON.stringify(groups));
		}

		var boundingSphere = this.boundingSphere;

		if (boundingSphere !== null) {

			data.data.boundingSphere = {
				center: boundingSphere.center.toArray(),
				radius: boundingSphere.radius
			};
		}

		return data;
	},

	clone: function clone() {

		/*
  // Handle primitives
  	var parameters = this.parameters;
  	if ( parameters !== undefined ) {
  		var values = [];
  		for ( var key in parameters ) {
  			values.push( parameters[ key ] );
  		}
  		var geometry = Object.create( this.constructor.prototype );
  	this.constructor.apply( geometry, values );
  	return geometry;
  	}
  	return new this.constructor().copy( this );
  */

		return new BufferGeometry().copy(this);
	},

	copy: function copy(source) {

		var index = source.index;

		if (index !== null) {

			this.setIndex(index.clone());
		}

		var attributes = source.attributes;

		for (var name in attributes) {

			var attribute = attributes[name];
			this.addAttribute(name, attribute.clone());
		}

		var groups = source.groups;

		for (var i = 0, l = groups.length; i < l; i++) {

			var group = groups[i];
			this.addGroup(group.start, group.count, group.materialIndex);
		}

		return this;
	},

	dispose: function dispose() {

		this.dispatchEvent({ type: 'dispose' });
	}

});

BufferGeometry.MaxIndex = 65535;

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 * @author mikael emtinger / http://gomo.se/
 * @author jonobr1 / http://jonobr1.com/
 */

function Mesh(geometry, material) {

	Object3D.call(this);

	this.type = 'Mesh';

	this.geometry = geometry !== undefined ? geometry : new BufferGeometry();
	this.material = material !== undefined ? material : new MeshBasicMaterial({ color: Math.random() * 0xffffff });

	this.drawMode = TrianglesDrawMode;

	this.updateMorphTargets();
}

Mesh.prototype = Object.assign(Object.create(Object3D.prototype), {

	constructor: Mesh,

	isMesh: true,

	setDrawMode: function setDrawMode(value) {

		this.drawMode = value;
	},

	copy: function copy(source) {

		Object3D.prototype.copy.call(this, source);

		this.drawMode = source.drawMode;

		return this;
	},

	updateMorphTargets: function updateMorphTargets() {

		var morphTargets = this.geometry.morphTargets;

		if (morphTargets !== undefined && morphTargets.length > 0) {

			this.morphTargetInfluences = [];
			this.morphTargetDictionary = {};

			for (var m = 0, ml = morphTargets.length; m < ml; m++) {

				this.morphTargetInfluences.push(0);
				this.morphTargetDictionary[morphTargets[m].name] = m;
			}
		}
	},

	raycast: function () {

		var inverseMatrix = new Matrix4();
		var ray = new Ray();
		var sphere = new Sphere();

		var vA = new Vector3();
		var vB = new Vector3();
		var vC = new Vector3();

		var tempA = new Vector3();
		var tempB = new Vector3();
		var tempC = new Vector3();

		var uvA = new Vector2();
		var uvB = new Vector2();
		var uvC = new Vector2();

		var barycoord = new Vector3();

		var intersectionPoint = new Vector3();
		var intersectionPointWorld = new Vector3();

		function uvIntersection(point, p1, p2, p3, uv1, uv2, uv3) {

			Triangle.barycoordFromPoint(point, p1, p2, p3, barycoord);

			uv1.multiplyScalar(barycoord.x);
			uv2.multiplyScalar(barycoord.y);
			uv3.multiplyScalar(barycoord.z);

			uv1.add(uv2).add(uv3);

			return uv1.clone();
		}

		function checkIntersection(object, raycaster, ray, pA, pB, pC, point) {

			var intersect;
			var material = object.material;

			if (material.side === BackSide) {

				intersect = ray.intersectTriangle(pC, pB, pA, true, point);
			} else {

				intersect = ray.intersectTriangle(pA, pB, pC, material.side !== DoubleSide, point);
			}

			if (intersect === null) return null;

			intersectionPointWorld.copy(point);
			intersectionPointWorld.applyMatrix4(object.matrixWorld);

			var distance = raycaster.ray.origin.distanceTo(intersectionPointWorld);

			if (distance < raycaster.near || distance > raycaster.far) return null;

			return {
				distance: distance,
				point: intersectionPointWorld.clone(),
				object: object
			};
		}

		function checkBufferGeometryIntersection(object, raycaster, ray, positions, uvs, a, b, c) {

			vA.fromArray(positions, a * 3);
			vB.fromArray(positions, b * 3);
			vC.fromArray(positions, c * 3);

			var intersection = checkIntersection(object, raycaster, ray, vA, vB, vC, intersectionPoint);

			if (intersection) {

				if (uvs) {

					uvA.fromArray(uvs, a * 2);
					uvB.fromArray(uvs, b * 2);
					uvC.fromArray(uvs, c * 2);

					intersection.uv = uvIntersection(intersectionPoint, vA, vB, vC, uvA, uvB, uvC);
				}

				intersection.face = new Face3(a, b, c, Triangle.normal(vA, vB, vC));
				intersection.faceIndex = a;
			}

			return intersection;
		}

		return function raycast(raycaster, intersects) {

			var geometry = this.geometry;
			var material = this.material;
			var matrixWorld = this.matrixWorld;

			if (material === undefined) return;

			// Checking boundingSphere distance to ray

			if (geometry.boundingSphere === null) geometry.computeBoundingSphere();

			sphere.copy(geometry.boundingSphere);
			sphere.applyMatrix4(matrixWorld);

			if (raycaster.ray.intersectsSphere(sphere) === false) return;

			//

			inverseMatrix.getInverse(matrixWorld);
			ray.copy(raycaster.ray).applyMatrix4(inverseMatrix);

			// Check boundingBox before continuing

			if (geometry.boundingBox !== null) {

				if (ray.intersectsBox(geometry.boundingBox) === false) return;
			}

			var uvs, intersection;

			if (geometry && geometry.isBufferGeometry) {

				var a, b, c;
				var index = geometry.index;
				var attributes = geometry.attributes;
				var positions = attributes.position.array;

				if (attributes.uv !== undefined) {

					uvs = attributes.uv.array;
				}

				if (index !== null) {

					var indices = index.array;

					for (var i = 0, l = indices.length; i < l; i += 3) {

						a = indices[i];
						b = indices[i + 1];
						c = indices[i + 2];

						intersection = checkBufferGeometryIntersection(this, raycaster, ray, positions, uvs, a, b, c);

						if (intersection) {

							intersection.faceIndex = Math.floor(i / 3); // triangle number in indices buffer semantics
							intersects.push(intersection);
						}
					}
				} else {

					for (var i = 0, l = positions.length; i < l; i += 9) {

						a = i / 3;
						b = a + 1;
						c = a + 2;

						intersection = checkBufferGeometryIntersection(this, raycaster, ray, positions, uvs, a, b, c);

						if (intersection) {

							intersection.index = a; // triangle number in positions buffer semantics
							intersects.push(intersection);
						}
					}
				}
			} else if (geometry && geometry.isGeometry) {

				var fvA, fvB, fvC;
				var isFaceMaterial = material && material.isMultiMaterial;
				var materials = isFaceMaterial === true ? material.materials : null;

				var vertices = geometry.vertices;
				var faces = geometry.faces;
				var faceVertexUvs = geometry.faceVertexUvs[0];
				if (faceVertexUvs.length > 0) uvs = faceVertexUvs;

				for (var f = 0, fl = faces.length; f < fl; f++) {

					var face = faces[f];
					var faceMaterial = isFaceMaterial === true ? materials[face.materialIndex] : material;

					if (faceMaterial === undefined) continue;

					fvA = vertices[face.a];
					fvB = vertices[face.b];
					fvC = vertices[face.c];

					if (faceMaterial.morphTargets === true) {

						var morphTargets = geometry.morphTargets;
						var morphInfluences = this.morphTargetInfluences;

						vA.set(0, 0, 0);
						vB.set(0, 0, 0);
						vC.set(0, 0, 0);

						for (var t = 0, tl = morphTargets.length; t < tl; t++) {

							var influence = morphInfluences[t];

							if (influence === 0) continue;

							var targets = morphTargets[t].vertices;

							vA.addScaledVector(tempA.subVectors(targets[face.a], fvA), influence);
							vB.addScaledVector(tempB.subVectors(targets[face.b], fvB), influence);
							vC.addScaledVector(tempC.subVectors(targets[face.c], fvC), influence);
						}

						vA.add(fvA);
						vB.add(fvB);
						vC.add(fvC);

						fvA = vA;
						fvB = vB;
						fvC = vC;
					}

					intersection = checkIntersection(this, raycaster, ray, fvA, fvB, fvC, intersectionPoint);

					if (intersection) {

						if (uvs) {

							var uvs_f = uvs[f];
							uvA.copy(uvs_f[0]);
							uvB.copy(uvs_f[1]);
							uvC.copy(uvs_f[2]);

							intersection.uv = uvIntersection(intersectionPoint, fvA, fvB, fvC, uvA, uvB, uvC);
						}

						intersection.face = face;
						intersection.faceIndex = f;
						intersects.push(intersection);
					}
				}
			}
		};
	}(),

	clone: function clone() {

		return new this.constructor(this.geometry, this.material).copy(this);
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 * @author szimek / https://github.com/szimek/
 */

function Texture(image, mapping, wrapS, wrapT, magFilter, minFilter, format, type, anisotropy, encoding) {

	Object.defineProperty(this, 'id', { value: TextureIdCount() });

	this.uuid = _Math.generateUUID();

	this.name = '';
	this.sourceFile = '';

	this.image = image !== undefined ? image : Texture.DEFAULT_IMAGE;
	this.mipmaps = [];

	this.mapping = mapping !== undefined ? mapping : Texture.DEFAULT_MAPPING;

	this.wrapS = wrapS !== undefined ? wrapS : ClampToEdgeWrapping;
	this.wrapT = wrapT !== undefined ? wrapT : ClampToEdgeWrapping;

	this.magFilter = magFilter !== undefined ? magFilter : LinearFilter;
	this.minFilter = minFilter !== undefined ? minFilter : LinearMipMapLinearFilter;

	this.anisotropy = anisotropy !== undefined ? anisotropy : 1;

	this.format = format !== undefined ? format : RGBAFormat;
	this.type = type !== undefined ? type : UnsignedByteType;

	this.offset = new Vector2(0, 0);
	this.repeat = new Vector2(1, 1);

	this.generateMipmaps = true;
	this.premultiplyAlpha = false;
	this.flipY = true;
	this.unpackAlignment = 4; // valid values: 1, 2, 4, 8 (see http://www.khronos.org/opengles/sdk/docs/man/xhtml/glPixelStorei.xml)


	// Values of encoding !== THREE.LinearEncoding only supported on map, envMap and emissiveMap.
	//
	// Also changing the encoding after already used by a Material will not automatically make the Material
	// update.  You need to explicitly call Material.needsUpdate to trigger it to recompile.
	this.encoding = encoding !== undefined ? encoding : LinearEncoding;

	this.version = 0;
	this.onUpdate = null;
}

Texture.DEFAULT_IMAGE = undefined;
Texture.DEFAULT_MAPPING = UVMapping;

Texture.prototype = {

	constructor: Texture,

	isTexture: true,

	set needsUpdate(value) {

		if (value === true) this.version++;
	},

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(source) {

		this.image = source.image;
		this.mipmaps = source.mipmaps.slice(0);

		this.mapping = source.mapping;

		this.wrapS = source.wrapS;
		this.wrapT = source.wrapT;

		this.magFilter = source.magFilter;
		this.minFilter = source.minFilter;

		this.anisotropy = source.anisotropy;

		this.format = source.format;
		this.type = source.type;

		this.offset.copy(source.offset);
		this.repeat.copy(source.repeat);

		this.generateMipmaps = source.generateMipmaps;
		this.premultiplyAlpha = source.premultiplyAlpha;
		this.flipY = source.flipY;
		this.unpackAlignment = source.unpackAlignment;
		this.encoding = source.encoding;

		return this;
	},

	toJSON: function toJSON(meta) {

		if (meta.textures[this.uuid] !== undefined) {

			return meta.textures[this.uuid];
		}

		function getDataURL(image) {

			var canvas;

			if (image.toDataURL !== undefined) {

				canvas = image;
			} else {

				canvas = document.createElementNS('http://www.w3.org/1999/xhtml', 'canvas');
				canvas.width = image.width;
				canvas.height = image.height;

				canvas.getContext('2d').drawImage(image, 0, 0, image.width, image.height);
			}

			if (canvas.width > 2048 || canvas.height > 2048) {

				return canvas.toDataURL('image/jpeg', 0.6);
			} else {

				return canvas.toDataURL('image/png');
			}
		}

		var output = {
			metadata: {
				version: 4.4,
				type: 'Texture',
				generator: 'Texture.toJSON'
			},

			uuid: this.uuid,
			name: this.name,

			mapping: this.mapping,

			repeat: [this.repeat.x, this.repeat.y],
			offset: [this.offset.x, this.offset.y],
			wrap: [this.wrapS, this.wrapT],

			minFilter: this.minFilter,
			magFilter: this.magFilter,
			anisotropy: this.anisotropy,

			flipY: this.flipY
		};

		if (this.image !== undefined) {

			// TODO: Move to THREE.Image

			var image = this.image;

			if (image.uuid === undefined) {

				image.uuid = _Math.generateUUID(); // UGH
			}

			if (meta.images[image.uuid] === undefined) {

				meta.images[image.uuid] = {
					uuid: image.uuid,
					url: getDataURL(image)
				};
			}

			output.image = image.uuid;
		}

		meta.textures[this.uuid] = output;

		return output;
	},

	dispose: function dispose() {

		this.dispatchEvent({ type: 'dispose' });
	},

	transformUv: function transformUv(uv) {

		if (this.mapping !== UVMapping) return;

		uv.multiply(this.repeat);
		uv.add(this.offset);

		if (uv.x < 0 || uv.x > 1) {

			switch (this.wrapS) {

				case RepeatWrapping:

					uv.x = uv.x - Math.floor(uv.x);
					break;

				case ClampToEdgeWrapping:

					uv.x = uv.x < 0 ? 0 : 1;
					break;

				case MirroredRepeatWrapping:

					if (Math.abs(Math.floor(uv.x) % 2) === 1) {

						uv.x = Math.ceil(uv.x) - uv.x;
					} else {

						uv.x = uv.x - Math.floor(uv.x);
					}
					break;

			}
		}

		if (uv.y < 0 || uv.y > 1) {

			switch (this.wrapT) {

				case RepeatWrapping:

					uv.y = uv.y - Math.floor(uv.y);
					break;

				case ClampToEdgeWrapping:

					uv.y = uv.y < 0 ? 0 : 1;
					break;

				case MirroredRepeatWrapping:

					if (Math.abs(Math.floor(uv.y) % 2) === 1) {

						uv.y = Math.ceil(uv.y) - uv.y;
					} else {

						uv.y = uv.y - Math.floor(uv.y);
					}
					break;

			}
		}

		if (this.flipY) {

			uv.y = 1 - uv.y;
		}
	}

};

Object.assign(Texture.prototype, EventDispatcher.prototype);

var count$3 = 0;
function TextureIdCount() {
	return count$3++;
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

var Cache = {

	enabled: false,

	files: {},

	add: function add(key, file) {

		if (this.enabled === false) return;

		// console.log( 'THREE.Cache', 'Adding key:', key );

		this.files[key] = file;
	},

	get: function get(key) {

		if (this.enabled === false) return;

		// console.log( 'THREE.Cache', 'Checking key:', key );

		return this.files[key];
	},

	remove: function remove(key) {

		delete this.files[key];
	},

	clear: function clear() {

		this.files = {};
	}

};

/**
 * @author mrdoob / http://mrdoob.com/
 */

function LoadingManager(onLoad, onProgress, onError) {

	var scope = this;

	var isLoading = false,
	    itemsLoaded = 0,
	    itemsTotal = 0;

	this.onStart = undefined;
	this.onLoad = onLoad;
	this.onProgress = onProgress;
	this.onError = onError;

	this.itemStart = function (url) {

		itemsTotal++;

		if (isLoading === false) {

			if (scope.onStart !== undefined) {

				scope.onStart(url, itemsLoaded, itemsTotal);
			}
		}

		isLoading = true;
	};

	this.itemEnd = function (url) {

		itemsLoaded++;

		if (scope.onProgress !== undefined) {

			scope.onProgress(url, itemsLoaded, itemsTotal);
		}

		if (itemsLoaded === itemsTotal) {

			isLoading = false;

			if (scope.onLoad !== undefined) {

				scope.onLoad();
			}
		}
	};

	this.itemError = function (url) {

		if (scope.onError !== undefined) {

			scope.onError(url);
		}
	};
}

var DefaultLoadingManager = new LoadingManager();

/**
 * @author mrdoob / http://mrdoob.com/
 */

function XHRLoader(manager) {

	this.manager = manager !== undefined ? manager : DefaultLoadingManager;
}

Object.assign(XHRLoader.prototype, {

	load: function load(url, onLoad, onProgress, onError) {

		if (url === undefined) url = '';

		if (this.path !== undefined) url = this.path + url;

		var scope = this;

		var cached = Cache.get(url);

		if (cached !== undefined) {

			scope.manager.itemStart(url);

			setTimeout(function () {

				if (onLoad) onLoad(cached);

				scope.manager.itemEnd(url);
			}, 0);

			return cached;
		}

		// Check for data: URI
		var dataUriRegex = /^data:(.*?)(;base64)?,(.*)$/;
		var dataUriRegexResult = url.match(dataUriRegex);

		// Safari can not handle Data URIs through XMLHttpRequest so process manually
		if (dataUriRegexResult) {

			var mimeType = dataUriRegexResult[1];
			var isBase64 = !!dataUriRegexResult[2];
			var data = dataUriRegexResult[3];

			data = window.decodeURIComponent(data);

			if (isBase64) {
				data = window.atob(data);
			}

			try {

				var response;
				var responseType = (this.responseType || '').toLowerCase();

				switch (responseType) {

					case 'arraybuffer':
					case 'blob':

						response = new ArrayBuffer(data.length);
						var view = new Uint8Array(response);
						for (var i = 0; i < data.length; i++) {

							view[i] = data.charCodeAt(i);
						}

						if (responseType === 'blob') {

							response = new Blob([response], { "type": mimeType });
						}

						break;

					case 'document':

						var parser = new DOMParser();
						response = parser.parseFromString(data, mimeType);

						break;

					case 'json':

						response = JSON.parse(data);

						break;

					default:
						// 'text' or other

						response = data;

						break;

				}

				// Wait for next browser tick
				window.setTimeout(function () {

					if (onLoad) onLoad(response);

					scope.manager.itemEnd(url);
				}, 0);
			} catch (error) {

				// Wait for next browser tick
				window.setTimeout(function () {

					if (onError) onError(error);

					scope.manager.itemError(url);
				}, 0);
			}
		} else {

			var request = new XMLHttpRequest();
			request.open('GET', url, true);

			request.addEventListener('load', function (event) {

				var response = event.target.response;

				Cache.add(url, response);

				if (this.status === 200) {

					if (onLoad) onLoad(response);

					scope.manager.itemEnd(url);
				} else if (this.status === 0) {

					// Some browsers return HTTP Status 0 when using non-http protocol
					// e.g. 'file://' or 'data://'. Handle as success.

					console.warn('THREE.XHRLoader: HTTP Status 0 received.');

					if (onLoad) onLoad(response);

					scope.manager.itemEnd(url);
				} else {

					if (onError) onError(event);

					scope.manager.itemError(url);
				}
			}, false);

			if (onProgress !== undefined) {

				request.addEventListener('progress', function (event) {

					onProgress(event);
				}, false);
			}

			request.addEventListener('error', function (event) {

				if (onError) onError(event);

				scope.manager.itemError(url);
			}, false);

			if (this.responseType !== undefined) request.responseType = this.responseType;
			if (this.withCredentials !== undefined) request.withCredentials = this.withCredentials;

			if (request.overrideMimeType) request.overrideMimeType('text/plain');

			request.send(null);
		}

		scope.manager.itemStart(url);

		return request;
	},

	setPath: function setPath(value) {

		this.path = value;
		return this;
	},

	setResponseType: function setResponseType(value) {

		this.responseType = value;
		return this;
	},

	setWithCredentials: function setWithCredentials(value) {

		this.withCredentials = value;
		return this;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function ImageLoader(manager) {

	this.manager = manager !== undefined ? manager : DefaultLoadingManager;
}

Object.assign(ImageLoader.prototype, {

	load: function load(url, onLoad, onProgress, onError) {

		var scope = this;

		var image = document.createElementNS('http://www.w3.org/1999/xhtml', 'img');
		image.onload = function () {

			image.onload = null;

			URL.revokeObjectURL(image.src);

			if (onLoad) onLoad(image);

			scope.manager.itemEnd(url);
		};
		image.onerror = onError;

		if (url.indexOf('data:') === 0) {

			image.src = url;
		} else {

			var loader = new XHRLoader();
			loader.setPath(this.path);
			loader.setResponseType('blob');
			loader.setWithCredentials(this.withCredentials);
			loader.load(url, function (blob) {

				image.src = URL.createObjectURL(blob);
			}, onProgress, onError);
		}

		scope.manager.itemStart(url);

		return image;
	},

	setCrossOrigin: function setCrossOrigin(value) {

		this.crossOrigin = value;
		return this;
	},

	setWithCredentials: function setWithCredentials(value) {

		this.withCredentials = value;
		return this;
	},

	setPath: function setPath(value) {

		this.path = value;
		return this;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function CubeTexture(images, mapping, wrapS, wrapT, magFilter, minFilter, format, type, anisotropy, encoding) {

	images = images !== undefined ? images : [];
	mapping = mapping !== undefined ? mapping : CubeReflectionMapping;

	Texture.call(this, images, mapping, wrapS, wrapT, magFilter, minFilter, format, type, anisotropy, encoding);

	this.flipY = false;
}

CubeTexture.prototype = Object.create(Texture.prototype);
CubeTexture.prototype.constructor = CubeTexture;

CubeTexture.prototype.isCubeTexture = true;

Object.defineProperty(CubeTexture.prototype, 'images', {

	get: function get() {

		return this.image;
	},

	set: function set(value) {

		this.image = value;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function CubeTextureLoader(manager) {

	this.manager = manager !== undefined ? manager : DefaultLoadingManager;
}

Object.assign(CubeTextureLoader.prototype, {

	load: function load(urls, onLoad, onProgress, onError) {

		var texture = new CubeTexture();

		var loader = new ImageLoader(this.manager);
		loader.setCrossOrigin(this.crossOrigin);
		loader.setPath(this.path);

		var loaded = 0;

		function loadTexture(i) {

			loader.load(urls[i], function (image) {

				texture.images[i] = image;

				loaded++;

				if (loaded === 6) {

					texture.needsUpdate = true;

					if (onLoad) onLoad(texture);
				}
			}, undefined, onError);
		}

		for (var i = 0; i < urls.length; ++i) {

			loadTexture(i);
		}

		return texture;
	},

	setCrossOrigin: function setCrossOrigin(value) {

		this.crossOrigin = value;
		return this;
	},

	setPath: function setPath(value) {

		this.path = value;
		return this;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function TextureLoader(manager) {

	this.manager = manager !== undefined ? manager : DefaultLoadingManager;
}

Object.assign(TextureLoader.prototype, {

	load: function load(url, onLoad, onProgress, onError) {

		var texture = new Texture();

		var loader = new ImageLoader(this.manager);
		loader.setCrossOrigin(this.crossOrigin);
		loader.setWithCredentials(this.withCredentials);
		loader.setPath(this.path);
		loader.load(url, function (image) {

			// JPEGs can't have an alpha channel, so memory can be saved by storing them as RGB.
			var isJPEG = url.search(/\.(jpg|jpeg)$/) > 0 || url.search(/^data\:image\/jpeg/) === 0;

			texture.format = isJPEG ? RGBFormat : RGBAFormat;
			texture.image = image;
			texture.needsUpdate = true;

			if (onLoad !== undefined) {

				onLoad(texture);
			}
		}, onProgress, onError);

		return texture;
	},

	setCrossOrigin: function setCrossOrigin(value) {

		this.crossOrigin = value;
		return this;
	},

	setWithCredentials: function setWithCredentials(value) {

		this.withCredentials = value;
		return this;
	},

	setPath: function setPath(value) {

		this.path = value;
		return this;
	}

});

function loadTexture$1(id, url, progress) {
  var textureLoader = null;
  if (url instanceof Array && url.length === 6) {
    textureLoader = new CubeTextureLoader();
  } else {
    if (url instanceof HTMLImageElement) {
      url = url.src;
    }

    if (typeof url === "string") {
      textureLoader = new TextureLoader();
    }
  }

  if (textureLoader) {
    textureLoader.setCrossOrigin("anonymous");
  }

  return cache("Texture(" + id + ")", function () {
    return new Promise(function (resolve, reject) {
      if (textureLoader) {
        textureLoader.load(url, resolve, progress, reject);
      } else {
        resolve(new Texture(url));
      }
    });
  });
}

function textured(geometry, txt, options) {
  options = Object.assign({}, {
    txtRepeatX: 1,
    txtRepeatY: 1,
    anisotropy: 1
  }, options);

  var txtID = (txt.id || txt).toString(),
      textureDescription = "Primrose.textured(" + txtID + ", " + options.txtRepeatX + ", " + options.txtRepeatY + ", " + options.anisotropy + ", " + options.scaleTextureWidth + ", " + options.scaleTextureHeight + ")";
  var texturePromise = cache(textureDescription, function () {
    if (typeof txt === "string" || txt instanceof Array || txt.length === 6) {
      return loadTexture$1(textureDescription, txt, options.progress);
    } else {
      var retValue = null;
      if (txt.isSurface) {
        if (!options.scaleTextureWidth || !options.scaleTextureHeight) {
          var imgWidth = txt.imageWidth,
              imgHeight = txt.imageHeight,
              dimX = Math.ceil(Math.log(imgWidth) / Math.LN2),
              dimY = Math.ceil(Math.log(imgHeight) / Math.LN2),
              newWidth = Math.pow(2, dimX),
              newHeight = Math.pow(2, dimY);

          if (options.scaleTexture) {
            newWidth *= options.scaleTexture;
            newHeight *= options.scaleTexture;
          }

          var scaleX = imgWidth / newWidth,
              scaleY = imgHeight / newHeight;

          if (scaleX !== 1 || scaleY !== 1) {
            if (scaleX !== 1) {
              options.scaleTextureWidth = scaleX;
            }

            if (scaleY !== 1) {
              options.scaleTextureHeight = scaleY;
            }

            txt.bounds.width = newWidth;
            txt.bounds.height = newHeight;
            txt.resize();
            txt.render(true);
          }
        }
        txt._material = mat;
        retValue = txt.texture;
      } else if (txt instanceof HTMLCanvasElement || txt instanceof HTMLVideoElement || txt instanceof HTMLImageElement) {
        retValue = new Texture(txt);
      } else if (txt.isTexture) {
        retValue = txt;
      } else {
        Promise.reject("Texture description couldn't be converted to a THREE.Texture object");
      }

      return Promise.resolve(retValue);
    }
  });

  var mat = material$1(textureDescription, options),
      obj = null;
  if (geometry.type.indexOf("Geometry") > -1) {
    obj = new Mesh(geometry, mat);
  } else if (geometry.isMesh) {
    obj = geometry;
    obj.material = mat;
    geometry = obj.geometry;
  }

  if (txt.isSurface) {
    obj.surface = txt;
  }

  if (options.shadow) {
    obj.receiveShadow = true;
    obj.castShadow = true;
  }

  if (options.scaleTextureWidth || options.scaleTextureHeight) {
    if (geometry.attributes && geometry.attributes.uv && geometry.attributes.uv.array) {
      var uv = geometry.attributes.uv,
          arr = uv.array,
          i;
      if (options.scaleTextureWidth) {
        for (i = 0; i < arr.length; i += uv.itemSize) {
          arr[i] *= options.scaleTextureWidth;
        }
      }
      if (options.scaleTextureHeight) {
        for (i = 1; i < arr.length; i += uv.itemSize) {
          arr[i] = 1 - (1 - arr[i]) * options.scaleTextureHeight;
        }
      }
    } else {
      console.trace(geometry, options);
    }
  }

  texturePromise.then(function (texture) {
    if (options.txtRepeatX * options.txtRepeatY > 1) {
      texture.wrapS = texture.wrapT = RepeatWrapping;
      texture.repeat.set(options.txtRepeatX, options.txtRepeatY);
    }

    texture.anisotropy = options.anisotropy;

    if (texture.isCubeTexture) {
      mat.envMap = texture;
    } else if (texture.isTexture) {
      mat.map = texture;
    }

    mat.needsUpdate = true;
    texture.needsUpdate = true;
    return texture;
  }).then(options.resolve).catch(options.reject);

  return obj;
}

function colored(geometry, color, options) {
  options = options || {};
  options.color = color;

  var mat = material$1("", options),
      obj = null;

  if (geometry.type.indexOf("Geometry") > -1) {
    obj = new Mesh(geometry, mat);
  } else if (geometry.isObject3D) {
    obj = geometry;
    obj.material = mat;
  }

  if (options.shadow) {
    obj.receiveShadow = true;
    obj.castShadow = true;
  }

  if (options.resolve) {
    options.resolve();
  }
  return obj;
}

/**
 * @author Mugen87 / https://github.com/Mugen87
 */

function BoxBufferGeometry(width, height, depth, widthSegments, heightSegments, depthSegments) {

	BufferGeometry.call(this);

	this.type = 'BoxBufferGeometry';

	this.parameters = {
		width: width,
		height: height,
		depth: depth,
		widthSegments: widthSegments,
		heightSegments: heightSegments,
		depthSegments: depthSegments
	};

	var scope = this;

	// segments
	widthSegments = Math.floor(widthSegments) || 1;
	heightSegments = Math.floor(heightSegments) || 1;
	depthSegments = Math.floor(depthSegments) || 1;

	// these are used to calculate buffer length
	var vertexCount = calculateVertexCount(widthSegments, heightSegments, depthSegments);
	var indexCount = calculateIndexCount(widthSegments, heightSegments, depthSegments);

	// buffers
	var indices = new (indexCount > 65535 ? Uint32Array : Uint16Array)(indexCount);
	var vertices = new Float32Array(vertexCount * 3);
	var normals = new Float32Array(vertexCount * 3);
	var uvs = new Float32Array(vertexCount * 2);

	// offset variables
	var vertexBufferOffset = 0;
	var uvBufferOffset = 0;
	var indexBufferOffset = 0;
	var numberOfVertices = 0;

	// group variables
	var groupStart = 0;

	// build each side of the box geometry
	buildPlane('z', 'y', 'x', -1, -1, depth, height, width, depthSegments, heightSegments, 0); // px
	buildPlane('z', 'y', 'x', 1, -1, depth, height, -width, depthSegments, heightSegments, 1); // nx
	buildPlane('x', 'z', 'y', 1, 1, width, depth, height, widthSegments, depthSegments, 2); // py
	buildPlane('x', 'z', 'y', 1, -1, width, depth, -height, widthSegments, depthSegments, 3); // ny
	buildPlane('x', 'y', 'z', 1, -1, width, height, depth, widthSegments, heightSegments, 4); // pz
	buildPlane('x', 'y', 'z', -1, -1, width, height, -depth, widthSegments, heightSegments, 5); // nz

	// build geometry
	this.setIndex(new BufferAttribute(indices, 1));
	this.addAttribute('position', new BufferAttribute(vertices, 3));
	this.addAttribute('normal', new BufferAttribute(normals, 3));
	this.addAttribute('uv', new BufferAttribute(uvs, 2));

	// helper functions

	function calculateVertexCount(w, h, d) {

		var vertices = 0;

		// calculate the amount of vertices for each side (plane)
		vertices += (w + 1) * (h + 1) * 2; // xy
		vertices += (w + 1) * (d + 1) * 2; // xz
		vertices += (d + 1) * (h + 1) * 2; // zy

		return vertices;
	}

	function calculateIndexCount(w, h, d) {

		var index = 0;

		// calculate the amount of squares for each side
		index += w * h * 2; // xy
		index += w * d * 2; // xz
		index += d * h * 2; // zy

		return index * 6; // two triangles per square => six vertices per square
	}

	function buildPlane(u, v, w, udir, vdir, width, height, depth, gridX, gridY, materialIndex) {

		var segmentWidth = width / gridX;
		var segmentHeight = height / gridY;

		var widthHalf = width / 2;
		var heightHalf = height / 2;
		var depthHalf = depth / 2;

		var gridX1 = gridX + 1;
		var gridY1 = gridY + 1;

		var vertexCounter = 0;
		var groupCount = 0;

		var vector = new Vector3();

		// generate vertices, normals and uvs

		for (var iy = 0; iy < gridY1; iy++) {

			var y = iy * segmentHeight - heightHalf;

			for (var ix = 0; ix < gridX1; ix++) {

				var x = ix * segmentWidth - widthHalf;

				// set values to correct vector component
				vector[u] = x * udir;
				vector[v] = y * vdir;
				vector[w] = depthHalf;

				// now apply vector to vertex buffer
				vertices[vertexBufferOffset] = vector.x;
				vertices[vertexBufferOffset + 1] = vector.y;
				vertices[vertexBufferOffset + 2] = vector.z;

				// set values to correct vector component
				vector[u] = 0;
				vector[v] = 0;
				vector[w] = depth > 0 ? 1 : -1;

				// now apply vector to normal buffer
				normals[vertexBufferOffset] = vector.x;
				normals[vertexBufferOffset + 1] = vector.y;
				normals[vertexBufferOffset + 2] = vector.z;

				// uvs
				uvs[uvBufferOffset] = ix / gridX;
				uvs[uvBufferOffset + 1] = 1 - iy / gridY;

				// update offsets and counters
				vertexBufferOffset += 3;
				uvBufferOffset += 2;
				vertexCounter += 1;
			}
		}

		// 1. you need three indices to draw a single face
		// 2. a single segment consists of two faces
		// 3. so we need to generate six (2*3) indices per segment

		for (iy = 0; iy < gridY; iy++) {

			for (ix = 0; ix < gridX; ix++) {

				// indices
				var a = numberOfVertices + ix + gridX1 * iy;
				var b = numberOfVertices + ix + gridX1 * (iy + 1);
				var c = numberOfVertices + (ix + 1) + gridX1 * (iy + 1);
				var d = numberOfVertices + (ix + 1) + gridX1 * iy;

				// face one
				indices[indexBufferOffset] = a;
				indices[indexBufferOffset + 1] = b;
				indices[indexBufferOffset + 2] = d;

				// face two
				indices[indexBufferOffset + 3] = b;
				indices[indexBufferOffset + 4] = c;
				indices[indexBufferOffset + 5] = d;

				// update offsets and counters
				indexBufferOffset += 6;
				groupCount += 6;
			}
		}

		// add a group to the geometry. this will ensure multi material support
		scope.addGroup(groupStart, groupCount, materialIndex);

		// calculate new start value for groups
		groupStart += groupCount;

		// update total number of vertices
		numberOfVertices += vertexCounter;
	}
}

BoxBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
BoxBufferGeometry.prototype.constructor = BoxBufferGeometry;

function box$1(width, height, length, t, u, v) {
  if (height === undefined) {
    height = width;
  }
  if (length === undefined) {
    length = width;
  }
  return cache("BoxBufferGeometry(" + width + ", " + height + ", " + length + ", " + t + ", " + u + ", " + v + ")", function () {
    return new BoxBufferGeometry(width, height, length, t, u, v);
  });
}

function brick(txt, width, height, length, options) {
  width = width || 1;
  height = height || 1;
  length = length || 1;
  options = Object.assign({}, {
    txtRepeatX: width,
    txtRepeatY: length,
    anisotropy: 8,
    transparent: true,
    opacity: 1
  }, options);
  var m = typeof txt === "number" ? colored : textured,
      obj = m(box$1(width, height, length), txt, options);
  return obj;
}

function axis(length, width) {
  var center = hub$1();
  put(brick(0xff0000, length, width, width)).on(center);
  put(brick(0x00ff00, width, length, width)).on(center);
  put(brick(0x0000ff, width, width, length)).on(center);
  return center;
}

/**
 * @author benaadams / https://twitter.com/ben_a_adams
 */

function CircleBufferGeometry(radius, segments, thetaStart, thetaLength) {

	BufferGeometry.call(this);

	this.type = 'CircleBufferGeometry';

	this.parameters = {
		radius: radius,
		segments: segments,
		thetaStart: thetaStart,
		thetaLength: thetaLength
	};

	radius = radius || 50;
	segments = segments !== undefined ? Math.max(3, segments) : 8;

	thetaStart = thetaStart !== undefined ? thetaStart : 0;
	thetaLength = thetaLength !== undefined ? thetaLength : Math.PI * 2;

	var vertices = segments + 2;

	var positions = new Float32Array(vertices * 3);
	var normals = new Float32Array(vertices * 3);
	var uvs = new Float32Array(vertices * 2);

	// center data is already zero, but need to set a few extras
	normals[2] = 1.0;
	uvs[0] = 0.5;
	uvs[1] = 0.5;

	for (var s = 0, i = 3, ii = 2; s <= segments; s++, i += 3, ii += 2) {

		var segment = thetaStart + s / segments * thetaLength;

		positions[i] = radius * Math.cos(segment);
		positions[i + 1] = radius * Math.sin(segment);

		normals[i + 2] = 1; // normal z

		uvs[ii] = (positions[i] / radius + 1) / 2;
		uvs[ii + 1] = (positions[i + 1] / radius + 1) / 2;
	}

	var indices = [];

	for (var i = 1; i <= segments; i++) {

		indices.push(i, i + 1, 0);
	}

	this.setIndex(new BufferAttribute(new Uint16Array(indices), 1));
	this.addAttribute('position', new BufferAttribute(positions, 3));
	this.addAttribute('normal', new BufferAttribute(normals, 3));
	this.addAttribute('uv', new BufferAttribute(uvs, 2));

	this.boundingSphere = new Sphere(new Vector3(), radius);
}

CircleBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
CircleBufferGeometry.prototype.constructor = CircleBufferGeometry;

function circle(r, sections, start, end) {
  r = r || 1;
  sections = sections || 18;
  return cache("CircleBufferGeometry(" + r + ", " + sections + ", " + start + ", " + end + ")", function () {
    return new CircleBufferGeometry(r, sections, start, end);
  });
}

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 *
 * parameters = {
 *  color: <hex>,
 *  opacity: <float>,
 *  map: new THREE.Texture( <Image> ),
 *
 *  size: <float>,
 *  sizeAttenuation: <bool>
 * }
 */

function PointsMaterial(parameters) {

	Material.call(this);

	this.type = 'PointsMaterial';

	this.color = new Color(0xffffff);

	this.map = null;

	this.size = 1;
	this.sizeAttenuation = true;

	this.lights = false;

	this.setValues(parameters);
}

PointsMaterial.prototype = Object.create(Material.prototype);
PointsMaterial.prototype.constructor = PointsMaterial;

PointsMaterial.prototype.isPointsMaterial = true;

PointsMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.color.copy(source.color);

	this.map = source.map;

	this.size = source.size;
	this.sizeAttenuation = source.sizeAttenuation;

	return this;
};

/**
 * @author alteredq / http://alteredqualia.com/
 */

function Points(geometry, material) {

	Object3D.call(this);

	this.type = 'Points';

	this.geometry = geometry !== undefined ? geometry : new BufferGeometry();
	this.material = material !== undefined ? material : new PointsMaterial({ color: Math.random() * 0xffffff });
}

Points.prototype = Object.assign(Object.create(Object3D.prototype), {

	constructor: Points,

	isPoints: true,

	raycast: function () {

		var inverseMatrix = new Matrix4();
		var ray = new Ray();
		var sphere = new Sphere();

		return function raycast(raycaster, intersects) {

			var object = this;
			var geometry = this.geometry;
			var matrixWorld = this.matrixWorld;
			var threshold = raycaster.params.Points.threshold;

			// Checking boundingSphere distance to ray

			if (geometry.boundingSphere === null) geometry.computeBoundingSphere();

			sphere.copy(geometry.boundingSphere);
			sphere.applyMatrix4(matrixWorld);

			if (raycaster.ray.intersectsSphere(sphere) === false) return;

			//

			inverseMatrix.getInverse(matrixWorld);
			ray.copy(raycaster.ray).applyMatrix4(inverseMatrix);

			var localThreshold = threshold / ((this.scale.x + this.scale.y + this.scale.z) / 3);
			var localThresholdSq = localThreshold * localThreshold;
			var position = new Vector3();

			function testPoint(point, index) {

				var rayPointDistanceSq = ray.distanceSqToPoint(point);

				if (rayPointDistanceSq < localThresholdSq) {

					var intersectPoint = ray.closestPointToPoint(point);
					intersectPoint.applyMatrix4(matrixWorld);

					var distance = raycaster.ray.origin.distanceTo(intersectPoint);

					if (distance < raycaster.near || distance > raycaster.far) return;

					intersects.push({

						distance: distance,
						distanceToRay: Math.sqrt(rayPointDistanceSq),
						point: intersectPoint.clone(),
						index: index,
						face: null,
						object: object

					});
				}
			}

			if (geometry && geometry.isBufferGeometry) {

				var index = geometry.index;
				var attributes = geometry.attributes;
				var positions = attributes.position.array;

				if (index !== null) {

					var indices = index.array;

					for (var i = 0, il = indices.length; i < il; i++) {

						var a = indices[i];

						position.fromArray(positions, a * 3);

						testPoint(position, a);
					}
				} else {

					for (var i = 0, l = positions.length / 3; i < l; i++) {

						position.fromArray(positions, i * 3);

						testPoint(position, i);
					}
				}
			} else {

				var vertices = geometry.vertices;

				for (var i = 0, l = vertices.length; i < l; i++) {

					testPoint(vertices[i], i);
				}
			}
		};
	}(),

	clone: function clone() {

		return new this.constructor(this.geometry, this.material).copy(this);
	}

});

function cloud(verts, c, s) {
  var geom = new Geometry();
  for (var i = 0; i < verts.length; ++i) {
    geom.vertices.push(verts[i]);
  }
  var mat = cache("PointsMaterial(" + c + ", " + s + ")", function () {
    return new PointsMaterial({
      color: c,
      size: s
    });
  });
  return new Points(geom, mat);
}

/**
 * @author Mugen87 / https://github.com/Mugen87
 */

function CylinderBufferGeometry(radiusTop, radiusBottom, height, radialSegments, heightSegments, openEnded, thetaStart, thetaLength) {

	BufferGeometry.call(this);

	this.type = 'CylinderBufferGeometry';

	this.parameters = {
		radiusTop: radiusTop,
		radiusBottom: radiusBottom,
		height: height,
		radialSegments: radialSegments,
		heightSegments: heightSegments,
		openEnded: openEnded,
		thetaStart: thetaStart,
		thetaLength: thetaLength
	};

	var scope = this;

	radiusTop = radiusTop !== undefined ? radiusTop : 20;
	radiusBottom = radiusBottom !== undefined ? radiusBottom : 20;
	height = height !== undefined ? height : 100;

	radialSegments = Math.floor(radialSegments) || 8;
	heightSegments = Math.floor(heightSegments) || 1;

	openEnded = openEnded !== undefined ? openEnded : false;
	thetaStart = thetaStart !== undefined ? thetaStart : 0.0;
	thetaLength = thetaLength !== undefined ? thetaLength : 2.0 * Math.PI;

	// used to calculate buffer length

	var nbCap = 0;

	if (openEnded === false) {

		if (radiusTop > 0) nbCap++;
		if (radiusBottom > 0) nbCap++;
	}

	var vertexCount = calculateVertexCount();
	var indexCount = calculateIndexCount();

	// buffers

	var indices = new BufferAttribute(new (indexCount > 65535 ? Uint32Array : Uint16Array)(indexCount), 1);
	var vertices = new BufferAttribute(new Float32Array(vertexCount * 3), 3);
	var normals = new BufferAttribute(new Float32Array(vertexCount * 3), 3);
	var uvs = new BufferAttribute(new Float32Array(vertexCount * 2), 2);

	// helper variables

	var index = 0,
	    indexOffset = 0,
	    indexArray = [],
	    halfHeight = height / 2;

	// group variables
	var groupStart = 0;

	// generate geometry

	generateTorso();

	if (openEnded === false) {

		if (radiusTop > 0) generateCap(true);
		if (radiusBottom > 0) generateCap(false);
	}

	// build geometry

	this.setIndex(indices);
	this.addAttribute('position', vertices);
	this.addAttribute('normal', normals);
	this.addAttribute('uv', uvs);

	// helper functions

	function calculateVertexCount() {

		var count = (radialSegments + 1) * (heightSegments + 1);

		if (openEnded === false) {

			count += (radialSegments + 1) * nbCap + radialSegments * nbCap;
		}

		return count;
	}

	function calculateIndexCount() {

		var count = radialSegments * heightSegments * 2 * 3;

		if (openEnded === false) {

			count += radialSegments * nbCap * 3;
		}

		return count;
	}

	function generateTorso() {

		var x, y;
		var normal = new Vector3();
		var vertex = new Vector3();

		var groupCount = 0;

		// this will be used to calculate the normal
		var slope = (radiusBottom - radiusTop) / height;

		// generate vertices, normals and uvs

		for (y = 0; y <= heightSegments; y++) {

			var indexRow = [];

			var v = y / heightSegments;

			// calculate the radius of the current row
			var radius = v * (radiusBottom - radiusTop) + radiusTop;

			for (x = 0; x <= radialSegments; x++) {

				var u = x / radialSegments;

				var theta = u * thetaLength + thetaStart;

				var sinTheta = Math.sin(theta);
				var cosTheta = Math.cos(theta);

				// vertex
				vertex.x = radius * sinTheta;
				vertex.y = -v * height + halfHeight;
				vertex.z = radius * cosTheta;
				vertices.setXYZ(index, vertex.x, vertex.y, vertex.z);

				// normal
				normal.set(sinTheta, slope, cosTheta).normalize();
				normals.setXYZ(index, normal.x, normal.y, normal.z);

				// uv
				uvs.setXY(index, u, 1 - v);

				// save index of vertex in respective row
				indexRow.push(index);

				// increase index
				index++;
			}

			// now save vertices of the row in our index array
			indexArray.push(indexRow);
		}

		// generate indices

		for (x = 0; x < radialSegments; x++) {

			for (y = 0; y < heightSegments; y++) {

				// we use the index array to access the correct indices
				var i1 = indexArray[y][x];
				var i2 = indexArray[y + 1][x];
				var i3 = indexArray[y + 1][x + 1];
				var i4 = indexArray[y][x + 1];

				// face one
				indices.setX(indexOffset, i1);indexOffset++;
				indices.setX(indexOffset, i2);indexOffset++;
				indices.setX(indexOffset, i4);indexOffset++;

				// face two
				indices.setX(indexOffset, i2);indexOffset++;
				indices.setX(indexOffset, i3);indexOffset++;
				indices.setX(indexOffset, i4);indexOffset++;

				// update counters
				groupCount += 6;
			}
		}

		// add a group to the geometry. this will ensure multi material support
		scope.addGroup(groupStart, groupCount, 0);

		// calculate new start value for groups
		groupStart += groupCount;
	}

	function generateCap(top) {

		var x, centerIndexStart, centerIndexEnd;

		var uv = new Vector2();
		var vertex = new Vector3();

		var groupCount = 0;

		var radius = top === true ? radiusTop : radiusBottom;
		var sign = top === true ? 1 : -1;

		// save the index of the first center vertex
		centerIndexStart = index;

		// first we generate the center vertex data of the cap.
		// because the geometry needs one set of uvs per face,
		// we must generate a center vertex per face/segment

		for (x = 1; x <= radialSegments; x++) {

			// vertex
			vertices.setXYZ(index, 0, halfHeight * sign, 0);

			// normal
			normals.setXYZ(index, 0, sign, 0);

			// uv
			uv.x = 0.5;
			uv.y = 0.5;

			uvs.setXY(index, uv.x, uv.y);

			// increase index
			index++;
		}

		// save the index of the last center vertex
		centerIndexEnd = index;

		// now we generate the surrounding vertices, normals and uvs

		for (x = 0; x <= radialSegments; x++) {

			var u = x / radialSegments;
			var theta = u * thetaLength + thetaStart;

			var cosTheta = Math.cos(theta);
			var sinTheta = Math.sin(theta);

			// vertex
			vertex.x = radius * sinTheta;
			vertex.y = halfHeight * sign;
			vertex.z = radius * cosTheta;
			vertices.setXYZ(index, vertex.x, vertex.y, vertex.z);

			// normal
			normals.setXYZ(index, 0, sign, 0);

			// uv
			uv.x = cosTheta * 0.5 + 0.5;
			uv.y = sinTheta * 0.5 * sign + 0.5;
			uvs.setXY(index, uv.x, uv.y);

			// increase index
			index++;
		}

		// generate indices

		for (x = 0; x < radialSegments; x++) {

			var c = centerIndexStart + x;
			var i = centerIndexEnd + x;

			if (top === true) {

				// face top
				indices.setX(indexOffset, i);indexOffset++;
				indices.setX(indexOffset, i + 1);indexOffset++;
				indices.setX(indexOffset, c);indexOffset++;
			} else {

				// face bottom
				indices.setX(indexOffset, i + 1);indexOffset++;
				indices.setX(indexOffset, i);indexOffset++;
				indices.setX(indexOffset, c);indexOffset++;
			}

			// update counters
			groupCount += 3;
		}

		// add a group to the geometry. this will ensure multi material support
		scope.addGroup(groupStart, groupCount, top === true ? 1 : 2);

		// calculate new start value for groups
		groupStart += groupCount;
	}
}

CylinderBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
CylinderBufferGeometry.prototype.constructor = CylinderBufferGeometry;

function cylinder(rT, rB, height, rS, hS, openEnded, thetaStart, thetaEnd) {
  if (rT === undefined) {
    rT = 0.5;
  }
  if (rB === undefined) {
    rB = 0.5;
  }
  if (height === undefined) {
    height = 1;
  }
  return cache("CylinderBufferGeometry(" + rT + ", " + rB + ", " + height + ", " + rS + ", " + hS + ", " + openEnded + ", " + thetaStart + ", " + thetaEnd + ")", function () {
    return new CylinderBufferGeometry(rT, rB, height, rS, hS, openEnded, thetaStart, thetaEnd);
  });
}

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 */

function Light(color, intensity) {

	Object3D.call(this);

	this.type = 'Light';

	this.color = new Color(color);
	this.intensity = intensity !== undefined ? intensity : 1;

	this.receiveShadow = undefined;
}

Light.prototype = Object.assign(Object.create(Object3D.prototype), {

	constructor: Light,

	isLight: true,

	copy: function copy(source) {

		Object3D.prototype.copy.call(this, source);

		this.color.copy(source.color);
		this.intensity = source.intensity;

		return this;
	},

	toJSON: function toJSON(meta) {

		var data = Object3D.prototype.toJSON.call(this, meta);

		data.object.color = this.color.getHex();
		data.object.intensity = this.intensity;

		if (this.groundColor !== undefined) data.object.groundColor = this.groundColor.getHex();

		if (this.distance !== undefined) data.object.distance = this.distance;
		if (this.angle !== undefined) data.object.angle = this.angle;
		if (this.decay !== undefined) data.object.decay = this.decay;
		if (this.penumbra !== undefined) data.object.penumbra = this.penumbra;

		if (this.shadow !== undefined) data.object.shadow = this.shadow.toJSON();

		return data;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 * @author mikael emtinger / http://gomo.se/
 * @author WestLangley / http://github.com/WestLangley
*/

function Camera() {

	Object3D.call(this);

	this.type = 'Camera';

	this.matrixWorldInverse = new Matrix4();
	this.projectionMatrix = new Matrix4();
}

Camera.prototype = Object.create(Object3D.prototype);
Camera.prototype.constructor = Camera;

Camera.prototype.isCamera = true;

Camera.prototype.getWorldDirection = function () {

	var quaternion = new Quaternion();

	return function getWorldDirection(optionalTarget) {

		var result = optionalTarget || new Vector3();

		this.getWorldQuaternion(quaternion);

		return result.set(0, 0, -1).applyQuaternion(quaternion);
	};
}();

Camera.prototype.lookAt = function () {

	// This routine does not support cameras with rotated and/or translated parent(s)

	var m1 = new Matrix4();

	return function lookAt(vector) {

		m1.lookAt(this.position, vector, this.up);

		this.quaternion.setFromRotationMatrix(m1);
	};
}();

Camera.prototype.clone = function () {

	return new this.constructor().copy(this);
};

Camera.prototype.copy = function (source) {

	Object3D.prototype.copy.call(this, source);

	this.matrixWorldInverse.copy(source.matrixWorldInverse);
	this.projectionMatrix.copy(source.projectionMatrix);

	return this;
};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author greggman / http://games.greggman.com/
 * @author zz85 / http://www.lab4games.net/zz85/blog
 * @author tschw
 */

function PerspectiveCamera(fov, aspect, near, far) {

	Camera.call(this);

	this.type = 'PerspectiveCamera';

	this.fov = fov !== undefined ? fov : 50;
	this.zoom = 1;

	this.near = near !== undefined ? near : 0.1;
	this.far = far !== undefined ? far : 2000;
	this.focus = 10;

	this.aspect = aspect !== undefined ? aspect : 1;
	this.view = null;

	this.filmGauge = 35; // width of the film (default in millimeters)
	this.filmOffset = 0; // horizontal film offset (same unit as gauge)

	this.updateProjectionMatrix();
}

PerspectiveCamera.prototype = Object.assign(Object.create(Camera.prototype), {

	constructor: PerspectiveCamera,

	isPerspectiveCamera: true,

	copy: function copy(source) {

		Camera.prototype.copy.call(this, source);

		this.fov = source.fov;
		this.zoom = source.zoom;

		this.near = source.near;
		this.far = source.far;
		this.focus = source.focus;

		this.aspect = source.aspect;
		this.view = source.view === null ? null : Object.assign({}, source.view);

		this.filmGauge = source.filmGauge;
		this.filmOffset = source.filmOffset;

		return this;
	},

	/**
  * Sets the FOV by focal length in respect to the current .filmGauge.
  *
  * The default film gauge is 35, so that the focal length can be specified for
  * a 35mm (full frame) camera.
  *
  * Values for focal length and film gauge must have the same unit.
  */
	setFocalLength: function setFocalLength(focalLength) {

		// see http://www.bobatkins.com/photography/technical/field_of_view.html
		var vExtentSlope = 0.5 * this.getFilmHeight() / focalLength;

		this.fov = _Math.RAD2DEG * 2 * Math.atan(vExtentSlope);
		this.updateProjectionMatrix();
	},

	/**
  * Calculates the focal length from the current .fov and .filmGauge.
  */
	getFocalLength: function getFocalLength() {

		var vExtentSlope = Math.tan(_Math.DEG2RAD * 0.5 * this.fov);

		return 0.5 * this.getFilmHeight() / vExtentSlope;
	},

	getEffectiveFOV: function getEffectiveFOV() {

		return _Math.RAD2DEG * 2 * Math.atan(Math.tan(_Math.DEG2RAD * 0.5 * this.fov) / this.zoom);
	},

	getFilmWidth: function getFilmWidth() {

		// film not completely covered in portrait format (aspect < 1)
		return this.filmGauge * Math.min(this.aspect, 1);
	},

	getFilmHeight: function getFilmHeight() {

		// film not completely covered in landscape format (aspect > 1)
		return this.filmGauge / Math.max(this.aspect, 1);
	},

	/**
  * Sets an offset in a larger frustum. This is useful for multi-window or
  * multi-monitor/multi-machine setups.
  *
  * For example, if you have 3x2 monitors and each monitor is 1920x1080 and
  * the monitors are in grid like this
  *
  *   +---+---+---+
  *   | A | B | C |
  *   +---+---+---+
  *   | D | E | F |
  *   +---+---+---+
  *
  * then for each monitor you would call it like this
  *
  *   var w = 1920;
  *   var h = 1080;
  *   var fullWidth = w * 3;
  *   var fullHeight = h * 2;
  *
  *   --A--
  *   camera.setOffset( fullWidth, fullHeight, w * 0, h * 0, w, h );
  *   --B--
  *   camera.setOffset( fullWidth, fullHeight, w * 1, h * 0, w, h );
  *   --C--
  *   camera.setOffset( fullWidth, fullHeight, w * 2, h * 0, w, h );
  *   --D--
  *   camera.setOffset( fullWidth, fullHeight, w * 0, h * 1, w, h );
  *   --E--
  *   camera.setOffset( fullWidth, fullHeight, w * 1, h * 1, w, h );
  *   --F--
  *   camera.setOffset( fullWidth, fullHeight, w * 2, h * 1, w, h );
  *
  *   Note there is no reason monitors have to be the same size or in a grid.
  */
	setViewOffset: function setViewOffset(fullWidth, fullHeight, x, y, width, height) {

		this.aspect = fullWidth / fullHeight;

		this.view = {
			fullWidth: fullWidth,
			fullHeight: fullHeight,
			offsetX: x,
			offsetY: y,
			width: width,
			height: height
		};

		this.updateProjectionMatrix();
	},

	clearViewOffset: function clearViewOffset() {

		this.view = null;
		this.updateProjectionMatrix();
	},

	updateProjectionMatrix: function updateProjectionMatrix() {

		var near = this.near,
		    top = near * Math.tan(_Math.DEG2RAD * 0.5 * this.fov) / this.zoom,
		    height = 2 * top,
		    width = this.aspect * height,
		    left = -0.5 * width,
		    view = this.view;

		if (view !== null) {

			var fullWidth = view.fullWidth,
			    fullHeight = view.fullHeight;

			left += view.offsetX * width / fullWidth;
			top -= view.offsetY * height / fullHeight;
			width *= view.width / fullWidth;
			height *= view.height / fullHeight;
		}

		var skew = this.filmOffset;
		if (skew !== 0) left += near * skew / this.getFilmWidth();

		this.projectionMatrix.makeFrustum(left, left + width, top - height, top, near, this.far);
	},

	toJSON: function toJSON(meta) {

		var data = Object3D.prototype.toJSON.call(this, meta);

		data.object.fov = this.fov;
		data.object.zoom = this.zoom;

		data.object.near = this.near;
		data.object.far = this.far;
		data.object.focus = this.focus;

		data.object.aspect = this.aspect;

		if (this.view !== null) data.object.view = Object.assign({}, this.view);

		data.object.filmGauge = this.filmGauge;
		data.object.filmOffset = this.filmOffset;

		return data;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function LightShadow(camera) {

	this.camera = camera;

	this.bias = 0;
	this.radius = 1;

	this.mapSize = new Vector2(512, 512);

	this.map = null;
	this.matrix = new Matrix4();
}

Object.assign(LightShadow.prototype, {

	copy: function copy(source) {

		this.camera = source.camera.clone();

		this.bias = source.bias;
		this.radius = source.radius;

		this.mapSize.copy(source.mapSize);

		return this;
	},

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	toJSON: function toJSON() {

		var object = {};

		if (this.bias !== 0) object.bias = this.bias;
		if (this.radius !== 1) object.radius = this.radius;
		if (this.mapSize.x !== 512 || this.mapSize.y !== 512) object.mapSize = this.mapSize.toArray();

		object.camera = this.camera.toJSON(false).object;
		delete object.camera.matrix;

		return object;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function PointLight(color, intensity, distance, decay) {

	Light.call(this, color, intensity);

	this.type = 'PointLight';

	Object.defineProperty(this, 'power', {
		get: function get() {
			// intensity = power per solid angle.
			// ref: equation (15) from http://www.frostbite.com/wp-content/uploads/2014/11/course_notes_moving_frostbite_to_pbr.pdf
			return this.intensity * 4 * Math.PI;
		},
		set: function set(power) {
			// intensity = power per solid angle.
			// ref: equation (15) from http://www.frostbite.com/wp-content/uploads/2014/11/course_notes_moving_frostbite_to_pbr.pdf
			this.intensity = power / (4 * Math.PI);
		}
	});

	this.distance = distance !== undefined ? distance : 0;
	this.decay = decay !== undefined ? decay : 1; // for physically correct lights, should be 2.

	this.shadow = new LightShadow(new PerspectiveCamera(90, 1, 0.5, 500));
}

PointLight.prototype = Object.assign(Object.create(Light.prototype), {

	constructor: PointLight,

	isPointLight: true,

	copy: function copy(source) {

		Light.prototype.copy.call(this, source);

		this.distance = source.distance;
		this.decay = source.decay;

		this.shadow = source.shadow.clone();

		return this;
	}

});

function light(color, intensity, distance, decay) {
  return new PointLight(color, intensity, distance, decay);
}

function fixGeometry(geometry, options) {
  options = options || {};
  var maxU = options.maxU || 1,
      maxV = options.maxV || 1,
      attrs = geometry.attributes || geometry._bufferGeometry && geometry._bufferGeometry.attributes;
  if (attrs && attrs.uv && attrs.uv.array) {
    var uv = attrs.uv,
        arr = uv.array;
    for (var j = 0; j < arr.length; j += uv.itemSize) {
      arr[j] *= maxU;
    }
    for (var _j = 1; _j < arr.length; _j += uv.itemSize) {
      arr[_j] = 1 - (1 - arr[_j]) * maxV;
    }
  } else if (geometry.faceVertexUvs) {
    var faces = geometry.faceVertexUvs;
    for (var i = 0; i < faces.length; ++i) {
      var face = faces[i];
      for (var _j2 = 0; _j2 < face.length; ++_j2) {
        var uvs = face[_j2];
        for (var k = 0; k < uvs.length; ++k) {
          var _uv = uvs[k];
          _uv.x *= maxU;
          _uv.y = 1 - (1 - _uv.y) * maxV;
        }
      }
    }
  }

  return geometry;
}

/**
 * @author mrdoob / http://mrdoob.com/
 * based on http://papervision3d.googlecode.com/svn/trunk/as3/trunk/src/org/papervision3d/objects/primitives/Plane.as
 */

function PlaneBufferGeometry(width, height, widthSegments, heightSegments) {

	BufferGeometry.call(this);

	this.type = 'PlaneBufferGeometry';

	this.parameters = {
		width: width,
		height: height,
		widthSegments: widthSegments,
		heightSegments: heightSegments
	};

	var width_half = width / 2;
	var height_half = height / 2;

	var gridX = Math.floor(widthSegments) || 1;
	var gridY = Math.floor(heightSegments) || 1;

	var gridX1 = gridX + 1;
	var gridY1 = gridY + 1;

	var segment_width = width / gridX;
	var segment_height = height / gridY;

	var vertices = new Float32Array(gridX1 * gridY1 * 3);
	var normals = new Float32Array(gridX1 * gridY1 * 3);
	var uvs = new Float32Array(gridX1 * gridY1 * 2);

	var offset = 0;
	var offset2 = 0;

	for (var iy = 0; iy < gridY1; iy++) {

		var y = iy * segment_height - height_half;

		for (var ix = 0; ix < gridX1; ix++) {

			var x = ix * segment_width - width_half;

			vertices[offset] = x;
			vertices[offset + 1] = -y;

			normals[offset + 2] = 1;

			uvs[offset2] = ix / gridX;
			uvs[offset2 + 1] = 1 - iy / gridY;

			offset += 3;
			offset2 += 2;
		}
	}

	offset = 0;

	var indices = new (vertices.length / 3 > 65535 ? Uint32Array : Uint16Array)(gridX * gridY * 6);

	for (var iy = 0; iy < gridY; iy++) {

		for (var ix = 0; ix < gridX; ix++) {

			var a = ix + gridX1 * iy;
			var b = ix + gridX1 * (iy + 1);
			var c = ix + 1 + gridX1 * (iy + 1);
			var d = ix + 1 + gridX1 * iy;

			indices[offset] = a;
			indices[offset + 1] = b;
			indices[offset + 2] = d;

			indices[offset + 3] = b;
			indices[offset + 4] = c;
			indices[offset + 5] = d;

			offset += 6;
		}
	}

	this.setIndex(new BufferAttribute(indices, 1));
	this.addAttribute('position', new BufferAttribute(vertices, 3));
	this.addAttribute('normal', new BufferAttribute(normals, 3));
	this.addAttribute('uv', new BufferAttribute(uvs, 2));
}

PlaneBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
PlaneBufferGeometry.prototype.constructor = PlaneBufferGeometry;

function quad$1(width, height, options) {

  if (width === undefined) {
    width = 1;
  }

  if (height === undefined) {
    height = width;
  }

  options = Object.assign({}, {
    s: 1,
    t: 1
  }, options);

  return cache("PlaneBufferGeometry(" + width + ", " + height + ", " + options.s + ", " + options.t + ", " + options.maxU + ", " + options.maxV + ")", function () {
    return fixGeometry(new PlaneBufferGeometry(width, height, options.s, options.t), options);
  });
}

function identity$1(obj) {
  return obj;
}

function range(n, m, s, t) {
  var n2 = s && n || 0,
      m2 = s && m || n,
      s2 = t && s || 1;
  var t2 = t || s || m,
      output = null;

  if (!(t2 instanceof Function)) {
    t2 = identity$1;
  }

  for (var i = n2; i < m2; i += s2) {
    var value = t2(i);
    if (output === null && value !== undefined) {
      output = [];
    }
    if (output !== null) {
      output.push(value);
    }
  }
  if (output !== null) {
    return output;
  }
}

/**
 * @author Mugen87 / https://github.com/Mugen87
 */

function RingBufferGeometry(innerRadius, outerRadius, thetaSegments, phiSegments, thetaStart, thetaLength) {

	BufferGeometry.call(this);

	this.type = 'RingBufferGeometry';

	this.parameters = {
		innerRadius: innerRadius,
		outerRadius: outerRadius,
		thetaSegments: thetaSegments,
		phiSegments: phiSegments,
		thetaStart: thetaStart,
		thetaLength: thetaLength
	};

	innerRadius = innerRadius || 20;
	outerRadius = outerRadius || 50;

	thetaStart = thetaStart !== undefined ? thetaStart : 0;
	thetaLength = thetaLength !== undefined ? thetaLength : Math.PI * 2;

	thetaSegments = thetaSegments !== undefined ? Math.max(3, thetaSegments) : 8;
	phiSegments = phiSegments !== undefined ? Math.max(1, phiSegments) : 1;

	// these are used to calculate buffer length
	var vertexCount = (thetaSegments + 1) * (phiSegments + 1);
	var indexCount = thetaSegments * phiSegments * 2 * 3;

	// buffers
	var indices = new BufferAttribute(new (indexCount > 65535 ? Uint32Array : Uint16Array)(indexCount), 1);
	var vertices = new BufferAttribute(new Float32Array(vertexCount * 3), 3);
	var normals = new BufferAttribute(new Float32Array(vertexCount * 3), 3);
	var uvs = new BufferAttribute(new Float32Array(vertexCount * 2), 2);

	// some helper variables
	var index = 0,
	    indexOffset = 0,
	    segment;
	var radius = innerRadius;
	var radiusStep = (outerRadius - innerRadius) / phiSegments;
	var vertex = new Vector3();
	var uv = new Vector2();
	var j, i;

	// generate vertices, normals and uvs

	// values are generate from the inside of the ring to the outside

	for (j = 0; j <= phiSegments; j++) {

		for (i = 0; i <= thetaSegments; i++) {

			segment = thetaStart + i / thetaSegments * thetaLength;

			// vertex
			vertex.x = radius * Math.cos(segment);
			vertex.y = radius * Math.sin(segment);
			vertices.setXYZ(index, vertex.x, vertex.y, vertex.z);

			// normal
			normals.setXYZ(index, 0, 0, 1);

			// uv
			uv.x = (vertex.x / outerRadius + 1) / 2;
			uv.y = (vertex.y / outerRadius + 1) / 2;
			uvs.setXY(index, uv.x, uv.y);

			// increase index
			index++;
		}

		// increase the radius for next row of vertices
		radius += radiusStep;
	}

	// generate indices

	for (j = 0; j < phiSegments; j++) {

		var thetaSegmentLevel = j * (thetaSegments + 1);

		for (i = 0; i < thetaSegments; i++) {

			segment = i + thetaSegmentLevel;

			// indices
			var a = segment;
			var b = segment + thetaSegments + 1;
			var c = segment + thetaSegments + 2;
			var d = segment + 1;

			// face one
			indices.setX(indexOffset, a);indexOffset++;
			indices.setX(indexOffset, b);indexOffset++;
			indices.setX(indexOffset, c);indexOffset++;

			// face two
			indices.setX(indexOffset, a);indexOffset++;
			indices.setX(indexOffset, c);indexOffset++;
			indices.setX(indexOffset, d);indexOffset++;
		}
	}

	// build geometry

	this.setIndex(indices);
	this.addAttribute('position', vertices);
	this.addAttribute('normal', normals);
	this.addAttribute('uv', uvs);
}

RingBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
RingBufferGeometry.prototype.constructor = RingBufferGeometry;

function ring(rInner, rOuter, sectors, rings, start, end) {
  if (rInner === undefined) {
    rInner = 0.5;
  }
  sectors = sectors || 18;
  rings = rings || 1;
  rOuter = rOuter || 1;
  start = start || 0;
  end = end || 2 * Math.PI;
  return cache("RingBufferGeometry(" + rInner + ", " + rOuter + ", " + sectors + ", " + rings + ", " + start + ", " + end + ")", function () {
    return new RingBufferGeometry(rInner, rOuter, sectors, rings, start, end);
  });
}

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};





var asyncGenerator = function () {
  function AwaitValue(value) {
    this.value = value;
  }

  function AsyncGenerator(gen) {
    var front, back;

    function send(key, arg) {
      return new Promise(function (resolve, reject) {
        var request = {
          key: key,
          arg: arg,
          resolve: resolve,
          reject: reject,
          next: null
        };

        if (back) {
          back = back.next = request;
        } else {
          front = back = request;
          resume(key, arg);
        }
      });
    }

    function resume(key, arg) {
      try {
        var result = gen[key](arg);
        var value = result.value;

        if (value instanceof AwaitValue) {
          Promise.resolve(value.value).then(function (arg) {
            resume("next", arg);
          }, function (arg) {
            resume("throw", arg);
          });
        } else {
          settle(result.done ? "return" : "normal", result.value);
        }
      } catch (err) {
        settle("throw", err);
      }
    }

    function settle(type, value) {
      switch (type) {
        case "return":
          front.resolve({
            value: value,
            done: true
          });
          break;

        case "throw":
          front.reject(value);
          break;

        default:
          front.resolve({
            value: value,
            done: false
          });
          break;
      }

      front = front.next;

      if (front) {
        resume(front.key, front.arg);
      } else {
        back = null;
      }
    }

    this._invoke = send;

    if (typeof gen.return !== "function") {
      this.return = undefined;
    }
  }

  if (typeof Symbol === "function" && Symbol.asyncIterator) {
    AsyncGenerator.prototype[Symbol.asyncIterator] = function () {
      return this;
    };
  }

  AsyncGenerator.prototype.next = function (arg) {
    return this._invoke("next", arg);
  };

  AsyncGenerator.prototype.throw = function (arg) {
    return this._invoke("throw", arg);
  };

  AsyncGenerator.prototype.return = function (arg) {
    return this._invoke("return", arg);
  };

  return {
    wrap: function (fn) {
      return function () {
        return new AsyncGenerator(fn.apply(this, arguments));
      };
    },
    await: function (value) {
      return new AwaitValue(value);
    }
  };
}();





var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();







var get$1 = function get$1(object, property, receiver) {
  if (object === null) object = Function.prototype;
  var desc = Object.getOwnPropertyDescriptor(object, property);

  if (desc === undefined) {
    var parent = Object.getPrototypeOf(object);

    if (parent === null) {
      return undefined;
    } else {
      return get$1(parent, property, receiver);
    }
  } else if ("value" in desc) {
    return desc.value;
  } else {
    var getter = desc.get;

    if (getter === undefined) {
      return undefined;
    }

    return getter.call(receiver);
  }
};

var inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};











var possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};



var set$1 = function set$1(object, property, value, receiver) {
  var desc = Object.getOwnPropertyDescriptor(object, property);

  if (desc === undefined) {
    var parent = Object.getPrototypeOf(object);

    if (parent !== null) {
      set$1(parent, property, value, receiver);
    }
  } else if ("value" in desc && desc.writable) {
    desc.value = value;
  } else {
    var setter = desc.set;

    if (setter !== undefined) {
      setter.call(receiver, value);
    }
  }

  return value;
};

var InsideSphereGeometry = function (_Geometry) {
  inherits(InsideSphereGeometry, _Geometry);

  function InsideSphereGeometry(radius, widthSegments, heightSegments, phiStart, phiLength, thetaStart, thetaLength) {
    classCallCheck(this, InsideSphereGeometry);

    var _this = possibleConstructorReturn(this, (InsideSphereGeometry.__proto__ || Object.getPrototypeOf(InsideSphereGeometry)).call(this));

    _this.type = 'InsideSphereGeometry';

    _this.parameters = {
      radius: radius,
      widthSegments: widthSegments,
      heightSegments: heightSegments,
      phiStart: phiStart,
      phiLength: phiLength,
      thetaStart: thetaStart,
      thetaLength: thetaLength
    };

    radius = radius || 50;

    widthSegments = Math.max(3, Math.floor(widthSegments) || 8);
    heightSegments = Math.max(2, Math.floor(heightSegments) || 6);

    phiStart = phiStart !== undefined ? phiStart : 0;
    phiLength = phiLength !== undefined ? phiLength : Math.PI * 2;

    thetaStart = thetaStart !== undefined ? thetaStart : 0;
    thetaLength = thetaLength !== undefined ? thetaLength : Math.PI;

    var x,
        y,
        vertices = [],
        uvs = [];

    for (y = 0; y <= heightSegments; y++) {

      var verticesRow = [];
      var uvsRow = [];

      for (x = widthSegments; x >= 0; x--) {

        var u = x / widthSegments;

        var v = y / heightSegments;

        var vertex = new Vector3();
        vertex.x = -radius * Math.cos(phiStart + u * phiLength) * Math.sin(thetaStart + v * thetaLength);
        vertex.y = radius * Math.cos(thetaStart + v * thetaLength);
        vertex.z = radius * Math.sin(phiStart + u * phiLength) * Math.sin(thetaStart + v * thetaLength);

        _this.vertices.push(vertex);

        verticesRow.push(_this.vertices.length - 1);
        uvsRow.push(new Vector2(1 - u, 1 - v));
      }

      vertices.push(verticesRow);
      uvs.push(uvsRow);
    }

    for (y = 0; y < heightSegments; y++) {

      for (x = 0; x < widthSegments; x++) {

        var v1 = vertices[y][x + 1];
        var v2 = vertices[y][x];
        var v3 = vertices[y + 1][x];
        var v4 = vertices[y + 1][x + 1];

        var n1 = _this.vertices[v1].clone().normalize();
        var n2 = _this.vertices[v2].clone().normalize();
        var n3 = _this.vertices[v3].clone().normalize();
        var n4 = _this.vertices[v4].clone().normalize();

        var uv1 = uvs[y][x + 1].clone();
        var uv2 = uvs[y][x].clone();
        var uv3 = uvs[y + 1][x].clone();
        var uv4 = uvs[y + 1][x + 1].clone();

        if (Math.abs(_this.vertices[v1].y) === radius) {

          uv1.x = (uv1.x + uv2.x) / 2;
          _this.faces.push(new Face3(v1, v3, v4, [n1, n3, n4]));
          _this.faceVertexUvs[0].push([uv1, uv3, uv4]);
        } else if (Math.abs(_this.vertices[v3].y) === radius) {

          uv3.x = (uv3.x + uv4.x) / 2;
          _this.faces.push(new Face3(v1, v2, v3, [n1, n2, n3]));
          _this.faceVertexUvs[0].push([uv1, uv2, uv3]);
        } else {

          _this.faces.push(new Face3(v1, v2, v4, [n1, n2, n4]));
          _this.faceVertexUvs[0].push([uv1, uv2, uv4]);

          _this.faces.push(new Face3(v2, v3, v4, [n2.clone(), n3, n4.clone()]));
          _this.faceVertexUvs[0].push([uv2.clone(), uv3, uv4.clone()]);
        }
      }
    }

    _this.computeFaceNormals();

    for (var i = 0; i < _this.faces.length; ++i) {
      var f = _this.faces[i];
      f.normal.multiplyScalar(-1);
      for (var j = 0; j < f.vertexNormals.length; ++j) {
        f.vertexNormals[j].multiplyScalar(-1);
      }
    }

    _this.boundingSphere = new Sphere(new Vector3(), radius);

    return _this;
  }

  return InsideSphereGeometry;
}(Geometry);

function shell$1(r, slices, rings, phi, theta, options) {
  var SLICE = 0.45;
  if (phi === undefined) {
    phi = Math.PI * SLICE;
  }
  if (theta === undefined) {
    theta = Math.PI * SLICE * 0.6;
  }
  var phiStart = 1.5 * Math.PI - phi * 0.5,
      thetaStart = (Math.PI - theta) * 0.5;
  options = options || {};
  return cache("InsideSphereGeometry(" + r + ", " + slices + ", " + rings + ", " + phi + ", " + theta + ")", function () {
    return fixGeometry(new InsideSphereGeometry(r, slices, rings, phiStart, phi, thetaStart, theta, true), options);
  });
}

/**
 * @author mrdoob / http://mrdoob.com/
 * @author bhouston / http://clara.io/
 * @author stephomi / http://stephaneginier.com/
 */

function Raycaster(origin, direction, near, far) {

	this.ray = new Ray(origin, direction);
	// direction is assumed to be normalized (for accurate distance calculations)

	this.near = near || 0;
	this.far = far || Infinity;

	this.params = {
		Mesh: {},
		Line: {},
		LOD: {},
		Points: { threshold: 1 },
		Sprite: {}
	};

	Object.defineProperties(this.params, {
		PointCloud: {
			get: function get() {
				console.warn('THREE.Raycaster: params.PointCloud has been renamed to params.Points.');
				return this.Points;
			}
		}
	});
}

function ascSort(a, b) {

	return a.distance - b.distance;
}

function _intersectObject(object, raycaster, intersects, recursive) {

	if (object.visible === false) return;

	object.raycast(raycaster, intersects);

	if (recursive === true) {

		var children = object.children;

		for (var i = 0, l = children.length; i < l; i++) {

			_intersectObject(children[i], raycaster, intersects, true);
		}
	}
}

//

Raycaster.prototype = {

	constructor: Raycaster,

	linePrecision: 1,

	set: function set(origin, direction) {

		// direction is assumed to be normalized (for accurate distance calculations)

		this.ray.set(origin, direction);
	},

	setFromCamera: function setFromCamera(coords, camera) {

		if (camera && camera.isPerspectiveCamera) {

			this.ray.origin.setFromMatrixPosition(camera.matrixWorld);
			this.ray.direction.set(coords.x, coords.y, 0.5).unproject(camera).sub(this.ray.origin).normalize();
		} else if (camera && camera.isOrthographicCamera) {

			this.ray.origin.set(coords.x, coords.y, (camera.near + camera.far) / (camera.near - camera.far)).unproject(camera); // set origin in plane of camera
			this.ray.direction.set(0, 0, -1).transformDirection(camera.matrixWorld);
		} else {

			console.error('THREE.Raycaster: Unsupported camera type.');
		}
	},

	intersectObject: function intersectObject(object, recursive) {

		var intersects = [];

		_intersectObject(object, this, intersects, recursive);

		intersects.sort(ascSort);

		return intersects;
	},

	intersectObjects: function intersectObjects(objects, recursive) {

		var intersects = [];

		if (Array.isArray(objects) === false) {

			console.warn('THREE.Raycaster.intersectObjects: objects is not an Array.');
			return intersects;
		}

		for (var i = 0, l = objects.length; i < l; i++) {

			_intersectObject(objects[i], this, intersects, recursive);
		}

		intersects.sort(ascSort);

		return intersects;
	}

};

function raycaster() {
  return new Raycaster();
}

/**
 * @author benaadams / https://twitter.com/ben_a_adams
 * based on THREE.SphereGeometry
 */

function SphereBufferGeometry(radius, widthSegments, heightSegments, phiStart, phiLength, thetaStart, thetaLength) {

	BufferGeometry.call(this);

	this.type = 'SphereBufferGeometry';

	this.parameters = {
		radius: radius,
		widthSegments: widthSegments,
		heightSegments: heightSegments,
		phiStart: phiStart,
		phiLength: phiLength,
		thetaStart: thetaStart,
		thetaLength: thetaLength
	};

	radius = radius || 50;

	widthSegments = Math.max(3, Math.floor(widthSegments) || 8);
	heightSegments = Math.max(2, Math.floor(heightSegments) || 6);

	phiStart = phiStart !== undefined ? phiStart : 0;
	phiLength = phiLength !== undefined ? phiLength : Math.PI * 2;

	thetaStart = thetaStart !== undefined ? thetaStart : 0;
	thetaLength = thetaLength !== undefined ? thetaLength : Math.PI;

	var thetaEnd = thetaStart + thetaLength;

	var vertexCount = (widthSegments + 1) * (heightSegments + 1);

	var positions = new BufferAttribute(new Float32Array(vertexCount * 3), 3);
	var normals = new BufferAttribute(new Float32Array(vertexCount * 3), 3);
	var uvs = new BufferAttribute(new Float32Array(vertexCount * 2), 2);

	var index = 0,
	    vertices = [],
	    normal = new Vector3();

	for (var y = 0; y <= heightSegments; y++) {

		var verticesRow = [];

		var v = y / heightSegments;

		for (var x = 0; x <= widthSegments; x++) {

			var u = x / widthSegments;

			var px = -radius * Math.cos(phiStart + u * phiLength) * Math.sin(thetaStart + v * thetaLength);
			var py = radius * Math.cos(thetaStart + v * thetaLength);
			var pz = radius * Math.sin(phiStart + u * phiLength) * Math.sin(thetaStart + v * thetaLength);

			normal.set(px, py, pz).normalize();

			positions.setXYZ(index, px, py, pz);
			normals.setXYZ(index, normal.x, normal.y, normal.z);
			uvs.setXY(index, u, 1 - v);

			verticesRow.push(index);

			index++;
		}

		vertices.push(verticesRow);
	}

	var indices = [];

	for (var y = 0; y < heightSegments; y++) {

		for (var x = 0; x < widthSegments; x++) {

			var v1 = vertices[y][x + 1];
			var v2 = vertices[y][x];
			var v3 = vertices[y + 1][x];
			var v4 = vertices[y + 1][x + 1];

			if (y !== 0 || thetaStart > 0) indices.push(v1, v2, v4);
			if (y !== heightSegments - 1 || thetaEnd < Math.PI) indices.push(v2, v3, v4);
		}
	}

	this.setIndex(new (positions.count > 65535 ? Uint32Attribute : Uint16Attribute)(indices, 1));
	this.addAttribute('position', positions);
	this.addAttribute('normal', normals);
	this.addAttribute('uv', uvs);

	this.boundingSphere = new Sphere(new Vector3(), radius);
}

SphereBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
SphereBufferGeometry.prototype.constructor = SphereBufferGeometry;

/**
 * @author mrdoob / http://mrdoob.com/
 */

function SphereGeometry(radius, widthSegments, heightSegments, phiStart, phiLength, thetaStart, thetaLength) {

	Geometry.call(this);

	this.type = 'SphereGeometry';

	this.parameters = {
		radius: radius,
		widthSegments: widthSegments,
		heightSegments: heightSegments,
		phiStart: phiStart,
		phiLength: phiLength,
		thetaStart: thetaStart,
		thetaLength: thetaLength
	};

	this.fromBufferGeometry(new SphereBufferGeometry(radius, widthSegments, heightSegments, phiStart, phiLength, thetaStart, thetaLength));
}

SphereGeometry.prototype = Object.create(Geometry.prototype);
SphereGeometry.prototype.constructor = SphereGeometry;

function sphere$1(r, slices, rings) {
  return cache("SphereGeometry(" + r + ", " + slices + ", " + rings + ")", function () {
    return new SphereGeometry(r, slices, rings);
  });
}

function v3(x, y, z) {
  return new Vector3(x, y, z);
}

var index$1 = {
  axis: axis,
  box: box$1,
  brick: brick,
  circle: circle,
  cloud: cloud,
  colored: colored,
  cylinder: cylinder,
  hub: hub$1,
  light: light,
  material: material$1,
  put: put,
  quad: quad$1,
  range: range,
  ring: ring,
  shell: shell$1,
  raycaster: raycaster,
  sphere: sphere$1,
  textured: textured,
  v3: v3
};

var liveAPI = Object.freeze({
	axis: axis,
	box: box$1,
	brick: brick,
	circle: circle,
	cloud: cloud,
	colored: colored,
	cylinder: cylinder,
	hub: hub$1,
	light: light,
	material: material$1,
	put: put,
	quad: quad$1,
	range: range,
	ring: ring,
	shell: shell$1,
	raycaster: raycaster,
	sphere: sphere$1,
	textured: textured,
	v3: v3,
	default: index$1
});

function findProperty(elem, arr) {
  for (var i = 0; i < arr.length; ++i) {
    if (elem[arr[i]] !== undefined) {
      return arr[i];
    }
  }
}

function AsyncLockRequest(name, elementOpts, changeEventOpts, errorEventOpts, requestMethodOpts, exitMethodOpts, testExtraParam) {
  var elementName = findProperty(document, elementOpts),
      changeEventName = findProperty(document, changeEventOpts),
      errorEventName = findProperty(document, errorEventOpts),
      requestMethodName = findProperty(document.documentElement, requestMethodOpts),
      exitMethodName = findProperty(document, exitMethodOpts),
      changeTimeout = null;

  changeEventName = changeEventName && changeEventName.substring(2);
  errorEventName = errorEventName && errorEventName.substring(2);

  var ns = {
    addChangeListener: function addChangeListener(thunk, bubbles) {
      return document.addEventListener(changeEventName, thunk, bubbles);
    },
    removeChangeListener: function removeChangeListener(thunk) {
      return document.removeEventListener(changeEventName, thunk);
    },
    addErrorListener: function addErrorListener(thunk, bubbles) {
      return document.addEventListener(errorEventName, thunk, bubbles);
    },
    removeErrorListener: function removeErrorListener(thunk) {
      return document.removeEventListener(errorEventName, thunk);
    },
    withChange: function withChange(act) {
      return new Promise(function (resolve, reject) {
        var onSuccess = function onSuccess() {
          setTimeout(tearDown);
          resolve(ns.element);
        },
            onError = function onError(evt) {
          setTimeout(tearDown);
          reject(evt);
        },
            stop = function stop() {
          if (changeTimeout) {
            clearTimeout(changeTimeout);
            changeTimeout = null;
          }
        },
            tearDown = function tearDown() {
          stop();
          ns.removeChangeListener(onSuccess);
          ns.removeErrorListener(onError);
        };

        ns.addChangeListener(onSuccess, false);
        ns.addErrorListener(onError, false);

        if (act()) {
          // we've already gotten lock, so don't wait for it.
          onSuccess();
        } else {
          // Timeout waiting on the lock to happen, for systems like iOS that
          // don't properly support it, even though they say they do.
          stop();
          changeTimeout = setTimeout(function () {
            return onError(name + " state did not change in allotted time");
          }, 1000);
        }
      });
    },
    request: function request(elem, extraParam) {
      if (testExtraParam) {
        extraParam = testExtraParam(extraParam);
      }
      return ns.withChange(function () {
        if (!requestMethodName) {
          throw new Error("No " + name + " API support.");
        } else if (ns.isActive) {
          return true;
        } else if (extraParam) {
          elem[requestMethodName](extraParam);
        } else {
          elem[requestMethodName]();
        }
      });
    },
    exit: function exit() {
      return ns.withChange(function () {
        if (!exitMethodName) {
          throw new Error("No Fullscreen API support.");
        } else if (!ns.isActive) {
          return true;
        } else {
          document[exitMethodName]();
        }
      });
    }
  };

  Object.defineProperties(ns, {
    element: {
      get: function get() {
        return document[elementName];
      }
    },
    isActive: {
      get: function get() {
        return !!document[elementName];
      }
    }
  });

  return ns;
}

var PointerLock = AsyncLockRequest("Pointer Lock", ["pointerLockElement", "mozPointerLockElement", "webkitPointerLockElement"], ["onpointerlockchange", "onmozpointerlockchange", "onwebkitpointerlockchange"], ["onpointerlockerror", "onmozpointerlockerror", "onwebkitpointerlockerror"], ["requestPointerLock", "mozRequestPointerLock", "webkitRequestPointerLock", "webkitRequestPointerLock"], ["exitPointerLock", "mozExitPointerLock", "webkitExitPointerLock", "webkitExitPointerLock"]);

function testUserAgent(a) {
  return (/(android|bb\d+|meego).+|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od|ad)|iris|kindle|lge |maemo|midp|mmp|mobile.+firefox|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows (ce|phone)|xda|xiino/i.test(a) || /1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i.test(a.substring(0, 4))
  );
}

var isMobile = testUserAgent(navigator.userAgent || navigator.vendor || window.opera);

var DEG2RAD = Math.PI / 180;
var RAD2DEG = 180 / Math.PI;

var Angle = function () {
  function Angle(v) {
    classCallCheck(this, Angle);

    if (typeof v !== "number") {
      throw new Error("Angle must be initialized with a number. Initial value was: " + v);
    }

    this._value = v;
    this._delta = 0;
    this._d1 = null;
    this._d2 = null;
    this._d3 = null;
  }

  createClass(Angle, [{
    key: "degrees",
    get: function get() {
      return this._value;
    },
    set: function set(newValue) {

      do {
        // figure out if it is adding the raw value, or whole
        // rotations of the value, that results in a smaller
        // magnitude of change.
        this._d1 = newValue + this._delta - this._value;
        this._d2 = Math.abs(this._d1 + 360);
        this._d3 = Math.abs(this._d1 - 360);
        this._d1 = Math.abs(this._d1);
        if (this._d2 < this._d1 && this._d2 < this._d3) {
          this._delta += 360;
        } else if (this._d3 < this._d1) {
          this._delta -= 360;
        }
      } while (this._d1 > this._d2 || this._d1 > this._d3);
      this._value = newValue + this._delta;
    }
  }, {
    key: "radians",
    get: function get() {
      return this.degrees * DEG2RAD;
    },
    set: function set(val) {

      this.degrees = val * RAD2DEG;
    }
  }]);
  return Angle;
}();

var AbstractEventEmitter = function () {
  function AbstractEventEmitter() {
    classCallCheck(this, AbstractEventEmitter);

    this._handlers = {};
  }

  createClass(AbstractEventEmitter, [{
    key: "addEventListener",
    value: function addEventListener(evt, thunk) {

      if (!this._handlers[evt]) {
        this._handlers[evt] = [];
      }
      this._handlers[evt].push(thunk);
    }
  }, {
    key: "removeEventListener",
    value: function removeEventListener(evt, thunk) {

      if (this._handlers[evt]) {
        var idx = this._handlers[evt].indexOf(thunk);
        if (idx > -1) {
          this._handlers[evt].splice(idx, 1);
        }
      }
    }
  }, {
    key: "forward",
    value: function forward(obj, evts) {
      var _this = this;

      evts.forEach(function (evt) {
        return _this.addEventListener(evt, obj.emit.bind(obj, evt));
      });
    }
  }, {
    key: "emit",
    value: function emit(evt, obj) {

      if (this._handlers[evt]) {
        if ((typeof obj === "undefined" ? "undefined" : _typeof(obj)) === "object" && !(obj instanceof Event)) {
          obj.type = evt;
          if (obj.defaultPrevented === undefined) {
            obj.defaultPrevented = false;
            obj.preventDefault = function () {
              return obj.defaultPrevented = true;
            };
          }
        }
        var h = this._handlers[evt];
        for (var i = 0; h && i < h.length; ++i) {
          h[i](obj);
          if (obj && obj.defaultPrevented) {
            return;
          }
        }
      }
    }
  }]);
  return AbstractEventEmitter;
}();

var TELEPORT_PAD_RADIUS = 0.4;
var FORWARD = new Vector3(0, 0, -1);
var LASER_WIDTH = 0.01;
var LASER_LENGTH = 3 * LASER_WIDTH;
var GAZE_RING_DISTANCE = -1.25;
var GAZE_RING_INNER = 0.015;
var GAZE_RING_OUTER = 0.03;
var VECTOR_TEMP = new Vector3();
var EULER_TEMP = new Euler();
var QUAT_TEMP = new Quaternion();

function hasGazeEvent(obj) {
  return !!obj.ongazecomplete || !!obj.onselect || !!obj.onclick || obj._handlers && (obj._handlers.gazecomplete && obj._handlers.gazecomplete.length > 0 || obj._handlers.select && obj._handlers.select.length > 0 || obj._handlers.click && obj._handlers.click.length > 0) || obj.button && hasGazeEvent(obj.button);
}

var Pointer = function (_AbstractEventEmitter) {
  inherits(Pointer, _AbstractEventEmitter);

  function Pointer(pointerName, color, highlight, s, devices, triggerDevices, options) {
    classCallCheck(this, Pointer);

    var _this = possibleConstructorReturn(this, (Pointer.__proto__ || Object.getPrototypeOf(Pointer)).call(this));

    _this.name = pointerName;
    _this.devices = devices.filter(identity$1);
    _this.triggerDevices = triggerDevices && triggerDevices.filter(identity$1) || _this.devices.slice();
    _this.options = options;
    _this.gazeTimeout = (_this.options.gazeLength || 1.5) * 1000;

    _this.unproject = null;

    _this.picker = new Raycaster();
    _this.showPointer = true;
    _this.color = color;
    _this.highlight = highlight;
    _this.velocity = new Vector3();
    _this.mesh = colored(box$1(LASER_WIDTH / s, LASER_WIDTH / s, LASER_LENGTH * s), _this.color, {
      unshaded: true
    });
    _this.mesh.position.z = -1.5;

    _this.disk = colored(sphere$1(TELEPORT_PAD_RADIUS, 128, 3), _this.color, {
      unshaded: true
    });
    _this.disk.geometry.computeBoundingBox();
    _this.disk.geometry.vertices.forEach(function (v) {
      v.y = 0.1 * (v.y - _this.disk.geometry.boundingBox.min.y);
    });
    _this.disk.visible = false;
    _this.disk.geometry.computeBoundingBox();

    _this.gazeInner = colored(circle(GAZE_RING_INNER / 2, 10), 0xc0c0c0, {
      unshaded: true
    });
    _this.gazeInner.position.set(0, 0, GAZE_RING_DISTANCE);

    _this.gazeInner.add(colored(ring(GAZE_RING_INNER * 0.5, GAZE_RING_INNER * 0.75, 10, 36, 0, 2 * Math.PI), 0xffffff, {
      unshaded: true
    }));

    _this.gazeOuter = colored(ring(GAZE_RING_INNER, GAZE_RING_OUTER, 10, 36, 0, 2 * Math.PI), 0xffffff, {
      unshaded: true
    });
    _this.gazeOuter.visible = false;
    _this.gazeInner.add(_this.gazeOuter);

    _this.root = new Object3D();
    _this.root.add(_this.mesh);
    _this.root.add(_this.gazeInner);

    _this.useGaze = _this.options.useGaze;
    _this.lastHit = null;
    return _this;
  }

  createClass(Pointer, [{
    key: "add",
    value: function add(obj) {
      this.root.add(obj);
    }
  }, {
    key: "addDevice",
    value: function addDevice(orientation, trigger) {
      if (orientation) {
        this.devices.push(orientation);
      }

      if (trigger) {
        this.triggerDevices.push(trigger);
      }
    }
  }, {
    key: "addToBrowserEnvironment",
    value: function addToBrowserEnvironment(env, scene) {
      scene.add(this.root);
      scene.add(this.disk);
    }
  }, {
    key: "updateMatrix",
    value: function updateMatrix() {
      return this.root.updateMatrix();
    }
  }, {
    key: "applyMatrix",
    value: function applyMatrix(m) {
      return this.root.applyMatrix(m);
    }
  }, {
    key: "update",
    value: function update() {
      this.position.set(0, 0, 0);

      if (this.unproject) {
        QUAT_TEMP.set(0, 1, 0, 0);
        VECTOR_TEMP.set(0, 0, 0);
        for (var i = 0; i < this.devices.length; ++i) {
          var obj = this.devices[i];
          if (obj.enabled && !obj.commands.U.disabled && !obj.commands.V.disabled) {
            VECTOR_TEMP.x += obj.getValue("U") - 1;
            VECTOR_TEMP.y += obj.getValue("V") - 1;
          }
        }
        VECTOR_TEMP.applyMatrix4(this.unproject).applyQuaternion(QUAT_TEMP);
        this.root.lookAt(VECTOR_TEMP);
      } else {
        this.quaternion.set(0, 0, 0, 1);
        EULER_TEMP.set(0, 0, 0, "YXZ");
        for (var _i = 0; _i < this.devices.length; ++_i) {
          var _obj = this.devices[_i];
          if (_obj.enabled) {
            if (_obj.quaternion) {
              this.quaternion.multiply(_obj.quaternion);
            }
            if (_obj.position) {
              this.position.add(_obj.position);
            }
          }
        }

        QUAT_TEMP.setFromEuler(EULER_TEMP);
        this.quaternion.multiply(QUAT_TEMP);
      }
      this.root.updateMatrixWorld();
    }
  }, {
    key: "moveTeleportPad",
    value: function moveTeleportPad(point) {
      this.disk.position.copy(point);
    }
  }, {
    key: "resolvePicking",
    value: function resolvePicking(objects) {
      this.mesh.visible = false;
      this.gazeInner.visible = false;
      this.mesh.material = material$1("", {
        color: this.color,
        unshaded: true
      });

      if (this.showPointer) {
        VECTOR_TEMP.set(0, 0, 0).applyMatrix4(this.root.matrixWorld);
        FORWARD.set(0, 0, -1).applyMatrix4(this.root.matrixWorld).sub(VECTOR_TEMP);
        this.picker.set(VECTOR_TEMP, FORWARD);
        var hits = this.picker.intersectObjects(objects, true),
            currentHit = hits[0],
            lastHit = this.lastHit,
            moved = lastHit && currentHit && (currentHit.point.x !== lastHit.point.x || currentHit.point.y !== lastHit.point.y || currentHit.point.z !== lastHit.point.z),
            dt = lastHit && lastHit.time && performance.now() - lastHit.time,
            changed = !lastHit && currentHit || lastHit && !currentHit || lastHit && currentHit && currentHit.object.id !== lastHit.object.id,
            enterEvt = {
          pointer: this,
          buttons: 0,
          hit: currentHit
        },
            leaveEvt = {
          pointer: this,
          buttons: 0,
          hit: lastHit
        };

        if (currentHit) {
          this.gazeInner.position.z = 0.02 - currentHit.distance;
        } else {
          this.gazeInner.position.z = GAZE_RING_DISTANCE;
        }
        this.mesh.position.z = this.gazeInner.position.z - 0.02;

        if (currentHit) {
          currentHit.time = performance.now();

          this.mesh.material = material$1("", {
            color: this.highlight,
            unshaded: true
          });
        }

        if (moved) {
          lastHit.point.copy(currentHit.point);
        }

        this.gazeInner.visible = this.useGaze;
        this.mesh.visible = !this.useGaze;

        var dButtons = 0;
        for (var i = 0; i < this.triggerDevices.length; ++i) {
          var obj = this.triggerDevices[i];
          if (obj.enabled) {
            enterEvt.buttons |= obj.getValue("buttons");
            dButtons |= obj.getValue("dButtons");
          }
        }

        leaveEvt.buttons = enterEvt.buttons;

        if (changed) {
          if (lastHit) {
            this.emit("exit", leaveEvt);
          }
          if (currentHit) {
            this.emit("enter", enterEvt);
          }
        }

        var selected = false;
        if (dButtons) {
          if (enterEvt.buttons) {
            this.emit("pointerstart", enterEvt);
            if (lastHit) {
              lastHit.time = performance.now();
            }
          } else {
            selected = !!currentHit;
            this.emit("pointerend", enterEvt);
          }
        } else if (moved) {
          this.emit("pointermove", enterEvt);
        }

        if (this.useGaze) {
          if (changed) {
            if (dt !== null && dt < this.gazeTimeout) {
              this.gazeOuter.visible = false;
              this.emit("gazecancel", leaveEvt);
            }
            if (currentHit) {
              this.gazeOuter.visible = true;
              this.emit("gazestart", enterEvt);
            }
          } else if (dt !== null) {
            if (dt >= this.gazeTimeout) {
              this.gazeOuter.visible = false;
              selected = !!currentHit;
              this.emit("gazecomplete", enterEvt);
              lastHit.time = null;
            } else if (currentHit && currentHit.object && hasGazeEvent(currentHit.object)) {
              var p = Math.round(36 * dt / this.gazeTimeout),
                  a = 2 * Math.PI * p / 36;
              this.gazeOuter.geometry = ring(GAZE_RING_INNER, GAZE_RING_OUTER, 36, p, 0, a);
              if (moved) {
                this.emit("gazemove", enterEvt);
              }
            } else {
              if (currentHit && currentHit.object) {
                hasGazeEvent(currentHit.object);
              }
              this.gazeOuter.visible = false;
            }
          }
        }

        if (selected) {
          this.emit("select", enterEvt);
        }

        if (changed) {
          this.lastHit = currentHit;
        }
      }
    }
  }, {
    key: "material",
    get: function get() {
      return this.mesh.material;
    },
    set: function set(v) {
      this.mesh.material = v;
      this.disk.material = v;
      this.gazeInner.material = v;
      this.gazeOuter.material = v;
    }
  }, {
    key: "position",
    get: function get() {
      return this.root.position;
    }
  }, {
    key: "quaternion",
    get: function get() {
      return this.root.quaternion;
    }
  }, {
    key: "rotation",
    get: function get() {
      return this.root.rotation;
    }
  }, {
    key: "matrix",
    get: function get() {
      return this.root.matrix;
    }
  }]);
  return Pointer;
}(AbstractEventEmitter);

Pointer.EVENTS = ["pointerstart", "pointerend", "pointermove", "gazestart", "gazemove", "gazecomplete", "gazecancel", "exit", "enter", "select"];

var Keys = {
  ANY: Number.MAX_VALUE,
  ///////////////////////////////////////////////////////////////////////////
  // modifiers
  ///////////////////////////////////////////////////////////////////////////
  MODIFIER_KEYS: ["ctrl", "shift", "alt", "meta", "meta_l", "meta_r"],
  SHIFT: 16,
  CTRL: 17,
  ALT: 18,
  META: 91,
  META_L: 91,
  META_R: 92,
  ///////////////////////////////////////////////////////////////////////////
  // whitespace
  ///////////////////////////////////////////////////////////////////////////
  BACKSPACE: 8,
  TAB: 9,
  ENTER: 13,
  SPACE: 32,
  DELETE: 46,
  ///////////////////////////////////////////////////////////////////////////
  // lock keys
  ///////////////////////////////////////////////////////////////////////////
  PAUSEBREAK: 19,
  CAPSLOCK: 20,
  NUMLOCK: 144,
  SCROLLLOCK: 145,
  INSERT: 45,
  ///////////////////////////////////////////////////////////////////////////
  // navigation keys
  ///////////////////////////////////////////////////////////////////////////
  ESCAPE: 27,
  PAGEUP: 33,
  PAGEDOWN: 34,
  END: 35,
  HOME: 36,
  LEFTARROW: 37,
  UPARROW: 38,
  RIGHTARROW: 39,
  DOWNARROW: 40,
  SELECTKEY: 93,
  ///////////////////////////////////////////////////////////////////////////
  // numbers
  ///////////////////////////////////////////////////////////////////////////
  NUMBER0: 48,
  NUMBER1: 49,
  NUMBER2: 50,
  NUMBER3: 51,
  NUMBER4: 52,
  NUMBER5: 53,
  NUMBER6: 54,
  NUMBER7: 55,
  NUMBER8: 56,
  NUMBER9: 57,
  ///////////////////////////////////////////////////////////////////////////
  // letters
  ///////////////////////////////////////////////////////////////////////////
  A: 65,
  B: 66,
  C: 67,
  D: 68,
  E: 69,
  F: 70,
  G: 71,
  H: 72,
  I: 73,
  J: 74,
  K: 75,
  L: 76,
  M: 77,
  N: 78,
  O: 79,
  P: 80,
  Q: 81,
  R: 82,
  S: 83,
  T: 84,
  U: 85,
  V: 86,
  W: 87,
  X: 88,
  Y: 89,
  Z: 90,
  ///////////////////////////////////////////////////////////////////////////
  // numpad
  ///////////////////////////////////////////////////////////////////////////
  NUMPAD0: 96,
  NUMPAD1: 97,
  NUMPAD2: 98,
  NUMPAD3: 99,
  NUMPAD4: 100,
  NUMPAD5: 101,
  NUMPAD6: 102,
  NUMPAD7: 103,
  NUMPAD8: 104,
  NUMPAD9: 105,
  MULTIPLY: 106,
  ADD: 107,
  SUBTRACT: 109,
  DECIMALPOINT: 110,
  DIVIDE: 111,
  ///////////////////////////////////////////////////////////////////////////
  // function keys
  ///////////////////////////////////////////////////////////////////////////
  F1: 112,
  F2: 113,
  F3: 114,
  F4: 115,
  F5: 116,
  F6: 117,
  F7: 118,
  F8: 119,
  F9: 120,
  F10: 121,
  F11: 122,
  F12: 123,
  ///////////////////////////////////////////////////////////////////////////
  // media keys
  ///////////////////////////////////////////////////////////////////////////
  VOLUME_DOWN: 174,
  VOLUME_UP: 175,
  TRACK_NEXT: 176,
  TRACK_PREVIOUS: 177
};

// create a reverse mapping from keyCode to name.
for (var key in Keys) {
  var val = Keys[key];
  if (Keys.hasOwnProperty(key) && typeof val === "number") {
    Keys[val] = key;
  }
}

var DEFAULT_SPEECH_SETTINGS = {
  remoteVoices: true,
  volume: 1,
  rate: 1,
  pitch: 1,
  voice: 0
};

var Speech = function () {
  function Speech(options) {
    var _this = this;

    classCallCheck(this, Speech);

    this.options = Object.assign({}, DEFAULT_SPEECH_SETTINGS, options);
    if (Speech.isAvailable) {
      var getVoices = function getVoices() {
        _this.voices = speechSynthesis.getVoices().filter(function (v) {
          return _this.options.remoteVoices || v.default || v.localService;
        });
        _this.voiceNames = _this.voices.map(function (voice) {
          return voice.name;
        });
      };

      getVoices();
      speechSynthesis.onvoiceschanged = getVoices;
    }
  }

  createClass(Speech, [{
    key: "speak",
    value: function speak(txt, opts) {
      var _this2 = this;

      if (Speech.isAvailable) {
        return new Promise(function (resolve, reject) {
          var msg = new SpeechSynthesisUtterance();
          msg.voice = _this2.voices[opts && opts.voice || _this2.options.voice];
          msg.volume = opts && opts.volume || _this2.options.volume;
          msg.rate = opts && opts.rate || _this2.options.rate;
          msg.pitch = opts && opts.pitch || _this2.options.pitch;
          msg.text = txt;
          msg.onend = resolve;
          msg.onerror = reject;
          speechSynthesis.speak(msg);
        });
      } else {
        return Promise.reject();
      }
    }
  }, {
    key: "speaking",
    get: function get() {
      return Speech.isAvailable && speechSynthesis.speaking;
    }
  }], [{
    key: "isAvailable",
    get: function get() {
      return !!window.speechSynthesis;
    }
  }]);
  return Speech;
}();

var isiOS = /iP(hone|od|ad)/.test(navigator.userAgent || "");

var PositionalSound = function () {
  function PositionalSound(ctx, mainVolume) {
    classCallCheck(this, PositionalSound);

    this.gn = ctx.createGain(), this.pnr = ctx.createPanner();
    this.gn.connect(this.pnr);
    this.pnr.connect(mainVolume);
    this.gn.gain.value = 0;
    this.ctx = ctx;
  }

  createClass(PositionalSound, [{
    key: "at",
    value: function at(x, y, z, dx, dy, dz) {
      x = x || 0;
      y = y || 0;
      z = z || 0;
      if (dx === undefined || dx === null) {
        dx = 0;
      }
      dy = dy || 0;
      dz = dz || 0;

      this.pnr.setPosition(x, y, z);
      this.pnr.setOrientation(dx, dy, dz);
      return this;
    }
  }]);
  return PositionalSound;
}();

var PIANO_BASE = Math.pow(2, 1 / 12);

var Note = function (_PositionalSound) {
  inherits(Note, _PositionalSound);
  createClass(Note, null, [{
    key: "piano",
    value: function piano(n) {
      return 440 * Math.pow(PIANO_BASE, n - 49);
    }
  }]);

  function Note(ctx, mainVolume, type) {
    classCallCheck(this, Note);

    var _this = possibleConstructorReturn(this, (Note.__proto__ || Object.getPrototypeOf(Note)).call(this, ctx, mainVolume));

    _this.osc = ctx.createOscillator(), _this.osc.type = type;
    _this.osc.frequency.value = 0;
    _this.osc.connect(_this.gn);
    _this.osc.start();
    return _this;
  }

  createClass(Note, [{
    key: "on",
    value: function on(i, volume, dt, ramp) {
      if (dt === undefined) {
        dt = 0;
      }
      var f = Note.piano(parseFloat(i) + 1),
          t = this.ctx.currentTime + dt;
      this.gn.gain.setValueAtTime(volume, t);
      if (ramp) {
        this.osc.frequency.exponentialRampToValueAtTime(f, t);
      } else {
        this.osc.frequency.setValueAtTime(f, t);
      }
      return this;
    }
  }, {
    key: "off",
    value: function off(dt) {
      if (dt === undefined) {
        dt = 0;
      }
      var t = this.ctx.currentTime + dt;
      this.gn.gain.setValueAtTime(0, t);
      this.osc.frequency.setValueAtTime(0, t);
      return this;
    }
  }, {
    key: "ready",
    get: function get() {
      return this.gn.gain.value === 0;
    }
  }]);
  return Note;
}(PositionalSound);

var MAX_NOTE_COUNT = (navigator.maxTouchPoints || 10) + 1;
var TYPES = ["sine", "square", "sawtooth", "triangle"];

var Music = function () {
  function Music(audio, numNotes) {
    var _this = this;

    classCallCheck(this, Music);

    if (numNotes === undefined) {
      numNotes = MAX_NOTE_COUNT;
    }

    this.oscillators = {};
    this.isAvailable = false;
    this.audio = audio;
    this.audio.ready.then(function () {
      var ctx = _this.audio.context;
      _this.mainVolume = ctx.createGain();
      _this.mainVolume.connect(_this.audio.mainVolume);
      _this.mainVolume.gain.value = 1;
      _this.numNotes = numNotes;
      TYPES.forEach(function (type) {
        var oscs = _this.oscillators[type] = [];
        _this[type] = _this.play.bind(_this, type);
        for (var i = 0; i < _this.numNotes; ++i) {
          oscs.push(new Note(ctx, _this.mainVolume, type));
        }
      });
      _this.isAvailable = true;
    });
  }

  createClass(Music, [{
    key: "getOsc",
    value: function getOsc(type) {
      var osc = this.oscillators[type];
      var n = void 0;
      for (n = 0; n < osc.length; ++n) {
        if (osc[n].ready) {
          break;
        }
      }

      return osc[n % this.numNotes];
    }
  }, {
    key: "play",
    value: function play(type, i, volume, duration, dt) {
      if (dt === undefined) {
        dt = 0;
      }
      var o = this.getOsc(type).on(i, volume, dt);
      dt += duration;
      o.off(dt);
      dt = this.audio.context.currentTime + dt - performance.now() / 1000;
      return o;
    }
  }, {
    key: "type",
    get: function get() {
      return this._type;
    },
    set: function set(v) {
      var _this2 = this;

      if (this.isAvailable) {
        this._type = v;
        this.oscillators.forEach(function (o) {
          return o.osc.type = _this2._type;
        });
      }
    }
  }]);
  return Music;
}();

Music.TYPES = TYPES;

var Sound = function (_PositionalSound) {
  inherits(Sound, _PositionalSound);

  function Sound(audio3D, sources, loop) {
    classCallCheck(this, Sound);

    var _this = possibleConstructorReturn(this, (Sound.__proto__ || Object.getPrototypeOf(Sound)).call(this, audio3D.context, audio3D.mainVolume));

    _this.audio = document.createElement("audio");
    _this.audio.autoplay = true;
    _this.audio.preload = "auto";
    _this.audio["webkit-playsinline"] = true;
    _this.audio.playsinline = true;
    _this.audio.loop = loop;
    _this.audio.crossOrigin = "anonymous";
    console.log("Loading " + sources);
    if (!(sources instanceof Array)) {
      sources = [sources];
    }
    sources.map(function (src) {
      var source = document.createElement("source");
      source.src = src;
      return source;
    }).forEach(_this.audio.appendChild.bind(_this.audio));
    _this.ready = new Promise(function (resolve, reject) {
      _this.audio.onerror = reject;
      _this.audio.oncanplay = function () {
        _this.audio.oncanplay = null;
        _this.node = _this.ctx.createMediaElementSource(_this.audio);
        _this.node.connect(_this.gn);
        _this.gn.gain.setValueAtTime(0, _this.ctx.currentTime);
        resolve(_this);
      };
      _this.audio.play();
    });

    document.body.appendChild(_this.audio);
    return _this;
  }

  createClass(Sound, [{
    key: "play",
    value: function play() {
      this.gn.gain.setValueAtTime(1, this.ctx.currentTime);
      this.audio.play();
      return this;
    }
  }]);
  return Sound;
}(PositionalSound);

var PIXEL_SCALES = [0.5, 0.25, 0.333333, 0.5, 1];

var Quality = {
  NONE: 0,
  VERYLOW: 1,
  LOW: 2,
  MEDIUM: 3,
  HIGH: 4,
  MAXIMUM: PIXEL_SCALES.length - 1
};

var ID$1 = 1;
var NUMBER_PATTERN = "([+-]?(?:(?:\\d+(?:\\.\\d*)?)|(?:\\.\\d+)))";
var DELIM = "\\s*,\\s*";
var UNITS = "(?:em|px)";
var TRANSLATE_PATTERN = new RegExp("translate3d\\s*\\(\\s*" + NUMBER_PATTERN + UNITS + DELIM + NUMBER_PATTERN + UNITS + DELIM + NUMBER_PATTERN + UNITS + "\\s*\\)", "i");
var ROTATE_PATTERN = new RegExp("rotate3d\\s*\\(\\s*" + NUMBER_PATTERN + DELIM + NUMBER_PATTERN + DELIM + NUMBER_PATTERN + DELIM + NUMBER_PATTERN + "rad\\s*\\)", "i");
var TEMP_VECTOR = new Vector3();

var BaseControl = function (_AbstractEventEmitter) {
  inherits(BaseControl, _AbstractEventEmitter);

  function BaseControl() {
    classCallCheck(this, BaseControl);

    var _this = possibleConstructorReturn(this, (BaseControl.__proto__ || Object.getPrototypeOf(BaseControl)).call(this));

    _this.controlID = ID$1++;

    _this.focused = false;
    return _this;
  }

  createClass(BaseControl, [{
    key: "focus",
    value: function focus() {

      this.focused = true;
      this.emit("focus", {
        target: this
      });
    }
  }, {
    key: "blur",
    value: function blur() {

      this.focused = false;
      this.emit("blur", {
        target: this
      });
    }
  }, {
    key: "copyElement",
    value: function copyElement(elem) {

      this.element = elem;
      if (elem.style.transform) {
        var match = TRANSLATE_PATTERN.exec(elem.style.transform);
        if (match) {
          this.position.set(parseFloat(match[1]), parseFloat(match[2]), parseFloat(match[3]));
        }
        match = ROTATE_PATTERN.exec(elem.style.transform);
        if (match) {
          TEMP_VECTOR.set(parseFloat(match[1]), parseFloat(match[2]), parseFloat(match[3]));
          this.quaternion.setFromAxisAngle(TEMP_VECTOR, parseFloat(match[4]));
        }
      }
    }
  }]);
  return BaseControl;
}(AbstractEventEmitter);

var isOpera = !!window.opera || navigator.userAgent.indexOf(' OPR/') >= 0;

var isChrome = !!window.chrome && !isOpera;

var entityKeys = [];
var entities = new WeakMap();

var Entity = function (_AbstractEventEmitter) {
  inherits(Entity, _AbstractEventEmitter);
  createClass(Entity, null, [{
    key: "registerEntity",
    value: function registerEntity(e) {
      entities.set(e._idObj, e);
      entityKeys.push(e._idObj);
      e.addEventListener("_idchanged", function (evt) {
        entityKeys.splice(entityKeys.indexOf(evt.oldID), 1);
        entities.delete(evt.oldID);
        entities.set(evt.entity._idObj, evt.entity);
        entityKeys.push(evt.entity._idObj);
      }, false);
    }
  }, {
    key: "eyeBlankAll",
    value: function eyeBlankAll(eye) {
      entityKeys.forEach(function (id) {
        entities.get(id).eyeBlank(eye);
      });
    }
  }]);

  function Entity(id) {
    classCallCheck(this, Entity);

    var _this = possibleConstructorReturn(this, (Entity.__proto__ || Object.getPrototypeOf(Entity)).call(this));

    _this.id = id;

    _this.parent = null;

    _this.children = [];

    _this.focused = false;

    _this.focusable = true;

    return _this;
  }

  createClass(Entity, [{
    key: "focus",
    value: function focus() {
      if (this.focusable) {
        this.focused = true;
        this.emit("focus", {
          target: this
        });
      }
    }
  }, {
    key: "blur",
    value: function blur() {
      if (this.focused) {
        this.focused = false;
        for (var i = 0; i < this.children.length; ++i) {
          if (this.children[i].focused) {
            this.children[i].blur();
          }
        }
        this.emit("blur", {
          target: this
        });
      }
    }
  }, {
    key: "appendChild",
    value: function appendChild(child) {
      if (child && !child.parent) {
        child.parent = this;
        this.children.push(child);
      }
    }
  }, {
    key: "removeChild",
    value: function removeChild(child) {
      var i = this.children.indexOf(child);
      if (0 <= i && i < this.children.length) {
        this.children.splice(i, 1);
        child.parent = null;
      }
    }
  }, {
    key: "eyeBlank",
    value: function eyeBlank(eye) {
      for (var i = 0; i < this.children.length; ++i) {
        this.children[i].eyeBlank(eye);
      }
    }
  }, {
    key: "_forFocusedChild",
    value: function _forFocusedChild(name, evt) {
      var elem = this.focusedElement;
      if (elem && elem !== this) {
        elem[name](evt);
      }
    }
  }, {
    key: "startUV",
    value: function startUV(evt) {
      this._forFocusedChild("startUV", evt);
    }
  }, {
    key: "moveUV",
    value: function moveUV(evt) {
      this._forFocusedChild("moveUV", evt);
    }
  }, {
    key: "endPointer",
    value: function endPointer(evt) {
      this._forFocusedChild("endPointer", evt);
    }
  }, {
    key: "dispatchEvent",
    value: function dispatchEvent(evt) {
      var _this2 = this;

      switch (evt.type) {
        case "pointerstart":
          this.startUV(evt.hit.uv);
          break;
        case "pointerend":
          this.endPointer(evt);
          break;
        case "pointermove":
        case "gazemove":
          this.moveUV(evt.hit.uv);
          break;
        case "gazecomplete":
          this.startUV(evt.hit.uv);
          setTimeout(function () {
            return _this2.endPointer(evt);
          }, 100);
          break;
      }
    }
  }, {
    key: "keyDown",
    value: function keyDown(evt) {
      this._forFocusedChild("keyDown", evt);
    }
  }, {
    key: "keyUp",
    value: function keyUp(evt) {
      this._forFocusedChild("keyUp", evt);
    }
  }, {
    key: "readClipboard",
    value: function readClipboard(evt) {
      this._forFocusedChild("readClipboard", evt);
    }
  }, {
    key: "copySelectedText",
    value: function copySelectedText(evt) {
      this._forFocusedChild("copySelectedText", evt);
    }
  }, {
    key: "cutSelectedText",
    value: function cutSelectedText(evt) {
      this._forFocusedChild("cutSelectedText", evt);
    }
  }, {
    key: "readWheel",
    value: function readWheel(evt) {
      this._forFocusedChild("readWheel", evt);
    }
  }, {
    key: "id",
    get: function get() {
      return this._id;
    },
    set: function set(v) {
      if (this._id !== v) {
        var oldID = this._idObj;
        this._id = v;
        this._idObj = new Object(v);
        // this `_idchanged` event is necessary to update the related ID in the WeakMap of entities for eye-blanking.
        this.emit("_idchanged", {
          oldID: oldID,
          entity: this
        });
      }
    }
  }, {
    key: "theme",
    get: function get() {
      return null;
    },
    set: function set(v) {
      for (var i = 0; i < this.children.length; ++i) {
        this.children[i].theme = v;
      }
    }
  }, {
    key: "lockMovement",
    get: function get() {
      var lock = false;
      for (var i = 0; i < this.children.length && !lock; ++i) {
        lock = lock || this.children[i].lockMovement;
      }
      return lock;
    }
  }, {
    key: "focusedElement",
    get: function get() {
      var result = null,
          head = this;
      while (head && head.focused) {
        result = head;
        var children = head.children;
        head = null;
        for (var i = 0; i < children.length; ++i) {
          var child = children[i];
          if (child.focused) {
            head = child;
          }
        }
      }
      return result;
    }
  }]);
  return Entity;
}(AbstractEventEmitter);

var Point = function () {
  function Point(x, y) {
    classCallCheck(this, Point);

    this.set(x || 0, y || 0);
  }

  createClass(Point, [{
    key: "set",
    value: function set(x, y) {
      this.x = x;
      this.y = y;
    }
  }, {
    key: "copy",
    value: function copy(p) {
      if (p) {
        this.x = p.x;
        this.y = p.y;
      }
    }
  }, {
    key: "clone",
    value: function clone() {
      return new Point(this.x, this.y);
    }
  }, {
    key: "toString",
    value: function toString() {
      return "(x:" + this.x + ", y:" + this.y + ")";
    }
  }]);
  return Point;
}();

var Size = function () {
  function Size(width, height) {
    classCallCheck(this, Size);

    this.set(width || 0, height || 0);
  }

  createClass(Size, [{
    key: "set",
    value: function set(width, height) {
      this.width = width;
      this.height = height;
    }
  }, {
    key: "copy",
    value: function copy(s) {
      if (s) {
        this.width = s.width;
        this.height = s.height;
      }
    }
  }, {
    key: "clone",
    value: function clone() {
      return new Size(this.width, this.height);
    }
  }, {
    key: "toString",
    value: function toString() {
      return "<w:" + this.width + ", h:" + this.height + ">";
    }
  }]);
  return Size;
}();

var Rectangle = function () {
  function Rectangle(x, y, width, height) {
    classCallCheck(this, Rectangle);

    this.isRectangle = true;
    this.point = new Point(x, y);
    this.size = new Size(width, height);
  }

  createClass(Rectangle, [{
    key: "set",
    value: function set(x, y, width, height) {
      this.point.set(x, y);
      this.size.set(width, height);
    }
  }, {
    key: "copy",
    value: function copy(r) {
      if (r) {
        this.point.copy(r.point);
        this.size.copy(r.size);
      }
    }
  }, {
    key: "clone",
    value: function clone() {
      return new Rectangle(this.point.x, this.point.y, this.size.width, this.size.height);
    }
  }, {
    key: "toString",
    value: function toString() {
      return "[" + this.point.toString() + " x " + this.size.toString() + "]";
    }
  }, {
    key: "overlap",
    value: function overlap(r) {
      var left = Math.max(this.left, r.left),
          top = Math.max(this.top, r.top),
          right = Math.min(this.right, r.right),
          bottom = Math.min(this.bottom, r.bottom);
      if (right > left && bottom > top) {
        return new Rectangle(left, top, right - left, bottom - top);
      }
    }
  }, {
    key: "x",
    get: function get() {
      return this.point.x;
    },
    set: function set(x) {
      this.point.x = x;
    }
  }, {
    key: "left",
    get: function get() {
      return this.point.x;
    },
    set: function set(x) {
      this.point.x = x;
    }
  }, {
    key: "width",
    get: function get() {
      return this.size.width;
    },
    set: function set(width) {
      this.size.width = width;
    }
  }, {
    key: "right",
    get: function get() {
      return this.point.x + this.size.width;
    },
    set: function set(right) {
      this.point.x = right - this.size.width;
    }
  }, {
    key: "y",
    get: function get() {
      return this.point.y;
    },
    set: function set(y) {
      this.point.y = y;
    }
  }, {
    key: "top",
    get: function get() {
      return this.point.y;
    },
    set: function set(y) {
      this.point.y = y;
    }
  }, {
    key: "height",
    get: function get() {
      return this.size.height;
    },
    set: function set(height) {
      this.size.height = height;
    }
  }, {
    key: "bottom",
    get: function get() {
      return this.point.y + this.size.height;
    },
    set: function set(bottom) {
      this.point.y = bottom - this.size.height;
    }
  }, {
    key: "area",
    get: function get() {
      return this.width * this.height;
    }
  }]);
  return Rectangle;
}();

var COUNTER$2 = 0;

var Surface = function (_Entity) {
  inherits(Surface, _Entity);
  createClass(Surface, null, [{
    key: "create",
    value: function create() {
      return new Surface();
    }
  }]);

  function Surface(options) {
    classCallCheck(this, Surface);

    var _this = possibleConstructorReturn(this, (Surface.__proto__ || Object.getPrototypeOf(Surface)).call(this));

    _this.isSurface = true;
    _this.options = Object.assign({}, {
      id: "Primrose.Controls.Surface[" + COUNTER$2++ + "]",
      bounds: new Rectangle()
    }, options);
    _this.bounds = _this.options.bounds;
    _this.canvas = null;
    _this.context = null;
    _this._opacity = 1;

    _this.style = {};

    Object.defineProperties(_this.style, {
      width: {
        get: function get() {
          return _this.bounds.width;
        },
        set: function set(v) {
          _this.bounds.width = v;
          _this.resize();
        }
      },
      height: {
        get: function get() {
          return _this.bounds.height;
        },
        set: function set(v) {
          _this.bounds.height = v;
          _this.resize();
        }
      },
      left: {
        get: function get() {
          return _this.bounds.left;
        },
        set: function set(v) {
          _this.bounds.left = v;
        }
      },
      top: {
        get: function get() {
          return _this.bounds.top;
        },
        set: function set(v) {
          _this.bounds.top = v;
        }
      },
      opacity: {
        get: function get() {
          return _this._opacity;
        },
        set: function set(v) {
          _this._opacity = v;
        }
      },
      fontSize: {
        get: function get() {
          return _this.fontSize;
        },
        set: function set(v) {
          _this.fontSize = v;
        }
      },
      backgroundColor: {
        get: function get() {
          return _this.backgroundColor;
        },
        set: function set(v) {
          _this.backgroundColor = v;
        }
      },
      color: {
        get: function get() {
          return _this.color;
        },
        set: function set(v) {
          _this.color = v;
        }
      }
    });

    if (_this.options.id.isSurface) {
      throw new Error("Object is already a Surface. Please don't try to wrap them.");
    } else if (_this.options.id instanceof CanvasRenderingContext2D) {
      _this.context = _this.options.id;
      _this.canvas = _this.context.canvas;
    } else if (_this.options.id instanceof HTMLCanvasElement) {
      _this.canvas = _this.options.id;
    } else if (typeof _this.options.id === "string" || _this.options.id instanceof String) {
      _this.canvas = document.getElementById(_this.options.id);
      if (_this.canvas === null) {
        _this.canvas = document.createElement("canvas");
        _this.canvas.id = _this.options.id;
      } else if (_this.canvas.tagName !== "CANVAS") {
        _this.canvas = null;
      }
    }

    if (_this.canvas === null) {
      console.error(_typeof(_this.options.id));
      console.error(_this.options.id);
      throw new Error(_this.options.id + " does not refer to a valid canvas element.");
    }

    _this.id = _this.canvas.id;

    if (_this.bounds.width === 0) {
      _this.bounds.width = _this.imageWidth;
      _this.bounds.height = _this.imageHeight;
    }

    _this.imageWidth = _this.bounds.width;
    _this.imageHeight = _this.bounds.height;

    if (_this.context === null) {
      _this.context = _this.canvas.getContext("2d");
    }

    _this.canvas.style.imageRendering = isChrome ? "pixelated" : "optimizespeed";
    _this.context.imageSmoothingEnabled = false;
    _this.context.textBaseline = "top";

    _this._texture = null;
    _this._material = null;
    _this._environment = null;
    return _this;
  }

  createClass(Surface, [{
    key: "addToBrowserEnvironment",
    value: function addToBrowserEnvironment(env, scene) {
      this._environment = env;
      var geom = this.className === "shell" ? shell(3, 10, 10) : quad(2, 2),
          mesh = textured(geom, this, {
        opacity: this._opacity
      });
      console.log(mesh, scene, env);
      scene.add(mesh);
      env.registerPickableObject(mesh);
      return mesh;
    }
  }, {
    key: "invalidate",
    value: function invalidate(bounds) {
      var useDefault = !bounds;
      if (!bounds) {
        bounds = this.bounds.clone();
        bounds.left = 0;
        bounds.top = 0;
      } else if (bounds.isRectangle) {
        bounds = bounds.clone();
      }
      for (var i = 0; i < this.children.length; ++i) {
        var child = this.children[i],
            overlap = bounds.overlap(child.bounds);
        if (overlap) {
          var x = overlap.left - child.bounds.left,
              y = overlap.top - child.bounds.top;
          this.context.drawImage(child.canvas, x, y, overlap.width, overlap.height, overlap.x, overlap.y, overlap.width, overlap.height);
        }
      }
      if (this._texture) {
        this._texture.needsUpdate = true;
      }
      if (this._material) {
        this._material.needsUpdate = true;
      }
      if (this.parent && this.parent.invalidate) {
        bounds.left += this.bounds.left;
        bounds.top += this.bounds.top;
        this.parent.invalidate(bounds);
      }
    }
  }, {
    key: "render",
    value: function render() {
      this.invalidate();
    }
  }, {
    key: "resize",
    value: function resize() {
      this.setSize(this.surfaceWidth, this.surfaceHeight);
    }
  }, {
    key: "setSize",
    value: function setSize(width, height) {
      var oldTextBaseline = this.context.textBaseline,
          oldTextAlign = this.context.textAlign;
      this.imageWidth = width;
      this.imageHeight = height;

      this.context.textBaseline = oldTextBaseline;
      this.context.textAlign = oldTextAlign;
    }
  }, {
    key: "appendChild",
    value: function appendChild(child) {
      if (!child.isSurface) {
        throw new Error("Can only append other Surfaces to a Surface. You gave: " + child);
      }
      get$1(Surface.prototype.__proto__ || Object.getPrototypeOf(Surface.prototype), "appendChild", this).call(this, child);
      this.invalidate();
    }
  }, {
    key: "mapUV",
    value: function mapUV(point) {
      if (point instanceof Array) {
        return {
          x: point[0] * this.imageWidth,
          y: (1 - point[1]) * this.imageHeight
        };
      } else if (point.isVector2) {
        return {
          x: point.x * this.imageWidth,
          y: (1 - point.y) * this.imageHeight
        };
      }
    }
  }, {
    key: "unmapUV",
    value: function unmapUV(point) {
      return [point.x / this.imageWidth, 1 - point.y / this.imageHeight];
    }
  }, {
    key: "_findChild",
    value: function _findChild(x, y, thunk) {
      var here = this.inBounds(x, y),
          found = null;
      for (var i = this.children.length - 1; i >= 0; --i) {
        var child = this.children[i];
        if (!found && child.inBounds(x - this.bounds.left, y - this.bounds.top)) {
          found = child;
        } else if (child.focused) {
          child.blur();
        }
      }
      return found || here && this;
    }
  }, {
    key: "inBounds",
    value: function inBounds(x, y) {
      return this.bounds.left <= x && x < this.bounds.right && this.bounds.top <= y && y < this.bounds.bottom;
    }
  }, {
    key: "startPointer",
    value: function startPointer(x, y) {
      if (this.inBounds(x, y)) {
        var target = this._findChild(x, y, function (child, x2, y2) {
          return child.startPointer(x2, y2);
        });
        if (target) {
          if (!this.focused) {
            this.focus();
          }
          this.emit("click", {
            target: target,
            x: x,
            y: y
          });
          if (target !== this) {
            target.startPointer(x - this.bounds.left, y - this.bounds.top);
          }
        } else if (this.focused) {
          this.blur();
        }
      }
    }
  }, {
    key: "movePointer",
    value: function movePointer(x, y) {
      var target = this._findChild(x, y, function (child, x2, y2) {
        return child.startPointer(x2, y2);
      });
      if (target) {
        this.emit("move", {
          target: target,
          x: x,
          y: y
        });
        if (target !== this) {
          target.movePointer(x - this.bounds.left, y - this.bounds.top);
        }
      }
    }
  }, {
    key: "startUV",
    value: function startUV(point) {
      var p = this.mapUV(point);
      this.startPointer(p.x, p.y);
    }
  }, {
    key: "moveUV",
    value: function moveUV(point) {
      var p = this.mapUV(point);
      this.movePointer(p.x, p.y);
    }
  }, {
    key: "imageWidth",
    get: function get() {
      return this.canvas.width;
    },
    set: function set(v) {
      this.canvas.width = v;
      this.bounds.width = v;
    }
  }, {
    key: "imageHeight",
    get: function get() {
      return this.canvas.height;
    },
    set: function set(v) {
      this.canvas.height = v;
      this.bounds.height = v;
    }
  }, {
    key: "elementWidth",
    get: function get() {
      return this.canvas.clientWidth * devicePixelRatio;
    },
    set: function set(v) {
      this.canvas.style.width = v / devicePixelRatio + "px";
    }
  }, {
    key: "elementHeight",
    get: function get() {
      return this.canvas.clientHeight * devicePixelRatio;
    },
    set: function set(v) {
      this.canvas.style.height = v / devicePixelRatio + "px";
    }
  }, {
    key: "surfaceWidth",
    get: function get() {
      return this.canvas.parentElement ? this.elementWidth : this.bounds.width;
    }
  }, {
    key: "surfaceHeight",
    get: function get() {
      return this.canvas.parentElement ? this.elementHeight : this.bounds.height;
    }
  }, {
    key: "resized",
    get: function get() {
      return this.imageWidth !== this.surfaceWidth || this.imageHeight !== this.surfaceHeight;
    }
  }, {
    key: "texture",
    get: function get() {
      if (!this._texture) {
        this._texture = new Texture(this.canvas);
      }
      return this._texture;
    }
  }, {
    key: "environment",
    get: function get() {
      var head = this;
      while (head) {
        if (head._environment) {
          if (head !== this) {
            this._environment = head._environment;
          }
          return this._environment;
        }
        head = head.parent;
      }
    }
  }]);
  return Surface;
}(Entity);

var Default = {
  name: "Light",
  fontFamily: "'Droid Sans Mono', 'Consolas', 'Lucida Console', 'Courier New', 'Courier', monospace",
  cursorColor: "black",
  fontSize: 16,
  lineNumbers: {
    foreColor: "black"
  },
  regular: {
    backColor: "white",
    foreColor: "black",
    currentRowBackColor: "#f0f0f0",
    selectedBackColor: "#c0c0c0",
    unfocused: "rgba(0, 0, 255, 0.25)"
  },
  strings: {
    foreColor: "#aa9900",
    fontStyle: "italic"
  },
  regexes: {
    foreColor: "#aa0099",
    fontStyle: "italic"
  },
  numbers: {
    foreColor: "green"
  },
  comments: {
    foreColor: "grey",
    fontStyle: "italic"
  },
  keywords: {
    foreColor: "blue"
  },
  functions: {
    foreColor: "brown",
    fontWeight: "bold"
  },
  members: {
    foreColor: "green"
  },
  error: {
    foreColor: "red",
    fontStyle: "underline italic"
  }
};

var COUNTER$1 = 0;

var Label = function (_Surface) {
  inherits(Label, _Surface);

  function Label(options) {
    classCallCheck(this, Label);

    ////////////////////////////////////////////////////////////////////////
    // initialization
    ///////////////////////////////////////////////////////////////////////

    var _this = possibleConstructorReturn(this, (Label.__proto__ || Object.getPrototypeOf(Label)).call(this, Object.assign({}, {
      id: "Primrose.Controls.Label[" + COUNTER$1++ + "]"
    }, options)));
    ////////////////////////////////////////////////////////////////////////
    // normalize input parameters
    ////////////////////////////////////////////////////////////////////////


    _this._lastFont = null;
    _this._lastText = null;
    _this._lastCharacterWidth = null;
    _this._lastCharacterHeight = null;
    _this._lastPadding = null;
    _this._lastWidth = -1;
    _this._lastHeight = -1;
    _this._lastTextAlign = null;

    _this.textAlign = _this.options.textAlign;
    _this.character = new Size();
    _this.theme = _this.options.theme;
    _this.fontSize = _this.options.fontSize || 16;
    _this.refreshCharacter();
    _this.backgroundColor = _this.options.backgroundColor || _this.theme.regular.backColor;
    _this.color = _this.options.color || _this.theme.regular.foreColor;
    _this.value = _this.options.value;
    return _this;
  }

  createClass(Label, [{
    key: "refreshCharacter",
    value: function refreshCharacter() {
      this.character.height = this.fontSize;
      this.context.font = this.character.height + "px " + this.theme.fontFamily;
      // measure 100 letter M's, then divide by 100, to get the width of an M
      // to two decimal places on systems that return integer values from
      // measureText.
      this.character.width = this.context.measureText("MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM").width / 100;
    }
  }, {
    key: "_isChanged",
    value: function _isChanged() {
      var textChanged = this._lastText !== this.value,
          characterWidthChanged = this.character.width !== this._lastCharacterWidth,
          characterHeightChanged = this.character.height !== this._lastCharacterHeight,
          fontChanged = this.context.font !== this._lastFont,
          alignChanged = this.textAlign !== this._lastTextAlign,
          changed = this.resized || textChanged || characterWidthChanged || characterHeightChanged || this.resized || fontChanged || alignChanged;
      return changed;
    }
  }, {
    key: "render",
    value: function render() {
      if (this.resized) {
        this.resize();
      }

      if (this.theme && this._isChanged) {
        this._lastText = this.value;
        this._lastCharacterWidth = this.character.width;
        this._lastCharacterHeight = this.character.height;
        this._lastWidth = this.imageWidth;
        this._lastHeight = this.imageHeight;
        this._lastFont = this.context.font;
        this._lastTextAlign = this.textAlign;

        this.context.textAlign = this.textAlign || "left";

        var clearFunc = this.backgroundColor ? "fillRect" : "clearRect";
        if (this.theme.regular.backColor) {
          this.context.fillStyle = this.backgroundColor;
        }

        this.context[clearFunc](0, 0, this.imageWidth, this.imageHeight);

        if (this.value) {
          var lines = this.value.split("\n");
          for (var y = 0; y < lines.length; ++y) {
            var line = lines[y],
                textY = (this.imageHeight - lines.length * this.character.height) / 2 + y * this.character.height;

            var textX = null;
            switch (this.textAlign) {
              case "right":
                textX = this.imageWidth;
                break;
              case "center":
                textX = this.imageWidth / 2;
                break;
              default:
                textX = 0;
            }

            var font = (this.theme.regular.fontWeight || "") + " " + (this.theme.regular.fontStyle || "") + " " + this.character.height + "px " + this.theme.fontFamily;
            this.context.font = font.trim();
            this.context.fillStyle = this.color;
            this.context.fillText(line, textX, textY);
          }
        }

        this.renderCanvasTrim();

        this.invalidate();
      }
    }
  }, {
    key: "renderCanvasTrim",
    value: function renderCanvasTrim() {}
  }, {
    key: "textAlign",
    get: function get() {
      return this.context.textAlign;
    },
    set: function set(v) {
      this.context.textAlign = v;
      this.render();
    }
  }, {
    key: "value",
    get: function get() {
      return this._value;
    },
    set: function set(txt) {
      txt = txt || "";
      this._value = txt.replace(/\r\n/g, "\n");
      this.render();
    }
  }, {
    key: "theme",
    get: function get() {
      return this._theme;
    },
    set: function set(t) {
      this._theme = Object.assign({}, Default, t);
      this._theme.fontSize = this.fontSize;
      this.refreshCharacter();
      this.render();
    }
  }]);
  return Label;
}(Surface);

var COUNTER = 0;

var Button2D = function (_Label) {
  inherits(Button2D, _Label);
  createClass(Button2D, null, [{
    key: "create",
    value: function create() {
      return new Button2D();
    }
  }]);

  function Button2D(options) {
    classCallCheck(this, Button2D);

    var _this = possibleConstructorReturn(this, (Button2D.__proto__ || Object.getPrototypeOf(Button2D)).call(this, Object.assign({}, {
      id: "Primrose.Controls.Button2D[" + COUNTER++ + "]",
      textAlign: "center"
    }, options)));

    _this._lastActivated = null;
    return _this;
  }

  createClass(Button2D, [{
    key: "addToBrowserEnvironment",
    value: function addToBrowserEnvironment(env, scene) {
      var btn3d = env.buttonFactory.create();
      btn3d._handlers = this._handlers;
      return env.appendChild(btn3d);
    }
  }, {
    key: "startPointer",
    value: function startPointer(x, y) {
      this.focus();
      this._activated = true;
      this.render();
    }
  }, {
    key: "endPointer",
    value: function endPointer() {
      if (this._activated) {
        this._activated = false;
        this.emit("click", {
          target: this
        });
        this.render();
      }
    }
  }, {
    key: "_isChanged",
    value: function _isChanged() {
      var activatedChanged = this._activated !== this._lastActivated,
          changed = get$1(Button2D.prototype.__proto__ || Object.getPrototypeOf(Button2D.prototype), "_isChanged", this) || activatedChanged;
      return changed;
    }
  }, {
    key: "renderCanvasTrim",
    value: function renderCanvasTrim() {
      this.context.lineWidth = this._activated ? 4 : 2;
      this.context.strokeStyle = this.theme.regular.foreColor || Primrose.Text.Themes.Default.regular.foreColor;
      this.context.strokeRect(0, 0, this.imageWidth, this.imageHeight);
    }
  }]);
  return Button2D;
}(Label);

var Button3D = function (_BaseControl) {
  inherits(Button3D, _BaseControl);

  function Button3D(model, buttonName, options) {
    classCallCheck(this, Button3D);

    var _this = possibleConstructorReturn(this, (Button3D.__proto__ || Object.getPrototypeOf(Button3D)).call(this));

    _this.options = Object.assign({}, Button3D, options);
    _this.options.minDeflection = Math.cos(_this.options.minDeflection);
    _this.options.colorUnpressed = new Color(_this.options.colorUnpressed);
    _this.options.colorPressed = new Color(_this.options.colorPressed);

    _this.base = model.children[1];

    _this.cap = model.children[0];
    _this.cap.name = buttonName;
    _this.cap.material = _this.cap.material.clone();
    _this.cap.button = _this;
    _this.cap.base = _this.base;

    _this.container = new Object3D();
    _this.container.add(_this.base);
    _this.container.add(_this.cap);

    _this.color = _this.cap.material.color;

    _this.name = buttonName;

    _this.element = null;
    return _this;
  }

  createClass(Button3D, [{
    key: "startUV",
    value: function startUV(point) {

      this.color.copy(this.options.colorPressed);
      if (this.element) {
        this.element.click();
      } else {
        this.emit("click", { source: this });
      }
    }
  }, {
    key: "endPointer",
    value: function endPointer(evt) {

      this.color.copy(this.options.colorUnpressed);
      this.emit("release", { source: this });
    }
  }, {
    key: "dispatchEvent",
    value: function dispatchEvent(evt) {
      var _this2 = this;

      switch (evt.type) {
        case "pointerstart":
          this.startUV();
          break;
        case "pointerend":
          this.endPointer(evt);
          break;
        case "gazecomplete":
          this.startUV();
          setTimeout(function () {
            return _this2.endPointer(evt);
          }, 100);
          break;
      }
    }
  }, {
    key: "addToBrowserEnvironment",
    value: function addToBrowserEnvironment(env, scene) {

      scene.add(this.container);
      env.registerPickableObject(this.cap);
      return this.container;
    }
  }, {
    key: "position",
    get: function get() {

      return this.container.position;
    }
  }]);
  return Button3D;
}(BaseControl);

Button3D.DEFAULTS = {
  maxThrow: 0.1,
  minDeflection: 10,
  colorUnpressed: 0x7f0000,
  colorPressed: 0x007f00,
  toggle: true
};

var buttonCount = 0;

var ButtonFactory = function () {
  function ButtonFactory(templateFile, options) {
    classCallCheck(this, ButtonFactory);

    this.options = options;

    this.template = templateFile;
  }

  createClass(ButtonFactory, [{
    key: "create",
    value: function create(toggle) {
      var name = "button" + ++buttonCount;
      var obj = this.template.clone();
      var btn = new Button3D(obj, name, this.options, toggle);
      return btn;
    }
  }]);
  return ButtonFactory;
}();

var COUNTER$3 = 0;

var Form = function (_Surface) {
  inherits(Form, _Surface);
  createClass(Form, null, [{
    key: "create",
    value: function create() {
      return new Form();
    }
  }]);

  function Form(options) {
    classCallCheck(this, Form);

    var _this = possibleConstructorReturn(this, (Form.__proto__ || Object.getPrototypeOf(Form)).call(this, Object.assign({}, {
      id: "Primrose.Controls.Form[" + COUNTER$3++ + "]"
    }, options)));

    _this._mesh = textured(quad$1(1, _this.bounds.height / _this.bounds.width), _this);
    _this._mesh.name = _this.id + "-mesh";
    Object.defineProperties(_this.style, {
      display: {
        get: function get() {
          return _this._mesh.visible ? "" : "none";
        },
        set: function set(v) {
          if (v === "none") {
            _this.hide();
          } else {
            _this.show();
          }
        }
      },
      visible: {
        get: function get() {
          return _this._mesh.visible ? "" : "hidden";
        },
        set: function set(v) {
          return _this.visible = v !== "hidden";
        }
      }
    });
    return _this;
  }

  createClass(Form, [{
    key: "addToBrowserEnvironment",
    value: function addToBrowserEnvironment(env, scene) {
      scene.add(this._mesh);
      env.registerPickableObject(this._mesh);
      return this._mesh;
    }
  }, {
    key: "hide",
    value: function hide() {
      this.visible = false;
      this.disabled = true;
    }
  }, {
    key: "show",
    value: function show() {
      this.visible = true;
      this.disabled = false;
    }
  }, {
    key: "position",
    get: function get() {
      return this._mesh.position;
    }
  }, {
    key: "visible",
    get: function get() {
      return this._mesh.visible;
    },
    set: function set(v) {
      this._mesh.visible = v;
    }
  }, {
    key: "disabled",
    get: function get() {
      return this._mesh.disabled;
    },
    set: function set(v) {
      this._mesh.disabled = v;
    }
  }, {
    key: "enabled",
    get: function get() {
      return !this.disabled;
    },
    set: function set(v) {
      this.disabled = !v;
    }
  }]);
  return Form;
}(Surface);

var index$2 = typeof Symbol === 'undefined' ? function (description) {
	return '@' + (description || '@') + Math.random();
} : Symbol;

/*! npm.im/intervalometer */
function intervalometer(cb, request, cancel, requestParameter) {
	var requestId;
	var previousLoopTime;
	function loop(now) {
		// must be requested before cb() because that might call .stop()
		requestId = request(loop, requestParameter);

		// called with "ms since last call". 0 on start()
		cb(now - (previousLoopTime || now));

		previousLoopTime = now;
	}
	return {
		start: function start() {
			if (!requestId) {
				// prevent double starts
				loop(0);
			}
		},
		stop: function stop() {
			cancel(requestId);
			requestId = null;
			previousLoopTime = 0;
		}
	};
}

function frameIntervalometer(cb) {
	return intervalometer(cb, requestAnimationFrame, cancelAnimationFrame);
}

/*! npm.im/iphone-inline-video */
function preventEvent(element, eventName, toggleProperty, preventWithProperty) {
	function handler(e) {
		if (Boolean(element[toggleProperty]) === Boolean(preventWithProperty)) {
			e.stopImmediatePropagation();
			// console.log(eventName, 'prevented on', element);
		}
		delete element[toggleProperty];
	}
	element.addEventListener(eventName, handler, false);

	// Return handler to allow to disable the prevention. Usage:
	// const preventionHandler = preventEvent(el, 'click');
	// el.removeEventHandler('click', preventionHandler);
	return handler;
}

function proxyProperty(object, propertyName, sourceObject, copyFirst) {
	function get() {
		return sourceObject[propertyName];
	}
	function set(value) {
		sourceObject[propertyName] = value;
	}

	if (copyFirst) {
		set(object[propertyName]);
	}

	Object.defineProperty(object, propertyName, { get: get, set: set });
}

function proxyEvent(object, eventName, sourceObject) {
	sourceObject.addEventListener(eventName, function () {
		return object.dispatchEvent(new Event(eventName));
	});
}

function dispatchEventAsync(element, type) {
	Promise.resolve().then(function () {
		element.dispatchEvent(new Event(type));
	});
}

// iOS 10 adds support for native inline playback + silent autoplay
var isWhitelisted = /iPhone|iPod/i.test(navigator.userAgent) && !matchMedia('(-webkit-video-playable-inline)').matches;

var ಠ = index$2();
var ಠevent = index$2();
var ಠplay = index$2('nativeplay');
var ಠpause = index$2('nativepause');

/**
 * UTILS
 */

function getAudioFromVideo(video) {
	var audio = new Audio();
	proxyEvent(video, 'play', audio);
	proxyEvent(video, 'playing', audio);
	proxyEvent(video, 'pause', audio);
	audio.crossOrigin = video.crossOrigin;

	// 'data:' causes audio.networkState > 0
	// which then allows to keep <audio> in a resumable playing state
	// i.e. once you set a real src it will keep playing if it was if .play() was called
	audio.src = video.src || video.currentSrc || 'data:';

	// if (audio.src === 'data:') {
	//   TODO: wait for video to be selected
	// }
	return audio;
}

var lastRequests = [];
var requestIndex = 0;
var lastTimeupdateEvent;

function setTime(video, time, rememberOnly) {
	// allow one timeupdate event every 200+ ms
	if ((lastTimeupdateEvent || 0) + 200 < Date.now()) {
		video[ಠevent] = true;
		lastTimeupdateEvent = Date.now();
	}
	if (!rememberOnly) {
		video.currentTime = time;
	}
	lastRequests[++requestIndex % 3] = time * 100 | 0 / 100;
}

function isPlayerEnded(player) {
	return player.driver.currentTime >= player.video.duration;
}

function update$1(timeDiff) {
	var player = this;
	// console.log('update', player.video.readyState, player.video.networkState, player.driver.readyState, player.driver.networkState, player.driver.paused);
	if (player.video.readyState >= player.video.HAVE_FUTURE_DATA) {
		if (!player.hasAudio) {
			player.driver.currentTime = player.video.currentTime + timeDiff * player.video.playbackRate / 1000;
			if (player.video.loop && isPlayerEnded(player)) {
				player.driver.currentTime = 0;
			}
		}
		setTime(player.video, player.driver.currentTime);
	} else if (player.video.networkState === player.video.NETWORK_IDLE && !player.video.buffered.length) {
		// this should happen when the source is available but:
		// - it's potentially playing (.paused === false)
		// - it's not ready to play
		// - it's not loading
		// If it hasAudio, that will be loaded in the 'emptied' handler below
		player.video.load();
		// console.log('Will load');
	}

	// console.assert(player.video.currentTime === player.driver.currentTime, 'Video not updating!');

	if (player.video.ended) {
		delete player.video[ಠevent]; // allow timeupdate event
		player.video.pause(true);
	}
}

/**
 * METHODS
 */

function play$1() {
	// console.log('play');
	var video = this;
	var player = video[ಠ];

	// if it's fullscreen, use the native player
	if (video.webkitDisplayingFullscreen) {
		video[ಠplay]();
		return;
	}

	if (player.driver.src !== 'data:' && player.driver.src !== video.src) {
		// console.log('src changed on play', video.src);
		setTime(video, 0, true);
		player.driver.src = video.src;
	}

	if (!video.paused) {
		return;
	}
	player.paused = false;

	if (!video.buffered.length) {
		// .load() causes the emptied event
		// the alternative is .play()+.pause() but that triggers play/pause events, even worse
		// possibly the alternative is preventing this event only once
		video.load();
	}

	player.driver.play();
	player.updater.start();

	if (!player.hasAudio) {
		dispatchEventAsync(video, 'play');
		if (player.video.readyState >= player.video.HAVE_ENOUGH_DATA) {
			// console.log('onplay');
			dispatchEventAsync(video, 'playing');
		}
	}
}
function pause$1(forceEvents) {
	// console.log('pause');
	var video = this;
	var player = video[ಠ];

	player.driver.pause();
	player.updater.stop();

	// if it's fullscreen, the developer the native player.pause()
	// This is at the end of pause() because it also
	// needs to make sure that the simulation is paused
	if (video.webkitDisplayingFullscreen) {
		video[ಠpause]();
	}

	if (player.paused && !forceEvents) {
		return;
	}

	player.paused = true;
	if (!player.hasAudio) {
		dispatchEventAsync(video, 'pause');
	}
	if (video.ended) {
		video[ಠevent] = true;
		dispatchEventAsync(video, 'ended');
	}
}

/**
 * SETUP
 */

function addPlayer(video, hasAudio) {
	var player = video[ಠ] = {};
	player.paused = true; // track whether 'pause' events have been fired
	player.hasAudio = hasAudio;
	player.video = video;
	player.updater = frameIntervalometer(update$1.bind(player));

	if (hasAudio) {
		player.driver = getAudioFromVideo(video);
	} else {
		video.addEventListener('canplay', function () {
			if (!video.paused) {
				// console.log('oncanplay');
				dispatchEventAsync(video, 'playing');
			}
		});
		player.driver = {
			src: video.src || video.currentSrc || 'data:',
			muted: true,
			paused: true,
			pause: function pause$1() {
				player.driver.paused = true;
			},
			play: function play$1() {
				player.driver.paused = false;
				// media automatically goes to 0 if .play() is called when it's done
				if (isPlayerEnded(player)) {
					setTime(video, 0);
				}
			},
			get ended() {
				return isPlayerEnded(player);
			}
		};
	}

	// .load() causes the emptied event
	video.addEventListener('emptied', function () {
		// console.log('driver src is', player.driver.src);
		var wasEmpty = !player.driver.src || player.driver.src === 'data:';
		if (player.driver.src && player.driver.src !== video.src) {
			// console.log('src changed to', video.src);
			setTime(video, 0, true);
			player.driver.src = video.src;
			// playing videos will only keep playing if no src was present when .play()’ed
			if (wasEmpty) {
				player.driver.play();
			} else {
				player.updater.stop();
			}
		}
	}, false);

	// stop programmatic player when OS takes over
	video.addEventListener('webkitbeginfullscreen', function () {
		if (!video.paused) {
			// make sure that the <audio> and the syncer/updater are stopped
			video.pause();

			// play video natively
			video[ಠplay]();
		} else if (hasAudio && !player.driver.buffered.length) {
			// if the first play is native,
			// the <audio> needs to be buffered manually
			// so when the fullscreen ends, it can be set to the same current time
			player.driver.load();
		}
	});
	if (hasAudio) {
		video.addEventListener('webkitendfullscreen', function () {
			// sync audio to new video position
			player.driver.currentTime = video.currentTime;
			// console.assert(player.driver.currentTime === video.currentTime, 'Audio not synced');
		});

		// allow seeking
		video.addEventListener('seeking', function () {
			if (lastRequests.indexOf(video.currentTime * 100 | 0 / 100) < 0) {
				// console.log('User-requested seeking');
				player.driver.currentTime = video.currentTime;
			}
		});
	}
}

function overloadAPI(video) {
	var player = video[ಠ];
	video[ಠplay] = video.play;
	video[ಠpause] = video.pause;
	video.play = play$1;
	video.pause = pause$1;
	proxyProperty(video, 'paused', player.driver);
	proxyProperty(video, 'muted', player.driver, true);
	proxyProperty(video, 'playbackRate', player.driver, true);
	proxyProperty(video, 'ended', player.driver);
	proxyProperty(video, 'loop', player.driver, true);
	preventEvent(video, 'seeking');
	preventEvent(video, 'seeked');
	preventEvent(video, 'timeupdate', ಠevent, false);
	preventEvent(video, 'ended', ಠevent, false); // prevent occasional native ended events
}

function enableInlineVideo(video, hasAudio, onlyWhitelisted) {
	if (hasAudio === void 0) hasAudio = true;
	if (onlyWhitelisted === void 0) onlyWhitelisted = true;

	if (onlyWhitelisted && !isWhitelisted || video[ಠ]) {
		return;
	}
	addPlayer(video, hasAudio);
	overloadAPI(video);
	video.classList.add('IIV');
	if (!hasAudio && video.autoplay) {
		video.play();
	}
	if (!/iPhone|iPod|iPad/.test(navigator.platform)) {
		console.warn('iphone-inline-video is not guaranteed to work in emulated environments');
	}
}

enableInlineVideo.isWhitelisted = isWhitelisted;

var COUNTER$4 = 0;

// Videos don't auto-play on mobile devices, so let's make them all play whenever
// we tap the screen.
var processedVideos = [];
function findAndFixVideo(evt) {
  var vids = document.querySelectorAll("video");
  for (var i = 0; i < vids.length; ++i) {
    fixVideo(vids[i]);
  }
  window.removeEventListener("touchend", findAndFixVideo);
  window.removeEventListener("mouseup", findAndFixVideo);
  window.removeEventListener("keyup", findAndFixVideo);
}

function fixVideo(vid) {
  if (processedVideos.indexOf(vid) === -1) {
    processedVideos.push(vid);
    enableInlineVideo(vid, false);
  }
}

window.addEventListener("touchend", findAndFixVideo, false);
window.addEventListener("mouseup", findAndFixVideo, false);
window.addEventListener("keyup", findAndFixVideo, false);

var Image = function (_Entity) {
  inherits(Image, _Entity);
  createClass(Image, null, [{
    key: "create",
    value: function create() {
      return new Image();
    }
  }]);

  function Image(options) {
    classCallCheck(this, Image);

    ////////////////////////////////////////////////////////////////////////
    // normalize input parameters
    ////////////////////////////////////////////////////////////////////////

    options = options || {};
    if (typeof options === "string") {
      options = {
        value: options
      };
    }

    options = Object.assign({}, {
      id: "Primrose.Controls.Image[" + COUNTER$4++ + "]"
    }, options);

    var _this = possibleConstructorReturn(this, (Image.__proto__ || Object.getPrototypeOf(Image)).call(this, options.id));

    _this.options = options;

    Entity.registerEntity(_this);

    ////////////////////////////////////////////////////////////////////////
    // initialization
    ///////////////////////////////////////////////////////////////////////
    _this._canvases = [];
    _this._contexts = [];
    _this._elements = [];
    _this._meshes = [];
    _this._textures = [];
    _this._currentImageIndex = 0;
    _this._lastTime = null;
    _this.isVideo = false;
    return _this;
  }

  createClass(Image, [{
    key: "add",
    value: function add(child) {
      this._meshes[0].add(child);
    }
  }, {
    key: "addToBrowserEnvironment",
    value: function addToBrowserEnvironment(env, scene) {
      var _this2 = this;

      this._meshes.forEach(function (mesh, i) {
        scene.add(mesh);
        mesh.name = _this2.id + "-" + i;
        if (i > 0) {
          mesh.visible = false;
        } else if (_this2.options.pickable) {
          env.registerPickableObject(mesh);
        }
      });
    }
  }, {
    key: "_setGeometry",
    value: function _setGeometry(options) {
      if (this.options.radius) {
        this.options.geometry = shell$1(this.options.radius, 72, 36, Math.PI * 2, Math.PI, options);
      } else if (!this.options.geometry) {
        if (!this.options.width) {
          this.options.width = 0.5;
        }
        if (!this.options.height) {
          this.options.height = 0.5;
        }
        this.options.geometry = quad$1(this.options.width, this.options.height, options);
      }
    }
  }, {
    key: "loadImages",
    value: function loadImages(images, progress) {
      var _this3 = this;

      return Promise.all(Array.prototype.map.call(images, function (src, i) {
        return loadTexture$1("Primrose.Controls.Image(" + src + ", " + i + ")", src, progress).then(function (txt) {
          _this3._textures[i] = txt;
          _this3._setGeometry();
          _this3._meshes[i] = textured(_this3.options.geometry, txt, _this3.options);
        });
      })).then(function () {
        return _this3.isVideo = false;
      }).then(function () {
        return _this3;
      });
    }
  }, {
    key: "loadVideos",
    value: function loadVideos(videos, progress) {
      var _this4 = this;

      this._elements.splice(0);
      this._canvases.splice(0);
      this._contexts.splice(0);
      this._textures.splice(0);

      return Promise.all(Array.prototype.map.call(videos, function (spec, i) {
        return new Promise(function (resolve, reject) {
          var video = null;
          if (typeof spec === "string") {
            video = document.querySelector("video[src='" + spec + "']");
            if (!video) {
              video = document.createElement("video");
              video.src = spec;
            }
          } else if (spec instanceof HTMLVideoElement) {
            video = spec;
          }
          video.onprogress = progress;
          video.onloadedmetadata = progress;
          video.onerror = reject;
          video.muted = true;
          video.preload = "auto";
          video.autoplay = true;
          video.loop = true;
          video.controls = true;
          video.setAttribute("playsinline", "");
          video.setAttribute("webkit-playsinline", "");
          video.oncanplay = function () {
            video.oncanplay = null;
            video.onerror = null;

            var width = video.videoWidth,
                height = video.videoHeight,
                p2Width = Math.pow(2, Math.ceil(Math.log2(width))),
                p2Height = Math.pow(2, Math.ceil(Math.log2(height)));

            _this4._elements[i] = video;

            _this4._setGeometry({
              maxU: width / p2Width,
              maxV: height / p2Height
            });

            if ((width !== p2Width || height !== p2Height) && !_this4.options.disableVideoCopying) {
              _this4._canvases[i] = document.createElement("canvas");
              _this4._canvases[i].id = (video.id || _this4.id) + "-canvas";
              _this4._canvases[i].width = p2Width;
              _this4._canvases[i].height = p2Height;

              _this4._contexts[i] = _this4._canvases[i].getContext("2d");
            }

            _this4._meshes[i] = textured(_this4.options.geometry, _this4._canvases[i] || _this4._elements[i], _this4.options);

            _this4._textures[i] = _this4._meshes[i].material.map;
            resolve();
          };
          if (!video.parentElement) {
            document.body.insertBefore(video, document.body.children[0]);
          }
          video.play();
        });
      })).then(function () {
        return _this4.isVideo = true;
      }).then(function () {
        return _this4;
      });
    }
  }, {
    key: "eyeBlank",
    value: function eyeBlank(eye) {
      if (this._meshes) {
        this._currentImageIndex = eye % this._meshes.length;
        var isVisible = this.visible;
        for (var i = 0; i < this._meshes.length; ++i) {
          var m = this._meshes[i];
          m.visible = isVisible && i === this._currentImageIndex;
          if (i > 0) {
            m.position.copy(this.position);
            m.quaternion.copy(this.quaternion);
            m.scale.copy(this.scale);
          }
        }
      }
    }
  }, {
    key: "update",
    value: function update() {
      if (this.isVideo) {
        for (var i = 0; i < this._textures.length; ++i) {
          var elem = this._elements[i];
          if (elem.currentTime !== this._lastTime) {
            if (i < this._contexts.length) {
              this._contexts[i].drawImage(elem, 0, 0);
            }
            this._textures[i].needsUpdate = true;
            this._lastTime = elem.currentTime;
          }
        }
      }
    }
  }, {
    key: "matrixWorld",
    get: function get() {
      return this._meshes[0].matrixWorld;
    }
  }, {
    key: "position",
    get: function get() {
      return this._meshes[0].position;
    }
  }, {
    key: "quaternion",
    get: function get() {
      return this._meshes[0].quaternion;
    }
  }, {
    key: "scale",
    get: function get() {
      return this._meshes[0].scale;
    }
  }, {
    key: "visible",
    get: function get() {
      for (var i = 0; i < this._meshes.length; ++i) {
        if (this._meshes[i].visible) {
          return true;
        }
      }
      return false;
    },
    set: function set(v) {
      var img = this._meshes[this._currentImageIndex];
      if (img) {
        img.visible = v;
      }
    }
  }]);
  return Image;
}(Entity);

/**
 * @author mrdoob / http://mrdoob.com/
 * based on http://papervision3d.googlecode.com/svn/trunk/as3/trunk/src/org/papervision3d/objects/primitives/Plane.as
 */

function PlaneGeometry(width, height, widthSegments, heightSegments) {

	Geometry.call(this);

	this.type = 'PlaneGeometry';

	this.parameters = {
		width: width,
		height: height,
		widthSegments: widthSegments,
		heightSegments: heightSegments
	};

	this.fromBufferGeometry(new PlaneBufferGeometry(width, height, widthSegments, heightSegments));
}

PlaneGeometry.prototype = Object.create(Geometry.prototype);
PlaneGeometry.prototype.constructor = PlaneGeometry;

var PlainText = function PlainText(text, size, fgcolor, bgcolor, x, y, z) {
  var hAlign = arguments.length > 7 && arguments[7] !== undefined ? arguments[7] : "center";
  classCallCheck(this, PlainText);

  text = text.replace(/\r\n/g, "\n");
  var lines = text.split("\n");
  var lineHeight = size * 1000;
  var boxHeight = lineHeight * lines.length;

  var textCanvas = document.createElement("canvas");
  var textContext = textCanvas.getContext("2d");
  textContext.font = lineHeight + "px Arial";
  var width = textContext.measureText(text).width;

  textCanvas.width = width;
  textCanvas.height = boxHeight;
  textContext.font = lineHeight * 0.8 + "px Arial";
  if (bgcolor !== "transparent") {
    textContext.fillStyle = bgcolor;
    textContext.fillRect(0, 0, textCanvas.width, textCanvas.height);
  }
  textContext.fillStyle = fgcolor;

  for (var i = 0; i < lines.length; ++i) {
    textContext.fillText(lines[i], 0, i * lineHeight);
  }

  var texture = new Texture(textCanvas);
  texture.needsUpdate = true;

  var material = new MeshBasicMaterial({
    map: texture,
    transparent: bgcolor === "transparent",
    useScreenCoordinates: false,
    color: 0xffffff,
    shading: FlatShading
  });

  var textGeometry = new PlaneGeometry(size * width / lineHeight, size * lines.length);
  textGeometry.computeBoundingBox();
  textGeometry.computeVertexNormals();

  var textMesh = new Mesh(textGeometry, material);
  if (hAlign === "left") {
    x -= textGeometry.boundingBox.min.x;
  } else if (hAlign === "right") {
    x += textGeometry.boundingBox.min.x;
  }
  textMesh.position.set(x, y, z);
  return textMesh;
};

var SIZE = 1;
var INSET = 0.8;
var PROPORTION = 10;
var SIZE_SMALL = SIZE / PROPORTION;
var INSET_LARGE = 1 - (1 - INSET) / PROPORTION;

var Progress = function () {
  function Progress(majorColor, minorColor) {
    classCallCheck(this, Progress);

    majorColor = majorColor || 0xffffff;
    minorColor = minorColor || 0x000000;
    var geom = box$1(SIZE, SIZE_SMALL, SIZE_SMALL);

    this.totalBar = colored(geom, minorColor, {
      unshaded: true,
      side: BackSide
    });

    this.valueBar = colored(geom, majorColor, {
      unshaded: true
    });
    this.valueBar.scale.set(0, INSET, INSET);

    this.totalBar.add(this.valueBar);

    this.fileState = null;
    this.reset();
  }

  createClass(Progress, [{
    key: "addToBrowserEnvironment",
    value: function addToBrowserEnvironment(env, scene) {
      scene.add(this.totalBar);
    }
  }, {
    key: "reset",
    value: function reset() {
      this.fileState = {};
      this.value = 0;
    }
  }, {
    key: "onProgress",
    value: function onProgress(evt) {
      var file = evt.target.responseURL || evt.target.currentSrc;
      if (file && evt.loaded !== undefined) {
        if (!this.fileState[file]) {
          this.fileState[file] = {};
        }
        var f = this.fileState[file];
        f.loaded = evt.loaded;
        f.total = evt.total;
      }

      var total = 0,
          loaded = 0;
      for (var key in this.fileState) {
        var _f = this.fileState[key];
        total += _f.total;
        loaded += _f.loaded;
      }

      if (total > 0) {
        this.value = loaded / total;
      } else {
        this.value = 0;
      }
    }
  }, {
    key: "visible",
    get: function get() {
      return this.totalBar.visible;
    },
    set: function set(v) {
      this.totalBar.visible = v;
    }
  }, {
    key: "position",
    get: function get() {
      return this.totalBar.position;
    }
  }, {
    key: "quaternion",
    get: function get() {
      return this.totalBar.quaternion;
    }
  }, {
    key: "value",
    get: function get() {
      return this.valueBar.scale.x / INSET_LARGE;
    },
    set: function set(v) {
      this.valueBar.scale.x = v * INSET_LARGE;
      this.valueBar.position.x = -SIZE * (1 - v) * INSET_LARGE / 2;
    }
  }]);
  return Progress;
}();

// unicode-aware string reverse
var reverse = function () {
  var combiningMarks = /(<%= allExceptCombiningMarks %>)(<%= combiningMarks %>+)/g,
      surrogatePair = /(<%= highSurrogates %>)(<%= lowSurrogates %>)/g;

  function reverse(str) {
    str = str.replace(combiningMarks, function (match, capture1, capture2) {
      return reverse(capture2) + capture1;
    }).replace(surrogatePair, "$2$1");
    var res = "";
    for (var i = str.length - 1; i >= 0; --i) {
      res += str[i];
    }
    return res;
  }
  return reverse;
}();

var Cursor = function () {
  createClass(Cursor, null, [{
    key: "min",
    value: function min(a, b) {
      if (a.i <= b.i) {
        return a;
      }
      return b;
    }
  }, {
    key: "max",
    value: function max(a, b) {
      if (a.i > b.i) {
        return a;
      }
      return b;
    }
  }]);

  function Cursor(i, x, y) {
    classCallCheck(this, Cursor);

    this.i = i || 0;
    this.x = x || 0;
    this.y = y || 0;
    this.moved = true;
  }

  createClass(Cursor, [{
    key: "clone",
    value: function clone() {
      return new Cursor(this.i, this.x, this.y);
    }
  }, {
    key: "toString",
    value: function toString() {
      return "[i:" + this.i + " x:" + this.x + " y:" + this.y + "]";
    }
  }, {
    key: "copy",
    value: function copy(cursor) {
      this.i = cursor.i;
      this.x = cursor.x;
      this.y = cursor.y;
      this.moved = false;
    }
  }, {
    key: "fullhome",
    value: function fullhome() {
      this.i = 0;
      this.x = 0;
      this.y = 0;
      this.moved = true;
    }
  }, {
    key: "fullend",
    value: function fullend(lines) {
      this.i = 0;
      var lastLength = 0;
      for (var y = 0; y < lines.length; ++y) {
        var line = lines[y];
        lastLength = line.length;
        this.i += lastLength;
      }
      this.y = lines.length - 1;
      this.x = lastLength;
      this.moved = true;
    }
  }, {
    key: "skipleft",
    value: function skipleft(lines) {
      if (this.x === 0) {
        this.left(lines);
      } else {
        var x = this.x - 1;
        var line = lines[this.y];
        var word = reverse(line.substring(0, x));
        var m = word.match(/(\s|\W)+/);
        var dx = m ? m.index + m[0].length + 1 : word.length;
        this.i -= dx;
        this.x -= dx;
      }
      this.moved = true;
    }
  }, {
    key: "left",
    value: function left(lines) {
      if (this.i > 0) {
        --this.i;
        --this.x;
        if (this.x < 0) {
          --this.y;
          var line = lines[this.y];
          this.x = line.length;
        }
        if (this.reverseFromNewline(lines)) {
          ++this.i;
        }
      }
      this.moved = true;
    }
  }, {
    key: "skipright",
    value: function skipright(lines) {
      var line = lines[this.y];
      if (this.x === line.length || line[this.x] === '\n') {
        this.right(lines);
      } else {
        var x = this.x + 1;
        line = line.substring(x);
        var m = line.match(/(\s|\W)+/);
        var dx = m ? m.index + m[0].length + 1 : line.length - this.x;
        this.i += dx;
        this.x += dx;
        this.reverseFromNewline(lines);
      }
      this.moved = true;
    }
  }, {
    key: "fixCursor",
    value: function fixCursor(lines) {
      this.x = this.i;
      this.y = 0;
      var total = 0;
      var line = lines[this.y];
      while (this.x > line.length) {
        this.x -= line.length;
        total += line.length;
        if (this.y >= lines.length - 1) {
          this.i = total;
          this.x = line.length;
          this.moved = true;
          break;
        }
        ++this.y;
        line = lines[this.y];
      }
      return this.moved;
    }
  }, {
    key: "right",
    value: function right(lines) {
      this.advanceN(lines, 1);
    }
  }, {
    key: "advanceN",
    value: function advanceN(lines, n) {
      var line = lines[this.y];
      if (this.y < lines.length - 1 || this.x < line.length) {
        this.i += n;
        this.fixCursor(lines);
        line = lines[this.y];
        if (this.x > 0 && line[this.x - 1] === '\n') {
          ++this.y;
          this.x = 0;
        }
      }
      this.moved = true;
    }
  }, {
    key: "home",
    value: function home() {
      this.i -= this.x;
      this.x = 0;
      this.moved = true;
    }
  }, {
    key: "end",
    value: function end(lines) {
      var line = lines[this.y];
      var dx = line.length - this.x;
      this.i += dx;
      this.x += dx;
      this.reverseFromNewline(lines);
      this.moved = true;
    }
  }, {
    key: "up",
    value: function up(lines) {
      if (this.y > 0) {
        --this.y;
        var line = lines[this.y];
        var dx = Math.min(0, line.length - this.x);
        this.x += dx;
        this.i -= line.length - dx;
        this.reverseFromNewline(lines);
      }
      this.moved = true;
    }
  }, {
    key: "down",
    value: function down(lines) {
      if (this.y < lines.length - 1) {
        ++this.y;
        var line = lines[this.y];
        var pLine = lines[this.y - 1];
        var dx = Math.min(0, line.length - this.x);
        this.x += dx;
        this.i += pLine.length + dx;
        this.reverseFromNewline(lines);
      }
      this.moved = true;
    }
  }, {
    key: "incY",
    value: function incY(dy, lines) {
      this.y = Math.max(0, Math.min(lines.length - 1, this.y + dy));
      var line = lines[this.y];
      this.x = Math.max(0, Math.min(line.length, this.x));
      this.i = this.x;
      for (var i = 0; i < this.y; ++i) {
        this.i += lines[i].length;
      }
      this.reverseFromNewline(lines);
      this.moved = true;
    }
  }, {
    key: "setXY",
    value: function setXY(x, y, lines) {
      this.y = Math.max(0, Math.min(lines.length - 1, y));
      var line = lines[this.y];
      this.x = Math.max(0, Math.min(line.length, x));
      this.i = this.x;
      for (var i = 0; i < this.y; ++i) {
        this.i += lines[i].length;
      }
      this.reverseFromNewline(lines);
      this.moved = true;
    }
  }, {
    key: "setI",
    value: function setI(i, lines) {
      this.i = i;
      this.fixCursor(lines);
      this.moved = true;
    }
  }, {
    key: "reverseFromNewline",
    value: function reverseFromNewline(lines) {
      var line = lines[this.y];
      if (this.x > 0 && line[this.x - 1] === '\n') {
        --this.x;
        --this.i;
        return true;
      }
      return false;
    }
  }]);
  return Cursor;
}();

var CommandPack = function CommandPack(commandPackName, commands) {
  classCallCheck(this, CommandPack);

  this.name = commandPackName;
  Object.assign(this, commands);
};

var BasicTextInput = function (_CommandPack) {
  inherits(BasicTextInput, _CommandPack);

  function BasicTextInput(additionalName, additionalCommands) {
    classCallCheck(this, BasicTextInput);

    var commands = {
      NORMAL_LEFTARROW: function NORMAL_LEFTARROW(prim, tokenRows) {
        prim.cursorLeft(tokenRows, prim.frontCursor);
      },
      NORMAL_SKIPLEFT: function NORMAL_SKIPLEFT(prim, tokenRows) {
        prim.cursorSkipLeft(tokenRows, prim.frontCursor);
      },
      NORMAL_RIGHTARROW: function NORMAL_RIGHTARROW(prim, tokenRows) {
        prim.cursorRight(tokenRows, prim.frontCursor);
      },
      NORMAL_SKIPRIGHT: function NORMAL_SKIPRIGHT(prim, tokenRows) {
        prim.cursorSkipRight(tokenRows, prim.frontCursor);
      },
      NORMAL_HOME: function NORMAL_HOME(prim, tokenRows) {
        prim.cursorHome(tokenRows, prim.frontCursor);
      },
      NORMAL_END: function NORMAL_END(prim, tokenRows) {
        prim.cursorEnd(tokenRows, prim.frontCursor);
      },
      NORMAL_BACKSPACE: function NORMAL_BACKSPACE(prim, tokenRows) {
        if (prim.frontCursor.i === prim.backCursor.i) {
          prim.frontCursor.left(tokenRows);
        }
        prim.selectedText = "";
        prim.scrollIntoView(prim.frontCursor);
      },
      NORMAL_ENTER: function NORMAL_ENTER(prim, tokenRows, currentToken) {
        prim.emit("change", {
          target: prim
        });
      },
      NORMAL_DELETE: function NORMAL_DELETE(prim, tokenRows) {
        if (prim.frontCursor.i === prim.backCursor.i) {
          prim.backCursor.right(tokenRows);
        }
        prim.selectedText = "";
        prim.scrollIntoView(prim.frontCursor);
      },
      NORMAL_TAB: function NORMAL_TAB(prim, tokenRows) {
        prim.selectedText = prim.tabString;
      },

      SHIFT_LEFTARROW: function SHIFT_LEFTARROW(prim, tokenRows) {
        prim.cursorLeft(tokenRows, prim.backCursor);
      },
      SHIFT_SKIPLEFT: function SHIFT_SKIPLEFT(prim, tokenRows) {
        prim.cursorSkipLeft(tokenRows, prim.backCursor);
      },
      SHIFT_RIGHTARROW: function SHIFT_RIGHTARROW(prim, tokenRows) {
        prim.cursorRight(tokenRows, prim.backCursor);
      },
      SHIFT_SKIPRIGHT: function SHIFT_SKIPRIGHT(prim, tokenRows) {
        prim.cursorSkipRight(tokenRows, prim.backCursor);
      },
      SHIFT_HOME: function SHIFT_HOME(prim, tokenRows) {
        prim.cursorHome(tokenRows, prim.backCursor);
      },
      SHIFT_END: function SHIFT_END(prim, tokenRows) {
        prim.cursorEnd(tokenRows, prim.backCursor);
      },
      SHIFT_DELETE: function SHIFT_DELETE(prim, tokenRows) {
        if (prim.frontCursor.i === prim.backCursor.i) {
          prim.frontCursor.home(tokenRows);
          prim.backCursor.end(tokenRows);
        }
        prim.selectedText = "";
        prim.scrollIntoView(prim.frontCursor);
      },
      CTRL_HOME: function CTRL_HOME(prim, tokenRows) {
        prim.cursorFullHome(tokenRows, prim.frontCursor);
      },
      CTRL_END: function CTRL_END(prim, tokenRows) {
        prim.cursorFullEnd(tokenRows, prim.frontCursor);
      },

      CTRLSHIFT_HOME: function CTRLSHIFT_HOME(prim, tokenRows) {
        prim.cursorFullHome(tokenRows, prim.backCursor);
      },
      CTRLSHIFT_END: function CTRLSHIFT_END(prim, tokenRows) {
        prim.cursorFullEnd(tokenRows, prim.backCursor);
      },

      SELECT_ALL: function SELECT_ALL(prim, tokenRows) {
        prim.frontCursor.fullhome(tokenRows);
        prim.backCursor.fullend(tokenRows);
      },

      REDO: function REDO(prim, tokenRows) {
        prim.redo();
        prim.scrollIntoView(prim.frontCursor);
      },
      UNDO: function UNDO(prim, tokenRows) {
        prim.undo();
        prim.scrollIntoView(prim.frontCursor);
      }
    };

    if (additionalCommands) {
      for (var key in additionalCommands) {
        commands[key] = additionalCommands[key];
      }
    }

    return possibleConstructorReturn(this, (BasicTextInput.__proto__ || Object.getPrototypeOf(BasicTextInput)).call(this, additionalName || "Text editor commands", commands));
  }

  return BasicTextInput;
}(CommandPack);

var TextEditor = new BasicTextInput("Text Area input commands", {
  NORMAL_UPARROW: function NORMAL_UPARROW(prim, tokenRows) {
    prim.cursorUp(tokenRows, prim.frontCursor);
  },
  NORMAL_DOWNARROW: function NORMAL_DOWNARROW(prim, tokenRows) {
    prim.cursorDown(tokenRows, prim.frontCursor);
  },
  NORMAL_PAGEUP: function NORMAL_PAGEUP(prim, tokenRows) {
    prim.cursorPageUp(tokenRows, prim.frontCursor);
  },
  NORMAL_PAGEDOWN: function NORMAL_PAGEDOWN(prim, tokenRows) {
    prim.cursorPageDown(tokenRows, prim.frontCursor);
  },
  NORMAL_ENTER: function NORMAL_ENTER(prim, tokenRows, currentToken) {
    var indent = "";
    var tokenRow = tokenRows[prim.frontCursor.y];
    if (tokenRow.length > 0 && tokenRow[0].type === "whitespace") {
      indent = tokenRow[0].value;
    }
    prim.selectedText = "\n" + indent;
    prim.scrollIntoView(prim.frontCursor);
  },

  SHIFT_UPARROW: function SHIFT_UPARROW(prim, tokenRows) {
    prim.cursorUp(tokenRows, prim.backCursor);
  },
  SHIFT_DOWNARROW: function SHIFT_DOWNARROW(prim, tokenRows) {
    prim.cursorDown(tokenRows, prim.backCursor);
  },
  SHIFT_PAGEUP: function SHIFT_PAGEUP(prim, tokenRows) {
    prim.cursorPageUp(tokenRows, prim.backCursor);
  },
  SHIFT_PAGEDOWN: function SHIFT_PAGEDOWN(prim, tokenRows) {
    prim.cursorPageDown(tokenRows, prim.backCursor);
  },

  WINDOW_SCROLL_DOWN: function WINDOW_SCROLL_DOWN(prim, tokenRows) {
    if (prim.scroll.y < tokenRows.length) {
      ++prim.scroll.y;
    }
  },
  WINDOW_SCROLL_UP: function WINDOW_SCROLL_UP(prim, tokenRows) {
    if (prim.scroll.y > 0) {
      --prim.scroll.y;
    }
  }
});

var Rule = function () {
  function Rule(name, test) {
    classCallCheck(this, Rule);

    this.name = name;
    this.test = test;
  }

  createClass(Rule, [{
    key: "carveOutMatchedToken",
    value: function carveOutMatchedToken(tokens, j) {
      var token = tokens[j];
      if (token.type === "regular") {
        var res = this.test.exec(token.value);
        if (res) {
          // Only use the last group that matches the regex, to allow for more
          // complex regexes that can match in special contexts, but not make
          // the context part of the token.
          var midx = res[res.length - 1],
              start = res.input.indexOf(midx),
              end = start + midx.length;
          if (start === 0) {
            // the rule matches the start of the token
            token.type = this.name;
            if (end < token.value.length) {
              // but not the end
              var next = token.splitAt(end);
              next.type = "regular";
              tokens.splice(j + 1, 0, next);
            }
          } else {
            // the rule matches from the middle of the token
            var mid = token.splitAt(start);
            if (midx.length < mid.value.length) {
              // but not the end
              var right = mid.splitAt(midx.length);
              tokens.splice(j + 1, 0, right);
            }
            mid.type = this.name;
            tokens.splice(j + 1, 0, mid);
          }
        }
      }
    }
  }]);
  return Rule;
}();

var Token = function () {
  function Token(value, type, index, line) {
    classCallCheck(this, Token);

    this.value = value;
    this.type = type;
    this.index = index;
    this.line = line;
  }

  createClass(Token, [{
    key: "clone",
    value: function clone() {
      return new Token(this.value, this.type, this.index, this.line);
    }
  }, {
    key: "splitAt",
    value: function splitAt(i) {
      var next = this.value.substring(i);
      this.value = this.value.substring(0, i);
      return new Token(next, this.type, this.index + i, this.line);
    }
  }, {
    key: "toString",
    value: function toString() {
      return "[" + this.type + ": " + this.value + "]";
    }
  }]);
  return Token;
}();

var Grammar = function () {
  function Grammar(grammarName, rules) {
    classCallCheck(this, Grammar);

    this.name = grammarName;

    // clone the preprocessing grammar to start a new grammar
    this.grammar = rules.map(function (rule) {
      return new Rule(rule[0], rule[1]);
    });

    function crudeParsing(tokens) {
      var commentDelim = null,
          stringDelim = null,
          line = 0,
          i,
          t;
      for (i = 0; i < tokens.length; ++i) {
        t = tokens[i];
        t.line = line;
        if (t.type === "newlines") {
          ++line;
        }

        if (stringDelim) {
          if (t.type === "stringDelim" && t.value === stringDelim && (i === 0 || tokens[i - 1].value[tokens[i - 1].value.length - 1] !== "\\")) {
            stringDelim = null;
          }
          if (t.type !== "newlines") {
            t.type = "strings";
          }
        } else if (commentDelim) {
          if (commentDelim === "startBlockComments" && t.type === "endBlockComments" || commentDelim === "startLineComments" && t.type === "newlines") {
            commentDelim = null;
          }
          if (t.type !== "newlines") {
            t.type = "comments";
          }
        } else if (t.type === "stringDelim") {
          stringDelim = t.value;
          t.type = "strings";
        } else if (t.type === "startBlockComments" || t.type === "startLineComments") {
          commentDelim = t.type;
          t.type = "comments";
        }
      }

      // recombine like-tokens
      for (i = tokens.length - 1; i > 0; --i) {
        var p = tokens[i - 1];
        t = tokens[i];
        if (p.type === t.type && p.type !== "newlines") {
          p.value += t.value;
          tokens.splice(i, 1);
        }
      }
    }

    this.tokenize = function (text) {
      // all text starts off as regular text, then gets cut up into tokens of
      // more specific type
      var tokens = [new Token(text, "regular", 0)];
      for (var i = 0; i < this.grammar.length; ++i) {
        var rule = this.grammar[i];
        for (var j = 0; j < tokens.length; ++j) {
          rule.carveOutMatchedToken(tokens, j);
        }
      }

      crudeParsing(tokens);
      return tokens;
    };
  }

  createClass(Grammar, [{
    key: "toHTML",
    value: function toHTML(txt) {
      var theme = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : Default;

      var tokenRows = this.tokenize(txt),
          temp = document.createElement("div");
      for (var y = 0; y < tokenRows.length; ++y) {
        // draw the tokens on this row
        var t = tokenRows[y];
        if (t.type === "newlines") {
          temp.appendChild(document.createElement("br"));
        } else {
          var style = theme[t.type] || {},
              elem = document.createElement("span");
          elem.style.fontWeight = style.fontWeight || theme.regular.fontWeight;
          elem.style.fontStyle = style.fontStyle || theme.regular.fontStyle || "";
          elem.style.color = style.foreColor || theme.regular.foreColor;
          elem.style.backgroundColor = style.backColor || theme.regular.backColor;
          elem.style.fontFamily = style.fontFamily || theme.fontFamily;
          elem.appendChild(document.createTextNode(t.value));
          temp.appendChild(elem);
        }
      }
      return temp.innerHTML;
    }
  }]);
  return Grammar;
}();

var JavaScript = new Grammar("JavaScript", [["newlines", /(?:\r\n|\r|\n)/], ["startBlockComments", /\/\*/], ["endBlockComments", /\*\//], ["regexes", /(?:^|,|;|\(|\[|\{)(?:\s*)(\/(?:\\\/|[^\n\/])+\/)/], ["stringDelim", /("|')/], ["startLineComments", /\/\/.*$/m], ["numbers", /-?(?:(?:\b\d*)?\.)?\b\d+\b/], ["keywords", /\b(?:break|case|catch|class|const|continue|debugger|default|delete|do|else|export|finally|for|function|if|import|in|instanceof|let|new|return|super|switch|this|throw|try|typeof|var|void|while|with)\b/], ["functions", /(\w+)(?:\s*\()/], ["members", /(\w+)\./], ["members", /((\w+\.)+)(\w+)/]]);

var isFirefox = typeof window.InstallTrigger !== 'undefined';

navigator.userAgent.indexOf("Mobile VR") > -1;

var isIE = false || !!document.documentMode;

var isMacOS = /Macintosh/.test(navigator.userAgent || "");

var isSafari = Object.prototype.toString.call(window.HTMLElement).indexOf('Constructor') > 0;

/Windows/.test(navigator.userAgent || "");

var SCROLL_SCALE = isFirefox ? 3 : 100;
var COUNTER$5 = 0;
var OFFSET = 0;

var TextBox = function (_Surface) {
  inherits(TextBox, _Surface);
  createClass(TextBox, null, [{
    key: "create",
    value: function create() {
      return new TextBox();
    }
  }]);

  function TextBox(options) {
    classCallCheck(this, TextBox);

    var _this = possibleConstructorReturn(this, (TextBox.__proto__ || Object.getPrototypeOf(TextBox)).call(this, Object.assign({}, {
      id: "Primrose.Controls.TextBox[" + COUNTER$5++ + "]"
    }, options)));

    _this.isTextBox = true;
    ////////////////////////////////////////////////////////////////////////
    // normalize input parameters
    ////////////////////////////////////////////////////////////////////////

    if (typeof options === "string") {
      _this.options = {
        value: _this.options
      };
    } else {
      _this.options = options || {};
    }

    _this.useCaching = !isFirefox || !isMobile;

    var makeCursorCommand = function makeCursorCommand(name) {
      var method = name.toLowerCase();
      this["cursor" + name] = function (lines, cursor) {
        cursor[method](lines);
        this.scrollIntoView(cursor);
      };
    };

    ["Left", "Right", "SkipLeft", "SkipRight", "Up", "Down", "Home", "End", "FullHome", "FullEnd"].map(makeCursorCommand.bind(_this));

    ////////////////////////////////////////////////////////////////////////
    // initialization
    ///////////////////////////////////////////////////////////////////////
    _this.tokens = null;
    _this.lines = null;
    _this._commandPack = null;
    _this._tokenRows = null;
    _this._tokenHashes = null;
    _this._tabString = null;
    _this._currentTouchID = null;
    _this._lineCountWidth = null;

    _this._lastFont = null;
    _this._lastText = null;
    _this._lastCharacterWidth = null;
    _this._lastCharacterHeight = null;
    _this._lastGridBounds = null;
    _this._lastPadding = null;
    _this._lastFrontCursor = null;
    _this._lastBackCursor = null;
    _this._lastWidth = -1;
    _this._lastHeight = -1;
    _this._lastScrollX = -1;
    _this._lastScrollY = -1;
    _this._lastFocused = false;
    _this._lastThemeName = null;
    _this._lastPointer = new Point();

    // different browsers have different sets of keycodes for less-frequently
    // used keys like curly brackets.
    _this._browser = isChrome ? "CHROMIUM" : isFirefox ? "FIREFOX" : isIE ? "IE" : isOpera ? "OPERA" : isSafari ? "SAFARI" : "UNKNOWN";
    _this._pointer = new Point();
    _this._deadKeyState = "";
    _this._history = [];
    _this._historyFrame = -1;
    _this._topLeftGutter = new Size();
    _this._bottomRightGutter = new Size();
    _this._dragging = false;
    _this._scrolling = false;
    _this._wheelScrollSpeed = 4;
    var subBounds = new Rectangle(0, 0, _this.bounds.width, _this.bounds.height);
    _this._fg = new Surface({
      id: _this.id + "-fore",
      bounds: subBounds
    });
    _this._fgCanvas = _this._fg.canvas;
    _this._fgfx = _this._fg.context;
    _this._bg = new Surface({
      id: _this.id + "-back",
      bounds: subBounds
    });
    _this._bgCanvas = _this._bg.canvas;
    _this._bgfx = _this._bg.context;
    _this._trim = new Surface({
      id: _this.id + "-trim",
      bounds: subBounds
    });
    _this._trimCanvas = _this._trim.canvas;
    _this._tgfx = _this._trim.context;
    _this._rowCache = {};
    _this._VSCROLL_WIDTH = 2;

    _this.tabWidth = _this.options.tabWidth;
    _this.showLineNumbers = !_this.options.hideLineNumbers;
    _this.showScrollBars = !_this.options.hideScrollBars;
    _this.wordWrap = !_this.options.disableWordWrap;
    _this.readOnly = !!_this.options.readOnly;
    _this.multiline = !_this.options.singleLine;
    _this.gridBounds = new Rectangle();
    _this.frontCursor = new Cursor();
    _this.backCursor = new Cursor();
    _this.scroll = new Point();
    _this.character = new Size();
    _this.theme = _this.options.theme;
    _this.fontSize = _this.options.fontSize;
    _this.tokenizer = _this.options.tokenizer;
    _this.commandPack = _this.options.commands || TextEditor;
    _this.value = _this.options.value;
    _this.padding = _this.options.padding || 1;

    _this.addEventListener("focus", _this.render.bind(_this), false);
    _this.addEventListener("blur", _this.render.bind(_this), false);
    return _this;
  }

  createClass(TextBox, [{
    key: "cursorPageUp",
    value: function cursorPageUp(lines, cursor) {
      cursor.incY(-this.gridBounds.height, lines);
      this.scrollIntoView(cursor);
    }
  }, {
    key: "cursorPageDown",
    value: function cursorPageDown(lines, cursor) {
      cursor.incY(this.gridBounds.height, lines);
      this.scrollIntoView(cursor);
    }
  }, {
    key: "setDeadKeyState",
    value: function setDeadKeyState(st) {
      this._deadKeyState = st || "";
    }
  }, {
    key: "pushUndo",
    value: function pushUndo(lines) {
      if (this._historyFrame < this._history.length - 1) {
        this._history.splice(this._historyFrame + 1);
      }
      this._history.push(lines);
      this._historyFrame = this._history.length - 1;
      this.refreshTokens();
      this.render();
    }
  }, {
    key: "redo",
    value: function redo() {
      if (this._historyFrame < this._history.length - 1) {
        ++this._historyFrame;
      }
      this.refreshTokens();
      this.fixCursor();
      this.render();
    }
  }, {
    key: "undo",
    value: function undo() {
      if (this._historyFrame > 0) {
        --this._historyFrame;
      }
      this.refreshTokens();
      this.fixCursor();
      this.render();
    }
  }, {
    key: "scrollIntoView",
    value: function scrollIntoView(currentCursor) {
      this.scroll.y += this.minDelta(currentCursor.y, this.scroll.y, this.scroll.y + this.gridBounds.height);
      if (!this.wordWrap) {
        this.scroll.x += this.minDelta(currentCursor.x, this.scroll.x, this.scroll.x + this.gridBounds.width);
      }
      this.clampScroll();
    }
  }, {
    key: "readWheel",
    value: function readWheel(evt) {
      if (this.focused) {
        if (evt.shiftKey || isChrome) {
          this.fontSize += -evt.deltaX / SCROLL_SCALE;
        }
        if (!evt.shiftKey || isChrome) {
          this.scroll.y += Math.floor(evt.deltaY * this._wheelScrollSpeed / SCROLL_SCALE);
        }
        this.clampScroll();
        this.render();
        evt.preventDefault();
      }
    }
  }, {
    key: "startPointer",
    value: function startPointer(x, y) {
      if (!get$1(TextBox.prototype.__proto__ || Object.getPrototypeOf(TextBox.prototype), "startPointer", this).call(this, x, y)) {
        this._dragging = true;
        this.setCursorXY(this.frontCursor, x, y);
      }
    }
  }, {
    key: "movePointer",
    value: function movePointer(x, y) {
      if (this._dragging) {
        this.setCursorXY(this.backCursor, x, y);
      }
    }
  }, {
    key: "endPointer",
    value: function endPointer() {
      get$1(TextBox.prototype.__proto__ || Object.getPrototypeOf(TextBox.prototype), "endPointer", this).call(this);
      this._dragging = false;
      this._scrolling = false;
    }
  }, {
    key: "copySelectedText",
    value: function copySelectedText(evt) {
      if (this.focused && this.frontCursor.i !== this.backCursor.i) {
        var clipboard = evt.clipboardData || window.clipboardData;
        clipboard.setData(window.clipboardData ? "Text" : "text/plain", this.selectedText);
        evt.returnValue = false;
      }
    }
  }, {
    key: "cutSelectedText",
    value: function cutSelectedText(evt) {
      if (this.focused) {
        this.copySelectedText(evt);
        if (!this.readOnly) {
          this.selectedText = "";
        }
      }
    }
  }, {
    key: "keyDown",
    value: function keyDown(evt) {
      this.environment.input.Keyboard.doTyping(this, evt);
    }
  }, {
    key: "execCommand",
    value: function execCommand(browser, codePage, commandName) {
      if (commandName && this.focused && !this.readOnly) {
        var altCommandName = browser + "_" + commandName,
            func = this.commandPack[altCommandName] || this.commandPack[commandName] || codePage[altCommandName] || codePage[commandName];

        if (func instanceof String || typeof func === "string") {
          console.log("okay");
          func = this.commandPack[func] || this.commandPack[func] || func;
        }

        if (func === undefined) {
          return false;
        } else {
          this.frontCursor.moved = false;
          this.backCursor.moved = false;
          if (func instanceof Function) {
            func(this, this.lines);
          } else if (func instanceof String || typeof func === "string") {
            console.log(func);
            this.selectedText = func;
          }
          if (this.frontCursor.moved && !this.backCursor.moved) {
            this.backCursor.copy(this.frontCursor);
          }
          this.clampScroll();
          this.render();
          return true;
        }
      }
    }
  }, {
    key: "readClipboard",
    value: function readClipboard(evt) {
      if (this.focused && !this.readOnly) {
        evt.returnValue = false;
        var clipboard = evt.clipboardData || window.clipboardData,
            str = clipboard.getData(window.clipboardData ? "Text" : "text/plain");
        if (str) {
          this.selectedText = str;
        }
      }
    }
  }, {
    key: "resize",
    value: function resize() {
      get$1(TextBox.prototype.__proto__ || Object.getPrototypeOf(TextBox.prototype), "resize", this).call(this);
      this._bg.setSize(this.surfaceWidth, this.surfaceHeight);
      this._fg.setSize(this.surfaceWidth, this.surfaceHeight);
      this._trim.setSize(this.surfaceWidth, this.surfaceHeight);
      if (this.theme) {
        this.character.height = this.fontSize;
        this.context.font = this.character.height + "px " + this.theme.fontFamily;
        // measure 100 letter M's, then divide by 100, to get the width of an M
        // to two decimal places on systems that return integer values from
        // measureText.
        this.character.width = this.context.measureText("MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM").width / 100;
      }
      this.render();
    }
  }, {
    key: "pixel2cell",
    value: function pixel2cell(point) {
      var x = point.x * this.imageWidth / this.surfaceWidth,
          y = point.y * this.imageHeight / this.surfaceHeight;
      point.set(Math.round(point.x / this.character.width) + this.scroll.x - this.gridBounds.x, Math.floor(point.y / this.character.height - 0.25) + this.scroll.y);
    }
  }, {
    key: "clampScroll",
    value: function clampScroll() {
      if (this.scroll.y < 0) {
        this.scroll.y = 0;
      } else {
        while (0 < this.scroll.y && this.scroll.y > this.lines.length - this.gridBounds.height) {
          --this.scroll.y;
        }
      }
    }
  }, {
    key: "refreshTokens",
    value: function refreshTokens() {
      this.tokens = this.tokenizer.tokenize(this.value);
    }
  }, {
    key: "fixCursor",
    value: function fixCursor() {
      var moved = this.frontCursor.fixCursor(this.lines) || this.backCursor.fixCursor(this.lines);
      if (moved) {
        this.render();
      }
    }
  }, {
    key: "setCursorXY",
    value: function setCursorXY(cursor, x, y) {
      x = Math.round(x);
      y = Math.round(y);
      this._pointer.set(x, y);
      this.pixel2cell(this._pointer, this.scroll, this.gridBounds);
      var gx = this._pointer.x - this.scroll.x,
          gy = this._pointer.y - this.scroll.y,
          onBottom = gy >= this.gridBounds.height,
          onLeft = gx < 0,
          onRight = this._pointer.x >= this.gridBounds.width;
      if (!this._scrolling && !onBottom && !onLeft && !onRight) {
        cursor.setXY(this._pointer.x, this._pointer.y, this.lines);
        this.backCursor.copy(cursor);
      } else if (this._scrolling || onRight && !onBottom) {
        this._scrolling = true;
        var scrollHeight = this.lines.length - this.gridBounds.height;
        if (gy >= 0 && scrollHeight >= 0) {
          var sy = gy * scrollHeight / this.gridBounds.height;
          this.scroll.y = Math.floor(sy);
        }
      } else if (onBottom && !onLeft) {
        var maxWidth = 0;
        for (var dy = 0; dy < this.lines.length; ++dy) {
          maxWidth = Math.max(maxWidth, this.lines[dy].length);
        }
        var scrollWidth = maxWidth - this.gridBounds.width;
        if (gx >= 0 && scrollWidth >= 0) {
          var sx = gx * scrollWidth / this.gridBounds.width;
          this.scroll.x = Math.floor(sx);
        }
      } else if (onLeft && !onBottom) {
        // clicked in number-line gutter
      } else {
          // clicked in the lower-left corner
        }
      this._lastPointer.copy(this._pointer);
      this.render();
    }
  }, {
    key: "setGutter",
    value: function setGutter() {
      if (this.showLineNumbers) {
        this._topLeftGutter.width = 1;
      } else {
        this._topLeftGutter.width = 0;
      }

      if (!this.showScrollBars) {
        this._bottomRightGutter.set(0, 0);
      } else if (this.wordWrap) {
        this._bottomRightGutter.set(this._VSCROLL_WIDTH, 0);
      } else {
        this._bottomRightGutter.set(this._VSCROLL_WIDTH, 1);
      }
    }
  }, {
    key: "refreshGridBounds",
    value: function refreshGridBounds() {
      this._lineCountWidth = 0;
      if (this.showLineNumbers) {
        this._lineCountWidth = Math.max(1, Math.ceil(Math.log(this._history[this._historyFrame].length) / Math.LN10));
      }

      var x = Math.floor(this._topLeftGutter.width + this._lineCountWidth + this.padding / this.character.width),
          y = Math.floor(this.padding / this.character.height),
          w = Math.floor((this.imageWidth - 2 * this.padding) / this.character.width) - x - this._bottomRightGutter.width,
          h = Math.floor((this.imageHeight - 2 * this.padding) / this.character.height) - y - this._bottomRightGutter.height;
      this.gridBounds.set(x, y, w, h);
    }
  }, {
    key: "performLayout",
    value: function performLayout() {

      // group the tokens into rows
      this._tokenRows = [[]];
      this._tokenHashes = [""];
      this.lines = [""];
      var currentRowWidth = 0;
      var tokenQueue = this.tokens.slice();
      for (var i = 0; i < tokenQueue.length; ++i) {
        var t = tokenQueue[i].clone();
        var widthLeft = this.gridBounds.width - currentRowWidth;
        var wrap = this.wordWrap && t.type !== "newlines" && t.value.length > widthLeft;
        var breakLine = t.type === "newlines" || wrap;
        if (wrap) {
          var split = t.value.length > this.gridBounds.width ? widthLeft : 0;
          tokenQueue.splice(i + 1, 0, t.splitAt(split));
        }

        if (t.value.length > 0) {
          this._tokenRows[this._tokenRows.length - 1].push(t);
          this._tokenHashes[this._tokenHashes.length - 1] += JSON.stringify(t);
          this.lines[this.lines.length - 1] += t.value;
          currentRowWidth += t.value.length;
        }

        if (breakLine) {
          this._tokenRows.push([]);
          this._tokenHashes.push("");
          this.lines.push("");
          currentRowWidth = 0;
        }
      }
    }
  }, {
    key: "minDelta",
    value: function minDelta(v, minV, maxV) {
      var dvMinV = v - minV,
          dvMaxV = v - maxV + 5,
          dv = 0;
      if (dvMinV < 0 || dvMaxV >= 0) {
        // compare the absolute values, so we get the smallest change
        // regardless of direction.
        dv = Math.abs(dvMinV) < Math.abs(dvMaxV) ? dvMinV : dvMaxV;
      }

      return dv;
    }
  }, {
    key: "fillRect",
    value: function fillRect(gfx, fill, x, y, w, h) {
      gfx.fillStyle = fill;
      gfx.fillRect(x * this.character.width, y * this.character.height, w * this.character.width + 1, h * this.character.height + 1);
    }
  }, {
    key: "strokeRect",
    value: function strokeRect(gfx, stroke, x, y, w, h) {
      gfx.strokeStyle = stroke;
      gfx.strokeRect(x * this.character.width, y * this.character.height, w * this.character.width + 1, h * this.character.height + 1);
    }
  }, {
    key: "renderCanvasBackground",
    value: function renderCanvasBackground() {
      var minCursor = Cursor.min(this.frontCursor, this.backCursor),
          maxCursor = Cursor.max(this.frontCursor, this.backCursor),
          tokenFront = new Cursor(),
          tokenBack = new Cursor(),
          clearFunc = this.theme.regular.backColor ? "fillRect" : "clearRect",
          OFFSETY = OFFSET / this.character.height;

      if (this.theme.regular.backColor) {
        this._bgfx.fillStyle = this.theme.regular.backColor;
      }

      this._bgfx[clearFunc](0, 0, this.imageWidth, this.imageHeight);
      this._bgfx.save();
      this._bgfx.translate((this.gridBounds.x - this.scroll.x) * this.character.width + this.padding, -this.scroll.y * this.character.height + this.padding);

      // draw the current row highlighter
      if (this.focused) {
        this.fillRect(this._bgfx, this.theme.regular.currentRowBackColor || Default.regular.currentRowBackColor, 0, minCursor.y + OFFSETY, this.gridBounds.width, maxCursor.y - minCursor.y + 1);
      }

      for (var y = 0; y < this._tokenRows.length; ++y) {
        // draw the tokens on this row
        var row = this._tokenRows[y];

        for (var i = 0; i < row.length; ++i) {
          var t = row[i];
          tokenBack.x += t.value.length;
          tokenBack.i += t.value.length;

          // skip drawing tokens that aren't in view
          if (this.scroll.y <= y && y < this.scroll.y + this.gridBounds.height && this.scroll.x <= tokenBack.x && tokenFront.x < this.scroll.x + this.gridBounds.width) {
            // draw the selection box
            var inSelection = minCursor.i <= tokenBack.i && tokenFront.i < maxCursor.i;
            if (inSelection) {
              var selectionFront = Cursor.max(minCursor, tokenFront);
              var selectionBack = Cursor.min(maxCursor, tokenBack);
              var cw = selectionBack.i - selectionFront.i;
              this.fillRect(this._bgfx, this.theme.regular.selectedBackColor || Default.regular.selectedBackColor, selectionFront.x, selectionFront.y + OFFSETY, cw, 1);
            }
          }

          tokenFront.copy(tokenBack);
        }

        tokenFront.x = 0;
        ++tokenFront.y;
        tokenBack.copy(tokenFront);
      }

      // draw the cursor caret
      if (this.focused) {
        var cc = this.theme.cursorColor || "black";
        var w = 1 / this.character.width;
        this.fillRect(this._bgfx, cc, minCursor.x, minCursor.y + OFFSETY, w, 1);
        this.fillRect(this._bgfx, cc, maxCursor.x, maxCursor.y + OFFSETY, w, 1);
      }
      this._bgfx.restore();
    }
  }, {
    key: "renderCanvasForeground",
    value: function renderCanvasForeground() {
      var tokenFront = new Cursor(),
          tokenBack = new Cursor();

      this._fgfx.clearRect(0, 0, this.imageWidth, this.imageHeight);
      this._fgfx.save();
      this._fgfx.translate((this.gridBounds.x - this.scroll.x) * this.character.width + this.padding, this.padding);
      for (var y = 0; y < this._tokenRows.length; ++y) {
        // draw the tokens on this row
        var line = this.lines[y] + this.padding,
            row = this._tokenRows[y],
            drawn = false,
            textY = (y - this.scroll.y) * this.character.height;

        for (var i = 0; i < row.length; ++i) {
          var t = row[i];
          tokenBack.x += t.value.length;
          tokenBack.i += t.value.length;

          // skip drawing tokens that aren't in view
          if (this.scroll.y <= y && y < this.scroll.y + this.gridBounds.height && this.scroll.x <= tokenBack.x && tokenFront.x < this.scroll.x + this.gridBounds.width) {

            // draw the text
            if (this.useCaching && this._rowCache[line] !== undefined) {
              if (i === 0) {
                this._fgfx.putImageData(this._rowCache[line], this.padding, textY + this.padding + OFFSET);
              }
            } else {
              var style = this.theme[t.type] || {};
              var font = (style.fontWeight || this.theme.regular.fontWeight || "") + " " + (style.fontStyle || this.theme.regular.fontStyle || "") + " " + this.character.height + "px " + this.theme.fontFamily;
              this._fgfx.font = font.trim();
              this._fgfx.fillStyle = style.foreColor || this.theme.regular.foreColor;
              this.drawText(this._fgfx, t.value, tokenFront.x * this.character.width, textY);
              drawn = true;
            }
          }

          tokenFront.copy(tokenBack);
        }

        tokenFront.x = 0;
        ++tokenFront.y;
        tokenBack.copy(tokenFront);
        if (this.useCaching && drawn && this._rowCache[line] === undefined) {
          this._rowCache[line] = this._fgfx.getImageData(this.padding, textY + this.padding + OFFSET, this.imageWidth - 2 * this.padding, this.character.height);
        }
      }

      this._fgfx.restore();
    }

    // provides a hook for TextInput to be able to override text drawing and spit out password blanking characters

  }, {
    key: "drawText",
    value: function drawText(ctx, txt, x, y) {
      ctx.fillText(txt, x, y);
    }
  }, {
    key: "renderCanvasTrim",
    value: function renderCanvasTrim() {
      var tokenFront = new Cursor(),
          tokenBack = new Cursor(),
          maxLineWidth = 0;

      this._tgfx.clearRect(0, 0, this.imageWidth, this.imageHeight);
      this._tgfx.save();
      this._tgfx.translate(this.padding, this.padding);
      this._tgfx.save();
      this._tgfx.lineWidth = 2;
      this._tgfx.translate(0, -this.scroll.y * this.character.height);
      for (var y = 0, lastLine = -1; y < this._tokenRows.length; ++y) {
        var row = this._tokenRows[y];

        for (var i = 0; i < row.length; ++i) {
          var t = row[i];
          tokenBack.x += t.value.length;
          tokenBack.i += t.value.length;
          tokenFront.copy(tokenBack);
        }

        maxLineWidth = Math.max(maxLineWidth, tokenBack.x);
        tokenFront.x = 0;
        ++tokenFront.y;
        tokenBack.copy(tokenFront);

        if (this.showLineNumbers && this.scroll.y <= y && y < this.scroll.y + this.gridBounds.height) {
          var currentLine = row.length > 0 ? row[0].line : lastLine + 1;
          // draw the left gutter
          var lineNumber = currentLine.toString();
          while (lineNumber.length < this._lineCountWidth) {
            lineNumber = " " + lineNumber;
          }
          this.fillRect(this._tgfx, this.theme.regular.selectedBackColor || Default.regular.selectedBackColor, 0, y, this.gridBounds.x, 1);
          this._tgfx.font = "bold " + this.character.height + "px " + this.theme.fontFamily;

          if (currentLine > lastLine) {
            this._tgfx.fillStyle = this.theme.regular.foreColor;
            this._tgfx.fillText(lineNumber, 0, y * this.character.height);
          }
          lastLine = currentLine;
        }
      }

      this._tgfx.restore();

      if (this.showLineNumbers) {
        this.strokeRect(this._tgfx, this.theme.regular.foreColor || Default.regular.foreColor, 0, 0, this.gridBounds.x, this.gridBounds.height);
      }

      // draw the scrollbars
      if (this.showScrollBars) {
        var drawWidth = this.gridBounds.width * this.character.width - this.padding,
            drawHeight = this.gridBounds.height * this.character.height,
            scrollX = this.scroll.x * drawWidth / maxLineWidth + this.gridBounds.x * this.character.width,
            scrollY = this.scroll.y * drawHeight / this._tokenRows.length;

        this._tgfx.fillStyle = this.theme.regular.selectedBackColor || Default.regular.selectedBackColor;
        // horizontal
        var bw;
        if (!this.wordWrap && maxLineWidth > this.gridBounds.width) {
          var scrollBarWidth = drawWidth * (this.gridBounds.width / maxLineWidth),
              by = this.gridBounds.height * this.character.height;
          bw = Math.max(this.character.width, scrollBarWidth);
          this._tgfx.fillRect(scrollX, by, bw, this.character.height);
          this._tgfx.strokeRect(scrollX, by, bw, this.character.height);
        }

        //vertical
        if (this._tokenRows.length > this.gridBounds.height) {
          var scrollBarHeight = drawHeight * (this.gridBounds.height / this._tokenRows.length),
              bx = this.image - this._VSCROLL_WIDTH * this.character.width - 2 * this.padding,
              bh = Math.max(this.character.height, scrollBarHeight);
          bw = this._VSCROLL_WIDTH * this.character.width;
          this._tgfx.fillRect(bx, scrollY, bw, bh);
          this._tgfx.strokeRect(bx, scrollY, bw, bh);
        }
      }

      this._tgfx.lineWidth = 2;
      this._tgfx.restore();
      this._tgfx.strokeRect(1, 1, this.imageWidth - 2, this.imageHeight - 2);
      if (!this.focused) {
        this._tgfx.fillStyle = this.theme.regular.unfocused || Default.regular.unfocused;
        this._tgfx.fillRect(0, 0, this.imageWidth, this.imageHeight);
      }
    }
  }, {
    key: "render",
    value: function render() {
      if (this.tokens && this.theme) {
        this.refreshGridBounds();
        var boundsChanged = this.gridBounds.toString() !== this._lastGridBounds,
            textChanged = this._lastText !== this.value,
            characterWidthChanged = this.character.width !== this._lastCharacterWidth,
            characterHeightChanged = this.character.height !== this._lastCharacterHeight,
            paddingChanged = this.padding !== this._lastPadding,
            cursorChanged = !this._lastFrontCursor || !this._lastBackCursor || this.frontCursor.i !== this._lastFrontCursor.i || this._lastBackCursor.i !== this.backCursor.i,
            scrollChanged = this.scroll.x !== this._lastScrollX || this.scroll.y !== this._lastScrollY,
            fontChanged = this.context.font !== this._lastFont,
            themeChanged = this.theme.name !== this._lastThemeName,
            focusChanged = this.focused !== this._lastFocused,
            changeBounds = null,
            layoutChanged = this.resized || boundsChanged || textChanged || characterWidthChanged || characterHeightChanged || paddingChanged,
            backgroundChanged = layoutChanged || cursorChanged || scrollChanged || themeChanged,
            foregroundChanged = backgroundChanged || textChanged,
            trimChanged = backgroundChanged || focusChanged,
            imageChanged = foregroundChanged || backgroundChanged || trimChanged;

        if (layoutChanged) {
          this.performLayout(this.gridBounds);
          this._rowCache = {};
        }

        if (imageChanged) {
          if (cursorChanged && !(layoutChanged || scrollChanged || themeChanged || focusChanged)) {
            var top = Math.min(this.frontCursor.y, this._lastFrontCursor.y, this.backCursor.y, this._lastBackCursor.y) - this.scroll.y + this.gridBounds.y,
                bottom = Math.max(this.frontCursor.y, this._lastFrontCursor.y, this.backCursor.y, this._lastBackCursor.y) - this.scroll.y + 1;
            changeBounds = new Rectangle(0, top * this.character.height, this.bounds.width, (bottom - top) * this.character.height + 2);
          }

          if (backgroundChanged) {
            this.renderCanvasBackground();
          }
          if (foregroundChanged) {
            this.renderCanvasForeground();
          }
          if (trimChanged) {
            this.renderCanvasTrim();
          }

          this.context.clearRect(0, 0, this.imageWidth, this.imageHeight);
          this.context.drawImage(this._bgCanvas, 0, 0);
          this.context.drawImage(this._fgCanvas, 0, 0);
          this.context.drawImage(this._trimCanvas, 0, 0);
          this.invalidate(changeBounds);
        }

        this._lastGridBounds = this.gridBounds.toString();
        this._lastText = this.value;
        this._lastCharacterWidth = this.character.width;
        this._lastCharacterHeight = this.character.height;
        this._lastWidth = this.imageWidth;
        this._lastHeight = this.imageHeight;
        this._lastPadding = this.padding;
        this._lastFrontCursor = this.frontCursor.clone();
        this._lastBackCursor = this.backCursor.clone();
        this._lastFocused = this.focused;
        this._lastFont = this.context.font;
        this._lastThemeName = this.theme.name;
        this._lastScrollX = this.scroll.x;
        this._lastScrollY = this.scroll.y;
      }
    }
  }, {
    key: "value",
    get: function get() {
      return this._history[this._historyFrame].join("\n");
    },
    set: function set(txt) {
      txt = txt || "";
      txt = txt.replace(/\r\n/g, "\n");
      if (!this.multiline) {
        txt = txt.replace(/\n/g, "");
      }
      var lines = txt.split("\n");
      this.pushUndo(lines);
      this.render();
      this.emit("change", {
        target: this
      });
    }
  }, {
    key: "selectedText",
    get: function get() {
      var minCursor = Cursor.min(this.frontCursor, this.backCursor),
          maxCursor = Cursor.max(this.frontCursor, this.backCursor);
      return this.value.substring(minCursor.i, maxCursor.i);
    },
    set: function set(str) {
      str = str || "";
      str = str.replace(/\r\n/g, "\n");

      if (this.frontCursor.i !== this.backCursor.i || str.length > 0) {
        var minCursor = Cursor.min(this.frontCursor, this.backCursor),
            maxCursor = Cursor.max(this.frontCursor, this.backCursor),

        // TODO: don't recalc the string first.
        text = this.value,
            left = text.substring(0, minCursor.i),
            right = text.substring(maxCursor.i);

        var v = left + str + right;
        this.value = v;
        this.refreshGridBounds();
        this.performLayout();
        minCursor.advanceN(this.lines, Math.max(0, str.length));
        this.scrollIntoView(maxCursor);
        this.clampScroll();
        maxCursor.copy(minCursor);
        this.render();
      }
    }
  }, {
    key: "padding",
    get: function get() {
      return this._padding;
    },
    set: function set(v) {
      this._padding = v;
      this.render();
    }
  }, {
    key: "wordWrap",
    get: function get() {
      return this._wordWrap;
    },
    set: function set(v) {
      this._wordWrap = v || false;
      this.setGutter();
    }
  }, {
    key: "showLineNumbers",
    get: function get() {
      return this._showLineNumbers;
    },
    set: function set(v) {
      this._showLineNumbers = v;
      this.setGutter();
    }
  }, {
    key: "showScrollBars",
    get: function get() {
      return this._showScrollBars;
    },
    set: function set(v) {
      this._showScrollBars = v;
      this.setGutter();
    }
  }, {
    key: "theme",
    get: function get() {
      return this._theme;
    },
    set: function set(t) {
      this._theme = Object.assign({}, Default, t);
      this._theme.fontSize = this.fontSize;
      this._rowCache = {};
      this.render();
    }
  }, {
    key: "commandPack",
    get: function get() {
      return this._commandPack;
    },
    set: function set(v) {
      this._commandPack = v;
    }
  }, {
    key: "selectionStart",
    get: function get() {
      return this.frontCursor.i;
    },
    set: function set(i) {
      this.frontCursor.setI(i, this.lines);
    }
  }, {
    key: "selectionEnd",
    get: function get() {
      return this.backCursor.i;
    },
    set: function set(i) {
      this.backCursor.setI(i, this.lines);
    }
  }, {
    key: "selectionDirection",
    get: function get() {
      return this.frontCursor.i <= this.backCursor.i ? "forward" : "backward";
    }
  }, {
    key: "tokenizer",
    get: function get() {
      return this._tokenizer;
    },
    set: function set(tk) {
      this._tokenizer = tk || JavaScript;
      if (this._history && this._history.length > 0) {
        this.refreshTokens();
        this.render();
      }
    }
  }, {
    key: "tabWidth",
    get: function get() {
      return this._tabWidth;
    },
    set: function set(tw) {
      this._tabWidth = tw || 2;
      this._tabString = "";
      for (var i = 0; i < this._tabWidth; ++i) {
        this._tabString += " ";
      }
    }
  }, {
    key: "tabString",
    get: function get() {
      return this._tabString;
    }
  }, {
    key: "fontSize",
    get: function get() {
      return this._fontSize || 16;
    },
    set: function set(v) {
      v = v || 16;
      this._fontSize = v;
      if (this.theme) {
        this.theme.fontSize = this._fontSize;
        this.resize();
        this.render();
      }
    }
  }, {
    key: "lockMovement",
    get: function get() {
      return this.focused && !this.readOnly;
    }
  }]);
  return TextBox;
}(Surface);

var PlainText$2 = new Grammar("PlainText", [["newlines", /(?:\r\n|\r|\n)/]]);

////
// For all of these commands, the "current" cursor is:
// If SHIFT is not held, then "front".
// If SHIFT is held, then "back"
//
var TextInput$2 = new BasicTextInput("Text Line input commands");

var COUNTER$6 = 0;

var TextInput = function (_TextBox) {
  inherits(TextInput, _TextBox);

  function TextInput(options) {
    classCallCheck(this, TextInput);

    var _this = possibleConstructorReturn(this, (TextInput.__proto__ || Object.getPrototypeOf(TextInput)).call(this, Object.assign({}, {
      id: "Primrose.Controls.TextInput[" + COUNTER$6++ + "]",
      padding: 5,
      singleLine: true,
      disableWordWrap: true,
      hideLineNumbers: true,
      hideScrollBars: true,
      tabWidth: 1,
      tokenizer: PlainText$2,
      commands: TextInput$2
    }), options));

    _this.passwordCharacter = _this.options.passwordCharacter;
    return _this;
  }

  createClass(TextInput, [{
    key: "drawText",
    value: function drawText(ctx, txt, x, y) {
      if (this.passwordCharacter) {
        var val = "";
        for (var i = 0; i < txt.length; ++i) {
          val += this.passwordCharacter;
        }
        txt = val;
      }
      get$1(TextInput.prototype.__proto__ || Object.getPrototypeOf(TextInput.prototype), "drawText", this).call(this, ctx, txt, x, y);
    }
  }, {
    key: "value",
    get: function get() {
      return get$1(TextInput.prototype.__proto__ || Object.getPrototypeOf(TextInput.prototype), "value", this);
    },
    set: function set(v) {
      v = v || "";
      v = v.replace(/\r?\n/g, "");
      set$1(TextInput.prototype.__proto__ || Object.getPrototypeOf(TextInput.prototype), "value", v, this);
    }
  }, {
    key: "selectedText",
    get: function get() {
      return get$1(TextInput.prototype.__proto__ || Object.getPrototypeOf(TextInput.prototype), "selectedText", this);
    },
    set: function set(v) {
      v = v || "";
      v = v.replace(/\r?\n/g, "");
      set$1(TextInput.prototype.__proto__ || Object.getPrototypeOf(TextInput.prototype), "selectedText", v, this);
    }
  }]);
  return TextInput;
}(TextBox);

function cascadeElement(id, tag, DOMClass, add) {
  var elem = null;
  if (id === null) {
    elem = document.createElement(tag);
    elem.id = id = "auto_" + tag + Date.now();
  } else if (DOMClass === undefined || id instanceof DOMClass) {
    elem = id;
  } else if (typeof id === "string") {
    elem = document.getElementById(id);
    if (elem === null) {
      elem = document.createElement(tag);
      elem.id = id;
      if (add) {
        document.body.appendChild(elem);
      }
    } else if (elem.tagName !== tag.toUpperCase()) {
      elem = null;
    }
  }

  if (elem === null) {
    throw new Error(id + " does not refer to a valid " + tag + " element.");
  }
  return elem;
}

function makeHidingContainer(id, obj) {
  var elem = cascadeElement(id, "div", window.HTMLDivElement);
  elem.style.position = "absolute";
  elem.style.left = 0;
  elem.style.top = 0;
  elem.style.width = 0;
  elem.style.height = 0;
  elem.style.overflow = "hidden";
  elem.appendChild(obj);
  return elem;
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function Group() {

	Object3D.call(this);

	this.type = 'Group';
}

Group.prototype = Object.assign(Object.create(Object3D.prototype), {

	constructor: Group

});

/**
 * @author alteredq / http://alteredqualia.com/
 *
 * parameters = {
 *  color: <hex>,
 *  opacity: <float>,
 *  map: new THREE.Texture( <Image> ),
 *
 *	uvOffset: new THREE.Vector2(),
 *	uvScale: new THREE.Vector2()
 * }
 */

function SpriteMaterial(parameters) {

	Material.call(this);

	this.type = 'SpriteMaterial';

	this.color = new Color(0xffffff);
	this.map = null;

	this.rotation = 0;

	this.fog = false;
	this.lights = false;

	this.setValues(parameters);
}

SpriteMaterial.prototype = Object.create(Material.prototype);
SpriteMaterial.prototype.constructor = SpriteMaterial;

SpriteMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.color.copy(source.color);
	this.map = source.map;

	this.rotation = source.rotation;

	return this;
};

/**
 * @author mikael emtinger / http://gomo.se/
 * @author alteredq / http://alteredqualia.com/
 */

function Sprite(material) {

	Object3D.call(this);

	this.type = 'Sprite';

	this.material = material !== undefined ? material : new SpriteMaterial();
}

Sprite.prototype = Object.assign(Object.create(Object3D.prototype), {

	constructor: Sprite,

	isSprite: true,

	raycast: function () {

		var matrixPosition = new Vector3();

		return function raycast(raycaster, intersects) {

			matrixPosition.setFromMatrixPosition(this.matrixWorld);

			var distanceSq = raycaster.ray.distanceSqToPoint(matrixPosition);
			var guessSizeSq = this.scale.x * this.scale.y / 4;

			if (distanceSq > guessSizeSq) {

				return;
			}

			intersects.push({

				distance: Math.sqrt(distanceSq),
				point: this.position,
				face: null,
				object: this

			});
		};
	}(),

	clone: function clone() {

		return new this.constructor(this.material).copy(this);
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 *
 * parameters = {
 *  color: <hex>,
 *  opacity: <float>,
 *
 *  linewidth: <float>,
 *  linecap: "round",
 *  linejoin: "round"
 * }
 */

function LineBasicMaterial(parameters) {

	Material.call(this);

	this.type = 'LineBasicMaterial';

	this.color = new Color(0xffffff);

	this.linewidth = 1;
	this.linecap = 'round';
	this.linejoin = 'round';

	this.lights = false;

	this.setValues(parameters);
}

LineBasicMaterial.prototype = Object.create(Material.prototype);
LineBasicMaterial.prototype.constructor = LineBasicMaterial;

LineBasicMaterial.prototype.isLineBasicMaterial = true;

LineBasicMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.color.copy(source.color);

	this.linewidth = source.linewidth;
	this.linecap = source.linecap;
	this.linejoin = source.linejoin;

	return this;
};

/**
 * @author mrdoob / http://mrdoob.com/
 */

function LineSegments(geometry, material) {

	Line.call(this, geometry, material);

	this.type = 'LineSegments';
}

LineSegments.prototype = Object.assign(Object.create(Line.prototype), {

	constructor: LineSegments,

	isLineSegments: true

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function Line(geometry, material, mode) {

	if (mode === 1) {

		console.warn('THREE.Line: parameter THREE.LinePieces no longer supported. Created THREE.LineSegments instead.');
		return new LineSegments(geometry, material);
	}

	Object3D.call(this);

	this.type = 'Line';

	this.geometry = geometry !== undefined ? geometry : new BufferGeometry();
	this.material = material !== undefined ? material : new LineBasicMaterial({ color: Math.random() * 0xffffff });
}

Line.prototype = Object.assign(Object.create(Object3D.prototype), {

	constructor: Line,

	isLine: true,

	raycast: function () {

		var inverseMatrix = new Matrix4();
		var ray = new Ray();
		var sphere = new Sphere();

		return function raycast(raycaster, intersects) {

			var precision = raycaster.linePrecision;
			var precisionSq = precision * precision;

			var geometry = this.geometry;
			var matrixWorld = this.matrixWorld;

			// Checking boundingSphere distance to ray

			if (geometry.boundingSphere === null) geometry.computeBoundingSphere();

			sphere.copy(geometry.boundingSphere);
			sphere.applyMatrix4(matrixWorld);

			if (raycaster.ray.intersectsSphere(sphere) === false) return;

			//

			inverseMatrix.getInverse(matrixWorld);
			ray.copy(raycaster.ray).applyMatrix4(inverseMatrix);

			var vStart = new Vector3();
			var vEnd = new Vector3();
			var interSegment = new Vector3();
			var interRay = new Vector3();
			var step = this && this.isLineSegments ? 2 : 1;

			if (geometry && geometry.isBufferGeometry) {

				var index = geometry.index;
				var attributes = geometry.attributes;
				var positions = attributes.position.array;

				if (index !== null) {

					var indices = index.array;

					for (var i = 0, l = indices.length - 1; i < l; i += step) {

						var a = indices[i];
						var b = indices[i + 1];

						vStart.fromArray(positions, a * 3);
						vEnd.fromArray(positions, b * 3);

						var distSq = ray.distanceSqToSegment(vStart, vEnd, interRay, interSegment);

						if (distSq > precisionSq) continue;

						interRay.applyMatrix4(this.matrixWorld); //Move back to world space for distance calculation

						var distance = raycaster.ray.origin.distanceTo(interRay);

						if (distance < raycaster.near || distance > raycaster.far) continue;

						intersects.push({

							distance: distance,
							// What do we want? intersection point on the ray or on the segment??
							// point: raycaster.ray.at( distance ),
							point: interSegment.clone().applyMatrix4(this.matrixWorld),
							index: i,
							face: null,
							faceIndex: null,
							object: this

						});
					}
				} else {

					for (var i = 0, l = positions.length / 3 - 1; i < l; i += step) {

						vStart.fromArray(positions, 3 * i);
						vEnd.fromArray(positions, 3 * i + 3);

						var distSq = ray.distanceSqToSegment(vStart, vEnd, interRay, interSegment);

						if (distSq > precisionSq) continue;

						interRay.applyMatrix4(this.matrixWorld); //Move back to world space for distance calculation

						var distance = raycaster.ray.origin.distanceTo(interRay);

						if (distance < raycaster.near || distance > raycaster.far) continue;

						intersects.push({

							distance: distance,
							// What do we want? intersection point on the ray or on the segment??
							// point: raycaster.ray.at( distance ),
							point: interSegment.clone().applyMatrix4(this.matrixWorld),
							index: i,
							face: null,
							faceIndex: null,
							object: this

						});
					}
				}
			} else if (geometry && geometry.isGeometry) {

				var vertices = geometry.vertices;
				var nbVertices = vertices.length;

				for (var i = 0; i < nbVertices - 1; i += step) {

					var distSq = ray.distanceSqToSegment(vertices[i], vertices[i + 1], interRay, interSegment);

					if (distSq > precisionSq) continue;

					interRay.applyMatrix4(this.matrixWorld); //Move back to world space for distance calculation

					var distance = raycaster.ray.origin.distanceTo(interRay);

					if (distance < raycaster.near || distance > raycaster.far) continue;

					intersects.push({

						distance: distance,
						// What do we want? intersection point on the ray or on the segment??
						// point: raycaster.ray.at( distance ),
						point: interSegment.clone().applyMatrix4(this.matrixWorld),
						index: i,
						face: null,
						faceIndex: null,
						object: this

					});
				}
			}
		};
	}(),

	clone: function clone() {

		return new this.constructor(this.geometry, this.material).copy(this);
	}

});

/**
 * @author mikael emtinger / http://gomo.se/
 * @author alteredq / http://alteredqualia.com/
 * @author mrdoob / http://mrdoob.com/
 */

function LOD() {

	Object3D.call(this);

	this.type = 'LOD';

	Object.defineProperties(this, {
		levels: {
			enumerable: true,
			value: []
		}
	});
}

LOD.prototype = Object.assign(Object.create(Object3D.prototype), {

	constructor: LOD,

	copy: function copy(source) {

		Object3D.prototype.copy.call(this, source, false);

		var levels = source.levels;

		for (var i = 0, l = levels.length; i < l; i++) {

			var level = levels[i];

			this.addLevel(level.object.clone(), level.distance);
		}

		return this;
	},

	addLevel: function addLevel(object, distance) {

		if (distance === undefined) distance = 0;

		distance = Math.abs(distance);

		var levels = this.levels;

		for (var l = 0; l < levels.length; l++) {

			if (distance < levels[l].distance) {

				break;
			}
		}

		levels.splice(l, 0, { distance: distance, object: object });

		this.add(object);
	},

	getObjectForDistance: function getObjectForDistance(distance) {

		var levels = this.levels;

		for (var i = 1, l = levels.length; i < l; i++) {

			if (distance < levels[i].distance) {

				break;
			}
		}

		return levels[i - 1].object;
	},

	raycast: function () {

		var matrixPosition = new Vector3();

		return function raycast(raycaster, intersects) {

			matrixPosition.setFromMatrixPosition(this.matrixWorld);

			var distance = raycaster.ray.origin.distanceTo(matrixPosition);

			this.getObjectForDistance(distance).raycast(raycaster, intersects);
		};
	}(),

	update: function () {

		var v1 = new Vector3();
		var v2 = new Vector3();

		return function update(camera) {

			var levels = this.levels;

			if (levels.length > 1) {

				v1.setFromMatrixPosition(camera.matrixWorld);
				v2.setFromMatrixPosition(this.matrixWorld);

				var distance = v1.distanceTo(v2);

				levels[0].object.visible = true;

				for (var i = 1, l = levels.length; i < l; i++) {

					if (distance >= levels[i].distance) {

						levels[i - 1].object.visible = false;
						levels[i].object.visible = true;
					} else {

						break;
					}
				}

				for (; i < l; i++) {

					levels[i].object.visible = false;
				}
			}
		};
	}(),

	toJSON: function toJSON(meta) {

		var data = Object3D.prototype.toJSON.call(this, meta);

		data.object.levels = [];

		var levels = this.levels;

		for (var i = 0, l = levels.length; i < l; i++) {

			var level = levels[i];

			data.object.levels.push({
				object: level.object.uuid,
				distance: level.distance
			});
		}

		return data;
	}

});

/**
 * @author alteredq / http://alteredqualia.com/
 */

function DataTexture(data, width, height, format, type, mapping, wrapS, wrapT, magFilter, minFilter, anisotropy, encoding) {

	Texture.call(this, null, mapping, wrapS, wrapT, magFilter, minFilter, format, type, anisotropy, encoding);

	this.image = { data: data, width: width, height: height };

	this.magFilter = magFilter !== undefined ? magFilter : NearestFilter;
	this.minFilter = minFilter !== undefined ? minFilter : NearestFilter;

	this.generateMipmaps = false;
	this.flipY = false;
	this.unpackAlignment = 1;
}

DataTexture.prototype = Object.create(Texture.prototype);
DataTexture.prototype.constructor = DataTexture;

DataTexture.prototype.isDataTexture = true;

/**
 * @author mikael emtinger / http://gomo.se/
 * @author alteredq / http://alteredqualia.com/
 * @author michael guerrero / http://realitymeltdown.com
 * @author ikerr / http://verold.com
 */

function Skeleton(bones, boneInverses, useVertexTexture) {

	this.useVertexTexture = useVertexTexture !== undefined ? useVertexTexture : true;

	this.identityMatrix = new Matrix4();

	// copy the bone array

	bones = bones || [];

	this.bones = bones.slice(0);

	// create a bone texture or an array of floats

	if (this.useVertexTexture) {

		// layout (1 matrix = 4 pixels)
		//      RGBA RGBA RGBA RGBA (=> column1, column2, column3, column4)
		//  with  8x8  pixel texture max   16 bones * 4 pixels =  (8 * 8)
		//       16x16 pixel texture max   64 bones * 4 pixels = (16 * 16)
		//       32x32 pixel texture max  256 bones * 4 pixels = (32 * 32)
		//       64x64 pixel texture max 1024 bones * 4 pixels = (64 * 64)


		var size = Math.sqrt(this.bones.length * 4); // 4 pixels needed for 1 matrix
		size = _Math.nextPowerOfTwo(Math.ceil(size));
		size = Math.max(size, 4);

		this.boneTextureWidth = size;
		this.boneTextureHeight = size;

		this.boneMatrices = new Float32Array(this.boneTextureWidth * this.boneTextureHeight * 4); // 4 floats per RGBA pixel
		this.boneTexture = new DataTexture(this.boneMatrices, this.boneTextureWidth, this.boneTextureHeight, RGBAFormat, FloatType);
	} else {

		this.boneMatrices = new Float32Array(16 * this.bones.length);
	}

	// use the supplied bone inverses or calculate the inverses

	if (boneInverses === undefined) {

		this.calculateInverses();
	} else {

		if (this.bones.length === boneInverses.length) {

			this.boneInverses = boneInverses.slice(0);
		} else {

			console.warn('THREE.Skeleton bonInverses is the wrong length.');

			this.boneInverses = [];

			for (var b = 0, bl = this.bones.length; b < bl; b++) {

				this.boneInverses.push(new Matrix4());
			}
		}
	}
}

Object.assign(Skeleton.prototype, {

	calculateInverses: function calculateInverses() {

		this.boneInverses = [];

		for (var b = 0, bl = this.bones.length; b < bl; b++) {

			var inverse = new Matrix4();

			if (this.bones[b]) {

				inverse.getInverse(this.bones[b].matrixWorld);
			}

			this.boneInverses.push(inverse);
		}
	},

	pose: function pose() {

		var bone;

		// recover the bind-time world matrices

		for (var b = 0, bl = this.bones.length; b < bl; b++) {

			bone = this.bones[b];

			if (bone) {

				bone.matrixWorld.getInverse(this.boneInverses[b]);
			}
		}

		// compute the local matrices, positions, rotations and scales

		for (var b = 0, bl = this.bones.length; b < bl; b++) {

			bone = this.bones[b];

			if (bone) {

				if (bone.parent && bone.parent.isBone) {

					bone.matrix.getInverse(bone.parent.matrixWorld);
					bone.matrix.multiply(bone.matrixWorld);
				} else {

					bone.matrix.copy(bone.matrixWorld);
				}

				bone.matrix.decompose(bone.position, bone.quaternion, bone.scale);
			}
		}
	},

	update: function () {

		var offsetMatrix = new Matrix4();

		return function update() {

			// flatten bone matrices to array

			for (var b = 0, bl = this.bones.length; b < bl; b++) {

				// compute the offset between the current and the original transform

				var matrix = this.bones[b] ? this.bones[b].matrixWorld : this.identityMatrix;

				offsetMatrix.multiplyMatrices(matrix, this.boneInverses[b]);
				offsetMatrix.toArray(this.boneMatrices, b * 16);
			}

			if (this.useVertexTexture) {

				this.boneTexture.needsUpdate = true;
			}
		};
	}(),

	clone: function clone() {

		return new Skeleton(this.bones, this.boneInverses, this.useVertexTexture);
	}

});

/**
 * @author mikael emtinger / http://gomo.se/
 * @author alteredq / http://alteredqualia.com/
 * @author ikerr / http://verold.com
 */

function Bone(skin) {

	Object3D.call(this);

	this.type = 'Bone';

	this.skin = skin;
}

Bone.prototype = Object.assign(Object.create(Object3D.prototype), {

	constructor: Bone,

	isBone: true,

	copy: function copy(source) {

		Object3D.prototype.copy.call(this, source);

		this.skin = source.skin;

		return this;
	}

});

/**
 * @author mikael emtinger / http://gomo.se/
 * @author alteredq / http://alteredqualia.com/
 * @author ikerr / http://verold.com
 */

function SkinnedMesh(geometry, material, useVertexTexture) {

	Mesh.call(this, geometry, material);

	this.type = 'SkinnedMesh';

	this.bindMode = "attached";
	this.bindMatrix = new Matrix4();
	this.bindMatrixInverse = new Matrix4();

	// init bones

	// TODO: remove bone creation as there is no reason (other than
	// convenience) for THREE.SkinnedMesh to do this.

	var bones = [];

	if (this.geometry && this.geometry.bones !== undefined) {

		var bone, gbone;

		for (var b = 0, bl = this.geometry.bones.length; b < bl; ++b) {

			gbone = this.geometry.bones[b];

			bone = new Bone(this);
			bones.push(bone);

			bone.name = gbone.name;
			bone.position.fromArray(gbone.pos);
			bone.quaternion.fromArray(gbone.rotq);
			if (gbone.scl !== undefined) bone.scale.fromArray(gbone.scl);
		}

		for (var b = 0, bl = this.geometry.bones.length; b < bl; ++b) {

			gbone = this.geometry.bones[b];

			if (gbone.parent !== -1 && gbone.parent !== null && bones[gbone.parent] !== undefined) {

				bones[gbone.parent].add(bones[b]);
			} else {

				this.add(bones[b]);
			}
		}
	}

	this.normalizeSkinWeights();

	this.updateMatrixWorld(true);
	this.bind(new Skeleton(bones, undefined, useVertexTexture), this.matrixWorld);
}

SkinnedMesh.prototype = Object.assign(Object.create(Mesh.prototype), {

	constructor: SkinnedMesh,

	isSkinnedMesh: true,

	bind: function bind(skeleton, bindMatrix) {

		this.skeleton = skeleton;

		if (bindMatrix === undefined) {

			this.updateMatrixWorld(true);

			this.skeleton.calculateInverses();

			bindMatrix = this.matrixWorld;
		}

		this.bindMatrix.copy(bindMatrix);
		this.bindMatrixInverse.getInverse(bindMatrix);
	},

	pose: function pose() {

		this.skeleton.pose();
	},

	normalizeSkinWeights: function normalizeSkinWeights() {

		if (this.geometry && this.geometry.isGeometry) {

			for (var i = 0; i < this.geometry.skinWeights.length; i++) {

				var sw = this.geometry.skinWeights[i];

				var scale = 1.0 / sw.lengthManhattan();

				if (scale !== Infinity) {

					sw.multiplyScalar(scale);
				} else {

					sw.set(1, 0, 0, 0); // do something reasonable
				}
			}
		} else if (this.geometry && this.geometry.isBufferGeometry) {

			var vec = new Vector4();

			var skinWeight = this.geometry.attributes.skinWeight;

			for (var i = 0; i < skinWeight.count; i++) {

				vec.x = skinWeight.getX(i);
				vec.y = skinWeight.getY(i);
				vec.z = skinWeight.getZ(i);
				vec.w = skinWeight.getW(i);

				var scale = 1.0 / vec.lengthManhattan();

				if (scale !== Infinity) {

					vec.multiplyScalar(scale);
				} else {

					vec.set(1, 0, 0, 0); // do something reasonable
				}

				skinWeight.setXYZW(i, vec.x, vec.y, vec.z, vec.w);
			}
		}
	},

	updateMatrixWorld: function updateMatrixWorld(force) {

		Mesh.prototype.updateMatrixWorld.call(this, true);

		if (this.bindMode === "attached") {

			this.bindMatrixInverse.getInverse(this.matrixWorld);
		} else if (this.bindMode === "detached") {

			this.bindMatrixInverse.getInverse(this.bindMatrix);
		} else {

			console.warn('THREE.SkinnedMesh unrecognized bindMode: ' + this.bindMode);
		}
	},

	clone: function clone() {

		return new this.constructor(this.geometry, this.material, this.skeleton.useVertexTexture).copy(this);
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 */

function Fog(color, near, far) {

	this.name = '';

	this.color = new Color(color);

	this.near = near !== undefined ? near : 1;
	this.far = far !== undefined ? far : 1000;
}

Fog.prototype.isFog = true;

Fog.prototype.clone = function () {

	return new Fog(this.color.getHex(), this.near, this.far);
};

Fog.prototype.toJSON = function (meta) {

	return {
		type: 'Fog',
		color: this.color.getHex(),
		near: this.near,
		far: this.far
	};
};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 */

function FogExp2(color, density) {

	this.name = '';

	this.color = new Color(color);
	this.density = density !== undefined ? density : 0.00025;
}

FogExp2.prototype.isFogExp2 = true;

FogExp2.prototype.clone = function () {

	return new FogExp2(this.color.getHex(), this.density);
};

FogExp2.prototype.toJSON = function (meta) {

	return {
		type: 'FogExp2',
		color: this.color.getHex(),
		density: this.density
	};
};

/**
 * @author alteredq / http://alteredqualia.com/
 */

function HemisphereLight(skyColor, groundColor, intensity) {

	Light.call(this, skyColor, intensity);

	this.type = 'HemisphereLight';

	this.castShadow = undefined;

	this.position.copy(Object3D.DefaultUp);
	this.updateMatrix();

	this.groundColor = new Color(groundColor);
}

HemisphereLight.prototype = Object.assign(Object.create(Light.prototype), {

	constructor: HemisphereLight,

	isHemisphereLight: true,

	copy: function copy(source) {

		Light.prototype.copy.call(this, source);

		this.groundColor.copy(source.groundColor);

		return this;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function SpotLightShadow() {

	LightShadow.call(this, new PerspectiveCamera(50, 1, 0.5, 500));
}

SpotLightShadow.prototype = Object.assign(Object.create(LightShadow.prototype), {

	constructor: SpotLightShadow,

	isSpotLightShadow: true,

	update: function update(light) {

		var fov = _Math.RAD2DEG * 2 * light.angle;
		var aspect = this.mapSize.width / this.mapSize.height;
		var far = light.distance || 500;

		var camera = this.camera;

		if (fov !== camera.fov || aspect !== camera.aspect || far !== camera.far) {

			camera.fov = fov;
			camera.aspect = aspect;
			camera.far = far;
			camera.updateProjectionMatrix();
		}
	}

});

/**
 * @author alteredq / http://alteredqualia.com/
 */

function SpotLight(color, intensity, distance, angle, penumbra, decay) {

	Light.call(this, color, intensity);

	this.type = 'SpotLight';

	this.position.copy(Object3D.DefaultUp);
	this.updateMatrix();

	this.target = new Object3D();

	Object.defineProperty(this, 'power', {
		get: function get() {
			// intensity = power per solid angle.
			// ref: equation (17) from http://www.frostbite.com/wp-content/uploads/2014/11/course_notes_moving_frostbite_to_pbr.pdf
			return this.intensity * Math.PI;
		},
		set: function set(power) {
			// intensity = power per solid angle.
			// ref: equation (17) from http://www.frostbite.com/wp-content/uploads/2014/11/course_notes_moving_frostbite_to_pbr.pdf
			this.intensity = power / Math.PI;
		}
	});

	this.distance = distance !== undefined ? distance : 0;
	this.angle = angle !== undefined ? angle : Math.PI / 3;
	this.penumbra = penumbra !== undefined ? penumbra : 0;
	this.decay = decay !== undefined ? decay : 1; // for physically correct lights, should be 2.

	this.shadow = new SpotLightShadow();
}

SpotLight.prototype = Object.assign(Object.create(Light.prototype), {

	constructor: SpotLight,

	isSpotLight: true,

	copy: function copy(source) {

		Light.prototype.copy.call(this, source);

		this.distance = source.distance;
		this.angle = source.angle;
		this.penumbra = source.penumbra;
		this.decay = source.decay;

		this.target = source.target.clone();

		this.shadow = source.shadow.clone();

		return this;
	}

});

/**
 * @author alteredq / http://alteredqualia.com/
 * @author arose / http://github.com/arose
 */

function OrthographicCamera(left, right, top, bottom, near, far) {

	Camera.call(this);

	this.type = 'OrthographicCamera';

	this.zoom = 1;
	this.view = null;

	this.left = left;
	this.right = right;
	this.top = top;
	this.bottom = bottom;

	this.near = near !== undefined ? near : 0.1;
	this.far = far !== undefined ? far : 2000;

	this.updateProjectionMatrix();
}

OrthographicCamera.prototype = Object.assign(Object.create(Camera.prototype), {

	constructor: OrthographicCamera,

	isOrthographicCamera: true,

	copy: function copy(source) {

		Camera.prototype.copy.call(this, source);

		this.left = source.left;
		this.right = source.right;
		this.top = source.top;
		this.bottom = source.bottom;
		this.near = source.near;
		this.far = source.far;

		this.zoom = source.zoom;
		this.view = source.view === null ? null : Object.assign({}, source.view);

		return this;
	},

	setViewOffset: function setViewOffset(fullWidth, fullHeight, x, y, width, height) {

		this.view = {
			fullWidth: fullWidth,
			fullHeight: fullHeight,
			offsetX: x,
			offsetY: y,
			width: width,
			height: height
		};

		this.updateProjectionMatrix();
	},

	clearViewOffset: function clearViewOffset() {

		this.view = null;
		this.updateProjectionMatrix();
	},

	updateProjectionMatrix: function updateProjectionMatrix() {

		var dx = (this.right - this.left) / (2 * this.zoom);
		var dy = (this.top - this.bottom) / (2 * this.zoom);
		var cx = (this.right + this.left) / 2;
		var cy = (this.top + this.bottom) / 2;

		var left = cx - dx;
		var right = cx + dx;
		var top = cy + dy;
		var bottom = cy - dy;

		if (this.view !== null) {

			var zoomW = this.zoom / (this.view.width / this.view.fullWidth);
			var zoomH = this.zoom / (this.view.height / this.view.fullHeight);
			var scaleW = (this.right - this.left) / this.view.width;
			var scaleH = (this.top - this.bottom) / this.view.height;

			left += scaleW * (this.view.offsetX / zoomW);
			right = left + scaleW * (this.view.width / zoomW);
			top -= scaleH * (this.view.offsetY / zoomH);
			bottom = top - scaleH * (this.view.height / zoomH);
		}

		this.projectionMatrix.makeOrthographic(left, right, top, bottom, this.near, this.far);
	},

	toJSON: function toJSON(meta) {

		var data = Object3D.prototype.toJSON.call(this, meta);

		data.object.zoom = this.zoom;
		data.object.left = this.left;
		data.object.right = this.right;
		data.object.top = this.top;
		data.object.bottom = this.bottom;
		data.object.near = this.near;
		data.object.far = this.far;

		if (this.view !== null) data.object.view = Object.assign({}, this.view);

		return data;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function DirectionalLightShadow(light) {

  LightShadow.call(this, new OrthographicCamera(-5, 5, 5, -5, 0.5, 500));
}

DirectionalLightShadow.prototype = Object.assign(Object.create(LightShadow.prototype), {

  constructor: DirectionalLightShadow

});

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 */

function DirectionalLight(color, intensity) {

	Light.call(this, color, intensity);

	this.type = 'DirectionalLight';

	this.position.copy(Object3D.DefaultUp);
	this.updateMatrix();

	this.target = new Object3D();

	this.shadow = new DirectionalLightShadow();
}

DirectionalLight.prototype = Object.assign(Object.create(Light.prototype), {

	constructor: DirectionalLight,

	isDirectionalLight: true,

	copy: function copy(source) {

		Light.prototype.copy.call(this, source);

		this.target = source.target.clone();

		this.shadow = source.shadow.clone();

		return this;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function AmbientLight(color, intensity) {

	Light.call(this, color, intensity);

	this.type = 'AmbientLight';

	this.castShadow = undefined;
}

AmbientLight.prototype = Object.assign(Object.create(Light.prototype), {

	constructor: AmbientLight,

	isAmbientLight: true

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function Scene() {

	Object3D.call(this);

	this.type = 'Scene';

	this.background = null;
	this.fog = null;
	this.overrideMaterial = null;

	this.autoUpdate = true; // checked by the renderer
}

Scene.prototype = Object.create(Object3D.prototype);

Scene.prototype.constructor = Scene;

Scene.prototype.copy = function (source, recursive) {

	Object3D.prototype.copy.call(this, source, recursive);

	if (source.background !== null) this.background = source.background.clone();
	if (source.fog !== null) this.fog = source.fog.clone();
	if (source.overrideMaterial !== null) this.overrideMaterial = source.overrideMaterial.clone();

	this.autoUpdate = source.autoUpdate;
	this.matrixAutoUpdate = source.matrixAutoUpdate;

	return this;
};

Scene.prototype.toJSON = function (meta) {

	var data = Object3D.prototype.toJSON.call(this, meta);

	if (this.background !== null) data.object.background = this.background.toJSON(meta);
	if (this.fog !== null) data.object.fog = this.fog.toJSON();

	return data;
};

/**
 * @author tschw
 * @author Ben Houston / http://clara.io/
 * @author David Sarno / http://lighthaus.us/
 */

var AnimationUtils = {

	// same as Array.prototype.slice, but also works on typed arrays
	arraySlice: function arraySlice(array, from, to) {

		if (AnimationUtils.isTypedArray(array)) {

			return new array.constructor(array.subarray(from, to));
		}

		return array.slice(from, to);
	},

	// converts an array to a specific type
	convertArray: function convertArray(array, type, forceClone) {

		if (!array || // let 'undefined' and 'null' pass
		!forceClone && array.constructor === type) return array;

		if (typeof type.BYTES_PER_ELEMENT === 'number') {

			return new type(array); // create typed array
		}

		return Array.prototype.slice.call(array); // create Array
	},

	isTypedArray: function isTypedArray(object) {

		return ArrayBuffer.isView(object) && !(object instanceof DataView);
	},

	// returns an array by which times and values can be sorted
	getKeyframeOrder: function getKeyframeOrder(times) {

		function compareTime(i, j) {

			return times[i] - times[j];
		}

		var n = times.length;
		var result = new Array(n);
		for (var i = 0; i !== n; ++i) {
			result[i] = i;
		}result.sort(compareTime);

		return result;
	},

	// uses the array previously returned by 'getKeyframeOrder' to sort data
	sortedArray: function sortedArray(values, stride, order) {

		var nValues = values.length;
		var result = new values.constructor(nValues);

		for (var i = 0, dstOffset = 0; dstOffset !== nValues; ++i) {

			var srcOffset = order[i] * stride;

			for (var j = 0; j !== stride; ++j) {

				result[dstOffset++] = values[srcOffset + j];
			}
		}

		return result;
	},

	// function for parsing AOS keyframe formats
	flattenJSON: function flattenJSON(jsonKeys, times, values, valuePropertyName) {

		var i = 1,
		    key = jsonKeys[0];

		while (key !== undefined && key[valuePropertyName] === undefined) {

			key = jsonKeys[i++];
		}

		if (key === undefined) return; // no data

		var value = key[valuePropertyName];
		if (value === undefined) return; // no data

		if (Array.isArray(value)) {

			do {

				value = key[valuePropertyName];

				if (value !== undefined) {

					times.push(key.time);
					values.push.apply(values, value); // push all elements
				}

				key = jsonKeys[i++];
			} while (key !== undefined);
		} else if (value.toArray !== undefined) {
			// ...assume THREE.Math-ish

			do {

				value = key[valuePropertyName];

				if (value !== undefined) {

					times.push(key.time);
					value.toArray(values, values.length);
				}

				key = jsonKeys[i++];
			} while (key !== undefined);
		} else {
			// otherwise push as-is

			do {

				value = key[valuePropertyName];

				if (value !== undefined) {

					times.push(key.time);
					values.push(value);
				}

				key = jsonKeys[i++];
			} while (key !== undefined);
		}
	}

};

/**
 * Abstract base class of interpolants over parametric samples.
 *
 * The parameter domain is one dimensional, typically the time or a path
 * along a curve defined by the data.
 *
 * The sample values can have any dimensionality and derived classes may
 * apply special interpretations to the data.
 *
 * This class provides the interval seek in a Template Method, deferring
 * the actual interpolation to derived classes.
 *
 * Time complexity is O(1) for linear access crossing at most two points
 * and O(log N) for random access, where N is the number of positions.
 *
 * References:
 *
 * 		http://www.oodesign.com/template-method-pattern.html
 *
 * @author tschw
 */

function Interpolant(parameterPositions, sampleValues, sampleSize, resultBuffer) {

	this.parameterPositions = parameterPositions;
	this._cachedIndex = 0;

	this.resultBuffer = resultBuffer !== undefined ? resultBuffer : new sampleValues.constructor(sampleSize);
	this.sampleValues = sampleValues;
	this.valueSize = sampleSize;
}

Interpolant.prototype = {

	constructor: Interpolant,

	evaluate: function evaluate(t) {

		var pp = this.parameterPositions,
		    i1 = this._cachedIndex,
		    t1 = pp[i1],
		    t0 = pp[i1 - 1];

		validate_interval: {

			seek: {

				var right;

				linear_scan: {
					//- See http://jsperf.com/comparison-to-undefined/3
					//- slower code:
					//-
					//- 				if ( t >= t1 || t1 === undefined ) {
					forward_scan: if (!(t < t1)) {

						for (var giveUpAt = i1 + 2;;) {

							if (t1 === undefined) {

								if (t < t0) break forward_scan;

								// after end

								i1 = pp.length;
								this._cachedIndex = i1;
								return this.afterEnd_(i1 - 1, t, t0);
							}

							if (i1 === giveUpAt) break; // this loop

							t0 = t1;
							t1 = pp[++i1];

							if (t < t1) {

								// we have arrived at the sought interval
								break seek;
							}
						}

						// prepare binary search on the right side of the index
						right = pp.length;
						break linear_scan;
					}

					//- slower code:
					//-					if ( t < t0 || t0 === undefined ) {
					if (!(t >= t0)) {

						// looping?

						var t1global = pp[1];

						if (t < t1global) {

							i1 = 2; // + 1, using the scan for the details
							t0 = t1global;
						}

						// linear reverse scan

						for (var giveUpAt = i1 - 2;;) {

							if (t0 === undefined) {

								// before start

								this._cachedIndex = 0;
								return this.beforeStart_(0, t, t1);
							}

							if (i1 === giveUpAt) break; // this loop

							t1 = t0;
							t0 = pp[--i1 - 1];

							if (t >= t0) {

								// we have arrived at the sought interval
								break seek;
							}
						}

						// prepare binary search on the left side of the index
						right = i1;
						i1 = 0;
						break linear_scan;
					}

					// the interval is valid

					break validate_interval;
				} // linear scan

				// binary search

				while (i1 < right) {

					var mid = i1 + right >>> 1;

					if (t < pp[mid]) {

						right = mid;
					} else {

						i1 = mid + 1;
					}
				}

				t1 = pp[i1];
				t0 = pp[i1 - 1];

				// check boundary cases, again

				if (t0 === undefined) {

					this._cachedIndex = 0;
					return this.beforeStart_(0, t, t1);
				}

				if (t1 === undefined) {

					i1 = pp.length;
					this._cachedIndex = i1;
					return this.afterEnd_(i1 - 1, t0, t);
				}
			} // seek

			this._cachedIndex = i1;

			this.intervalChanged_(i1, t0, t1);
		} // validate_interval

		return this.interpolate_(i1, t0, t, t1);
	},

	settings: null, // optional, subclass-specific settings structure
	// Note: The indirection allows central control of many interpolants.

	// --- Protected interface

	DefaultSettings_: {},

	getSettings_: function getSettings_() {

		return this.settings || this.DefaultSettings_;
	},

	copySampleValue_: function copySampleValue_(index) {

		// copies a sample value to the result buffer

		var result = this.resultBuffer,
		    values = this.sampleValues,
		    stride = this.valueSize,
		    offset = index * stride;

		for (var i = 0; i !== stride; ++i) {

			result[i] = values[offset + i];
		}

		return result;
	},

	// Template methods for derived classes:

	interpolate_: function interpolate_(i1, t0, t, t1) {

		throw new Error("call to abstract method");
		// implementations shall return this.resultBuffer
	},

	intervalChanged_: function intervalChanged_(i1, t0, t1) {

		// empty

	}

};

Object.assign(Interpolant.prototype, {

	beforeStart_: //( 0, t, t0 ), returns this.resultBuffer
	Interpolant.prototype.copySampleValue_,

	afterEnd_: //( N-1, tN-1, t ), returns this.resultBuffer
	Interpolant.prototype.copySampleValue_

});

/**
 * Fast and simple cubic spline interpolant.
 *
 * It was derived from a Hermitian construction setting the first derivative
 * at each sample position to the linear slope between neighboring positions
 * over their parameter interval.
 *
 * @author tschw
 */

function CubicInterpolant(parameterPositions, sampleValues, sampleSize, resultBuffer) {

	Interpolant.call(this, parameterPositions, sampleValues, sampleSize, resultBuffer);

	this._weightPrev = -0;
	this._offsetPrev = -0;
	this._weightNext = -0;
	this._offsetNext = -0;
}

CubicInterpolant.prototype = Object.assign(Object.create(Interpolant.prototype), {

	constructor: CubicInterpolant,

	DefaultSettings_: {

		endingStart: ZeroCurvatureEnding,
		endingEnd: ZeroCurvatureEnding

	},

	intervalChanged_: function intervalChanged_(i1, t0, t1) {

		var pp = this.parameterPositions,
		    iPrev = i1 - 2,
		    iNext = i1 + 1,
		    tPrev = pp[iPrev],
		    tNext = pp[iNext];

		if (tPrev === undefined) {

			switch (this.getSettings_().endingStart) {

				case ZeroSlopeEnding:

					// f'(t0) = 0
					iPrev = i1;
					tPrev = 2 * t0 - t1;

					break;

				case WrapAroundEnding:

					// use the other end of the curve
					iPrev = pp.length - 2;
					tPrev = t0 + pp[iPrev] - pp[iPrev + 1];

					break;

				default:
					// ZeroCurvatureEnding

					// f''(t0) = 0 a.k.a. Natural Spline
					iPrev = i1;
					tPrev = t1;

			}
		}

		if (tNext === undefined) {

			switch (this.getSettings_().endingEnd) {

				case ZeroSlopeEnding:

					// f'(tN) = 0
					iNext = i1;
					tNext = 2 * t1 - t0;

					break;

				case WrapAroundEnding:

					// use the other end of the curve
					iNext = 1;
					tNext = t1 + pp[1] - pp[0];

					break;

				default:
					// ZeroCurvatureEnding

					// f''(tN) = 0, a.k.a. Natural Spline
					iNext = i1 - 1;
					tNext = t0;

			}
		}

		var halfDt = (t1 - t0) * 0.5,
		    stride = this.valueSize;

		this._weightPrev = halfDt / (t0 - tPrev);
		this._weightNext = halfDt / (tNext - t1);
		this._offsetPrev = iPrev * stride;
		this._offsetNext = iNext * stride;
	},

	interpolate_: function interpolate_(i1, t0, t, t1) {

		var result = this.resultBuffer,
		    values = this.sampleValues,
		    stride = this.valueSize,
		    o1 = i1 * stride,
		    o0 = o1 - stride,
		    oP = this._offsetPrev,
		    oN = this._offsetNext,
		    wP = this._weightPrev,
		    wN = this._weightNext,
		    p = (t - t0) / (t1 - t0),
		    pp = p * p,
		    ppp = pp * p;

		// evaluate polynomials

		var sP = -wP * ppp + 2 * wP * pp - wP * p;
		var s0 = (1 + wP) * ppp + (-1.5 - 2 * wP) * pp + (-0.5 + wP) * p + 1;
		var s1 = (-1 - wN) * ppp + (1.5 + wN) * pp + 0.5 * p;
		var sN = wN * ppp - wN * pp;

		// combine data linearly

		for (var i = 0; i !== stride; ++i) {

			result[i] = sP * values[oP + i] + s0 * values[o0 + i] + s1 * values[o1 + i] + sN * values[oN + i];
		}

		return result;
	}

});

/**
 * @author tschw
 */

function LinearInterpolant(parameterPositions, sampleValues, sampleSize, resultBuffer) {

	Interpolant.call(this, parameterPositions, sampleValues, sampleSize, resultBuffer);
}

LinearInterpolant.prototype = Object.assign(Object.create(Interpolant.prototype), {

	constructor: LinearInterpolant,

	interpolate_: function interpolate_(i1, t0, t, t1) {

		var result = this.resultBuffer,
		    values = this.sampleValues,
		    stride = this.valueSize,
		    offset1 = i1 * stride,
		    offset0 = offset1 - stride,
		    weight1 = (t - t0) / (t1 - t0),
		    weight0 = 1 - weight1;

		for (var i = 0; i !== stride; ++i) {

			result[i] = values[offset0 + i] * weight0 + values[offset1 + i] * weight1;
		}

		return result;
	}

});

/**
 *
 * Interpolant that evaluates to the sample value at the position preceeding
 * the parameter.
 *
 * @author tschw
 */

function DiscreteInterpolant(parameterPositions, sampleValues, sampleSize, resultBuffer) {

	Interpolant.call(this, parameterPositions, sampleValues, sampleSize, resultBuffer);
}

DiscreteInterpolant.prototype = Object.assign(Object.create(Interpolant.prototype), {

	constructor: DiscreteInterpolant,

	interpolate_: function interpolate_(i1, t0, t, t1) {

		return this.copySampleValue_(i1 - 1);
	}

});

var KeyframeTrackPrototype;

KeyframeTrackPrototype = {

	TimeBufferType: Float32Array,
	ValueBufferType: Float32Array,

	DefaultInterpolation: InterpolateLinear,

	InterpolantFactoryMethodDiscrete: function InterpolantFactoryMethodDiscrete(result) {

		return new DiscreteInterpolant(this.times, this.values, this.getValueSize(), result);
	},

	InterpolantFactoryMethodLinear: function InterpolantFactoryMethodLinear(result) {

		return new LinearInterpolant(this.times, this.values, this.getValueSize(), result);
	},

	InterpolantFactoryMethodSmooth: function InterpolantFactoryMethodSmooth(result) {

		return new CubicInterpolant(this.times, this.values, this.getValueSize(), result);
	},

	setInterpolation: function setInterpolation(interpolation) {

		var factoryMethod;

		switch (interpolation) {

			case InterpolateDiscrete:

				factoryMethod = this.InterpolantFactoryMethodDiscrete;

				break;

			case InterpolateLinear:

				factoryMethod = this.InterpolantFactoryMethodLinear;

				break;

			case InterpolateSmooth:

				factoryMethod = this.InterpolantFactoryMethodSmooth;

				break;

		}

		if (factoryMethod === undefined) {

			var message = "unsupported interpolation for " + this.ValueTypeName + " keyframe track named " + this.name;

			if (this.createInterpolant === undefined) {

				// fall back to default, unless the default itself is messed up
				if (interpolation !== this.DefaultInterpolation) {

					this.setInterpolation(this.DefaultInterpolation);
				} else {

					throw new Error(message); // fatal, in this case
				}
			}

			console.warn(message);
			return;
		}

		this.createInterpolant = factoryMethod;
	},

	getInterpolation: function getInterpolation() {

		switch (this.createInterpolant) {

			case this.InterpolantFactoryMethodDiscrete:

				return InterpolateDiscrete;

			case this.InterpolantFactoryMethodLinear:

				return InterpolateLinear;

			case this.InterpolantFactoryMethodSmooth:

				return InterpolateSmooth;

		}
	},

	getValueSize: function getValueSize() {

		return this.values.length / this.times.length;
	},

	// move all keyframes either forwards or backwards in time
	shift: function shift(timeOffset) {

		if (timeOffset !== 0.0) {

			var times = this.times;

			for (var i = 0, n = times.length; i !== n; ++i) {

				times[i] += timeOffset;
			}
		}

		return this;
	},

	// scale all keyframe times by a factor (useful for frame <-> seconds conversions)
	scale: function scale(timeScale) {

		if (timeScale !== 1.0) {

			var times = this.times;

			for (var i = 0, n = times.length; i !== n; ++i) {

				times[i] *= timeScale;
			}
		}

		return this;
	},

	// removes keyframes before and after animation without changing any values within the range [startTime, endTime].
	// IMPORTANT: We do not shift around keys to the start of the track time, because for interpolated keys this will change their values
	trim: function trim(startTime, endTime) {

		var times = this.times,
		    nKeys = times.length,
		    from = 0,
		    to = nKeys - 1;

		while (from !== nKeys && times[from] < startTime) {
			++from;
		}while (to !== -1 && times[to] > endTime) {
			--to;
		}++to; // inclusive -> exclusive bound

		if (from !== 0 || to !== nKeys) {

			// empty tracks are forbidden, so keep at least one keyframe
			if (from >= to) to = Math.max(to, 1), from = to - 1;

			var stride = this.getValueSize();
			this.times = AnimationUtils.arraySlice(times, from, to);
			this.values = AnimationUtils.arraySlice(this.values, from * stride, to * stride);
		}

		return this;
	},

	// ensure we do not get a GarbageInGarbageOut situation, make sure tracks are at least minimally viable
	validate: function validate() {

		var valid = true;

		var valueSize = this.getValueSize();
		if (valueSize - Math.floor(valueSize) !== 0) {

			console.error("invalid value size in track", this);
			valid = false;
		}

		var times = this.times,
		    values = this.values,
		    nKeys = times.length;

		if (nKeys === 0) {

			console.error("track is empty", this);
			valid = false;
		}

		var prevTime = null;

		for (var i = 0; i !== nKeys; i++) {

			var currTime = times[i];

			if (typeof currTime === 'number' && isNaN(currTime)) {

				console.error("time is not a valid number", this, i, currTime);
				valid = false;
				break;
			}

			if (prevTime !== null && prevTime > currTime) {

				console.error("out of order keys", this, i, currTime, prevTime);
				valid = false;
				break;
			}

			prevTime = currTime;
		}

		if (values !== undefined) {

			if (AnimationUtils.isTypedArray(values)) {

				for (var i = 0, n = values.length; i !== n; ++i) {

					var value = values[i];

					if (isNaN(value)) {

						console.error("value is not a valid number", this, i, value);
						valid = false;
						break;
					}
				}
			}
		}

		return valid;
	},

	// removes equivalent sequential keys as common in morph target sequences
	// (0,0,0,0,1,1,1,0,0,0,0,0,0,0) --> (0,0,1,1,0,0)
	optimize: function optimize() {

		var times = this.times,
		    values = this.values,
		    stride = this.getValueSize(),
		    smoothInterpolation = this.getInterpolation() === InterpolateSmooth,
		    writeIndex = 1,
		    lastIndex = times.length - 1;

		for (var i = 1; i < lastIndex; ++i) {

			var keep = false;

			var time = times[i];
			var timeNext = times[i + 1];

			// remove adjacent keyframes scheduled at the same time

			if (time !== timeNext && (i !== 1 || time !== time[0])) {

				if (!smoothInterpolation) {

					// remove unnecessary keyframes same as their neighbors

					var offset = i * stride,
					    offsetP = offset - stride,
					    offsetN = offset + stride;

					for (var j = 0; j !== stride; ++j) {

						var value = values[offset + j];

						if (value !== values[offsetP + j] || value !== values[offsetN + j]) {

							keep = true;
							break;
						}
					}
				} else keep = true;
			}

			// in-place compaction

			if (keep) {

				if (i !== writeIndex) {

					times[writeIndex] = times[i];

					var readOffset = i * stride,
					    writeOffset = writeIndex * stride;

					for (var j = 0; j !== stride; ++j) {

						values[writeOffset + j] = values[readOffset + j];
					}
				}

				++writeIndex;
			}
		}

		// flush last keyframe (compaction looks ahead)

		if (lastIndex > 0) {

			times[writeIndex] = times[lastIndex];

			for (var readOffset = lastIndex * stride, writeOffset = writeIndex * stride, j = 0; j !== stride; ++j) {

				values[writeOffset + j] = values[readOffset + j];
			}++writeIndex;
		}

		if (writeIndex !== times.length) {

			this.times = AnimationUtils.arraySlice(times, 0, writeIndex);
			this.values = AnimationUtils.arraySlice(values, 0, writeIndex * stride);
		}

		return this;
	}

};

function KeyframeTrackConstructor(name, times, values, interpolation) {

	if (name === undefined) throw new Error("track name is undefined");

	if (times === undefined || times.length === 0) {

		throw new Error("no keyframes in track named " + name);
	}

	this.name = name;

	this.times = AnimationUtils.convertArray(times, this.TimeBufferType);
	this.values = AnimationUtils.convertArray(values, this.ValueBufferType);

	this.setInterpolation(interpolation || this.DefaultInterpolation);

	this.validate();
	this.optimize();
}

/**
 *
 * A Track of vectored keyframe values.
 *
 *
 * @author Ben Houston / http://clara.io/
 * @author David Sarno / http://lighthaus.us/
 * @author tschw
 */

function VectorKeyframeTrack(name, times, values, interpolation) {

  KeyframeTrackConstructor.call(this, name, times, values, interpolation);
}

VectorKeyframeTrack.prototype = Object.assign(Object.create(KeyframeTrackPrototype), {

  constructor: VectorKeyframeTrack,

  ValueTypeName: 'vector'

  // ValueBufferType is inherited

  // DefaultInterpolation is inherited

});

/**
 * Spherical linear unit quaternion interpolant.
 *
 * @author tschw
 */

function QuaternionLinearInterpolant(parameterPositions, sampleValues, sampleSize, resultBuffer) {

	Interpolant.call(this, parameterPositions, sampleValues, sampleSize, resultBuffer);
}

QuaternionLinearInterpolant.prototype = Object.assign(Object.create(Interpolant.prototype), {

	constructor: QuaternionLinearInterpolant,

	interpolate_: function interpolate_(i1, t0, t, t1) {

		var result = this.resultBuffer,
		    values = this.sampleValues,
		    stride = this.valueSize,
		    offset = i1 * stride,
		    alpha = (t - t0) / (t1 - t0);

		for (var end = offset + stride; offset !== end; offset += 4) {

			Quaternion.slerpFlat(result, 0, values, offset - stride, values, offset, alpha);
		}

		return result;
	}

});

/**
 *
 * A Track of quaternion keyframe values.
 *
 * @author Ben Houston / http://clara.io/
 * @author David Sarno / http://lighthaus.us/
 * @author tschw
 */

function QuaternionKeyframeTrack(name, times, values, interpolation) {

	KeyframeTrackConstructor.call(this, name, times, values, interpolation);
}

QuaternionKeyframeTrack.prototype = Object.assign(Object.create(KeyframeTrackPrototype), {

	constructor: QuaternionKeyframeTrack,

	ValueTypeName: 'quaternion',

	// ValueBufferType is inherited

	DefaultInterpolation: InterpolateLinear,

	InterpolantFactoryMethodLinear: function InterpolantFactoryMethodLinear(result) {

		return new QuaternionLinearInterpolant(this.times, this.values, this.getValueSize(), result);
	},

	InterpolantFactoryMethodSmooth: undefined // not yet implemented

});

/**
 *
 * A Track of numeric keyframe values.
 *
 * @author Ben Houston / http://clara.io/
 * @author David Sarno / http://lighthaus.us/
 * @author tschw
 */

function NumberKeyframeTrack(name, times, values, interpolation) {

  KeyframeTrackConstructor.call(this, name, times, values, interpolation);
}

NumberKeyframeTrack.prototype = Object.assign(Object.create(KeyframeTrackPrototype), {

  constructor: NumberKeyframeTrack,

  ValueTypeName: 'number'

});

/**
 *
 * A Track that interpolates Strings
 *
 *
 * @author Ben Houston / http://clara.io/
 * @author David Sarno / http://lighthaus.us/
 * @author tschw
 */

function StringKeyframeTrack(name, times, values, interpolation) {

  KeyframeTrackConstructor.call(this, name, times, values, interpolation);
}

StringKeyframeTrack.prototype = Object.assign(Object.create(KeyframeTrackPrototype), {

  constructor: StringKeyframeTrack,

  ValueTypeName: 'string',
  ValueBufferType: Array,

  DefaultInterpolation: InterpolateDiscrete,

  InterpolantFactoryMethodLinear: undefined,

  InterpolantFactoryMethodSmooth: undefined

});

/**
 *
 * A Track of Boolean keyframe values.
 *
 *
 * @author Ben Houston / http://clara.io/
 * @author David Sarno / http://lighthaus.us/
 * @author tschw
 */

function BooleanKeyframeTrack(name, times, values) {

	KeyframeTrackConstructor.call(this, name, times, values);
}

BooleanKeyframeTrack.prototype = Object.assign(Object.create(KeyframeTrackPrototype), {

	constructor: BooleanKeyframeTrack,

	ValueTypeName: 'bool',
	ValueBufferType: Array,

	DefaultInterpolation: InterpolateDiscrete,

	InterpolantFactoryMethodLinear: undefined,
	InterpolantFactoryMethodSmooth: undefined

	// Note: Actually this track could have a optimized / compressed
	// representation of a single value and a custom interpolant that
	// computes "firstValue ^ isOdd( index )".

});

/**
 *
 * A Track of keyframe values that represent color.
 *
 *
 * @author Ben Houston / http://clara.io/
 * @author David Sarno / http://lighthaus.us/
 * @author tschw
 */

function ColorKeyframeTrack(name, times, values, interpolation) {

  KeyframeTrackConstructor.call(this, name, times, values, interpolation);
}

ColorKeyframeTrack.prototype = Object.assign(Object.create(KeyframeTrackPrototype), {

  constructor: ColorKeyframeTrack,

  ValueTypeName: 'color'

  // ValueBufferType is inherited

  // DefaultInterpolation is inherited


  // Note: Very basic implementation and nothing special yet.
  // However, this is the place for color space parameterization.

});

/**
 *
 * A timed sequence of keyframes for a specific property.
 *
 *
 * @author Ben Houston / http://clara.io/
 * @author David Sarno / http://lighthaus.us/
 * @author tschw
 */

function KeyframeTrack(name, times, values, interpolation) {

	KeyframeTrackConstructor.apply(this, arguments);
}

KeyframeTrack.prototype = KeyframeTrackPrototype;
KeyframeTrackPrototype.constructor = KeyframeTrack;

// Static methods:

Object.assign(KeyframeTrack, {

	// Serialization (in static context, because of constructor invocation
	// and automatic invocation of .toJSON):

	parse: function parse(json) {

		if (json.type === undefined) {

			throw new Error("track type undefined, can not parse");
		}

		var trackType = KeyframeTrack._getTrackTypeForValueTypeName(json.type);

		if (json.times === undefined) {

			var times = [],
			    values = [];

			AnimationUtils.flattenJSON(json.keys, times, values, 'value');

			json.times = times;
			json.values = values;
		}

		// derived classes can define a static parse method
		if (trackType.parse !== undefined) {

			return trackType.parse(json);
		} else {

			// by default, we asssume a constructor compatible with the base
			return new trackType(json.name, json.times, json.values, json.interpolation);
		}
	},

	toJSON: function toJSON(track) {

		var trackType = track.constructor;

		var json;

		// derived classes can define a static toJSON method
		if (trackType.toJSON !== undefined) {

			json = trackType.toJSON(track);
		} else {

			// by default, we assume the data can be serialized as-is
			json = {

				'name': track.name,
				'times': AnimationUtils.convertArray(track.times, Array),
				'values': AnimationUtils.convertArray(track.values, Array)

			};

			var interpolation = track.getInterpolation();

			if (interpolation !== track.DefaultInterpolation) {

				json.interpolation = interpolation;
			}
		}

		json.type = track.ValueTypeName; // mandatory

		return json;
	},

	_getTrackTypeForValueTypeName: function _getTrackTypeForValueTypeName(typeName) {

		switch (typeName.toLowerCase()) {

			case "scalar":
			case "double":
			case "float":
			case "number":
			case "integer":

				return NumberKeyframeTrack;

			case "vector":
			case "vector2":
			case "vector3":
			case "vector4":

				return VectorKeyframeTrack;

			case "color":

				return ColorKeyframeTrack;

			case "quaternion":

				return QuaternionKeyframeTrack;

			case "bool":
			case "boolean":

				return BooleanKeyframeTrack;

			case "string":

				return StringKeyframeTrack;

		}

		throw new Error("Unsupported typeName: " + typeName);
	}

});

/**
 *
 * Reusable set of Tracks that represent an animation.
 *
 * @author Ben Houston / http://clara.io/
 * @author David Sarno / http://lighthaus.us/
 */

function AnimationClip(name, duration, tracks) {

	this.name = name;
	this.tracks = tracks;
	this.duration = duration !== undefined ? duration : -1;

	this.uuid = _Math.generateUUID();

	// this means it should figure out its duration by scanning the tracks
	if (this.duration < 0) {

		this.resetDuration();
	}

	this.optimize();
}

AnimationClip.prototype = {

	constructor: AnimationClip,

	resetDuration: function resetDuration() {

		var tracks = this.tracks,
		    duration = 0;

		for (var i = 0, n = tracks.length; i !== n; ++i) {

			var track = this.tracks[i];

			duration = Math.max(duration, track.times[track.times.length - 1]);
		}

		this.duration = duration;
	},

	trim: function trim() {

		for (var i = 0; i < this.tracks.length; i++) {

			this.tracks[i].trim(0, this.duration);
		}

		return this;
	},

	optimize: function optimize() {

		for (var i = 0; i < this.tracks.length; i++) {

			this.tracks[i].optimize();
		}

		return this;
	}

};

// Static methods:

Object.assign(AnimationClip, {

	parse: function parse(json) {

		var tracks = [],
		    jsonTracks = json.tracks,
		    frameTime = 1.0 / (json.fps || 1.0);

		for (var i = 0, n = jsonTracks.length; i !== n; ++i) {

			tracks.push(KeyframeTrack.parse(jsonTracks[i]).scale(frameTime));
		}

		return new AnimationClip(json.name, json.duration, tracks);
	},

	toJSON: function toJSON(clip) {

		var tracks = [],
		    clipTracks = clip.tracks;

		var json = {

			'name': clip.name,
			'duration': clip.duration,
			'tracks': tracks

		};

		for (var i = 0, n = clipTracks.length; i !== n; ++i) {

			tracks.push(KeyframeTrack.toJSON(clipTracks[i]));
		}

		return json;
	},

	CreateFromMorphTargetSequence: function CreateFromMorphTargetSequence(name, morphTargetSequence, fps, noLoop) {

		var numMorphTargets = morphTargetSequence.length;
		var tracks = [];

		for (var i = 0; i < numMorphTargets; i++) {

			var times = [];
			var values = [];

			times.push((i + numMorphTargets - 1) % numMorphTargets, i, (i + 1) % numMorphTargets);

			values.push(0, 1, 0);

			var order = AnimationUtils.getKeyframeOrder(times);
			times = AnimationUtils.sortedArray(times, 1, order);
			values = AnimationUtils.sortedArray(values, 1, order);

			// if there is a key at the first frame, duplicate it as the
			// last frame as well for perfect loop.
			if (!noLoop && times[0] === 0) {

				times.push(numMorphTargets);
				values.push(values[0]);
			}

			tracks.push(new NumberKeyframeTrack('.morphTargetInfluences[' + morphTargetSequence[i].name + ']', times, values).scale(1.0 / fps));
		}

		return new AnimationClip(name, -1, tracks);
	},

	findByName: function findByName(objectOrClipArray, name) {

		var clipArray = objectOrClipArray;

		if (!Array.isArray(objectOrClipArray)) {

			var o = objectOrClipArray;
			clipArray = o.geometry && o.geometry.animations || o.animations;
		}

		for (var i = 0; i < clipArray.length; i++) {

			if (clipArray[i].name === name) {

				return clipArray[i];
			}
		}

		return null;
	},

	CreateClipsFromMorphTargetSequences: function CreateClipsFromMorphTargetSequences(morphTargets, fps, noLoop) {

		var animationToMorphTargets = {};

		// tested with https://regex101.com/ on trick sequences
		// such flamingo_flyA_003, flamingo_run1_003, crdeath0059
		var pattern = /^([\w-]*?)([\d]+)$/;

		// sort morph target names into animation groups based
		// patterns like Walk_001, Walk_002, Run_001, Run_002
		for (var i = 0, il = morphTargets.length; i < il; i++) {

			var morphTarget = morphTargets[i];
			var parts = morphTarget.name.match(pattern);

			if (parts && parts.length > 1) {

				var name = parts[1];

				var animationMorphTargets = animationToMorphTargets[name];
				if (!animationMorphTargets) {

					animationToMorphTargets[name] = animationMorphTargets = [];
				}

				animationMorphTargets.push(morphTarget);
			}
		}

		var clips = [];

		for (var name in animationToMorphTargets) {

			clips.push(AnimationClip.CreateFromMorphTargetSequence(name, animationToMorphTargets[name], fps, noLoop));
		}

		return clips;
	},

	// parse the animation.hierarchy format
	parseAnimation: function parseAnimation(animation, bones) {

		if (!animation) {

			console.error("  no animation in JSONLoader data");
			return null;
		}

		var addNonemptyTrack = function addNonemptyTrack(trackType, trackName, animationKeys, propertyName, destTracks) {

			// only return track if there are actually keys.
			if (animationKeys.length !== 0) {

				var times = [];
				var values = [];

				AnimationUtils.flattenJSON(animationKeys, times, values, propertyName);

				// empty keys are filtered out, so check again
				if (times.length !== 0) {

					destTracks.push(new trackType(trackName, times, values));
				}
			}
		};

		var tracks = [];

		var clipName = animation.name || 'default';
		// automatic length determination in AnimationClip.
		var duration = animation.length || -1;
		var fps = animation.fps || 30;

		var hierarchyTracks = animation.hierarchy || [];

		for (var h = 0; h < hierarchyTracks.length; h++) {

			var animationKeys = hierarchyTracks[h].keys;

			// skip empty tracks
			if (!animationKeys || animationKeys.length === 0) continue;

			// process morph targets in a way exactly compatible
			// with AnimationHandler.init( animation )
			if (animationKeys[0].morphTargets) {

				// figure out all morph targets used in this track
				var morphTargetNames = {};
				for (var k = 0; k < animationKeys.length; k++) {

					if (animationKeys[k].morphTargets) {

						for (var m = 0; m < animationKeys[k].morphTargets.length; m++) {

							morphTargetNames[animationKeys[k].morphTargets[m]] = -1;
						}
					}
				}

				// create a track for each morph target with all zero
				// morphTargetInfluences except for the keys in which
				// the morphTarget is named.
				for (var morphTargetName in morphTargetNames) {

					var times = [];
					var values = [];

					for (var m = 0; m !== animationKeys[k].morphTargets.length; ++m) {

						var animationKey = animationKeys[k];

						times.push(animationKey.time);
						values.push(animationKey.morphTarget === morphTargetName ? 1 : 0);
					}

					tracks.push(new NumberKeyframeTrack('.morphTargetInfluence[' + morphTargetName + ']', times, values));
				}

				duration = morphTargetNames.length * (fps || 1.0);
			} else {
				// ...assume skeletal animation

				var boneName = '.bones[' + bones[h].name + ']';

				addNonemptyTrack(VectorKeyframeTrack, boneName + '.position', animationKeys, 'pos', tracks);

				addNonemptyTrack(QuaternionKeyframeTrack, boneName + '.quaternion', animationKeys, 'rot', tracks);

				addNonemptyTrack(VectorKeyframeTrack, boneName + '.scale', animationKeys, 'scl', tracks);
			}
		}

		if (tracks.length === 0) {

			return null;
		}

		var clip = new AnimationClip(clipName, duration, tracks);

		return clip;
	}

});

/**
 * Uniform Utilities
 */

var UniformsUtils = {

	merge: function merge(uniforms) {

		var merged = {};

		for (var u = 0; u < uniforms.length; u++) {

			var tmp = this.clone(uniforms[u]);

			for (var p in tmp) {

				merged[p] = tmp[p];
			}
		}

		return merged;
	},

	clone: function clone(uniforms_src) {

		var uniforms_dst = {};

		for (var u in uniforms_src) {

			uniforms_dst[u] = {};

			for (var p in uniforms_src[u]) {

				var parameter_src = uniforms_src[u][p];

				if (parameter_src && (parameter_src.isColor || parameter_src.isMatrix3 || parameter_src.isMatrix4 || parameter_src.isVector2 || parameter_src.isVector3 || parameter_src.isVector4 || parameter_src.isTexture)) {

					uniforms_dst[u][p] = parameter_src.clone();
				} else if (Array.isArray(parameter_src)) {

					uniforms_dst[u][p] = parameter_src.slice();
				} else {

					uniforms_dst[u][p] = parameter_src;
				}
			}
		}

		return uniforms_dst;
	}

};

/**
 * @author alteredq / http://alteredqualia.com/
 *
 * parameters = {
 *  defines: { "label" : "value" },
 *  uniforms: { "parameter1": { value: 1.0 }, "parameter2": { value2: 2 } },
 *
 *  fragmentShader: <string>,
 *  vertexShader: <string>,
 *
 *  wireframe: <boolean>,
 *  wireframeLinewidth: <float>,
 *
 *  lights: <bool>,
 *
 *  skinning: <bool>,
 *  morphTargets: <bool>,
 *  morphNormals: <bool>
 * }
 */

function ShaderMaterial(parameters) {

	Material.call(this);

	this.type = 'ShaderMaterial';

	this.defines = {};
	this.uniforms = {};

	this.vertexShader = 'void main() {\n\tgl_Position = projectionMatrix * modelViewMatrix * vec4( position, 1.0 );\n}';
	this.fragmentShader = 'void main() {\n\tgl_FragColor = vec4( 1.0, 0.0, 0.0, 1.0 );\n}';

	this.linewidth = 1;

	this.wireframe = false;
	this.wireframeLinewidth = 1;

	this.fog = false; // set to use scene fog
	this.lights = false; // set to use scene lights
	this.clipping = false; // set to use user-defined clipping planes

	this.skinning = false; // set to use skinning attribute streams
	this.morphTargets = false; // set to use morph targets
	this.morphNormals = false; // set to use morph normals

	this.extensions = {
		derivatives: false, // set to use derivatives
		fragDepth: false, // set to use fragment depth values
		drawBuffers: false, // set to use draw buffers
		shaderTextureLOD: false // set to use shader texture LOD
	};

	// When rendered geometry doesn't include these attributes but the material does,
	// use these default values in WebGL. This avoids errors when buffer data is missing.
	this.defaultAttributeValues = {
		'color': [1, 1, 1],
		'uv': [0, 0],
		'uv2': [0, 0]
	};

	this.index0AttributeName = undefined;

	if (parameters !== undefined) {

		if (parameters.attributes !== undefined) {

			console.error('THREE.ShaderMaterial: attributes should now be defined in THREE.BufferGeometry instead.');
		}

		this.setValues(parameters);
	}
}

ShaderMaterial.prototype = Object.create(Material.prototype);
ShaderMaterial.prototype.constructor = ShaderMaterial;

ShaderMaterial.prototype.isShaderMaterial = true;

ShaderMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.fragmentShader = source.fragmentShader;
	this.vertexShader = source.vertexShader;

	this.uniforms = UniformsUtils.clone(source.uniforms);

	this.defines = source.defines;

	this.wireframe = source.wireframe;
	this.wireframeLinewidth = source.wireframeLinewidth;

	this.lights = source.lights;
	this.clipping = source.clipping;

	this.skinning = source.skinning;

	this.morphTargets = source.morphTargets;
	this.morphNormals = source.morphNormals;

	this.extensions = source.extensions;

	return this;
};

ShaderMaterial.prototype.toJSON = function (meta) {

	var data = Material.prototype.toJSON.call(this, meta);

	data.uniforms = this.uniforms;
	data.vertexShader = this.vertexShader;
	data.fragmentShader = this.fragmentShader;

	return data;
};

var alphamap_fragment = "#ifdef USE_ALPHAMAP\n\tdiffuseColor.a *= texture2D( alphaMap, vUv ).g;\n#endif\n";

var alphamap_pars_fragment = "#ifdef USE_ALPHAMAP\n\tuniform sampler2D alphaMap;\n#endif\n";

var alphatest_fragment = "#ifdef ALPHATEST\n\tif ( diffuseColor.a < ALPHATEST ) discard;\n#endif\n";

var aomap_fragment = "#ifdef USE_AOMAP\n\tfloat ambientOcclusion = ( texture2D( aoMap, vUv2 ).r - 1.0 ) * aoMapIntensity + 1.0;\n\treflectedLight.indirectDiffuse *= ambientOcclusion;\n\t#if defined( USE_ENVMAP ) && defined( PHYSICAL )\n\t\tfloat dotNV = saturate( dot( geometry.normal, geometry.viewDir ) );\n\t\treflectedLight.indirectSpecular *= computeSpecularOcclusion( dotNV, ambientOcclusion, material.specularRoughness );\n\t#endif\n#endif\n";

var aomap_pars_fragment = "#ifdef USE_AOMAP\n\tuniform sampler2D aoMap;\n\tuniform float aoMapIntensity;\n#endif";

var begin_vertex = "\nvec3 transformed = vec3( position );\n";

var beginnormal_vertex = "\nvec3 objectNormal = vec3( normal );\n";

var bsdfs = "bool testLightInRange( const in float lightDistance, const in float cutoffDistance ) {\n\treturn any( bvec2( cutoffDistance == 0.0, lightDistance < cutoffDistance ) );\n}\nfloat punctualLightIntensityToIrradianceFactor( const in float lightDistance, const in float cutoffDistance, const in float decayExponent ) {\n\t\tif( decayExponent > 0.0 ) {\n#if defined ( PHYSICALLY_CORRECT_LIGHTS )\n\t\t\tfloat distanceFalloff = 1.0 / max( pow( lightDistance, decayExponent ), 0.01 );\n\t\t\tfloat maxDistanceCutoffFactor = pow2( saturate( 1.0 - pow4( lightDistance / cutoffDistance ) ) );\n\t\t\treturn distanceFalloff * maxDistanceCutoffFactor;\n#else\n\t\t\treturn pow( saturate( -lightDistance / cutoffDistance + 1.0 ), decayExponent );\n#endif\n\t\t}\n\t\treturn 1.0;\n}\nvec3 BRDF_Diffuse_Lambert( const in vec3 diffuseColor ) {\n\treturn RECIPROCAL_PI * diffuseColor;\n}\nvec3 F_Schlick( const in vec3 specularColor, const in float dotLH ) {\n\tfloat fresnel = exp2( ( -5.55473 * dotLH - 6.98316 ) * dotLH );\n\treturn ( 1.0 - specularColor ) * fresnel + specularColor;\n}\nfloat G_GGX_Smith( const in float alpha, const in float dotNL, const in float dotNV ) {\n\tfloat a2 = pow2( alpha );\n\tfloat gl = dotNL + sqrt( a2 + ( 1.0 - a2 ) * pow2( dotNL ) );\n\tfloat gv = dotNV + sqrt( a2 + ( 1.0 - a2 ) * pow2( dotNV ) );\n\treturn 1.0 / ( gl * gv );\n}\nfloat G_GGX_SmithCorrelated( const in float alpha, const in float dotNL, const in float dotNV ) {\n\tfloat a2 = pow2( alpha );\n\tfloat gv = dotNL * sqrt( a2 + ( 1.0 - a2 ) * pow2( dotNV ) );\n\tfloat gl = dotNV * sqrt( a2 + ( 1.0 - a2 ) * pow2( dotNL ) );\n\treturn 0.5 / max( gv + gl, EPSILON );\n}\nfloat D_GGX( const in float alpha, const in float dotNH ) {\n\tfloat a2 = pow2( alpha );\n\tfloat denom = pow2( dotNH ) * ( a2 - 1.0 ) + 1.0;\n\treturn RECIPROCAL_PI * a2 / pow2( denom );\n}\nvec3 BRDF_Specular_GGX( const in IncidentLight incidentLight, const in GeometricContext geometry, const in vec3 specularColor, const in float roughness ) {\n\tfloat alpha = pow2( roughness );\n\tvec3 halfDir = normalize( incidentLight.direction + geometry.viewDir );\n\tfloat dotNL = saturate( dot( geometry.normal, incidentLight.direction ) );\n\tfloat dotNV = saturate( dot( geometry.normal, geometry.viewDir ) );\n\tfloat dotNH = saturate( dot( geometry.normal, halfDir ) );\n\tfloat dotLH = saturate( dot( incidentLight.direction, halfDir ) );\n\tvec3 F = F_Schlick( specularColor, dotLH );\n\tfloat G = G_GGX_SmithCorrelated( alpha, dotNL, dotNV );\n\tfloat D = D_GGX( alpha, dotNH );\n\treturn F * ( G * D );\n}\nvec3 BRDF_Specular_GGX_Environment( const in GeometricContext geometry, const in vec3 specularColor, const in float roughness ) {\n\tfloat dotNV = saturate( dot( geometry.normal, geometry.viewDir ) );\n\tconst vec4 c0 = vec4( - 1, - 0.0275, - 0.572, 0.022 );\n\tconst vec4 c1 = vec4( 1, 0.0425, 1.04, - 0.04 );\n\tvec4 r = roughness * c0 + c1;\n\tfloat a004 = min( r.x * r.x, exp2( - 9.28 * dotNV ) ) * r.x + r.y;\n\tvec2 AB = vec2( -1.04, 1.04 ) * a004 + r.zw;\n\treturn specularColor * AB.x + AB.y;\n}\nfloat G_BlinnPhong_Implicit( ) {\n\treturn 0.25;\n}\nfloat D_BlinnPhong( const in float shininess, const in float dotNH ) {\n\treturn RECIPROCAL_PI * ( shininess * 0.5 + 1.0 ) * pow( dotNH, shininess );\n}\nvec3 BRDF_Specular_BlinnPhong( const in IncidentLight incidentLight, const in GeometricContext geometry, const in vec3 specularColor, const in float shininess ) {\n\tvec3 halfDir = normalize( incidentLight.direction + geometry.viewDir );\n\tfloat dotNH = saturate( dot( geometry.normal, halfDir ) );\n\tfloat dotLH = saturate( dot( incidentLight.direction, halfDir ) );\n\tvec3 F = F_Schlick( specularColor, dotLH );\n\tfloat G = G_BlinnPhong_Implicit( );\n\tfloat D = D_BlinnPhong( shininess, dotNH );\n\treturn F * ( G * D );\n}\nfloat GGXRoughnessToBlinnExponent( const in float ggxRoughness ) {\n\treturn ( 2.0 / pow2( ggxRoughness + 0.0001 ) - 2.0 );\n}\nfloat BlinnExponentToGGXRoughness( const in float blinnExponent ) {\n\treturn sqrt( 2.0 / ( blinnExponent + 2.0 ) );\n}\n";

var bumpmap_pars_fragment = "#ifdef USE_BUMPMAP\n\tuniform sampler2D bumpMap;\n\tuniform float bumpScale;\n\tvec2 dHdxy_fwd() {\n\t\tvec2 dSTdx = dFdx( vUv );\n\t\tvec2 dSTdy = dFdy( vUv );\n\t\tfloat Hll = bumpScale * texture2D( bumpMap, vUv ).x;\n\t\tfloat dBx = bumpScale * texture2D( bumpMap, vUv + dSTdx ).x - Hll;\n\t\tfloat dBy = bumpScale * texture2D( bumpMap, vUv + dSTdy ).x - Hll;\n\t\treturn vec2( dBx, dBy );\n\t}\n\tvec3 perturbNormalArb( vec3 surf_pos, vec3 surf_norm, vec2 dHdxy ) {\n\t\tvec3 vSigmaX = dFdx( surf_pos );\n\t\tvec3 vSigmaY = dFdy( surf_pos );\n\t\tvec3 vN = surf_norm;\n\t\tvec3 R1 = cross( vSigmaY, vN );\n\t\tvec3 R2 = cross( vN, vSigmaX );\n\t\tfloat fDet = dot( vSigmaX, R1 );\n\t\tvec3 vGrad = sign( fDet ) * ( dHdxy.x * R1 + dHdxy.y * R2 );\n\t\treturn normalize( abs( fDet ) * surf_norm - vGrad );\n\t}\n#endif\n";

var clipping_planes_fragment = "#if NUM_CLIPPING_PLANES > 0\n\tfor ( int i = 0; i < UNION_CLIPPING_PLANES; ++ i ) {\n\t\tvec4 plane = clippingPlanes[ i ];\n\t\tif ( dot( vViewPosition, plane.xyz ) > plane.w ) discard;\n\t}\n\t\t\n\t#if UNION_CLIPPING_PLANES < NUM_CLIPPING_PLANES\n\t\tbool clipped = true;\n\t\tfor ( int i = UNION_CLIPPING_PLANES; i < NUM_CLIPPING_PLANES; ++ i ) {\n\t\t\tvec4 plane = clippingPlanes[ i ];\n\t\t\tclipped = ( dot( vViewPosition, plane.xyz ) > plane.w ) && clipped;\n\t\t}\n\t\tif ( clipped ) discard;\n\t\n\t#endif\n#endif\n";

var clipping_planes_pars_fragment = "#if NUM_CLIPPING_PLANES > 0\n\t#if ! defined( PHYSICAL ) && ! defined( PHONG )\n\t\tvarying vec3 vViewPosition;\n\t#endif\n\tuniform vec4 clippingPlanes[ NUM_CLIPPING_PLANES ];\n#endif\n";

var clipping_planes_pars_vertex = "#if NUM_CLIPPING_PLANES > 0 && ! defined( PHYSICAL ) && ! defined( PHONG )\n\tvarying vec3 vViewPosition;\n#endif\n";

var clipping_planes_vertex = "#if NUM_CLIPPING_PLANES > 0 && ! defined( PHYSICAL ) && ! defined( PHONG )\n\tvViewPosition = - mvPosition.xyz;\n#endif\n";

var color_fragment = "#ifdef USE_COLOR\n\tdiffuseColor.rgb *= vColor;\n#endif";

var color_pars_fragment = "#ifdef USE_COLOR\n\tvarying vec3 vColor;\n#endif\n";

var color_pars_vertex = "#ifdef USE_COLOR\n\tvarying vec3 vColor;\n#endif";

var color_vertex = "#ifdef USE_COLOR\n\tvColor.xyz = color.xyz;\n#endif";

var common = "#define PI 3.14159265359\n#define PI2 6.28318530718\n#define RECIPROCAL_PI 0.31830988618\n#define RECIPROCAL_PI2 0.15915494\n#define LOG2 1.442695\n#define EPSILON 1e-6\n#define saturate(a) clamp( a, 0.0, 1.0 )\n#define whiteCompliment(a) ( 1.0 - saturate( a ) )\nfloat pow2( const in float x ) { return x*x; }\nfloat pow3( const in float x ) { return x*x*x; }\nfloat pow4( const in float x ) { float x2 = x*x; return x2*x2; }\nfloat average( const in vec3 color ) { return dot( color, vec3( 0.3333 ) ); }\nhighp float rand( const in vec2 uv ) {\n\tconst highp float a = 12.9898, b = 78.233, c = 43758.5453;\n\thighp float dt = dot( uv.xy, vec2( a,b ) ), sn = mod( dt, PI );\n\treturn fract(sin(sn) * c);\n}\nstruct IncidentLight {\n\tvec3 color;\n\tvec3 direction;\n\tbool visible;\n};\nstruct ReflectedLight {\n\tvec3 directDiffuse;\n\tvec3 directSpecular;\n\tvec3 indirectDiffuse;\n\tvec3 indirectSpecular;\n};\nstruct GeometricContext {\n\tvec3 position;\n\tvec3 normal;\n\tvec3 viewDir;\n};\nvec3 transformDirection( in vec3 dir, in mat4 matrix ) {\n\treturn normalize( ( matrix * vec4( dir, 0.0 ) ).xyz );\n}\nvec3 inverseTransformDirection( in vec3 dir, in mat4 matrix ) {\n\treturn normalize( ( vec4( dir, 0.0 ) * matrix ).xyz );\n}\nvec3 projectOnPlane(in vec3 point, in vec3 pointOnPlane, in vec3 planeNormal ) {\n\tfloat distance = dot( planeNormal, point - pointOnPlane );\n\treturn - distance * planeNormal + point;\n}\nfloat sideOfPlane( in vec3 point, in vec3 pointOnPlane, in vec3 planeNormal ) {\n\treturn sign( dot( point - pointOnPlane, planeNormal ) );\n}\nvec3 linePlaneIntersect( in vec3 pointOnLine, in vec3 lineDirection, in vec3 pointOnPlane, in vec3 planeNormal ) {\n\treturn lineDirection * ( dot( planeNormal, pointOnPlane - pointOnLine ) / dot( planeNormal, lineDirection ) ) + pointOnLine;\n}\n";

var cube_uv_reflection_fragment = "#ifdef ENVMAP_TYPE_CUBE_UV\n#define cubeUV_textureSize (1024.0)\nint getFaceFromDirection(vec3 direction) {\n\tvec3 absDirection = abs(direction);\n\tint face = -1;\n\tif( absDirection.x > absDirection.z ) {\n\t\tif(absDirection.x > absDirection.y )\n\t\t\tface = direction.x > 0.0 ? 0 : 3;\n\t\telse\n\t\t\tface = direction.y > 0.0 ? 1 : 4;\n\t}\n\telse {\n\t\tif(absDirection.z > absDirection.y )\n\t\t\tface = direction.z > 0.0 ? 2 : 5;\n\t\telse\n\t\t\tface = direction.y > 0.0 ? 1 : 4;\n\t}\n\treturn face;\n}\n#define cubeUV_maxLods1  (log2(cubeUV_textureSize*0.25) - 1.0)\n#define cubeUV_rangeClamp (exp2((6.0 - 1.0) * 2.0))\nvec2 MipLevelInfo( vec3 vec, float roughnessLevel, float roughness ) {\n\tfloat scale = exp2(cubeUV_maxLods1 - roughnessLevel);\n\tfloat dxRoughness = dFdx(roughness);\n\tfloat dyRoughness = dFdy(roughness);\n\tvec3 dx = dFdx( vec * scale * dxRoughness );\n\tvec3 dy = dFdy( vec * scale * dyRoughness );\n\tfloat d = max( dot( dx, dx ), dot( dy, dy ) );\n\td = clamp(d, 1.0, cubeUV_rangeClamp);\n\tfloat mipLevel = 0.5 * log2(d);\n\treturn vec2(floor(mipLevel), fract(mipLevel));\n}\n#define cubeUV_maxLods2 (log2(cubeUV_textureSize*0.25) - 2.0)\n#define cubeUV_rcpTextureSize (1.0 / cubeUV_textureSize)\nvec2 getCubeUV(vec3 direction, float roughnessLevel, float mipLevel) {\n\tmipLevel = roughnessLevel > cubeUV_maxLods2 - 3.0 ? 0.0 : mipLevel;\n\tfloat a = 16.0 * cubeUV_rcpTextureSize;\n\tvec2 exp2_packed = exp2( vec2( roughnessLevel, mipLevel ) );\n\tvec2 rcp_exp2_packed = vec2( 1.0 ) / exp2_packed;\n\tfloat powScale = exp2_packed.x * exp2_packed.y;\n\tfloat scale = rcp_exp2_packed.x * rcp_exp2_packed.y * 0.25;\n\tfloat mipOffset = 0.75*(1.0 - rcp_exp2_packed.y) * rcp_exp2_packed.x;\n\tbool bRes = mipLevel == 0.0;\n\tscale =  bRes && (scale < a) ? a : scale;\n\tvec3 r;\n\tvec2 offset;\n\tint face = getFaceFromDirection(direction);\n\tfloat rcpPowScale = 1.0 / powScale;\n\tif( face == 0) {\n\t\tr = vec3(direction.x, -direction.z, direction.y);\n\t\toffset = vec2(0.0+mipOffset,0.75 * rcpPowScale);\n\t\toffset.y = bRes && (offset.y < 2.0*a) ?  a : offset.y;\n\t}\n\telse if( face == 1) {\n\t\tr = vec3(direction.y, direction.x, direction.z);\n\t\toffset = vec2(scale+mipOffset, 0.75 * rcpPowScale);\n\t\toffset.y = bRes && (offset.y < 2.0*a) ?  a : offset.y;\n\t}\n\telse if( face == 2) {\n\t\tr = vec3(direction.z, direction.x, direction.y);\n\t\toffset = vec2(2.0*scale+mipOffset, 0.75 * rcpPowScale);\n\t\toffset.y = bRes && (offset.y < 2.0*a) ?  a : offset.y;\n\t}\n\telse if( face == 3) {\n\t\tr = vec3(direction.x, direction.z, direction.y);\n\t\toffset = vec2(0.0+mipOffset,0.5 * rcpPowScale);\n\t\toffset.y = bRes && (offset.y < 2.0*a) ?  0.0 : offset.y;\n\t}\n\telse if( face == 4) {\n\t\tr = vec3(direction.y, direction.x, -direction.z);\n\t\toffset = vec2(scale+mipOffset, 0.5 * rcpPowScale);\n\t\toffset.y = bRes && (offset.y < 2.0*a) ?  0.0 : offset.y;\n\t}\n\telse {\n\t\tr = vec3(direction.z, -direction.x, direction.y);\n\t\toffset = vec2(2.0*scale+mipOffset, 0.5 * rcpPowScale);\n\t\toffset.y = bRes && (offset.y < 2.0*a) ?  0.0 : offset.y;\n\t}\n\tr = normalize(r);\n\tfloat texelOffset = 0.5 * cubeUV_rcpTextureSize;\n\tvec2 s = ( r.yz / abs( r.x ) + vec2( 1.0 ) ) * 0.5;\n\tvec2 base = offset + vec2( texelOffset );\n\treturn base + s * ( scale - 2.0 * texelOffset );\n}\n#define cubeUV_maxLods3 (log2(cubeUV_textureSize*0.25) - 3.0)\nvec4 textureCubeUV(vec3 reflectedDirection, float roughness ) {\n\tfloat roughnessVal = roughness* cubeUV_maxLods3;\n\tfloat r1 = floor(roughnessVal);\n\tfloat r2 = r1 + 1.0;\n\tfloat t = fract(roughnessVal);\n\tvec2 mipInfo = MipLevelInfo(reflectedDirection, r1, roughness);\n\tfloat s = mipInfo.y;\n\tfloat level0 = mipInfo.x;\n\tfloat level1 = level0 + 1.0;\n\tlevel1 = level1 > 5.0 ? 5.0 : level1;\n\tlevel0 += min( floor( s + 0.5 ), 5.0 );\n\tvec2 uv_10 = getCubeUV(reflectedDirection, r1, level0);\n\tvec4 color10 = envMapTexelToLinear(texture2D(envMap, uv_10));\n\tvec2 uv_20 = getCubeUV(reflectedDirection, r2, level0);\n\tvec4 color20 = envMapTexelToLinear(texture2D(envMap, uv_20));\n\tvec4 result = mix(color10, color20, t);\n\treturn vec4(result.rgb, 1.0);\n}\n#endif\n";

var defaultnormal_vertex = "#ifdef FLIP_SIDED\n\tobjectNormal = -objectNormal;\n#endif\nvec3 transformedNormal = normalMatrix * objectNormal;\n";

var displacementmap_pars_vertex = "#ifdef USE_DISPLACEMENTMAP\n\tuniform sampler2D displacementMap;\n\tuniform float displacementScale;\n\tuniform float displacementBias;\n#endif\n";

var displacementmap_vertex = "#ifdef USE_DISPLACEMENTMAP\n\ttransformed += normal * ( texture2D( displacementMap, uv ).x * displacementScale + displacementBias );\n#endif\n";

var emissivemap_fragment = "#ifdef USE_EMISSIVEMAP\n\tvec4 emissiveColor = texture2D( emissiveMap, vUv );\n\temissiveColor.rgb = emissiveMapTexelToLinear( emissiveColor ).rgb;\n\ttotalEmissiveRadiance *= emissiveColor.rgb;\n#endif\n";

var emissivemap_pars_fragment = "#ifdef USE_EMISSIVEMAP\n\tuniform sampler2D emissiveMap;\n#endif\n";

var encodings_fragment = "  gl_FragColor = linearToOutputTexel( gl_FragColor );\n";

var encodings_pars_fragment = "\nvec4 LinearToLinear( in vec4 value ) {\n  return value;\n}\nvec4 GammaToLinear( in vec4 value, in float gammaFactor ) {\n  return vec4( pow( value.xyz, vec3( gammaFactor ) ), value.w );\n}\nvec4 LinearToGamma( in vec4 value, in float gammaFactor ) {\n  return vec4( pow( value.xyz, vec3( 1.0 / gammaFactor ) ), value.w );\n}\nvec4 sRGBToLinear( in vec4 value ) {\n  return vec4( mix( pow( value.rgb * 0.9478672986 + vec3( 0.0521327014 ), vec3( 2.4 ) ), value.rgb * 0.0773993808, vec3( lessThanEqual( value.rgb, vec3( 0.04045 ) ) ) ), value.w );\n}\nvec4 LinearTosRGB( in vec4 value ) {\n  return vec4( mix( pow( value.rgb, vec3( 0.41666 ) ) * 1.055 - vec3( 0.055 ), value.rgb * 12.92, vec3( lessThanEqual( value.rgb, vec3( 0.0031308 ) ) ) ), value.w );\n}\nvec4 RGBEToLinear( in vec4 value ) {\n  return vec4( value.rgb * exp2( value.a * 255.0 - 128.0 ), 1.0 );\n}\nvec4 LinearToRGBE( in vec4 value ) {\n  float maxComponent = max( max( value.r, value.g ), value.b );\n  float fExp = clamp( ceil( log2( maxComponent ) ), -128.0, 127.0 );\n  return vec4( value.rgb / exp2( fExp ), ( fExp + 128.0 ) / 255.0 );\n}\nvec4 RGBMToLinear( in vec4 value, in float maxRange ) {\n  return vec4( value.xyz * value.w * maxRange, 1.0 );\n}\nvec4 LinearToRGBM( in vec4 value, in float maxRange ) {\n  float maxRGB = max( value.x, max( value.g, value.b ) );\n  float M      = clamp( maxRGB / maxRange, 0.0, 1.0 );\n  M            = ceil( M * 255.0 ) / 255.0;\n  return vec4( value.rgb / ( M * maxRange ), M );\n}\nvec4 RGBDToLinear( in vec4 value, in float maxRange ) {\n    return vec4( value.rgb * ( ( maxRange / 255.0 ) / value.a ), 1.0 );\n}\nvec4 LinearToRGBD( in vec4 value, in float maxRange ) {\n    float maxRGB = max( value.x, max( value.g, value.b ) );\n    float D      = max( maxRange / maxRGB, 1.0 );\n    D            = min( floor( D ) / 255.0, 1.0 );\n    return vec4( value.rgb * ( D * ( 255.0 / maxRange ) ), D );\n}\nconst mat3 cLogLuvM = mat3( 0.2209, 0.3390, 0.4184, 0.1138, 0.6780, 0.7319, 0.0102, 0.1130, 0.2969 );\nvec4 LinearToLogLuv( in vec4 value )  {\n  vec3 Xp_Y_XYZp = value.rgb * cLogLuvM;\n  Xp_Y_XYZp = max(Xp_Y_XYZp, vec3(1e-6, 1e-6, 1e-6));\n  vec4 vResult;\n  vResult.xy = Xp_Y_XYZp.xy / Xp_Y_XYZp.z;\n  float Le = 2.0 * log2(Xp_Y_XYZp.y) + 127.0;\n  vResult.w = fract(Le);\n  vResult.z = (Le - (floor(vResult.w*255.0))/255.0)/255.0;\n  return vResult;\n}\nconst mat3 cLogLuvInverseM = mat3( 6.0014, -2.7008, -1.7996, -1.3320, 3.1029, -5.7721, 0.3008, -1.0882, 5.6268 );\nvec4 LogLuvToLinear( in vec4 value ) {\n  float Le = value.z * 255.0 + value.w;\n  vec3 Xp_Y_XYZp;\n  Xp_Y_XYZp.y = exp2((Le - 127.0) / 2.0);\n  Xp_Y_XYZp.z = Xp_Y_XYZp.y / value.y;\n  Xp_Y_XYZp.x = value.x * Xp_Y_XYZp.z;\n  vec3 vRGB = Xp_Y_XYZp.rgb * cLogLuvInverseM;\n  return vec4( max(vRGB, 0.0), 1.0 );\n}\n";

var envmap_fragment = "#ifdef USE_ENVMAP\n\t#if defined( USE_BUMPMAP ) || defined( USE_NORMALMAP ) || defined( PHONG )\n\t\tvec3 cameraToVertex = normalize( vWorldPosition - cameraPosition );\n\t\tvec3 worldNormal = inverseTransformDirection( normal, viewMatrix );\n\t\t#ifdef ENVMAP_MODE_REFLECTION\n\t\t\tvec3 reflectVec = reflect( cameraToVertex, worldNormal );\n\t\t#else\n\t\t\tvec3 reflectVec = refract( cameraToVertex, worldNormal, refractionRatio );\n\t\t#endif\n\t#else\n\t\tvec3 reflectVec = vReflect;\n\t#endif\n\t#ifdef ENVMAP_TYPE_CUBE\n\t\tvec4 envColor = textureCube( envMap, flipNormal * vec3( flipEnvMap * reflectVec.x, reflectVec.yz ) );\n\t#elif defined( ENVMAP_TYPE_EQUIREC )\n\t\tvec2 sampleUV;\n\t\tsampleUV.y = saturate( flipNormal * reflectVec.y * 0.5 + 0.5 );\n\t\tsampleUV.x = atan( flipNormal * reflectVec.z, flipNormal * reflectVec.x ) * RECIPROCAL_PI2 + 0.5;\n\t\tvec4 envColor = texture2D( envMap, sampleUV );\n\t#elif defined( ENVMAP_TYPE_SPHERE )\n\t\tvec3 reflectView = flipNormal * normalize( ( viewMatrix * vec4( reflectVec, 0.0 ) ).xyz + vec3( 0.0, 0.0, 1.0 ) );\n\t\tvec4 envColor = texture2D( envMap, reflectView.xy * 0.5 + 0.5 );\n\t#else\n\t\tvec4 envColor = vec4( 0.0 );\n\t#endif\n\tenvColor = envMapTexelToLinear( envColor );\n\t#ifdef ENVMAP_BLENDING_MULTIPLY\n\t\toutgoingLight = mix( outgoingLight, outgoingLight * envColor.xyz, specularStrength * reflectivity );\n\t#elif defined( ENVMAP_BLENDING_MIX )\n\t\toutgoingLight = mix( outgoingLight, envColor.xyz, specularStrength * reflectivity );\n\t#elif defined( ENVMAP_BLENDING_ADD )\n\t\toutgoingLight += envColor.xyz * specularStrength * reflectivity;\n\t#endif\n#endif\n";

var envmap_pars_fragment = "#if defined( USE_ENVMAP ) || defined( PHYSICAL )\n\tuniform float reflectivity;\n\tuniform float envMapIntenstiy;\n#endif\n#ifdef USE_ENVMAP\n\t#if ! defined( PHYSICAL ) && ( defined( USE_BUMPMAP ) || defined( USE_NORMALMAP ) || defined( PHONG ) )\n\t\tvarying vec3 vWorldPosition;\n\t#endif\n\t#ifdef ENVMAP_TYPE_CUBE\n\t\tuniform samplerCube envMap;\n\t#else\n\t\tuniform sampler2D envMap;\n\t#endif\n\tuniform float flipEnvMap;\n\t#if defined( USE_BUMPMAP ) || defined( USE_NORMALMAP ) || defined( PHONG ) || defined( PHYSICAL )\n\t\tuniform float refractionRatio;\n\t#else\n\t\tvarying vec3 vReflect;\n\t#endif\n#endif\n";

var envmap_pars_vertex = "#ifdef USE_ENVMAP\n\t#if defined( USE_BUMPMAP ) || defined( USE_NORMALMAP ) || defined( PHONG )\n\t\tvarying vec3 vWorldPosition;\n\t#else\n\t\tvarying vec3 vReflect;\n\t\tuniform float refractionRatio;\n\t#endif\n#endif\n";

var envmap_vertex = "#ifdef USE_ENVMAP\n\t#if defined( USE_BUMPMAP ) || defined( USE_NORMALMAP ) || defined( PHONG )\n\t\tvWorldPosition = worldPosition.xyz;\n\t#else\n\t\tvec3 cameraToVertex = normalize( worldPosition.xyz - cameraPosition );\n\t\tvec3 worldNormal = inverseTransformDirection( transformedNormal, viewMatrix );\n\t\t#ifdef ENVMAP_MODE_REFLECTION\n\t\t\tvReflect = reflect( cameraToVertex, worldNormal );\n\t\t#else\n\t\t\tvReflect = refract( cameraToVertex, worldNormal, refractionRatio );\n\t\t#endif\n\t#endif\n#endif\n";

var fog_fragment = "#ifdef USE_FOG\n\t#ifdef USE_LOGDEPTHBUF_EXT\n\t\tfloat depth = gl_FragDepthEXT / gl_FragCoord.w;\n\t#else\n\t\tfloat depth = gl_FragCoord.z / gl_FragCoord.w;\n\t#endif\n\t#ifdef FOG_EXP2\n\t\tfloat fogFactor = whiteCompliment( exp2( - fogDensity * fogDensity * depth * depth * LOG2 ) );\n\t#else\n\t\tfloat fogFactor = smoothstep( fogNear, fogFar, depth );\n\t#endif\n\tgl_FragColor.rgb = mix( gl_FragColor.rgb, fogColor, fogFactor );\n#endif\n";

var fog_pars_fragment = "#ifdef USE_FOG\n\tuniform vec3 fogColor;\n\t#ifdef FOG_EXP2\n\t\tuniform float fogDensity;\n\t#else\n\t\tuniform float fogNear;\n\t\tuniform float fogFar;\n\t#endif\n#endif";

var lightmap_fragment = "#ifdef USE_LIGHTMAP\n\treflectedLight.indirectDiffuse += PI * texture2D( lightMap, vUv2 ).xyz * lightMapIntensity;\n#endif\n";

var lightmap_pars_fragment = "#ifdef USE_LIGHTMAP\n\tuniform sampler2D lightMap;\n\tuniform float lightMapIntensity;\n#endif";

var lights_lambert_vertex = "vec3 diffuse = vec3( 1.0 );\nGeometricContext geometry;\ngeometry.position = mvPosition.xyz;\ngeometry.normal = normalize( transformedNormal );\ngeometry.viewDir = normalize( -mvPosition.xyz );\nGeometricContext backGeometry;\nbackGeometry.position = geometry.position;\nbackGeometry.normal = -geometry.normal;\nbackGeometry.viewDir = geometry.viewDir;\nvLightFront = vec3( 0.0 );\n#ifdef DOUBLE_SIDED\n\tvLightBack = vec3( 0.0 );\n#endif\nIncidentLight directLight;\nfloat dotNL;\nvec3 directLightColor_Diffuse;\n#if NUM_POINT_LIGHTS > 0\n\tfor ( int i = 0; i < NUM_POINT_LIGHTS; i ++ ) {\n\t\tgetPointDirectLightIrradiance( pointLights[ i ], geometry, directLight );\n\t\tdotNL = dot( geometry.normal, directLight.direction );\n\t\tdirectLightColor_Diffuse = PI * directLight.color;\n\t\tvLightFront += saturate( dotNL ) * directLightColor_Diffuse;\n\t\t#ifdef DOUBLE_SIDED\n\t\t\tvLightBack += saturate( -dotNL ) * directLightColor_Diffuse;\n\t\t#endif\n\t}\n#endif\n#if NUM_SPOT_LIGHTS > 0\n\tfor ( int i = 0; i < NUM_SPOT_LIGHTS; i ++ ) {\n\t\tgetSpotDirectLightIrradiance( spotLights[ i ], geometry, directLight );\n\t\tdotNL = dot( geometry.normal, directLight.direction );\n\t\tdirectLightColor_Diffuse = PI * directLight.color;\n\t\tvLightFront += saturate( dotNL ) * directLightColor_Diffuse;\n\t\t#ifdef DOUBLE_SIDED\n\t\t\tvLightBack += saturate( -dotNL ) * directLightColor_Diffuse;\n\t\t#endif\n\t}\n#endif\n#if NUM_DIR_LIGHTS > 0\n\tfor ( int i = 0; i < NUM_DIR_LIGHTS; i ++ ) {\n\t\tgetDirectionalDirectLightIrradiance( directionalLights[ i ], geometry, directLight );\n\t\tdotNL = dot( geometry.normal, directLight.direction );\n\t\tdirectLightColor_Diffuse = PI * directLight.color;\n\t\tvLightFront += saturate( dotNL ) * directLightColor_Diffuse;\n\t\t#ifdef DOUBLE_SIDED\n\t\t\tvLightBack += saturate( -dotNL ) * directLightColor_Diffuse;\n\t\t#endif\n\t}\n#endif\n#if NUM_HEMI_LIGHTS > 0\n\tfor ( int i = 0; i < NUM_HEMI_LIGHTS; i ++ ) {\n\t\tvLightFront += getHemisphereLightIrradiance( hemisphereLights[ i ], geometry );\n\t\t#ifdef DOUBLE_SIDED\n\t\t\tvLightBack += getHemisphereLightIrradiance( hemisphereLights[ i ], backGeometry );\n\t\t#endif\n\t}\n#endif\n";

var lights_pars = "uniform vec3 ambientLightColor;\nvec3 getAmbientLightIrradiance( const in vec3 ambientLightColor ) {\n\tvec3 irradiance = ambientLightColor;\n\t#ifndef PHYSICALLY_CORRECT_LIGHTS\n\t\tirradiance *= PI;\n\t#endif\n\treturn irradiance;\n}\n#if NUM_DIR_LIGHTS > 0\n\tstruct DirectionalLight {\n\t\tvec3 direction;\n\t\tvec3 color;\n\t\tint shadow;\n\t\tfloat shadowBias;\n\t\tfloat shadowRadius;\n\t\tvec2 shadowMapSize;\n\t};\n\tuniform DirectionalLight directionalLights[ NUM_DIR_LIGHTS ];\n\tvoid getDirectionalDirectLightIrradiance( const in DirectionalLight directionalLight, const in GeometricContext geometry, out IncidentLight directLight ) {\n\t\tdirectLight.color = directionalLight.color;\n\t\tdirectLight.direction = directionalLight.direction;\n\t\tdirectLight.visible = true;\n\t}\n#endif\n#if NUM_POINT_LIGHTS > 0\n\tstruct PointLight {\n\t\tvec3 position;\n\t\tvec3 color;\n\t\tfloat distance;\n\t\tfloat decay;\n\t\tint shadow;\n\t\tfloat shadowBias;\n\t\tfloat shadowRadius;\n\t\tvec2 shadowMapSize;\n\t};\n\tuniform PointLight pointLights[ NUM_POINT_LIGHTS ];\n\tvoid getPointDirectLightIrradiance( const in PointLight pointLight, const in GeometricContext geometry, out IncidentLight directLight ) {\n\t\tvec3 lVector = pointLight.position - geometry.position;\n\t\tdirectLight.direction = normalize( lVector );\n\t\tfloat lightDistance = length( lVector );\n\t\tif ( testLightInRange( lightDistance, pointLight.distance ) ) {\n\t\t\tdirectLight.color = pointLight.color;\n\t\t\tdirectLight.color *= punctualLightIntensityToIrradianceFactor( lightDistance, pointLight.distance, pointLight.decay );\n\t\t\tdirectLight.visible = true;\n\t\t} else {\n\t\t\tdirectLight.color = vec3( 0.0 );\n\t\t\tdirectLight.visible = false;\n\t\t}\n\t}\n#endif\n#if NUM_SPOT_LIGHTS > 0\n\tstruct SpotLight {\n\t\tvec3 position;\n\t\tvec3 direction;\n\t\tvec3 color;\n\t\tfloat distance;\n\t\tfloat decay;\n\t\tfloat coneCos;\n\t\tfloat penumbraCos;\n\t\tint shadow;\n\t\tfloat shadowBias;\n\t\tfloat shadowRadius;\n\t\tvec2 shadowMapSize;\n\t};\n\tuniform SpotLight spotLights[ NUM_SPOT_LIGHTS ];\n\tvoid getSpotDirectLightIrradiance( const in SpotLight spotLight, const in GeometricContext geometry, out IncidentLight directLight  ) {\n\t\tvec3 lVector = spotLight.position - geometry.position;\n\t\tdirectLight.direction = normalize( lVector );\n\t\tfloat lightDistance = length( lVector );\n\t\tfloat angleCos = dot( directLight.direction, spotLight.direction );\n\t\tif ( all( bvec2( angleCos > spotLight.coneCos, testLightInRange( lightDistance, spotLight.distance ) ) ) ) {\n\t\t\tfloat spotEffect = smoothstep( spotLight.coneCos, spotLight.penumbraCos, angleCos );\n\t\t\tdirectLight.color = spotLight.color;\n\t\t\tdirectLight.color *= spotEffect * punctualLightIntensityToIrradianceFactor( lightDistance, spotLight.distance, spotLight.decay );\n\t\t\tdirectLight.visible = true;\n\t\t} else {\n\t\t\tdirectLight.color = vec3( 0.0 );\n\t\t\tdirectLight.visible = false;\n\t\t}\n\t}\n#endif\n#if NUM_HEMI_LIGHTS > 0\n\tstruct HemisphereLight {\n\t\tvec3 direction;\n\t\tvec3 skyColor;\n\t\tvec3 groundColor;\n\t};\n\tuniform HemisphereLight hemisphereLights[ NUM_HEMI_LIGHTS ];\n\tvec3 getHemisphereLightIrradiance( const in HemisphereLight hemiLight, const in GeometricContext geometry ) {\n\t\tfloat dotNL = dot( geometry.normal, hemiLight.direction );\n\t\tfloat hemiDiffuseWeight = 0.5 * dotNL + 0.5;\n\t\tvec3 irradiance = mix( hemiLight.groundColor, hemiLight.skyColor, hemiDiffuseWeight );\n\t\t#ifndef PHYSICALLY_CORRECT_LIGHTS\n\t\t\tirradiance *= PI;\n\t\t#endif\n\t\treturn irradiance;\n\t}\n#endif\n#if defined( USE_ENVMAP ) && defined( PHYSICAL )\n\tvec3 getLightProbeIndirectIrradiance( const in GeometricContext geometry, const in int maxMIPLevel ) {\n\t\t#include <normal_flip>\n\t\tvec3 worldNormal = inverseTransformDirection( geometry.normal, viewMatrix );\n\t\t#ifdef ENVMAP_TYPE_CUBE\n\t\t\tvec3 queryVec = flipNormal * vec3( flipEnvMap * worldNormal.x, worldNormal.yz );\n\t\t\t#ifdef TEXTURE_LOD_EXT\n\t\t\t\tvec4 envMapColor = textureCubeLodEXT( envMap, queryVec, float( maxMIPLevel ) );\n\t\t\t#else\n\t\t\t\tvec4 envMapColor = textureCube( envMap, queryVec, float( maxMIPLevel ) );\n\t\t\t#endif\n\t\t\tenvMapColor.rgb = envMapTexelToLinear( envMapColor ).rgb;\n\t\t#elif defined( ENVMAP_TYPE_CUBE_UV )\n\t\t\tvec3 queryVec = flipNormal * vec3( flipEnvMap * worldNormal.x, worldNormal.yz );\n\t\t\tvec4 envMapColor = textureCubeUV( queryVec, 1.0 );\n\t\t#else\n\t\t\tvec4 envMapColor = vec4( 0.0 );\n\t\t#endif\n\t\treturn PI * envMapColor.rgb * envMapIntensity;\n\t}\n\tfloat getSpecularMIPLevel( const in float blinnShininessExponent, const in int maxMIPLevel ) {\n\t\tfloat maxMIPLevelScalar = float( maxMIPLevel );\n\t\tfloat desiredMIPLevel = maxMIPLevelScalar - 0.79248 - 0.5 * log2( pow2( blinnShininessExponent ) + 1.0 );\n\t\treturn clamp( desiredMIPLevel, 0.0, maxMIPLevelScalar );\n\t}\n\tvec3 getLightProbeIndirectRadiance( const in GeometricContext geometry, const in float blinnShininessExponent, const in int maxMIPLevel ) {\n\t\t#ifdef ENVMAP_MODE_REFLECTION\n\t\t\tvec3 reflectVec = reflect( -geometry.viewDir, geometry.normal );\n\t\t#else\n\t\t\tvec3 reflectVec = refract( -geometry.viewDir, geometry.normal, refractionRatio );\n\t\t#endif\n\t\t#include <normal_flip>\n\t\treflectVec = inverseTransformDirection( reflectVec, viewMatrix );\n\t\tfloat specularMIPLevel = getSpecularMIPLevel( blinnShininessExponent, maxMIPLevel );\n\t\t#ifdef ENVMAP_TYPE_CUBE\n\t\t\tvec3 queryReflectVec = flipNormal * vec3( flipEnvMap * reflectVec.x, reflectVec.yz );\n\t\t\t#ifdef TEXTURE_LOD_EXT\n\t\t\t\tvec4 envMapColor = textureCubeLodEXT( envMap, queryReflectVec, specularMIPLevel );\n\t\t\t#else\n\t\t\t\tvec4 envMapColor = textureCube( envMap, queryReflectVec, specularMIPLevel );\n\t\t\t#endif\n\t\t\tenvMapColor.rgb = envMapTexelToLinear( envMapColor ).rgb;\n\t\t#elif defined( ENVMAP_TYPE_CUBE_UV )\n\t\t\tvec3 queryReflectVec = flipNormal * vec3( flipEnvMap * reflectVec.x, reflectVec.yz );\n\t\t\tvec4 envMapColor = textureCubeUV(queryReflectVec, BlinnExponentToGGXRoughness(blinnShininessExponent));\n\t\t#elif defined( ENVMAP_TYPE_EQUIREC )\n\t\t\tvec2 sampleUV;\n\t\t\tsampleUV.y = saturate( flipNormal * reflectVec.y * 0.5 + 0.5 );\n\t\t\tsampleUV.x = atan( flipNormal * reflectVec.z, flipNormal * reflectVec.x ) * RECIPROCAL_PI2 + 0.5;\n\t\t\t#ifdef TEXTURE_LOD_EXT\n\t\t\t\tvec4 envMapColor = texture2DLodEXT( envMap, sampleUV, specularMIPLevel );\n\t\t\t#else\n\t\t\t\tvec4 envMapColor = texture2D( envMap, sampleUV, specularMIPLevel );\n\t\t\t#endif\n\t\t\tenvMapColor.rgb = envMapTexelToLinear( envMapColor ).rgb;\n\t\t#elif defined( ENVMAP_TYPE_SPHERE )\n\t\t\tvec3 reflectView = flipNormal * normalize( ( viewMatrix * vec4( reflectVec, 0.0 ) ).xyz + vec3( 0.0,0.0,1.0 ) );\n\t\t\t#ifdef TEXTURE_LOD_EXT\n\t\t\t\tvec4 envMapColor = texture2DLodEXT( envMap, reflectView.xy * 0.5 + 0.5, specularMIPLevel );\n\t\t\t#else\n\t\t\t\tvec4 envMapColor = texture2D( envMap, reflectView.xy * 0.5 + 0.5, specularMIPLevel );\n\t\t\t#endif\n\t\t\tenvMapColor.rgb = envMapTexelToLinear( envMapColor ).rgb;\n\t\t#endif\n\t\treturn envMapColor.rgb * envMapIntensity;\n\t}\n#endif\n";

var lights_phong_fragment = "BlinnPhongMaterial material;\nmaterial.diffuseColor = diffuseColor.rgb;\nmaterial.specularColor = specular;\nmaterial.specularShininess = shininess;\nmaterial.specularStrength = specularStrength;\n";

var lights_phong_pars_fragment = "varying vec3 vViewPosition;\n#ifndef FLAT_SHADED\n\tvarying vec3 vNormal;\n#endif\nstruct BlinnPhongMaterial {\n\tvec3\tdiffuseColor;\n\tvec3\tspecularColor;\n\tfloat\tspecularShininess;\n\tfloat\tspecularStrength;\n};\nvoid RE_Direct_BlinnPhong( const in IncidentLight directLight, const in GeometricContext geometry, const in BlinnPhongMaterial material, inout ReflectedLight reflectedLight ) {\n\tfloat dotNL = saturate( dot( geometry.normal, directLight.direction ) );\n\tvec3 irradiance = dotNL * directLight.color;\n\t#ifndef PHYSICALLY_CORRECT_LIGHTS\n\t\tirradiance *= PI;\n\t#endif\n\treflectedLight.directDiffuse += irradiance * BRDF_Diffuse_Lambert( material.diffuseColor );\n\treflectedLight.directSpecular += irradiance * BRDF_Specular_BlinnPhong( directLight, geometry, material.specularColor, material.specularShininess ) * material.specularStrength;\n}\nvoid RE_IndirectDiffuse_BlinnPhong( const in vec3 irradiance, const in GeometricContext geometry, const in BlinnPhongMaterial material, inout ReflectedLight reflectedLight ) {\n\treflectedLight.indirectDiffuse += irradiance * BRDF_Diffuse_Lambert( material.diffuseColor );\n}\n#define RE_Direct\t\t\t\tRE_Direct_BlinnPhong\n#define RE_IndirectDiffuse\t\tRE_IndirectDiffuse_BlinnPhong\n#define Material_LightProbeLOD( material )\t(0)\n";

var lights_physical_fragment = "PhysicalMaterial material;\nmaterial.diffuseColor = diffuseColor.rgb * ( 1.0 - metalnessFactor );\nmaterial.specularRoughness = clamp( roughnessFactor, 0.04, 1.0 );\n#ifdef STANDARD\n\tmaterial.specularColor = mix( vec3( DEFAULT_SPECULAR_COEFFICIENT ), diffuseColor.rgb, metalnessFactor );\n#else\n\tmaterial.specularColor = mix( vec3( MAXIMUM_SPECULAR_COEFFICIENT * pow2( reflectivity ) ), diffuseColor.rgb, metalnessFactor );\n\tmaterial.clearCoat = saturate( clearCoat );\tmaterial.clearCoatRoughness = clamp( clearCoatRoughness, 0.04, 1.0 );\n#endif\n";

var lights_physical_pars_fragment = "struct PhysicalMaterial {\n\tvec3\tdiffuseColor;\n\tfloat\tspecularRoughness;\n\tvec3\tspecularColor;\n\t#ifndef STANDARD\n\t\tfloat clearCoat;\n\t\tfloat clearCoatRoughness;\n\t#endif\n};\n#define MAXIMUM_SPECULAR_COEFFICIENT 0.16\n#define DEFAULT_SPECULAR_COEFFICIENT 0.04\nfloat clearCoatDHRApprox( const in float roughness, const in float dotNL ) {\n\treturn DEFAULT_SPECULAR_COEFFICIENT + ( 1.0 - DEFAULT_SPECULAR_COEFFICIENT ) * ( pow( 1.0 - dotNL, 5.0 ) * pow( 1.0 - roughness, 2.0 ) );\n}\nvoid RE_Direct_Physical( const in IncidentLight directLight, const in GeometricContext geometry, const in PhysicalMaterial material, inout ReflectedLight reflectedLight ) {\n\tfloat dotNL = saturate( dot( geometry.normal, directLight.direction ) );\n\tvec3 irradiance = dotNL * directLight.color;\n\t#ifndef PHYSICALLY_CORRECT_LIGHTS\n\t\tirradiance *= PI;\n\t#endif\n\t#ifndef STANDARD\n\t\tfloat clearCoatDHR = material.clearCoat * clearCoatDHRApprox( material.clearCoatRoughness, dotNL );\n\t#else\n\t\tfloat clearCoatDHR = 0.0;\n\t#endif\n\treflectedLight.directSpecular += ( 1.0 - clearCoatDHR ) * irradiance * BRDF_Specular_GGX( directLight, geometry, material.specularColor, material.specularRoughness );\n\treflectedLight.directDiffuse += ( 1.0 - clearCoatDHR ) * irradiance * BRDF_Diffuse_Lambert( material.diffuseColor );\n\t#ifndef STANDARD\n\t\treflectedLight.directSpecular += irradiance * material.clearCoat * BRDF_Specular_GGX( directLight, geometry, vec3( DEFAULT_SPECULAR_COEFFICIENT ), material.clearCoatRoughness );\n\t#endif\n}\nvoid RE_IndirectDiffuse_Physical( const in vec3 irradiance, const in GeometricContext geometry, const in PhysicalMaterial material, inout ReflectedLight reflectedLight ) {\n\treflectedLight.indirectDiffuse += irradiance * BRDF_Diffuse_Lambert( material.diffuseColor );\n}\nvoid RE_IndirectSpecular_Physical( const in vec3 radiance, const in vec3 clearCoatRadiance, const in GeometricContext geometry, const in PhysicalMaterial material, inout ReflectedLight reflectedLight ) {\n\t#ifndef STANDARD\n\t\tfloat dotNV = saturate( dot( geometry.normal, geometry.viewDir ) );\n\t\tfloat dotNL = dotNV;\n\t\tfloat clearCoatDHR = material.clearCoat * clearCoatDHRApprox( material.clearCoatRoughness, dotNL );\n\t#else\n\t\tfloat clearCoatDHR = 0.0;\n\t#endif\n\treflectedLight.indirectSpecular += ( 1.0 - clearCoatDHR ) * radiance * BRDF_Specular_GGX_Environment( geometry, material.specularColor, material.specularRoughness );\n\t#ifndef STANDARD\n\t\treflectedLight.indirectSpecular += clearCoatRadiance * material.clearCoat * BRDF_Specular_GGX_Environment( geometry, vec3( DEFAULT_SPECULAR_COEFFICIENT ), material.clearCoatRoughness );\n\t#endif\n}\n#define RE_Direct\t\t\t\tRE_Direct_Physical\n#define RE_IndirectDiffuse\t\tRE_IndirectDiffuse_Physical\n#define RE_IndirectSpecular\t\tRE_IndirectSpecular_Physical\n#define Material_BlinnShininessExponent( material )   GGXRoughnessToBlinnExponent( material.specularRoughness )\n#define Material_ClearCoat_BlinnShininessExponent( material )   GGXRoughnessToBlinnExponent( material.clearCoatRoughness )\nfloat computeSpecularOcclusion( const in float dotNV, const in float ambientOcclusion, const in float roughness ) {\n\treturn saturate( pow( dotNV + ambientOcclusion, exp2( - 16.0 * roughness - 1.0 ) ) - 1.0 + ambientOcclusion );\n}\n";

var lights_template = "\nGeometricContext geometry;\ngeometry.position = - vViewPosition;\ngeometry.normal = normal;\ngeometry.viewDir = normalize( vViewPosition );\nIncidentLight directLight;\n#if ( NUM_POINT_LIGHTS > 0 ) && defined( RE_Direct )\n\tPointLight pointLight;\n\tfor ( int i = 0; i < NUM_POINT_LIGHTS; i ++ ) {\n\t\tpointLight = pointLights[ i ];\n\t\tgetPointDirectLightIrradiance( pointLight, geometry, directLight );\n\t\t#ifdef USE_SHADOWMAP\n\t\tdirectLight.color *= all( bvec2( pointLight.shadow, directLight.visible ) ) ? getPointShadow( pointShadowMap[ i ], pointLight.shadowMapSize, pointLight.shadowBias, pointLight.shadowRadius, vPointShadowCoord[ i ] ) : 1.0;\n\t\t#endif\n\t\tRE_Direct( directLight, geometry, material, reflectedLight );\n\t}\n#endif\n#if ( NUM_SPOT_LIGHTS > 0 ) && defined( RE_Direct )\n\tSpotLight spotLight;\n\tfor ( int i = 0; i < NUM_SPOT_LIGHTS; i ++ ) {\n\t\tspotLight = spotLights[ i ];\n\t\tgetSpotDirectLightIrradiance( spotLight, geometry, directLight );\n\t\t#ifdef USE_SHADOWMAP\n\t\tdirectLight.color *= all( bvec2( spotLight.shadow, directLight.visible ) ) ? getShadow( spotShadowMap[ i ], spotLight.shadowMapSize, spotLight.shadowBias, spotLight.shadowRadius, vSpotShadowCoord[ i ] ) : 1.0;\n\t\t#endif\n\t\tRE_Direct( directLight, geometry, material, reflectedLight );\n\t}\n#endif\n#if ( NUM_DIR_LIGHTS > 0 ) && defined( RE_Direct )\n\tDirectionalLight directionalLight;\n\tfor ( int i = 0; i < NUM_DIR_LIGHTS; i ++ ) {\n\t\tdirectionalLight = directionalLights[ i ];\n\t\tgetDirectionalDirectLightIrradiance( directionalLight, geometry, directLight );\n\t\t#ifdef USE_SHADOWMAP\n\t\tdirectLight.color *= all( bvec2( directionalLight.shadow, directLight.visible ) ) ? getShadow( directionalShadowMap[ i ], directionalLight.shadowMapSize, directionalLight.shadowBias, directionalLight.shadowRadius, vDirectionalShadowCoord[ i ] ) : 1.0;\n\t\t#endif\n\t\tRE_Direct( directLight, geometry, material, reflectedLight );\n\t}\n#endif\n#if defined( RE_IndirectDiffuse )\n\tvec3 irradiance = getAmbientLightIrradiance( ambientLightColor );\n\t#ifdef USE_LIGHTMAP\n\t\tvec3 lightMapIrradiance = texture2D( lightMap, vUv2 ).xyz * lightMapIntensity;\n\t\t#ifndef PHYSICALLY_CORRECT_LIGHTS\n\t\t\tlightMapIrradiance *= PI;\n\t\t#endif\n\t\tirradiance += lightMapIrradiance;\n\t#endif\n\t#if ( NUM_HEMI_LIGHTS > 0 )\n\t\tfor ( int i = 0; i < NUM_HEMI_LIGHTS; i ++ ) {\n\t\t\tirradiance += getHemisphereLightIrradiance( hemisphereLights[ i ], geometry );\n\t\t}\n\t#endif\n\t#if defined( USE_ENVMAP ) && defined( PHYSICAL ) && defined( ENVMAP_TYPE_CUBE_UV )\n\t \tirradiance += getLightProbeIndirectIrradiance( geometry, 8 );\n\t#endif\n\tRE_IndirectDiffuse( irradiance, geometry, material, reflectedLight );\n#endif\n#if defined( USE_ENVMAP ) && defined( RE_IndirectSpecular )\n\tvec3 radiance = getLightProbeIndirectRadiance( geometry, Material_BlinnShininessExponent( material ), 8 );\n\t#ifndef STANDARD\n\t\tvec3 clearCoatRadiance = getLightProbeIndirectRadiance( geometry, Material_ClearCoat_BlinnShininessExponent( material ), 8 );\n\t#else\n\t\tvec3 clearCoatRadiance = vec3( 0.0 );\n\t#endif\n\t\t\n\tRE_IndirectSpecular( radiance, clearCoatRadiance, geometry, material, reflectedLight );\n#endif\n";

var logdepthbuf_fragment = "#if defined(USE_LOGDEPTHBUF) && defined(USE_LOGDEPTHBUF_EXT)\n\tgl_FragDepthEXT = log2(vFragDepth) * logDepthBufFC * 0.5;\n#endif";

var logdepthbuf_pars_fragment = "#ifdef USE_LOGDEPTHBUF\n\tuniform float logDepthBufFC;\n\t#ifdef USE_LOGDEPTHBUF_EXT\n\t\tvarying float vFragDepth;\n\t#endif\n#endif\n";

var logdepthbuf_pars_vertex = "#ifdef USE_LOGDEPTHBUF\n\t#ifdef USE_LOGDEPTHBUF_EXT\n\t\tvarying float vFragDepth;\n\t#endif\n\tuniform float logDepthBufFC;\n#endif";

var logdepthbuf_vertex = "#ifdef USE_LOGDEPTHBUF\n\tgl_Position.z = log2(max( EPSILON, gl_Position.w + 1.0 )) * logDepthBufFC;\n\t#ifdef USE_LOGDEPTHBUF_EXT\n\t\tvFragDepth = 1.0 + gl_Position.w;\n\t#else\n\t\tgl_Position.z = (gl_Position.z - 1.0) * gl_Position.w;\n\t#endif\n#endif\n";

var map_fragment = "#ifdef USE_MAP\n\tvec4 texelColor = texture2D( map, vUv );\n\ttexelColor = mapTexelToLinear( texelColor );\n\tdiffuseColor *= texelColor;\n#endif\n";

var map_pars_fragment = "#ifdef USE_MAP\n\tuniform sampler2D map;\n#endif\n";

var map_particle_fragment = "#ifdef USE_MAP\n\tvec4 mapTexel = texture2D( map, vec2( gl_PointCoord.x, 1.0 - gl_PointCoord.y ) * offsetRepeat.zw + offsetRepeat.xy );\n\tdiffuseColor *= mapTexelToLinear( mapTexel );\n#endif\n";

var map_particle_pars_fragment = "#ifdef USE_MAP\n\tuniform vec4 offsetRepeat;\n\tuniform sampler2D map;\n#endif\n";

var metalnessmap_fragment = "float metalnessFactor = metalness;\n#ifdef USE_METALNESSMAP\n\tvec4 texelMetalness = texture2D( metalnessMap, vUv );\n\tmetalnessFactor *= texelMetalness.r;\n#endif\n";

var metalnessmap_pars_fragment = "#ifdef USE_METALNESSMAP\n\tuniform sampler2D metalnessMap;\n#endif";

var morphnormal_vertex = "#ifdef USE_MORPHNORMALS\n\tobjectNormal += ( morphNormal0 - normal ) * morphTargetInfluences[ 0 ];\n\tobjectNormal += ( morphNormal1 - normal ) * morphTargetInfluences[ 1 ];\n\tobjectNormal += ( morphNormal2 - normal ) * morphTargetInfluences[ 2 ];\n\tobjectNormal += ( morphNormal3 - normal ) * morphTargetInfluences[ 3 ];\n#endif\n";

var morphtarget_pars_vertex = "#ifdef USE_MORPHTARGETS\n\t#ifndef USE_MORPHNORMALS\n\tuniform float morphTargetInfluences[ 8 ];\n\t#else\n\tuniform float morphTargetInfluences[ 4 ];\n\t#endif\n#endif";

var morphtarget_vertex = "#ifdef USE_MORPHTARGETS\n\ttransformed += ( morphTarget0 - position ) * morphTargetInfluences[ 0 ];\n\ttransformed += ( morphTarget1 - position ) * morphTargetInfluences[ 1 ];\n\ttransformed += ( morphTarget2 - position ) * morphTargetInfluences[ 2 ];\n\ttransformed += ( morphTarget3 - position ) * morphTargetInfluences[ 3 ];\n\t#ifndef USE_MORPHNORMALS\n\ttransformed += ( morphTarget4 - position ) * morphTargetInfluences[ 4 ];\n\ttransformed += ( morphTarget5 - position ) * morphTargetInfluences[ 5 ];\n\ttransformed += ( morphTarget6 - position ) * morphTargetInfluences[ 6 ];\n\ttransformed += ( morphTarget7 - position ) * morphTargetInfluences[ 7 ];\n\t#endif\n#endif\n";

var normal_flip = "#ifdef DOUBLE_SIDED\n\tfloat flipNormal = ( float( gl_FrontFacing ) * 2.0 - 1.0 );\n#else\n\tfloat flipNormal = 1.0;\n#endif\n";

var normal_fragment = "#ifdef FLAT_SHADED\n\tvec3 fdx = vec3( dFdx( vViewPosition.x ), dFdx( vViewPosition.y ), dFdx( vViewPosition.z ) );\n\tvec3 fdy = vec3( dFdy( vViewPosition.x ), dFdy( vViewPosition.y ), dFdy( vViewPosition.z ) );\n\tvec3 normal = normalize( cross( fdx, fdy ) );\n#else\n\tvec3 normal = normalize( vNormal ) * flipNormal;\n#endif\n#ifdef USE_NORMALMAP\n\tnormal = perturbNormal2Arb( -vViewPosition, normal );\n#elif defined( USE_BUMPMAP )\n\tnormal = perturbNormalArb( -vViewPosition, normal, dHdxy_fwd() );\n#endif\n";

var normalmap_pars_fragment = "#ifdef USE_NORMALMAP\n\tuniform sampler2D normalMap;\n\tuniform vec2 normalScale;\n\tvec3 perturbNormal2Arb( vec3 eye_pos, vec3 surf_norm ) {\n\t\tvec3 q0 = dFdx( eye_pos.xyz );\n\t\tvec3 q1 = dFdy( eye_pos.xyz );\n\t\tvec2 st0 = dFdx( vUv.st );\n\t\tvec2 st1 = dFdy( vUv.st );\n\t\tvec3 S = normalize( q0 * st1.t - q1 * st0.t );\n\t\tvec3 T = normalize( -q0 * st1.s + q1 * st0.s );\n\t\tvec3 N = normalize( surf_norm );\n\t\tvec3 mapN = texture2D( normalMap, vUv ).xyz * 2.0 - 1.0;\n\t\tmapN.xy = normalScale * mapN.xy;\n\t\tmat3 tsn = mat3( S, T, N );\n\t\treturn normalize( tsn * mapN );\n\t}\n#endif\n";

var packing = "vec3 packNormalToRGB( const in vec3 normal ) {\n  return normalize( normal ) * 0.5 + 0.5;\n}\nvec3 unpackRGBToNormal( const in vec3 rgb ) {\n  return 1.0 - 2.0 * rgb.xyz;\n}\nconst float PackUpscale = 256. / 255.;const float UnpackDownscale = 255. / 256.;\nconst vec3 PackFactors = vec3( 256. * 256. * 256., 256. * 256.,  256. );\nconst vec4 UnpackFactors = UnpackDownscale / vec4( PackFactors, 1. );\nconst float ShiftRight8 = 1. / 256.;\nvec4 packDepthToRGBA( const in float v ) {\n\tvec4 r = vec4( fract( v * PackFactors ), v );\n\tr.yzw -= r.xyz * ShiftRight8;\treturn r * PackUpscale;\n}\nfloat unpackRGBAToDepth( const in vec4 v ) {\n\treturn dot( v, UnpackFactors );\n}\nfloat viewZToOrthographicDepth( const in float viewZ, const in float near, const in float far ) {\n  return ( viewZ + near ) / ( near - far );\n}\nfloat orthographicDepthToViewZ( const in float linearClipZ, const in float near, const in float far ) {\n  return linearClipZ * ( near - far ) - near;\n}\nfloat viewZToPerspectiveDepth( const in float viewZ, const in float near, const in float far ) {\n  return (( near + viewZ ) * far ) / (( far - near ) * viewZ );\n}\nfloat perspectiveDepthToViewZ( const in float invClipZ, const in float near, const in float far ) {\n  return ( near * far ) / ( ( far - near ) * invClipZ - far );\n}\n";

var premultiplied_alpha_fragment = "#ifdef PREMULTIPLIED_ALPHA\n\tgl_FragColor.rgb *= gl_FragColor.a;\n#endif\n";

var project_vertex = "#ifdef USE_SKINNING\n\tvec4 mvPosition = modelViewMatrix * skinned;\n#else\n\tvec4 mvPosition = modelViewMatrix * vec4( transformed, 1.0 );\n#endif\ngl_Position = projectionMatrix * mvPosition;\n";

var roughnessmap_fragment = "float roughnessFactor = roughness;\n#ifdef USE_ROUGHNESSMAP\n\tvec4 texelRoughness = texture2D( roughnessMap, vUv );\n\troughnessFactor *= texelRoughness.r;\n#endif\n";

var roughnessmap_pars_fragment = "#ifdef USE_ROUGHNESSMAP\n\tuniform sampler2D roughnessMap;\n#endif";

var shadowmap_pars_fragment = "#ifdef USE_SHADOWMAP\n\t#if NUM_DIR_LIGHTS > 0\n\t\tuniform sampler2D directionalShadowMap[ NUM_DIR_LIGHTS ];\n\t\tvarying vec4 vDirectionalShadowCoord[ NUM_DIR_LIGHTS ];\n\t#endif\n\t#if NUM_SPOT_LIGHTS > 0\n\t\tuniform sampler2D spotShadowMap[ NUM_SPOT_LIGHTS ];\n\t\tvarying vec4 vSpotShadowCoord[ NUM_SPOT_LIGHTS ];\n\t#endif\n\t#if NUM_POINT_LIGHTS > 0\n\t\tuniform sampler2D pointShadowMap[ NUM_POINT_LIGHTS ];\n\t\tvarying vec4 vPointShadowCoord[ NUM_POINT_LIGHTS ];\n\t#endif\n\tfloat texture2DCompare( sampler2D depths, vec2 uv, float compare ) {\n\t\treturn step( compare, unpackRGBAToDepth( texture2D( depths, uv ) ) );\n\t}\n\tfloat texture2DShadowLerp( sampler2D depths, vec2 size, vec2 uv, float compare ) {\n\t\tconst vec2 offset = vec2( 0.0, 1.0 );\n\t\tvec2 texelSize = vec2( 1.0 ) / size;\n\t\tvec2 centroidUV = floor( uv * size + 0.5 ) / size;\n\t\tfloat lb = texture2DCompare( depths, centroidUV + texelSize * offset.xx, compare );\n\t\tfloat lt = texture2DCompare( depths, centroidUV + texelSize * offset.xy, compare );\n\t\tfloat rb = texture2DCompare( depths, centroidUV + texelSize * offset.yx, compare );\n\t\tfloat rt = texture2DCompare( depths, centroidUV + texelSize * offset.yy, compare );\n\t\tvec2 f = fract( uv * size + 0.5 );\n\t\tfloat a = mix( lb, lt, f.y );\n\t\tfloat b = mix( rb, rt, f.y );\n\t\tfloat c = mix( a, b, f.x );\n\t\treturn c;\n\t}\n\tfloat getShadow( sampler2D shadowMap, vec2 shadowMapSize, float shadowBias, float shadowRadius, vec4 shadowCoord ) {\n\t\tshadowCoord.xyz /= shadowCoord.w;\n\t\tshadowCoord.z += shadowBias;\n\t\tbvec4 inFrustumVec = bvec4 ( shadowCoord.x >= 0.0, shadowCoord.x <= 1.0, shadowCoord.y >= 0.0, shadowCoord.y <= 1.0 );\n\t\tbool inFrustum = all( inFrustumVec );\n\t\tbvec2 frustumTestVec = bvec2( inFrustum, shadowCoord.z <= 1.0 );\n\t\tbool frustumTest = all( frustumTestVec );\n\t\tif ( frustumTest ) {\n\t\t#if defined( SHADOWMAP_TYPE_PCF )\n\t\t\tvec2 texelSize = vec2( 1.0 ) / shadowMapSize;\n\t\t\tfloat dx0 = - texelSize.x * shadowRadius;\n\t\t\tfloat dy0 = - texelSize.y * shadowRadius;\n\t\t\tfloat dx1 = + texelSize.x * shadowRadius;\n\t\t\tfloat dy1 = + texelSize.y * shadowRadius;\n\t\t\treturn (\n\t\t\t\ttexture2DCompare( shadowMap, shadowCoord.xy + vec2( dx0, dy0 ), shadowCoord.z ) +\n\t\t\t\ttexture2DCompare( shadowMap, shadowCoord.xy + vec2( 0.0, dy0 ), shadowCoord.z ) +\n\t\t\t\ttexture2DCompare( shadowMap, shadowCoord.xy + vec2( dx1, dy0 ), shadowCoord.z ) +\n\t\t\t\ttexture2DCompare( shadowMap, shadowCoord.xy + vec2( dx0, 0.0 ), shadowCoord.z ) +\n\t\t\t\ttexture2DCompare( shadowMap, shadowCoord.xy, shadowCoord.z ) +\n\t\t\t\ttexture2DCompare( shadowMap, shadowCoord.xy + vec2( dx1, 0.0 ), shadowCoord.z ) +\n\t\t\t\ttexture2DCompare( shadowMap, shadowCoord.xy + vec2( dx0, dy1 ), shadowCoord.z ) +\n\t\t\t\ttexture2DCompare( shadowMap, shadowCoord.xy + vec2( 0.0, dy1 ), shadowCoord.z ) +\n\t\t\t\ttexture2DCompare( shadowMap, shadowCoord.xy + vec2( dx1, dy1 ), shadowCoord.z )\n\t\t\t) * ( 1.0 / 9.0 );\n\t\t#elif defined( SHADOWMAP_TYPE_PCF_SOFT )\n\t\t\tvec2 texelSize = vec2( 1.0 ) / shadowMapSize;\n\t\t\tfloat dx0 = - texelSize.x * shadowRadius;\n\t\t\tfloat dy0 = - texelSize.y * shadowRadius;\n\t\t\tfloat dx1 = + texelSize.x * shadowRadius;\n\t\t\tfloat dy1 = + texelSize.y * shadowRadius;\n\t\t\treturn (\n\t\t\t\ttexture2DShadowLerp( shadowMap, shadowMapSize, shadowCoord.xy + vec2( dx0, dy0 ), shadowCoord.z ) +\n\t\t\t\ttexture2DShadowLerp( shadowMap, shadowMapSize, shadowCoord.xy + vec2( 0.0, dy0 ), shadowCoord.z ) +\n\t\t\t\ttexture2DShadowLerp( shadowMap, shadowMapSize, shadowCoord.xy + vec2( dx1, dy0 ), shadowCoord.z ) +\n\t\t\t\ttexture2DShadowLerp( shadowMap, shadowMapSize, shadowCoord.xy + vec2( dx0, 0.0 ), shadowCoord.z ) +\n\t\t\t\ttexture2DShadowLerp( shadowMap, shadowMapSize, shadowCoord.xy, shadowCoord.z ) +\n\t\t\t\ttexture2DShadowLerp( shadowMap, shadowMapSize, shadowCoord.xy + vec2( dx1, 0.0 ), shadowCoord.z ) +\n\t\t\t\ttexture2DShadowLerp( shadowMap, shadowMapSize, shadowCoord.xy + vec2( dx0, dy1 ), shadowCoord.z ) +\n\t\t\t\ttexture2DShadowLerp( shadowMap, shadowMapSize, shadowCoord.xy + vec2( 0.0, dy1 ), shadowCoord.z ) +\n\t\t\t\ttexture2DShadowLerp( shadowMap, shadowMapSize, shadowCoord.xy + vec2( dx1, dy1 ), shadowCoord.z )\n\t\t\t) * ( 1.0 / 9.0 );\n\t\t#else\n\t\t\treturn texture2DCompare( shadowMap, shadowCoord.xy, shadowCoord.z );\n\t\t#endif\n\t\t}\n\t\treturn 1.0;\n\t}\n\tvec2 cubeToUV( vec3 v, float texelSizeY ) {\n\t\tvec3 absV = abs( v );\n\t\tfloat scaleToCube = 1.0 / max( absV.x, max( absV.y, absV.z ) );\n\t\tabsV *= scaleToCube;\n\t\tv *= scaleToCube * ( 1.0 - 2.0 * texelSizeY );\n\t\tvec2 planar = v.xy;\n\t\tfloat almostATexel = 1.5 * texelSizeY;\n\t\tfloat almostOne = 1.0 - almostATexel;\n\t\tif ( absV.z >= almostOne ) {\n\t\t\tif ( v.z > 0.0 )\n\t\t\t\tplanar.x = 4.0 - v.x;\n\t\t} else if ( absV.x >= almostOne ) {\n\t\t\tfloat signX = sign( v.x );\n\t\t\tplanar.x = v.z * signX + 2.0 * signX;\n\t\t} else if ( absV.y >= almostOne ) {\n\t\t\tfloat signY = sign( v.y );\n\t\t\tplanar.x = v.x + 2.0 * signY + 2.0;\n\t\t\tplanar.y = v.z * signY - 2.0;\n\t\t}\n\t\treturn vec2( 0.125, 0.25 ) * planar + vec2( 0.375, 0.75 );\n\t}\n\tfloat getPointShadow( sampler2D shadowMap, vec2 shadowMapSize, float shadowBias, float shadowRadius, vec4 shadowCoord ) {\n\t\tvec2 texelSize = vec2( 1.0 ) / ( shadowMapSize * vec2( 4.0, 2.0 ) );\n\t\tvec3 lightToPosition = shadowCoord.xyz;\n\t\tvec3 bd3D = normalize( lightToPosition );\n\t\tfloat dp = ( length( lightToPosition ) - shadowBias ) / 1000.0;\n\t\t#if defined( SHADOWMAP_TYPE_PCF ) || defined( SHADOWMAP_TYPE_PCF_SOFT )\n\t\t\tvec2 offset = vec2( - 1, 1 ) * shadowRadius * texelSize.y;\n\t\t\treturn (\n\t\t\t\ttexture2DCompare( shadowMap, cubeToUV( bd3D + offset.xyy, texelSize.y ), dp ) +\n\t\t\t\ttexture2DCompare( shadowMap, cubeToUV( bd3D + offset.yyy, texelSize.y ), dp ) +\n\t\t\t\ttexture2DCompare( shadowMap, cubeToUV( bd3D + offset.xyx, texelSize.y ), dp ) +\n\t\t\t\ttexture2DCompare( shadowMap, cubeToUV( bd3D + offset.yyx, texelSize.y ), dp ) +\n\t\t\t\ttexture2DCompare( shadowMap, cubeToUV( bd3D, texelSize.y ), dp ) +\n\t\t\t\ttexture2DCompare( shadowMap, cubeToUV( bd3D + offset.xxy, texelSize.y ), dp ) +\n\t\t\t\ttexture2DCompare( shadowMap, cubeToUV( bd3D + offset.yxy, texelSize.y ), dp ) +\n\t\t\t\ttexture2DCompare( shadowMap, cubeToUV( bd3D + offset.xxx, texelSize.y ), dp ) +\n\t\t\t\ttexture2DCompare( shadowMap, cubeToUV( bd3D + offset.yxx, texelSize.y ), dp )\n\t\t\t) * ( 1.0 / 9.0 );\n\t\t#else\n\t\t\treturn texture2DCompare( shadowMap, cubeToUV( bd3D, texelSize.y ), dp );\n\t\t#endif\n\t}\n#endif\n";

var shadowmap_pars_vertex = "#ifdef USE_SHADOWMAP\n\t#if NUM_DIR_LIGHTS > 0\n\t\tuniform mat4 directionalShadowMatrix[ NUM_DIR_LIGHTS ];\n\t\tvarying vec4 vDirectionalShadowCoord[ NUM_DIR_LIGHTS ];\n\t#endif\n\t#if NUM_SPOT_LIGHTS > 0\n\t\tuniform mat4 spotShadowMatrix[ NUM_SPOT_LIGHTS ];\n\t\tvarying vec4 vSpotShadowCoord[ NUM_SPOT_LIGHTS ];\n\t#endif\n\t#if NUM_POINT_LIGHTS > 0\n\t\tuniform mat4 pointShadowMatrix[ NUM_POINT_LIGHTS ];\n\t\tvarying vec4 vPointShadowCoord[ NUM_POINT_LIGHTS ];\n\t#endif\n#endif\n";

var shadowmap_vertex = "#ifdef USE_SHADOWMAP\n\t#if NUM_DIR_LIGHTS > 0\n\tfor ( int i = 0; i < NUM_DIR_LIGHTS; i ++ ) {\n\t\tvDirectionalShadowCoord[ i ] = directionalShadowMatrix[ i ] * worldPosition;\n\t}\n\t#endif\n\t#if NUM_SPOT_LIGHTS > 0\n\tfor ( int i = 0; i < NUM_SPOT_LIGHTS; i ++ ) {\n\t\tvSpotShadowCoord[ i ] = spotShadowMatrix[ i ] * worldPosition;\n\t}\n\t#endif\n\t#if NUM_POINT_LIGHTS > 0\n\tfor ( int i = 0; i < NUM_POINT_LIGHTS; i ++ ) {\n\t\tvPointShadowCoord[ i ] = pointShadowMatrix[ i ] * worldPosition;\n\t}\n\t#endif\n#endif\n";

var shadowmask_pars_fragment = "float getShadowMask() {\n\tfloat shadow = 1.0;\n\t#ifdef USE_SHADOWMAP\n\t#if NUM_DIR_LIGHTS > 0\n\tDirectionalLight directionalLight;\n\tfor ( int i = 0; i < NUM_DIR_LIGHTS; i ++ ) {\n\t\tdirectionalLight = directionalLights[ i ];\n\t\tshadow *= bool( directionalLight.shadow ) ? getShadow( directionalShadowMap[ i ], directionalLight.shadowMapSize, directionalLight.shadowBias, directionalLight.shadowRadius, vDirectionalShadowCoord[ i ] ) : 1.0;\n\t}\n\t#endif\n\t#if NUM_SPOT_LIGHTS > 0\n\tSpotLight spotLight;\n\tfor ( int i = 0; i < NUM_SPOT_LIGHTS; i ++ ) {\n\t\tspotLight = spotLights[ i ];\n\t\tshadow *= bool( spotLight.shadow ) ? getShadow( spotShadowMap[ i ], spotLight.shadowMapSize, spotLight.shadowBias, spotLight.shadowRadius, vSpotShadowCoord[ i ] ) : 1.0;\n\t}\n\t#endif\n\t#if NUM_POINT_LIGHTS > 0\n\tPointLight pointLight;\n\tfor ( int i = 0; i < NUM_POINT_LIGHTS; i ++ ) {\n\t\tpointLight = pointLights[ i ];\n\t\tshadow *= bool( pointLight.shadow ) ? getPointShadow( pointShadowMap[ i ], pointLight.shadowMapSize, pointLight.shadowBias, pointLight.shadowRadius, vPointShadowCoord[ i ] ) : 1.0;\n\t}\n\t#endif\n\t#endif\n\treturn shadow;\n}\n";

var skinbase_vertex = "#ifdef USE_SKINNING\n\tmat4 boneMatX = getBoneMatrix( skinIndex.x );\n\tmat4 boneMatY = getBoneMatrix( skinIndex.y );\n\tmat4 boneMatZ = getBoneMatrix( skinIndex.z );\n\tmat4 boneMatW = getBoneMatrix( skinIndex.w );\n#endif";

var skinning_pars_vertex = "#ifdef USE_SKINNING\n\tuniform mat4 bindMatrix;\n\tuniform mat4 bindMatrixInverse;\n\t#ifdef BONE_TEXTURE\n\t\tuniform sampler2D boneTexture;\n\t\tuniform int boneTextureWidth;\n\t\tuniform int boneTextureHeight;\n\t\tmat4 getBoneMatrix( const in float i ) {\n\t\t\tfloat j = i * 4.0;\n\t\t\tfloat x = mod( j, float( boneTextureWidth ) );\n\t\t\tfloat y = floor( j / float( boneTextureWidth ) );\n\t\t\tfloat dx = 1.0 / float( boneTextureWidth );\n\t\t\tfloat dy = 1.0 / float( boneTextureHeight );\n\t\t\ty = dy * ( y + 0.5 );\n\t\t\tvec4 v1 = texture2D( boneTexture, vec2( dx * ( x + 0.5 ), y ) );\n\t\t\tvec4 v2 = texture2D( boneTexture, vec2( dx * ( x + 1.5 ), y ) );\n\t\t\tvec4 v3 = texture2D( boneTexture, vec2( dx * ( x + 2.5 ), y ) );\n\t\t\tvec4 v4 = texture2D( boneTexture, vec2( dx * ( x + 3.5 ), y ) );\n\t\t\tmat4 bone = mat4( v1, v2, v3, v4 );\n\t\t\treturn bone;\n\t\t}\n\t#else\n\t\tuniform mat4 boneMatrices[ MAX_BONES ];\n\t\tmat4 getBoneMatrix( const in float i ) {\n\t\t\tmat4 bone = boneMatrices[ int(i) ];\n\t\t\treturn bone;\n\t\t}\n\t#endif\n#endif\n";

var skinning_vertex = "#ifdef USE_SKINNING\n\tvec4 skinVertex = bindMatrix * vec4( transformed, 1.0 );\n\tvec4 skinned = vec4( 0.0 );\n\tskinned += boneMatX * skinVertex * skinWeight.x;\n\tskinned += boneMatY * skinVertex * skinWeight.y;\n\tskinned += boneMatZ * skinVertex * skinWeight.z;\n\tskinned += boneMatW * skinVertex * skinWeight.w;\n\tskinned  = bindMatrixInverse * skinned;\n#endif\n";

var skinnormal_vertex = "#ifdef USE_SKINNING\n\tmat4 skinMatrix = mat4( 0.0 );\n\tskinMatrix += skinWeight.x * boneMatX;\n\tskinMatrix += skinWeight.y * boneMatY;\n\tskinMatrix += skinWeight.z * boneMatZ;\n\tskinMatrix += skinWeight.w * boneMatW;\n\tskinMatrix  = bindMatrixInverse * skinMatrix * bindMatrix;\n\tobjectNormal = vec4( skinMatrix * vec4( objectNormal, 0.0 ) ).xyz;\n#endif\n";

var specularmap_fragment = "float specularStrength;\n#ifdef USE_SPECULARMAP\n\tvec4 texelSpecular = texture2D( specularMap, vUv );\n\tspecularStrength = texelSpecular.r;\n#else\n\tspecularStrength = 1.0;\n#endif";

var specularmap_pars_fragment = "#ifdef USE_SPECULARMAP\n\tuniform sampler2D specularMap;\n#endif";

var tonemapping_fragment = "#if defined( TONE_MAPPING )\n  gl_FragColor.rgb = toneMapping( gl_FragColor.rgb );\n#endif\n";

var tonemapping_pars_fragment = "#define saturate(a) clamp( a, 0.0, 1.0 )\nuniform float toneMappingExposure;\nuniform float toneMappingWhitePoint;\nvec3 LinearToneMapping( vec3 color ) {\n  return toneMappingExposure * color;\n}\nvec3 ReinhardToneMapping( vec3 color ) {\n  color *= toneMappingExposure;\n  return saturate( color / ( vec3( 1.0 ) + color ) );\n}\n#define Uncharted2Helper( x ) max( ( ( x * ( 0.15 * x + 0.10 * 0.50 ) + 0.20 * 0.02 ) / ( x * ( 0.15 * x + 0.50 ) + 0.20 * 0.30 ) ) - 0.02 / 0.30, vec3( 0.0 ) )\nvec3 Uncharted2ToneMapping( vec3 color ) {\n  color *= toneMappingExposure;\n  return saturate( Uncharted2Helper( color ) / Uncharted2Helper( vec3( toneMappingWhitePoint ) ) );\n}\nvec3 OptimizedCineonToneMapping( vec3 color ) {\n  color *= toneMappingExposure;\n  color = max( vec3( 0.0 ), color - 0.004 );\n  return pow( ( color * ( 6.2 * color + 0.5 ) ) / ( color * ( 6.2 * color + 1.7 ) + 0.06 ), vec3( 2.2 ) );\n}\n";

var uv_pars_fragment = "#if defined( USE_MAP ) || defined( USE_BUMPMAP ) || defined( USE_NORMALMAP ) || defined( USE_SPECULARMAP ) || defined( USE_ALPHAMAP ) || defined( USE_EMISSIVEMAP ) || defined( USE_ROUGHNESSMAP ) || defined( USE_METALNESSMAP )\n\tvarying vec2 vUv;\n#endif";

var uv_pars_vertex = "#if defined( USE_MAP ) || defined( USE_BUMPMAP ) || defined( USE_NORMALMAP ) || defined( USE_SPECULARMAP ) || defined( USE_ALPHAMAP ) || defined( USE_EMISSIVEMAP ) || defined( USE_ROUGHNESSMAP ) || defined( USE_METALNESSMAP )\n\tvarying vec2 vUv;\n\tuniform vec4 offsetRepeat;\n#endif\n";

var uv_vertex = "#if defined( USE_MAP ) || defined( USE_BUMPMAP ) || defined( USE_NORMALMAP ) || defined( USE_SPECULARMAP ) || defined( USE_ALPHAMAP ) || defined( USE_EMISSIVEMAP ) || defined( USE_ROUGHNESSMAP ) || defined( USE_METALNESSMAP )\n\tvUv = uv * offsetRepeat.zw + offsetRepeat.xy;\n#endif";

var uv2_pars_fragment = "#if defined( USE_LIGHTMAP ) || defined( USE_AOMAP )\n\tvarying vec2 vUv2;\n#endif";

var uv2_pars_vertex = "#if defined( USE_LIGHTMAP ) || defined( USE_AOMAP )\n\tattribute vec2 uv2;\n\tvarying vec2 vUv2;\n#endif";

var uv2_vertex = "#if defined( USE_LIGHTMAP ) || defined( USE_AOMAP )\n\tvUv2 = uv2;\n#endif";

var worldpos_vertex = "#if defined( USE_ENVMAP ) || defined( PHONG ) || defined( PHYSICAL ) || defined( LAMBERT ) || defined ( USE_SHADOWMAP )\n\t#ifdef USE_SKINNING\n\t\tvec4 worldPosition = modelMatrix * skinned;\n\t#else\n\t\tvec4 worldPosition = modelMatrix * vec4( transformed, 1.0 );\n\t#endif\n#endif\n";

var cube_frag = "uniform samplerCube tCube;\nuniform float tFlip;\nuniform float opacity;\nvarying vec3 vWorldPosition;\n#include <common>\nvoid main() {\n\tgl_FragColor = textureCube( tCube, vec3( tFlip * vWorldPosition.x, vWorldPosition.yz ) );\n\tgl_FragColor.a *= opacity;\n}\n";

var cube_vert = "varying vec3 vWorldPosition;\n#include <common>\nvoid main() {\n\tvWorldPosition = transformDirection( position, modelMatrix );\n\t#include <begin_vertex>\n\t#include <project_vertex>\n}\n";

var depth_frag = "#if DEPTH_PACKING == 3200\n\tuniform float opacity;\n#endif\n#include <common>\n#include <packing>\n#include <uv_pars_fragment>\n#include <map_pars_fragment>\n#include <alphamap_pars_fragment>\n#include <logdepthbuf_pars_fragment>\n#include <clipping_planes_pars_fragment>\nvoid main() {\n\t#include <clipping_planes_fragment>\n\tvec4 diffuseColor = vec4( 1.0 );\n\t#if DEPTH_PACKING == 3200\n\t\tdiffuseColor.a = opacity;\n\t#endif\n\t#include <map_fragment>\n\t#include <alphamap_fragment>\n\t#include <alphatest_fragment>\n\t#include <logdepthbuf_fragment>\n\t#if DEPTH_PACKING == 3200\n\t\tgl_FragColor = vec4( vec3( gl_FragCoord.z ), opacity );\n\t#elif DEPTH_PACKING == 3201\n\t\tgl_FragColor = packDepthToRGBA( gl_FragCoord.z );\n\t#endif\n}\n";

var depth_vert = "#include <common>\n#include <uv_pars_vertex>\n#include <displacementmap_pars_vertex>\n#include <morphtarget_pars_vertex>\n#include <skinning_pars_vertex>\n#include <logdepthbuf_pars_vertex>\n#include <clipping_planes_pars_vertex>\nvoid main() {\n\t#include <uv_vertex>\n\t#include <skinbase_vertex>\n\t#include <begin_vertex>\n\t#include <displacementmap_vertex>\n\t#include <morphtarget_vertex>\n\t#include <skinning_vertex>\n\t#include <project_vertex>\n\t#include <logdepthbuf_vertex>\n\t#include <clipping_planes_vertex>\n}\n";

var distanceRGBA_frag = "uniform vec3 lightPos;\nvarying vec4 vWorldPosition;\n#include <common>\n#include <packing>\n#include <clipping_planes_pars_fragment>\nvoid main () {\n\t#include <clipping_planes_fragment>\n\tgl_FragColor = packDepthToRGBA( length( vWorldPosition.xyz - lightPos.xyz ) / 1000.0 );\n}\n";

var distanceRGBA_vert = "varying vec4 vWorldPosition;\n#include <common>\n#include <morphtarget_pars_vertex>\n#include <skinning_pars_vertex>\n#include <clipping_planes_pars_vertex>\nvoid main() {\n\t#include <skinbase_vertex>\n\t#include <begin_vertex>\n\t#include <morphtarget_vertex>\n\t#include <skinning_vertex>\n\t#include <project_vertex>\n\t#include <worldpos_vertex>\n\t#include <clipping_planes_vertex>\n\tvWorldPosition = worldPosition;\n}\n";

var equirect_frag = "uniform sampler2D tEquirect;\nuniform float tFlip;\nvarying vec3 vWorldPosition;\n#include <common>\nvoid main() {\n\tvec3 direction = normalize( vWorldPosition );\n\tvec2 sampleUV;\n\tsampleUV.y = saturate( tFlip * direction.y * -0.5 + 0.5 );\n\tsampleUV.x = atan( direction.z, direction.x ) * RECIPROCAL_PI2 + 0.5;\n\tgl_FragColor = texture2D( tEquirect, sampleUV );\n}\n";

var equirect_vert = "varying vec3 vWorldPosition;\n#include <common>\nvoid main() {\n\tvWorldPosition = transformDirection( position, modelMatrix );\n\t#include <begin_vertex>\n\t#include <project_vertex>\n}\n";

var linedashed_frag = "uniform vec3 diffuse;\nuniform float opacity;\nuniform float dashSize;\nuniform float totalSize;\nvarying float vLineDistance;\n#include <common>\n#include <color_pars_fragment>\n#include <fog_pars_fragment>\n#include <logdepthbuf_pars_fragment>\n#include <clipping_planes_pars_fragment>\nvoid main() {\n\t#include <clipping_planes_fragment>\n\tif ( mod( vLineDistance, totalSize ) > dashSize ) {\n\t\tdiscard;\n\t}\n\tvec3 outgoingLight = vec3( 0.0 );\n\tvec4 diffuseColor = vec4( diffuse, opacity );\n\t#include <logdepthbuf_fragment>\n\t#include <color_fragment>\n\toutgoingLight = diffuseColor.rgb;\n\tgl_FragColor = vec4( outgoingLight, diffuseColor.a );\n\t#include <premultiplied_alpha_fragment>\n\t#include <tonemapping_fragment>\n\t#include <encodings_fragment>\n\t#include <fog_fragment>\n}\n";

var linedashed_vert = "uniform float scale;\nattribute float lineDistance;\nvarying float vLineDistance;\n#include <common>\n#include <color_pars_vertex>\n#include <logdepthbuf_pars_vertex>\n#include <clipping_planes_pars_vertex>\nvoid main() {\n\t#include <color_vertex>\n\tvLineDistance = scale * lineDistance;\n\tvec4 mvPosition = modelViewMatrix * vec4( position, 1.0 );\n\tgl_Position = projectionMatrix * mvPosition;\n\t#include <logdepthbuf_vertex>\n\t#include <clipping_planes_vertex>\n}\n";

var meshbasic_frag = "uniform vec3 diffuse;\nuniform float opacity;\n#ifndef FLAT_SHADED\n\tvarying vec3 vNormal;\n#endif\n#include <common>\n#include <color_pars_fragment>\n#include <uv_pars_fragment>\n#include <uv2_pars_fragment>\n#include <map_pars_fragment>\n#include <alphamap_pars_fragment>\n#include <aomap_pars_fragment>\n#include <envmap_pars_fragment>\n#include <fog_pars_fragment>\n#include <specularmap_pars_fragment>\n#include <logdepthbuf_pars_fragment>\n#include <clipping_planes_pars_fragment>\nvoid main() {\n\t#include <clipping_planes_fragment>\n\tvec4 diffuseColor = vec4( diffuse, opacity );\n\t#include <logdepthbuf_fragment>\n\t#include <map_fragment>\n\t#include <color_fragment>\n\t#include <alphamap_fragment>\n\t#include <alphatest_fragment>\n\t#include <specularmap_fragment>\n\tReflectedLight reflectedLight;\n\treflectedLight.directDiffuse = vec3( 0.0 );\n\treflectedLight.directSpecular = vec3( 0.0 );\n\treflectedLight.indirectDiffuse = diffuseColor.rgb;\n\treflectedLight.indirectSpecular = vec3( 0.0 );\n\t#include <aomap_fragment>\n\tvec3 outgoingLight = reflectedLight.indirectDiffuse;\n\t#include <normal_flip>\n\t#include <envmap_fragment>\n\tgl_FragColor = vec4( outgoingLight, diffuseColor.a );\n\t#include <premultiplied_alpha_fragment>\n\t#include <tonemapping_fragment>\n\t#include <encodings_fragment>\n\t#include <fog_fragment>\n}\n";

var meshbasic_vert = "#include <common>\n#include <uv_pars_vertex>\n#include <uv2_pars_vertex>\n#include <envmap_pars_vertex>\n#include <color_pars_vertex>\n#include <morphtarget_pars_vertex>\n#include <skinning_pars_vertex>\n#include <logdepthbuf_pars_vertex>\n#include <clipping_planes_pars_vertex>\nvoid main() {\n\t#include <uv_vertex>\n\t#include <uv2_vertex>\n\t#include <color_vertex>\n\t#include <skinbase_vertex>\n\t#ifdef USE_ENVMAP\n\t#include <beginnormal_vertex>\n\t#include <morphnormal_vertex>\n\t#include <skinnormal_vertex>\n\t#include <defaultnormal_vertex>\n\t#endif\n\t#include <begin_vertex>\n\t#include <morphtarget_vertex>\n\t#include <skinning_vertex>\n\t#include <project_vertex>\n\t#include <logdepthbuf_vertex>\n\t#include <worldpos_vertex>\n\t#include <clipping_planes_vertex>\n\t#include <envmap_vertex>\n}\n";

var meshlambert_frag = "uniform vec3 diffuse;\nuniform vec3 emissive;\nuniform float opacity;\nvarying vec3 vLightFront;\n#ifdef DOUBLE_SIDED\n\tvarying vec3 vLightBack;\n#endif\n#include <common>\n#include <packing>\n#include <color_pars_fragment>\n#include <uv_pars_fragment>\n#include <uv2_pars_fragment>\n#include <map_pars_fragment>\n#include <alphamap_pars_fragment>\n#include <aomap_pars_fragment>\n#include <lightmap_pars_fragment>\n#include <emissivemap_pars_fragment>\n#include <envmap_pars_fragment>\n#include <bsdfs>\n#include <lights_pars>\n#include <fog_pars_fragment>\n#include <shadowmap_pars_fragment>\n#include <shadowmask_pars_fragment>\n#include <specularmap_pars_fragment>\n#include <logdepthbuf_pars_fragment>\n#include <clipping_planes_pars_fragment>\nvoid main() {\n\t#include <clipping_planes_fragment>\n\tvec4 diffuseColor = vec4( diffuse, opacity );\n\tReflectedLight reflectedLight = ReflectedLight( vec3( 0.0 ), vec3( 0.0 ), vec3( 0.0 ), vec3( 0.0 ) );\n\tvec3 totalEmissiveRadiance = emissive;\n\t#include <logdepthbuf_fragment>\n\t#include <map_fragment>\n\t#include <color_fragment>\n\t#include <alphamap_fragment>\n\t#include <alphatest_fragment>\n\t#include <specularmap_fragment>\n\t#include <emissivemap_fragment>\n\treflectedLight.indirectDiffuse = getAmbientLightIrradiance( ambientLightColor );\n\t#include <lightmap_fragment>\n\treflectedLight.indirectDiffuse *= BRDF_Diffuse_Lambert( diffuseColor.rgb );\n\t#ifdef DOUBLE_SIDED\n\t\treflectedLight.directDiffuse = ( gl_FrontFacing ) ? vLightFront : vLightBack;\n\t#else\n\t\treflectedLight.directDiffuse = vLightFront;\n\t#endif\n\treflectedLight.directDiffuse *= BRDF_Diffuse_Lambert( diffuseColor.rgb ) * getShadowMask();\n\t#include <aomap_fragment>\n\tvec3 outgoingLight = reflectedLight.directDiffuse + reflectedLight.indirectDiffuse + totalEmissiveRadiance;\n\t#include <normal_flip>\n\t#include <envmap_fragment>\n\tgl_FragColor = vec4( outgoingLight, diffuseColor.a );\n\t#include <premultiplied_alpha_fragment>\n\t#include <tonemapping_fragment>\n\t#include <encodings_fragment>\n\t#include <fog_fragment>\n}\n";

var meshlambert_vert = "#define LAMBERT\nvarying vec3 vLightFront;\n#ifdef DOUBLE_SIDED\n\tvarying vec3 vLightBack;\n#endif\n#include <common>\n#include <uv_pars_vertex>\n#include <uv2_pars_vertex>\n#include <envmap_pars_vertex>\n#include <bsdfs>\n#include <lights_pars>\n#include <color_pars_vertex>\n#include <morphtarget_pars_vertex>\n#include <skinning_pars_vertex>\n#include <shadowmap_pars_vertex>\n#include <logdepthbuf_pars_vertex>\n#include <clipping_planes_pars_vertex>\nvoid main() {\n\t#include <uv_vertex>\n\t#include <uv2_vertex>\n\t#include <color_vertex>\n\t#include <beginnormal_vertex>\n\t#include <morphnormal_vertex>\n\t#include <skinbase_vertex>\n\t#include <skinnormal_vertex>\n\t#include <defaultnormal_vertex>\n\t#include <begin_vertex>\n\t#include <morphtarget_vertex>\n\t#include <skinning_vertex>\n\t#include <project_vertex>\n\t#include <logdepthbuf_vertex>\n\t#include <clipping_planes_vertex>\n\t#include <worldpos_vertex>\n\t#include <envmap_vertex>\n\t#include <lights_lambert_vertex>\n\t#include <shadowmap_vertex>\n}\n";

var meshphong_frag = "#define PHONG\nuniform vec3 diffuse;\nuniform vec3 emissive;\nuniform vec3 specular;\nuniform float shininess;\nuniform float opacity;\n#include <common>\n#include <packing>\n#include <color_pars_fragment>\n#include <uv_pars_fragment>\n#include <uv2_pars_fragment>\n#include <map_pars_fragment>\n#include <alphamap_pars_fragment>\n#include <aomap_pars_fragment>\n#include <lightmap_pars_fragment>\n#include <emissivemap_pars_fragment>\n#include <envmap_pars_fragment>\n#include <fog_pars_fragment>\n#include <bsdfs>\n#include <lights_pars>\n#include <lights_phong_pars_fragment>\n#include <shadowmap_pars_fragment>\n#include <bumpmap_pars_fragment>\n#include <normalmap_pars_fragment>\n#include <specularmap_pars_fragment>\n#include <logdepthbuf_pars_fragment>\n#include <clipping_planes_pars_fragment>\nvoid main() {\n\t#include <clipping_planes_fragment>\n\tvec4 diffuseColor = vec4( diffuse, opacity );\n\tReflectedLight reflectedLight = ReflectedLight( vec3( 0.0 ), vec3( 0.0 ), vec3( 0.0 ), vec3( 0.0 ) );\n\tvec3 totalEmissiveRadiance = emissive;\n\t#include <logdepthbuf_fragment>\n\t#include <map_fragment>\n\t#include <color_fragment>\n\t#include <alphamap_fragment>\n\t#include <alphatest_fragment>\n\t#include <specularmap_fragment>\n\t#include <normal_flip>\n\t#include <normal_fragment>\n\t#include <emissivemap_fragment>\n\t#include <lights_phong_fragment>\n\t#include <lights_template>\n\t#include <aomap_fragment>\n\tvec3 outgoingLight = reflectedLight.directDiffuse + reflectedLight.indirectDiffuse + reflectedLight.directSpecular + reflectedLight.indirectSpecular + totalEmissiveRadiance;\n\t#include <envmap_fragment>\n\tgl_FragColor = vec4( outgoingLight, diffuseColor.a );\n\t#include <premultiplied_alpha_fragment>\n\t#include <tonemapping_fragment>\n\t#include <encodings_fragment>\n\t#include <fog_fragment>\n}\n";

var meshphong_vert = "#define PHONG\nvarying vec3 vViewPosition;\n#ifndef FLAT_SHADED\n\tvarying vec3 vNormal;\n#endif\n#include <common>\n#include <uv_pars_vertex>\n#include <uv2_pars_vertex>\n#include <displacementmap_pars_vertex>\n#include <envmap_pars_vertex>\n#include <color_pars_vertex>\n#include <morphtarget_pars_vertex>\n#include <skinning_pars_vertex>\n#include <shadowmap_pars_vertex>\n#include <logdepthbuf_pars_vertex>\n#include <clipping_planes_pars_vertex>\nvoid main() {\n\t#include <uv_vertex>\n\t#include <uv2_vertex>\n\t#include <color_vertex>\n\t#include <beginnormal_vertex>\n\t#include <morphnormal_vertex>\n\t#include <skinbase_vertex>\n\t#include <skinnormal_vertex>\n\t#include <defaultnormal_vertex>\n#ifndef FLAT_SHADED\n\tvNormal = normalize( transformedNormal );\n#endif\n\t#include <begin_vertex>\n\t#include <displacementmap_vertex>\n\t#include <morphtarget_vertex>\n\t#include <skinning_vertex>\n\t#include <project_vertex>\n\t#include <logdepthbuf_vertex>\n\t#include <clipping_planes_vertex>\n\tvViewPosition = - mvPosition.xyz;\n\t#include <worldpos_vertex>\n\t#include <envmap_vertex>\n\t#include <shadowmap_vertex>\n}\n";

var meshphysical_frag = "#define PHYSICAL\nuniform vec3 diffuse;\nuniform vec3 emissive;\nuniform float roughness;\nuniform float metalness;\nuniform float opacity;\n#ifndef STANDARD\n\tuniform float clearCoat;\n\tuniform float clearCoatRoughness;\n#endif\nuniform float envMapIntensity;\nvarying vec3 vViewPosition;\n#ifndef FLAT_SHADED\n\tvarying vec3 vNormal;\n#endif\n#include <common>\n#include <packing>\n#include <color_pars_fragment>\n#include <uv_pars_fragment>\n#include <uv2_pars_fragment>\n#include <map_pars_fragment>\n#include <alphamap_pars_fragment>\n#include <aomap_pars_fragment>\n#include <lightmap_pars_fragment>\n#include <emissivemap_pars_fragment>\n#include <envmap_pars_fragment>\n#include <fog_pars_fragment>\n#include <bsdfs>\n#include <cube_uv_reflection_fragment>\n#include <lights_pars>\n#include <lights_physical_pars_fragment>\n#include <shadowmap_pars_fragment>\n#include <bumpmap_pars_fragment>\n#include <normalmap_pars_fragment>\n#include <roughnessmap_pars_fragment>\n#include <metalnessmap_pars_fragment>\n#include <logdepthbuf_pars_fragment>\n#include <clipping_planes_pars_fragment>\nvoid main() {\n\t#include <clipping_planes_fragment>\n\tvec4 diffuseColor = vec4( diffuse, opacity );\n\tReflectedLight reflectedLight = ReflectedLight( vec3( 0.0 ), vec3( 0.0 ), vec3( 0.0 ), vec3( 0.0 ) );\n\tvec3 totalEmissiveRadiance = emissive;\n\t#include <logdepthbuf_fragment>\n\t#include <map_fragment>\n\t#include <color_fragment>\n\t#include <alphamap_fragment>\n\t#include <alphatest_fragment>\n\t#include <specularmap_fragment>\n\t#include <roughnessmap_fragment>\n\t#include <metalnessmap_fragment>\n\t#include <normal_flip>\n\t#include <normal_fragment>\n\t#include <emissivemap_fragment>\n\t#include <lights_physical_fragment>\n\t#include <lights_template>\n\t#include <aomap_fragment>\n\tvec3 outgoingLight = reflectedLight.directDiffuse + reflectedLight.indirectDiffuse + reflectedLight.directSpecular + reflectedLight.indirectSpecular + totalEmissiveRadiance;\n\tgl_FragColor = vec4( outgoingLight, diffuseColor.a );\n\t#include <premultiplied_alpha_fragment>\n\t#include <tonemapping_fragment>\n\t#include <encodings_fragment>\n\t#include <fog_fragment>\n}\n";

var meshphysical_vert = "#define PHYSICAL\nvarying vec3 vViewPosition;\n#ifndef FLAT_SHADED\n\tvarying vec3 vNormal;\n#endif\n#include <common>\n#include <uv_pars_vertex>\n#include <uv2_pars_vertex>\n#include <displacementmap_pars_vertex>\n#include <color_pars_vertex>\n#include <morphtarget_pars_vertex>\n#include <skinning_pars_vertex>\n#include <shadowmap_pars_vertex>\n#include <specularmap_pars_fragment>\n#include <logdepthbuf_pars_vertex>\n#include <clipping_planes_pars_vertex>\nvoid main() {\n\t#include <uv_vertex>\n\t#include <uv2_vertex>\n\t#include <color_vertex>\n\t#include <beginnormal_vertex>\n\t#include <morphnormal_vertex>\n\t#include <skinbase_vertex>\n\t#include <skinnormal_vertex>\n\t#include <defaultnormal_vertex>\n#ifndef FLAT_SHADED\n\tvNormal = normalize( transformedNormal );\n#endif\n\t#include <begin_vertex>\n\t#include <displacementmap_vertex>\n\t#include <morphtarget_vertex>\n\t#include <skinning_vertex>\n\t#include <project_vertex>\n\t#include <logdepthbuf_vertex>\n\t#include <clipping_planes_vertex>\n\tvViewPosition = - mvPosition.xyz;\n\t#include <worldpos_vertex>\n\t#include <shadowmap_vertex>\n}\n";

var normal_frag = "uniform float opacity;\nvarying vec3 vNormal;\n#include <common>\n#include <packing>\n#include <logdepthbuf_pars_fragment>\n#include <clipping_planes_pars_fragment>\nvoid main() {\n\t#include <clipping_planes_fragment>\n\tgl_FragColor = vec4( packNormalToRGB( vNormal ), opacity );\n\t#include <logdepthbuf_fragment>\n}\n";

var normal_vert = "varying vec3 vNormal;\n#include <common>\n#include <morphtarget_pars_vertex>\n#include <logdepthbuf_pars_vertex>\n#include <clipping_planes_pars_vertex>\nvoid main() {\n\tvNormal = normalize( normalMatrix * normal );\n\t#include <begin_vertex>\n\t#include <morphtarget_vertex>\n\t#include <project_vertex>\n\t#include <logdepthbuf_vertex>\n\t#include <clipping_planes_vertex>\n}\n";

var points_frag = "uniform vec3 diffuse;\nuniform float opacity;\n#include <common>\n#include <packing>\n#include <color_pars_fragment>\n#include <map_particle_pars_fragment>\n#include <fog_pars_fragment>\n#include <shadowmap_pars_fragment>\n#include <logdepthbuf_pars_fragment>\n#include <clipping_planes_pars_fragment>\nvoid main() {\n\t#include <clipping_planes_fragment>\n\tvec3 outgoingLight = vec3( 0.0 );\n\tvec4 diffuseColor = vec4( diffuse, opacity );\n\t#include <logdepthbuf_fragment>\n\t#include <map_particle_fragment>\n\t#include <color_fragment>\n\t#include <alphatest_fragment>\n\toutgoingLight = diffuseColor.rgb;\n\tgl_FragColor = vec4( outgoingLight, diffuseColor.a );\n\t#include <premultiplied_alpha_fragment>\n\t#include <tonemapping_fragment>\n\t#include <encodings_fragment>\n\t#include <fog_fragment>\n}\n";

var points_vert = "uniform float size;\nuniform float scale;\n#include <common>\n#include <color_pars_vertex>\n#include <shadowmap_pars_vertex>\n#include <logdepthbuf_pars_vertex>\n#include <clipping_planes_pars_vertex>\nvoid main() {\n\t#include <color_vertex>\n\t#include <begin_vertex>\n\t#include <project_vertex>\n\t#ifdef USE_SIZEATTENUATION\n\t\tgl_PointSize = size * ( scale / - mvPosition.z );\n\t#else\n\t\tgl_PointSize = size;\n\t#endif\n\t#include <logdepthbuf_vertex>\n\t#include <clipping_planes_vertex>\n\t#include <worldpos_vertex>\n\t#include <shadowmap_vertex>\n}\n";

var shadow_frag = "uniform float opacity;\n#include <common>\n#include <packing>\n#include <bsdfs>\n#include <lights_pars>\n#include <shadowmap_pars_fragment>\n#include <shadowmask_pars_fragment>\nvoid main() {\n\tgl_FragColor = vec4( 0.0, 0.0, 0.0, opacity * ( 1.0  - getShadowMask() ) );\n}\n";

var shadow_vert = "#include <shadowmap_pars_vertex>\nvoid main() {\n\t#include <begin_vertex>\n\t#include <project_vertex>\n\t#include <worldpos_vertex>\n\t#include <shadowmap_vertex>\n}\n";

var ShaderChunk = {
	alphamap_fragment: alphamap_fragment,
	alphamap_pars_fragment: alphamap_pars_fragment,
	alphatest_fragment: alphatest_fragment,
	aomap_fragment: aomap_fragment,
	aomap_pars_fragment: aomap_pars_fragment,
	begin_vertex: begin_vertex,
	beginnormal_vertex: beginnormal_vertex,
	bsdfs: bsdfs,
	bumpmap_pars_fragment: bumpmap_pars_fragment,
	clipping_planes_fragment: clipping_planes_fragment,
	clipping_planes_pars_fragment: clipping_planes_pars_fragment,
	clipping_planes_pars_vertex: clipping_planes_pars_vertex,
	clipping_planes_vertex: clipping_planes_vertex,
	color_fragment: color_fragment,
	color_pars_fragment: color_pars_fragment,
	color_pars_vertex: color_pars_vertex,
	color_vertex: color_vertex,
	common: common,
	cube_uv_reflection_fragment: cube_uv_reflection_fragment,
	defaultnormal_vertex: defaultnormal_vertex,
	displacementmap_pars_vertex: displacementmap_pars_vertex,
	displacementmap_vertex: displacementmap_vertex,
	emissivemap_fragment: emissivemap_fragment,
	emissivemap_pars_fragment: emissivemap_pars_fragment,
	encodings_fragment: encodings_fragment,
	encodings_pars_fragment: encodings_pars_fragment,
	envmap_fragment: envmap_fragment,
	envmap_pars_fragment: envmap_pars_fragment,
	envmap_pars_vertex: envmap_pars_vertex,
	envmap_vertex: envmap_vertex,
	fog_fragment: fog_fragment,
	fog_pars_fragment: fog_pars_fragment,
	lightmap_fragment: lightmap_fragment,
	lightmap_pars_fragment: lightmap_pars_fragment,
	lights_lambert_vertex: lights_lambert_vertex,
	lights_pars: lights_pars,
	lights_phong_fragment: lights_phong_fragment,
	lights_phong_pars_fragment: lights_phong_pars_fragment,
	lights_physical_fragment: lights_physical_fragment,
	lights_physical_pars_fragment: lights_physical_pars_fragment,
	lights_template: lights_template,
	logdepthbuf_fragment: logdepthbuf_fragment,
	logdepthbuf_pars_fragment: logdepthbuf_pars_fragment,
	logdepthbuf_pars_vertex: logdepthbuf_pars_vertex,
	logdepthbuf_vertex: logdepthbuf_vertex,
	map_fragment: map_fragment,
	map_pars_fragment: map_pars_fragment,
	map_particle_fragment: map_particle_fragment,
	map_particle_pars_fragment: map_particle_pars_fragment,
	metalnessmap_fragment: metalnessmap_fragment,
	metalnessmap_pars_fragment: metalnessmap_pars_fragment,
	morphnormal_vertex: morphnormal_vertex,
	morphtarget_pars_vertex: morphtarget_pars_vertex,
	morphtarget_vertex: morphtarget_vertex,
	normal_flip: normal_flip,
	normal_fragment: normal_fragment,
	normalmap_pars_fragment: normalmap_pars_fragment,
	packing: packing,
	premultiplied_alpha_fragment: premultiplied_alpha_fragment,
	project_vertex: project_vertex,
	roughnessmap_fragment: roughnessmap_fragment,
	roughnessmap_pars_fragment: roughnessmap_pars_fragment,
	shadowmap_pars_fragment: shadowmap_pars_fragment,
	shadowmap_pars_vertex: shadowmap_pars_vertex,
	shadowmap_vertex: shadowmap_vertex,
	shadowmask_pars_fragment: shadowmask_pars_fragment,
	skinbase_vertex: skinbase_vertex,
	skinning_pars_vertex: skinning_pars_vertex,
	skinning_vertex: skinning_vertex,
	skinnormal_vertex: skinnormal_vertex,
	specularmap_fragment: specularmap_fragment,
	specularmap_pars_fragment: specularmap_pars_fragment,
	tonemapping_fragment: tonemapping_fragment,
	tonemapping_pars_fragment: tonemapping_pars_fragment,
	uv_pars_fragment: uv_pars_fragment,
	uv_pars_vertex: uv_pars_vertex,
	uv_vertex: uv_vertex,
	uv2_pars_fragment: uv2_pars_fragment,
	uv2_pars_vertex: uv2_pars_vertex,
	uv2_vertex: uv2_vertex,
	worldpos_vertex: worldpos_vertex,

	cube_frag: cube_frag,
	cube_vert: cube_vert,
	depth_frag: depth_frag,
	depth_vert: depth_vert,
	distanceRGBA_frag: distanceRGBA_frag,
	distanceRGBA_vert: distanceRGBA_vert,
	equirect_frag: equirect_frag,
	equirect_vert: equirect_vert,
	linedashed_frag: linedashed_frag,
	linedashed_vert: linedashed_vert,
	meshbasic_frag: meshbasic_frag,
	meshbasic_vert: meshbasic_vert,
	meshlambert_frag: meshlambert_frag,
	meshlambert_vert: meshlambert_vert,
	meshphong_frag: meshphong_frag,
	meshphong_vert: meshphong_vert,
	meshphysical_frag: meshphysical_frag,
	meshphysical_vert: meshphysical_vert,
	normal_frag: normal_frag,
	normal_vert: normal_vert,
	points_frag: points_frag,
	points_vert: points_vert,
	shadow_frag: shadow_frag,
	shadow_vert: shadow_vert
};

/**
 * Uniforms library for shared webgl shaders
 */

var UniformsLib = {

	common: {

		diffuse: { value: new Color(0xeeeeee) },
		opacity: { value: 1.0 },

		map: { value: null },
		offsetRepeat: { value: new Vector4(0, 0, 1, 1) },

		specularMap: { value: null },
		alphaMap: { value: null },

		envMap: { value: null },
		flipEnvMap: { value: -1 },
		reflectivity: { value: 1.0 },
		refractionRatio: { value: 0.98 }

	},

	aomap: {

		aoMap: { value: null },
		aoMapIntensity: { value: 1 }

	},

	lightmap: {

		lightMap: { value: null },
		lightMapIntensity: { value: 1 }

	},

	emissivemap: {

		emissiveMap: { value: null }

	},

	bumpmap: {

		bumpMap: { value: null },
		bumpScale: { value: 1 }

	},

	normalmap: {

		normalMap: { value: null },
		normalScale: { value: new Vector2(1, 1) }

	},

	displacementmap: {

		displacementMap: { value: null },
		displacementScale: { value: 1 },
		displacementBias: { value: 0 }

	},

	roughnessmap: {

		roughnessMap: { value: null }

	},

	metalnessmap: {

		metalnessMap: { value: null }

	},

	fog: {

		fogDensity: { value: 0.00025 },
		fogNear: { value: 1 },
		fogFar: { value: 2000 },
		fogColor: { value: new Color(0xffffff) }

	},

	lights: {

		ambientLightColor: { value: [] },

		directionalLights: { value: [], properties: {
				direction: {},
				color: {},

				shadow: {},
				shadowBias: {},
				shadowRadius: {},
				shadowMapSize: {}
			} },

		directionalShadowMap: { value: [] },
		directionalShadowMatrix: { value: [] },

		spotLights: { value: [], properties: {
				color: {},
				position: {},
				direction: {},
				distance: {},
				coneCos: {},
				penumbraCos: {},
				decay: {},

				shadow: {},
				shadowBias: {},
				shadowRadius: {},
				shadowMapSize: {}
			} },

		spotShadowMap: { value: [] },
		spotShadowMatrix: { value: [] },

		pointLights: { value: [], properties: {
				color: {},
				position: {},
				decay: {},
				distance: {},

				shadow: {},
				shadowBias: {},
				shadowRadius: {},
				shadowMapSize: {}
			} },

		pointShadowMap: { value: [] },
		pointShadowMatrix: { value: [] },

		hemisphereLights: { value: [], properties: {
				direction: {},
				skyColor: {},
				groundColor: {}
			} }

	},

	points: {

		diffuse: { value: new Color(0xeeeeee) },
		opacity: { value: 1.0 },
		size: { value: 1.0 },
		scale: { value: 1.0 },
		map: { value: null },
		offsetRepeat: { value: new Vector4(0, 0, 1, 1) }

	}

};

/**
 * @author mrdoob / http://mrdoob.com/
 */

function ShadowMaterial() {

	ShaderMaterial.call(this, {
		uniforms: UniformsUtils.merge([UniformsLib["lights"], {
			opacity: { value: 1.0 }
		}]),
		vertexShader: ShaderChunk['shadow_vert'],
		fragmentShader: ShaderChunk['shadow_frag']
	});

	this.lights = true;
	this.transparent = true;

	Object.defineProperties(this, {
		opacity: {
			enumerable: true,
			get: function get() {
				return this.uniforms.opacity.value;
			},
			set: function set(value) {
				this.uniforms.opacity.value = value;
			}
		}
	});
}

ShadowMaterial.prototype = Object.create(ShaderMaterial.prototype);
ShadowMaterial.prototype.constructor = ShadowMaterial;

ShadowMaterial.prototype.isShadowMaterial = true;

/**
 * @author mrdoob / http://mrdoob.com/
 */

function RawShaderMaterial(parameters) {

  ShaderMaterial.call(this, parameters);

  this.type = 'RawShaderMaterial';
}

RawShaderMaterial.prototype = Object.create(ShaderMaterial.prototype);
RawShaderMaterial.prototype.constructor = RawShaderMaterial;

RawShaderMaterial.prototype.isRawShaderMaterial = true;

/**
 * @author mrdoob / http://mrdoob.com/
 */

function MultiMaterial(materials) {

	this.uuid = _Math.generateUUID();

	this.type = 'MultiMaterial';

	this.materials = materials instanceof Array ? materials : [];

	this.visible = true;
}

MultiMaterial.prototype = {

	constructor: MultiMaterial,

	isMultiMaterial: true,

	toJSON: function toJSON(meta) {

		var output = {
			metadata: {
				version: 4.2,
				type: 'material',
				generator: 'MaterialExporter'
			},
			uuid: this.uuid,
			type: this.type,
			materials: []
		};

		var materials = this.materials;

		for (var i = 0, l = materials.length; i < l; i++) {

			var material = materials[i].toJSON(meta);
			delete material.metadata;

			output.materials.push(material);
		}

		output.visible = this.visible;

		return output;
	},

	clone: function clone() {

		var material = new this.constructor();

		for (var i = 0; i < this.materials.length; i++) {

			material.materials.push(this.materials[i].clone());
		}

		material.visible = this.visible;

		return material;
	}

};

/**
 * @author WestLangley / http://github.com/WestLangley
 *
 * parameters = {
 *  reflectivity: <float>
 * }
 */

function MeshPhysicalMaterial(parameters) {

	MeshStandardMaterial.call(this);

	this.defines = { 'PHYSICAL': '' };

	this.type = 'MeshPhysicalMaterial';

	this.reflectivity = 0.5; // maps to F0 = 0.04

	this.clearCoat = 0.0;
	this.clearCoatRoughness = 0.0;

	this.setValues(parameters);
}

MeshPhysicalMaterial.prototype = Object.create(MeshStandardMaterial.prototype);
MeshPhysicalMaterial.prototype.constructor = MeshPhysicalMaterial;

MeshPhysicalMaterial.prototype.isMeshPhysicalMaterial = true;

MeshPhysicalMaterial.prototype.copy = function (source) {

	MeshStandardMaterial.prototype.copy.call(this, source);

	this.defines = { 'PHYSICAL': '' };

	this.reflectivity = source.reflectivity;

	this.clearCoat = source.clearCoat;
	this.clearCoatRoughness = source.clearCoatRoughness;

	return this;
};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 *
 * parameters = {
 *  color: <hex>,
 *  specular: <hex>,
 *  shininess: <float>,
 *  opacity: <float>,
 *
 *  map: new THREE.Texture( <Image> ),
 *
 *  lightMap: new THREE.Texture( <Image> ),
 *  lightMapIntensity: <float>
 *
 *  aoMap: new THREE.Texture( <Image> ),
 *  aoMapIntensity: <float>
 *
 *  emissive: <hex>,
 *  emissiveIntensity: <float>
 *  emissiveMap: new THREE.Texture( <Image> ),
 *
 *  bumpMap: new THREE.Texture( <Image> ),
 *  bumpScale: <float>,
 *
 *  normalMap: new THREE.Texture( <Image> ),
 *  normalScale: <Vector2>,
 *
 *  displacementMap: new THREE.Texture( <Image> ),
 *  displacementScale: <float>,
 *  displacementBias: <float>,
 *
 *  specularMap: new THREE.Texture( <Image> ),
 *
 *  alphaMap: new THREE.Texture( <Image> ),
 *
 *  envMap: new THREE.TextureCube( [posx, negx, posy, negy, posz, negz] ),
 *  combine: THREE.Multiply,
 *  reflectivity: <float>,
 *  refractionRatio: <float>,
 *
 *  wireframe: <boolean>,
 *  wireframeLinewidth: <float>,
 *
 *  skinning: <bool>,
 *  morphTargets: <bool>,
 *  morphNormals: <bool>
 * }
 */

function MeshPhongMaterial(parameters) {

	Material.call(this);

	this.type = 'MeshPhongMaterial';

	this.color = new Color(0xffffff); // diffuse
	this.specular = new Color(0x111111);
	this.shininess = 30;

	this.map = null;

	this.lightMap = null;
	this.lightMapIntensity = 1.0;

	this.aoMap = null;
	this.aoMapIntensity = 1.0;

	this.emissive = new Color(0x000000);
	this.emissiveIntensity = 1.0;
	this.emissiveMap = null;

	this.bumpMap = null;
	this.bumpScale = 1;

	this.normalMap = null;
	this.normalScale = new Vector2(1, 1);

	this.displacementMap = null;
	this.displacementScale = 1;
	this.displacementBias = 0;

	this.specularMap = null;

	this.alphaMap = null;

	this.envMap = null;
	this.combine = MultiplyOperation;
	this.reflectivity = 1;
	this.refractionRatio = 0.98;

	this.wireframe = false;
	this.wireframeLinewidth = 1;
	this.wireframeLinecap = 'round';
	this.wireframeLinejoin = 'round';

	this.skinning = false;
	this.morphTargets = false;
	this.morphNormals = false;

	this.setValues(parameters);
}

MeshPhongMaterial.prototype = Object.create(Material.prototype);
MeshPhongMaterial.prototype.constructor = MeshPhongMaterial;

MeshPhongMaterial.prototype.isMeshPhongMaterial = true;

MeshPhongMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.color.copy(source.color);
	this.specular.copy(source.specular);
	this.shininess = source.shininess;

	this.map = source.map;

	this.lightMap = source.lightMap;
	this.lightMapIntensity = source.lightMapIntensity;

	this.aoMap = source.aoMap;
	this.aoMapIntensity = source.aoMapIntensity;

	this.emissive.copy(source.emissive);
	this.emissiveMap = source.emissiveMap;
	this.emissiveIntensity = source.emissiveIntensity;

	this.bumpMap = source.bumpMap;
	this.bumpScale = source.bumpScale;

	this.normalMap = source.normalMap;
	this.normalScale.copy(source.normalScale);

	this.displacementMap = source.displacementMap;
	this.displacementScale = source.displacementScale;
	this.displacementBias = source.displacementBias;

	this.specularMap = source.specularMap;

	this.alphaMap = source.alphaMap;

	this.envMap = source.envMap;
	this.combine = source.combine;
	this.reflectivity = source.reflectivity;
	this.refractionRatio = source.refractionRatio;

	this.wireframe = source.wireframe;
	this.wireframeLinewidth = source.wireframeLinewidth;
	this.wireframeLinecap = source.wireframeLinecap;
	this.wireframeLinejoin = source.wireframeLinejoin;

	this.skinning = source.skinning;
	this.morphTargets = source.morphTargets;
	this.morphNormals = source.morphNormals;

	return this;
};

/**
 * @author mrdoob / http://mrdoob.com/
 *
 * parameters = {
 *  opacity: <float>,
 *
 *  wireframe: <boolean>,
 *  wireframeLinewidth: <float>
 * }
 */

function MeshNormalMaterial(parameters) {

	Material.call(this, parameters);

	this.type = 'MeshNormalMaterial';

	this.wireframe = false;
	this.wireframeLinewidth = 1;

	this.fog = false;
	this.lights = false;
	this.morphTargets = false;

	this.setValues(parameters);
}

MeshNormalMaterial.prototype = Object.create(Material.prototype);
MeshNormalMaterial.prototype.constructor = MeshNormalMaterial;

MeshNormalMaterial.prototype.isMeshNormalMaterial = true;

MeshNormalMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.wireframe = source.wireframe;
	this.wireframeLinewidth = source.wireframeLinewidth;

	return this;
};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 *
 * parameters = {
 *  color: <hex>,
 *  opacity: <float>,
 *
 *  map: new THREE.Texture( <Image> ),
 *
 *  lightMap: new THREE.Texture( <Image> ),
 *  lightMapIntensity: <float>
 *
 *  aoMap: new THREE.Texture( <Image> ),
 *  aoMapIntensity: <float>
 *
 *  emissive: <hex>,
 *  emissiveIntensity: <float>
 *  emissiveMap: new THREE.Texture( <Image> ),
 *
 *  specularMap: new THREE.Texture( <Image> ),
 *
 *  alphaMap: new THREE.Texture( <Image> ),
 *
 *  envMap: new THREE.TextureCube( [posx, negx, posy, negy, posz, negz] ),
 *  combine: THREE.Multiply,
 *  reflectivity: <float>,
 *  refractionRatio: <float>,
 *
 *  wireframe: <boolean>,
 *  wireframeLinewidth: <float>,
 *
 *  skinning: <bool>,
 *  morphTargets: <bool>,
 *  morphNormals: <bool>
 * }
 */

function MeshLambertMaterial(parameters) {

	Material.call(this);

	this.type = 'MeshLambertMaterial';

	this.color = new Color(0xffffff); // diffuse

	this.map = null;

	this.lightMap = null;
	this.lightMapIntensity = 1.0;

	this.aoMap = null;
	this.aoMapIntensity = 1.0;

	this.emissive = new Color(0x000000);
	this.emissiveIntensity = 1.0;
	this.emissiveMap = null;

	this.specularMap = null;

	this.alphaMap = null;

	this.envMap = null;
	this.combine = MultiplyOperation;
	this.reflectivity = 1;
	this.refractionRatio = 0.98;

	this.wireframe = false;
	this.wireframeLinewidth = 1;
	this.wireframeLinecap = 'round';
	this.wireframeLinejoin = 'round';

	this.skinning = false;
	this.morphTargets = false;
	this.morphNormals = false;

	this.setValues(parameters);
}

MeshLambertMaterial.prototype = Object.create(Material.prototype);
MeshLambertMaterial.prototype.constructor = MeshLambertMaterial;

MeshLambertMaterial.prototype.isMeshLambertMaterial = true;

MeshLambertMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.color.copy(source.color);

	this.map = source.map;

	this.lightMap = source.lightMap;
	this.lightMapIntensity = source.lightMapIntensity;

	this.aoMap = source.aoMap;
	this.aoMapIntensity = source.aoMapIntensity;

	this.emissive.copy(source.emissive);
	this.emissiveMap = source.emissiveMap;
	this.emissiveIntensity = source.emissiveIntensity;

	this.specularMap = source.specularMap;

	this.alphaMap = source.alphaMap;

	this.envMap = source.envMap;
	this.combine = source.combine;
	this.reflectivity = source.reflectivity;
	this.refractionRatio = source.refractionRatio;

	this.wireframe = source.wireframe;
	this.wireframeLinewidth = source.wireframeLinewidth;
	this.wireframeLinecap = source.wireframeLinecap;
	this.wireframeLinejoin = source.wireframeLinejoin;

	this.skinning = source.skinning;
	this.morphTargets = source.morphTargets;
	this.morphNormals = source.morphNormals;

	return this;
};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 * @author bhouston / https://clara.io
 * @author WestLangley / http://github.com/WestLangley
 *
 * parameters = {
 *
 *  opacity: <float>,
 *
 *  map: new THREE.Texture( <Image> ),
 *
 *  alphaMap: new THREE.Texture( <Image> ),
 *
 *  displacementMap: new THREE.Texture( <Image> ),
 *  displacementScale: <float>,
 *  displacementBias: <float>,
 *
 *  wireframe: <boolean>,
 *  wireframeLinewidth: <float>
 * }
 */

function MeshDepthMaterial(parameters) {

	Material.call(this);

	this.type = 'MeshDepthMaterial';

	this.depthPacking = BasicDepthPacking;

	this.skinning = false;
	this.morphTargets = false;

	this.map = null;

	this.alphaMap = null;

	this.displacementMap = null;
	this.displacementScale = 1;
	this.displacementBias = 0;

	this.wireframe = false;
	this.wireframeLinewidth = 1;

	this.fog = false;
	this.lights = false;

	this.setValues(parameters);
}

MeshDepthMaterial.prototype = Object.create(Material.prototype);
MeshDepthMaterial.prototype.constructor = MeshDepthMaterial;

MeshDepthMaterial.prototype.isMeshDepthMaterial = true;

MeshDepthMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.depthPacking = source.depthPacking;

	this.skinning = source.skinning;
	this.morphTargets = source.morphTargets;

	this.map = source.map;

	this.alphaMap = source.alphaMap;

	this.displacementMap = source.displacementMap;
	this.displacementScale = source.displacementScale;
	this.displacementBias = source.displacementBias;

	this.wireframe = source.wireframe;
	this.wireframeLinewidth = source.wireframeLinewidth;

	return this;
};

/**
 * @author alteredq / http://alteredqualia.com/
 *
 * parameters = {
 *  color: <hex>,
 *  opacity: <float>,
 *
 *  linewidth: <float>,
 *
 *  scale: <float>,
 *  dashSize: <float>,
 *  gapSize: <float>
 * }
 */

function LineDashedMaterial(parameters) {

	Material.call(this);

	this.type = 'LineDashedMaterial';

	this.color = new Color(0xffffff);

	this.linewidth = 1;

	this.scale = 1;
	this.dashSize = 3;
	this.gapSize = 1;

	this.lights = false;

	this.setValues(parameters);
}

LineDashedMaterial.prototype = Object.create(Material.prototype);
LineDashedMaterial.prototype.constructor = LineDashedMaterial;

LineDashedMaterial.prototype.isLineDashedMaterial = true;

LineDashedMaterial.prototype.copy = function (source) {

	Material.prototype.copy.call(this, source);

	this.color.copy(source.color);

	this.linewidth = source.linewidth;

	this.scale = source.scale;
	this.dashSize = source.dashSize;
	this.gapSize = source.gapSize;

	return this;
};



var Materials = Object.freeze({
	ShadowMaterial: ShadowMaterial,
	SpriteMaterial: SpriteMaterial,
	RawShaderMaterial: RawShaderMaterial,
	ShaderMaterial: ShaderMaterial,
	PointsMaterial: PointsMaterial,
	MultiMaterial: MultiMaterial,
	MeshPhysicalMaterial: MeshPhysicalMaterial,
	MeshStandardMaterial: MeshStandardMaterial,
	MeshPhongMaterial: MeshPhongMaterial,
	MeshNormalMaterial: MeshNormalMaterial,
	MeshLambertMaterial: MeshLambertMaterial,
	MeshDepthMaterial: MeshDepthMaterial,
	MeshBasicMaterial: MeshBasicMaterial,
	LineDashedMaterial: LineDashedMaterial,
	LineBasicMaterial: LineBasicMaterial,
	Material: Material
});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function MaterialLoader(manager) {

	this.manager = manager !== undefined ? manager : DefaultLoadingManager;
	this.textures = {};
}

Object.assign(MaterialLoader.prototype, {

	load: function load(url, onLoad, onProgress, onError) {

		var scope = this;

		var loader = new XHRLoader(scope.manager);
		loader.load(url, function (text) {

			onLoad(scope.parse(JSON.parse(text)));
		}, onProgress, onError);
	},

	setTextures: function setTextures(value) {

		this.textures = value;
	},

	parse: function parse(json) {

		var textures = this.textures;

		function getTexture(name) {

			if (textures[name] === undefined) {

				console.warn('THREE.MaterialLoader: Undefined texture', name);
			}

			return textures[name];
		}

		var material = new Materials[json.type]();

		if (json.uuid !== undefined) material.uuid = json.uuid;
		if (json.name !== undefined) material.name = json.name;
		if (json.color !== undefined) material.color.setHex(json.color);
		if (json.roughness !== undefined) material.roughness = json.roughness;
		if (json.metalness !== undefined) material.metalness = json.metalness;
		if (json.emissive !== undefined) material.emissive.setHex(json.emissive);
		if (json.specular !== undefined) material.specular.setHex(json.specular);
		if (json.shininess !== undefined) material.shininess = json.shininess;
		if (json.uniforms !== undefined) material.uniforms = json.uniforms;
		if (json.vertexShader !== undefined) material.vertexShader = json.vertexShader;
		if (json.fragmentShader !== undefined) material.fragmentShader = json.fragmentShader;
		if (json.vertexColors !== undefined) material.vertexColors = json.vertexColors;
		if (json.fog !== undefined) material.fog = json.fog;
		if (json.shading !== undefined) material.shading = json.shading;
		if (json.blending !== undefined) material.blending = json.blending;
		if (json.side !== undefined) material.side = json.side;
		if (json.opacity !== undefined) material.opacity = json.opacity;
		if (json.transparent !== undefined) material.transparent = json.transparent;
		if (json.alphaTest !== undefined) material.alphaTest = json.alphaTest;
		if (json.depthTest !== undefined) material.depthTest = json.depthTest;
		if (json.depthWrite !== undefined) material.depthWrite = json.depthWrite;
		if (json.colorWrite !== undefined) material.colorWrite = json.colorWrite;
		if (json.wireframe !== undefined) material.wireframe = json.wireframe;
		if (json.wireframeLinewidth !== undefined) material.wireframeLinewidth = json.wireframeLinewidth;
		if (json.wireframeLinecap !== undefined) material.wireframeLinecap = json.wireframeLinecap;
		if (json.wireframeLinejoin !== undefined) material.wireframeLinejoin = json.wireframeLinejoin;
		if (json.skinning !== undefined) material.skinning = json.skinning;
		if (json.morphTargets !== undefined) material.morphTargets = json.morphTargets;

		// for PointsMaterial

		if (json.size !== undefined) material.size = json.size;
		if (json.sizeAttenuation !== undefined) material.sizeAttenuation = json.sizeAttenuation;

		// maps

		if (json.map !== undefined) material.map = getTexture(json.map);

		if (json.alphaMap !== undefined) {

			material.alphaMap = getTexture(json.alphaMap);
			material.transparent = true;
		}

		if (json.bumpMap !== undefined) material.bumpMap = getTexture(json.bumpMap);
		if (json.bumpScale !== undefined) material.bumpScale = json.bumpScale;

		if (json.normalMap !== undefined) material.normalMap = getTexture(json.normalMap);
		if (json.normalScale !== undefined) {

			var normalScale = json.normalScale;

			if (Array.isArray(normalScale) === false) {

				// Blender exporter used to export a scalar. See #7459

				normalScale = [normalScale, normalScale];
			}

			material.normalScale = new Vector2().fromArray(normalScale);
		}

		if (json.displacementMap !== undefined) material.displacementMap = getTexture(json.displacementMap);
		if (json.displacementScale !== undefined) material.displacementScale = json.displacementScale;
		if (json.displacementBias !== undefined) material.displacementBias = json.displacementBias;

		if (json.roughnessMap !== undefined) material.roughnessMap = getTexture(json.roughnessMap);
		if (json.metalnessMap !== undefined) material.metalnessMap = getTexture(json.metalnessMap);

		if (json.emissiveMap !== undefined) material.emissiveMap = getTexture(json.emissiveMap);
		if (json.emissiveIntensity !== undefined) material.emissiveIntensity = json.emissiveIntensity;

		if (json.specularMap !== undefined) material.specularMap = getTexture(json.specularMap);

		if (json.envMap !== undefined) material.envMap = getTexture(json.envMap);

		if (json.reflectivity !== undefined) material.reflectivity = json.reflectivity;

		if (json.lightMap !== undefined) material.lightMap = getTexture(json.lightMap);
		if (json.lightMapIntensity !== undefined) material.lightMapIntensity = json.lightMapIntensity;

		if (json.aoMap !== undefined) material.aoMap = getTexture(json.aoMap);
		if (json.aoMapIntensity !== undefined) material.aoMapIntensity = json.aoMapIntensity;

		// MultiMaterial

		if (json.materials !== undefined) {

			for (var i = 0, l = json.materials.length; i < l; i++) {

				material.materials.push(this.parse(json.materials[i]));
			}
		}

		return material;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function BufferGeometryLoader(manager) {

	this.manager = manager !== undefined ? manager : DefaultLoadingManager;
}

Object.assign(BufferGeometryLoader.prototype, {

	load: function load(url, onLoad, onProgress, onError) {

		var scope = this;

		var loader = new XHRLoader(scope.manager);
		loader.load(url, function (text) {

			onLoad(scope.parse(JSON.parse(text)));
		}, onProgress, onError);
	},

	parse: function parse(json) {

		var geometry = new BufferGeometry();

		var index = json.data.index;

		var TYPED_ARRAYS = {
			'Int8Array': Int8Array,
			'Uint8Array': Uint8Array,
			'Uint8ClampedArray': Uint8ClampedArray,
			'Int16Array': Int16Array,
			'Uint16Array': Uint16Array,
			'Int32Array': Int32Array,
			'Uint32Array': Uint32Array,
			'Float32Array': Float32Array,
			'Float64Array': Float64Array
		};

		if (index !== undefined) {

			var typedArray = new TYPED_ARRAYS[index.type](index.array);
			geometry.setIndex(new BufferAttribute(typedArray, 1));
		}

		var attributes = json.data.attributes;

		for (var key in attributes) {

			var attribute = attributes[key];
			var typedArray = new TYPED_ARRAYS[attribute.type](attribute.array);

			geometry.addAttribute(key, new BufferAttribute(typedArray, attribute.itemSize, attribute.normalized));
		}

		var groups = json.data.groups || json.data.drawcalls || json.data.offsets;

		if (groups !== undefined) {

			for (var i = 0, n = groups.length; i !== n; ++i) {

				var group = groups[i];

				geometry.addGroup(group.start, group.count, group.materialIndex);
			}
		}

		var boundingSphere = json.data.boundingSphere;

		if (boundingSphere !== undefined) {

			var center = new Vector3();

			if (boundingSphere.center !== undefined) {

				center.fromArray(boundingSphere.center);
			}

			geometry.boundingSphere = new Sphere(center, boundingSphere.radius);
		}

		return geometry;
	}

});

/**
 * @author alteredq / http://alteredqualia.com/
 */

function Loader() {

	this.onLoadStart = function () {};
	this.onLoadProgress = function () {};
	this.onLoadComplete = function () {};
}

Loader.prototype = {

	constructor: Loader,

	crossOrigin: undefined,

	extractUrlBase: function extractUrlBase(url) {

		var parts = url.split('/');

		if (parts.length === 1) return './';

		parts.pop();

		return parts.join('/') + '/';
	},

	initMaterials: function initMaterials(materials, texturePath, crossOrigin) {

		var array = [];

		for (var i = 0; i < materials.length; ++i) {

			array[i] = this.createMaterial(materials[i], texturePath, crossOrigin);
		}

		return array;
	},

	createMaterial: function () {

		var color, textureLoader, materialLoader;

		return function createMaterial(m, texturePath, crossOrigin) {

			if (color === undefined) color = new Color();
			if (textureLoader === undefined) textureLoader = new TextureLoader();
			if (materialLoader === undefined) materialLoader = new MaterialLoader();

			// convert from old material format

			var textures = {};

			function loadTexture(path, repeat, offset, wrap, anisotropy) {

				var fullPath = texturePath + path;
				var loader = Loader.Handlers.get(fullPath);

				var texture;

				if (loader !== null) {

					texture = loader.load(fullPath);
				} else {

					textureLoader.setCrossOrigin(crossOrigin);
					texture = textureLoader.load(fullPath);
				}

				if (repeat !== undefined) {

					texture.repeat.fromArray(repeat);

					if (repeat[0] !== 1) texture.wrapS = RepeatWrapping;
					if (repeat[1] !== 1) texture.wrapT = RepeatWrapping;
				}

				if (offset !== undefined) {

					texture.offset.fromArray(offset);
				}

				if (wrap !== undefined) {

					if (wrap[0] === 'repeat') texture.wrapS = RepeatWrapping;
					if (wrap[0] === 'mirror') texture.wrapS = MirroredRepeatWrapping;

					if (wrap[1] === 'repeat') texture.wrapT = RepeatWrapping;
					if (wrap[1] === 'mirror') texture.wrapT = MirroredRepeatWrapping;
				}

				if (anisotropy !== undefined) {

					texture.anisotropy = anisotropy;
				}

				var uuid = _Math.generateUUID();

				textures[uuid] = texture;

				return uuid;
			}

			//

			var json = {
				uuid: _Math.generateUUID(),
				type: 'MeshLambertMaterial'
			};

			for (var name in m) {

				var value = m[name];

				switch (name) {
					case 'DbgColor':
					case 'DbgIndex':
					case 'opticalDensity':
					case 'illumination':
						break;
					case 'DbgName':
						json.name = value;
						break;
					case 'blending':
						json.blending = BlendingMode[value];
						break;
					case 'colorAmbient':
					case 'mapAmbient':
						console.warn('THREE.Loader.createMaterial:', name, 'is no longer supported.');
						break;
					case 'colorDiffuse':
						json.color = color.fromArray(value).getHex();
						break;
					case 'colorSpecular':
						json.specular = color.fromArray(value).getHex();
						break;
					case 'colorEmissive':
						json.emissive = color.fromArray(value).getHex();
						break;
					case 'specularCoef':
						json.shininess = value;
						break;
					case 'shading':
						if (value.toLowerCase() === 'basic') json.type = 'MeshBasicMaterial';
						if (value.toLowerCase() === 'phong') json.type = 'MeshPhongMaterial';
						if (value.toLowerCase() === 'standard') json.type = 'MeshStandardMaterial';
						break;
					case 'mapDiffuse':
						json.map = loadTexture(value, m.mapDiffuseRepeat, m.mapDiffuseOffset, m.mapDiffuseWrap, m.mapDiffuseAnisotropy);
						break;
					case 'mapDiffuseRepeat':
					case 'mapDiffuseOffset':
					case 'mapDiffuseWrap':
					case 'mapDiffuseAnisotropy':
						break;
					case 'mapEmissive':
						json.emissiveMap = loadTexture(value, m.mapEmissiveRepeat, m.mapEmissiveOffset, m.mapEmissiveWrap, m.mapEmissiveAnisotropy);
						break;
					case 'mapEmissiveRepeat':
					case 'mapEmissiveOffset':
					case 'mapEmissiveWrap':
					case 'mapEmissiveAnisotropy':
						break;
					case 'mapLight':
						json.lightMap = loadTexture(value, m.mapLightRepeat, m.mapLightOffset, m.mapLightWrap, m.mapLightAnisotropy);
						break;
					case 'mapLightRepeat':
					case 'mapLightOffset':
					case 'mapLightWrap':
					case 'mapLightAnisotropy':
						break;
					case 'mapAO':
						json.aoMap = loadTexture(value, m.mapAORepeat, m.mapAOOffset, m.mapAOWrap, m.mapAOAnisotropy);
						break;
					case 'mapAORepeat':
					case 'mapAOOffset':
					case 'mapAOWrap':
					case 'mapAOAnisotropy':
						break;
					case 'mapBump':
						json.bumpMap = loadTexture(value, m.mapBumpRepeat, m.mapBumpOffset, m.mapBumpWrap, m.mapBumpAnisotropy);
						break;
					case 'mapBumpScale':
						json.bumpScale = value;
						break;
					case 'mapBumpRepeat':
					case 'mapBumpOffset':
					case 'mapBumpWrap':
					case 'mapBumpAnisotropy':
						break;
					case 'mapNormal':
						json.normalMap = loadTexture(value, m.mapNormalRepeat, m.mapNormalOffset, m.mapNormalWrap, m.mapNormalAnisotropy);
						break;
					case 'mapNormalFactor':
						json.normalScale = [value, value];
						break;
					case 'mapNormalRepeat':
					case 'mapNormalOffset':
					case 'mapNormalWrap':
					case 'mapNormalAnisotropy':
						break;
					case 'mapSpecular':
						json.specularMap = loadTexture(value, m.mapSpecularRepeat, m.mapSpecularOffset, m.mapSpecularWrap, m.mapSpecularAnisotropy);
						break;
					case 'mapSpecularRepeat':
					case 'mapSpecularOffset':
					case 'mapSpecularWrap':
					case 'mapSpecularAnisotropy':
						break;
					case 'mapMetalness':
						json.metalnessMap = loadTexture(value, m.mapMetalnessRepeat, m.mapMetalnessOffset, m.mapMetalnessWrap, m.mapMetalnessAnisotropy);
						break;
					case 'mapMetalnessRepeat':
					case 'mapMetalnessOffset':
					case 'mapMetalnessWrap':
					case 'mapMetalnessAnisotropy':
						break;
					case 'mapRoughness':
						json.roughnessMap = loadTexture(value, m.mapRoughnessRepeat, m.mapRoughnessOffset, m.mapRoughnessWrap, m.mapRoughnessAnisotropy);
						break;
					case 'mapRoughnessRepeat':
					case 'mapRoughnessOffset':
					case 'mapRoughnessWrap':
					case 'mapRoughnessAnisotropy':
						break;
					case 'mapAlpha':
						json.alphaMap = loadTexture(value, m.mapAlphaRepeat, m.mapAlphaOffset, m.mapAlphaWrap, m.mapAlphaAnisotropy);
						break;
					case 'mapAlphaRepeat':
					case 'mapAlphaOffset':
					case 'mapAlphaWrap':
					case 'mapAlphaAnisotropy':
						break;
					case 'flipSided':
						json.side = BackSide;
						break;
					case 'doubleSided':
						json.side = DoubleSide;
						break;
					case 'transparency':
						console.warn('THREE.Loader.createMaterial: transparency has been renamed to opacity');
						json.opacity = value;
						break;
					case 'depthTest':
					case 'depthWrite':
					case 'colorWrite':
					case 'opacity':
					case 'reflectivity':
					case 'transparent':
					case 'visible':
					case 'wireframe':
						json[name] = value;
						break;
					case 'vertexColors':
						if (value === true) json.vertexColors = VertexColors;
						if (value === 'face') json.vertexColors = FaceColors;
						break;
					default:
						console.error('THREE.Loader.createMaterial: Unsupported', name, value);
						break;
				}
			}

			if (json.type === 'MeshBasicMaterial') delete json.emissive;
			if (json.type !== 'MeshPhongMaterial') delete json.specular;

			if (json.opacity < 1) json.transparent = true;

			materialLoader.setTextures(textures);

			return materialLoader.parse(json);
		};
	}()

};

Loader.Handlers = {

	handlers: [],

	add: function add(regex, loader) {

		this.handlers.push(regex, loader);
	},

	get: function get(file) {

		var handlers = this.handlers;

		for (var i = 0, l = handlers.length; i < l; i += 2) {

			var regex = handlers[i];
			var loader = handlers[i + 1];

			if (regex.test(file)) {

				return loader;
			}
		}

		return null;
	}

};

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 */

function JSONLoader(manager) {

	if (typeof manager === 'boolean') {

		console.warn('THREE.JSONLoader: showStatus parameter has been removed from constructor.');
		manager = undefined;
	}

	this.manager = manager !== undefined ? manager : DefaultLoadingManager;

	this.withCredentials = false;
}

Object.assign(JSONLoader.prototype, {

	load: function load(url, onLoad, onProgress, onError) {

		var scope = this;

		var texturePath = this.texturePath && typeof this.texturePath === "string" ? this.texturePath : Loader.prototype.extractUrlBase(url);

		var loader = new XHRLoader(this.manager);
		loader.setWithCredentials(this.withCredentials);
		loader.load(url, function (text) {

			var json = JSON.parse(text);
			var metadata = json.metadata;

			if (metadata !== undefined) {

				var type = metadata.type;

				if (type !== undefined) {

					if (type.toLowerCase() === 'object') {

						console.error('THREE.JSONLoader: ' + url + ' should be loaded with THREE.ObjectLoader instead.');
						return;
					}

					if (type.toLowerCase() === 'scene') {

						console.error('THREE.JSONLoader: ' + url + ' should be loaded with THREE.SceneLoader instead.');
						return;
					}
				}
			}

			var object = scope.parse(json, texturePath);
			onLoad(object.geometry, object.materials);
		}, onProgress, onError);
	},

	setTexturePath: function setTexturePath(value) {

		this.texturePath = value;
	},

	parse: function parse(json, texturePath) {

		var geometry = new Geometry(),
		    scale = json.scale !== undefined ? 1.0 / json.scale : 1.0;

		parseModel(scale);

		parseSkin();
		parseMorphing(scale);
		parseAnimations();

		geometry.computeFaceNormals();
		geometry.computeBoundingSphere();

		function parseModel(scale) {

			function isBitSet(value, position) {

				return value & 1 << position;
			}

			var i,
			    j,
			    fi,
			    offset,
			    zLength,
			    colorIndex,
			    normalIndex,
			    uvIndex,
			    materialIndex,
			    type,
			    isQuad,
			    hasMaterial,
			    hasFaceVertexUv,
			    hasFaceNormal,
			    hasFaceVertexNormal,
			    hasFaceColor,
			    hasFaceVertexColor,
			    vertex,
			    face,
			    faceA,
			    faceB,
			    hex,
			    normal,
			    uvLayer,
			    uv,
			    u,
			    v,
			    faces = json.faces,
			    vertices = json.vertices,
			    normals = json.normals,
			    colors = json.colors,
			    nUvLayers = 0;

			if (json.uvs !== undefined) {

				// disregard empty arrays

				for (i = 0; i < json.uvs.length; i++) {

					if (json.uvs[i].length) nUvLayers++;
				}

				for (i = 0; i < nUvLayers; i++) {

					geometry.faceVertexUvs[i] = [];
				}
			}

			offset = 0;
			zLength = vertices.length;

			while (offset < zLength) {

				vertex = new Vector3();

				vertex.x = vertices[offset++] * scale;
				vertex.y = vertices[offset++] * scale;
				vertex.z = vertices[offset++] * scale;

				geometry.vertices.push(vertex);
			}

			offset = 0;
			zLength = faces.length;

			while (offset < zLength) {

				type = faces[offset++];

				isQuad = isBitSet(type, 0);
				hasMaterial = isBitSet(type, 1);
				hasFaceVertexUv = isBitSet(type, 3);
				hasFaceNormal = isBitSet(type, 4);
				hasFaceVertexNormal = isBitSet(type, 5);
				hasFaceColor = isBitSet(type, 6);
				hasFaceVertexColor = isBitSet(type, 7);

				// console.log("type", type, "bits", isQuad, hasMaterial, hasFaceVertexUv, hasFaceNormal, hasFaceVertexNormal, hasFaceColor, hasFaceVertexColor);

				if (isQuad) {

					faceA = new Face3();
					faceA.a = faces[offset];
					faceA.b = faces[offset + 1];
					faceA.c = faces[offset + 3];

					faceB = new Face3();
					faceB.a = faces[offset + 1];
					faceB.b = faces[offset + 2];
					faceB.c = faces[offset + 3];

					offset += 4;

					if (hasMaterial) {

						materialIndex = faces[offset++];
						faceA.materialIndex = materialIndex;
						faceB.materialIndex = materialIndex;
					}

					// to get face <=> uv index correspondence

					fi = geometry.faces.length;

					if (hasFaceVertexUv) {

						for (i = 0; i < nUvLayers; i++) {

							uvLayer = json.uvs[i];

							geometry.faceVertexUvs[i][fi] = [];
							geometry.faceVertexUvs[i][fi + 1] = [];

							for (j = 0; j < 4; j++) {

								uvIndex = faces[offset++];

								u = uvLayer[uvIndex * 2];
								v = uvLayer[uvIndex * 2 + 1];

								uv = new Vector2(u, v);

								if (j !== 2) geometry.faceVertexUvs[i][fi].push(uv);
								if (j !== 0) geometry.faceVertexUvs[i][fi + 1].push(uv);
							}
						}
					}

					if (hasFaceNormal) {

						normalIndex = faces[offset++] * 3;

						faceA.normal.set(normals[normalIndex++], normals[normalIndex++], normals[normalIndex]);

						faceB.normal.copy(faceA.normal);
					}

					if (hasFaceVertexNormal) {

						for (i = 0; i < 4; i++) {

							normalIndex = faces[offset++] * 3;

							normal = new Vector3(normals[normalIndex++], normals[normalIndex++], normals[normalIndex]);

							if (i !== 2) faceA.vertexNormals.push(normal);
							if (i !== 0) faceB.vertexNormals.push(normal);
						}
					}

					if (hasFaceColor) {

						colorIndex = faces[offset++];
						hex = colors[colorIndex];

						faceA.color.setHex(hex);
						faceB.color.setHex(hex);
					}

					if (hasFaceVertexColor) {

						for (i = 0; i < 4; i++) {

							colorIndex = faces[offset++];
							hex = colors[colorIndex];

							if (i !== 2) faceA.vertexColors.push(new Color(hex));
							if (i !== 0) faceB.vertexColors.push(new Color(hex));
						}
					}

					geometry.faces.push(faceA);
					geometry.faces.push(faceB);
				} else {

					face = new Face3();
					face.a = faces[offset++];
					face.b = faces[offset++];
					face.c = faces[offset++];

					if (hasMaterial) {

						materialIndex = faces[offset++];
						face.materialIndex = materialIndex;
					}

					// to get face <=> uv index correspondence

					fi = geometry.faces.length;

					if (hasFaceVertexUv) {

						for (i = 0; i < nUvLayers; i++) {

							uvLayer = json.uvs[i];

							geometry.faceVertexUvs[i][fi] = [];

							for (j = 0; j < 3; j++) {

								uvIndex = faces[offset++];

								u = uvLayer[uvIndex * 2];
								v = uvLayer[uvIndex * 2 + 1];

								uv = new Vector2(u, v);

								geometry.faceVertexUvs[i][fi].push(uv);
							}
						}
					}

					if (hasFaceNormal) {

						normalIndex = faces[offset++] * 3;

						face.normal.set(normals[normalIndex++], normals[normalIndex++], normals[normalIndex]);
					}

					if (hasFaceVertexNormal) {

						for (i = 0; i < 3; i++) {

							normalIndex = faces[offset++] * 3;

							normal = new Vector3(normals[normalIndex++], normals[normalIndex++], normals[normalIndex]);

							face.vertexNormals.push(normal);
						}
					}

					if (hasFaceColor) {

						colorIndex = faces[offset++];
						face.color.setHex(colors[colorIndex]);
					}

					if (hasFaceVertexColor) {

						for (i = 0; i < 3; i++) {

							colorIndex = faces[offset++];
							face.vertexColors.push(new Color(colors[colorIndex]));
						}
					}

					geometry.faces.push(face);
				}
			}
		}

		function parseSkin() {

			var influencesPerVertex = json.influencesPerVertex !== undefined ? json.influencesPerVertex : 2;

			if (json.skinWeights) {

				for (var i = 0, l = json.skinWeights.length; i < l; i += influencesPerVertex) {

					var x = json.skinWeights[i];
					var y = influencesPerVertex > 1 ? json.skinWeights[i + 1] : 0;
					var z = influencesPerVertex > 2 ? json.skinWeights[i + 2] : 0;
					var w = influencesPerVertex > 3 ? json.skinWeights[i + 3] : 0;

					geometry.skinWeights.push(new Vector4(x, y, z, w));
				}
			}

			if (json.skinIndices) {

				for (var i = 0, l = json.skinIndices.length; i < l; i += influencesPerVertex) {

					var a = json.skinIndices[i];
					var b = influencesPerVertex > 1 ? json.skinIndices[i + 1] : 0;
					var c = influencesPerVertex > 2 ? json.skinIndices[i + 2] : 0;
					var d = influencesPerVertex > 3 ? json.skinIndices[i + 3] : 0;

					geometry.skinIndices.push(new Vector4(a, b, c, d));
				}
			}

			geometry.bones = json.bones;

			if (geometry.bones && geometry.bones.length > 0 && (geometry.skinWeights.length !== geometry.skinIndices.length || geometry.skinIndices.length !== geometry.vertices.length)) {

				console.warn('When skinning, number of vertices (' + geometry.vertices.length + '), skinIndices (' + geometry.skinIndices.length + '), and skinWeights (' + geometry.skinWeights.length + ') should match.');
			}
		}

		function parseMorphing(scale) {

			if (json.morphTargets !== undefined) {

				for (var i = 0, l = json.morphTargets.length; i < l; i++) {

					geometry.morphTargets[i] = {};
					geometry.morphTargets[i].name = json.morphTargets[i].name;
					geometry.morphTargets[i].vertices = [];

					var dstVertices = geometry.morphTargets[i].vertices;
					var srcVertices = json.morphTargets[i].vertices;

					for (var v = 0, vl = srcVertices.length; v < vl; v += 3) {

						var vertex = new Vector3();
						vertex.x = srcVertices[v] * scale;
						vertex.y = srcVertices[v + 1] * scale;
						vertex.z = srcVertices[v + 2] * scale;

						dstVertices.push(vertex);
					}
				}
			}

			if (json.morphColors !== undefined && json.morphColors.length > 0) {

				console.warn('THREE.JSONLoader: "morphColors" no longer supported. Using them as face colors.');

				var faces = geometry.faces;
				var morphColors = json.morphColors[0].colors;

				for (var i = 0, l = faces.length; i < l; i++) {

					faces[i].color.fromArray(morphColors, i * 3);
				}
			}
		}

		function parseAnimations() {

			var outputAnimations = [];

			// parse old style Bone/Hierarchy animations
			var animations = [];

			if (json.animation !== undefined) {

				animations.push(json.animation);
			}

			if (json.animations !== undefined) {

				if (json.animations.length) {

					animations = animations.concat(json.animations);
				} else {

					animations.push(json.animations);
				}
			}

			for (var i = 0; i < animations.length; i++) {

				var clip = AnimationClip.parseAnimation(animations[i], geometry.bones);
				if (clip) outputAnimations.push(clip);
			}

			// parse implicit morph animations
			if (geometry.morphTargets) {

				// TODO: Figure out what an appropraite FPS is for morph target animations -- defaulting to 10, but really it is completely arbitrary.
				var morphAnimationClips = AnimationClip.CreateClipsFromMorphTargetSequences(geometry.morphTargets, 10);
				outputAnimations = outputAnimations.concat(morphAnimationClips);
			}

			if (outputAnimations.length > 0) geometry.animations = outputAnimations;
		}

		if (json.materials === undefined || json.materials.length === 0) {

			return { geometry: geometry };
		} else {

			var materials = Loader.prototype.initMaterials(json.materials, texturePath, this.crossOrigin);

			return { geometry: geometry, materials: materials };
		}
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WireframeGeometry(geometry) {

	BufferGeometry.call(this);

	var edge = [0, 0],
	    hash = {};

	function sortFunction(a, b) {

		return a - b;
	}

	var keys = ['a', 'b', 'c'];

	if (geometry && geometry.isGeometry) {

		var vertices = geometry.vertices;
		var faces = geometry.faces;
		var numEdges = 0;

		// allocate maximal size
		var edges = new Uint32Array(6 * faces.length);

		for (var i = 0, l = faces.length; i < l; i++) {

			var face = faces[i];

			for (var j = 0; j < 3; j++) {

				edge[0] = face[keys[j]];
				edge[1] = face[keys[(j + 1) % 3]];
				edge.sort(sortFunction);

				var key = edge.toString();

				if (hash[key] === undefined) {

					edges[2 * numEdges] = edge[0];
					edges[2 * numEdges + 1] = edge[1];
					hash[key] = true;
					numEdges++;
				}
			}
		}

		var coords = new Float32Array(numEdges * 2 * 3);

		for (var i = 0, l = numEdges; i < l; i++) {

			for (var j = 0; j < 2; j++) {

				var vertex = vertices[edges[2 * i + j]];

				var index = 6 * i + 3 * j;
				coords[index + 0] = vertex.x;
				coords[index + 1] = vertex.y;
				coords[index + 2] = vertex.z;
			}
		}

		this.addAttribute('position', new BufferAttribute(coords, 3));
	} else if (geometry && geometry.isBufferGeometry) {

		if (geometry.index !== null) {

			// Indexed BufferGeometry

			var indices = geometry.index.array;
			var vertices = geometry.attributes.position;
			var groups = geometry.groups;
			var numEdges = 0;

			if (groups.length === 0) {

				geometry.addGroup(0, indices.length);
			}

			// allocate maximal size
			var edges = new Uint32Array(2 * indices.length);

			for (var o = 0, ol = groups.length; o < ol; ++o) {

				var group = groups[o];

				var start = group.start;
				var count = group.count;

				for (var i = start, il = start + count; i < il; i += 3) {

					for (var j = 0; j < 3; j++) {

						edge[0] = indices[i + j];
						edge[1] = indices[i + (j + 1) % 3];
						edge.sort(sortFunction);

						var key = edge.toString();

						if (hash[key] === undefined) {

							edges[2 * numEdges] = edge[0];
							edges[2 * numEdges + 1] = edge[1];
							hash[key] = true;
							numEdges++;
						}
					}
				}
			}

			var coords = new Float32Array(numEdges * 2 * 3);

			for (var i = 0, l = numEdges; i < l; i++) {

				for (var j = 0; j < 2; j++) {

					var index = 6 * i + 3 * j;
					var index2 = edges[2 * i + j];

					coords[index + 0] = vertices.getX(index2);
					coords[index + 1] = vertices.getY(index2);
					coords[index + 2] = vertices.getZ(index2);
				}
			}

			this.addAttribute('position', new BufferAttribute(coords, 3));
		} else {

			// non-indexed BufferGeometry

			var vertices = geometry.attributes.position.array;
			var numEdges = vertices.length / 3;
			var numTris = numEdges / 3;

			var coords = new Float32Array(numEdges * 2 * 3);

			for (var i = 0, l = numTris; i < l; i++) {

				for (var j = 0; j < 3; j++) {

					var index = 18 * i + 6 * j;

					var index1 = 9 * i + 3 * j;
					coords[index + 0] = vertices[index1];
					coords[index + 1] = vertices[index1 + 1];
					coords[index + 2] = vertices[index1 + 2];

					var index2 = 9 * i + 3 * ((j + 1) % 3);
					coords[index + 3] = vertices[index2];
					coords[index + 4] = vertices[index2 + 1];
					coords[index + 5] = vertices[index2 + 2];
				}
			}

			this.addAttribute('position', new BufferAttribute(coords, 3));
		}
	}
}

WireframeGeometry.prototype = Object.create(BufferGeometry.prototype);
WireframeGeometry.prototype.constructor = WireframeGeometry;

/**
 * @author Mugen87 / https://github.com/Mugen87
 *
 * Parametric Surfaces Geometry
 * based on the brilliant article by @prideout http://prideout.net/blog/?p=44
 */

function ParametricBufferGeometry(func, slices, stacks) {

	BufferGeometry.call(this);

	this.type = 'ParametricBufferGeometry';

	this.parameters = {
		func: func,
		slices: slices,
		stacks: stacks
	};

	// generate vertices and uvs

	var vertices = [];
	var uvs = [];

	var i, j, p;
	var u, v;

	var sliceCount = slices + 1;

	for (i = 0; i <= stacks; i++) {

		v = i / stacks;

		for (j = 0; j <= slices; j++) {

			u = j / slices;

			p = func(u, v);
			vertices.push(p.x, p.y, p.z);

			uvs.push(u, v);
		}
	}

	// generate indices

	var indices = [];
	var a, b, c, d;

	for (i = 0; i < stacks; i++) {

		for (j = 0; j < slices; j++) {

			a = i * sliceCount + j;
			b = i * sliceCount + j + 1;
			c = (i + 1) * sliceCount + j + 1;
			d = (i + 1) * sliceCount + j;

			// faces one and two

			indices.push(a, b, d);
			indices.push(b, c, d);
		}
	}

	// build geometry

	this.setIndex((indices.length > 65535 ? Uint32Attribute : Uint16Attribute)(indices, 1));
	this.addAttribute('position', Float32Attribute(vertices, 3));
	this.addAttribute('uv', Float32Attribute(uvs, 2));

	// generate normals

	this.computeVertexNormals();
}

ParametricBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
ParametricBufferGeometry.prototype.constructor = ParametricBufferGeometry;

/**
 * @author zz85 / https://github.com/zz85
 *
 * Parametric Surfaces Geometry
 * based on the brilliant article by @prideout http://prideout.net/blog/?p=44
 */

function ParametricGeometry(func, slices, stacks) {

	Geometry.call(this);

	this.type = 'ParametricGeometry';

	this.parameters = {
		func: func,
		slices: slices,
		stacks: stacks
	};

	this.fromBufferGeometry(new ParametricBufferGeometry(func, slices, stacks));
	this.mergeVertices();
}

ParametricGeometry.prototype = Object.create(Geometry.prototype);
ParametricGeometry.prototype.constructor = ParametricGeometry;

/**
 * @author Mugen87 / https://github.com/Mugen87
 */

function PolyhedronBufferGeometry(vertices, indices, radius, detail) {

	BufferGeometry.call(this);

	this.type = 'PolyhedronBufferGeometry';

	this.parameters = {
		vertices: vertices,
		indices: indices,
		radius: radius,
		detail: detail
	};

	radius = radius || 1;
	detail = detail || 0;

	// default buffer data

	var vertexBuffer = [];
	var uvBuffer = [];

	// the subdivision creates the vertex buffer data

	subdivide(detail);

	// all vertices should lie on a conceptual sphere with a given radius

	appplyRadius(radius);

	// finally, create the uv data

	generateUVs();

	// build non-indexed geometry

	this.addAttribute('position', Float32Attribute(vertexBuffer, 3));
	this.addAttribute('normal', Float32Attribute(vertexBuffer.slice(), 3));
	this.addAttribute('uv', Float32Attribute(uvBuffer, 2));
	this.normalizeNormals();

	this.boundingSphere = new Sphere(new Vector3(), radius);

	// helper functions

	function subdivide(detail) {

		var a = new Vector3();
		var b = new Vector3();
		var c = new Vector3();

		// iterate over all faces and apply a subdivison with the given detail value

		for (var i = 0; i < indices.length; i += 3) {

			// get the vertices of the face

			getVertexByIndex(indices[i + 0], a);
			getVertexByIndex(indices[i + 1], b);
			getVertexByIndex(indices[i + 2], c);

			// perform subdivision

			subdivideFace(a, b, c, detail);
		}
	}

	function subdivideFace(a, b, c, detail) {

		var cols = Math.pow(2, detail);

		// we use this multidimensional array as a data structure for creating the subdivision

		var v = [];

		var i, j;

		// construct all of the vertices for this subdivision

		for (i = 0; i <= cols; i++) {

			v[i] = [];

			var aj = a.clone().lerp(c, i / cols);
			var bj = b.clone().lerp(c, i / cols);

			var rows = cols - i;

			for (j = 0; j <= rows; j++) {

				if (j === 0 && i === cols) {

					v[i][j] = aj;
				} else {

					v[i][j] = aj.clone().lerp(bj, j / rows);
				}
			}
		}

		// construct all of the faces

		for (i = 0; i < cols; i++) {

			for (j = 0; j < 2 * (cols - i) - 1; j++) {

				var k = Math.floor(j / 2);

				if (j % 2 === 0) {

					pushVertex(v[i][k + 1]);
					pushVertex(v[i + 1][k]);
					pushVertex(v[i][k]);
				} else {

					pushVertex(v[i][k + 1]);
					pushVertex(v[i + 1][k + 1]);
					pushVertex(v[i + 1][k]);
				}
			}
		}
	}

	function appplyRadius(radius) {

		var vertex = new Vector3();

		// iterate over the entire buffer and apply the radius to each vertex

		for (var i = 0; i < vertexBuffer.length; i += 3) {

			vertex.x = vertexBuffer[i + 0];
			vertex.y = vertexBuffer[i + 1];
			vertex.z = vertexBuffer[i + 2];

			vertex.normalize().multiplyScalar(radius);

			vertexBuffer[i + 0] = vertex.x;
			vertexBuffer[i + 1] = vertex.y;
			vertexBuffer[i + 2] = vertex.z;
		}
	}

	function generateUVs() {

		var vertex = new Vector3();

		for (var i = 0; i < vertexBuffer.length; i += 3) {

			vertex.x = vertexBuffer[i + 0];
			vertex.y = vertexBuffer[i + 1];
			vertex.z = vertexBuffer[i + 2];

			var u = azimuth(vertex) / 2 / Math.PI + 0.5;
			var v = inclination(vertex) / Math.PI + 0.5;
			uvBuffer.push(u, 1 - v);
		}

		correctUVs();

		correctSeam();
	}

	function correctSeam() {

		// handle case when face straddles the seam, see #3269

		for (var i = 0; i < uvBuffer.length; i += 6) {

			// uv data of a single face

			var x0 = uvBuffer[i + 0];
			var x1 = uvBuffer[i + 2];
			var x2 = uvBuffer[i + 4];

			var max = Math.max(x0, x1, x2);
			var min = Math.min(x0, x1, x2);

			// 0.9 is somewhat arbitrary

			if (max > 0.9 && min < 0.1) {

				if (x0 < 0.2) uvBuffer[i + 0] += 1;
				if (x1 < 0.2) uvBuffer[i + 2] += 1;
				if (x2 < 0.2) uvBuffer[i + 4] += 1;
			}
		}
	}

	function pushVertex(vertex) {

		vertexBuffer.push(vertex.x, vertex.y, vertex.z);
	}

	function getVertexByIndex(index, vertex) {

		var stride = index * 3;

		vertex.x = vertices[stride + 0];
		vertex.y = vertices[stride + 1];
		vertex.z = vertices[stride + 2];
	}

	function correctUVs() {

		var a = new Vector3();
		var b = new Vector3();
		var c = new Vector3();

		var centroid = new Vector3();

		var uvA = new Vector2();
		var uvB = new Vector2();
		var uvC = new Vector2();

		for (var i = 0, j = 0; i < vertexBuffer.length; i += 9, j += 6) {

			a.set(vertexBuffer[i + 0], vertexBuffer[i + 1], vertexBuffer[i + 2]);
			b.set(vertexBuffer[i + 3], vertexBuffer[i + 4], vertexBuffer[i + 5]);
			c.set(vertexBuffer[i + 6], vertexBuffer[i + 7], vertexBuffer[i + 8]);

			uvA.set(uvBuffer[j + 0], uvBuffer[j + 1]);
			uvB.set(uvBuffer[j + 2], uvBuffer[j + 3]);
			uvC.set(uvBuffer[j + 4], uvBuffer[j + 5]);

			centroid.copy(a).add(b).add(c).divideScalar(3);

			var azi = azimuth(centroid);

			correctUV(uvA, j + 0, a, azi);
			correctUV(uvB, j + 2, b, azi);
			correctUV(uvC, j + 4, c, azi);
		}
	}

	function correctUV(uv, stride, vector, azimuth) {

		if (azimuth < 0 && uv.x === 1) {

			uvBuffer[stride] = uv.x - 1;
		}

		if (vector.x === 0 && vector.z === 0) {

			uvBuffer[stride] = azimuth / 2 / Math.PI + 0.5;
		}
	}

	// Angle around the Y axis, counter-clockwise when looking from above.

	function azimuth(vector) {

		return Math.atan2(vector.z, -vector.x);
	}

	// Angle above the XZ plane.

	function inclination(vector) {

		return Math.atan2(-vector.y, Math.sqrt(vector.x * vector.x + vector.z * vector.z));
	}
}

PolyhedronBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
PolyhedronBufferGeometry.prototype.constructor = PolyhedronBufferGeometry;

/**
 * @author Mugen87 / https://github.com/Mugen87
 */

function TetrahedronBufferGeometry(radius, detail) {

	var vertices = [1, 1, 1, -1, -1, 1, -1, 1, -1, 1, -1, -1];

	var indices = [2, 1, 0, 0, 3, 2, 1, 3, 0, 2, 3, 1];

	PolyhedronBufferGeometry.call(this, vertices, indices, radius, detail);

	this.type = 'TetrahedronBufferGeometry';

	this.parameters = {
		radius: radius,
		detail: detail
	};
}

TetrahedronBufferGeometry.prototype = Object.create(PolyhedronBufferGeometry.prototype);
TetrahedronBufferGeometry.prototype.constructor = TetrahedronBufferGeometry;

/**
 * @author timothypratley / https://github.com/timothypratley
 */

function TetrahedronGeometry(radius, detail) {

	Geometry.call(this);

	this.type = 'TetrahedronGeometry';

	this.parameters = {
		radius: radius,
		detail: detail
	};

	this.fromBufferGeometry(new TetrahedronBufferGeometry(radius, detail));
	this.mergeVertices();
}

TetrahedronGeometry.prototype = Object.create(Geometry.prototype);
TetrahedronGeometry.prototype.constructor = TetrahedronGeometry;

/**
 * @author Mugen87 / https://github.com/Mugen87
 */

function OctahedronBufferGeometry(radius, detail) {

	var vertices = [1, 0, 0, -1, 0, 0, 0, 1, 0, 0, -1, 0, 0, 0, 1, 0, 0, -1];

	var indices = [0, 2, 4, 0, 4, 3, 0, 3, 5, 0, 5, 2, 1, 2, 5, 1, 5, 3, 1, 3, 4, 1, 4, 2];

	PolyhedronBufferGeometry.call(this, vertices, indices, radius, detail);

	this.type = 'OctahedronBufferGeometry';

	this.parameters = {
		radius: radius,
		detail: detail
	};
}

OctahedronBufferGeometry.prototype = Object.create(PolyhedronBufferGeometry.prototype);
OctahedronBufferGeometry.prototype.constructor = OctahedronBufferGeometry;

/**
 * @author timothypratley / https://github.com/timothypratley
 */

function OctahedronGeometry(radius, detail) {

	Geometry.call(this);

	this.type = 'OctahedronGeometry';

	this.parameters = {
		radius: radius,
		detail: detail
	};

	this.fromBufferGeometry(new OctahedronBufferGeometry(radius, detail));
	this.mergeVertices();
}

OctahedronGeometry.prototype = Object.create(Geometry.prototype);
OctahedronGeometry.prototype.constructor = OctahedronGeometry;

/**
 * @author Mugen87 / https://github.com/Mugen87
 */

function IcosahedronBufferGeometry(radius, detail) {

	var t = (1 + Math.sqrt(5)) / 2;

	var vertices = [-1, t, 0, 1, t, 0, -1, -t, 0, 1, -t, 0, 0, -1, t, 0, 1, t, 0, -1, -t, 0, 1, -t, t, 0, -1, t, 0, 1, -t, 0, -1, -t, 0, 1];

	var indices = [0, 11, 5, 0, 5, 1, 0, 1, 7, 0, 7, 10, 0, 10, 11, 1, 5, 9, 5, 11, 4, 11, 10, 2, 10, 7, 6, 7, 1, 8, 3, 9, 4, 3, 4, 2, 3, 2, 6, 3, 6, 8, 3, 8, 9, 4, 9, 5, 2, 4, 11, 6, 2, 10, 8, 6, 7, 9, 8, 1];

	PolyhedronBufferGeometry.call(this, vertices, indices, radius, detail);

	this.type = 'IcosahedronBufferGeometry';

	this.parameters = {
		radius: radius,
		detail: detail
	};
}

IcosahedronBufferGeometry.prototype = Object.create(PolyhedronBufferGeometry.prototype);
IcosahedronBufferGeometry.prototype.constructor = IcosahedronBufferGeometry;

/**
 * @author timothypratley / https://github.com/timothypratley
 */

function IcosahedronGeometry(radius, detail) {

	Geometry.call(this);

	this.type = 'IcosahedronGeometry';

	this.parameters = {
		radius: radius,
		detail: detail
	};

	this.fromBufferGeometry(new IcosahedronBufferGeometry(radius, detail));
	this.mergeVertices();
}

IcosahedronGeometry.prototype = Object.create(Geometry.prototype);
IcosahedronGeometry.prototype.constructor = IcosahedronGeometry;

/**
 * @author Mugen87 / https://github.com/Mugen87
 */

function DodecahedronBufferGeometry(radius, detail) {

	var t = (1 + Math.sqrt(5)) / 2;
	var r = 1 / t;

	var vertices = [

	// (±1, ±1, ±1)
	-1, -1, -1, -1, -1, 1, -1, 1, -1, -1, 1, 1, 1, -1, -1, 1, -1, 1, 1, 1, -1, 1, 1, 1,

	// (0, ±1/φ, ±φ)
	0, -r, -t, 0, -r, t, 0, r, -t, 0, r, t,

	// (±1/φ, ±φ, 0)
	-r, -t, 0, -r, t, 0, r, -t, 0, r, t, 0,

	// (±φ, 0, ±1/φ)
	-t, 0, -r, t, 0, -r, -t, 0, r, t, 0, r];

	var indices = [3, 11, 7, 3, 7, 15, 3, 15, 13, 7, 19, 17, 7, 17, 6, 7, 6, 15, 17, 4, 8, 17, 8, 10, 17, 10, 6, 8, 0, 16, 8, 16, 2, 8, 2, 10, 0, 12, 1, 0, 1, 18, 0, 18, 16, 6, 10, 2, 6, 2, 13, 6, 13, 15, 2, 16, 18, 2, 18, 3, 2, 3, 13, 18, 1, 9, 18, 9, 11, 18, 11, 3, 4, 14, 12, 4, 12, 0, 4, 0, 8, 11, 9, 5, 11, 5, 19, 11, 19, 7, 19, 5, 14, 19, 14, 4, 19, 4, 17, 1, 12, 14, 1, 14, 5, 1, 5, 9];

	PolyhedronBufferGeometry.call(this, vertices, indices, radius, detail);

	this.type = 'DodecahedronBufferGeometry';

	this.parameters = {
		radius: radius,
		detail: detail
	};
}

DodecahedronBufferGeometry.prototype = Object.create(PolyhedronBufferGeometry.prototype);
DodecahedronBufferGeometry.prototype.constructor = DodecahedronBufferGeometry;

/**
 * @author Abe Pazos / https://hamoid.com
 */

function DodecahedronGeometry(radius, detail) {

	Geometry.call(this);

	this.type = 'DodecahedronGeometry';

	this.parameters = {
		radius: radius,
		detail: detail
	};

	this.fromBufferGeometry(new DodecahedronBufferGeometry(radius, detail));
	this.mergeVertices();
}

DodecahedronGeometry.prototype = Object.create(Geometry.prototype);
DodecahedronGeometry.prototype.constructor = DodecahedronGeometry;

/**
 * @author clockworkgeek / https://github.com/clockworkgeek
 * @author timothypratley / https://github.com/timothypratley
 * @author WestLangley / http://github.com/WestLangley
*/

function PolyhedronGeometry(vertices, indices, radius, detail) {

	Geometry.call(this);

	this.type = 'PolyhedronGeometry';

	this.parameters = {
		vertices: vertices,
		indices: indices,
		radius: radius,
		detail: detail
	};

	this.fromBufferGeometry(new PolyhedronBufferGeometry(vertices, indices, radius, detail));
	this.mergeVertices();
}

PolyhedronGeometry.prototype = Object.create(Geometry.prototype);
PolyhedronGeometry.prototype.constructor = PolyhedronGeometry;

/**
 * @author Mugen87 / https://github.com/Mugen87
 *
 * Creates a tube which extrudes along a 3d spline.
 *
 */

function TubeBufferGeometry(path, tubularSegments, radius, radialSegments, closed) {

	BufferGeometry.call(this);

	this.type = 'TubeBufferGeometry';

	this.parameters = {
		path: path,
		tubularSegments: tubularSegments,
		radius: radius,
		radialSegments: radialSegments,
		closed: closed
	};

	tubularSegments = tubularSegments || 64;
	radius = radius || 1;
	radialSegments = radialSegments || 8;
	closed = closed || false;

	var frames = path.computeFrenetFrames(tubularSegments, closed);

	// expose internals

	this.tangents = frames.tangents;
	this.normals = frames.normals;
	this.binormals = frames.binormals;

	// helper variables

	var vertex = new Vector3();
	var normal = new Vector3();
	var uv = new Vector2();

	var i, j;

	// buffer

	var vertices = [];
	var normals = [];
	var uvs = [];
	var indices = [];

	// create buffer data

	generateBufferData();

	// build geometry

	this.setIndex((indices.length > 65535 ? Uint32Attribute : Uint16Attribute)(indices, 1));
	this.addAttribute('position', Float32Attribute(vertices, 3));
	this.addAttribute('normal', Float32Attribute(normals, 3));
	this.addAttribute('uv', Float32Attribute(uvs, 2));

	// functions

	function generateBufferData() {

		for (i = 0; i < tubularSegments; i++) {

			generateSegment(i);
		}

		// if the geometry is not closed, generate the last row of vertices and normals
		// at the regular position on the given path
		//
		// if the geometry is closed, duplicate the first row of vertices and normals (uvs will differ)

		generateSegment(closed === false ? tubularSegments : 0);

		// uvs are generated in a separate function.
		// this makes it easy compute correct values for closed geometries

		generateUVs();

		// finally create faces

		generateIndices();
	}

	function generateSegment(i) {

		// we use getPointAt to sample evenly distributed points from the given path

		var P = path.getPointAt(i / tubularSegments);

		// retrieve corresponding normal and binormal

		var N = frames.normals[i];
		var B = frames.binormals[i];

		// generate normals and vertices for the current segment

		for (j = 0; j <= radialSegments; j++) {

			var v = j / radialSegments * Math.PI * 2;

			var sin = Math.sin(v);
			var cos = -Math.cos(v);

			// normal

			normal.x = cos * N.x + sin * B.x;
			normal.y = cos * N.y + sin * B.y;
			normal.z = cos * N.z + sin * B.z;
			normal.normalize();

			normals.push(normal.x, normal.y, normal.z);

			// vertex

			vertex.x = P.x + radius * normal.x;
			vertex.y = P.y + radius * normal.y;
			vertex.z = P.z + radius * normal.z;

			vertices.push(vertex.x, vertex.y, vertex.z);
		}
	}

	function generateIndices() {

		for (j = 1; j <= tubularSegments; j++) {

			for (i = 1; i <= radialSegments; i++) {

				var a = (radialSegments + 1) * (j - 1) + (i - 1);
				var b = (radialSegments + 1) * j + (i - 1);
				var c = (radialSegments + 1) * j + i;
				var d = (radialSegments + 1) * (j - 1) + i;

				// faces

				indices.push(a, b, d);
				indices.push(b, c, d);
			}
		}
	}

	function generateUVs() {

		for (i = 0; i <= tubularSegments; i++) {

			for (j = 0; j <= radialSegments; j++) {

				uv.x = i / tubularSegments;
				uv.y = j / radialSegments;

				uvs.push(uv.x, uv.y);
			}
		}
	}
}

TubeBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
TubeBufferGeometry.prototype.constructor = TubeBufferGeometry;

/**
 * @author oosmoxiecode / https://github.com/oosmoxiecode
 * @author WestLangley / https://github.com/WestLangley
 * @author zz85 / https://github.com/zz85
 * @author miningold / https://github.com/miningold
 * @author jonobr1 / https://github.com/jonobr1
 *
 * Creates a tube which extrudes along a 3d spline.
 */

function TubeGeometry(path, tubularSegments, radius, radialSegments, closed, taper) {

	Geometry.call(this);

	this.type = 'TubeGeometry';

	this.parameters = {
		path: path,
		tubularSegments: tubularSegments,
		radius: radius,
		radialSegments: radialSegments,
		closed: closed
	};

	if (taper !== undefined) console.warn('THREE.TubeGeometry: taper has been removed.');

	var bufferGeometry = new TubeBufferGeometry(path, tubularSegments, radius, radialSegments, closed);

	// expose internals

	this.tangents = bufferGeometry.tangents;
	this.normals = bufferGeometry.normals;
	this.binormals = bufferGeometry.binormals;

	// create geometry

	this.fromBufferGeometry(bufferGeometry);
	this.mergeVertices();
}

TubeGeometry.prototype = Object.create(Geometry.prototype);
TubeGeometry.prototype.constructor = TubeGeometry;

/**
 * @author Mugen87 / https://github.com/Mugen87
 *
 * see: http://www.blackpawn.com/texts/pqtorus/
 */
function TorusKnotBufferGeometry(radius, tube, tubularSegments, radialSegments, p, q) {

	BufferGeometry.call(this);

	this.type = 'TorusKnotBufferGeometry';

	this.parameters = {
		radius: radius,
		tube: tube,
		tubularSegments: tubularSegments,
		radialSegments: radialSegments,
		p: p,
		q: q
	};

	radius = radius || 100;
	tube = tube || 40;
	tubularSegments = Math.floor(tubularSegments) || 64;
	radialSegments = Math.floor(radialSegments) || 8;
	p = p || 2;
	q = q || 3;

	// used to calculate buffer length
	var vertexCount = (radialSegments + 1) * (tubularSegments + 1);
	var indexCount = radialSegments * tubularSegments * 2 * 3;

	// buffers
	var indices = new BufferAttribute(new (indexCount > 65535 ? Uint32Array : Uint16Array)(indexCount), 1);
	var vertices = new BufferAttribute(new Float32Array(vertexCount * 3), 3);
	var normals = new BufferAttribute(new Float32Array(vertexCount * 3), 3);
	var uvs = new BufferAttribute(new Float32Array(vertexCount * 2), 2);

	// helper variables
	var i,
	    j,
	    index = 0,
	    indexOffset = 0;

	var vertex = new Vector3();
	var normal = new Vector3();
	var uv = new Vector2();

	var P1 = new Vector3();
	var P2 = new Vector3();

	var B = new Vector3();
	var T = new Vector3();
	var N = new Vector3();

	// generate vertices, normals and uvs

	for (i = 0; i <= tubularSegments; ++i) {

		// the radian "u" is used to calculate the position on the torus curve of the current tubular segement

		var u = i / tubularSegments * p * Math.PI * 2;

		// now we calculate two points. P1 is our current position on the curve, P2 is a little farther ahead.
		// these points are used to create a special "coordinate space", which is necessary to calculate the correct vertex positions

		calculatePositionOnCurve(u, p, q, radius, P1);
		calculatePositionOnCurve(u + 0.01, p, q, radius, P2);

		// calculate orthonormal basis

		T.subVectors(P2, P1);
		N.addVectors(P2, P1);
		B.crossVectors(T, N);
		N.crossVectors(B, T);

		// normalize B, N. T can be ignored, we don't use it

		B.normalize();
		N.normalize();

		for (j = 0; j <= radialSegments; ++j) {

			// now calculate the vertices. they are nothing more than an extrusion of the torus curve.
			// because we extrude a shape in the xy-plane, there is no need to calculate a z-value.

			var v = j / radialSegments * Math.PI * 2;
			var cx = -tube * Math.cos(v);
			var cy = tube * Math.sin(v);

			// now calculate the final vertex position.
			// first we orient the extrusion with our basis vectos, then we add it to the current position on the curve

			vertex.x = P1.x + (cx * N.x + cy * B.x);
			vertex.y = P1.y + (cx * N.y + cy * B.y);
			vertex.z = P1.z + (cx * N.z + cy * B.z);

			// vertex
			vertices.setXYZ(index, vertex.x, vertex.y, vertex.z);

			// normal (P1 is always the center/origin of the extrusion, thus we can use it to calculate the normal)
			normal.subVectors(vertex, P1).normalize();
			normals.setXYZ(index, normal.x, normal.y, normal.z);

			// uv
			uv.x = i / tubularSegments;
			uv.y = j / radialSegments;
			uvs.setXY(index, uv.x, uv.y);

			// increase index
			index++;
		}
	}

	// generate indices

	for (j = 1; j <= tubularSegments; j++) {

		for (i = 1; i <= radialSegments; i++) {

			// indices
			var a = (radialSegments + 1) * (j - 1) + (i - 1);
			var b = (radialSegments + 1) * j + (i - 1);
			var c = (radialSegments + 1) * j + i;
			var d = (radialSegments + 1) * (j - 1) + i;

			// face one
			indices.setX(indexOffset, a);indexOffset++;
			indices.setX(indexOffset, b);indexOffset++;
			indices.setX(indexOffset, d);indexOffset++;

			// face two
			indices.setX(indexOffset, b);indexOffset++;
			indices.setX(indexOffset, c);indexOffset++;
			indices.setX(indexOffset, d);indexOffset++;
		}
	}

	// build geometry

	this.setIndex(indices);
	this.addAttribute('position', vertices);
	this.addAttribute('normal', normals);
	this.addAttribute('uv', uvs);

	// this function calculates the current position on the torus curve

	function calculatePositionOnCurve(u, p, q, radius, position) {

		var cu = Math.cos(u);
		var su = Math.sin(u);
		var quOverP = q / p * u;
		var cs = Math.cos(quOverP);

		position.x = radius * (2 + cs) * 0.5 * cu;
		position.y = radius * (2 + cs) * su * 0.5;
		position.z = radius * Math.sin(quOverP) * 0.5;
	}
}

TorusKnotBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
TorusKnotBufferGeometry.prototype.constructor = TorusKnotBufferGeometry;

/**
 * @author oosmoxiecode
 */

function TorusKnotGeometry(radius, tube, tubularSegments, radialSegments, p, q, heightScale) {

	Geometry.call(this);

	this.type = 'TorusKnotGeometry';

	this.parameters = {
		radius: radius,
		tube: tube,
		tubularSegments: tubularSegments,
		radialSegments: radialSegments,
		p: p,
		q: q
	};

	if (heightScale !== undefined) console.warn('THREE.TorusKnotGeometry: heightScale has been deprecated. Use .scale( x, y, z ) instead.');

	this.fromBufferGeometry(new TorusKnotBufferGeometry(radius, tube, tubularSegments, radialSegments, p, q));
	this.mergeVertices();
}

TorusKnotGeometry.prototype = Object.create(Geometry.prototype);
TorusKnotGeometry.prototype.constructor = TorusKnotGeometry;

/**
 * @author Mugen87 / https://github.com/Mugen87
 */

function TorusBufferGeometry(radius, tube, radialSegments, tubularSegments, arc) {

	BufferGeometry.call(this);

	this.type = 'TorusBufferGeometry';

	this.parameters = {
		radius: radius,
		tube: tube,
		radialSegments: radialSegments,
		tubularSegments: tubularSegments,
		arc: arc
	};

	radius = radius || 100;
	tube = tube || 40;
	radialSegments = Math.floor(radialSegments) || 8;
	tubularSegments = Math.floor(tubularSegments) || 6;
	arc = arc || Math.PI * 2;

	// used to calculate buffer length
	var vertexCount = (radialSegments + 1) * (tubularSegments + 1);
	var indexCount = radialSegments * tubularSegments * 2 * 3;

	// buffers
	var indices = new (indexCount > 65535 ? Uint32Array : Uint16Array)(indexCount);
	var vertices = new Float32Array(vertexCount * 3);
	var normals = new Float32Array(vertexCount * 3);
	var uvs = new Float32Array(vertexCount * 2);

	// offset variables
	var vertexBufferOffset = 0;
	var uvBufferOffset = 0;
	var indexBufferOffset = 0;

	// helper variables
	var center = new Vector3();
	var vertex = new Vector3();
	var normal = new Vector3();

	var j, i;

	// generate vertices, normals and uvs

	for (j = 0; j <= radialSegments; j++) {

		for (i = 0; i <= tubularSegments; i++) {

			var u = i / tubularSegments * arc;
			var v = j / radialSegments * Math.PI * 2;

			// vertex
			vertex.x = (radius + tube * Math.cos(v)) * Math.cos(u);
			vertex.y = (radius + tube * Math.cos(v)) * Math.sin(u);
			vertex.z = tube * Math.sin(v);

			vertices[vertexBufferOffset] = vertex.x;
			vertices[vertexBufferOffset + 1] = vertex.y;
			vertices[vertexBufferOffset + 2] = vertex.z;

			// this vector is used to calculate the normal
			center.x = radius * Math.cos(u);
			center.y = radius * Math.sin(u);

			// normal
			normal.subVectors(vertex, center).normalize();

			normals[vertexBufferOffset] = normal.x;
			normals[vertexBufferOffset + 1] = normal.y;
			normals[vertexBufferOffset + 2] = normal.z;

			// uv
			uvs[uvBufferOffset] = i / tubularSegments;
			uvs[uvBufferOffset + 1] = j / radialSegments;

			// update offsets
			vertexBufferOffset += 3;
			uvBufferOffset += 2;
		}
	}

	// generate indices

	for (j = 1; j <= radialSegments; j++) {

		for (i = 1; i <= tubularSegments; i++) {

			// indices
			var a = (tubularSegments + 1) * j + i - 1;
			var b = (tubularSegments + 1) * (j - 1) + i - 1;
			var c = (tubularSegments + 1) * (j - 1) + i;
			var d = (tubularSegments + 1) * j + i;

			// face one
			indices[indexBufferOffset] = a;
			indices[indexBufferOffset + 1] = b;
			indices[indexBufferOffset + 2] = d;

			// face two
			indices[indexBufferOffset + 3] = b;
			indices[indexBufferOffset + 4] = c;
			indices[indexBufferOffset + 5] = d;

			// update offset
			indexBufferOffset += 6;
		}
	}

	// build geometry
	this.setIndex(new BufferAttribute(indices, 1));
	this.addAttribute('position', new BufferAttribute(vertices, 3));
	this.addAttribute('normal', new BufferAttribute(normals, 3));
	this.addAttribute('uv', new BufferAttribute(uvs, 2));
}

TorusBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
TorusBufferGeometry.prototype.constructor = TorusBufferGeometry;

/**
 * @author oosmoxiecode
 * @author mrdoob / http://mrdoob.com/
 * based on http://code.google.com/p/away3d/source/browse/trunk/fp10/Away3DLite/src/away3dlite/primitives/Torus.as?r=2888
 */

function TorusGeometry(radius, tube, radialSegments, tubularSegments, arc) {

	Geometry.call(this);

	this.type = 'TorusGeometry';

	this.parameters = {
		radius: radius,
		tube: tube,
		radialSegments: radialSegments,
		tubularSegments: tubularSegments,
		arc: arc
	};

	this.fromBufferGeometry(new TorusBufferGeometry(radius, tube, radialSegments, tubularSegments, arc));
}

TorusGeometry.prototype = Object.create(Geometry.prototype);
TorusGeometry.prototype.constructor = TorusGeometry;

/**
 * @author zz85 / http://www.lab4games.net/zz85/blog
 */

var ShapeUtils = {

	// calculate area of the contour polygon

	area: function area(contour) {

		var n = contour.length;
		var a = 0.0;

		for (var p = n - 1, q = 0; q < n; p = q++) {

			a += contour[p].x * contour[q].y - contour[q].x * contour[p].y;
		}

		return a * 0.5;
	},

	triangulate: function () {

		/**
   * This code is a quick port of code written in C++ which was submitted to
   * flipcode.com by John W. Ratcliff  // July 22, 2000
   * See original code and more information here:
   * http://www.flipcode.com/archives/Efficient_Polygon_Triangulation.shtml
   *
   * ported to actionscript by Zevan Rosser
   * www.actionsnippet.com
   *
   * ported to javascript by Joshua Koo
   * http://www.lab4games.net/zz85/blog
   *
   */

		function snip(contour, u, v, w, n, verts) {

			var p;
			var ax, ay, bx, by;
			var cx, cy, px, py;

			ax = contour[verts[u]].x;
			ay = contour[verts[u]].y;

			bx = contour[verts[v]].x;
			by = contour[verts[v]].y;

			cx = contour[verts[w]].x;
			cy = contour[verts[w]].y;

			if ((bx - ax) * (cy - ay) - (by - ay) * (cx - ax) <= 0) return false;

			var aX, aY, bX, bY, cX, cY;
			var apx, apy, bpx, bpy, cpx, cpy;
			var cCROSSap, bCROSScp, aCROSSbp;

			aX = cx - bx;aY = cy - by;
			bX = ax - cx;bY = ay - cy;
			cX = bx - ax;cY = by - ay;

			for (p = 0; p < n; p++) {

				px = contour[verts[p]].x;
				py = contour[verts[p]].y;

				if (px === ax && py === ay || px === bx && py === by || px === cx && py === cy) continue;

				apx = px - ax;apy = py - ay;
				bpx = px - bx;bpy = py - by;
				cpx = px - cx;cpy = py - cy;

				// see if p is inside triangle abc

				aCROSSbp = aX * bpy - aY * bpx;
				cCROSSap = cX * apy - cY * apx;
				bCROSScp = bX * cpy - bY * cpx;

				if (aCROSSbp >= -Number.EPSILON && bCROSScp >= -Number.EPSILON && cCROSSap >= -Number.EPSILON) return false;
			}

			return true;
		}

		// takes in an contour array and returns

		return function triangulate(contour, indices) {

			var n = contour.length;

			if (n < 3) return null;

			var result = [],
			    verts = [],
			    vertIndices = [];

			/* we want a counter-clockwise polygon in verts */

			var u, v, w;

			if (ShapeUtils.area(contour) > 0.0) {

				for (v = 0; v < n; v++) {
					verts[v] = v;
				}
			} else {

				for (v = 0; v < n; v++) {
					verts[v] = n - 1 - v;
				}
			}

			var nv = n;

			/*  remove nv - 2 vertices, creating 1 triangle every time */

			var count = 2 * nv; /* error detection */

			for (v = nv - 1; nv > 2;) {

				/* if we loop, it is probably a non-simple polygon */

				if (count-- <= 0) {

					//** Triangulate: ERROR - probable bad polygon!

					//throw ( "Warning, unable to triangulate polygon!" );
					//return null;
					// Sometimes warning is fine, especially polygons are triangulated in reverse.
					console.warn('THREE.ShapeUtils: Unable to triangulate polygon! in triangulate()');

					if (indices) return vertIndices;
					return result;
				}

				/* three consecutive vertices in current polygon, <u,v,w> */

				u = v;if (nv <= u) u = 0; /* previous */
				v = u + 1;if (nv <= v) v = 0; /* new v    */
				w = v + 1;if (nv <= w) w = 0; /* next     */

				if (snip(contour, u, v, w, nv, verts)) {

					var a, b, c, s, t;

					/* true names of the vertices */

					a = verts[u];
					b = verts[v];
					c = verts[w];

					/* output Triangle */

					result.push([contour[a], contour[b], contour[c]]);

					vertIndices.push([verts[u], verts[v], verts[w]]);

					/* remove v from the remaining polygon */

					for (s = v, t = v + 1; t < nv; s++, t++) {

						verts[s] = verts[t];
					}

					nv--;

					/* reset error detection counter */

					count = 2 * nv;
				}
			}

			if (indices) return vertIndices;
			return result;
		};
	}(),

	triangulateShape: function triangulateShape(contour, holes) {

		function removeDupEndPts(points) {

			var l = points.length;

			if (l > 2 && points[l - 1].equals(points[0])) {

				points.pop();
			}
		}

		removeDupEndPts(contour);
		holes.forEach(removeDupEndPts);

		function point_in_segment_2D_colin(inSegPt1, inSegPt2, inOtherPt) {

			// inOtherPt needs to be collinear to the inSegment
			if (inSegPt1.x !== inSegPt2.x) {

				if (inSegPt1.x < inSegPt2.x) {

					return inSegPt1.x <= inOtherPt.x && inOtherPt.x <= inSegPt2.x;
				} else {

					return inSegPt2.x <= inOtherPt.x && inOtherPt.x <= inSegPt1.x;
				}
			} else {

				if (inSegPt1.y < inSegPt2.y) {

					return inSegPt1.y <= inOtherPt.y && inOtherPt.y <= inSegPt2.y;
				} else {

					return inSegPt2.y <= inOtherPt.y && inOtherPt.y <= inSegPt1.y;
				}
			}
		}

		function intersect_segments_2D(inSeg1Pt1, inSeg1Pt2, inSeg2Pt1, inSeg2Pt2, inExcludeAdjacentSegs) {

			var seg1dx = inSeg1Pt2.x - inSeg1Pt1.x,
			    seg1dy = inSeg1Pt2.y - inSeg1Pt1.y;
			var seg2dx = inSeg2Pt2.x - inSeg2Pt1.x,
			    seg2dy = inSeg2Pt2.y - inSeg2Pt1.y;

			var seg1seg2dx = inSeg1Pt1.x - inSeg2Pt1.x;
			var seg1seg2dy = inSeg1Pt1.y - inSeg2Pt1.y;

			var limit = seg1dy * seg2dx - seg1dx * seg2dy;
			var perpSeg1 = seg1dy * seg1seg2dx - seg1dx * seg1seg2dy;

			if (Math.abs(limit) > Number.EPSILON) {

				// not parallel

				var perpSeg2;
				if (limit > 0) {

					if (perpSeg1 < 0 || perpSeg1 > limit) return [];
					perpSeg2 = seg2dy * seg1seg2dx - seg2dx * seg1seg2dy;
					if (perpSeg2 < 0 || perpSeg2 > limit) return [];
				} else {

					if (perpSeg1 > 0 || perpSeg1 < limit) return [];
					perpSeg2 = seg2dy * seg1seg2dx - seg2dx * seg1seg2dy;
					if (perpSeg2 > 0 || perpSeg2 < limit) return [];
				}

				// i.e. to reduce rounding errors
				// intersection at endpoint of segment#1?
				if (perpSeg2 === 0) {

					if (inExcludeAdjacentSegs && (perpSeg1 === 0 || perpSeg1 === limit)) return [];
					return [inSeg1Pt1];
				}
				if (perpSeg2 === limit) {

					if (inExcludeAdjacentSegs && (perpSeg1 === 0 || perpSeg1 === limit)) return [];
					return [inSeg1Pt2];
				}
				// intersection at endpoint of segment#2?
				if (perpSeg1 === 0) return [inSeg2Pt1];
				if (perpSeg1 === limit) return [inSeg2Pt2];

				// return real intersection point
				var factorSeg1 = perpSeg2 / limit;
				return [{ x: inSeg1Pt1.x + factorSeg1 * seg1dx,
					y: inSeg1Pt1.y + factorSeg1 * seg1dy }];
			} else {

				// parallel or collinear
				if (perpSeg1 !== 0 || seg2dy * seg1seg2dx !== seg2dx * seg1seg2dy) return [];

				// they are collinear or degenerate
				var seg1Pt = seg1dx === 0 && seg1dy === 0; // segment1 is just a point?
				var seg2Pt = seg2dx === 0 && seg2dy === 0; // segment2 is just a point?
				// both segments are points
				if (seg1Pt && seg2Pt) {

					if (inSeg1Pt1.x !== inSeg2Pt1.x || inSeg1Pt1.y !== inSeg2Pt1.y) return []; // they are distinct  points
					return [inSeg1Pt1]; // they are the same point
				}
				// segment#1  is a single point
				if (seg1Pt) {

					if (!point_in_segment_2D_colin(inSeg2Pt1, inSeg2Pt2, inSeg1Pt1)) return []; // but not in segment#2
					return [inSeg1Pt1];
				}
				// segment#2  is a single point
				if (seg2Pt) {

					if (!point_in_segment_2D_colin(inSeg1Pt1, inSeg1Pt2, inSeg2Pt1)) return []; // but not in segment#1
					return [inSeg2Pt1];
				}

				// they are collinear segments, which might overlap
				var seg1min, seg1max, seg1minVal, seg1maxVal;
				var seg2min, seg2max, seg2minVal, seg2maxVal;
				if (seg1dx !== 0) {

					// the segments are NOT on a vertical line
					if (inSeg1Pt1.x < inSeg1Pt2.x) {

						seg1min = inSeg1Pt1;seg1minVal = inSeg1Pt1.x;
						seg1max = inSeg1Pt2;seg1maxVal = inSeg1Pt2.x;
					} else {

						seg1min = inSeg1Pt2;seg1minVal = inSeg1Pt2.x;
						seg1max = inSeg1Pt1;seg1maxVal = inSeg1Pt1.x;
					}
					if (inSeg2Pt1.x < inSeg2Pt2.x) {

						seg2min = inSeg2Pt1;seg2minVal = inSeg2Pt1.x;
						seg2max = inSeg2Pt2;seg2maxVal = inSeg2Pt2.x;
					} else {

						seg2min = inSeg2Pt2;seg2minVal = inSeg2Pt2.x;
						seg2max = inSeg2Pt1;seg2maxVal = inSeg2Pt1.x;
					}
				} else {

					// the segments are on a vertical line
					if (inSeg1Pt1.y < inSeg1Pt2.y) {

						seg1min = inSeg1Pt1;seg1minVal = inSeg1Pt1.y;
						seg1max = inSeg1Pt2;seg1maxVal = inSeg1Pt2.y;
					} else {

						seg1min = inSeg1Pt2;seg1minVal = inSeg1Pt2.y;
						seg1max = inSeg1Pt1;seg1maxVal = inSeg1Pt1.y;
					}
					if (inSeg2Pt1.y < inSeg2Pt2.y) {

						seg2min = inSeg2Pt1;seg2minVal = inSeg2Pt1.y;
						seg2max = inSeg2Pt2;seg2maxVal = inSeg2Pt2.y;
					} else {

						seg2min = inSeg2Pt2;seg2minVal = inSeg2Pt2.y;
						seg2max = inSeg2Pt1;seg2maxVal = inSeg2Pt1.y;
					}
				}
				if (seg1minVal <= seg2minVal) {

					if (seg1maxVal < seg2minVal) return [];
					if (seg1maxVal === seg2minVal) {

						if (inExcludeAdjacentSegs) return [];
						return [seg2min];
					}
					if (seg1maxVal <= seg2maxVal) return [seg2min, seg1max];
					return [seg2min, seg2max];
				} else {

					if (seg1minVal > seg2maxVal) return [];
					if (seg1minVal === seg2maxVal) {

						if (inExcludeAdjacentSegs) return [];
						return [seg1min];
					}
					if (seg1maxVal <= seg2maxVal) return [seg1min, seg1max];
					return [seg1min, seg2max];
				}
			}
		}

		function isPointInsideAngle(inVertex, inLegFromPt, inLegToPt, inOtherPt) {

			// The order of legs is important

			// translation of all points, so that Vertex is at (0,0)
			var legFromPtX = inLegFromPt.x - inVertex.x,
			    legFromPtY = inLegFromPt.y - inVertex.y;
			var legToPtX = inLegToPt.x - inVertex.x,
			    legToPtY = inLegToPt.y - inVertex.y;
			var otherPtX = inOtherPt.x - inVertex.x,
			    otherPtY = inOtherPt.y - inVertex.y;

			// main angle >0: < 180 deg.; 0: 180 deg.; <0: > 180 deg.
			var from2toAngle = legFromPtX * legToPtY - legFromPtY * legToPtX;
			var from2otherAngle = legFromPtX * otherPtY - legFromPtY * otherPtX;

			if (Math.abs(from2toAngle) > Number.EPSILON) {

				// angle != 180 deg.

				var other2toAngle = otherPtX * legToPtY - otherPtY * legToPtX;
				// console.log( "from2to: " + from2toAngle + ", from2other: " + from2otherAngle + ", other2to: " + other2toAngle );

				if (from2toAngle > 0) {

					// main angle < 180 deg.
					return from2otherAngle >= 0 && other2toAngle >= 0;
				} else {

					// main angle > 180 deg.
					return from2otherAngle >= 0 || other2toAngle >= 0;
				}
			} else {

				// angle == 180 deg.
				// console.log( "from2to: 180 deg., from2other: " + from2otherAngle  );
				return from2otherAngle > 0;
			}
		}

		function removeHoles(contour, holes) {

			var shape = contour.concat(); // work on this shape
			var hole;

			function isCutLineInsideAngles(inShapeIdx, inHoleIdx) {

				// Check if hole point lies within angle around shape point
				var lastShapeIdx = shape.length - 1;

				var prevShapeIdx = inShapeIdx - 1;
				if (prevShapeIdx < 0) prevShapeIdx = lastShapeIdx;

				var nextShapeIdx = inShapeIdx + 1;
				if (nextShapeIdx > lastShapeIdx) nextShapeIdx = 0;

				var insideAngle = isPointInsideAngle(shape[inShapeIdx], shape[prevShapeIdx], shape[nextShapeIdx], hole[inHoleIdx]);
				if (!insideAngle) {

					// console.log( "Vertex (Shape): " + inShapeIdx + ", Point: " + hole[inHoleIdx].x + "/" + hole[inHoleIdx].y );
					return false;
				}

				// Check if shape point lies within angle around hole point
				var lastHoleIdx = hole.length - 1;

				var prevHoleIdx = inHoleIdx - 1;
				if (prevHoleIdx < 0) prevHoleIdx = lastHoleIdx;

				var nextHoleIdx = inHoleIdx + 1;
				if (nextHoleIdx > lastHoleIdx) nextHoleIdx = 0;

				insideAngle = isPointInsideAngle(hole[inHoleIdx], hole[prevHoleIdx], hole[nextHoleIdx], shape[inShapeIdx]);
				if (!insideAngle) {

					// console.log( "Vertex (Hole): " + inHoleIdx + ", Point: " + shape[inShapeIdx].x + "/" + shape[inShapeIdx].y );
					return false;
				}

				return true;
			}

			function intersectsShapeEdge(inShapePt, inHolePt) {

				// checks for intersections with shape edges
				var sIdx, nextIdx, intersection;
				for (sIdx = 0; sIdx < shape.length; sIdx++) {

					nextIdx = sIdx + 1;nextIdx %= shape.length;
					intersection = intersect_segments_2D(inShapePt, inHolePt, shape[sIdx], shape[nextIdx], true);
					if (intersection.length > 0) return true;
				}

				return false;
			}

			var indepHoles = [];

			function intersectsHoleEdge(inShapePt, inHolePt) {

				// checks for intersections with hole edges
				var ihIdx, chkHole, hIdx, nextIdx, intersection;
				for (ihIdx = 0; ihIdx < indepHoles.length; ihIdx++) {

					chkHole = holes[indepHoles[ihIdx]];
					for (hIdx = 0; hIdx < chkHole.length; hIdx++) {

						nextIdx = hIdx + 1;nextIdx %= chkHole.length;
						intersection = intersect_segments_2D(inShapePt, inHolePt, chkHole[hIdx], chkHole[nextIdx], true);
						if (intersection.length > 0) return true;
					}
				}
				return false;
			}

			var holeIndex,
			    shapeIndex,
			    shapePt,
			    holePt,
			    holeIdx,
			    cutKey,
			    failedCuts = [],
			    tmpShape1,
			    tmpShape2,
			    tmpHole1,
			    tmpHole2;

			for (var h = 0, hl = holes.length; h < hl; h++) {

				indepHoles.push(h);
			}

			var minShapeIndex = 0;
			var counter = indepHoles.length * 2;
			while (indepHoles.length > 0) {

				counter--;
				if (counter < 0) {

					console.log("Infinite Loop! Holes left:" + indepHoles.length + ", Probably Hole outside Shape!");
					break;
				}

				// search for shape-vertex and hole-vertex,
				// which can be connected without intersections
				for (shapeIndex = minShapeIndex; shapeIndex < shape.length; shapeIndex++) {

					shapePt = shape[shapeIndex];
					holeIndex = -1;

					// search for hole which can be reached without intersections
					for (var h = 0; h < indepHoles.length; h++) {

						holeIdx = indepHoles[h];

						// prevent multiple checks
						cutKey = shapePt.x + ":" + shapePt.y + ":" + holeIdx;
						if (failedCuts[cutKey] !== undefined) continue;

						hole = holes[holeIdx];
						for (var h2 = 0; h2 < hole.length; h2++) {

							holePt = hole[h2];
							if (!isCutLineInsideAngles(shapeIndex, h2)) continue;
							if (intersectsShapeEdge(shapePt, holePt)) continue;
							if (intersectsHoleEdge(shapePt, holePt)) continue;

							holeIndex = h2;
							indepHoles.splice(h, 1);

							tmpShape1 = shape.slice(0, shapeIndex + 1);
							tmpShape2 = shape.slice(shapeIndex);
							tmpHole1 = hole.slice(holeIndex);
							tmpHole2 = hole.slice(0, holeIndex + 1);

							shape = tmpShape1.concat(tmpHole1).concat(tmpHole2).concat(tmpShape2);

							minShapeIndex = shapeIndex;

							// Debug only, to show the selected cuts
							// glob_CutLines.push( [ shapePt, holePt ] );

							break;
						}
						if (holeIndex >= 0) break; // hole-vertex found

						failedCuts[cutKey] = true; // remember failure
					}
					if (holeIndex >= 0) break; // hole-vertex found
				}
			}

			return shape; /* shape with no holes */
		}

		var i,
		    il,
		    f,
		    face,
		    key,
		    index,
		    allPointsMap = {};

		// To maintain reference to old shape, one must match coordinates, or offset the indices from original arrays. It's probably easier to do the first.

		var allpoints = contour.concat();

		for (var h = 0, hl = holes.length; h < hl; h++) {

			Array.prototype.push.apply(allpoints, holes[h]);
		}

		//console.log( "allpoints",allpoints, allpoints.length );

		// prepare all points map

		for (i = 0, il = allpoints.length; i < il; i++) {

			key = allpoints[i].x + ":" + allpoints[i].y;

			if (allPointsMap[key] !== undefined) {

				console.warn("THREE.ShapeUtils: Duplicate point", key, i);
			}

			allPointsMap[key] = i;
		}

		// remove holes by cutting paths to holes and adding them to the shape
		var shapeWithoutHoles = removeHoles(contour, holes);

		var triangles = ShapeUtils.triangulate(shapeWithoutHoles, false); // True returns indices for points of spooled shape
		//console.log( "triangles",triangles, triangles.length );

		// check all face vertices against all points map

		for (i = 0, il = triangles.length; i < il; i++) {

			face = triangles[i];

			for (f = 0; f < 3; f++) {

				key = face[f].x + ":" + face[f].y;

				index = allPointsMap[key];

				if (index !== undefined) {

					face[f] = index;
				}
			}
		}

		return triangles.concat();
	},

	isClockWise: function isClockWise(pts) {

		return ShapeUtils.area(pts) < 0;
	},

	// Bezier Curves formulas obtained from
	// http://en.wikipedia.org/wiki/B%C3%A9zier_curve

	// Quad Bezier Functions

	b2: function () {

		function b2p0(t, p) {

			var k = 1 - t;
			return k * k * p;
		}

		function b2p1(t, p) {

			return 2 * (1 - t) * t * p;
		}

		function b2p2(t, p) {

			return t * t * p;
		}

		return function b2(t, p0, p1, p2) {

			return b2p0(t, p0) + b2p1(t, p1) + b2p2(t, p2);
		};
	}(),

	// Cubic Bezier Functions

	b3: function () {

		function b3p0(t, p) {

			var k = 1 - t;
			return k * k * k * p;
		}

		function b3p1(t, p) {

			var k = 1 - t;
			return 3 * k * k * t * p;
		}

		function b3p2(t, p) {

			var k = 1 - t;
			return 3 * k * t * t * p;
		}

		function b3p3(t, p) {

			return t * t * t * p;
		}

		return function b3(t, p0, p1, p2, p3) {

			return b3p0(t, p0) + b3p1(t, p1) + b3p2(t, p2) + b3p3(t, p3);
		};
	}()

};

/**
 * @author zz85 / http://www.lab4games.net/zz85/blog
 *
 * Creates extruded geometry from a path shape.
 *
 * parameters = {
 *
 *  curveSegments: <int>, // number of points on the curves
 *  steps: <int>, // number of points for z-side extrusions / used for subdividing segments of extrude spline too
 *  amount: <int>, // Depth to extrude the shape
 *
 *  bevelEnabled: <bool>, // turn on bevel
 *  bevelThickness: <float>, // how deep into the original shape bevel goes
 *  bevelSize: <float>, // how far from shape outline is bevel
 *  bevelSegments: <int>, // number of bevel layers
 *
 *  extrudePath: <THREE.CurvePath> // 3d spline path to extrude shape along. (creates Frames if .frames aren't defined)
 *  frames: <Object> // containing arrays of tangents, normals, binormals
 *
 *  uvGenerator: <Object> // object that provides UV generator functions
 *
 * }
 **/

function ExtrudeGeometry(shapes, options) {

	if (typeof shapes === "undefined") {

		shapes = [];
		return;
	}

	Geometry.call(this);

	this.type = 'ExtrudeGeometry';

	shapes = Array.isArray(shapes) ? shapes : [shapes];

	this.addShapeList(shapes, options);

	this.computeFaceNormals();

	// can't really use automatic vertex normals
	// as then front and back sides get smoothed too
	// should do separate smoothing just for sides

	//this.computeVertexNormals();

	//console.log( "took", ( Date.now() - startTime ) );
}

ExtrudeGeometry.prototype = Object.create(Geometry.prototype);
ExtrudeGeometry.prototype.constructor = ExtrudeGeometry;

ExtrudeGeometry.prototype.addShapeList = function (shapes, options) {

	var sl = shapes.length;

	for (var s = 0; s < sl; s++) {

		var shape = shapes[s];
		this.addShape(shape, options);
	}
};

ExtrudeGeometry.prototype.addShape = function (shape, options) {

	var amount = options.amount !== undefined ? options.amount : 100;

	var bevelThickness = options.bevelThickness !== undefined ? options.bevelThickness : 6; // 10
	var bevelSize = options.bevelSize !== undefined ? options.bevelSize : bevelThickness - 2; // 8
	var bevelSegments = options.bevelSegments !== undefined ? options.bevelSegments : 3;

	var bevelEnabled = options.bevelEnabled !== undefined ? options.bevelEnabled : true; // false

	var curveSegments = options.curveSegments !== undefined ? options.curveSegments : 12;

	var steps = options.steps !== undefined ? options.steps : 1;

	var extrudePath = options.extrudePath;
	var extrudePts,
	    extrudeByPath = false;

	// Use default WorldUVGenerator if no UV generators are specified.
	var uvgen = options.UVGenerator !== undefined ? options.UVGenerator : ExtrudeGeometry.WorldUVGenerator;

	var splineTube, binormal, normal, position2;
	if (extrudePath) {

		extrudePts = extrudePath.getSpacedPoints(steps);

		extrudeByPath = true;
		bevelEnabled = false; // bevels not supported for path extrusion

		// SETUP TNB variables

		// TODO1 - have a .isClosed in spline?

		splineTube = options.frames !== undefined ? options.frames : extrudePath.computeFrenetFrames(steps, false);

		// console.log(splineTube, 'splineTube', splineTube.normals.length, 'steps', steps, 'extrudePts', extrudePts.length);

		binormal = new Vector3();
		normal = new Vector3();
		position2 = new Vector3();
	}

	// Safeguards if bevels are not enabled

	if (!bevelEnabled) {

		bevelSegments = 0;
		bevelThickness = 0;
		bevelSize = 0;
	}

	// Variables initialization

	var ahole, h, hl; // looping of holes
	var scope = this;

	var shapesOffset = this.vertices.length;

	var shapePoints = shape.extractPoints(curveSegments);

	var vertices = shapePoints.shape;
	var holes = shapePoints.holes;

	var reverse = !ShapeUtils.isClockWise(vertices);

	if (reverse) {

		vertices = vertices.reverse();

		// Maybe we should also check if holes are in the opposite direction, just to be safe ...

		for (h = 0, hl = holes.length; h < hl; h++) {

			ahole = holes[h];

			if (ShapeUtils.isClockWise(ahole)) {

				holes[h] = ahole.reverse();
			}
		}

		reverse = false; // If vertices are in order now, we shouldn't need to worry about them again (hopefully)!
	}

	var faces = ShapeUtils.triangulateShape(vertices, holes);

	/* Vertices */

	var contour = vertices; // vertices has all points but contour has only points of circumference

	for (h = 0, hl = holes.length; h < hl; h++) {

		ahole = holes[h];

		vertices = vertices.concat(ahole);
	}

	function scalePt2(pt, vec, size) {

		if (!vec) console.error("THREE.ExtrudeGeometry: vec does not exist");

		return vec.clone().multiplyScalar(size).add(pt);
	}

	var b,
	    bs,
	    t,
	    z,
	    vert,
	    vlen = vertices.length,
	    face,
	    flen = faces.length;

	// Find directions for point movement


	function getBevelVec(inPt, inPrev, inNext) {

		// computes for inPt the corresponding point inPt' on a new contour
		//   shifted by 1 unit (length of normalized vector) to the left
		// if we walk along contour clockwise, this new contour is outside the old one
		//
		// inPt' is the intersection of the two lines parallel to the two
		//  adjacent edges of inPt at a distance of 1 unit on the left side.

		var v_trans_x,
		    v_trans_y,
		    shrink_by = 1; // resulting translation vector for inPt

		// good reading for geometry algorithms (here: line-line intersection)
		// http://geomalgorithms.com/a05-_intersect-1.html

		var v_prev_x = inPt.x - inPrev.x,
		    v_prev_y = inPt.y - inPrev.y;
		var v_next_x = inNext.x - inPt.x,
		    v_next_y = inNext.y - inPt.y;

		var v_prev_lensq = v_prev_x * v_prev_x + v_prev_y * v_prev_y;

		// check for collinear edges
		var collinear0 = v_prev_x * v_next_y - v_prev_y * v_next_x;

		if (Math.abs(collinear0) > Number.EPSILON) {

			// not collinear

			// length of vectors for normalizing

			var v_prev_len = Math.sqrt(v_prev_lensq);
			var v_next_len = Math.sqrt(v_next_x * v_next_x + v_next_y * v_next_y);

			// shift adjacent points by unit vectors to the left

			var ptPrevShift_x = inPrev.x - v_prev_y / v_prev_len;
			var ptPrevShift_y = inPrev.y + v_prev_x / v_prev_len;

			var ptNextShift_x = inNext.x - v_next_y / v_next_len;
			var ptNextShift_y = inNext.y + v_next_x / v_next_len;

			// scaling factor for v_prev to intersection point

			var sf = ((ptNextShift_x - ptPrevShift_x) * v_next_y - (ptNextShift_y - ptPrevShift_y) * v_next_x) / (v_prev_x * v_next_y - v_prev_y * v_next_x);

			// vector from inPt to intersection point

			v_trans_x = ptPrevShift_x + v_prev_x * sf - inPt.x;
			v_trans_y = ptPrevShift_y + v_prev_y * sf - inPt.y;

			// Don't normalize!, otherwise sharp corners become ugly
			//  but prevent crazy spikes
			var v_trans_lensq = v_trans_x * v_trans_x + v_trans_y * v_trans_y;
			if (v_trans_lensq <= 2) {

				return new Vector2(v_trans_x, v_trans_y);
			} else {

				shrink_by = Math.sqrt(v_trans_lensq / 2);
			}
		} else {

			// handle special case of collinear edges

			var direction_eq = false; // assumes: opposite
			if (v_prev_x > Number.EPSILON) {

				if (v_next_x > Number.EPSILON) {

					direction_eq = true;
				}
			} else {

				if (v_prev_x < -Number.EPSILON) {

					if (v_next_x < -Number.EPSILON) {

						direction_eq = true;
					}
				} else {

					if (Math.sign(v_prev_y) === Math.sign(v_next_y)) {

						direction_eq = true;
					}
				}
			}

			if (direction_eq) {

				// console.log("Warning: lines are a straight sequence");
				v_trans_x = -v_prev_y;
				v_trans_y = v_prev_x;
				shrink_by = Math.sqrt(v_prev_lensq);
			} else {

				// console.log("Warning: lines are a straight spike");
				v_trans_x = v_prev_x;
				v_trans_y = v_prev_y;
				shrink_by = Math.sqrt(v_prev_lensq / 2);
			}
		}

		return new Vector2(v_trans_x / shrink_by, v_trans_y / shrink_by);
	}

	var contourMovements = [];

	for (var i = 0, il = contour.length, j = il - 1, k = i + 1; i < il; i++, j++, k++) {

		if (j === il) j = 0;
		if (k === il) k = 0;

		//  (j)---(i)---(k)
		// console.log('i,j,k', i, j , k)

		contourMovements[i] = getBevelVec(contour[i], contour[j], contour[k]);
	}

	var holesMovements = [],
	    oneHoleMovements,
	    verticesMovements = contourMovements.concat();

	for (h = 0, hl = holes.length; h < hl; h++) {

		ahole = holes[h];

		oneHoleMovements = [];

		for (i = 0, il = ahole.length, j = il - 1, k = i + 1; i < il; i++, j++, k++) {

			if (j === il) j = 0;
			if (k === il) k = 0;

			//  (j)---(i)---(k)
			oneHoleMovements[i] = getBevelVec(ahole[i], ahole[j], ahole[k]);
		}

		holesMovements.push(oneHoleMovements);
		verticesMovements = verticesMovements.concat(oneHoleMovements);
	}

	// Loop bevelSegments, 1 for the front, 1 for the back

	for (b = 0; b < bevelSegments; b++) {

		//for ( b = bevelSegments; b > 0; b -- ) {

		t = b / bevelSegments;
		z = bevelThickness * Math.cos(t * Math.PI / 2);
		bs = bevelSize * Math.sin(t * Math.PI / 2);

		// contract shape

		for (i = 0, il = contour.length; i < il; i++) {

			vert = scalePt2(contour[i], contourMovements[i], bs);

			v(vert.x, vert.y, -z);
		}

		// expand holes

		for (h = 0, hl = holes.length; h < hl; h++) {

			ahole = holes[h];
			oneHoleMovements = holesMovements[h];

			for (i = 0, il = ahole.length; i < il; i++) {

				vert = scalePt2(ahole[i], oneHoleMovements[i], bs);

				v(vert.x, vert.y, -z);
			}
		}
	}

	bs = bevelSize;

	// Back facing vertices

	for (i = 0; i < vlen; i++) {

		vert = bevelEnabled ? scalePt2(vertices[i], verticesMovements[i], bs) : vertices[i];

		if (!extrudeByPath) {

			v(vert.x, vert.y, 0);
		} else {

			// v( vert.x, vert.y + extrudePts[ 0 ].y, extrudePts[ 0 ].x );

			normal.copy(splineTube.normals[0]).multiplyScalar(vert.x);
			binormal.copy(splineTube.binormals[0]).multiplyScalar(vert.y);

			position2.copy(extrudePts[0]).add(normal).add(binormal);

			v(position2.x, position2.y, position2.z);
		}
	}

	// Add stepped vertices...
	// Including front facing vertices

	var s;

	for (s = 1; s <= steps; s++) {

		for (i = 0; i < vlen; i++) {

			vert = bevelEnabled ? scalePt2(vertices[i], verticesMovements[i], bs) : vertices[i];

			if (!extrudeByPath) {

				v(vert.x, vert.y, amount / steps * s);
			} else {

				// v( vert.x, vert.y + extrudePts[ s - 1 ].y, extrudePts[ s - 1 ].x );

				normal.copy(splineTube.normals[s]).multiplyScalar(vert.x);
				binormal.copy(splineTube.binormals[s]).multiplyScalar(vert.y);

				position2.copy(extrudePts[s]).add(normal).add(binormal);

				v(position2.x, position2.y, position2.z);
			}
		}
	}

	// Add bevel segments planes

	//for ( b = 1; b <= bevelSegments; b ++ ) {
	for (b = bevelSegments - 1; b >= 0; b--) {

		t = b / bevelSegments;
		z = bevelThickness * Math.cos(t * Math.PI / 2);
		bs = bevelSize * Math.sin(t * Math.PI / 2);

		// contract shape

		for (i = 0, il = contour.length; i < il; i++) {

			vert = scalePt2(contour[i], contourMovements[i], bs);
			v(vert.x, vert.y, amount + z);
		}

		// expand holes

		for (h = 0, hl = holes.length; h < hl; h++) {

			ahole = holes[h];
			oneHoleMovements = holesMovements[h];

			for (i = 0, il = ahole.length; i < il; i++) {

				vert = scalePt2(ahole[i], oneHoleMovements[i], bs);

				if (!extrudeByPath) {

					v(vert.x, vert.y, amount + z);
				} else {

					v(vert.x, vert.y + extrudePts[steps - 1].y, extrudePts[steps - 1].x + z);
				}
			}
		}
	}

	/* Faces */

	// Top and bottom faces

	buildLidFaces();

	// Sides faces

	buildSideFaces();

	/////  Internal functions

	function buildLidFaces() {

		if (bevelEnabled) {

			var layer = 0; // steps + 1
			var offset = vlen * layer;

			// Bottom faces

			for (i = 0; i < flen; i++) {

				face = faces[i];
				f3(face[2] + offset, face[1] + offset, face[0] + offset);
			}

			layer = steps + bevelSegments * 2;
			offset = vlen * layer;

			// Top faces

			for (i = 0; i < flen; i++) {

				face = faces[i];
				f3(face[0] + offset, face[1] + offset, face[2] + offset);
			}
		} else {

			// Bottom faces

			for (i = 0; i < flen; i++) {

				face = faces[i];
				f3(face[2], face[1], face[0]);
			}

			// Top faces

			for (i = 0; i < flen; i++) {

				face = faces[i];
				f3(face[0] + vlen * steps, face[1] + vlen * steps, face[2] + vlen * steps);
			}
		}
	}

	// Create faces for the z-sides of the shape

	function buildSideFaces() {

		var layeroffset = 0;
		sidewalls(contour, layeroffset);
		layeroffset += contour.length;

		for (h = 0, hl = holes.length; h < hl; h++) {

			ahole = holes[h];
			sidewalls(ahole, layeroffset);

			//, true
			layeroffset += ahole.length;
		}
	}

	function sidewalls(contour, layeroffset) {

		var j, k;
		i = contour.length;

		while (--i >= 0) {

			j = i;
			k = i - 1;
			if (k < 0) k = contour.length - 1;

			//console.log('b', i,j, i-1, k,vertices.length);

			var s = 0,
			    sl = steps + bevelSegments * 2;

			for (s = 0; s < sl; s++) {

				var slen1 = vlen * s;
				var slen2 = vlen * (s + 1);

				var a = layeroffset + j + slen1,
				    b = layeroffset + k + slen1,
				    c = layeroffset + k + slen2,
				    d = layeroffset + j + slen2;

				f4(a, b, c, d, contour, s, sl, j, k);
			}
		}
	}

	function v(x, y, z) {

		scope.vertices.push(new Vector3(x, y, z));
	}

	function f3(a, b, c) {

		a += shapesOffset;
		b += shapesOffset;
		c += shapesOffset;

		scope.faces.push(new Face3(a, b, c, null, null, 0));

		var uvs = uvgen.generateTopUV(scope, a, b, c);

		scope.faceVertexUvs[0].push(uvs);
	}

	function f4(a, b, c, d, wallContour, stepIndex, stepsLength, contourIndex1, contourIndex2) {

		a += shapesOffset;
		b += shapesOffset;
		c += shapesOffset;
		d += shapesOffset;

		scope.faces.push(new Face3(a, b, d, null, null, 1));
		scope.faces.push(new Face3(b, c, d, null, null, 1));

		var uvs = uvgen.generateSideWallUV(scope, a, b, c, d);

		scope.faceVertexUvs[0].push([uvs[0], uvs[1], uvs[3]]);
		scope.faceVertexUvs[0].push([uvs[1], uvs[2], uvs[3]]);
	}
};

ExtrudeGeometry.WorldUVGenerator = {

	generateTopUV: function generateTopUV(geometry, indexA, indexB, indexC) {

		var vertices = geometry.vertices;

		var a = vertices[indexA];
		var b = vertices[indexB];
		var c = vertices[indexC];

		return [new Vector2(a.x, a.y), new Vector2(b.x, b.y), new Vector2(c.x, c.y)];
	},

	generateSideWallUV: function generateSideWallUV(geometry, indexA, indexB, indexC, indexD) {

		var vertices = geometry.vertices;

		var a = vertices[indexA];
		var b = vertices[indexB];
		var c = vertices[indexC];
		var d = vertices[indexD];

		if (Math.abs(a.y - b.y) < 0.01) {

			return [new Vector2(a.x, 1 - a.z), new Vector2(b.x, 1 - b.z), new Vector2(c.x, 1 - c.z), new Vector2(d.x, 1 - d.z)];
		} else {

			return [new Vector2(a.y, 1 - a.z), new Vector2(b.y, 1 - b.z), new Vector2(c.y, 1 - c.z), new Vector2(d.y, 1 - d.z)];
		}
	}
};

/**
 * @author zz85 / http://www.lab4games.net/zz85/blog
 * @author alteredq / http://alteredqualia.com/
 *
 * Text = 3D Text
 *
 * parameters = {
 *  font: <THREE.Font>, // font
 *
 *  size: <float>, // size of the text
 *  height: <float>, // thickness to extrude text
 *  curveSegments: <int>, // number of points on the curves
 *
 *  bevelEnabled: <bool>, // turn on bevel
 *  bevelThickness: <float>, // how deep into text bevel goes
 *  bevelSize: <float> // how far from text outline is bevel
 * }
 */

function TextGeometry(text, parameters) {

  parameters = parameters || {};

  var font = parameters.font;

  if ((font && font.isFont) === false) {

    console.error('THREE.TextGeometry: font parameter is not an instance of THREE.Font.');
    return new Geometry();
  }

  var shapes = font.generateShapes(text, parameters.size, parameters.curveSegments);

  // translate parameters to ExtrudeGeometry API

  parameters.amount = parameters.height !== undefined ? parameters.height : 50;

  // defaults

  if (parameters.bevelThickness === undefined) parameters.bevelThickness = 10;
  if (parameters.bevelSize === undefined) parameters.bevelSize = 8;
  if (parameters.bevelEnabled === undefined) parameters.bevelEnabled = false;

  ExtrudeGeometry.call(this, shapes, parameters);

  this.type = 'TextGeometry';
}

TextGeometry.prototype = Object.create(ExtrudeGeometry.prototype);
TextGeometry.prototype.constructor = TextGeometry;

/**
 * @author Kaleb Murphy
 */

function RingGeometry(innerRadius, outerRadius, thetaSegments, phiSegments, thetaStart, thetaLength) {

	Geometry.call(this);

	this.type = 'RingGeometry';

	this.parameters = {
		innerRadius: innerRadius,
		outerRadius: outerRadius,
		thetaSegments: thetaSegments,
		phiSegments: phiSegments,
		thetaStart: thetaStart,
		thetaLength: thetaLength
	};

	this.fromBufferGeometry(new RingBufferGeometry(innerRadius, outerRadius, thetaSegments, phiSegments, thetaStart, thetaLength));
}

RingGeometry.prototype = Object.create(Geometry.prototype);
RingGeometry.prototype.constructor = RingGeometry;

/**
 * @author Mugen87 / https://github.com/Mugen87
 */

// points - to create a closed torus, one must use a set of points
//    like so: [ a, b, c, d, a ], see first is the same as last.
// segments - the number of circumference segments to create
// phiStart - the starting radian
// phiLength - the radian (0 to 2PI) range of the lathed section
//    2PI is a closed lathe, less than 2PI is a portion.

function LatheBufferGeometry(points, segments, phiStart, phiLength) {

	BufferGeometry.call(this);

	this.type = 'LatheBufferGeometry';

	this.parameters = {
		points: points,
		segments: segments,
		phiStart: phiStart,
		phiLength: phiLength
	};

	segments = Math.floor(segments) || 12;
	phiStart = phiStart || 0;
	phiLength = phiLength || Math.PI * 2;

	// clamp phiLength so it's in range of [ 0, 2PI ]
	phiLength = _Math.clamp(phiLength, 0, Math.PI * 2);

	// these are used to calculate buffer length
	var vertexCount = (segments + 1) * points.length;
	var indexCount = segments * points.length * 2 * 3;

	// buffers
	var indices = new BufferAttribute(new (indexCount > 65535 ? Uint32Array : Uint16Array)(indexCount), 1);
	var vertices = new BufferAttribute(new Float32Array(vertexCount * 3), 3);
	var uvs = new BufferAttribute(new Float32Array(vertexCount * 2), 2);

	// helper variables
	var index = 0,
	    indexOffset = 0,
	    base;
	var inverseSegments = 1.0 / segments;
	var vertex = new Vector3();
	var uv = new Vector2();
	var i, j;

	// generate vertices and uvs

	for (i = 0; i <= segments; i++) {

		var phi = phiStart + i * inverseSegments * phiLength;

		var sin = Math.sin(phi);
		var cos = Math.cos(phi);

		for (j = 0; j <= points.length - 1; j++) {

			// vertex
			vertex.x = points[j].x * sin;
			vertex.y = points[j].y;
			vertex.z = points[j].x * cos;
			vertices.setXYZ(index, vertex.x, vertex.y, vertex.z);

			// uv
			uv.x = i / segments;
			uv.y = j / (points.length - 1);
			uvs.setXY(index, uv.x, uv.y);

			// increase index
			index++;
		}
	}

	// generate indices

	for (i = 0; i < segments; i++) {

		for (j = 0; j < points.length - 1; j++) {

			base = j + i * points.length;

			// indices
			var a = base;
			var b = base + points.length;
			var c = base + points.length + 1;
			var d = base + 1;

			// face one
			indices.setX(indexOffset, a);indexOffset++;
			indices.setX(indexOffset, b);indexOffset++;
			indices.setX(indexOffset, d);indexOffset++;

			// face two
			indices.setX(indexOffset, b);indexOffset++;
			indices.setX(indexOffset, c);indexOffset++;
			indices.setX(indexOffset, d);indexOffset++;
		}
	}

	// build geometry

	this.setIndex(indices);
	this.addAttribute('position', vertices);
	this.addAttribute('uv', uvs);

	// generate normals

	this.computeVertexNormals();

	// if the geometry is closed, we need to average the normals along the seam.
	// because the corresponding vertices are identical (but still have different UVs).

	if (phiLength === Math.PI * 2) {

		var normals = this.attributes.normal.array;
		var n1 = new Vector3();
		var n2 = new Vector3();
		var n = new Vector3();

		// this is the buffer offset for the last line of vertices
		base = segments * points.length * 3;

		for (i = 0, j = 0; i < points.length; i++, j += 3) {

			// select the normal of the vertex in the first line
			n1.x = normals[j + 0];
			n1.y = normals[j + 1];
			n1.z = normals[j + 2];

			// select the normal of the vertex in the last line
			n2.x = normals[base + j + 0];
			n2.y = normals[base + j + 1];
			n2.z = normals[base + j + 2];

			// average normals
			n.addVectors(n1, n2).normalize();

			// assign the new values to both normals
			normals[j + 0] = normals[base + j + 0] = n.x;
			normals[j + 1] = normals[base + j + 1] = n.y;
			normals[j + 2] = normals[base + j + 2] = n.z;
		} // next row
	}
}

LatheBufferGeometry.prototype = Object.create(BufferGeometry.prototype);
LatheBufferGeometry.prototype.constructor = LatheBufferGeometry;

/**
 * @author astrodud / http://astrodud.isgreat.org/
 * @author zz85 / https://github.com/zz85
 * @author bhouston / http://clara.io
 */

// points - to create a closed torus, one must use a set of points
//    like so: [ a, b, c, d, a ], see first is the same as last.
// segments - the number of circumference segments to create
// phiStart - the starting radian
// phiLength - the radian (0 to 2PI) range of the lathed section
//    2PI is a closed lathe, less than 2PI is a portion.

function LatheGeometry(points, segments, phiStart, phiLength) {

	Geometry.call(this);

	this.type = 'LatheGeometry';

	this.parameters = {
		points: points,
		segments: segments,
		phiStart: phiStart,
		phiLength: phiLength
	};

	this.fromBufferGeometry(new LatheBufferGeometry(points, segments, phiStart, phiLength));
	this.mergeVertices();
}

LatheGeometry.prototype = Object.create(Geometry.prototype);
LatheGeometry.prototype.constructor = LatheGeometry;

/**
 * @author jonobr1 / http://jonobr1.com
 *
 * Creates a one-sided polygonal geometry from a path shape. Similar to
 * ExtrudeGeometry.
 *
 * parameters = {
 *
 *	curveSegments: <int>, // number of points on the curves. NOT USED AT THE MOMENT.
 *
 *	material: <int> // material index for front and back faces
 *	uvGenerator: <Object> // object that provides UV generator functions
 *
 * }
 **/

function ShapeGeometry(shapes, options) {

	Geometry.call(this);

	this.type = 'ShapeGeometry';

	if (Array.isArray(shapes) === false) shapes = [shapes];

	this.addShapeList(shapes, options);

	this.computeFaceNormals();
}

ShapeGeometry.prototype = Object.create(Geometry.prototype);
ShapeGeometry.prototype.constructor = ShapeGeometry;

/**
 * Add an array of shapes to THREE.ShapeGeometry.
 */
ShapeGeometry.prototype.addShapeList = function (shapes, options) {

	for (var i = 0, l = shapes.length; i < l; i++) {

		this.addShape(shapes[i], options);
	}

	return this;
};

/**
 * Adds a shape to THREE.ShapeGeometry, based on THREE.ExtrudeGeometry.
 */
ShapeGeometry.prototype.addShape = function (shape, options) {

	if (options === undefined) options = {};
	var curveSegments = options.curveSegments !== undefined ? options.curveSegments : 12;

	var material = options.material;
	var uvgen = options.UVGenerator === undefined ? ExtrudeGeometry.WorldUVGenerator : options.UVGenerator;

	//

	var i, l, hole;

	var shapesOffset = this.vertices.length;
	var shapePoints = shape.extractPoints(curveSegments);

	var vertices = shapePoints.shape;
	var holes = shapePoints.holes;

	var reverse = !ShapeUtils.isClockWise(vertices);

	if (reverse) {

		vertices = vertices.reverse();

		// Maybe we should also check if holes are in the opposite direction, just to be safe...

		for (i = 0, l = holes.length; i < l; i++) {

			hole = holes[i];

			if (ShapeUtils.isClockWise(hole)) {

				holes[i] = hole.reverse();
			}
		}

		reverse = false;
	}

	var faces = ShapeUtils.triangulateShape(vertices, holes);

	// Vertices

	for (i = 0, l = holes.length; i < l; i++) {

		hole = holes[i];
		vertices = vertices.concat(hole);
	}

	//

	var vert,
	    vlen = vertices.length;
	var face,
	    flen = faces.length;

	for (i = 0; i < vlen; i++) {

		vert = vertices[i];

		this.vertices.push(new Vector3(vert.x, vert.y, 0));
	}

	for (i = 0; i < flen; i++) {

		face = faces[i];

		var a = face[0] + shapesOffset;
		var b = face[1] + shapesOffset;
		var c = face[2] + shapesOffset;

		this.faces.push(new Face3(a, b, c, null, null, material));
		this.faceVertexUvs[0].push(uvgen.generateTopUV(this, a, b, c));
	}
};

/**
 * @author WestLangley / http://github.com/WestLangley
 */

function EdgesGeometry(geometry, thresholdAngle) {

	BufferGeometry.call(this);

	thresholdAngle = thresholdAngle !== undefined ? thresholdAngle : 1;

	var thresholdDot = Math.cos(_Math.DEG2RAD * thresholdAngle);

	var edge = [0, 0],
	    hash = {};

	function sortFunction(a, b) {

		return a - b;
	}

	var keys = ['a', 'b', 'c'];

	var geometry2;

	if (geometry && geometry.isBufferGeometry) {

		geometry2 = new Geometry();
		geometry2.fromBufferGeometry(geometry);
	} else {

		geometry2 = geometry.clone();
	}

	geometry2.mergeVertices();
	geometry2.computeFaceNormals();

	var vertices = geometry2.vertices;
	var faces = geometry2.faces;

	for (var i = 0, l = faces.length; i < l; i++) {

		var face = faces[i];

		for (var j = 0; j < 3; j++) {

			edge[0] = face[keys[j]];
			edge[1] = face[keys[(j + 1) % 3]];
			edge.sort(sortFunction);

			var key = edge.toString();

			if (hash[key] === undefined) {

				hash[key] = { vert1: edge[0], vert2: edge[1], face1: i, face2: undefined };
			} else {

				hash[key].face2 = i;
			}
		}
	}

	var coords = [];

	for (var key in hash) {

		var h = hash[key];

		if (h.face2 === undefined || faces[h.face1].normal.dot(faces[h.face2].normal) <= thresholdDot) {

			var vertex = vertices[h.vert1];
			coords.push(vertex.x);
			coords.push(vertex.y);
			coords.push(vertex.z);

			vertex = vertices[h.vert2];
			coords.push(vertex.x);
			coords.push(vertex.y);
			coords.push(vertex.z);
		}
	}

	this.addAttribute('position', new BufferAttribute(new Float32Array(coords), 3));
}

EdgesGeometry.prototype = Object.create(BufferGeometry.prototype);
EdgesGeometry.prototype.constructor = EdgesGeometry;

/**
 * @author mrdoob / http://mrdoob.com/
 */

function CylinderGeometry(radiusTop, radiusBottom, height, radialSegments, heightSegments, openEnded, thetaStart, thetaLength) {

	Geometry.call(this);

	this.type = 'CylinderGeometry';

	this.parameters = {
		radiusTop: radiusTop,
		radiusBottom: radiusBottom,
		height: height,
		radialSegments: radialSegments,
		heightSegments: heightSegments,
		openEnded: openEnded,
		thetaStart: thetaStart,
		thetaLength: thetaLength
	};

	this.fromBufferGeometry(new CylinderBufferGeometry(radiusTop, radiusBottom, height, radialSegments, heightSegments, openEnded, thetaStart, thetaLength));
	this.mergeVertices();
}

CylinderGeometry.prototype = Object.create(Geometry.prototype);
CylinderGeometry.prototype.constructor = CylinderGeometry;

/**
 * @author abelnation / http://github.com/abelnation
 */

function ConeGeometry(radius, height, radialSegments, heightSegments, openEnded, thetaStart, thetaLength) {

	CylinderGeometry.call(this, 0, radius, height, radialSegments, heightSegments, openEnded, thetaStart, thetaLength);

	this.type = 'ConeGeometry';

	this.parameters = {
		radius: radius,
		height: height,
		radialSegments: radialSegments,
		heightSegments: heightSegments,
		openEnded: openEnded,
		thetaStart: thetaStart,
		thetaLength: thetaLength
	};
}

ConeGeometry.prototype = Object.create(CylinderGeometry.prototype);
ConeGeometry.prototype.constructor = ConeGeometry;

/**
 * @author: abelnation / http://github.com/abelnation
 */

function ConeBufferGeometry(radius, height, radialSegments, heightSegments, openEnded, thetaStart, thetaLength) {

	CylinderBufferGeometry.call(this, 0, radius, height, radialSegments, heightSegments, openEnded, thetaStart, thetaLength);

	this.type = 'ConeBufferGeometry';

	this.parameters = {
		radius: radius,
		height: height,
		radialSegments: radialSegments,
		heightSegments: heightSegments,
		openEnded: openEnded,
		thetaStart: thetaStart,
		thetaLength: thetaLength
	};
}

ConeBufferGeometry.prototype = Object.create(CylinderBufferGeometry.prototype);
ConeBufferGeometry.prototype.constructor = ConeBufferGeometry;

/**
 * @author hughes
 */

function CircleGeometry(radius, segments, thetaStart, thetaLength) {

	Geometry.call(this);

	this.type = 'CircleGeometry';

	this.parameters = {
		radius: radius,
		segments: segments,
		thetaStart: thetaStart,
		thetaLength: thetaLength
	};

	this.fromBufferGeometry(new CircleBufferGeometry(radius, segments, thetaStart, thetaLength));
}

CircleGeometry.prototype = Object.create(Geometry.prototype);
CircleGeometry.prototype.constructor = CircleGeometry;

/**
 * @author mrdoob / http://mrdoob.com/
 * based on http://papervision3d.googlecode.com/svn/trunk/as3/trunk/src/org/papervision3d/objects/primitives/Cube.as
 */

function BoxGeometry(width, height, depth, widthSegments, heightSegments, depthSegments) {

	Geometry.call(this);

	this.type = 'BoxGeometry';

	this.parameters = {
		width: width,
		height: height,
		depth: depth,
		widthSegments: widthSegments,
		heightSegments: heightSegments,
		depthSegments: depthSegments
	};

	this.fromBufferGeometry(new BoxBufferGeometry(width, height, depth, widthSegments, heightSegments, depthSegments));
	this.mergeVertices();
}

BoxGeometry.prototype = Object.create(Geometry.prototype);
BoxGeometry.prototype.constructor = BoxGeometry;



var Geometries = Object.freeze({
	WireframeGeometry: WireframeGeometry,
	ParametricGeometry: ParametricGeometry,
	ParametricBufferGeometry: ParametricBufferGeometry,
	TetrahedronGeometry: TetrahedronGeometry,
	TetrahedronBufferGeometry: TetrahedronBufferGeometry,
	OctahedronGeometry: OctahedronGeometry,
	OctahedronBufferGeometry: OctahedronBufferGeometry,
	IcosahedronGeometry: IcosahedronGeometry,
	IcosahedronBufferGeometry: IcosahedronBufferGeometry,
	DodecahedronGeometry: DodecahedronGeometry,
	DodecahedronBufferGeometry: DodecahedronBufferGeometry,
	PolyhedronGeometry: PolyhedronGeometry,
	PolyhedronBufferGeometry: PolyhedronBufferGeometry,
	TubeGeometry: TubeGeometry,
	TubeBufferGeometry: TubeBufferGeometry,
	TorusKnotGeometry: TorusKnotGeometry,
	TorusKnotBufferGeometry: TorusKnotBufferGeometry,
	TorusGeometry: TorusGeometry,
	TorusBufferGeometry: TorusBufferGeometry,
	TextGeometry: TextGeometry,
	SphereBufferGeometry: SphereBufferGeometry,
	SphereGeometry: SphereGeometry,
	RingGeometry: RingGeometry,
	RingBufferGeometry: RingBufferGeometry,
	PlaneBufferGeometry: PlaneBufferGeometry,
	PlaneGeometry: PlaneGeometry,
	LatheGeometry: LatheGeometry,
	LatheBufferGeometry: LatheBufferGeometry,
	ShapeGeometry: ShapeGeometry,
	ExtrudeGeometry: ExtrudeGeometry,
	EdgesGeometry: EdgesGeometry,
	ConeGeometry: ConeGeometry,
	ConeBufferGeometry: ConeBufferGeometry,
	CylinderGeometry: CylinderGeometry,
	CylinderBufferGeometry: CylinderBufferGeometry,
	CircleBufferGeometry: CircleBufferGeometry,
	CircleGeometry: CircleGeometry,
	BoxBufferGeometry: BoxBufferGeometry,
	BoxGeometry: BoxGeometry
});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function ObjectLoader(manager) {

	this.manager = manager !== undefined ? manager : DefaultLoadingManager;
	this.texturePath = '';
}

Object.assign(ObjectLoader.prototype, {

	load: function load(url, onLoad, onProgress, onError) {

		if (this.texturePath === '') {

			this.texturePath = url.substring(0, url.lastIndexOf('/') + 1);
		}

		var scope = this;

		var loader = new XHRLoader(scope.manager);
		loader.load(url, function (text) {

			scope.parse(JSON.parse(text), onLoad);
		}, onProgress, onError);
	},

	setTexturePath: function setTexturePath(value) {

		this.texturePath = value;
	},

	setCrossOrigin: function setCrossOrigin(value) {

		this.crossOrigin = value;
	},

	parse: function parse(json, onLoad) {

		var geometries = this.parseGeometries(json.geometries);

		var images = this.parseImages(json.images, function () {

			if (onLoad !== undefined) onLoad(object);
		});

		var textures = this.parseTextures(json.textures, images);
		var materials = this.parseMaterials(json.materials, textures);

		var object = this.parseObject(json.object, geometries, materials);

		if (json.animations) {

			object.animations = this.parseAnimations(json.animations);
		}

		if (json.images === undefined || json.images.length === 0) {

			if (onLoad !== undefined) onLoad(object);
		}

		return object;
	},

	parseGeometries: function parseGeometries(json) {

		var geometries = {};

		if (json !== undefined) {

			var geometryLoader = new JSONLoader();
			var bufferGeometryLoader = new BufferGeometryLoader();

			for (var i = 0, l = json.length; i < l; i++) {

				var geometry;
				var data = json[i];

				switch (data.type) {

					case 'PlaneGeometry':
					case 'PlaneBufferGeometry':

						geometry = new Geometries[data.type](data.width, data.height, data.widthSegments, data.heightSegments);

						break;

					case 'BoxGeometry':
					case 'BoxBufferGeometry':
					case 'CubeGeometry':
						// backwards compatible

						geometry = new Geometries[data.type](data.width, data.height, data.depth, data.widthSegments, data.heightSegments, data.depthSegments);

						break;

					case 'CircleGeometry':
					case 'CircleBufferGeometry':

						geometry = new Geometries[data.type](data.radius, data.segments, data.thetaStart, data.thetaLength);

						break;

					case 'CylinderGeometry':
					case 'CylinderBufferGeometry':

						geometry = new Geometries[data.type](data.radiusTop, data.radiusBottom, data.height, data.radialSegments, data.heightSegments, data.openEnded, data.thetaStart, data.thetaLength);

						break;

					case 'ConeGeometry':
					case 'ConeBufferGeometry':

						geometry = new Geometries[data.type](data.radius, data.height, data.radialSegments, data.heightSegments, data.openEnded, data.thetaStart, data.thetaLength);

						break;

					case 'SphereGeometry':
					case 'SphereBufferGeometry':

						geometry = new Geometries[data.type](data.radius, data.widthSegments, data.heightSegments, data.phiStart, data.phiLength, data.thetaStart, data.thetaLength);

						break;

					case 'DodecahedronGeometry':
					case 'IcosahedronGeometry':
					case 'OctahedronGeometry':
					case 'TetrahedronGeometry':

						geometry = new Geometries[data.type](data.radius, data.detail);

						break;

					case 'RingGeometry':
					case 'RingBufferGeometry':

						geometry = new Geometries[data.type](data.innerRadius, data.outerRadius, data.thetaSegments, data.phiSegments, data.thetaStart, data.thetaLength);

						break;

					case 'TorusGeometry':
					case 'TorusBufferGeometry':

						geometry = new Geometries[data.type](data.radius, data.tube, data.radialSegments, data.tubularSegments, data.arc);

						break;

					case 'TorusKnotGeometry':
					case 'TorusKnotBufferGeometry':

						geometry = new Geometries[data.type](data.radius, data.tube, data.tubularSegments, data.radialSegments, data.p, data.q);

						break;

					case 'LatheGeometry':
					case 'LatheBufferGeometry':

						geometry = new Geometries[data.type](data.points, data.segments, data.phiStart, data.phiLength);

						break;

					case 'BufferGeometry':

						geometry = bufferGeometryLoader.parse(data);

						break;

					case 'Geometry':

						geometry = geometryLoader.parse(data.data, this.texturePath).geometry;

						break;

					default:

						console.warn('THREE.ObjectLoader: Unsupported geometry type "' + data.type + '"');

						continue;

				}

				geometry.uuid = data.uuid;

				if (data.name !== undefined) geometry.name = data.name;

				geometries[data.uuid] = geometry;
			}
		}

		return geometries;
	},

	parseMaterials: function parseMaterials(json, textures) {

		var materials = {};

		if (json !== undefined) {

			var loader = new MaterialLoader();
			loader.setTextures(textures);

			for (var i = 0, l = json.length; i < l; i++) {

				var material = loader.parse(json[i]);
				materials[material.uuid] = material;
			}
		}

		return materials;
	},

	parseAnimations: function parseAnimations(json) {

		var animations = [];

		for (var i = 0; i < json.length; i++) {

			var clip = AnimationClip.parse(json[i]);

			animations.push(clip);
		}

		return animations;
	},

	parseImages: function parseImages(json, onLoad) {

		var scope = this;
		var images = {};

		function loadImage(url) {

			scope.manager.itemStart(url);

			return loader.load(url, function () {

				scope.manager.itemEnd(url);
			}, undefined, function () {

				scope.manager.itemError(url);
			});
		}

		if (json !== undefined && json.length > 0) {

			var manager = new LoadingManager(onLoad);

			var loader = new ImageLoader(manager);
			loader.setCrossOrigin(this.crossOrigin);

			for (var i = 0, l = json.length; i < l; i++) {

				var image = json[i];
				var path = /^(\/\/)|([a-z]+:(\/\/)?)/i.test(image.url) ? image.url : scope.texturePath + image.url;

				images[image.uuid] = loadImage(path);
			}
		}

		return images;
	},

	parseTextures: function parseTextures(json, images) {

		function parseConstant(value, type) {

			if (typeof value === 'number') return value;

			console.warn('THREE.ObjectLoader.parseTexture: Constant should be in numeric form.', value);

			return type[value];
		}

		var textures = {};

		if (json !== undefined) {

			for (var i = 0, l = json.length; i < l; i++) {

				var data = json[i];

				if (data.image === undefined) {

					console.warn('THREE.ObjectLoader: No "image" specified for', data.uuid);
				}

				if (images[data.image] === undefined) {

					console.warn('THREE.ObjectLoader: Undefined image', data.image);
				}

				var texture = new Texture(images[data.image]);
				texture.needsUpdate = true;

				texture.uuid = data.uuid;

				if (data.name !== undefined) texture.name = data.name;

				if (data.mapping !== undefined) texture.mapping = parseConstant(data.mapping, TextureMapping);

				if (data.offset !== undefined) texture.offset.fromArray(data.offset);
				if (data.repeat !== undefined) texture.repeat.fromArray(data.repeat);
				if (data.wrap !== undefined) {

					texture.wrapS = parseConstant(data.wrap[0], TextureWrapping);
					texture.wrapT = parseConstant(data.wrap[1], TextureWrapping);
				}

				if (data.minFilter !== undefined) texture.minFilter = parseConstant(data.minFilter, TextureFilter);
				if (data.magFilter !== undefined) texture.magFilter = parseConstant(data.magFilter, TextureFilter);
				if (data.anisotropy !== undefined) texture.anisotropy = data.anisotropy;

				if (data.flipY !== undefined) texture.flipY = data.flipY;

				textures[data.uuid] = texture;
			}
		}

		return textures;
	},

	parseObject: function () {

		var matrix = new Matrix4();

		return function parseObject(data, geometries, materials) {

			var object;

			function getGeometry(name) {

				if (geometries[name] === undefined) {

					console.warn('THREE.ObjectLoader: Undefined geometry', name);
				}

				return geometries[name];
			}

			function getMaterial(name) {

				if (name === undefined) return undefined;

				if (materials[name] === undefined) {

					console.warn('THREE.ObjectLoader: Undefined material', name);
				}

				return materials[name];
			}

			switch (data.type) {

				case 'Scene':

					object = new Scene();

					if (data.background !== undefined) {

						if (Number.isInteger(data.background)) {

							object.background = new Color(data.background);
						}
					}

					if (data.fog !== undefined) {

						if (data.fog.type === 'Fog') {

							object.fog = new Fog(data.fog.color, data.fog.near, data.fog.far);
						} else if (data.fog.type === 'FogExp2') {

							object.fog = new FogExp2(data.fog.color, data.fog.density);
						}
					}

					break;

				case 'PerspectiveCamera':

					object = new PerspectiveCamera(data.fov, data.aspect, data.near, data.far);

					if (data.focus !== undefined) object.focus = data.focus;
					if (data.zoom !== undefined) object.zoom = data.zoom;
					if (data.filmGauge !== undefined) object.filmGauge = data.filmGauge;
					if (data.filmOffset !== undefined) object.filmOffset = data.filmOffset;
					if (data.view !== undefined) object.view = Object.assign({}, data.view);

					break;

				case 'OrthographicCamera':

					object = new OrthographicCamera(data.left, data.right, data.top, data.bottom, data.near, data.far);

					break;

				case 'AmbientLight':

					object = new AmbientLight(data.color, data.intensity);

					break;

				case 'DirectionalLight':

					object = new DirectionalLight(data.color, data.intensity);

					break;

				case 'PointLight':

					object = new PointLight(data.color, data.intensity, data.distance, data.decay);

					break;

				case 'SpotLight':

					object = new SpotLight(data.color, data.intensity, data.distance, data.angle, data.penumbra, data.decay);

					break;

				case 'HemisphereLight':

					object = new HemisphereLight(data.color, data.groundColor, data.intensity);

					break;

				case 'Mesh':

					var geometry = getGeometry(data.geometry);
					var material = getMaterial(data.material);

					if (geometry.bones && geometry.bones.length > 0) {

						object = new SkinnedMesh(geometry, material);
					} else {

						object = new Mesh(geometry, material);
					}

					break;

				case 'LOD':

					object = new LOD();

					break;

				case 'Line':

					object = new Line(getGeometry(data.geometry), getMaterial(data.material), data.mode);

					break;

				case 'LineSegments':

					object = new LineSegments(getGeometry(data.geometry), getMaterial(data.material));

					break;

				case 'PointCloud':
				case 'Points':

					object = new Points(getGeometry(data.geometry), getMaterial(data.material));

					break;

				case 'Sprite':

					object = new Sprite(getMaterial(data.material));

					break;

				case 'Group':

					object = new Group();

					break;

				default:

					object = new Object3D();

			}

			object.uuid = data.uuid;

			if (data.name !== undefined) object.name = data.name;
			if (data.matrix !== undefined) {

				matrix.fromArray(data.matrix);
				matrix.decompose(object.position, object.quaternion, object.scale);
			} else {

				if (data.position !== undefined) object.position.fromArray(data.position);
				if (data.rotation !== undefined) object.rotation.fromArray(data.rotation);
				if (data.quaternion !== undefined) object.quaternion.fromArray(data.quaternion);
				if (data.scale !== undefined) object.scale.fromArray(data.scale);
			}

			if (data.castShadow !== undefined) object.castShadow = data.castShadow;
			if (data.receiveShadow !== undefined) object.receiveShadow = data.receiveShadow;

			if (data.shadow) {

				if (data.shadow.bias !== undefined) object.shadow.bias = data.shadow.bias;
				if (data.shadow.radius !== undefined) object.shadow.radius = data.shadow.radius;
				if (data.shadow.mapSize !== undefined) object.shadow.mapSize.fromArray(data.shadow.mapSize);
				if (data.shadow.camera !== undefined) object.shadow.camera = this.parseObject(data.shadow.camera);
			}

			if (data.visible !== undefined) object.visible = data.visible;
			if (data.userData !== undefined) object.userData = data.userData;

			if (data.children !== undefined) {

				for (var child in data.children) {

					object.add(this.parseObject(data.children[child], geometries, materials));
				}
			}

			if (data.type === 'LOD') {

				var levels = data.levels;

				for (var l = 0; l < levels.length; l++) {

					var level = levels[l];
					var child = object.getObjectByProperty('uuid', level.object);

					if (child !== undefined) {

						object.addLevel(child, level.distance);
					}
				}
			}

			return object;
		};
	}()

});

/**
 * @author zz85 / http://www.lab4games.net/zz85/blog
 * Extensible curve object
 *
 * Some common of Curve methods
 * .getPoint(t), getTangent(t)
 * .getPointAt(u), getTangentAt(u)
 * .getPoints(), .getSpacedPoints()
 * .getLength()
 * .updateArcLengths()
 *
 * This following classes subclasses THREE.Curve:
 *
 * -- 2d classes --
 * THREE.LineCurve
 * THREE.QuadraticBezierCurve
 * THREE.CubicBezierCurve
 * THREE.SplineCurve
 * THREE.ArcCurve
 * THREE.EllipseCurve
 *
 * -- 3d classes --
 * THREE.LineCurve3
 * THREE.QuadraticBezierCurve3
 * THREE.CubicBezierCurve3
 * THREE.SplineCurve3
 *
 * A series of curves can be represented as a THREE.CurvePath
 *
 **/

/**************************************************************
 *	Abstract Curve base class
 **************************************************************/

function Curve() {}

Curve.prototype = {

	constructor: Curve,

	// Virtual base class method to overwrite and implement in subclasses
	//	- t [0 .. 1]

	getPoint: function getPoint(t) {

		console.warn("THREE.Curve: Warning, getPoint() not implemented!");
		return null;
	},

	// Get point at relative position in curve according to arc length
	// - u [0 .. 1]

	getPointAt: function getPointAt(u) {

		var t = this.getUtoTmapping(u);
		return this.getPoint(t);
	},

	// Get sequence of points using getPoint( t )

	getPoints: function getPoints(divisions) {

		if (!divisions) divisions = 5;

		var points = [];

		for (var d = 0; d <= divisions; d++) {

			points.push(this.getPoint(d / divisions));
		}

		return points;
	},

	// Get sequence of points using getPointAt( u )

	getSpacedPoints: function getSpacedPoints(divisions) {

		if (!divisions) divisions = 5;

		var points = [];

		for (var d = 0; d <= divisions; d++) {

			points.push(this.getPointAt(d / divisions));
		}

		return points;
	},

	// Get total curve arc length

	getLength: function getLength() {

		var lengths = this.getLengths();
		return lengths[lengths.length - 1];
	},

	// Get list of cumulative segment lengths

	getLengths: function getLengths(divisions) {

		if (!divisions) divisions = this.__arcLengthDivisions ? this.__arcLengthDivisions : 200;

		if (this.cacheArcLengths && this.cacheArcLengths.length === divisions + 1 && !this.needsUpdate) {

			//console.log( "cached", this.cacheArcLengths );
			return this.cacheArcLengths;
		}

		this.needsUpdate = false;

		var cache = [];
		var current,
		    last = this.getPoint(0);
		var p,
		    sum = 0;

		cache.push(0);

		for (p = 1; p <= divisions; p++) {

			current = this.getPoint(p / divisions);
			sum += current.distanceTo(last);
			cache.push(sum);
			last = current;
		}

		this.cacheArcLengths = cache;

		return cache; // { sums: cache, sum:sum }; Sum is in the last element.
	},

	updateArcLengths: function updateArcLengths() {

		this.needsUpdate = true;
		this.getLengths();
	},

	// Given u ( 0 .. 1 ), get a t to find p. This gives you points which are equidistant

	getUtoTmapping: function getUtoTmapping(u, distance) {

		var arcLengths = this.getLengths();

		var i = 0,
		    il = arcLengths.length;

		var targetArcLength; // The targeted u distance value to get

		if (distance) {

			targetArcLength = distance;
		} else {

			targetArcLength = u * arcLengths[il - 1];
		}

		//var time = Date.now();

		// binary search for the index with largest value smaller than target u distance

		var low = 0,
		    high = il - 1,
		    comparison;

		while (low <= high) {

			i = Math.floor(low + (high - low) / 2); // less likely to overflow, though probably not issue here, JS doesn't really have integers, all numbers are floats

			comparison = arcLengths[i] - targetArcLength;

			if (comparison < 0) {

				low = i + 1;
			} else if (comparison > 0) {

				high = i - 1;
			} else {

				high = i;
				break;

				// DONE
			}
		}

		i = high;

		//console.log('b' , i, low, high, Date.now()- time);

		if (arcLengths[i] === targetArcLength) {

			var t = i / (il - 1);
			return t;
		}

		// we could get finer grain at lengths, or use simple interpolation between two points

		var lengthBefore = arcLengths[i];
		var lengthAfter = arcLengths[i + 1];

		var segmentLength = lengthAfter - lengthBefore;

		// determine where we are between the 'before' and 'after' points

		var segmentFraction = (targetArcLength - lengthBefore) / segmentLength;

		// add that fractional amount to t

		var t = (i + segmentFraction) / (il - 1);

		return t;
	},

	// Returns a unit vector tangent at t
	// In case any sub curve does not implement its tangent derivation,
	// 2 points a small delta apart will be used to find its gradient
	// which seems to give a reasonable approximation

	getTangent: function getTangent(t) {

		var delta = 0.0001;
		var t1 = t - delta;
		var t2 = t + delta;

		// Capping in case of danger

		if (t1 < 0) t1 = 0;
		if (t2 > 1) t2 = 1;

		var pt1 = this.getPoint(t1);
		var pt2 = this.getPoint(t2);

		var vec = pt2.clone().sub(pt1);
		return vec.normalize();
	},

	getTangentAt: function getTangentAt(u) {

		var t = this.getUtoTmapping(u);
		return this.getTangent(t);
	},

	computeFrenetFrames: function computeFrenetFrames(segments, closed) {

		// see http://www.cs.indiana.edu/pub/techreports/TR425.pdf

		var normal = new Vector3();

		var tangents = [];
		var normals = [];
		var binormals = [];

		var vec = new Vector3();
		var mat = new Matrix4();

		var i, u, theta;

		// compute the tangent vectors for each segment on the curve

		for (i = 0; i <= segments; i++) {

			u = i / segments;

			tangents[i] = this.getTangentAt(u);
			tangents[i].normalize();
		}

		// select an initial normal vector perpendicular to the first tangent vector,
		// and in the direction of the minimum tangent xyz component

		normals[0] = new Vector3();
		binormals[0] = new Vector3();
		var min = Number.MAX_VALUE;
		var tx = Math.abs(tangents[0].x);
		var ty = Math.abs(tangents[0].y);
		var tz = Math.abs(tangents[0].z);

		if (tx <= min) {

			min = tx;
			normal.set(1, 0, 0);
		}

		if (ty <= min) {

			min = ty;
			normal.set(0, 1, 0);
		}

		if (tz <= min) {

			normal.set(0, 0, 1);
		}

		vec.crossVectors(tangents[0], normal).normalize();

		normals[0].crossVectors(tangents[0], vec);
		binormals[0].crossVectors(tangents[0], normals[0]);

		// compute the slowly-varying normal and binormal vectors for each segment on the curve

		for (i = 1; i <= segments; i++) {

			normals[i] = normals[i - 1].clone();

			binormals[i] = binormals[i - 1].clone();

			vec.crossVectors(tangents[i - 1], tangents[i]);

			if (vec.length() > Number.EPSILON) {

				vec.normalize();

				theta = Math.acos(_Math.clamp(tangents[i - 1].dot(tangents[i]), -1, 1)); // clamp for floating pt errors

				normals[i].applyMatrix4(mat.makeRotationAxis(vec, theta));
			}

			binormals[i].crossVectors(tangents[i], normals[i]);
		}

		// if the curve is closed, postprocess the vectors so the first and last normal vectors are the same

		if (closed === true) {

			theta = Math.acos(_Math.clamp(normals[0].dot(normals[segments]), -1, 1));
			theta /= segments;

			if (tangents[0].dot(vec.crossVectors(normals[0], normals[segments])) > 0) {

				theta = -theta;
			}

			for (i = 1; i <= segments; i++) {

				// twist a little...
				normals[i].applyMatrix4(mat.makeRotationAxis(tangents[i], theta * i));
				binormals[i].crossVectors(tangents[i], normals[i]);
			}
		}

		return {
			tangents: tangents,
			normals: normals,
			binormals: binormals
		};
	}

};

// TODO: Transformation for Curves?

/**************************************************************
 *	3D Curves
 **************************************************************/

// A Factory method for creating new curve subclasses

Curve.create = function (constructor, getPointFunc) {

	constructor.prototype = Object.create(Curve.prototype);
	constructor.prototype.constructor = constructor;
	constructor.prototype.getPoint = getPointFunc;

	return constructor;
};

/**************************************************************
 *	Line
 **************************************************************/

function LineCurve(v1, v2) {

	this.v1 = v1;
	this.v2 = v2;
}

LineCurve.prototype = Object.create(Curve.prototype);
LineCurve.prototype.constructor = LineCurve;

LineCurve.prototype.isLineCurve = true;

LineCurve.prototype.getPoint = function (t) {

	if (t === 1) {

		return this.v2.clone();
	}

	var point = this.v2.clone().sub(this.v1);
	point.multiplyScalar(t).add(this.v1);

	return point;
};

// Line curve is linear, so we can overwrite default getPointAt

LineCurve.prototype.getPointAt = function (u) {

	return this.getPoint(u);
};

LineCurve.prototype.getTangent = function (t) {

	var tangent = this.v2.clone().sub(this.v1);

	return tangent.normalize();
};

/**
 * @author zz85 / http://www.lab4games.net/zz85/blog
 *
 **/

/**************************************************************
 *	Curved Path - a curve path is simply a array of connected
 *  curves, but retains the api of a curve
 **************************************************************/

function CurvePath() {

	this.curves = [];

	this.autoClose = false; // Automatically closes the path
}

CurvePath.prototype = Object.assign(Object.create(Curve.prototype), {

	constructor: CurvePath,

	add: function add(curve) {

		this.curves.push(curve);
	},

	closePath: function closePath() {

		// Add a line curve if start and end of lines are not connected
		var startPoint = this.curves[0].getPoint(0);
		var endPoint = this.curves[this.curves.length - 1].getPoint(1);

		if (!startPoint.equals(endPoint)) {

			this.curves.push(new LineCurve(endPoint, startPoint));
		}
	},

	// To get accurate point with reference to
	// entire path distance at time t,
	// following has to be done:

	// 1. Length of each sub path have to be known
	// 2. Locate and identify type of curve
	// 3. Get t for the curve
	// 4. Return curve.getPointAt(t')

	getPoint: function getPoint(t) {

		var d = t * this.getLength();
		var curveLengths = this.getCurveLengths();
		var i = 0;

		// To think about boundaries points.

		while (i < curveLengths.length) {

			if (curveLengths[i] >= d) {

				var diff = curveLengths[i] - d;
				var curve = this.curves[i];

				var segmentLength = curve.getLength();
				var u = segmentLength === 0 ? 0 : 1 - diff / segmentLength;

				return curve.getPointAt(u);
			}

			i++;
		}

		return null;

		// loop where sum != 0, sum > d , sum+1 <d
	},

	// We cannot use the default THREE.Curve getPoint() with getLength() because in
	// THREE.Curve, getLength() depends on getPoint() but in THREE.CurvePath
	// getPoint() depends on getLength

	getLength: function getLength() {

		var lens = this.getCurveLengths();
		return lens[lens.length - 1];
	},

	// cacheLengths must be recalculated.
	updateArcLengths: function updateArcLengths() {

		this.needsUpdate = true;
		this.cacheLengths = null;
		this.getLengths();
	},

	// Compute lengths and cache them
	// We cannot overwrite getLengths() because UtoT mapping uses it.

	getCurveLengths: function getCurveLengths() {

		// We use cache values if curves and cache array are same length

		if (this.cacheLengths && this.cacheLengths.length === this.curves.length) {

			return this.cacheLengths;
		}

		// Get length of sub-curve
		// Push sums into cached array

		var lengths = [],
		    sums = 0;

		for (var i = 0, l = this.curves.length; i < l; i++) {

			sums += this.curves[i].getLength();
			lengths.push(sums);
		}

		this.cacheLengths = lengths;

		return lengths;
	},

	getSpacedPoints: function getSpacedPoints(divisions) {

		if (!divisions) divisions = 40;

		var points = [];

		for (var i = 0; i <= divisions; i++) {

			points.push(this.getPoint(i / divisions));
		}

		if (this.autoClose) {

			points.push(points[0]);
		}

		return points;
	},

	getPoints: function getPoints(divisions) {

		divisions = divisions || 12;

		var points = [],
		    last;

		for (var i = 0, curves = this.curves; i < curves.length; i++) {

			var curve = curves[i];
			var resolution = curve && curve.isEllipseCurve ? divisions * 2 : curve && curve.isLineCurve ? 1 : curve && curve.isSplineCurve ? divisions * curve.points.length : divisions;

			var pts = curve.getPoints(resolution);

			for (var j = 0; j < pts.length; j++) {

				var point = pts[j];

				if (last && last.equals(point)) continue; // ensures no consecutive points are duplicates

				points.push(point);
				last = point;
			}
		}

		if (this.autoClose && points.length > 1 && !points[points.length - 1].equals(points[0])) {

			points.push(points[0]);
		}

		return points;
	},

	/**************************************************************
  *	Create Geometries Helpers
  **************************************************************/

	/// Generate geometry from path points (for Line or Points objects)

	createPointsGeometry: function createPointsGeometry(divisions) {

		var pts = this.getPoints(divisions);
		return this.createGeometry(pts);
	},

	// Generate geometry from equidistant sampling along the path

	createSpacedPointsGeometry: function createSpacedPointsGeometry(divisions) {

		var pts = this.getSpacedPoints(divisions);
		return this.createGeometry(pts);
	},

	createGeometry: function createGeometry(points) {

		var geometry = new Geometry();

		for (var i = 0, l = points.length; i < l; i++) {

			var point = points[i];
			geometry.vertices.push(new Vector3(point.x, point.y, point.z || 0));
		}

		return geometry;
	}

});

/**************************************************************
 *	Ellipse curve
 **************************************************************/

function EllipseCurve(aX, aY, xRadius, yRadius, aStartAngle, aEndAngle, aClockwise, aRotation) {

	this.aX = aX;
	this.aY = aY;

	this.xRadius = xRadius;
	this.yRadius = yRadius;

	this.aStartAngle = aStartAngle;
	this.aEndAngle = aEndAngle;

	this.aClockwise = aClockwise;

	this.aRotation = aRotation || 0;
}

EllipseCurve.prototype = Object.create(Curve.prototype);
EllipseCurve.prototype.constructor = EllipseCurve;

EllipseCurve.prototype.isEllipseCurve = true;

EllipseCurve.prototype.getPoint = function (t) {

	var twoPi = Math.PI * 2;
	var deltaAngle = this.aEndAngle - this.aStartAngle;
	var samePoints = Math.abs(deltaAngle) < Number.EPSILON;

	// ensures that deltaAngle is 0 .. 2 PI
	while (deltaAngle < 0) {
		deltaAngle += twoPi;
	}while (deltaAngle > twoPi) {
		deltaAngle -= twoPi;
	}if (deltaAngle < Number.EPSILON) {

		if (samePoints) {

			deltaAngle = 0;
		} else {

			deltaAngle = twoPi;
		}
	}

	if (this.aClockwise === true && !samePoints) {

		if (deltaAngle === twoPi) {

			deltaAngle = -twoPi;
		} else {

			deltaAngle = deltaAngle - twoPi;
		}
	}

	var angle = this.aStartAngle + t * deltaAngle;
	var x = this.aX + this.xRadius * Math.cos(angle);
	var y = this.aY + this.yRadius * Math.sin(angle);

	if (this.aRotation !== 0) {

		var cos = Math.cos(this.aRotation);
		var sin = Math.sin(this.aRotation);

		var tx = x - this.aX;
		var ty = y - this.aY;

		// Rotate the point about the center of the ellipse.
		x = tx * cos - ty * sin + this.aX;
		y = tx * sin + ty * cos + this.aY;
	}

	return new Vector2(x, y);
};

/**
 * @author zz85 / http://www.lab4games.net/zz85/blog
 */

var CurveUtils = {

	tangentQuadraticBezier: function tangentQuadraticBezier(t, p0, p1, p2) {

		return 2 * (1 - t) * (p1 - p0) + 2 * t * (p2 - p1);
	},

	// Puay Bing, thanks for helping with this derivative!

	tangentCubicBezier: function tangentCubicBezier(t, p0, p1, p2, p3) {

		return -3 * p0 * (1 - t) * (1 - t) + 3 * p1 * (1 - t) * (1 - t) - 6 * t * p1 * (1 - t) + 6 * t * p2 * (1 - t) - 3 * t * t * p2 + 3 * t * t * p3;
	},

	tangentSpline: function tangentSpline(t, p0, p1, p2, p3) {

		// To check if my formulas are correct

		var h00 = 6 * t * t - 6 * t; // derived from 2t^3 − 3t^2 + 1
		var h10 = 3 * t * t - 4 * t + 1; // t^3 − 2t^2 + t
		var h01 = -6 * t * t + 6 * t; // − 2t3 + 3t2
		var h11 = 3 * t * t - 2 * t; // t3 − t2

		return h00 + h10 + h01 + h11;
	},

	// Catmull-Rom

	interpolate: function interpolate(p0, p1, p2, p3, t) {

		var v0 = (p2 - p0) * 0.5;
		var v1 = (p3 - p1) * 0.5;
		var t2 = t * t;
		var t3 = t * t2;
		return (2 * p1 - 2 * p2 + v0 + v1) * t3 + (-3 * p1 + 3 * p2 - 2 * v0 - v1) * t2 + v0 * t + p1;
	}

};

/**************************************************************
 *	Spline curve
 **************************************************************/

function SplineCurve(points /* array of Vector2 */) {

	this.points = points === undefined ? [] : points;
}

SplineCurve.prototype = Object.create(Curve.prototype);
SplineCurve.prototype.constructor = SplineCurve;

SplineCurve.prototype.isSplineCurve = true;

SplineCurve.prototype.getPoint = function (t) {

	var points = this.points;
	var point = (points.length - 1) * t;

	var intPoint = Math.floor(point);
	var weight = point - intPoint;

	var point0 = points[intPoint === 0 ? intPoint : intPoint - 1];
	var point1 = points[intPoint];
	var point2 = points[intPoint > points.length - 2 ? points.length - 1 : intPoint + 1];
	var point3 = points[intPoint > points.length - 3 ? points.length - 1 : intPoint + 2];

	var interpolate = CurveUtils.interpolate;

	return new Vector2(interpolate(point0.x, point1.x, point2.x, point3.x, weight), interpolate(point0.y, point1.y, point2.y, point3.y, weight));
};

/**************************************************************
 *	Cubic Bezier curve
 **************************************************************/

function CubicBezierCurve(v0, v1, v2, v3) {

	this.v0 = v0;
	this.v1 = v1;
	this.v2 = v2;
	this.v3 = v3;
}

CubicBezierCurve.prototype = Object.create(Curve.prototype);
CubicBezierCurve.prototype.constructor = CubicBezierCurve;

CubicBezierCurve.prototype.getPoint = function (t) {

	var b3 = ShapeUtils.b3;

	return new Vector2(b3(t, this.v0.x, this.v1.x, this.v2.x, this.v3.x), b3(t, this.v0.y, this.v1.y, this.v2.y, this.v3.y));
};

CubicBezierCurve.prototype.getTangent = function (t) {

	var tangentCubicBezier = CurveUtils.tangentCubicBezier;

	return new Vector2(tangentCubicBezier(t, this.v0.x, this.v1.x, this.v2.x, this.v3.x), tangentCubicBezier(t, this.v0.y, this.v1.y, this.v2.y, this.v3.y)).normalize();
};

/**************************************************************
 *	Quadratic Bezier curve
 **************************************************************/

function QuadraticBezierCurve(v0, v1, v2) {

	this.v0 = v0;
	this.v1 = v1;
	this.v2 = v2;
}

QuadraticBezierCurve.prototype = Object.create(Curve.prototype);
QuadraticBezierCurve.prototype.constructor = QuadraticBezierCurve;

QuadraticBezierCurve.prototype.getPoint = function (t) {

	var b2 = ShapeUtils.b2;

	return new Vector2(b2(t, this.v0.x, this.v1.x, this.v2.x), b2(t, this.v0.y, this.v1.y, this.v2.y));
};

QuadraticBezierCurve.prototype.getTangent = function (t) {

	var tangentQuadraticBezier = CurveUtils.tangentQuadraticBezier;

	return new Vector2(tangentQuadraticBezier(t, this.v0.x, this.v1.x, this.v2.x), tangentQuadraticBezier(t, this.v0.y, this.v1.y, this.v2.y)).normalize();
};

var PathPrototype = Object.assign(Object.create(CurvePath.prototype), {

	fromPoints: function fromPoints(vectors) {

		this.moveTo(vectors[0].x, vectors[0].y);

		for (var i = 1, l = vectors.length; i < l; i++) {

			this.lineTo(vectors[i].x, vectors[i].y);
		}
	},

	moveTo: function moveTo(x, y) {

		this.currentPoint.set(x, y); // TODO consider referencing vectors instead of copying?
	},

	lineTo: function lineTo(x, y) {

		var curve = new LineCurve(this.currentPoint.clone(), new Vector2(x, y));
		this.curves.push(curve);

		this.currentPoint.set(x, y);
	},

	quadraticCurveTo: function quadraticCurveTo(aCPx, aCPy, aX, aY) {

		var curve = new QuadraticBezierCurve(this.currentPoint.clone(), new Vector2(aCPx, aCPy), new Vector2(aX, aY));

		this.curves.push(curve);

		this.currentPoint.set(aX, aY);
	},

	bezierCurveTo: function bezierCurveTo(aCP1x, aCP1y, aCP2x, aCP2y, aX, aY) {

		var curve = new CubicBezierCurve(this.currentPoint.clone(), new Vector2(aCP1x, aCP1y), new Vector2(aCP2x, aCP2y), new Vector2(aX, aY));

		this.curves.push(curve);

		this.currentPoint.set(aX, aY);
	},

	splineThru: function splineThru(pts /*Array of Vector*/) {

		var npts = [this.currentPoint.clone()].concat(pts);

		var curve = new SplineCurve(npts);
		this.curves.push(curve);

		this.currentPoint.copy(pts[pts.length - 1]);
	},

	arc: function arc(aX, aY, aRadius, aStartAngle, aEndAngle, aClockwise) {

		var x0 = this.currentPoint.x;
		var y0 = this.currentPoint.y;

		this.absarc(aX + x0, aY + y0, aRadius, aStartAngle, aEndAngle, aClockwise);
	},

	absarc: function absarc(aX, aY, aRadius, aStartAngle, aEndAngle, aClockwise) {

		this.absellipse(aX, aY, aRadius, aRadius, aStartAngle, aEndAngle, aClockwise);
	},

	ellipse: function ellipse(aX, aY, xRadius, yRadius, aStartAngle, aEndAngle, aClockwise, aRotation) {

		var x0 = this.currentPoint.x;
		var y0 = this.currentPoint.y;

		this.absellipse(aX + x0, aY + y0, xRadius, yRadius, aStartAngle, aEndAngle, aClockwise, aRotation);
	},

	absellipse: function absellipse(aX, aY, xRadius, yRadius, aStartAngle, aEndAngle, aClockwise, aRotation) {

		var curve = new EllipseCurve(aX, aY, xRadius, yRadius, aStartAngle, aEndAngle, aClockwise, aRotation);

		if (this.curves.length > 0) {

			// if a previous curve is present, attempt to join
			var firstPoint = curve.getPoint(0);

			if (!firstPoint.equals(this.currentPoint)) {

				this.lineTo(firstPoint.x, firstPoint.y);
			}
		}

		this.curves.push(curve);

		var lastPoint = curve.getPoint(1);
		this.currentPoint.copy(lastPoint);
	}

});

/**
 * @author zz85 / http://www.lab4games.net/zz85/blog
 * Defines a 2d shape plane using paths.
 **/

// STEP 1 Create a path.
// STEP 2 Turn path into shape.
// STEP 3 ExtrudeGeometry takes in Shape/Shapes
// STEP 3a - Extract points from each shape, turn to vertices
// STEP 3b - Triangulate each shape, add faces.

function Shape() {

	Path.apply(this, arguments);

	this.holes = [];
}

Shape.prototype = Object.assign(Object.create(PathPrototype), {

	constructor: Shape,

	getPointsHoles: function getPointsHoles(divisions) {

		var holesPts = [];

		for (var i = 0, l = this.holes.length; i < l; i++) {

			holesPts[i] = this.holes[i].getPoints(divisions);
		}

		return holesPts;
	},

	// Get points of shape and holes (keypoints based on segments parameter)

	extractAllPoints: function extractAllPoints(divisions) {

		return {

			shape: this.getPoints(divisions),
			holes: this.getPointsHoles(divisions)

		};
	},

	extractPoints: function extractPoints(divisions) {

		return this.extractAllPoints(divisions);
	}

});

/**
 * @author zz85 / http://www.lab4games.net/zz85/blog
 * Creates free form 2d path using series of points, lines or curves.
 *
 **/

function Path(points) {

	CurvePath.call(this);
	this.currentPoint = new Vector2();

	if (points) {

		this.fromPoints(points);
	}
}

Path.prototype = PathPrototype;
PathPrototype.constructor = Path;

// minimal class for proxing functions to Path. Replaces old "extractSubpaths()"
function ShapePath() {
	this.subPaths = [];
	this.currentPath = null;
}

ShapePath.prototype = {
	moveTo: function moveTo(x, y) {
		this.currentPath = new Path();
		this.subPaths.push(this.currentPath);
		this.currentPath.moveTo(x, y);
	},
	lineTo: function lineTo(x, y) {
		this.currentPath.lineTo(x, y);
	},
	quadraticCurveTo: function quadraticCurveTo(aCPx, aCPy, aX, aY) {
		this.currentPath.quadraticCurveTo(aCPx, aCPy, aX, aY);
	},
	bezierCurveTo: function bezierCurveTo(aCP1x, aCP1y, aCP2x, aCP2y, aX, aY) {
		this.currentPath.bezierCurveTo(aCP1x, aCP1y, aCP2x, aCP2y, aX, aY);
	},
	splineThru: function splineThru(pts) {
		this.currentPath.splineThru(pts);
	},

	toShapes: function toShapes(isCCW, noHoles) {

		function toShapesNoHoles(inSubpaths) {

			var shapes = [];

			for (var i = 0, l = inSubpaths.length; i < l; i++) {

				var tmpPath = inSubpaths[i];

				var tmpShape = new Shape();
				tmpShape.curves = tmpPath.curves;

				shapes.push(tmpShape);
			}

			return shapes;
		}

		function isPointInsidePolygon(inPt, inPolygon) {

			var polyLen = inPolygon.length;

			// inPt on polygon contour => immediate success    or
			// toggling of inside/outside at every single! intersection point of an edge
			//  with the horizontal line through inPt, left of inPt
			//  not counting lowerY endpoints of edges and whole edges on that line
			var inside = false;
			for (var p = polyLen - 1, q = 0; q < polyLen; p = q++) {

				var edgeLowPt = inPolygon[p];
				var edgeHighPt = inPolygon[q];

				var edgeDx = edgeHighPt.x - edgeLowPt.x;
				var edgeDy = edgeHighPt.y - edgeLowPt.y;

				if (Math.abs(edgeDy) > Number.EPSILON) {

					// not parallel
					if (edgeDy < 0) {

						edgeLowPt = inPolygon[q];edgeDx = -edgeDx;
						edgeHighPt = inPolygon[p];edgeDy = -edgeDy;
					}
					if (inPt.y < edgeLowPt.y || inPt.y > edgeHighPt.y) continue;

					if (inPt.y === edgeLowPt.y) {

						if (inPt.x === edgeLowPt.x) return true; // inPt is on contour ?
						// continue;				// no intersection or edgeLowPt => doesn't count !!!
					} else {

						var perpEdge = edgeDy * (inPt.x - edgeLowPt.x) - edgeDx * (inPt.y - edgeLowPt.y);
						if (perpEdge === 0) return true; // inPt is on contour ?
						if (perpEdge < 0) continue;
						inside = !inside; // true intersection left of inPt
					}
				} else {

					// parallel or collinear
					if (inPt.y !== edgeLowPt.y) continue; // parallel
					// edge lies on the same horizontal line as inPt
					if (edgeHighPt.x <= inPt.x && inPt.x <= edgeLowPt.x || edgeLowPt.x <= inPt.x && inPt.x <= edgeHighPt.x) return true; // inPt: Point on contour !
					// continue;
				}
			}

			return inside;
		}

		var isClockWise = ShapeUtils.isClockWise;

		var subPaths = this.subPaths;
		if (subPaths.length === 0) return [];

		if (noHoles === true) return toShapesNoHoles(subPaths);

		var solid,
		    tmpPath,
		    tmpShape,
		    shapes = [];

		if (subPaths.length === 1) {

			tmpPath = subPaths[0];
			tmpShape = new Shape();
			tmpShape.curves = tmpPath.curves;
			shapes.push(tmpShape);
			return shapes;
		}

		var holesFirst = !isClockWise(subPaths[0].getPoints());
		holesFirst = isCCW ? !holesFirst : holesFirst;

		// console.log("Holes first", holesFirst);

		var betterShapeHoles = [];
		var newShapes = [];
		var newShapeHoles = [];
		var mainIdx = 0;
		var tmpPoints;

		newShapes[mainIdx] = undefined;
		newShapeHoles[mainIdx] = [];

		for (var i = 0, l = subPaths.length; i < l; i++) {

			tmpPath = subPaths[i];
			tmpPoints = tmpPath.getPoints();
			solid = isClockWise(tmpPoints);
			solid = isCCW ? !solid : solid;

			if (solid) {

				if (!holesFirst && newShapes[mainIdx]) mainIdx++;

				newShapes[mainIdx] = { s: new Shape(), p: tmpPoints };
				newShapes[mainIdx].s.curves = tmpPath.curves;

				if (holesFirst) mainIdx++;
				newShapeHoles[mainIdx] = [];

				//console.log('cw', i);
			} else {

				newShapeHoles[mainIdx].push({ h: tmpPath, p: tmpPoints[0] });

				//console.log('ccw', i);
			}
		}

		// only Holes? -> probably all Shapes with wrong orientation
		if (!newShapes[0]) return toShapesNoHoles(subPaths);

		if (newShapes.length > 1) {

			var ambiguous = false;
			var toChange = [];

			for (var sIdx = 0, sLen = newShapes.length; sIdx < sLen; sIdx++) {

				betterShapeHoles[sIdx] = [];
			}

			for (var sIdx = 0, sLen = newShapes.length; sIdx < sLen; sIdx++) {

				var sho = newShapeHoles[sIdx];

				for (var hIdx = 0; hIdx < sho.length; hIdx++) {

					var ho = sho[hIdx];
					var hole_unassigned = true;

					for (var s2Idx = 0; s2Idx < newShapes.length; s2Idx++) {

						if (isPointInsidePolygon(ho.p, newShapes[s2Idx].p)) {

							if (sIdx !== s2Idx) toChange.push({ froms: sIdx, tos: s2Idx, hole: hIdx });
							if (hole_unassigned) {

								hole_unassigned = false;
								betterShapeHoles[s2Idx].push(ho);
							} else {

								ambiguous = true;
							}
						}
					}
					if (hole_unassigned) {

						betterShapeHoles[sIdx].push(ho);
					}
				}
			}
			// console.log("ambiguous: ", ambiguous);
			if (toChange.length > 0) {

				// console.log("to change: ", toChange);
				if (!ambiguous) newShapeHoles = betterShapeHoles;
			}
		}

		var tmpHoles;

		for (var i = 0, il = newShapes.length; i < il; i++) {

			tmpShape = newShapes[i].s;
			shapes.push(tmpShape);
			tmpHoles = newShapeHoles[i];

			for (var j = 0, jl = tmpHoles.length; j < jl; j++) {

				tmpShape.holes.push(tmpHoles[j].h);
			}
		}

		//console.log("shape", shapes);

		return shapes;
	}
};

/**
 * @author zz85 / http://www.lab4games.net/zz85/blog
 * @author mrdoob / http://mrdoob.com/
 */

function Font(data) {

	this.data = data;
}

Object.assign(Font.prototype, {

	isFont: true,

	generateShapes: function generateShapes(text, size, divisions) {

		function createPaths(text) {

			var chars = String(text).split('');
			var scale = size / data.resolution;
			var offset = 0;

			var paths = [];

			for (var i = 0; i < chars.length; i++) {

				var ret = createPath(chars[i], scale, offset);
				offset += ret.offset;

				paths.push(ret.path);
			}

			return paths;
		}

		function createPath(c, scale, offset) {

			var glyph = data.glyphs[c] || data.glyphs['?'];

			if (!glyph) return;

			var path = new ShapePath();

			var pts = [],
			    b2 = ShapeUtils.b2,
			    b3 = ShapeUtils.b3;
			var x, y, cpx, cpy, cpx0, cpy0, cpx1, cpy1, cpx2, cpy2, laste;

			if (glyph.o) {

				var outline = glyph._cachedOutline || (glyph._cachedOutline = glyph.o.split(' '));

				for (var i = 0, l = outline.length; i < l;) {

					var action = outline[i++];

					switch (action) {

						case 'm':
							// moveTo

							x = outline[i++] * scale + offset;
							y = outline[i++] * scale;

							path.moveTo(x, y);

							break;

						case 'l':
							// lineTo

							x = outline[i++] * scale + offset;
							y = outline[i++] * scale;

							path.lineTo(x, y);

							break;

						case 'q':
							// quadraticCurveTo

							cpx = outline[i++] * scale + offset;
							cpy = outline[i++] * scale;
							cpx1 = outline[i++] * scale + offset;
							cpy1 = outline[i++] * scale;

							path.quadraticCurveTo(cpx1, cpy1, cpx, cpy);

							laste = pts[pts.length - 1];

							if (laste) {

								cpx0 = laste.x;
								cpy0 = laste.y;

								for (var i2 = 1; i2 <= divisions; i2++) {

									var t = i2 / divisions;
									b2(t, cpx0, cpx1, cpx);
									b2(t, cpy0, cpy1, cpy);
								}
							}

							break;

						case 'b':
							// bezierCurveTo

							cpx = outline[i++] * scale + offset;
							cpy = outline[i++] * scale;
							cpx1 = outline[i++] * scale + offset;
							cpy1 = outline[i++] * scale;
							cpx2 = outline[i++] * scale + offset;
							cpy2 = outline[i++] * scale;

							path.bezierCurveTo(cpx1, cpy1, cpx2, cpy2, cpx, cpy);

							laste = pts[pts.length - 1];

							if (laste) {

								cpx0 = laste.x;
								cpy0 = laste.y;

								for (var i2 = 1; i2 <= divisions; i2++) {

									var t = i2 / divisions;
									b3(t, cpx0, cpx1, cpx2, cpx);
									b3(t, cpy0, cpy1, cpy2, cpy);
								}
							}

							break;

					}
				}
			}

			return { offset: glyph.ha * scale, path: path };
		}

		//

		if (size === undefined) size = 100;
		if (divisions === undefined) divisions = 4;

		var data = this.data;

		var paths = createPaths(text);
		var shapes = [];

		for (var p = 0, pl = paths.length; p < pl; p++) {

			Array.prototype.push.apply(shapes, paths[p].toShapes());
		}

		return shapes;
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 */

function FontLoader(manager) {

	this.manager = manager !== undefined ? manager : DefaultLoadingManager;
}

Object.assign(FontLoader.prototype, {

	load: function load(url, onLoad, onProgress, onError) {

		var scope = this;

		var loader = new XHRLoader(this.manager);
		loader.load(url, function (text) {

			var json;

			try {

				json = JSON.parse(text);
			} catch (e) {

				console.warn('THREE.FontLoader: typeface.js support is being deprecated. Use typeface.json instead.');
				json = JSON.parse(text.substring(65, text.length - 2));
			}

			var font = scope.parse(json);

			if (onLoad) onLoad(font);
		}, onProgress, onError);
	},

	parse: function parse(json) {

		return new Font(json);
	}

});

BufferGeometry.prototype.center = Geometry.prototype.center = function () {
  this.computeBoundingBox();
  var b = this.boundingBox,
      dx = (b.max.x + b.min.x) / 2,
      dy = (b.max.y + b.min.y) / 2,
      dz = (b.max.z + b.min.z) / 2;
  return this.offset(-dx, -dy, -dz);
};

BufferGeometry.prototype.colored = Geometry.prototype.colored = Mesh.prototype.colored = function (color, options) {
  return colored(this, color, options);
};

CubeTextureLoader.prototype.load = function (urls, onLoad, onProgress, onError) {
  var texture = new CubeTexture();
  var loader = new ImageLoader(this.manager);
  loader.setCrossOrigin(this.crossOrigin);
  loader.setPath(this.path);
  var loaded = 0;
  function loadTexture(i) {
    loader.load(urls[i], function (image) {
      texture.images[i] = image;
      ++loaded;
      if (loaded === 6) {
        texture.needsUpdate = true;
        if (onLoad) onLoad(texture);
      }
    }, onProgress, onError);
  }

  for (var i = 0; i < urls.length; ++i) {
    loadTexture(i);
  }

  return texture;
};

Matrix4.prototype.toString = function (digits) {
  if (digits === undefined) {
    digits = 10;
  }
  this.transpose();
  var parts = this.toArray();
  this.transpose();
  if (digits !== undefined) {
    for (var i = 0; i < parts.length; ++i) {}
  }
  var output = "";
  for (var _i = 0; _i < parts.length; ++_i) {
    if (_i % 4 === 0) {
      output += "| ";
    }
    if (Math.sign(parts[_i]) === -1) {
      output += "-";
    } else {
      output += " ";
    }

    if (parts[_i] !== null && parts[_i] !== undefined) {
      output += Math.abs(parts[_i]).toFixed(digits);
    } else {
      output += "undefined".substring(0, digits);
    }

    if (_i % 4 === 3) {
      output += " |\n";
    } else {
      output += ", ";
    }
  }
  return output;
};

/**
 * Loads a Wavefront .mtl file specifying materials
 *
 * @author angelxuanchang
 *
 * Converted to ES2015 by @capnmidnight
 *
 */
var MTLLoader = function (_EventDispatcher) {
    inherits(MTLLoader, _EventDispatcher);

    function MTLLoader(manager) {
        classCallCheck(this, MTLLoader);

        var _this = possibleConstructorReturn(this, (MTLLoader.__proto__ || Object.getPrototypeOf(MTLLoader)).call(this));

        _this.manager = manager !== undefined ? manager : DefaultLoadingManager;

        return _this;
    }

    /**
     * Loads and parses a MTL asset from a URL.
     *
     * @param {String} url - URL to the MTL file.
     * @param {Function} [onLoad] - Callback invoked with the loaded object.
     * @param {Function} [onProgress] - Callback for download progress.
     * @param {Function} [onError] - Callback for download errors.
     *
     * @see setPath setTexturePath
     *
     * @note In order for relative texture references to resolve correctly
     * you must call setPath and/or setTexturePath explicitly prior to load.
     */


    createClass(MTLLoader, [{
        key: "load",
        value: function load(url, onLoad, onProgress, onError) {

            var scope = this;

            var loader = new XHRLoader(this.manager);
            loader.setPath(this.path);
            loader.load(url, function (text) {

                onLoad(scope.parse(text));
            }, onProgress, onError);
        }

        /**
         * Set base path for resolving references.
         * If set this path will be prepended to each loaded and found reference.
         *
         * @see setTexturePath
         * @param {String} path
         *
         * @example
         *     mtlLoader.setPath( 'assets/obj/' );
         *     mtlLoader.load( 'my.mtl', ... );
         */

    }, {
        key: "setPath",
        value: function setPath(path) {

            this.path = path;
        }

        /**
         * Set base path for resolving texture references.
         * If set this path will be prepended found texture reference.
         * If not set and setPath is, it will be used as texture base path.
         *
         * @see setPath
         * @param {String} path
         *
         * @example
         *     mtlLoader.setPath( 'assets/obj/' );
         *     mtlLoader.setTexturePath( 'assets/textures/' );
         *     mtlLoader.load( 'my.mtl', ... );
         */

    }, {
        key: "setTexturePath",
        value: function setTexturePath(path) {

            this.texturePath = path;
        }
    }, {
        key: "setBaseUrl",
        value: function setBaseUrl(path) {

            console.warn('MTLLoader: .setBaseUrl() is deprecated. Use .setTexturePath( path ) for texture path or .setPath( path ) for general base path instead.');

            this.setTexturePath(path);
        }
    }, {
        key: "setCrossOrigin",
        value: function setCrossOrigin(value) {

            this.crossOrigin = value;
        }
    }, {
        key: "setMaterialOptions",
        value: function setMaterialOptions(value) {

            this.materialOptions = value;
        }

        /**
         * Parses a MTL file.
         *
         * @param {String} text - Content of MTL file
         * @return {MTLLoader.MaterialCreator}
         *
         * @see setPath setTexturePath
         *
         * @note In order for relative texture references to resolve correctly
         * you must call setPath and/or setTexturePath explicitly prior to parse.
         */

    }, {
        key: "parse",
        value: function parse(text) {

            var lines = text.split('\n');
            var info = {};
            var delimiter_pattern = /\s+/;
            var materialsInfo = {};

            for (var i = 0; i < lines.length; i++) {

                var line = lines[i];
                line = line.trim();

                if (line.length === 0 || line.charAt(0) === '#') {

                    // Blank line or comment ignore
                    continue;
                }

                var pos = line.indexOf(' ');

                var key = pos >= 0 ? line.substring(0, pos) : line;
                key = key.toLowerCase();

                var value = pos >= 0 ? line.substring(pos + 1) : '';
                value = value.trim();

                if (key === 'newmtl') {

                    // New material

                    info = { name: value };
                    materialsInfo[value] = info;
                } else if (info) {

                    if (key === 'ka' || key === 'kd' || key === 'ks') {

                        var ss = value.split(delimiter_pattern, 3);
                        info[key] = [parseFloat(ss[0]), parseFloat(ss[1]), parseFloat(ss[2])];
                    } else {

                        info[key] = value;
                    }
                }
            }

            var materialCreator = new MaterialCreator(this.texturePath || this.path, this.materialOptions);
            materialCreator.setCrossOrigin(this.crossOrigin);
            materialCreator.setManager(this.manager);
            materialCreator.setMaterials(materialsInfo);
            return materialCreator;
        }
    }]);
    return MTLLoader;
}(EventDispatcher);

/**
 * Create a new MTLLoader.MaterialCreator
 * @param baseUrl - Url relative to which textures are loaded
 * @param options - Set of options on how to construct the materials
 *                  side: Which side to apply the material
 *                        THREE.FrontSide (default), THREE.BackSide, THREE.DoubleSide
 *                  wrap: What type of wrapping to apply for textures
 *                        THREE.RepeatWrapping (default), THREE.ClampToEdgeWrapping, THREE.MirroredRepeatWrapping
 *                  normalizeRGB: RGBs need to be normalized to 0-1 from 0-255
 *                                Default: false, assumed to be already normalized
 *                  ignoreZeroRGBs: Ignore values of RGBs (Ka,Kd,Ks) that are all 0's
 *                                  Default: false
 * @constructor
 */

var MaterialCreator = function () {
    function MaterialCreator(baseUrl, options) {
        classCallCheck(this, MaterialCreator);


        this.baseUrl = baseUrl || '';
        this.options = options;
        this.materialsInfo = {};
        this.materials = {};
        this.materialsArray = [];
        this.nameLookup = {};

        this.side = this.options && this.options.side ? this.options.side : FrontSide;
        this.wrap = this.options && this.options.wrap ? this.options.wrap : RepeatWrapping;
    }

    createClass(MaterialCreator, [{
        key: "setCrossOrigin",
        value: function setCrossOrigin(value) {

            this.crossOrigin = value;
        }
    }, {
        key: "setManager",
        value: function setManager(value) {

            this.manager = value;
        }
    }, {
        key: "setMaterials",
        value: function setMaterials(materialsInfo) {

            this.materialsInfo = this.convert(materialsInfo);
            this.materials = {};
            this.materialsArray = [];
            this.nameLookup = {};
        }
    }, {
        key: "convert",
        value: function convert(materialsInfo) {

            if (!this.options) return materialsInfo;

            var converted = {};

            for (var mn in materialsInfo) {

                // Convert materials info into normalized form based on options

                var mat = materialsInfo[mn];

                var covmat = {};

                converted[mn] = covmat;

                for (var prop in mat) {

                    var save = true;
                    var value = mat[prop];
                    var lprop = prop.toLowerCase();

                    switch (lprop) {

                        case 'kd':
                        case 'ka':
                        case 'ks':

                            // Diffuse color (color under white light) using RGB values

                            if (this.options && this.options.normalizeRGB) {

                                value = [value[0] / 255, value[1] / 255, value[2] / 255];
                            }

                            if (this.options && this.options.ignoreZeroRGBs) {

                                if (value[0] === 0 && value[1] === 0 && value[2] === 0) {

                                    // ignore

                                    save = false;
                                }
                            }

                            break;

                        default:

                            break;
                    }

                    if (save) {

                        covmat[lprop] = value;
                    }
                }
            }

            return converted;
        }
    }, {
        key: "preload",
        value: function preload() {

            for (var mn in this.materialsInfo) {

                this.create(mn);
            }
        }
    }, {
        key: "getIndex",
        value: function getIndex(materialName) {

            return this.nameLookup[materialName];
        }
    }, {
        key: "getAsArray",
        value: function getAsArray() {

            var index = 0;

            for (var mn in this.materialsInfo) {

                this.materialsArray[index] = this.create(mn);
                this.nameLookup[mn] = index;
                index++;
            }

            return this.materialsArray;
        }
    }, {
        key: "create",
        value: function create(materialName) {

            if (this.materials[materialName] === undefined) {

                this.createMaterial_(materialName);
            }

            return this.materials[materialName];
        }
    }, {
        key: "createMaterial_",
        value: function createMaterial_(materialName) {

            // Create material

            var TMaterial = MeshPhongMaterial;
            var scope = this;
            var mat = this.materialsInfo[materialName];
            var params = {

                name: materialName,
                side: this.side

            };

            var resolveURL = function resolveURL(baseUrl, url) {

                if (typeof url !== 'string' || url === '') return '';

                // Absolute URL
                if (/^https?:\/\//i.test(url)) {
                    return url;
                }

                return baseUrl + url;
            };

            function setMapForType(mapType, value) {

                if (params[mapType]) return; // Keep the first encountered texture

                var texParams = scope.getTextureParams(value, params);
                var map = scope.loadTexture(resolveURL(scope.baseUrl, texParams.url));

                map.repeat.copy(texParams.scale);
                map.offset.copy(texParams.offset);

                map.wrapS = scope.wrap;
                map.wrapT = scope.wrap;

                params[mapType] = map;
            }

            for (var prop in mat) {

                var value = mat[prop];

                if (value === '') continue;

                switch (prop.toLowerCase()) {

                    // Ns is material specular exponent

                    case 'kd':

                        // Diffuse color (color under white light) using RGB values

                        params.color = new Color().fromArray(value);

                        break;

                    case 'ks':

                        // Specular color (color when light is reflected from shiny surface) using RGB values
                        params.specular = new Color().fromArray(value);

                        break;

                    case 'map_kd':

                        // Diffuse texture map

                        setMapForType("map", value);

                        break;

                    case 'map_ks':

                        // Specular map

                        setMapForType("specularMap", value);

                        break;

                    case 'map_bump':
                    case 'bump':

                        // Bump texture map

                        setMapForType("bumpMap", value);

                        break;

                    case 'ns':

                        // The specular exponent (defines the focus of the specular highlight)
                        // A high exponent results in a tight, concentrated highlight. Ns values normally range from 0 to 1000.

                        params.shininess = parseFloat(value);

                        break;

                    case 'd':

                        if (value < 1) {

                            params.opacity = value;
                            params.transparent = true;
                        }

                        break;

                    case 'illum':

                        value = parseFloat(value);

                        if (value === MTLLoader.COLOR_ON_AND_AMBIENT_OFF) {

                            TMaterial = MeshBasicMaterial;
                        }

                        break;

                    case 'Tr':

                        if (value > 0) {

                            params.opacity = 1 - value;
                            params.transparent = true;
                        }

                        break;

                    default:
                        break;

                }
            }

            if (TMaterial === MeshBasicMaterial) {

                ["shininess", "specular"].forEach(function (attribute) {

                    if (attribute in params) {

                        delete params[attribute];
                    }
                });
            }

            this.materials[materialName] = new TMaterial(params);
            return this.materials[materialName];
        }
    }, {
        key: "getTextureParams",
        value: function getTextureParams(value, matParams) {

            var texParams = {

                scale: new Vector2(1, 1),
                offset: new Vector2(0, 0)

            };

            var items = value.split(/\s+/);
            var pos;

            pos = items.indexOf('-bm');
            if (pos >= 0) {

                matParams.bumpScale = parseFloat(items[pos + 1]);
                items.splice(pos, 2);
            }

            pos = items.indexOf('-s');
            if (pos >= 0) {

                texParams.scale.set(parseFloat(items[pos + 1]), parseFloat(items[pos + 2]));
                items.splice(pos, 4); // we expect 3 parameters here!
            }

            pos = items.indexOf('-o');
            if (pos >= 0) {

                texParams.offset.set(parseFloat(items[pos + 1]), parseFloat(items[pos + 2]));
                items.splice(pos, 4); // we expect 3 parameters here!
            }

            texParams.url = items.join(' ').trim();
            return texParams;
        }
    }, {
        key: "loadTexture",
        value: function loadTexture(url, mapping, onLoad, onProgress, onError) {

            var texture;
            var loader = Loader.Handlers.get(url);
            var manager = this.manager !== undefined ? this.manager : DefaultLoadingManager;

            if (loader === null) {

                loader = new TextureLoader(manager);
            }

            if (loader.setCrossOrigin) loader.setCrossOrigin(this.crossOrigin);
            texture = loader.load(url, onLoad, onProgress, onError);

            if (mapping !== undefined) texture.mapping = mapping;

            return texture;
        }
    }]);
    return MaterialCreator;
}();



// http://paulbourke.net/dataformats/mtl/
Object.assign(MTLLoader, {
    COLOR_ON_AND_AMBIENT_OFF: 0,
    COLOR_ON_AND_AMBIENT_ON: 1,
    HIGHLIGHT_ON: 2,
    REFLECTION_ON_AND_RAY_TRACE_ON: 3,
    TRANSPARENCY_GLASS_ON_REFLECTION_RAY_TRACE_ON: 4,
    REFLECTION_FRESNEL_ON_AND_RAY_TRACE_ON: 5,
    TRANSPARENCY_REFRACTION_ON_REFLECTION_FRESNEL_OFF_AND_RAY_TRACE_ON: 6,
    TRANSPARENCY_REFRACTION_ON_REFLECTION_FRESNEL_ON_AND_RAY_TRACE_ON: 7,
    REFLECTION_ON_AND_RAY_TRACE_OFF: 8,
    TRANSPARENCY_GLASS_ON_REFLECTION_RAY_TRACE_OFF: 9,
    CASTS_SHADOWS_ONTO_INVISIBLE_SURFACES: 10
});

Object3D.prototype.addToBrowserEnvironment = function (env, scene) {
  var _this = this;

  scene.add(this);
  // this has to be done as a lambda expression because it needs to capture the
  // env variable provided in the addToBrowserEnvironment call;

  this.appendChild = function (child) {
    if (child.addToBrowserEnvironment) {
      return child.addToBrowserEnvironment(env, _this);
    } else {
      _this.add(child);
      env.registerPickableObject(child);
      return child;
    }
  };
};

Object3D.prototype.latLon = function (lat, lon, r) {
  lat = -Math.PI * (lat || 0) / 180;
  lon = Math.PI * (lon || 0) / 180;
  r = r || 1.5;
  this.rotation.set(lat, lon, 0, "XYZ");
  this.position.set(0, 0, -r);
  this.position.applyQuaternion(this.quaternion);
  return this;
};

Object3D.prototype.named = function (name) {
  this.name = name;
  return this;
};

/**
 * @author mrdoob / http://mrdoob.com/
 */

var OBJLoader = function () {
		function OBJLoader(manager) {
				classCallCheck(this, OBJLoader);


				this.manager = manager !== undefined ? manager : DefaultLoadingManager;

				this.materials = null;

				this.regexp = {
						// v float float float
						vertex_pattern: /^v\s+([\d|\.|\+|\-|e|E]+)\s+([\d|\.|\+|\-|e|E]+)\s+([\d|\.|\+|\-|e|E]+)/,
						// vn float float float
						normal_pattern: /^vn\s+([\d|\.|\+|\-|e|E]+)\s+([\d|\.|\+|\-|e|E]+)\s+([\d|\.|\+|\-|e|E]+)/,
						// vt float float
						uv_pattern: /^vt\s+([\d|\.|\+|\-|e|E]+)\s+([\d|\.|\+|\-|e|E]+)/,
						// f vertex vertex vertex
						face_vertex: /^f\s+(-?\d+)\s+(-?\d+)\s+(-?\d+)(?:\s+(-?\d+))?/,
						// f vertex/uv vertex/uv vertex/uv
						face_vertex_uv: /^f\s+(-?\d+)\/(-?\d+)\s+(-?\d+)\/(-?\d+)\s+(-?\d+)\/(-?\d+)(?:\s+(-?\d+)\/(-?\d+))?/,
						// f vertex/uv/normal vertex/uv/normal vertex/uv/normal
						face_vertex_uv_normal: /^f\s+(-?\d+)\/(-?\d+)\/(-?\d+)\s+(-?\d+)\/(-?\d+)\/(-?\d+)\s+(-?\d+)\/(-?\d+)\/(-?\d+)(?:\s+(-?\d+)\/(-?\d+)\/(-?\d+))?/,
						// f vertex//normal vertex//normal vertex//normal
						face_vertex_normal: /^f\s+(-?\d+)\/\/(-?\d+)\s+(-?\d+)\/\/(-?\d+)\s+(-?\d+)\/\/(-?\d+)(?:\s+(-?\d+)\/\/(-?\d+))?/,
						// o object_name | g group_name
						object_pattern: /^[og]\s*(.+)?/,
						// s boolean
						smoothing_pattern: /^s\s+(\d+|on|off)/,
						// mtllib file_reference
						material_library_pattern: /^mtllib /,
						// usemtl material_name
						material_use_pattern: /^usemtl /
				};
		}

		createClass(OBJLoader, [{
				key: "load",
				value: function load(url, onLoad, onProgress, onError) {

						var scope = this;

						var loader = new XHRLoader(scope.manager);
						loader.setPath(this.path);
						loader.load(url, function (text) {

								onLoad(scope.parse(text));
						}, onProgress, onError);
				}
		}, {
				key: "setPath",
				value: function setPath(value) {

						this.path = value;
				}
		}, {
				key: "setMaterials",
				value: function setMaterials(materials) {

						this.materials = materials;
				}
		}, {
				key: "_createParserState",
				value: function _createParserState() {

						var state = new OBJParserState();

						state.startObject('', false);

						return state;
				}
		}, {
				key: "parse",
				value: function parse(text) {

						console.time('OBJLoader');

						var state = this._createParserState();

						if (text.indexOf('\r\n') !== -1) {

								// This is faster than String.split with regex that splits on both
								text = text.replace('\r\n', '\n');
						}

						var lines = text.split('\n');
						var line = '',
						    lineFirstChar = '',
						    lineSecondChar = '';
						var lineLength = 0;
						var result = [];

						// Faster to just trim left side of the line. Use if available.
						var trimLeft = typeof ''.trimLeft === 'function';

						for (var i = 0, l = lines.length; i < l; i++) {

								line = lines[i];

								line = trimLeft ? line.trimLeft() : line.trim();

								lineLength = line.length;

								if (lineLength === 0) continue;

								lineFirstChar = line.charAt(0);

								// @todo invoke passed in handler if any
								if (lineFirstChar === '#') continue;

								if (lineFirstChar === 'v') {

										lineSecondChar = line.charAt(1);

										if (lineSecondChar === ' ' && (result = this.regexp.vertex_pattern.exec(line)) !== null) {

												// 0                  1      2      3
												// ["v 1.0 2.0 3.0", "1.0", "2.0", "3.0"]

												state.vertices.push(parseFloat(result[1]), parseFloat(result[2]), parseFloat(result[3]));
										} else if (lineSecondChar === 'n' && (result = this.regexp.normal_pattern.exec(line)) !== null) {

												// 0                   1      2      3
												// ["vn 1.0 2.0 3.0", "1.0", "2.0", "3.0"]

												state.normals.push(parseFloat(result[1]), parseFloat(result[2]), parseFloat(result[3]));
										} else if (lineSecondChar === 't' && (result = this.regexp.uv_pattern.exec(line)) !== null) {

												// 0               1      2
												// ["vt 0.1 0.2", "0.1", "0.2"]

												state.uvs.push(parseFloat(result[1]), parseFloat(result[2]));
										} else {

												throw new Error("Unexpected vertex/normal/uv line: '" + line + "'");
										}
								} else if (lineFirstChar === "f") {

										if ((result = this.regexp.face_vertex_uv_normal.exec(line)) !== null) {

												// f vertex/uv/normal vertex/uv/normal vertex/uv/normal
												// 0                        1    2    3    4    5    6    7    8    9   10         11         12
												// ["f 1/1/1 2/2/2 3/3/3", "1", "1", "1", "2", "2", "2", "3", "3", "3", undefined, undefined, undefined]

												state.addFace(result[1], result[4], result[7], result[10], result[2], result[5], result[8], result[11], result[3], result[6], result[9], result[12]);
										} else if ((result = this.regexp.face_vertex_uv.exec(line)) !== null) {

												// f vertex/uv vertex/uv vertex/uv
												// 0                  1    2    3    4    5    6   7          8
												// ["f 1/1 2/2 3/3", "1", "1", "2", "2", "3", "3", undefined, undefined]

												state.addFace(result[1], result[3], result[5], result[7], result[2], result[4], result[6], result[8]);
										} else if ((result = this.regexp.face_vertex_normal.exec(line)) !== null) {

												// f vertex//normal vertex//normal vertex//normal
												// 0                     1    2    3    4    5    6   7          8
												// ["f 1//1 2//2 3//3", "1", "1", "2", "2", "3", "3", undefined, undefined]

												state.addFace(result[1], result[3], result[5], result[7], undefined, undefined, undefined, undefined, result[2], result[4], result[6], result[8]);
										} else if ((result = this.regexp.face_vertex.exec(line)) !== null) {

												// f vertex vertex vertex
												// 0            1    2    3   4
												// ["f 1 2 3", "1", "2", "3", undefined]

												state.addFace(result[1], result[2], result[3], result[4]);
										} else {

												throw new Error("Unexpected face line: '" + line + "'");
										}
								} else if (lineFirstChar === "l") {

										var lineParts = line.substring(1).trim().split(" ");
										var lineVertices = [],
										    lineUVs = [];

										if (line.indexOf("/") === -1) {

												lineVertices = lineParts;
										} else {

												for (var li = 0, llen = lineParts.length; li < llen; li++) {

														var parts = lineParts[li].split("/");

														if (parts[0] !== "") lineVertices.push(parts[0]);
														if (parts[1] !== "") lineUVs.push(parts[1]);
												}
										}
										state.addLineGeometry(lineVertices, lineUVs);
								} else if ((result = this.regexp.object_pattern.exec(line)) !== null) {

										// o object_name
										// or
										// g group_name

										var name = result[0].substr(1).trim();
										state.startObject(name);
								} else if (this.regexp.material_use_pattern.test(line)) {

										// material

										state.object.startMaterial(line.substring(7).trim(), state.materialLibraries);
								} else if (this.regexp.material_library_pattern.test(line)) {

										// mtl file

										state.materialLibraries.push(line.substring(7).trim());
								} else if ((result = this.regexp.smoothing_pattern.exec(line)) !== null) {

										// smooth shading

										// @todo Handle files that have varying smooth values for a set of faces inside one geometry,
										// but does not define a usemtl for each face set.
										// This should be detected and a dummy material created (later MultiMaterial and geometry groups).
										// This requires some care to not create extra material on each smooth value for "normal" obj files.
										// where explicit usemtl defines geometry groups.
										// Example asset: examples/models/obj/cerberus/Cerberus.obj

										var value = result[1].trim().toLowerCase();
										state.object.smooth = value === '1' || value === 'on';

										var material = state.object.currentMaterial();
										if (material) {

												material.smooth = state.object.smooth;
										}
								} else {

										// Handle null terminated files without exception
										if (line === '\0') continue;

										throw new Error("Unexpected line: '" + line + "'");
								}
						}

						state.finalize();

						var container = new Group();
						container.materialLibraries = [].concat(state.materialLibraries);

						for (var i = 0, l = state.objects.length; i < l; i++) {

								var object = state.objects[i];
								var geometry = object.geometry;
								var materials = object.materials;
								var isLine = geometry.type === 'Line';

								// Skip o/g line declarations that did not follow with any faces
								if (geometry.vertices.length === 0) continue;

								var buffergeometry = new BufferGeometry();

								buffergeometry.addAttribute('position', new BufferAttribute(new Float32Array(geometry.vertices), 3));

								if (geometry.normals.length > 0) {

										buffergeometry.addAttribute('normal', new BufferAttribute(new Float32Array(geometry.normals), 3));
								} else {

										buffergeometry.computeVertexNormals();
								}

								if (geometry.uvs.length > 0) {

										buffergeometry.addAttribute('uv', new BufferAttribute(new Float32Array(geometry.uvs), 2));
								}

								// Create materials

								var createdMaterials = [];

								for (var mi = 0, miLen = materials.length; mi < miLen; mi++) {

										var sourceMaterial = materials[mi];
										var material = undefined;

										if (this.materials !== null) {

												material = this.materials.create(sourceMaterial.name);

												// mtl etc. loaders probably can't create line materials correctly, copy properties to a line material.
												if (isLine && material && !material.isLineBasicMaterial) {

														var materialLine = new LineBasicMaterial();
														materialLine.copy(material);
														material = materialLine;
												}
										}

										if (!material) {

												material = !isLine ? new MeshPhongMaterial() : new LineBasicMaterial();
												material.name = sourceMaterial.name;
										}

										material.shading = sourceMaterial.smooth ? SmoothShading : FlatShading;

										createdMaterials.push(material);
								}

								// Create mesh

								var mesh;

								if (createdMaterials.length > 1) {

										for (var mi = 0, miLen = materials.length; mi < miLen; mi++) {

												var sourceMaterial = materials[mi];
												buffergeometry.addGroup(sourceMaterial.groupStart, sourceMaterial.groupCount, mi);
										}

										var multiMaterial = new MultiMaterial(createdMaterials);
										mesh = !isLine ? new Mesh(buffergeometry, multiMaterial) : new LineSegments(buffergeometry, multiMaterial);
								} else {

										mesh = !isLine ? new Mesh(buffergeometry, createdMaterials[0]) : new LineSegments(buffergeometry, createdMaterials[0]);
								}

								mesh.name = object.name;

								container.add(mesh);
						}

						console.timeEnd('OBJLoader');

						return container;
				}
		}]);
		return OBJLoader;
}();

var OBJParserState = function () {
		function OBJParserState() {
				classCallCheck(this, OBJParserState);

				this.objects = [];
				this.object = {};

				this.vertices = [];
				this.normals = [];
				this.uvs = [];

				this.materialLibraries = [];
		}

		createClass(OBJParserState, [{
				key: "startObject",
				value: function startObject(name, fromDeclaration) {

						// If the current object (initial from reset) is not from a g/o declaration in the parsed
						// file. We need to use it for the first parsed g/o to keep things in sync.
						if (this.object && this.object.fromDeclaration === false) {

								this.object.name = name;
								this.object.fromDeclaration = fromDeclaration !== false;
								return;
						}

						if (this.object && typeof this.object._finalize === 'function') {

								this.object._finalize();
						}

						var previousMaterial = this.object && typeof this.object.currentMaterial === 'function' ? this.object.currentMaterial() : undefined;

						this.object = new OBJ(name, fromDeclaration);

						// Inherit previous objects material.
						// Spec tells us that a declared material must be set to all objects until a new material is declared.
						// If a usemtl declaration is encountered while this new object is being parsed, it will
						// overwrite the inherited material. Exception being that there was already face declarations
						// to the inherited material, then it will be preserved for proper MultiMaterial continuation.

						if (previousMaterial && previousMaterial.name && typeof previousMaterial.clone === "function") {

								var declared = previousMaterial.clone(0);
								declared.inherited = true;
								this.object.materials.push(declared);
						}

						this.objects.push(this.object);
				}
		}, {
				key: "finalize",
				value: function finalize() {

						if (this.object && typeof this.object._finalize === 'function') {

								this.object._finalize();
						}
				}
		}, {
				key: "parseVertexIndex",
				value: function parseVertexIndex(value, len) {

						var index = parseInt(value, 10);
						return (index >= 0 ? index - 1 : index + len / 3) * 3;
				}
		}, {
				key: "parseNormalIndex",
				value: function parseNormalIndex(value, len) {

						var index = parseInt(value, 10);
						return (index >= 0 ? index - 1 : index + len / 3) * 3;
				}
		}, {
				key: "parseUVIndex",
				value: function parseUVIndex(value, len) {

						var index = parseInt(value, 10);
						return (index >= 0 ? index - 1 : index + len / 2) * 2;
				}
		}, {
				key: "addVertex",
				value: function addVertex(a, b, c) {

						var src = this.vertices;
						var dst = this.object.geometry.vertices;

						dst.push(src[a + 0]);
						dst.push(src[a + 1]);
						dst.push(src[a + 2]);
						dst.push(src[b + 0]);
						dst.push(src[b + 1]);
						dst.push(src[b + 2]);
						dst.push(src[c + 0]);
						dst.push(src[c + 1]);
						dst.push(src[c + 2]);
				}
		}, {
				key: "addVertexLine",
				value: function addVertexLine(a) {

						var src = this.vertices;
						var dst = this.object.geometry.vertices;

						dst.push(src[a + 0]);
						dst.push(src[a + 1]);
						dst.push(src[a + 2]);
				}
		}, {
				key: "addNormal",
				value: function addNormal(a, b, c) {

						var src = this.normals;
						var dst = this.object.geometry.normals;

						dst.push(src[a + 0]);
						dst.push(src[a + 1]);
						dst.push(src[a + 2]);
						dst.push(src[b + 0]);
						dst.push(src[b + 1]);
						dst.push(src[b + 2]);
						dst.push(src[c + 0]);
						dst.push(src[c + 1]);
						dst.push(src[c + 2]);
				}
		}, {
				key: "addUV",
				value: function addUV(a, b, c) {

						var src = this.uvs;
						var dst = this.object.geometry.uvs;

						dst.push(src[a + 0]);
						dst.push(src[a + 1]);
						dst.push(src[b + 0]);
						dst.push(src[b + 1]);
						dst.push(src[c + 0]);
						dst.push(src[c + 1]);
				}
		}, {
				key: "addUVLine",
				value: function addUVLine(a) {

						var src = this.uvs;
						var dst = this.object.geometry.uvs;

						dst.push(src[a + 0]);
						dst.push(src[a + 1]);
				}
		}, {
				key: "addFace",
				value: function addFace(a, b, c, d, ua, ub, uc, ud, na, nb, nc, nd) {

						var vLen = this.vertices.length;

						var ia = this.parseVertexIndex(a, vLen);
						var ib = this.parseVertexIndex(b, vLen);
						var ic = this.parseVertexIndex(c, vLen);
						var id;

						if (d === undefined) {

								this.addVertex(ia, ib, ic);
						} else {

								id = this.parseVertexIndex(d, vLen);

								this.addVertex(ia, ib, id);
								this.addVertex(ib, ic, id);
						}

						if (ua !== undefined) {

								var uvLen = this.uvs.length;

								ia = this.parseUVIndex(ua, uvLen);
								ib = this.parseUVIndex(ub, uvLen);
								ic = this.parseUVIndex(uc, uvLen);

								if (d === undefined) {

										this.addUV(ia, ib, ic);
								} else {

										id = this.parseUVIndex(ud, uvLen);

										this.addUV(ia, ib, id);
										this.addUV(ib, ic, id);
								}
						}

						if (na !== undefined) {

								// Normals are many times the same. If so, skip function call and parseInt.
								var nLen = this.normals.length;
								ia = this.parseNormalIndex(na, nLen);

								ib = na === nb ? ia : this.parseNormalIndex(nb, nLen);
								ic = na === nc ? ia : this.parseNormalIndex(nc, nLen);

								if (d === undefined) {

										this.addNormal(ia, ib, ic);
								} else {

										id = this.parseNormalIndex(nd, nLen);

										this.addNormal(ia, ib, id);
										this.addNormal(ib, ic, id);
								}
						}
				}
		}, {
				key: "addLineGeometry",
				value: function addLineGeometry(vertices, uvs) {

						this.object.geometry.type = 'Line';

						var vLen = this.vertices.length;
						var uvLen = this.uvs.length;

						for (var vi = 0, l = vertices.length; vi < l; vi++) {

								this.addVertexLine(this.parseVertexIndex(vertices[vi], vLen));
						}

						for (var uvi = 0, l = uvs.length; uvi < l; uvi++) {

								this.addUVLine(this.parseUVIndex(uvs[uvi], uvLen));
						}
				}
		}]);
		return OBJParserState;
}();

var OBJ = function () {
		function OBJ(name, fromDeclaration) {
				classCallCheck(this, OBJ);


				this.name = name || '';
				this.fromDeclaration = fromDeclaration !== false;

				this.geometry = {
						vertices: [],
						normals: [],
						uvs: []
				};
				this.materials = [];
				this.smooth = true;
		}

		createClass(OBJ, [{
				key: "startMaterial",
				value: function startMaterial(name, libraries) {

						var previous = this._finalize(false);

						// New usemtl declaration overwrites an inherited material, except if faces were declared
						// after the material, then it must be preserved for proper MultiMaterial continuation.
						if (previous && (previous.inherited || previous.groupCount <= 0)) {

								this.materials.splice(previous.index, 1);
						}

						var material = {
								index: this.materials.length,
								name: name || '',
								mtllib: Array.isArray(libraries) && libraries.length > 0 ? libraries[libraries.length - 1] : '',
								smooth: previous !== undefined ? previous.smooth : this.smooth,
								groupStart: previous !== undefined ? previous.groupEnd : 0,
								groupEnd: -1,
								groupCount: -1,
								inherited: false,

								clone: function clone(index) {
										return {
												index: typeof index === 'number' ? index : this.index,
												name: this.name,
												mtllib: this.mtllib,
												smooth: this.smooth,
												groupStart: this.groupEnd,
												groupEnd: -1,
												groupCount: -1,
												inherited: false
										};
								}
						};

						this.materials.push(material);

						return material;
				}
		}, {
				key: "currentMaterial",
				value: function currentMaterial() {

						if (this.materials.length > 0) {
								return this.materials[this.materials.length - 1];
						}

						return undefined;
				}
		}, {
				key: "_finalize",
				value: function _finalize(end) {

						var lastMultiMaterial = this.currentMaterial();
						if (lastMultiMaterial && lastMultiMaterial.groupEnd === -1) {

								lastMultiMaterial.groupEnd = this.geometry.vertices.length / 3;
								lastMultiMaterial.groupCount = lastMultiMaterial.groupEnd - lastMultiMaterial.groupStart;
								lastMultiMaterial.inherited = false;
						}

						// Guarantee at least one empty material, this makes the creation later more straight forward.
						if (end !== false && this.materials.length === 0) {
								this.materials.push({
										name: '',
										smooth: this.smooth
								});
						}

						return lastMultiMaterial;
				}
		}]);
		return OBJ;
}();

Geometry.prototype.offset = function (x, y, z) {
  var arr = this.vertices;
  for (var i = 0; i < arr.length; ++i) {
    var vert = arr[i];
    vert.x += x;
    vert.y += y;
    vert.z += z;
  }
  return this;
};

BufferGeometry.prototype.offset = function (x, y, z) {
  var arr = this.attributes.position.array,
      l = this.attributes.position.itemSize;
  for (var i = 0; i < arr.length; i += l) {
    arr[i] += x;
    arr[i + 1] += y;
    arr[i + 2] += z;
  }
  return this;
};

BufferGeometry.prototype.textured = Geometry.prototype.textured = Mesh.prototype.textured = function (texture, options) {
  return textured(this, texture, options);
};

Euler.prototype.toString = Quaternion.prototype.toString = Vector2.prototype.toString = Vector3.prototype.toString = Vector4.prototype.toString = function (digits) {
  var parts = this.toArray();
  if (digits !== undefined) {
    for (var i = 0; i < parts.length; ++i) {
      if (parts[i] !== null && parts[i] !== undefined) {
        parts[i] = parts[i].toFixed(digits);
      } else {
        parts[i] = "undefined";
      }
    }
  }
  return "<" + parts.join(", ") + ">";
};

var cache$1 = {};
Euler.prototype.debug = Quaternion.prototype.debug = Vector2.prototype.debug = Vector3.prototype.debug = Vector4.prototype.debug = function (label, digits) {
  var val = this.toString(digits);
  if (val !== cache$1[label]) {
    cache$1[label] = val;
    console.log(label + "\n" + val);
  }
  return this;
};

// The JSON format object loader is not always included in the Three.js distribution,
// so we have to first check for it.
var loaders = null;
var PATH_PATTERN = /((?:https?:\/\/)?(?:[^/]+\/)+)(\w+)(\.(?:\w+))$/;
var EXTENSION_PATTERN = /(\.(?:\w+))+$/;

// Sometimes, the properties that export out of Blender and into Three.js don't
// come out correctly, so we need to do a correction.
function fixJSONScene(json) {
  json.traverse(function (obj) {
    if (obj.geometry) {
      obj.geometry.computeBoundingSphere();
      obj.geometry.computeBoundingBox();
    }
  });
  return json;
}

var propertyTests = {
  isButton: function isButton(obj) {
    return obj.material && obj.material.name.match(/^button\d+$/);
  },
  isSolid: function isSolid(obj) {
    return !obj.name.match(/^(water|sky)/);
  },
  isGround: function isGround(obj) {
    return obj.material && obj.material.name && obj.material.name.match(/\bground\b/);
  }
};

function setProperties(object) {
  object.traverse(function (obj) {
    if (obj.isMesh) {
      for (var prop in propertyTests) {
        obj[prop] = obj[prop] || propertyTests[prop](obj);
      }
    }
  });
  return object;
}

var ModelLoader = function () {
  createClass(ModelLoader, null, [{
    key: "loadModel",
    value: function loadModel(src, type, progress) {
      return ModelLoader.loadObject(src, type, progress).then(function (scene) {
        return new ModelLoader(scene);
      });
    }
  }, {
    key: "loadObject",
    value: function loadObject(src, type, progress) {

      var extMatch = src.match(EXTENSION_PATTERN),
          extension = type && "." + type || extMatch[0];
      if (!extension) {
        return Promise.reject("File path `" + src + "` does not have a file extension, and a type was not provided as a parameter, so we can't determine the type.");
      } else {
        extension = extension.toLowerCase();
        if (loaders === null) {
          loaders = {
            ".json": ObjectLoader,
            ".mtl": MTLLoader,
            ".obj": OBJLoader,
            ".typeface.json": FontLoader
          };
        }
        var Loader = new loaders[extension]();
        if (!Loader) {
          return Promise.reject("There is no loader type for the file extension: " + extension);
        } else {
          var name = src.substring(0, extMatch.index),
              elemID = name + "_" + extension.toLowerCase(),
              elem = document.getElementById(elemID),
              promise = Promise.resolve();
          if (extension === ".obj") {
            var newPath = src.replace(EXTENSION_PATTERN, ".mtl");
            promise = promise.then(function () {
              return ModelLoader.loadObject(newPath, "mtl", progress);
            }).then(function (materials) {
              materials.preload();
              Loader.setMaterials(materials);
            });
          } else if (extension === ".mtl") {
            var match = src.match(PATH_PATTERN),
                dir = match[1];
            src = match[2] + match[3];
            Loader.setTexturePath(dir);
            Loader.setPath(dir);
          }

          if (elem) {
            var elemSource = elem.innerHTML.split(/\r?\n/g).map(function (s) {
              return s.trim();
            }).join("\n");
            promise = promise.then(function () {
              return Loader.parse(elemSource);
            });
          } else {
            if (Loader.setCrossOrigin) {
              Loader.setCrossOrigin("anonymous");
            }
            promise = promise.then(function () {
              return new Promise(function (resolve, reject) {
                return Loader.load(src, resolve, progress, reject);
              });
            });
          }

          if (extension === ".json") {
            promise = promise.then(fixJSONScene);
          }

          if (extension !== ".mtl" && extension !== ".typeface.json") {
            promise = promise.then(setProperties);
          }
          promise = promise.catch(console.error.bind(console, "MODEL_ERR", src));
          return promise;
        }
      }
    }
  }, {
    key: "loadObjects",
    value: function loadObjects(map) {

      var output = {},
          promise = Promise.resolve(output);
      for (var key in map) {
        if (map[key]) {
          promise = promise.then(loader(map, key));
        }
      }
      return promise;
    }
  }]);

  function ModelLoader(template) {
    classCallCheck(this, ModelLoader);

    this.template = template;
  }

  createClass(ModelLoader, [{
    key: "clone",
    value: function clone() {
      var _this = this;

      var obj = this.template.clone();

      obj.traverse(function (child) {
        if (child.isSkinnedMesh) {
          obj.animation = new AnimationClip(child, child.geometry.animation);
          if (!_this.template.originalAnimationClipData && obj.animation.data) {
            _this.template.originalAnimationClipData = obj.animation.data;
          }
          if (!obj.animation.data) {
            obj.animation.data = _this.template.originalAnimationClipData;
          }
        }
      });

      setProperties(obj);
      return obj;
    }
  }]);
  return ModelLoader;
}();

function loader(map, key) {
  return function (obj) {
    return ModelLoader.loadObject(map[key]).then(function (model) {
      obj[key] = model;
      return obj;
    });
  };
}

function initState() {
  this.inputState = {
    buttons: [],
    axes: [],
    ctrl: false,
    alt: false,
    shift: false,
    meta: false
  };
  this.lastInputState = {
    buttons: [],
    axes: [],
    ctrl: false,
    alt: false,
    shift: false,
    meta: false
  };
}

function filterMetaKey(k) {
  for (var i = 0; i < Keys.MODIFIER_KEYS.length; ++i) {
    var m = Keys.MODIFIER_KEYS[i];
    if (Math.abs(k) === Keys[m.toLocaleUpperCase()]) {
      return Math.sign(k) * (i + 1);
    }
  }
}

function filterValue(elem) {
  var t = typeof elem === "undefined" ? "undefined" : _typeof(elem);
  var index = 0,
      toggle = false,
      sign = 1;

  if (t === "number") {
    index = Math.abs(elem) - 1;
    toggle = elem < 0;
    sign = elem < 0 ? -1 : 1;
  } else if (t === "string") {
    if (elem[0] === "-") {
      sign = -1;
      elem = elem.substring(1);
    }
    index = this.axisNames.indexOf(elem);
  } else {
    throw new Error("Cannot clone command spec. Element was type: " + t, elem);
  }

  return {
    index: index,
    toggle: toggle,
    sign: sign
  };
}

function swap(a, b) {
  for (var i = 0; i < this.inputState.buttons.length; ++i) {
    this[a].buttons[i] = this[b].buttons[i];
  }
  for (var _i = 0; _i < this.inputState.axes.length; ++_i) {
    this[a].axes[_i] = this[b].axes[_i];
  }
  for (var _i2 = 0; _i2 < Keys.MODIFIER_KEYS.length; ++_i2) {
    var m = Keys.MODIFIER_KEYS[_i2];
    this[a][m] = this[b][m];
  }
}

function resetInputState() {
  swap.call(this, "inputState", "lastInputState");
}

function recordLastState() {
  swap.call(this, "lastInputState", "inputState");
}

var CommandState = function CommandState() {
  classCallCheck(this, CommandState);

  this.value = null;
  this.pressed = false;
  this.wasPressed = false;
  this.fireAgain = false;
  this.lt = 0;
  this.ct = 0;
  this.repeatCount = 0;
};

var InputProcessor = function (_AbstractEventEmitter) {
  inherits(InputProcessor, _AbstractEventEmitter);

  function InputProcessor(name, commands, axisNames) {
    classCallCheck(this, InputProcessor);

    var _this = possibleConstructorReturn(this, (InputProcessor.__proto__ || Object.getPrototypeOf(InputProcessor)).call(this));

    _this.name = name;
    _this.commands = {};
    _this.commandNames = [];
    _this.enabled = true;
    _this.paused = false;
    _this.ready = true;
    _this.inPhysicalUse = false;
    initState.call(_this);

    var readMetaKeys = function readMetaKeys(event) {
      for (var i = 0; i < Keys.MODIFIER_KEYS.length; ++i) {
        var m = Keys.MODIFIER_KEYS[i];
        _this.inputState[m] = event[m + "Key"];
      }
    };

    window.addEventListener("keydown", readMetaKeys, false);
    window.addEventListener("keyup", readMetaKeys, false);
    window.addEventListener("focus", readMetaKeys, false);

    _this.axisNames = axisNames || [];

    for (var i = 0; i < _this.axisNames.length; ++i) {
      _this.inputState.axes[i] = 0;
    }

    for (var cmdName in commands) {
      _this.addCommand(cmdName, commands[cmdName]);
    }

    for (var _i3 = 0; _i3 < Keys.MODIFIER_KEYS.length; ++_i3) {
      _this.inputState[Keys.MODIFIER_KEYS[_i3]] = false;
    }
    return _this;
  }

  createClass(InputProcessor, [{
    key: "addCommand",
    value: function addCommand(name, cmd) {
      cmd.name = name;
      cmd = this.cloneCommand(cmd);
      if (typeof cmd.repetitions === "undefined") {
        cmd.repetitions = 1;
      }
      cmd.state = new CommandState();
      this.commands[name] = cmd;
      this.commandNames.push(name);
    }
  }, {
    key: "cloneCommand",
    value: function cloneCommand(cmd) {
      return {
        name: cmd.name,
        disabled: !!cmd.disabled,
        dt: cmd.dt || 0,
        deadzone: cmd.deadzone || 0,
        threshold: cmd.threshold || 0,
        repetitions: cmd.repetitions,
        scale: cmd.scale,
        offset: cmd.offset,
        min: cmd.min,
        max: cmd.max,
        integrate: !!cmd.integrate,
        delta: !!cmd.delta,
        axes: this.maybeClone(cmd.axes),
        commands: cmd.commands && cmd.commands.slice() || [],
        buttons: this.maybeClone(cmd.buttons),
        metaKeys: this.maybeClone(cmd.metaKeys && cmd.metaKeys.map(filterMetaKey)),
        commandDown: cmd.commandDown,
        commandUp: cmd.commandUp
      };
    }
  }, {
    key: "maybeClone",
    value: function maybeClone(arr) {
      var output = [];
      if (arr) {
        for (var i = 0; i < arr.length; ++i) {
          output[i] = filterValue.call(this, arr[i]);
        }
      }
      return output;
    }
  }, {
    key: "update",
    value: function update(dt) {
      if (this.enabled && this.ready && this.inPhysicalUse && !this.paused && dt > 0) {

        this.inputState.buttons[Keys.ANY] = false;
        for (var n in this.inputState.buttons) {
          if (this.inputState.buttons[n]) {
            this.inputState.buttons[Keys.ANY] = true;
            break;
          }
        }

        var stateMod = recordLastState;
        for (var name in this.commands) {
          var cmd = this.commands[name];
          cmd.state.wasPressed = cmd.state.pressed;
          cmd.state.pressed = false;
          if (!cmd.disabled) {
            var pressed = true,
                value = 0;
            if (cmd.metaKeys) {
              for (var _n = 0; _n < cmd.metaKeys.length && pressed; ++_n) {
                var m = cmd.metaKeys[_n];
                pressed = pressed && (this.inputState[Keys.MODIFIER_KEYS[m.index]] && !m.toggle || !this.inputState[Keys.MODIFIER_KEYS[m.index]] && m.toggle);
              }
            }

            if (pressed) {
              if (cmd.buttons.length > 0) {
                for (var _n2 = 0; _n2 < cmd.buttons.length; ++_n2) {
                  var btn = cmd.buttons[_n2],
                      code = btn.index + 1,
                      p = !!this.inputState.buttons[code];

                  var temp = p ? btn.sign : 0;
                  pressed = pressed && (p && !btn.toggle || !p && btn.toggle);
                  if (Math.abs(temp) > Math.abs(value)) {
                    value = temp;
                  }
                }
              }

              if (cmd.buttons.length === 0 || value !== 0) {
                if (cmd.axes.length > 0) {
                  value = 0;
                  for (var _n3 = 0; _n3 < cmd.axes.length; ++_n3) {
                    var a = cmd.axes[_n3];
                    var _temp = a.sign * this.inputState.axes[a.index];
                    if (Math.abs(_temp) > Math.abs(value)) {
                      value = _temp;
                    }
                  }
                } else if (cmd.commands.length > 0) {
                  value = 0;
                  for (var _n4 = 0; _n4 < cmd.commands.length; ++_n4) {
                    var _temp2 = this.getValue(cmd.commands[_n4]);
                    if (Math.abs(_temp2) > Math.abs(value)) {
                      value = _temp2;
                    }
                  }
                }

                if (cmd.scale !== undefined) {
                  value *= cmd.scale;
                }

                if (cmd.offset !== undefined) {
                  value += cmd.offset;
                }

                if (cmd.deadzone && Math.abs(value) < cmd.deadzone) {
                  value = 0;
                }

                if (cmd.integrate) {
                  value = this.getValue(cmd.name) + value * dt;
                } else if (cmd.delta) {
                  var ov = value;
                  if (cmd.state.lv !== undefined) {
                    value = value - cmd.state.lv;
                  }
                  cmd.state.lv = ov;
                }

                if (cmd.min !== undefined && value < cmd.min) {
                  value = cmd.min;
                  stateMod = resetInputState;
                }

                if (cmd.max !== undefined && value > cmd.max) {
                  value = cmd.max;
                  stateMod = resetInputState;
                }

                if (cmd.threshold) {
                  pressed = pressed && value > cmd.threshold;
                }
              }
            }

            cmd.state.pressed = pressed;
            cmd.state.value = value;
            cmd.state.lt += dt;

            cmd.state.fireAgain = cmd.state.pressed && cmd.state.lt >= cmd.dt && (cmd.repetitions === -1 || cmd.state.repeatCount < cmd.repetitions);

            if (cmd.state.fireAgain) {
              cmd.state.lt = 0;
              ++cmd.state.repeatCount;
            } else if (!cmd.state.pressed) {
              cmd.state.repeatCount = 0;
            }
          }
        }
        stateMod.call(this);

        this.fireCommands();
      }
    }
  }, {
    key: "zero",
    value: function zero() {
      initState.call(this);
      for (var key in this.commands) {
        this.commands[key].state = new CommandState();
      }
    }
  }, {
    key: "fireCommands",
    value: function fireCommands() {
      if (this.ready && !this.paused) {
        for (var name in this.commands) {
          var cmd = this.commands[name];
          if (cmd.state.fireAgain && cmd.commandDown) {
            cmd.commandDown(this.name);
          }

          if (!cmd.state.pressed && cmd.state.wasPressed && cmd.commandUp) {
            cmd.commandUp(this.name);
          }
        }
      }
    }
  }, {
    key: "setProperty",
    value: function setProperty(key, name, value) {
      if (this.commands[name]) {
        this.commands[name][key] = value;
      }
    }
  }, {
    key: "setDeadzone",
    value: function setDeadzone(name, value) {
      this.setProperty("deadzone", name, value);
    }
  }, {
    key: "setScale",
    value: function setScale(name, value) {
      this.setProperty("scale", name, value);
    }
  }, {
    key: "setDT",
    value: function setDT(name, value) {
      this.setProperty("dt", name, value);
    }
  }, {
    key: "setMin",
    value: function setMin(name, value) {
      this.setProperty("min", name, value);
    }
  }, {
    key: "setMax",
    value: function setMax(name, value) {
      this.setProperty("max", name, value);
    }
  }, {
    key: "addMetaKey",
    value: function addMetaKey(name, value) {
      this.addToArray("metaKeys", name, filterMetaKey(value));
    }
  }, {
    key: "addAxis",
    value: function addAxis(name, value) {
      this.addToArray("axes", name, value);
    }
  }, {
    key: "addButton",
    value: function addButton(name, value) {
      this.addToArray("buttons", name, value);
    }
  }, {
    key: "removeMetaKey",
    value: function removeMetaKey(name, value) {
      this.removeFromArray("metaKeys", name, value);
    }
  }, {
    key: "removeAxis",
    value: function removeAxis(name, value) {
      this.removeFromArray("axes", name, value);
    }
  }, {
    key: "removeButton",
    value: function removeButton(name, value) {
      this.removeFromArray("buttons", name, value);
    }
  }, {
    key: "invertAxis",
    value: function invertAxis(name, value) {
      this.invertInArray("axes", name, value);
    }
  }, {
    key: "invertButton",
    value: function invertButton(name, value) {
      this.invertInArray("buttons", name, value);
    }
  }, {
    key: "invertMetaKey",
    value: function invertMetaKey(name, value) {
      this.invertInArray("metaKeys", name, value);
    }
  }, {
    key: "addToArray",
    value: function addToArray(key, name, value) {
      if (this.commands[name] && this.commands[name][key]) {
        this.commands[name][key].push(filterValue(value));
      }
    }
  }, {
    key: "removeFromArray",
    value: function removeFromArray(key, name, value) {
      if (this.commands[name] && this.commands[name][key]) {
        --value;
        var arr = this.commands[name][key];
        for (var i = 0; i < arr.length; ++i) {
          var elem = arr[i];
          if (elem.index === value) {
            return arr.splice(i, 1);
          }
        }
      }
    }
  }, {
    key: "invertInArray",
    value: function invertInArray(key, name, value) {
      if (this.commands[name] && this.commands[name][key]) {
        var arr = this.commands[name][key],
            n = arr.indexOf(value);
        for (var i = 0; i < arr.length; ++i) {
          var elem = arr[i];
          if (elem.index === value) {
            elem.sign *= -1;
            return;
          }
        }
      }
    }
  }, {
    key: "pause",
    value: function pause(v) {
      this.paused = v;
    }
  }, {
    key: "isPaused",
    value: function isPaused() {
      return this.paused;
    }
  }, {
    key: "enable",
    value: function enable(k, v) {
      if (v === undefined || v === null) {
        v = k;
        k = null;
      }

      if (k) {
        this.setProperty("disabled", k, !v);
      } else {
        this.enabled = v;
      }
    }
  }, {
    key: "isEnabled",
    value: function isEnabled(name) {
      return name && this.commands[name] && !this.commands[name].disabled;
    }
  }, {
    key: "getAxis",
    value: function getAxis(name) {
      var i = this.axisNames.indexOf(name);
      if (i > -1) {
        var value = this.inputState.axes[i] || 0;
        return value;
      }
      return null;
    }
  }, {
    key: "setAxis",
    value: function setAxis(name, value) {
      var i = this.axisNames.indexOf(name);
      if (i > -1 && (this.inPhysicalUse || value !== 0)) {
        this.inPhysicalUse = true;
        this.inputState.axes[i] = value;
      }
    }
  }, {
    key: "setButton",
    value: function setButton(index, pressed) {
      if (this.inPhysicalUse || pressed) {
        this.inPhysicalUse = true;
        this.inputState.buttons[index] = pressed;
      }
    }
  }, {
    key: "isDown",
    value: function isDown(name) {
      return this.enabled && this.isEnabled(name) && this.commands[name].state.pressed;
    }
  }, {
    key: "isUp",
    value: function isUp(name) {
      return this.enabled && this.isEnabled(name) && this.commands[name].state.pressed;
    }
  }, {
    key: "getValue",
    value: function getValue(name) {
      return this.enabled && this.isEnabled(name) && (this.commands[name].state.value || this.getAxis(name)) || 0;
    }
  }, {
    key: "setValue",
    value: function setValue(name, value) {
      var j = this.axisNames.indexOf(name);
      if (!this.commands[name] && j > -1) {
        this.setAxis(name, value);
      } else if (this.commands[name] && !this.commands[name].disabled) {
        this.commands[name].state.value = value;
      }
    }
  }]);
  return InputProcessor;
}(AbstractEventEmitter);

function setCursorCommand(obj, mod, key, func, cur) {
  var name = mod + "_" + key;
  obj[name] = function (prim, tokenRows) {
    prim["cursor" + func](tokenRows, prim[cur + "Cursor"]);
  };
}

var OperatingSystem = function () {
  function OperatingSystem(osName, pre1, pre2, redo, pre3, home, end, pre5) {
    classCallCheck(this, OperatingSystem);

    this.name = osName;

    var pre4 = pre3;
    pre3 = pre3.length > 0 ? pre3 : "NORMAL";

    this[pre1 + "_a"] = "SELECT_ALL";
    this[pre1 + "_c"] = "COPY";
    this[pre1 + "_x"] = "CUT";
    this[pre1 + "_v"] = "PASTE";
    this[redo] = "REDO";
    this[pre1 + "_z"] = "UNDO";
    this[pre1 + "_DOWNARROW"] = "WINDOW_SCROLL_DOWN";
    this[pre1 + "_UPARROW"] = "WINDOW_SCROLL_UP";
    this[pre2 + "_LEFTARROW"] = "NORMAL_SKIPLEFT";
    this[pre2 + "SHIFT_LEFTARROW"] = "SHIFT_SKIPLEFT";
    this[pre2 + "_RIGHTARROW"] = "NORMAL_SKIPRIGHT";
    this[pre2 + "SHIFT_RIGHTARROW"] = "SHIFT_SKIPRIGHT";
    this[pre3 + "_HOME"] = "NORMAL_HOME";
    this[pre4 + "SHIFT_HOME"] = "SHIFT_HOME";
    this[pre3 + "_END"] = "NORMAL_END";
    this[pre4 + "SHIFT_END"] = "SHIFT_END";
    this[pre5 + "_HOME"] = "CTRL_HOME";
    this[pre5 + "SHIFT_HOME"] = "CTRLSHIFT_HOME";
    this[pre5 + "_END"] = "CTRL_END";
    this[pre5 + "SHIFT_END"] = "CTRLSHIFT_END";

    this._deadKeyState = "";
  }

  createClass(OperatingSystem, [{
    key: "makeCommandName",
    value: function makeCommandName(evt, codePage) {
      var key = evt.keyCode;
      if (key !== Keys.CTRL && key !== Keys.ALT && key !== Keys.META_L && key !== Keys.META_R && key !== Keys.SHIFT) {

        var oldDeadKeyState = this._deadKeyState,
            commandName = this._deadKeyState;

        if (evt.ctrlKey) {
          commandName += "CTRL";
        }
        if (evt.altKey) {
          commandName += "ALT";
        }
        if (evt.metaKey) {
          commandName += "META";
        }
        if (evt.shiftKey) {
          commandName += "SHIFT";
        }
        if (commandName === this._deadKeyState) {
          commandName += "NORMAL";
        }

        commandName += "_" + codePage.keyNames[key];

        return this[commandName] || commandName;
      }
    }
  }]);
  return OperatingSystem;
}();

var Windows = new OperatingSystem("Windows", "CTRL", "CTRL", "CTRL_y", "", "HOME", "END", "CTRL", "HOME", "END");

var macOS = new OperatingSystem("macOS", "META", "ALT", "METASHIFT_z", "META", "LEFTARROW", "RIGHTARROW", "META", "UPARROW", "DOWNARROW");

var CodePage = function CodePage(codePageName, lang, options) {
  classCallCheck(this, CodePage);

  this.name = codePageName;
  this.language = lang;

  var commands = {
    NORMAL: {
      "65": "a",
      "66": "b",
      "67": "c",
      "68": "d",
      "69": "e",
      "70": "f",
      "71": "g",
      "72": "h",
      "73": "i",
      "74": "j",
      "75": "k",
      "76": "l",
      "77": "m",
      "78": "n",
      "79": "o",
      "80": "p",
      "81": "q",
      "82": "r",
      "83": "s",
      "84": "t",
      "85": "u",
      "86": "v",
      "87": "w",
      "88": "x",
      "89": "y",
      "90": "z"
    },
    SHIFT: {
      "65": "A",
      "66": "B",
      "67": "C",
      "68": "D",
      "69": "E",
      "70": "F",
      "71": "G",
      "72": "H",
      "73": "I",
      "74": "J",
      "75": "K",
      "76": "L",
      "77": "M",
      "78": "N",
      "79": "O",
      "80": "P",
      "81": "Q",
      "82": "R",
      "83": "S",
      "84": "T",
      "85": "U",
      "86": "V",
      "87": "W",
      "88": "X",
      "89": "Y",
      "90": "Z"
    }
  };

  for (var key in options) {
    commands[key] = Object.assign({}, commands[key], options[key]);
  }

  var char, code, cmdName;
  for (var i = 0; i <= 9; ++i) {
    code = Keys["NUMPAD" + i];
    commands.NORMAL[code] = i.toString();
  }

  commands.NORMAL[Keys.MULTIPLY] = "*";
  commands.NORMAL[Keys.ADD] = "+";
  commands.NORMAL[Keys.SUBTRACT] = "-";
  commands.NORMAL[Keys.DECIMALPOINT] = ".";
  commands.NORMAL[Keys.DIVIDE] = "/";

  this.keyNames = {};
  this.commandNames = [];
  for (char in Keys) {
    code = Keys[char];
    if (!isNaN(code)) {
      this.keyNames[code] = char;
    }
  }

  function overwriteText(txt, prim, lines) {
    prim.selectedText = txt;
  }

  for (var type in commands) {
    var codes = commands[type];
    if ((typeof codes === "undefined" ? "undefined" : _typeof(codes)) === "object") {
      for (code in codes) {
        if (code.indexOf("_") > -1) {
          var parts = code.split(' '),
              browser = parts[0];
          code = parts[1];
          char = commands.NORMAL[code];
          cmdName = browser + "_" + type + " " + char;
        } else {
          char = commands.NORMAL[code];
          cmdName = type + "_" + char;
        }
        this.commandNames.push(cmdName);
        this.keyNames[code] = char;
        var func = codes[code];
        if (typeof func !== "function") {
          func = overwriteText.bind(null, func);
        }
        this[cmdName] = func;
      }
    }
  }
};



CodePage.DEAD = function (key) {
  return function (prim) {
    prim.setDeadKeyState("DEAD" + key);
  };
};

var DE_QWERTZ = new CodePage("Deutsch: QWERTZ", "de", {
  deadKeys: [220, 221, 160, 192],
  NORMAL: {
    "32": " ",
    "48": "0",
    "49": "1",
    "50": "2",
    "51": "3",
    "52": "4",
    "53": "5",
    "54": "6",
    "55": "7",
    "56": "8",
    "57": "9",
    "60": "<",
    "63": "ß",
    "160": CodePage.DEAD(3),
    "163": "#",
    "171": "+",
    "173": "-",
    "186": "ü",
    "187": "+",
    "188": ",",
    "189": "-",
    "190": ".",
    "191": "#",
    "192": CodePage.DEAD(4),
    "219": "ß",
    "220": CodePage.DEAD(1),
    "221": CodePage.DEAD(2),
    "222": "ä",
    "226": "<"
  },
  DEAD1NORMAL: {
    "65": "â",
    "69": "ê",
    "73": "î",
    "79": "ô",
    "85": "û",
    "190": "."
  },
  DEAD2NORMAL: {
    "65": "á",
    "69": "é",
    "73": "í",
    "79": "ó",
    "83": "s",
    "85": "ú",
    "89": "ý"
  },
  SHIFT: {
    "32": " ",
    "48": "=",
    "49": "!",
    "50": "\"",
    "51": "§",
    "52": "$",
    "53": "%",
    "54": "&",
    "55": "/",
    "56": "(",
    "57": ")",
    "60": ">",
    "63": "?",
    "163": "'",
    "171": "*",
    "173": "_",
    "186": "Ü",
    "187": "*",
    "188": ";",
    "189": "_",
    "190": ":",
    "191": "'",
    "192": "Ö",
    "219": "?",
    "222": "Ä",
    "226": ">"
  },
  CTRLALT: {
    "48": "}",
    "50": "²",
    "51": "³",
    "55": "{",
    "56": "[",
    "57": "]",
    "60": "|",
    "63": "\\",
    "69": "€",
    "77": "µ",
    "81": "@",
    "171": "~",
    "187": "~",
    "219": "\\",
    "226": "|"
  },
  CTRLALTSHIFT: {
    "63": "ẞ",
    "219": "ẞ"
  },
  DEAD3NORMAL: {
    "65": "a",
    "69": "e",
    "73": "i",
    "79": "o",
    "85": "u",
    "190": "."
  },
  DEAD4NORMAL: {
    "65": "a",
    "69": "e",
    "73": "i",
    "79": "o",
    "83": "s",
    "85": "u",
    "89": "y"
  }
});

var EN_UKX = new CodePage("English: UK Extended", "en-GB", {
  CTRLALT: {
    "52": "€",
    "65": "á",
    "69": "é",
    "73": "í",
    "79": "ó",
    "85": "ú",
    "163": "\\",
    "192": "¦",
    "222": "\\",
    "223": "¦"
  },
  CTRLALTSHIFT: {
    "65": "Á",
    "69": "É",
    "73": "Í",
    "79": "Ó",
    "85": "Ú",
    "222": "|"
  },
  NORMAL: {
    "32": " ",
    "48": "0",
    "49": "1",
    "50": "2",
    "51": "3",
    "52": "4",
    "53": "5",
    "54": "6",
    "55": "7",
    "56": "8",
    "57": "9",
    "59": ";",
    "61": "=",
    "163": "#",
    "173": "-",
    "186": ";",
    "187": "=",
    "188": ",",
    "189": "-",
    "190": ".",
    "191": "/",
    "192": "'",
    "219": "[",
    "220": "\\",
    "221": "]",
    "222": "#",
    "223": "`"
  },
  SHIFT: {
    "32": " ",
    "48": ")",
    "49": "!",
    "50": "\"",
    "51": "£",
    "52": "$",
    "53": "%",
    "54": "^",
    "55": "&",
    "56": "*",
    "57": "(",
    "59": ":",
    "61": "+",
    "163": "~",
    "173": "_",
    "186": ":",
    "187": "+",
    "188": "<",
    "189": "_",
    "190": ">",
    "191": "?",
    "192": "@",
    "219": "{",
    "220": "|",
    "221": "}",
    "222": "~",
    "223": "¬"
  }
});

var EN_US = new CodePage("English: USA", "en-US", {
  NORMAL: {
    "32": " ",
    "48": "0",
    "49": "1",
    "50": "2",
    "51": "3",
    "52": "4",
    "53": "5",
    "54": "6",
    "55": "7",
    "56": "8",
    "57": "9",
    "59": ";",
    "61": "=",
    "173": "-",
    "186": ";",
    "187": "=",
    "188": ",",
    "189": "-",
    "190": ".",
    "191": "/",
    "219": "[",
    "220": "\\",
    "221": "]",
    "222": "'"
  },
  SHIFT: {
    "32": " ",
    "48": ")",
    "49": "!",
    "50": "@",
    "51": "#",
    "52": "$",
    "53": "%",
    "54": "^",
    "55": "&",
    "56": "*",
    "57": "(",
    "59": ":",
    "61": "+",
    "173": "_",
    "186": ":",
    "187": "+",
    "188": "<",
    "189": "_",
    "190": ">",
    "191": "?",
    "219": "{",
    "220": "|",
    "221": "}",
    "222": "\""
  }
});

var FR_AZERTY = new CodePage("Français: AZERTY", "fr", {
  deadKeys: [221, 50, 55],
  NORMAL: {
    "32": " ",
    "48": "à",
    "49": "&",
    "50": "é",
    "51": "\"",
    "52": "'",
    "53": "(",
    "54": "-",
    "55": "è",
    "56": "_",
    "57": "ç",
    "186": "$",
    "187": "=",
    "188": ",",
    "190": ";",
    "191": ":",
    "192": "ù",
    "219": ")",
    "220": "*",
    "221": CodePage.DEAD(1),
    "222": "²",
    "223": "!",
    "226": "<"
  },
  SHIFT: {
    "32": " ",
    "48": "0",
    "49": "1",
    "50": "2",
    "51": "3",
    "52": "4",
    "53": "5",
    "54": "6",
    "55": "7",
    "56": "8",
    "57": "9",
    "186": "£",
    "187": "+",
    "188": "?",
    "190": ".",
    "191": "/",
    "192": "%",
    "219": "°",
    "220": "µ",
    "223": "§",
    "226": ">"
  },
  CTRLALT: {
    "48": "@",
    "50": CodePage.DEAD(2),
    "51": "#",
    "52": "{",
    "53": "[",
    "54": "|",
    "55": CodePage.DEAD(3),
    "56": "\\",
    "57": "^",
    "69": "€",
    "186": "¤",
    "187": "}",
    "219": "]"
  },
  DEAD1NORMAL: {
    "65": "â",
    "69": "ê",
    "73": "î",
    "79": "ô",
    "85": "û"
  },
  DEAD2NORMAL: {
    "65": "ã",
    "78": "ñ",
    "79": "õ"
  },
  DEAD3NORMAL: {
    "48": "à",
    "50": "é",
    "55": "è",
    "65": "à",
    "69": "è",
    "73": "ì",
    "79": "ò",
    "85": "ù"
  }
});

var CodePages = {
  CodePage: CodePage,
  DE_QWERTZ: DE_QWERTZ,
  EN_UKX: EN_UKX,
  EN_US: EN_US,
  FR_AZERTY: FR_AZERTY
};

var Keyboard = function (_InputProcessor) {
  inherits(Keyboard, _InputProcessor);

  function Keyboard(input, commands) {
    classCallCheck(this, Keyboard);

    var _this = possibleConstructorReturn(this, (Keyboard.__proto__ || Object.getPrototypeOf(Keyboard)).call(this, "Keyboard", commands));

    _this._operatingSystem = null;
    _this.browser = isChrome ? "CHROMIUM" : isFirefox ? "FIREFOX" : isIE ? "IE" : isOpera ? "OPERA" : isSafari ? "SAFARI" : "UNKNOWN";
    _this._codePage = null;
    return _this;
  }

  createClass(Keyboard, [{
    key: "dispatchEvent",
    value: function dispatchEvent(evt) {
      this.setButton(evt.keyCode, evt.type === "keydown");
    }
  }, {
    key: "doTyping",
    value: function doTyping(elem, evt) {
      if (elem && elem.execCommand && this.operatingSystem && this.browser && this.codePage) {
        var oldDeadKeyState = this.operatingSystem._deadKeyState,
            cmdName = this.operatingSystem.makeCommandName(evt, this.codePage);
        if (elem.execCommand(this.browser, this.codePage, cmdName)) {
          evt.preventDefault();
        }
        if (this.operatingSystem._deadKeyState === oldDeadKeyState) {
          this.operatingSystem._deadKeyState = "";
        }
      }
    }
  }, {
    key: "operatingSystem",
    get: function get() {
      return this._operatingSystem;
    },
    set: function set(os) {
      this._operatingSystem = os || (isMacOS ? macOS : Windows);
    }
  }, {
    key: "codePage",
    get: function get() {
      return this._codePage;
    },
    set: function set(cp) {
      var key, code, char, name;
      this._codePage = cp;
      if (!this._codePage) {
        var lang = navigator.languages && navigator.languages[0] || navigator.language || navigator.userLanguage || navigator.browserLanguage;

        if (!lang || lang === "en") {
          lang = "en-US";
        }

        for (key in CodePages) {
          cp = CodePages[key];
          if (cp.language === lang) {
            this._codePage = cp;
            break;
          }
        }

        if (!this._codePage) {
          this._codePage = CodePages.EN_US;
        }
      }
    }
  }]);
  return Keyboard;
}(InputProcessor);

var Mouse = function (_InputProcessor) {
  inherits(Mouse, _InputProcessor);

  function Mouse(DOMElement, commands) {
    classCallCheck(this, Mouse);

    var _this = possibleConstructorReturn(this, (Mouse.__proto__ || Object.getPrototypeOf(Mouse)).call(this, "Mouse", commands, ["BUTTONS", "X", "Y", "Z", "W"]));

    _this.timer = null;

    var setState = function setState(stateChange, event) {
      var state = event.buttons;
      for (var button = 0; button < Mouse.NUM_BUTTONS; ++button) {
        var isDown = state & 0x1 !== 0;
        if (isDown && stateChange || !isDown && !stateChange) {
          _this.setButton(button, stateChange);
        }
        state >>= 1;
      }
      _this.setAxis("BUTTONS", event.buttons << 10);
      if (event.target === DOMElement) {
        event.preventDefault();
      }
    };

    DOMElement.addEventListener("mousedown", setState.bind(_this, true), false);
    DOMElement.addEventListener("mouseup", setState.bind(_this, false), false);
    DOMElement.addEventListener("contextmenu", function (event) {
      return !(event.ctrlKey && event.shiftKey) && event.preventDefault();
    }, false);
    DOMElement.addEventListener("mousemove", function (event) {
      setState(true, event);

      if (PointerLock.isActive) {
        var mx = event.movementX,
            my = event.movementY;

        if (mx === undefined) {
          mx = event.webkitMovementX || event.mozMovementX || 0;
          my = event.webkitMovementY || event.mozMovementY || 0;
        }
        _this.setAxis("X", _this.getAxis("X") + mx);
        _this.setAxis("Y", _this.getAxis("Y") + my);
      } else {
        _this.setAxis("X", event.layerX);
        _this.setAxis("Y", event.layerY);
      }
    }, false);

    DOMElement.addEventListener("wheel", function (event) {
      if (isChrome) {
        _this.W += event.deltaX;
        _this.Z += event.deltaY;
      } else if (event.shiftKey) {
        _this.W += event.deltaY;
      } else {
        _this.Z += event.deltaY;
      }
      if (event.target === DOMElement) {
        event.preventDefault();
      }
    }, false);
    return _this;
  }

  return Mouse;
}(InputProcessor);



Mouse.NUM_BUTTONS = 3;

var DEFAULT_POSE = {
  position: [0, 0, 0],
  orientation: [0, 0, 0, 1]
};
var EMPTY_SCALE = new Vector3();
var IE_CORRECTION = new Quaternion(1, 0, 0, 0);

var PoseInputProcessor = function (_InputProcessor) {
  inherits(PoseInputProcessor, _InputProcessor);

  function PoseInputProcessor(name, commands, axisNames) {
    classCallCheck(this, PoseInputProcessor);

    var _this = possibleConstructorReturn(this, (PoseInputProcessor.__proto__ || Object.getPrototypeOf(PoseInputProcessor)).call(this, name, commands, axisNames));

    _this.currentDevice = null;
    _this.lastPose = null;
    _this.currentPose = null;
    _this.posePosition = new Vector3();
    _this.poseQuaternion = new Quaternion();
    _this.position = new Vector3();
    _this.quaternion = new Quaternion();
    _this.matrix = new Matrix4();
    return _this;
  }

  createClass(PoseInputProcessor, [{
    key: "update",
    value: function update(dt) {
      get$1(PoseInputProcessor.prototype.__proto__ || Object.getPrototypeOf(PoseInputProcessor.prototype), "update", this).call(this, dt);

      if (this.currentDevice) {
        var pose = this.currentPose || this.lastPose || DEFAULT_POSE;
        this.lastPose = pose;
        this.inPhysicalUse = this.hasOrientation || this.inPhysicalUse;
        var orient = this.currentPose && this.currentPose.orientation,
            pos = this.currentPose && this.currentPose.position;
        if (orient) {
          this.poseQuaternion.fromArray(orient);
          if (isMobile && isIE) {
            this.poseQuaternion.multiply(IE_CORRECTION);
          }
        } else {
          this.poseQuaternion.set(0, 0, 0, 1);
        }
        if (pos) {
          this.posePosition.fromArray(pos);
        } else {
          this.posePosition.set(0, 0, 0);
        }
      }
    }
  }, {
    key: "updateStage",
    value: function updateStage(stageMatrix) {
      this.matrix.makeRotationFromQuaternion(this.poseQuaternion);
      this.matrix.setPosition(this.posePosition);
      this.matrix.multiplyMatrices(stageMatrix, this.matrix);
      this.matrix.decompose(this.position, this.quaternion, EMPTY_SCALE);
    }
  }, {
    key: "hasPose",
    get: function get() {
      return !!this.currentPose;
    }
  }]);
  return PoseInputProcessor;
}(InputProcessor);

navigator.getGamepads = navigator.getGamepads || navigator.webkitGetGamepads;

function playPattern(devices, pattern, pause) {
  if (pattern.length > 0) {
    var length = pattern.shift();
    if (!pause) {
      for (var i = 0; i < devices.length; ++i) {
        devices[0].vibrate(1, length);
      }
    }
    setTimeout(playPattern, length, devices, pattern, !pause);
  }
}

var Gamepad = function (_PoseInputProcessor) {
  inherits(Gamepad, _PoseInputProcessor);
  createClass(Gamepad, null, [{
    key: "ID",
    value: function ID(pad) {
      var id = pad.id;
      if (id === "OpenVR Gamepad") {
        id = "Vive";
      } else if (id.indexOf("Rift") === 0) {
        id = "Rift";
      } else if (id.indexOf("Unknown") === 0) {
        id = "Unknown";
      } else {
        id = "Gamepad";
      }
      id = (id + "_" + (pad.index || 0)).replace(/\s+/g, "_");
      return id;
    }
  }, {
    key: "isMotionController",
    value: function isMotionController(pad) {
      if (pad) {
        var obj = pad.capabilities || pad.pose;
        return obj && obj.hasOrientation;
      }
      return false;
    }
  }]);

  function Gamepad(mgr, pad, axisOffset, commands) {
    classCallCheck(this, Gamepad);

    var padID = Gamepad.ID(pad);

    var _this = possibleConstructorReturn(this, (Gamepad.__proto__ || Object.getPrototypeOf(Gamepad)).call(this, padID, commands, ["LSX", "LSY", "RSX", "RSY", "IDK1", "IDK2", "Z", "BUTTONS"]));

    mgr.registerPad(padID, _this);

    _this.currentDevice = pad;
    _this.axisOffset = axisOffset;
    return _this;
  }

  createClass(Gamepad, [{
    key: "checkDevice",
    value: function checkDevice(pad) {
      var i,
          j,
          buttonMap = 0;
      this.currentDevice = pad;
      this.currentPose = this.hasOrientation && this.currentDevice.pose;
      for (i = 0, j = pad.buttons.length; i < pad.buttons.length; ++i, ++j) {
        var btn = pad.buttons[i];
        this.setButton(i, btn.pressed);
        if (btn.pressed) {
          buttonMap |= 0x1 << i;
        }

        this.setButton(j, btn.touched);
        if (btn.touched) {
          buttonMap |= 0x1 << j;
        }
      }
      this.setAxis("BUTTONS", buttonMap);
      for (i = 0; i < pad.axes.length; ++i) {
        var axisName = this.axisNames[this.axisOffset * pad.axes.length + i],
            axisValue = pad.axes[i];
        this.setAxis(axisName, axisValue);
      }
    }
  }, {
    key: "vibratePattern",
    value: function vibratePattern(pattern) {
      if (this.currentDevice) {
        if (this.currentDevice.vibrate) {
          this.currentDevice.vibrate(pattern);
        } else if (this.currentDevice.haptics && this.currentDevice.haptics.length > 0) {
          playPattern(this.currentDevice.haptics, pattern);
        }
      }
    }
  }, {
    key: "hasOrientation",
    get: function get() {
      return Gamepad.isMotionController(this.currentDevice);
    }
  }, {
    key: "haptics",
    get: function get() {
      return this.currentDevice && this.currentDevice.haptics;
    }
  }]);
  return Gamepad;
}(PoseInputProcessor);

Gamepad.XBOX_360_BUTTONS = {
  A: 1,
  B: 2,
  X: 3,
  Y: 4,
  LEFT_BUMPER: 5,
  RIGHT_BUMPER: 6,
  LEFT_TRIGGER: 7,
  RIGHT_TRIGGER: 8,
  BACK: 9,
  START: 10,
  LEFT_STICK: 11,
  RIGHT_STICK: 12,
  UP_DPAD: 13,
  DOWN_DPAD: 14,
  LEFT_DPAD: 15,
  RIGHT_DPAD: 16
};

Gamepad.XBOX_ONE_BUTTONS = {
  A: 1,
  B: 2,
  X: 3,
  Y: 4,
  LEFT_BUMPER: 5,
  RIGHT_BUMPER: 6,
  LEFT_TRIGGER: 7,
  RIGHT_TRIGGER: 8,
  BACK: 9,
  START: 10,
  LEFT_STICK: 11,
  RIGHT_STICK: 12,
  UP_DPAD: 13,
  DOWN_DPAD: 14,
  LEFT_DPAD: 15,
  RIGHT_DPAD: 16
};

Gamepad.VIVE_BUTTONS = {
  TOUCHPAD_PRESSED: 0,
  TRIGGER_PRESSED: 1,
  GRIP_PRESSED: 2,
  MENU_PRESSED: 3,

  TOUCHPAD_TOUCHED: 4,
  //TRIGGER_TOUCHED: 5, // doesn't ever actually trigger in the current version of Chromium - STM 6/22/2016
  GRIP_TOUCHED: 6,
  MENU_TOUCHED: 7
};

var GamepadManager = function (_AbstractEventEmitter) {
  inherits(GamepadManager, _AbstractEventEmitter);
  createClass(GamepadManager, null, [{
    key: "isAvailable",
    get: function get() {
      return !!navigator.getGamepads;
    }
  }]);

  function GamepadManager() {
    classCallCheck(this, GamepadManager);

    var _this = possibleConstructorReturn(this, (GamepadManager.__proto__ || Object.getPrototypeOf(GamepadManager)).call(this));

    _this.currentDevices = [];
    _this.currentDeviceIDs = [];
    _this.currentManagers = {};
    return _this;
  }

  createClass(GamepadManager, [{
    key: "poll",
    value: function poll() {
      if (GamepadManager.isAvailable) {
        var maybePads = navigator.getGamepads(),
            pads = [],
            padIDs = [],
            newPads = [],
            oldPads = [],
            i,
            padID;

        if (maybePads) {
          for (i = 0; i < maybePads.length; ++i) {
            var maybePad = maybePads[i];
            if (maybePad) {
              padID = Gamepad.ID(maybePad);
              var padIdx = this.currentDeviceIDs.indexOf(padID);
              pads.push(maybePad);
              padIDs.push(padID);
              if (padIdx === -1) {
                newPads.push(maybePad);
                this.currentDeviceIDs.push(padID);
                this.currentDevices.push(maybePad);
                delete this.currentManagers[padID];
              } else {
                this.currentDevices[padIdx] = maybePad;
              }
            }
          }
        }

        for (i = this.currentDeviceIDs.length - 1; i >= 0; --i) {
          padID = this.currentDeviceIDs[i];
          var mgr = this.currentManagers[padID],
              pad = this.currentDevices[i];
          if (padIDs.indexOf(padID) === -1) {
            oldPads.push(padID);
            this.currentDevices.splice(i, 1);
            this.currentDeviceIDs.splice(i, 1);
          } else if (mgr) {
            mgr.checkDevice(pad);
          }
        }

        newPads.forEach(this.emit.bind(this, "gamepadconnected"));
        oldPads.forEach(this.emit.bind(this, "gamepaddisconnected"));
      }
    }
  }, {
    key: "registerPad",
    value: function registerPad(id, mgr) {
      this.currentManagers[id] = mgr;
    }
  }, {
    key: "pads",
    get: function get() {
      return this.currentDevices;
    }
  }]);
  return GamepadManager;
}(AbstractEventEmitter);

var TEMP$1 = new Vector2();

var Touch = function (_InputProcessor) {
  inherits(Touch, _InputProcessor);

  function Touch(DOMElement, commands) {
    classCallCheck(this, Touch);

    var axes = ["FINGERS"];
    for (var i = 0; i < 10; ++i) {
      axes.push("X" + i);
      axes.push("Y" + i);
      axes.push("LX" + i);
      axes.push("LY" + i);
      axes.push("DX" + i);
      axes.push("DY" + i);
    }

    var _this = possibleConstructorReturn(this, (Touch.__proto__ || Object.getPrototypeOf(Touch)).call(this, "Touch", commands, axes));

    var setState = function setState(stateChange, setAxis, event) {
      // We have to find the minimum identifier value because iOS uses a very
      // large number that changes after every gesture. Every other platform
      // just numbers them 0 through 9.
      var touches = event.changedTouches,
          minIdentifier = Number.MAX_VALUE;
      for (var _i = 0; _i < touches.length; ++_i) {
        minIdentifier = Math.min(minIdentifier, touches[_i].identifier);
      }

      for (var _i2 = 0; _i2 < touches.length; ++_i2) {
        var t = touches[_i2],
            id = t.identifier - minIdentifier,
            x = t.pageX,
            y = t.pageY;
        _this.setAxis("X" + id, x);
        _this.setAxis("Y" + id, y);
        _this.setButton("FINGER" + t.identifier, stateChange);

        if (setAxis) {
          var lx = _this.getAxis("LX" + id),
              ly = _this.getAxis("LY" + id);
          _this.setAxis("DX" + id, x - lx);
          _this.setAxis("DY" + id, y - ly);
        }

        _this.setAxis("LX" + id, x);
        _this.setAxis("LY" + id, y);
      }

      touches = event.touches;
      var fingerState = 0;
      for (var _i3 = 0; _i3 < touches.length; ++_i3) {
        var _t = touches[_i3];
        fingerState |= 1 << _t.identifier;
      }
      _this.setAxis("FINGERS", fingerState);

      if (event.target === DOMElement) {
        event.preventDefault();
      }
    };

    DOMElement.addEventListener("touchstart", setState.bind(_this, true, false), false);
    DOMElement.addEventListener("touchend", setState.bind(_this, false, true), false);
    DOMElement.addEventListener("touchmove", setState.bind(_this, true, true), false);
    return _this;
  }

  createClass(Touch, [{
    key: "update",
    value: function update(dt) {
      get$1(Touch.prototype.__proto__ || Object.getPrototypeOf(Touch.prototype), "update", this).call(this, dt);
      for (var id = 0; id < 10; ++id) {
        var x = this.getAxis("X" + id),
            y = this.getAxis("Y" + id),
            lx = this.getAxis("LX" + id),
            ly = this.getAxis("LY" + id);
        this.setAxis("DX" + id, x - lx);
        this.setAxis("DY" + id, y - ly);
        this.setAxis("LX" + id, x);
        this.setAxis("LY" + id, y);
      }
    }
  }]);
  return Touch;
}(InputProcessor);

var Speech$2 = function (_InputProcessor) {
  inherits(Speech, _InputProcessor);

  function Speech(commands) {
    classCallCheck(this, Speech);

    var _this = possibleConstructorReturn(this, (Speech.__proto__ || Object.getPrototypeOf(Speech)).call(this, "Speech", commands));

    var running = false,
        recognition = null,
        errorMessage = null;

    function warn() {
      var msg = "Failed to initialize speech engine. Reason: " + errorMessage.message;
      console.error(msg);
      return false;
    }

    function start() {
      if (!available) {
        return warn();
      } else if (!running) {
        running = true;
        recognition.start();
        return true;
      }
      return false;
    }

    function stop() {
      if (!available) {
        return warn();
      }
      if (running) {
        recognition.stop();
        return true;
      }
      return false;
    }

    _this.check = function () {
      if (this.enabled && !running) {
        start();
      } else if (!this.enabled && running) {
        stop();
      }
    };

    _this.getErrorMessage = function () {
      return errorMessage;
    };

    try {
      if (window.SpeechRecognition) {
        // just in case this ever gets standardized
        recognition = new SpeechRecognition();
      } else {
        // purposefully don't check the existance so it errors out and setup fails.
        recognition = new webkitSpeechRecognition();
      }
      recognition.continuous = true;
      recognition.interimResults = true;
      recognition.lang = "en-US";
      var restart = false;
      recognition.addEventListener("start", function () {
        console.log("speech started");
        command = "";
      }.bind(_this), true);

      recognition.addEventListener("error", function (evt) {
        restart = true;
        console.log("speech error", evt);
        running = false;
        command = "speech error";
      }.bind(_this), true);

      recognition.addEventListener("end", function (evt) {
        console.log("speech ended", evt);
        running = false;
        command = "speech ended";
        if (restart) {
          restart = false;
          this.enable(true);
        }
      }.bind(_this), true);

      recognition.addEventListener("result", function (evt) {
        var newCommand = [];
        var result = evt.results[evt.resultIndex];
        var max = 0;
        var maxI = -1;
        if (result && result.isFinal) {
          for (var i = 0; i < result.length; ++i) {
            var alt = result[i];
            if (alt.confidence > max) {
              max = alt.confidence;
              maxI = i;
            }
          }
        }

        if (max > 0.85) {
          newCommand.push(result[maxI].transcript.trim());
        }

        newCommand = newCommand.join(" ");

        if (newCommand !== this.inputState) {
          this.inputState.text = newCommand;
        }
        this.update();
      }.bind(_this), true);

      available = true;
    } catch (exp) {
      console.error(exp);
      errorMessage = exp;
      available = false;
    }
    return _this;
  }

  createClass(Speech, [{
    key: "cloneCommand",
    value: function cloneCommand(cmd) {
      return {
        name: cmd.name,
        preamble: cmd.preamble,
        keywords: Speech.maybeClone(cmd.keywords),
        commandUp: cmd.commandUp,
        disabled: cmd.disabled
      };
    }
  }, {
    key: "evalCommand",
    value: function evalCommand(cmd, cmdState, metaKeysSet, dt) {
      if (metaKeysSet && this.inputState.text) {
        for (var i = 0; i < cmd.keywords.length; ++i) {
          if (this.inputState.text.indexOf(cmd.keywords[i]) === 0 && (cmd.preamble || cmd.keywords[i].length === this.inputState.text.length)) {
            cmdState.pressed = true;
            cmdState.value = this.inputState.text.substring(cmd.keywords[i].length).trim();
            this.inputState.text = null;
          }
        }
      }
    }
  }, {
    key: "enable",
    value: function enable(k, v) {
      get$1(Speech.prototype.__proto__ || Object.getPrototypeOf(Speech.prototype), "enable", this).call(this, k, v);
      this.check();
    }
  }], [{
    key: "maybeClone",
    value: function maybeClone(arr) {
      return arr && arr.slice() || [];
    }
  }]);
  return Speech;
}(InputProcessor);

function lock(element) {
  var type = screen.orientation && screen.orientation.type || screen.mozOrientation || "";
  if (type.indexOf("landscape") === -1) {
    type = "landscape-primary";
  }
  if (screen.orientation && screen.orientation.lock) {
    return screen.orientation.lock(type);
  } else if (screen.mozLockOrientation) {
    var locked = screen.mozLockOrientation(type);
    if (locked) {
      return Promise.resolve(element);
    }
  } else {
    return Promise.reject(new Error("Pointer lock not supported."));
  }
}

function unlock() {
  if (screen.orientation && screen.orientation.unlock) {
    screen.orientation.unlock();
  } else if (screen.mozUnlockOrientation) {
    screen.mozUnlockOrientation();
  }
}

var FullScreen = AsyncLockRequest("Fullscreen", ["fullscreenElement", "mozFullScreenElement", "webkitFullscreenElement", "msFullscreenElement"], ["onfullscreenchange", "onmozfullscreenchange", "onwebkitfullscreenchange", "onmsfullscreenchange"], ["onfullscreenerror", "onmozfullscreenerror", "onwebkitfullscreenerror", "onmsfullscreenerror"], ["requestFullscreen", "mozRequestFullScreen", "webkitRequestFullscreen", "webkitRequestFullScreen", "msRequestFullscreen"], ["exitFullscreen", "mozExitFullScreen", "webkitExitFullscreen", "webkitExitFullScreen", "msExitFullscreen"], function (arg) {
  return arg || window.Element && window.Element.ALLOW_KEYBOARD_INPUT || undefined;
});

function immutable(value) {
  var getter = typeof value === "function" ? value : function () {
    return value;
  };
  return {
    enumerable: true,
    get: getter,
    set: function set() {
      throw new Error("This value is immutable and may only be read, not written.");
    }
  };
}

function mutable(value, type) {
  return {
    enumerable: true,
    get: function get() {
      return value;
    },
    set: function set(v) {
      var t = typeof v === "undefined" ? "undefined" : _typeof(v);
      if (t !== type) {
        throw new Error("Value must be a " + type + ". An " + t + " was provided instead: " + v);
      }
      value = v;
    }
  };
}

var isMobile$1 = (function (a) {
  return (/(android|bb\d+|meego).+|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od|ad)|iris|kindle|lge |maemo|midp|mmp|mobile.+firefox|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows (ce|phone)|xda|xiino/i.test(a) || /1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i.test(a.substring(0, 4))
  );
})(navigator.userAgent || navigator.vendor || window.opera);

var Orientation = { lock: lock, unlock: unlock };

function warn(msg) {
  return function (exp) {
    console.warn(msg, exp);
  };
}

function defaultPose() {
  return {
    position: [0, 0, 0],
    orientation: [0, 0, 0, 1],
    linearVelocity: null,
    linearAcceleration: null,
    angularVelocity: null,
    angularAcceleration: null
  };
}

function fireDisplayPresentChange(evt) {
  if (!FullScreen.isActive) {
    FullScreen.removeChangeListener(fireDisplayPresentChange);
  }
  window.dispatchEvent(new Event("vrdisplaypresentchange"));
}

var WebVRStandardMonitor = function () {
  createClass(WebVRStandardMonitor, null, [{
    key: "standardFullScreenBehavior",
    value: function standardFullScreenBehavior(elem) {
      return FullScreen.request(elem).catch(warn("FullScreen failed")).then(WebVRStandardMonitor.standardLockBehavior);
    }
  }, {
    key: "standardLockBehavior",
    value: function standardLockBehavior(elem) {
      if (isMobile$1) {
        return Orientation.lock(elem).catch(warn("OrientationLock failed"));
      } else {
        return PointerLock.request(elem).catch(warn("PointerLock failed"));
      }
    }
  }, {
    key: "standardExitFullScreenBehavior",
    value: function standardExitFullScreenBehavior() {
      return WebVRStandardMonitor.standardUnlockBehavior().then(function () {
        return FullScreen.exit();
      }).catch(warn("FullScreen failed"));
    }
  }, {
    key: "standardUnlockBehavior",
    value: function standardUnlockBehavior() {
      if (isMobile$1) {
        Orientation.unlock();
        return Promise.resolve();
      } else {
        return PointerLock.exit().catch(warn("PointerLock exit failed"));
      }
    }
  }]);

  function WebVRStandardMonitor(display) {
    classCallCheck(this, WebVRStandardMonitor);

    if (this !== window && this !== undefined) {
      this._currentLayers = [];
      this._display = display;

      Object.defineProperties(this, {
        capabilities: immutable(Object.defineProperties({}, {
          hasPosition: immutable(false),
          hasOrientation: immutable(isMobile$1),
          hasExternalDisplay: immutable(false),
          canPresent: immutable(true),
          maxLayers: immutable(1)
        })),
        isPolyfilled: immutable(display && display.isPolyfilled || false),
        displayId: immutable(0),
        displayName: immutable(isMobile$1 && "Magic Window" || "Standard Monitor"),
        isConnected: immutable(true),
        stageParameters: immutable(null),
        isPresenting: immutable(function () {
          return FullScreen.isActive;
        }),

        depthNear: mutable(0.01, "number"),
        depthFar: mutable(10000.0, "number")
      });
    }
  }

  createClass(WebVRStandardMonitor, [{
    key: "requestAnimationFrame",
    value: function requestAnimationFrame(thunk) {
      return window.requestAnimationFrame(thunk);
    }
  }, {
    key: "cancelAnimationFrame",
    value: function cancelAnimationFrame(handle) {
      window.cancelAnimationFrame(handle);
    }
  }, {
    key: "submitFrame",
    value: function submitFrame() {}
  }, {
    key: "getPose",
    value: function getPose() {
      var display = isMobile$1 && this._display;
      if (display) {
        return display.getPose();
      } else {
        return defaultPose();
      }
    }
  }, {
    key: "getImmediatePose",
    value: function getImmediatePose() {
      var display = isMobile$1 && this._display;
      if (display) {
        return display.getImmediatePose();
      } else {
        return defaultPose();
      }
    }
  }, {
    key: "resetPose",
    value: function resetPose() {
      var display = isMobile$1 && this._display;
      if (display) {
        return display.resetPose();
      }
    }
  }, {
    key: "requestPresent",
    value: function requestPresent(layers) {
      for (var i = 0; i < this.capabilities.maxLayers && i < layers.length; ++i) {
        this._currentLayers[i] = layers[i];
      }
      var elem = layers[0].source;
      if (isMobile$1) {
        return this._display.requestPresent(layers).then(function () {
          return WebVRStandardMonitor.standardLockBehavior(elem);
        });
      } else {
        FullScreen.addChangeListener(fireDisplayPresentChange);
        return WebVRStandardMonitor.standardFullScreenBehavior(elem);
      }
    }
  }, {
    key: "getLayers",
    value: function getLayers() {
      return this._currentLayers.slice();
    }
  }, {
    key: "exitPresent",
    value: function exitPresent() {
      var _this = this;

      this._currentLayers.splice(0);

      if (isMobile$1) {
        return WebVRStandardMonitor.standardUnlockBehavior().then(function () {
          return _this._display.exitPresent();
        });
      } else {
        return WebVRStandardMonitor.standardExitFullScreenBehavior();
      }
    }
  }, {
    key: "getEyeParameters",
    value: function getEyeParameters(side) {
      if (side === "left") {
        var curLayer = this.getLayers()[0],
            elem = curLayer && curLayer.source || document.body,
            width = FullScreen.isActive ? screen.width : elem.clientWidth,
            height = FullScreen.isActive ? screen.height : elem.clientHeight,
            aspect = width / height,
            vFOV = 25,
            hFOV = vFOV * aspect;
        return {
          renderWidth: width * devicePixelRatio,
          renderHeight: height * devicePixelRatio,
          offset: new Float32Array([0, 0, 0]),
          fieldOfView: {
            upDegrees: vFOV,
            downDegrees: vFOV,
            leftDegrees: hFOV,
            rightDegrees: hFOV
          }
        };
      }
    }
  }]);
  return WebVRStandardMonitor;
}();

WebVRStandardMonitor._shimSetup = false;

function install() {
  if (!WebVRStandardMonitor._shimSetup) {
    WebVRStandardMonitor._shimSetup = true;
    var oldGetVRDisplays = navigator.getVRDisplays || Promise.resolve.bind(Promise, []);
    navigator.getVRDisplays = function () {
      return oldGetVRDisplays.call(navigator).then(function (displays) {
        var created = false;
        for (var i = 0; i < displays.length && !created; ++i) {
          var dsp = displays[i];
          created = dsp instanceof WebVRStandardMonitor;
        }
        if (!created) {
          displays.unshift(new WebVRStandardMonitor(displays[0]));
        }
        return displays;
      });
    };
  }
}

var VR = function (_PoseInputProcessor) {
  inherits(VR, _PoseInputProcessor);
  createClass(VR, null, [{
    key: "isStereoDisplay",
    value: function isStereoDisplay(display) {
      var leftParams = display.getEyeParameters("left"),
          rightParams = display.getEyeParameters("right");
      return !!(leftParams && rightParams);
    }
  }]);

  function VR(avatarHeight) {
    classCallCheck(this, VR);

    var _this = possibleConstructorReturn(this, (VR.__proto__ || Object.getPrototypeOf(VR)).call(this, "VR"));

    _this._requestPresent = function (layers) {
      return _this.currentDevice.requestPresent(layers).catch(function (exp) {
        return console.warn("requstPresent", exp);
      });
    };

    _this.displays = [];
    _this._transformers = [];
    _this.currentDeviceIndex = -1;
    _this.movePlayer = new Matrix4();
    _this.defaultAvatarHeight = avatarHeight;
    _this.stage = null;
    _this.lastStageWidth = null;
    _this.lastStageDepth = null;
    _this.isStereo = false;

    install();
    _this.ready = navigator.getVRDisplays().then(function (displays) {
      // We skip the Standard Monitor and Magic Window on iOS because we can't
      // go full screen on those systems.
      _this.displays.push.apply(_this.displays, displays.filter(function (display) {
        return !isiOS || VR.isStereoDisplay(display);
      }));
      return _this.displays;
    });
    return _this;
  }

  createClass(VR, [{
    key: "connect",
    value: function connect(selectedIndex) {
      this.currentDevice = null;
      this.currentDeviceIndex = selectedIndex;
      this.currentPose = null;
      if (0 <= selectedIndex && selectedIndex <= this.displays.length) {
        this.currentDevice = this.displays[selectedIndex];
        this.currentPose = this.currentDevice.getPose();
        this.isStereo = VR.isStereoDisplay(this.currentDevice);
      }
    }
  }, {
    key: "requestPresent",
    value: function requestPresent(opts) {
      var _this2 = this;

      if (!this.currentDevice) {
        return Promise.reject("No display");
      } else {
        var promise;

        var _ret = function () {
          var layers = opts,
              elem = opts[0].source;

          if (!(layers instanceof Array)) {
            layers = [layers];
          }

          // A hack to deal with a bug in the current build of Chromium
          if (_this2.isNativeMobileWebVR && _this2.isStereo) {
            layers = layers[0];
          }

          var rp = _this2._requestPresent;
          promise = null;

          if (isiOS) {
            promise = rp(layers);
          } else if (_this2.currentDevice.capabilities.hasExternalDisplay) {
            // PCs with HMD should also make the browser window on the main
            // display full-screen, so we can then also lock pointer.
            promise = WebVRStandardMonitor.standardFullScreenBehavior(elem).then(function () {
              return rp(layers);
            });
          } else {
            promise = rp(layers).then(WebVRStandardMonitor.standardLockBehavior);
          }
          return {
            v: promise
          };
        }();

        if ((typeof _ret === "undefined" ? "undefined" : _typeof(_ret)) === "object") return _ret.v;
      }
    }
  }, {
    key: "cancel",
    value: function cancel() {
      var _this3 = this;

      var promise = null;
      if (this.isPresenting) {
        promise = this.currentDevice.exitPresent();
        this.currentDevice = null;
        this.currentDeviceIndex = -1;
        this.currentPose = null;
      } else {
        promise = Promise.resolve();
      }

      if (this.isNativeMobileWebVR) {
        promise = promise.then(unlock);
      }

      return promise.then(PointerLock.exit).catch(function (exp) {
        return console.warn(exp);
      }).then(function () {
        return _this3.connect(0);
      });
    }
  }, {
    key: "zero",
    value: function zero() {
      get$1(VR.prototype.__proto__ || Object.getPrototypeOf(VR.prototype), "zero", this).call(this);
      if (this.currentDevice) {
        this.currentDevice.resetPose();
      }
    }
  }, {
    key: "update",
    value: function update(dt) {
      var x, z, stage;

      if (this.currentDevice) {
        this.currentPose = this.currentDevice.getPose();
        stage = this.currentDevice.stageParameters;
      } else {
        stage = null;
      }

      get$1(VR.prototype.__proto__ || Object.getPrototypeOf(VR.prototype), "update", this).call(this, dt);

      if (stage) {
        this.movePlayer.fromArray(stage.sittingToStandingTransform);
        x = stage.sizeX;
        z = stage.sizeZ;
      } else {
        this.movePlayer.makeTranslation(0, this.defaultAvatarHeight, 0);
        x = 0;
        z = 0;
      }

      var s = {
        matrix: this.movePlayer,
        sizeX: x,
        sizeZ: z
      };

      if (!this.stage || s.sizeX !== this.stage.sizeX || s.sizeZ !== this.stage.sizeZ) {
        this.stage = s;
      }
    }
  }, {
    key: "submitFrame",
    value: function submitFrame() {
      if (this.currentDevice) {
        this.currentDevice.submitFrame(this.currentPose);
      }
    }
  }, {
    key: "getTransforms",
    value: function getTransforms(near, far) {
      if (this.currentDevice) {
        if (!this._transformers[this.currentDeviceIndex]) {
          this._transformers[this.currentDeviceIndex] = new ViewCameraTransform(this.currentDevice);
        }
        this.currentDevice.depthNear = near;
        this.currentDevice.depthFar = far;
        return this._transformers[this.currentDeviceIndex].getTransforms(near, far);
      }
    }
  }, {
    key: "isNativeMobileWebVR",
    get: function get() {
      return !(this.currentDevice && this.currentDevice.isPolyfilled) && isChrome && isMobile;
    }
  }, {
    key: "hasStage",
    get: function get() {
      return this.stage && this.stage.sizeX * this.stage.sizeZ > 0;
    }
  }, {
    key: "canMirror",
    get: function get() {
      return this.currentDevice && this.currentDevice.capabilities.hasExternalDisplay;
    }
  }, {
    key: "isPolyfilled",
    get: function get() {
      return this.currentDevice && this.currentDevice.isPolyfilled;
    }
  }, {
    key: "isPresenting",
    get: function get() {
      return this.currentDevice && this.currentDevice.isPresenting;
    }
  }, {
    key: "hasOrientation",
    get: function get() {
      return this.currentDevice && this.currentDevice.capabilities.hasOrientation;
    }
  }, {
    key: "currentCanvas",
    get: function get() {
      if (this.isPresenting) {
        var layers = this.currentDevice.getLayers();
        if (layers.length > 0) {
          return layers[0].source;
        }
      }
      return null;
    }
  }]);
  return VR;
}(PoseInputProcessor);

var ViewCameraTransform = function () {
  createClass(ViewCameraTransform, null, [{
    key: "makeTransform",
    value: function makeTransform(eye, near, far) {
      return {
        translation: new Vector3().fromArray(eye.offset),
        projection: ViewCameraTransform.fieldOfViewToProjectionMatrix(eye.fieldOfView, near, far),
        viewport: {
          left: 0,
          top: 0,
          width: eye.renderWidth,
          height: eye.renderHeight
        }
      };
    }
  }, {
    key: "fieldOfViewToProjectionMatrix",
    value: function fieldOfViewToProjectionMatrix(fov, zNear, zFar) {
      var upTan = Math.tan(fov.upDegrees * Math.PI / 180.0),
          downTan = Math.tan(fov.downDegrees * Math.PI / 180.0),
          leftTan = Math.tan(fov.leftDegrees * Math.PI / 180.0),
          rightTan = Math.tan(fov.rightDegrees * Math.PI / 180.0),
          xScale = 2.0 / (leftTan + rightTan),
          yScale = 2.0 / (upTan + downTan),
          matrix = new Matrix4();

      matrix.elements[0] = xScale;
      matrix.elements[1] = 0.0;
      matrix.elements[2] = 0.0;
      matrix.elements[3] = 0.0;
      matrix.elements[4] = 0.0;
      matrix.elements[5] = yScale;
      matrix.elements[6] = 0.0;
      matrix.elements[7] = 0.0;
      matrix.elements[8] = -((leftTan - rightTan) * xScale * 0.5);
      matrix.elements[9] = (upTan - downTan) * yScale * 0.5;
      matrix.elements[10] = -(zNear + zFar) / (zFar - zNear);
      matrix.elements[11] = -1.0;
      matrix.elements[12] = 0.0;
      matrix.elements[13] = 0.0;
      matrix.elements[14] = -(2.0 * zFar * zNear) / (zFar - zNear);
      matrix.elements[15] = 0.0;

      return matrix;
    }
  }]);

  function ViewCameraTransform(display) {
    classCallCheck(this, ViewCameraTransform);

    this.display = display;
  }

  createClass(ViewCameraTransform, [{
    key: "getTransforms",
    value: function getTransforms(near, far) {
      var l = this.display.getEyeParameters("left"),
          r = this.display.getEyeParameters("right"),
          params = [ViewCameraTransform.makeTransform(l, near, far)];
      if (r) {
        params.push(ViewCameraTransform.makeTransform(r, near, far));
      }
      for (var i = 1; i < params.length; ++i) {
        params[i].viewport.left = params[i - 1].viewport.left + params[i - 1].viewport.width;
      }
      return params;
    }
  }]);
  return ViewCameraTransform;
}();

var DISPLACEMENT = new Vector3();
var EULER_TEMP$1 = new Euler();
var QUAT_TEMP$1 = new Quaternion();
var WEDGE = Math.PI / 3;

var FPSInput = function (_AbstractEventEmitter) {
  inherits(FPSInput, _AbstractEventEmitter);

  function FPSInput(DOMElement, options) {
    classCallCheck(this, FPSInput);

    var _this = possibleConstructorReturn(this, (FPSInput.__proto__ || Object.getPrototypeOf(FPSInput)).call(this));

    _this.options = options;
    _this._handlers.zero = [];
    _this._handlers.motioncontroller = [];
    _this._handlers.gamepad = [];

    _this.managers = [];
    _this.newState = [];
    _this.pointers = [];
    _this.motionDevices = [];
    _this.velocity = new Vector3();
    _this.matrix = new Matrix4();

    if (!_this.options.disableKeyboard) {
      _this.add(new Keyboard(_this, {
        strafeLeft: {
          buttons: [-Keys.A, -Keys.LEFTARROW]
        },
        strafeRight: {
          buttons: [Keys.D, Keys.RIGHTARROW]
        },
        strafe: {
          commands: ["strafeLeft", "strafeRight"]
        },
        lift: {
          buttons: [Keys.E],
          scale: 12
        },
        driveForward: {
          buttons: [-Keys.W, -Keys.UPARROW]
        },
        driveBack: {
          buttons: [Keys.S, Keys.DOWNARROW]
        },
        drive: {
          commands: ["driveForward", "driveBack"]
        },
        select: {
          buttons: [Keys.ENTER]
        },
        dSelect: {
          buttons: [Keys.ENTER],
          delta: true
        },
        zero: {
          buttons: [Keys.Z],
          metaKeys: [-Keys.CTRL, -Keys.ALT, -Keys.SHIFT, -Keys.META],
          commandUp: _this.emit.bind(_this, "zero")
        }
      }));

      _this.Keyboard.operatingSystem = _this.options.os;
      _this.Keyboard.codePage = _this.options.language;
    }

    _this.add(new Touch(DOMElement, {
      buttons: {
        axes: ["FINGERS"]
      },
      dButtons: {
        axes: ["FINGERS"],
        delta: true
      },
      heading: {
        axes: ["DX0"],
        integrate: true
      },
      pitch: {
        axes: ["DY0"],
        integrate: true,
        min: -Math.PI * 0.5,
        max: Math.PI * 0.5
      }
    }));

    _this.add(new Mouse(DOMElement, {
      U: { axes: ["X"], min: 0, max: 2, offset: 0 },
      V: { axes: ["Y"], min: 0, max: 2 },
      buttons: {
        axes: ["BUTTONS"]
      },
      dButtons: {
        axes: ["BUTTONS"],
        delta: true
      },
      _dx: {
        axes: ["X"],
        delta: true,
        scale: 0.25
      },
      dx: {
        buttons: [0],
        commands: ["_dx"]
      },
      heading: {
        commands: ["dx"],
        integrate: true
      },
      _dy: {
        axes: ["Y"],
        delta: true,
        scale: 0.25
      },
      dy: {
        buttons: [0],
        commands: ["_dy"]
      },
      pitch: {
        commands: ["dy"],
        integrate: true,
        min: -Math.PI * 0.5,
        max: Math.PI * 0.5
      }
    }));

    _this.add(new VR(_this.options.avatarHeight));
    _this.motionDevices.push(_this.VR);

    if (!_this.options.disableGamepad && GamepadManager.isAvailable) {
      _this.gamepadMgr = new GamepadManager();
      _this.gamepadMgr.addEventListener("gamepadconnected", function (pad) {
        var padID = Gamepad.ID(pad);
        var mgr = null;

        if (padID !== "Unknown" && padID !== "Rift") {
          if (Gamepad.isMotionController(pad)) {
            var controllerNumber = 0;
            for (var i = 0; i < _this.managers.length; ++i) {
              mgr = _this.managers[i];
              if (mgr.currentPad && mgr.currentPad.id === pad.id) {
                ++controllerNumber;
              }
            }

            mgr = new Gamepad(_this.gamepadMgr, pad, controllerNumber, {
              buttons: {
                axes: ["BUTTONS"]
              },
              dButtons: {
                axes: ["BUTTONS"],
                delta: true
              },
              zero: {
                buttons: [Gamepad.VIVE_BUTTONS.GRIP_PRESSED],
                commandUp: emit.bind(_this, "zero")
              }
            });

            _this.add(mgr);
            _this.motionDevices.push(mgr);

            var shift = (_this.motionDevices.length - 2) * 8,
                color = 0x0000ff << shift,
                highlight = 0xff0000 >> shift,
                ptr = new Pointer(padID + "Pointer", color, 1, highlight, [mgr], null, _this.options);
            ptr.add(colored(box$1(0.1, 0.025, 0.2), color, {
              emissive: highlight
            }));

            _this.pointers.push(ptr);
            ptr.addToBrowserEnvironment(null, _this.options.scene);
            ptr.forward(_this, Pointer.EVENTS);
          } else {
            mgr = new Gamepad(_this.gamepadMgr, pad, 0, {
              buttons: {
                axes: ["BUTTONS"]
              },
              dButtons: {
                axes: ["BUTTONS"],
                delta: true
              },
              strafe: {
                axes: ["LSX"],
                deadzone: 0.2
              },
              drive: {
                axes: ["LSY"],
                deadzone: 0.2
              },
              heading: {
                axes: ["RSX"],
                scale: -1,
                deadzone: 0.2,
                integrate: true
              },
              dHeading: {
                commands: ["heading"],
                delta: true
              },
              pitch: {
                axes: ["RSY"],
                scale: -1,
                deadzone: 0.2,
                integrate: true
              },
              zero: {
                buttons: [Gamepad.XBOX_ONE_BUTTONS.BACK],
                commandUp: emit.bind(_this, "zero")
              }
            });
            _this.add(mgr);
            _this.mousePointer.addDevice(mgr, mgr);
          }
        }
      });

      _this.gamepadMgr.addEventListener("gamepaddisconnected", _this.remove.bind(_this));
    }

    _this.stage = hub$1();

    _this.head = new Pointer("GazePointer", 0xffff00, 0x0000ff, 0.8, [_this.VR], [_this.Mouse, _this.Touch, _this.Keyboard], _this.options);

    _this.head.rotation.order = "YXZ";
    _this.head.useGaze = _this.options.useGaze;
    _this.pointers.push(_this.head);
    _this.options.scene.add(_this.head.root);
    _this.options.scene.add(_this.head.disk);

    _this.mousePointer = new Pointer("MousePointer", 0xff0000, 0x00ff00, 1, [_this.Mouse], null, _this.options);
    _this.mousePointer.unproject = new Matrix4();
    _this.pointers.push(_this.mousePointer);
    _this.head.add(_this.mousePointer.root);
    _this.options.scene.add(_this.mousePointer.disk);

    _this.pointers.forEach(function (ptr) {
      return ptr.forward(_this, Pointer.EVENTS);
    });
    _this.ready = Promise.all(_this.managers.map(function (mgr) {
      return mgr.ready;
    }).filter(identity$1));
    return _this;
  }

  createClass(FPSInput, [{
    key: "remove",
    value: function remove(id) {
      var mgr = this[id],
          mgrIdx = this.managers.indexOf(mgr);
      if (mgrIdx > -1) {
        this.managers.splice(mgrIdx, 1);
        delete this[id];
      }
      console.log("removed", mgr);
    }
  }, {
    key: "add",
    value: function add(mgr) {
      for (var i = this.managers.length - 1; i >= 0; --i) {
        if (this.managers[i].name === mgr.name) {
          this.managers.splice(i, 1);
        }
      }
      this.managers.push(mgr);
      this[mgr.name] = mgr;
    }
  }, {
    key: "zero",
    value: function zero() {
      for (var i = 0; i < this.managers.length; ++i) {
        this.managers[i].zero();
      }
    }
  }, {
    key: "submitFrame",
    value: function submitFrame() {
      this.VR.submitFrame();
    }
  }, {
    key: "update",
    value: function update(dt) {
      var hadGamepad = this.hasGamepad;
      if (this.gamepadMgr) {
        this.gamepadMgr.poll();
      }
      for (var i = 0; i < this.managers.length; ++i) {
        this.managers[i].update(dt);
      }

      if (!hadGamepad && this.hasGamepad) {
        this.Mouse.inPhysicalUse = false;
      }

      this.head.showPointer = this.VR.hasOrientation && this.options.showHeadPointer;
      this.mousePointer.showPointer = (this.hasMouse || this.hasGamepad) && !this.hasMotionControllers;

      if (this.Keyboard) {
        this.Keyboard.enabled = this.Touch.enabled = this.Mouse.enabled = !this.hasMotionControllers;
      }

      if (this.Gamepad_0) {
        this.Gamepad_0.enabled = !this.hasMotionControllers;
      }

      this.updateStage(dt);
      this.stage.position.y = this.options.avatarHeight;
      for (var _i = 0; _i < this.motionDevices.length; ++_i) {
        this.motionDevices[_i].posePosition.y -= this.options.avatarHeight;
      }

      // update the motionDevices
      this.stage.updateMatrix();
      this.matrix.multiplyMatrices(this.stage.matrix, this.VR.stage.matrix);
      for (var _i2 = 0; _i2 < this.motionDevices.length; ++_i2) {
        this.motionDevices[_i2].updateStage(this.matrix);
      }

      for (var _i3 = 0; _i3 < this.pointers.length; ++_i3) {
        this.pointers[_i3].update();
      }

      // record the position and orientation of the user
      this.newState = [];
      this.head.updateMatrix();
      this.stage.rotation.x = 0;
      this.stage.rotation.z = 0;
      this.stage.quaternion.setFromEuler(this.stage.rotation);
      this.stage.updateMatrix();
      this.head.position.toArray(this.newState, 0);
      this.head.quaternion.toArray(this.newState, 3);
    }
  }, {
    key: "updateStage",
    value: function updateStage(dt) {
      // get the linear movement from the mouse/keyboard/gamepad
      var heading = 0,
          pitch = 0,
          strafe = 0,
          drive = 0,
          lift = 0;
      for (var i = 0; i < this.managers.length; ++i) {
        var mgr = this.managers[i];
        if (mgr.enabled) {
          if (mgr.name !== "Mouse") {
            heading += mgr.getValue("heading");
          }
          pitch += mgr.getValue("pitch");
          strafe += mgr.getValue("strafe");
          drive += mgr.getValue("drive");
          lift += mgr.getValue("lift");
        }
      }

      if (this.hasMouse) {
        var mouseHeading = null;
        if (this.VR.hasOrientation) {
          mouseHeading = this.mousePointer.root.rotation.y;
          var newMouseHeading = WEDGE * Math.floor(mouseHeading / WEDGE + 0.5);
          if (newMouseHeading !== 0) {
            this.Mouse.commands.U.offset -= this.Mouse.getValue("U") - 1;
          }
          mouseHeading = newMouseHeading + this.Mouse.commands.U.offset * 2;
        } else {
          mouseHeading = this.Mouse.getValue("heading");
        }
        heading += mouseHeading;
      }
      if (this.VR.hasOrientation) {
        pitch = 0;
      }

      // move stage according to heading and thrust
      EULER_TEMP$1.set(pitch, heading, 0, "YXZ");
      this.stage.quaternion.setFromEuler(EULER_TEMP$1);

      // update the stage's velocity
      this.velocity.set(strafe, lift, drive);

      if (this.stage.isOnGround) {
        if (this.velocity.y > 0) {
          this.stage.isOnGround = false;
        }
      } else {
        this.velocity.y -= this.options.gravity;
        if (this.stage.position.y < 0) {
          this.velocity.y = 0;
          this.stage.position.y = 0;
          this.stage.isOnGround = true;
        }
      }

      QUAT_TEMP$1.copy(this.head.quaternion);
      EULER_TEMP$1.setFromQuaternion(QUAT_TEMP$1);
      EULER_TEMP$1.x = 0;
      EULER_TEMP$1.z = 0;
      QUAT_TEMP$1.setFromEuler(EULER_TEMP$1);

      this.moveStage(DISPLACEMENT.copy(this.velocity).multiplyScalar(dt).applyQuaternion(QUAT_TEMP$1).add(this.head.position));
    }
  }, {
    key: "cancelVR",
    value: function cancelVR() {
      this.VR.cancel();
      this.Mouse.commands.U.offset = 0;
    }
  }, {
    key: "moveStage",
    value: function moveStage(position) {
      DISPLACEMENT.copy(position).sub(this.head.position);
      this.stage.position.add(DISPLACEMENT);
    }
  }, {
    key: "resolvePicking",
    value: function resolvePicking(objects) {
      for (var i = 0; i < this.pointers.length; ++i) {
        var ptr = this.pointers[i];
        ptr.resolvePicking(objects);
      }
    }
  }, {
    key: "hasMotionControllers",
    get: function get() {
      return !!(this.Vive_0 && this.Vive_0.enabled && this.Vive_0.inPhysicalUse || this.Vive_1 && this.Vive_1.enabled && this.Vive_1.inPhysicalUse);
    }
  }, {
    key: "hasGamepad",
    get: function get() {
      return !!(this.Gamepad_0 && this.Gamepad_0.enabled && this.Gamepad_0.inPhysicalUse);
    }
  }, {
    key: "hasMouse",
    get: function get() {
      return !this.hasTouch && !!(this.Mouse && this.Mouse.enabled && this.Mouse.inPhysicalUse);
    }
  }, {
    key: "hasTouch",
    get: function get() {
      return !!(this.Touch && this.Touch.enabled && this.Touch.inPhysicalUse);
    }
  }, {
    key: "segments",
    get: function get() {
      var segments = [];
      for (var i = 0; i < this.pointers.length; ++i) {
        var seg = this.pointers[i].segment;
        if (seg) {
          segments.push(seg);
        }
      }
      return segments;
    }
  }]);
  return FPSInput;
}(AbstractEventEmitter);

var Location = function (_InputProcessor) {
  inherits(Location, _InputProcessor);

  function Location(commands, options) {
    classCallCheck(this, Location);

    var _this = possibleConstructorReturn(this, (Location.__proto__ || Object.getPrototypeOf(Location)).call(this, "Location", commands, ["LONGITUDE", "LATITUDE", "ALTITUDE", "HEADING", "SPEED"]));

    _this.options = Object.assign({}, Location.DEFAULTS, options);

    _this.available = !!navigator.geolocation;
    if (_this.available) {
      navigator.geolocation.watchPosition(_this.setState.bind(_this), function () {
        return _this.available = false;
      }, _this.options);
    }
    return _this;
  }

  createClass(Location, [{
    key: "setState",
    value: function setState(location) {
      for (var p in location.coords) {
        var k = p.toUpperCase();
        if (this.axisNames.indexOf(k) > -1) {
          this.setAxis(k, location.coords[p]);
        }
      }
      this.update();
    }
  }]);
  return Location;
}(InputProcessor);



Location.DEFAULTS = {
  enableHighAccuracy: true,
  maximumAge: 30000,
  timeout: 25000
};

function number(min, max, power) {
  power = power || 1;
  if (max === undefined) {
    max = min;
    min = 0;
  }
  var delta = max - min,
      n = Math.pow(Math.random(), power);
  return min + n * delta;
}

function int(min, max, power) {
  return Math.floor(number(min, max, power));
}

function color() {
  var r = int(0, 256),
      g = int(0, 256),
      b = int(0, 256);
  return r << 16 | g << 8 | b;
}

var RemoteUser = function (_AbstractEventEmitter) {
  inherits(RemoteUser, _AbstractEventEmitter);

  function RemoteUser(userName, modelFactory, nameMaterial, disableWebRTC, requestICEPath, microphone, localUserName, goSecond) {
    classCallCheck(this, RemoteUser);

    var _this = possibleConstructorReturn(this, (RemoteUser.__proto__ || Object.getPrototypeOf(RemoteUser)).call(this));

    _this.time = 0;

    _this.userName = userName;
    _this.peeringError = null;
    _this.peering = false;
    _this.peered = false;
    _this.stage = modelFactory.clone();
    _this.stage.traverse(function (obj) {
      if (obj.name === "AvatarBelt") {
        colored(obj, color());
      } else if (obj.name === "AvatarHead") {
        _this.head = obj;
      }
    });

    _this.nameObject = colored(text3D(0.1, userName), nameMaterial);
    var bounds = _this.nameObject.geometry.boundingBox.max;
    _this.nameObject.rotation.set(0, Math.PI, 0);
    _this.nameObject.position.set(bounds.x / 2, bounds.y, 0);
    _this.head.add(_this.nameObject);

    _this.dStageQuaternion = new Quaternion();
    _this.dHeadPosition = new Vector3();
    _this.dHeadQuaternion = new Quaternion();

    _this.lastStageQuaternion = new Quaternion();
    _this.lastHeadPosition = new Vector3();
    _this.lastHeadQuaternion = new Quaternion();

    _this.headPosition = {
      arr1: [],
      arr2: [],
      last: _this.lastHeadPosition,
      delta: _this.dHeadPosition,
      curr: _this.head.position
    };
    _this.headQuaternion = {
      arr1: [],
      arr2: [],
      last: _this.lastHeadQuaternion,
      delta: _this.dHeadQuaternion,
      curr: _this.head.quaternion
    };

    _this.audioChannel = null;
    _this.audioElement = null;
    _this.audioStream = null;
    _this.gain = null;
    _this.panner = null;
    _this.analyzer = null;
    return _this;
  }

  createClass(RemoteUser, [{
    key: "setAudio",
    value: function setAudio(audio, audioSource) {
      if (audioSource instanceof Element) {
        this.audioElement = audioSource;
        Audio3D.setAudioProperties(this.audioElement);
        this.audioStream = audio.context.createMediaElementSource(this.audioElement);
      } else {
        this.audioElement = Audio3D.setAudioStream(audioSource, "audio" + this.userName);
        this.audioStream = audio.context.createMediaStreamSource(audioSource);
      }
      this.gain = audio.context.createGain();
      this.panner = audio.context.createPanner();

      this.audioStream.connect(this.gain);
      this.gain.connect(this.panner);
      this.panner.connect(audio.mainVolume);
      this.panner.coneInnerAngle = 180;
      this.panner.coneOuterAngle = 360;
      this.panner.coneOuterGain = 0.1;
      this.panner.panningModel = "HRTF";
      this.panner.distanceModel = "exponential";
    }
  }, {
    key: "unpeer",
    value: function unpeer() {
      if (this.audioChannel) {
        this.audioChannel.close();
        if (this.audioElement) {
          document.body.removeChild(this.audioElement);
          if (this.panner) {
            this.panner.disconnect();
            this.gain.disconnect();
            this.audioStream.disconnect();
          }
        }
      }
    }
  }, {
    key: "_updateV",
    value: function _updateV(v, dt, fade) {
      v.curr.toArray(v.arr1);
      v.delta.toArray(v.arr2);
      for (var i = 0; i < v.arr1.length; ++i) {
        if (fade) {
          v.arr2[i] *= RemoteUser.FADE_FACTOR;
        }
        v.arr1[i] += v.arr2[i] * dt;
      }

      v.curr.fromArray(v.arr1);
      v.delta.fromArray(v.arr2);
    }
  }, {
    key: "_predict",
    value: function _predict(v, state, off) {
      v.delta.fromArray(state, off);
      v.delta.toArray(v.arr1);
      v.curr.toArray(v.arr2);
      for (var i = 0; i < v.arr1.length; ++i) {
        v.arr1[i] = (v.arr1[i] - v.arr2[i]) * RemoteUser.NETWORK_DT_INV;
      }
      v.delta.fromArray(v.arr1);
    }
  }, {
    key: "update",
    value: function update(dt) {
      this.time += dt;
      var fade = this.time >= RemoteUser.NETWORK_DT;
      this._updateV(this.headPosition, dt, fade);
      this._updateV(this.headQuaternion, dt, fade);
      this.stage.rotation.setFromQuaternion(this.headQuaternion.curr);
      this.stage.rotation.x = 0;
      this.stage.rotation.z = 0;
      this.stage.position.copy(this.headPosition.curr);
      this.stage.position.y = 0;
      if (this.panner) {
        this.panner.setPosition(this.stage.position.x, this.stage.position.y, this.stage.position.z);
        this.panner.setOrientation(Math.sin(this.stage.rotation.y), 0, Math.cos(this.stage.rotation.y));
      }
    }
  }, {
    key: "setState",
    value: function setState(v) {
      this.time = 0;
      this._predict(this.headPosition, v, 1);
      this._predict(this.headQuaternion, v, 4);
    }
  }, {
    key: "toString",
    value: function toString(digits) {
      return this.stage.position.curr.toString(digits) + " " + this.headPosition.curr.toString(digits);
    }
  }]);
  return RemoteUser;
}(AbstractEventEmitter);

RemoteUser.FADE_FACTOR = 0.5;
RemoteUser.NETWORK_DT = 0.10;
RemoteUser.NETWORK_DT_INV = 1 / RemoteUser.NETWORK_DT;

var Manager = function (_AbstractEventEmitter) {
  inherits(Manager, _AbstractEventEmitter);

  function Manager(localUser, audio, factories, options) {
    classCallCheck(this, Manager);

    var _this = possibleConstructorReturn(this, (Manager.__proto__ || Object.getPrototypeOf(Manager)).call(this));

    _this.localUser = localUser;
    _this.audio = audio;
    _this.factories = factories;
    _this.options = options;
    _this.lastNetworkUpdate = 0;
    _this.oldState = [];
    _this.users = {};
    _this.waitForLastUser = Promise.resolve();
    _this._socket = null;
    _this.userName = null;
    _this.microphone = null;
    _this.audioHeap = {};
    return _this;
  }

  createClass(Manager, [{
    key: "update",
    value: function update(dt) {
      if (this._socket && this.deviceIndex === 0) {
        this.lastNetworkUpdate += dt;
        if (this.lastNetworkUpdate >= RemoteUser.NETWORK_DT) {
          this.lastNetworkUpdate -= RemoteUser.NETWORK_DT;
          for (var i = 0; i < this.localUser.newState.length; ++i) {
            if (this.oldState[i] !== this.localUser.newState[i]) {
              this._socket.emit("userState", this.localUser.newState);
              this.oldState = this.localUser.newState;
              break;
            }
          }
        }
      }
      for (var key in this.users) {
        var user = this.users[key];
        user.update(dt);
        if (this.audioHeap[key]) {
          user.setAudio(this.audio, this.audioHeap[key]);
          delete this.audioHeap[key];
        }
      }
    }
  }, {
    key: "updateUser",
    value: function updateUser(state) {
      var key = state[0];
      if (key !== this.userName) {
        var user = this.users[key];
        if (user) {
          user.setState(state);
        }
      } else if (this.deviceIndex > 0) {
        this.localUser.stage.position.fromArray(state, 1);
        this.localUser.stage.quaternion.fromArray(state, 4);
        this.localUser.head.position.fromArray(state, 8);
        this.localUser.head.quaternion.fromArray(state, 11);
      }
    }
  }, {
    key: "connect",
    value: function connect(socket, userName) {
      this.userName = userName.toLocaleUpperCase();
      if (!this.microphone) {
        this.microphone = navigator.mediaDevices.getUserMedia({
          audio: true,
          video: false
        }).catch(console.warn.bind(console, "Can't get audio"));
      }
      if (!this._socket) {
        this._socket = socket;
        this._socket.on("userList", this.listUsers.bind(this));
        this._socket.on("userJoin", this.addUser.bind(this));
        this._socket.on("deviceAdded", this.addDevice.bind(this));
        this._socket.on("deviceIndex", this.setDeviceIndex.bind(this));
        this._socket.on("chat", this.receiveChat.bind(this));
        this._socket.on("userState", this.updateUser.bind(this));
        this._socket.on("userLeft", this.removeUser.bind(this));
        this._socket.on("connection_lost", this.lostConnection.bind(this));
        this._socket.emit("listUsers");
        this._socket.emit("getDeviceIndex");
      }
    }
  }, {
    key: "disconnect",
    value: function disconnect() {
      this.userName = null;
      this._socket.close();
      this._socket = null;
    }
  }, {
    key: "addUser",
    value: function addUser(state, goSecond) {
      console.log("User %s logging on.", state[0]);
      var toUserName = state[0],
          user = new RemoteUser(toUserName, this.factories.avatar, this.options.foregroundColor, this.options.disableWebRTC, this.options.webRTC, this.microphone, this.userName, goSecond);
      this.users[toUserName] = user;
      this.updateUser(state);
      this.emit("addavatar", user);
    }
  }, {
    key: "removeUser",
    value: function removeUser(key) {
      console.log("User %s logging off.", key);
      var user = this.users[key];
      if (user) {
        if (user.peered) {
          user.unpeer();
        }
        delete this.users[key];
        this.emit("removeavatar", user);
      }
    }
  }, {
    key: "listUsers",
    value: function listUsers(newUsers) {
      Object.keys(this.users).forEach(this.removeUser.bind(this));
      while (newUsers.length > 0) {
        this.addUser(newUsers.shift(), true);
      }
      this.emit("authorizationsucceeded");
    }
  }, {
    key: "receiveChat",
    value: function receiveChat(evt) {
      console.log("chat", evt);
    }
  }, {
    key: "lostConnection",
    value: function lostConnection() {
      this.deviceIndex = null;
    }
  }, {
    key: "addDevice",
    value: function addDevice(index) {
      console.log("addDevice", index);
    }
  }, {
    key: "setDeviceIndex",
    value: function setDeviceIndex(index) {
      this.deviceIndex = index;
    }
  }, {
    key: "setAudioFromUser",
    value: function setAudioFromUser(userName, audioElement) {
      this.audioHeap[userName] = audioElement;
    }
  }]);
  return Manager;
}(AbstractEventEmitter);

var Basic = new Grammar("BASIC",
// Grammar rules are applied in the order they are specified.
[
// Text needs at least the newlines token, or else every line will attempt to render as a single line and the line count won't work.
["newlines", /(?:\r\n|\r|\n)/],
// BASIC programs used to require the programmer type in her own line numbers. The start at the beginning of the line.
["lineNumbers", /^\d+\s+/],
// Comments were lines that started with the keyword "REM" (for REMARK) and ran to the end of the line. They did not have to be numbered, because they were not executable and were stripped out by the interpreter.
["startLineComments", /^REM\s/],
// Both double-quoted and single-quoted strings were not always supported, but in this case, I'm just demonstrating how it would be done for both.
["strings", /"(?:\\"|[^"])*"/], ["strings", /'(?:\\'|[^'])*'/],
// Numbers are an optional dash, followed by a optional digits, followed by optional period, followed by 1 or more required digits. This allows us to match both integers and decimal numbers, both positive and negative, with or without leading zeroes for decimal numbers between (-1, 1).
["numbers", /-?(?:(?:\b\d*)?\.)?\b\d+\b/],
// Keywords are really just a list of different words we want to match, surrounded by the "word boundary" selector "\b".
["keywords", /\b(?:RESTORE|REPEAT|RETURN|LOAD|LABEL|DATA|READ|THEN|ELSE|FOR|DIM|LET|IF|TO|STEP|NEXT|WHILE|WEND|UNTIL|GOTO|GOSUB|ON|TAB|AT|END|STOP|PRINT|INPUT|RND|INT|CLS|CLK|LEN)\b/],
// Sometimes things we want to treat as keywords have different meanings in different locations. We can specify rules for tokens more than once.
["keywords", /^DEF FN/],
// These are all treated as mathematical operations.
["operators", /(?:\+|;|,|-|\*\*|\*|\/|>=|<=|=|<>|<|>|OR|AND|NOT|MOD|\(|\)|\[|\])/],
// Once everything else has been matched, the left over blocks of words are treated as variable and function names.
["identifiers", /\w+\$?/]]);
var oldTokenize = Basic.tokenize;
Basic.tokenize = function (code) {
  return oldTokenize.call(this, code.toUpperCase());
};

Basic.interpret = function (sourceCode, input, output, errorOut, next, clearScreen, loadFile, done) {
  var tokens = this.tokenize(sourceCode),
      EQUAL_SIGN = new Token("=", "operators"),
      counter = 0,
      isDone = false,
      program = {},
      lineNumbers = [],
      currentLine = [],
      lines = [currentLine],
      data = [],
      returnStack = [],
      forLoopCounters = {},
      dataCounter = 0,
      state = {
    INT: function INT(v) {
      return v | 0;
    },
    RND: function RND() {
      return Math.random();
    },
    CLK: function CLK() {
      return Date.now() / 3600000;
    },
    LEN: function LEN(id) {
      return id.length;
    },
    LINE: function LINE() {
      return lineNumbers[counter];
    },
    TAB: function TAB(v) {
      var str = "";
      for (var i = 0; i < v; ++i) {
        str += " ";
      }
      return str;
    },
    POW: function POW(a, b) {
      return Math.pow(a, b);
    }
  };

  function toNum(ln) {
    return new Token(ln.toString(), "numbers");
  }

  function toStr(str) {
    return new Token("\"" + str.replace("\n", "\\n").replace("\"", "\\\"") + "\"", "strings");
  }

  var tokenMap = {
    "OR": "||",
    "AND": "&&",
    "NOT": "!",
    "MOD": "%",
    "<>": "!="
  };

  while (tokens.length > 0) {
    var token = tokens.shift();
    if (token.type === "newlines") {
      currentLine = [];
      lines.push(currentLine);
    } else if (token.type !== "regular" && token.type !== "comments") {
      token.value = tokenMap[token.value] || token.value;
      currentLine.push(token);
    }
  }

  for (var i = 0; i < lines.length; ++i) {
    var line = lines[i];
    if (line.length > 0) {
      var lastLine = lineNumbers[lineNumbers.length - 1];
      var lineNumber = line.shift();

      if (lineNumber.type !== "lineNumbers") {
        line.unshift(lineNumber);

        if (lastLine === undefined) {
          lastLine = -1;
        }

        lineNumber = toNum(lastLine + 1);
      }

      lineNumber = parseFloat(lineNumber.value);
      if (lastLine && lineNumber <= lastLine) {
        throw new Error("expected line number greater than " + lastLine + ", but received " + lineNumber + ".");
      } else if (line.length > 0) {
        lineNumbers.push(lineNumber);
        program[lineNumber] = line;
      }
    }
  }

  function process(line) {
    if (line && line.length > 0) {
      var op = line.shift();
      if (op) {
        if (commands.hasOwnProperty(op.value)) {
          return commands[op.value](line);
        } else if (!isNaN(op.value)) {
          return setProgramCounter([op]);
        } else if (state[op.value] || line.length > 0 && line[0].type === "operators" && line[0].value === "=") {
          line.unshift(op);
          return translate(line);
        } else {
          error("Unknown command. >>> " + op.value);
        }
      }
    }
    return pauseBeforeComplete();
  }

  function error(msg) {
    errorOut("At line " + lineNumbers[counter] + ": " + msg);
  }

  function getLine(i) {
    var lineNumber = lineNumbers[i];
    var line = program[lineNumber];
    return line && line.slice();
  }

  function evaluate(line) {
    var script = "";
    for (var i = 0; i < line.length; ++i) {
      var t = line[i];
      var nest = 0;
      if (t.type === "identifiers" && typeof state[t.value] !== "function" && i < line.length - 1 && line[i + 1].value === "(") {
        for (var j = i + 1; j < line.length; ++j) {
          var t2 = line[j];
          if (t2.value === "(") {
            if (nest === 0) {
              t2.value = "[";
            }
            ++nest;
          } else if (t2.value === ")") {
            --nest;
            if (nest === 0) {
              t2.value = "]";
            }
          } else if (t2.value === "," && nest === 1) {
            t2.value = "][";
          }

          if (nest === 0) {
            break;
          }
        }
      }
      script += t.value;
    }
    //with ( state ) { // jshint ignore:line
    try {
      return eval(script); // jshint ignore:line
    } catch (exp) {
      console.error(exp);
      console.debug(line.join(", "));
      console.error(script);
      error(exp.message + ": " + script);
    }
    //}
  }

  function declareVariable(line) {
    var decl = [],
        decls = [decl],
        nest = 0,
        i;
    for (i = 0; i < line.length; ++i) {
      var t = line[i];
      if (t.value === "(") {
        ++nest;
      } else if (t.value === ")") {
        --nest;
      }
      if (nest === 0 && t.value === ",") {
        decl = [];
        decls.push(decl);
      } else {
        decl.push(t);
      }
    }
    for (i = 0; i < decls.length; ++i) {
      decl = decls[i];
      var id = decl.shift();
      if (id.type !== "identifiers") {
        error("Identifier expected: " + id.value);
      } else {
        var val = null,
            j;
        id = id.value;
        if (decl[0].value === "(" && decl[decl.length - 1].value === ")") {
          var sizes = [];
          for (j = 1; j < decl.length - 1; ++j) {
            if (decl[j].type === "numbers") {
              sizes.push(decl[j].value | 0);
            }
          }
          if (sizes.length === 0) {
            val = [];
          } else {
            val = new Array(sizes[0]);
            var queue = [val];
            for (j = 1; j < sizes.length; ++j) {
              var size = sizes[j];
              for (var k = 0, l = queue.length; k < l; ++k) {
                var arr = queue.shift();
                for (var m = 0; m < arr.length; ++m) {
                  arr[m] = new Array(size);
                  if (j < sizes.length - 1) {
                    queue.push(arr[m]);
                  }
                }
              }
            }
          }
        }
        state[id] = val;
        return true;
      }
    }
  }

  function print(line) {
    var endLine = "\n";
    var nest = 0;
    line = line.map(function (t, i) {
      t = t.clone();
      if (t.type === "operators") {
        if (t.value === ",") {
          if (nest === 0) {
            t.value = "+ \", \" + ";
          }
        } else if (t.value === ";") {
          t.value = "+ \" \"";
          if (i < line.length - 1) {
            t.value += " + ";
          } else {
            endLine = "";
          }
        } else if (t.value === "(") {
          ++nest;
        } else if (t.value === ")") {
          --nest;
        }
      }
      return t;
    });
    var txt = evaluate(line);
    if (txt === undefined) {
      txt = "";
    }
    output(txt + endLine);
    return true;
  }

  function setProgramCounter(line) {
    var lineNumber = parseFloat(evaluate(line));
    counter = -1;
    while (counter < lineNumbers.length - 1 && lineNumbers[counter + 1] < lineNumber) {
      ++counter;
    }

    return true;
  }

  function checkConditional(line) {
    var thenIndex = -1,
        elseIndex = -1,
        i;
    for (i = 0; i < line.length; ++i) {
      if (line[i].type === "keywords" && line[i].value === "THEN") {
        thenIndex = i;
      } else if (line[i].type === "keywords" && line[i].value === "ELSE") {
        elseIndex = i;
      }
    }
    if (thenIndex === -1) {
      error("Expected THEN clause.");
    } else {
      var condition = line.slice(0, thenIndex);
      for (i = 0; i < condition.length; ++i) {
        var t = condition[i];
        if (t.type === "operators" && t.value === "=") {
          t.value = "==";
        }
      }
      var thenClause, elseClause;
      if (elseIndex === -1) {
        thenClause = line.slice(thenIndex + 1);
      } else {
        thenClause = line.slice(thenIndex + 1, elseIndex);
        elseClause = line.slice(elseIndex + 1);
      }
      if (evaluate(condition)) {
        return process(thenClause);
      } else if (elseClause) {
        return process(elseClause);
      }
    }

    return true;
  }

  function pauseBeforeComplete() {
    output("PROGRAM COMPLETE - PRESS RETURN TO FINISH.");
    input(function () {
      isDone = true;
      if (done) {
        done();
      }
    });
    return false;
  }

  function labelLine(line) {
    line.push(EQUAL_SIGN);
    line.push(toNum(lineNumbers[counter]));
    return translate(line);
  }

  function waitForInput(line) {
    var toVar = line.pop();
    if (line.length > 0) {
      print(line);
    }
    input(function (str) {
      str = str.toUpperCase();
      var valueToken = null;
      if (!isNaN(str)) {
        valueToken = toNum(str);
      } else {
        valueToken = toStr(str);
      }
      evaluate([toVar, EQUAL_SIGN, valueToken]);
      if (next) {
        next();
      }
    });
    return false;
  }

  function onStatement(line) {
    var idxExpr = [],
        idx = null,
        targets = [];
    try {
      while (line.length > 0 && (line[0].type !== "keywords" || line[0].value !== "GOTO")) {
        idxExpr.push(line.shift());
      }

      if (line.length > 0) {
        line.shift(); // burn the goto;

        for (var i = 0; i < line.length; ++i) {
          var t = line[i];
          if (t.type !== "operators" || t.value !== ",") {
            targets.push(t);
          }
        }

        idx = evaluate(idxExpr) - 1;

        if (0 <= idx && idx < targets.length) {
          return setProgramCounter([targets[idx]]);
        }
      }
    } catch (exp) {
      console.error(exp);
    }
    return true;
  }

  function gotoSubroutine(line) {
    returnStack.push(toNum(lineNumbers[counter + 1]));
    return setProgramCounter(line);
  }

  function setRepeat() {
    returnStack.push(toNum(lineNumbers[counter]));
    return true;
  }

  function conditionalReturn(cond) {
    var ret = true;
    var val = returnStack.pop();
    if (val && cond) {
      ret = setProgramCounter([val]);
    }
    return ret;
  }

  function untilLoop(line) {
    var cond = !evaluate(line);
    return conditionalReturn(cond);
  }

  function findNext(str) {
    for (i = counter + 1; i < lineNumbers.length; ++i) {
      var l = getLine(i);
      if (l[0].value === str) {
        return i;
      }
    }
    return lineNumbers.length;
  }

  function whileLoop(line) {
    var cond = evaluate(line);
    if (!cond) {
      counter = findNext("WEND");
    } else {
      returnStack.push(toNum(lineNumbers[counter]));
    }
    return true;
  }

  var FOR_LOOP_DELIMS = ["=", "TO", "STEP"];

  function forLoop(line) {
    var n = lineNumbers[counter];
    var varExpr = [];
    var fromExpr = [];
    var toExpr = [];
    var skipExpr = [];
    var arrs = [varExpr, fromExpr, toExpr, skipExpr];
    var a = 0;
    var i = 0;
    for (i = 0; i < line.length; ++i) {
      var t = line[i];
      if (t.value === FOR_LOOP_DELIMS[a]) {
        if (a === 0) {
          varExpr.push(t);
        }
        ++a;
      } else {
        arrs[a].push(t);
      }
    }

    var skip = 1;
    if (skipExpr.length > 0) {
      skip = evaluate(skipExpr);
    }

    if (forLoopCounters[n] === undefined) {
      forLoopCounters[n] = evaluate(fromExpr);
    }

    var end = evaluate(toExpr);
    var cond = forLoopCounters[n] <= end;
    if (!cond) {
      delete forLoopCounters[n];
      counter = findNext("NEXT");
    } else {
      varExpr.push(toNum(forLoopCounters[n]));
      process(varExpr);
      forLoopCounters[n] += skip;
      returnStack.push(toNum(lineNumbers[counter]));
    }
    return true;
  }

  function stackReturn() {
    return conditionalReturn(true);
  }

  function loadCodeFile(line) {
    loadFile(evaluate(line)).then(next);
    return false;
  }

  function noop() {
    return true;
  }

  function loadData(line) {
    while (line.length > 0) {
      var t = line.shift();
      if (t.type !== "operators") {
        data.push(t.value);
      }
    }
    return true;
  }

  function readData(line) {
    if (data.length === 0) {
      var dataLine = findNext("DATA");
      process(getLine(dataLine));
    }
    var value = data[dataCounter];
    ++dataCounter;
    line.push(EQUAL_SIGN);
    line.push(toNum(value));
    return translate(line);
  }

  function restoreData() {
    dataCounter = 0;
    return true;
  }

  function defineFunction(line) {
    var name = line.shift().value;
    var signature = "";
    var body = "";
    var fillSig = true;
    for (var i = 0; i < line.length; ++i) {
      var t = line[i];
      if (t.type === "operators" && t.value === "=") {
        fillSig = false;
      } else if (fillSig) {
        signature += t.value;
      } else {
        body += t.value;
      }
    }
    name = "FN" + name;
    var script = "(function " + name + signature + "{ return " + body + "; })";
    state[name] = eval(script); // jshint ignore:line
    return true;
  }

  function translate(line) {
    evaluate(line);
    return true;
  }

  var commands = {
    DIM: declareVariable,
    LET: translate,
    PRINT: print,
    GOTO: setProgramCounter,
    IF: checkConditional,
    INPUT: waitForInput,
    END: pauseBeforeComplete,
    STOP: pauseBeforeComplete,
    REM: noop,
    "'": noop,
    CLS: clearScreen,
    ON: onStatement,
    GOSUB: gotoSubroutine,
    RETURN: stackReturn,
    LOAD: loadCodeFile,
    DATA: loadData,
    READ: readData,
    RESTORE: restoreData,
    REPEAT: setRepeat,
    UNTIL: untilLoop,
    "DEF FN": defineFunction,
    WHILE: whileLoop,
    WEND: stackReturn,
    FOR: forLoop,
    NEXT: stackReturn,
    LABEL: labelLine
  };

  return function () {
    if (!isDone) {
      var goNext = true;
      while (goNext) {
        var line = getLine(counter);
        goNext = process(line);
        ++counter;
      }
    }
  };
};

new Grammar("TestResults", [["newlines", /(?:\r\n|\r|\n)/, true], ["numbers", /(\[)(o+)/, true], ["numbers", /(\d+ succeeded), 0 failed/, true], ["numbers", /^    Successes:/, true], ["functions", /(x+)\]/, true], ["functions", /[1-9]\d* failed/, true], ["functions", /^    Failures:/, true], ["comments", /(\d+ms:)(.*)/, true], ["keywords", /(Test results for )(\w+):/, true], ["strings", /        \w+/, true]]);

var Terminal = function Terminal(inputEditor, outputEditor) {
  classCallCheck(this, Terminal);

  outputEditor = outputEditor || inputEditor;

  var inputCallback = null,
      currentProgram = null,
      originalGrammar = null,
      currentEditIndex = 0,
      pageSize = 40,
      outputQueue = [],
      buffer = "",
      restoreInput = inputEditor === outputEditor,
      self = this;

  this.running = false;
  this.waitingForInput = false;

  function toEnd(editor) {
    editor.selectionStart = editor.selectionEnd = editor.value.length;
    editor.scrollIntoView(editor.frontCursor);
  }

  function done() {
    if (self.running) {
      flush();
      self.running = false;
      if (restoreInput) {
        inputEditor.tokenizer = originalGrammar;
        inputEditor.value = currentProgram;
      }
      toEnd(inputEditor);
    }
  }

  function clearScreen() {
    outputEditor.selectionStart = outputEditor.selectionEnd = 0;
    outputEditor.value = "";
    return true;
  }

  function flush() {
    if (buffer.length > 0) {
      var lines = buffer.split("\n");
      for (var i = 0; i < pageSize && lines.length > 0; ++i) {
        outputQueue.push(lines.shift());
      }
      if (lines.length > 0) {
        outputQueue.push(" ----- more -----");
      }
      buffer = lines.join("\n");
    }
  }

  function input(callback) {
    inputCallback = callback;
    self.waitingForInput = true;
    flush();
  }

  function stdout(str) {
    buffer += str;
  }

  this.sendInput = function (evt) {
    if (buffer.length > 0) {
      flush();
    } else {
      outputEditor.keyDown(evt);
      var str = outputEditor.value.substring(currentEditIndex);
      inputCallback(str.trim());
      inputCallback = null;
      this.waitingForInput = false;
    }
  };

  this.execute = function () {
    pageSize = 10;
    originalGrammar = inputEditor.tokenizer;
    if (originalGrammar && originalGrammar.interpret) {
      this.running = true;
      var looper,
          next = function next() {
        if (self.running) {
          setTimeout(looper, 1);
        }
      };

      currentProgram = inputEditor.value;
      looper = originalGrammar.interpret(currentProgram, input, stdout, stdout, next, clearScreen, this.loadFile.bind(this), done);
      outputEditor.tokenizer = PlainText$2;
      clearScreen();
      next();
    }
  };

  this.loadFile = function (fileName) {
    return getText(fileName.toLowerCase()).then(function (file) {
      if (isMacOS) {
        file = file.replace("CTRL+SHIFT+SPACE", "CMD+OPT+E");
      }
      inputEditor.value = currentProgram = file;
      return file;
    });
  };

  this.update = function () {
    if (outputQueue.length > 0) {
      outputEditor.value += outputQueue.shift() + "\n";
      toEnd(outputEditor);
      currentEditIndex = outputEditor.selectionStart;
    }
  };
};

function XHR(method, type, url, options) {
  return new Promise(function (resolve, reject) {
    options = options || {};
    options.headers = options.headers || {};
    if (method === "POST") {
      options.headers["Content-Type"] = options.headers["Content-Type"] || type;
    }

    var req = new XMLHttpRequest();
    req.onerror = function (evt) {
      return reject(new Error("Request error: " + evt.message));
    };
    req.onabort = function (evt) {
      return reject(new Error("Request abort: " + evt.message));
    };
    req.onload = function () {
      // The other error events are client-errors. If there was a server error,
      // we'd find out about it during this event. We need to only respond to
      // successful requests, i.e. those with HTTP status code in the 200 or 300
      // range.
      if (req.status < 400) {
        resolve(req.response);
      } else {
        reject(req);
      }
    };

    // The order of these operations is very explicit. You have to call open
    // first. It seems counter intuitive, but think of it more like you're opening
    // an HTTP document to be able to write to it, and then you finish by sending
    // the document. The "open" method does not refer to a network connection.
    req.open(method, url);
    if (type) {
      req.responseType = type;
    }

    req.onprogress = options.progress;

    for (var key in options.headers) {
      req.setRequestHeader(key, options.headers[key]);
    }

    req.withCredentials = !!options.withCredentials;

    if (options.data) {
      req.send(JSON.stringify(options.data));
    } else {
      req.send();
    }
  });
}

function get$2(type, url, options) {
  return XHR("GET", type || "text", url, options);
}

function getBuffer(url, options) {
  return get$2("arraybuffer", url, options);
}

function getText(url, options) {
  return get$2("text", url, options);
}

// polyfill
window.AudioContext = window.AudioContext || window.webkitAudioContext;

var VECTOR = new Vector3();
var UP = new Vector3();
var TEMP = new Matrix4();

var Audio3D = function () {
  createClass(Audio3D, null, [{
    key: "setAudioStream",
    value: function setAudioStream(stream, id) {
      var audioElementCount = document.querySelectorAll("audio").length,
          element = cascadeElement(id || "audioStream" + audioElementCount, "audio", HTMLAudioElement, true);
      setAudioProperties(element);
      element.srcObject = stream;
      return element;
    }
  }, {
    key: "setAudioProperties",
    value: function setAudioProperties(element) {
      element.autoplay = true;
      element.controls = false;
      element.crossOrigin = "anonymous";
      element.muted = true;
      element.setAttribute("muted", "");
    }
  }]);

  function Audio3D() {
    var _this = this;

    classCallCheck(this, Audio3D);

    this.ready = new Promise(function (resolve, reject) {
      try {
        if (Audio3D.isAvailable) {
          (function () {
            var finishSetup = function finishSetup() {
              try {
                _this.sampleRate = _this.context.sampleRate;
                _this.mainVolume = _this.context.createGain();
                _this.start();
                resolve();
              } catch (exp) {
                reject(exp);
              }
            };

            if (!isiOS) {
              _this.context = new AudioContext();
              finishSetup();
            } else {
              (function () {
                var unlock = function unlock() {
                  try {
                    (function () {
                      _this.context = _this.context || new AudioContext();
                      var source = _this.context.createBufferSource();
                      source.buffer = _this.createRawSound([[0]]);
                      source.connect(_this.context.destination);
                      source.sine(0, 0);
                      setTimeout(function () {
                        if (source.playbackState === source.PLAYING_STATE || source.playbackState === source.FINISHED_STATE) {
                          window.removeEventListener("mouseup", unlock);
                          window.removeEventListener("touchend", unlock);
                          window.removeEventListener("keyup", unlock);
                          finishSetup();
                        }
                      }, 0);
                    })();
                  } catch (exp) {
                    reject(exp);
                  }
                };

                window.addEventListener("mouseup", unlock, false);
                window.addEventListener("touchend", unlock, false);
                window.addEventListener("keyup", unlock, false);
              })();
            }
          })();
        }
      } catch (exp) {
        reject(exp);
      }
    }).then(function () {
      return console.log("Audio ready");
    });
  }

  createClass(Audio3D, [{
    key: "setVelocity",
    value: function setVelocity(x, y, z) {
      if (this.context) {
        this.context.listener.setVelocity(x, y, z);
      }
    }
  }, {
    key: "setPlayer",
    value: function setPlayer(obj) {
      if (this.context && this.context.listener) {
        obj.updateMatrixWorld();
        TEMP.copy(obj.matrixWorld);
        var mx = TEMP.elements[12],
            my = TEMP.elements[13],
            mz = TEMP.elements[14];

        this.context.listener.setPosition(mx, my, mz);

        VECTOR.set(0, 0, -1).applyMatrix4(TEMP).normalize();
        UP.set(0, 1, 0).applyMatrix4(TEMP).normalize();

        this.context.listener.setOrientation(VECTOR.x, VECTOR.y, VECTOR.z, UP.x, UP.y, UP.z);
      }
    }
  }, {
    key: "start",
    value: function start() {
      if (this.mainVolume) {
        this.mainVolume.connect(this.context.destination);
      }
    }
  }, {
    key: "stop",
    value: function stop() {
      if (this.mainVolume) {
        this.mainVolume.disconnect();
      }
    }
  }, {
    key: "loadURL",
    value: function loadURL(src) {
      var _this2 = this;

      return this.ready.then(function () {
        console.log("Loading " + src + " from URL");
        getBuffer(src);
      }).then(function (data) {
        return new Promise(function (resolve, reject) {
          return _this2.context.decodeAudioData(data, resolve, reject);
        });
      }).then(function (dat) {
        console.log(src + " loaded");
        return dat;
      }).catch(function (err) {
        console.error("Couldn't load " + src + ". Reason: " + err);
      });
    }
  }, {
    key: "createRawSound",
    value: function createRawSound(pcmData) {
      if (pcmData.length !== 1 && pcmData.length !== 2) {
        throw new Error("Incorrect number of channels. Expected 1 or 2, got " + pcmData.length);
      }

      var frameCount = pcmData[0].length;
      if (pcmData.length > 1 && pcmData[1].length !== frameCount) {
        throw new Error("Second channel is not the same length as the first channel. Expected " + frameCount + ", but was " + pcmData[1].length);
      }

      var buffer = this.context.createBuffer(pcmData.length, frameCount, this.sampleRate || 22050);
      for (var c = 0; c < pcmData.length; ++c) {
        var channel = buffer.getChannelData(c);
        for (var i = 0; i < frameCount; ++i) {
          channel[i] = pcmData[c][i];
        }
      }
      return buffer;
    }
  }, {
    key: "create3DSound",
    value: function create3DSound(x, y, z, snd) {
      snd.panner = this.context.createPanner();
      snd.panner.setPosition(x, y, z);
      snd.panner.connect(this.mainVolume);
      snd.volume.connect(snd.panner);
      return snd;
    }
  }, {
    key: "createFixedSound",
    value: function createFixedSound(snd) {
      snd.volume.connect(this.mainVolume);
      return snd;
    }
  }, {
    key: "loadSource",
    value: function loadSource(sources, loop) {
      var _this3 = this;

      return this.ready.then(function () {
        return new Promise(function (resolve, reject) {
          console.log("Loading " + sources);
          if (!(sources instanceof Array)) {
            sources = [sources];
          }
          var audio = document.createElement("audio");
          audio.autoplay = true;
          audio.preload = "auto";
          audio["webkit-playsinline"] = true;
          audio.playsinline = true;
          audio.loop = loop;
          audio.crossOrigin = "anonymous";
          sources.map(function (src) {
            var source = document.createElement("source");
            source.src = src;
            return source;
          }).forEach(audio.appendChild.bind(audio));
          audio.onerror = reject;
          audio.oncanplay = function () {
            audio.oncanplay = null;
            var snd = {
              volume: _this3.context.createGain(),
              source: _this3.context.createMediaElementSource(audio)
            };
            snd.source.connect(snd.volume);
            resolve(snd);
          };
          audio.play();
          document.body.appendChild(audio);
        });
      }).then(function (dat) {
        console.log(sources + " loaded");
        return dat;
      }).catch(function (err) {
        console.error("Couldn't load " + sources + ". Reason: " + err);
      });
    }
  }, {
    key: "load3DSound",
    value: function load3DSound(src, loop, x, y, z) {
      return this.loadSource(src, loop).then(this.create3DSound.bind(this, x, y, z));
    }
  }, {
    key: "loadFixedSound",
    value: function loadFixedSound(src, loop) {
      return this.loadSource(src, loop).then(this.createFixedSound.bind(this));
    }
  }]);
  return Audio3D;
}();

Audio3D.isAvailable = !!window.AudioContext && !!AudioContext.prototype.createGain;

/**
 * @author tschw
 *
 * Uniforms of a program.
 * Those form a tree structure with a special top-level container for the root,
 * which you get by calling 'new WebGLUniforms( gl, program, renderer )'.
 *
 *
 * Properties of inner nodes including the top-level container:
 *
 * .seq - array of nested uniforms
 * .map - nested uniforms by name
 *
 *
 * Methods of all nodes except the top-level container:
 *
 * .setValue( gl, value, [renderer] )
 *
 * 		uploads a uniform value(s)
 *  	the 'renderer' parameter is needed for sampler uniforms
 *
 *
 * Static methods of the top-level container (renderer factorizations):
 *
 * .upload( gl, seq, values, renderer )
 *
 * 		sets uniforms in 'seq' to 'values[id].value'
 *
 * .seqWithValue( seq, values ) : filteredSeq
 *
 * 		filters 'seq' entries with corresponding entry in values
 *
 *
 * Methods of the top-level container (renderer factorizations):
 *
 * .setValue( gl, name, value )
 *
 * 		sets uniform with  name 'name' to 'value'
 *
 * .set( gl, obj, prop )
 *
 * 		sets uniform from object and property with same name than uniform
 *
 * .setOptional( gl, obj, prop )
 *
 * 		like .set for an optional property of the object
 *
 */

var emptyTexture = new Texture();
var emptyCubeTexture = new CubeTexture();

// --- Base for inner nodes (including the root) ---

function UniformContainer() {

	this.seq = [];
	this.map = {};
}

// --- Utilities ---

// Array Caches (provide typed arrays for temporary by size)

var arrayCacheF32 = [];
var arrayCacheI32 = [];

// Flattening for arrays of vectors and matrices

function flatten(array, nBlocks, blockSize) {

	var firstElem = array[0];

	if (firstElem <= 0 || firstElem > 0) return array;
	// unoptimized: ! isNaN( firstElem )
	// see http://jacksondunstan.com/articles/983

	var n = nBlocks * blockSize,
	    r = arrayCacheF32[n];

	if (r === undefined) {

		r = new Float32Array(n);
		arrayCacheF32[n] = r;
	}

	if (nBlocks !== 0) {

		firstElem.toArray(r, 0);

		for (var i = 1, offset = 0; i !== nBlocks; ++i) {

			offset += blockSize;
			array[i].toArray(r, offset);
		}
	}

	return r;
}

// Texture unit allocation

function allocTexUnits(renderer, n) {

	var r = arrayCacheI32[n];

	if (r === undefined) {

		r = new Int32Array(n);
		arrayCacheI32[n] = r;
	}

	for (var i = 0; i !== n; ++i) {
		r[i] = renderer.allocTextureUnit();
	}return r;
}

// --- Setters ---

// Note: Defining these methods externally, because they come in a bunch
// and this way their names minify.

// Single scalar

function setValue1f(gl, v) {
	gl.uniform1f(this.addr, v);
}
function setValue1i(gl, v) {
	gl.uniform1i(this.addr, v);
}

// Single float vector (from flat array or THREE.VectorN)

function setValue2fv(gl, v) {

	if (v.x === undefined) gl.uniform2fv(this.addr, v);else gl.uniform2f(this.addr, v.x, v.y);
}

function setValue3fv(gl, v) {

	if (v.x !== undefined) gl.uniform3f(this.addr, v.x, v.y, v.z);else if (v.r !== undefined) gl.uniform3f(this.addr, v.r, v.g, v.b);else gl.uniform3fv(this.addr, v);
}

function setValue4fv(gl, v) {

	if (v.x === undefined) gl.uniform4fv(this.addr, v);else gl.uniform4f(this.addr, v.x, v.y, v.z, v.w);
}

// Single matrix (from flat array or MatrixN)

function setValue2fm(gl, v) {

	gl.uniformMatrix2fv(this.addr, false, v.elements || v);
}

function setValue3fm(gl, v) {

	gl.uniformMatrix3fv(this.addr, false, v.elements || v);
}

function setValue4fm(gl, v) {

	gl.uniformMatrix4fv(this.addr, false, v.elements || v);
}

// Single texture (2D / Cube)

function setValueT1(gl, v, renderer) {

	var unit = renderer.allocTextureUnit();
	gl.uniform1i(this.addr, unit);
	renderer.setTexture2D(v || emptyTexture, unit);
}

function setValueT6(gl, v, renderer) {

	var unit = renderer.allocTextureUnit();
	gl.uniform1i(this.addr, unit);
	renderer.setTextureCube(v || emptyCubeTexture, unit);
}

// Integer / Boolean vectors or arrays thereof (always flat arrays)

function setValue2iv(gl, v) {
	gl.uniform2iv(this.addr, v);
}
function setValue3iv(gl, v) {
	gl.uniform3iv(this.addr, v);
}
function setValue4iv(gl, v) {
	gl.uniform4iv(this.addr, v);
}

// Helper to pick the right setter for the singular case

function getSingularSetter(type) {

	switch (type) {

		case 0x1406:
			return setValue1f; // FLOAT
		case 0x8b50:
			return setValue2fv; // _VEC2
		case 0x8b51:
			return setValue3fv; // _VEC3
		case 0x8b52:
			return setValue4fv; // _VEC4

		case 0x8b5a:
			return setValue2fm; // _MAT2
		case 0x8b5b:
			return setValue3fm; // _MAT3
		case 0x8b5c:
			return setValue4fm; // _MAT4

		case 0x8b5e:
			return setValueT1; // SAMPLER_2D
		case 0x8b60:
			return setValueT6; // SAMPLER_CUBE

		case 0x1404:case 0x8b56:
			return setValue1i; // INT, BOOL
		case 0x8b53:case 0x8b57:
			return setValue2iv; // _VEC2
		case 0x8b54:case 0x8b58:
			return setValue3iv; // _VEC3
		case 0x8b55:case 0x8b59:
			return setValue4iv; // _VEC4

	}
}

// Array of scalars

function setValue1fv(gl, v) {
	gl.uniform1fv(this.addr, v);
}
function setValue1iv(gl, v) {
	gl.uniform1iv(this.addr, v);
}

// Array of vectors (flat or from THREE classes)

function setValueV2a(gl, v) {

	gl.uniform2fv(this.addr, flatten(v, this.size, 2));
}

function setValueV3a(gl, v) {

	gl.uniform3fv(this.addr, flatten(v, this.size, 3));
}

function setValueV4a(gl, v) {

	gl.uniform4fv(this.addr, flatten(v, this.size, 4));
}

// Array of matrices (flat or from THREE clases)

function setValueM2a(gl, v) {

	gl.uniformMatrix2fv(this.addr, false, flatten(v, this.size, 4));
}

function setValueM3a(gl, v) {

	gl.uniformMatrix3fv(this.addr, false, flatten(v, this.size, 9));
}

function setValueM4a(gl, v) {

	gl.uniformMatrix4fv(this.addr, false, flatten(v, this.size, 16));
}

// Array of textures (2D / Cube)

function setValueT1a(gl, v, renderer) {

	var n = v.length,
	    units = allocTexUnits(renderer, n);

	gl.uniform1iv(this.addr, units);

	for (var i = 0; i !== n; ++i) {

		renderer.setTexture2D(v[i] || emptyTexture, units[i]);
	}
}

function setValueT6a(gl, v, renderer) {

	var n = v.length,
	    units = allocTexUnits(renderer, n);

	gl.uniform1iv(this.addr, units);

	for (var i = 0; i !== n; ++i) {

		renderer.setTextureCube(v[i] || emptyCubeTexture, units[i]);
	}
}

// Helper to pick the right setter for a pure (bottom-level) array

function getPureArraySetter(type) {

	switch (type) {

		case 0x1406:
			return setValue1fv; // FLOAT
		case 0x8b50:
			return setValueV2a; // _VEC2
		case 0x8b51:
			return setValueV3a; // _VEC3
		case 0x8b52:
			return setValueV4a; // _VEC4

		case 0x8b5a:
			return setValueM2a; // _MAT2
		case 0x8b5b:
			return setValueM3a; // _MAT3
		case 0x8b5c:
			return setValueM4a; // _MAT4

		case 0x8b5e:
			return setValueT1a; // SAMPLER_2D
		case 0x8b60:
			return setValueT6a; // SAMPLER_CUBE

		case 0x1404:case 0x8b56:
			return setValue1iv; // INT, BOOL
		case 0x8b53:case 0x8b57:
			return setValue2iv; // _VEC2
		case 0x8b54:case 0x8b58:
			return setValue3iv; // _VEC3
		case 0x8b55:case 0x8b59:
			return setValue4iv; // _VEC4

	}
}

// --- Uniform Classes ---

function SingleUniform(id, activeInfo, addr) {

	this.id = id;
	this.addr = addr;
	this.setValue = getSingularSetter(activeInfo.type);

	// this.path = activeInfo.name; // DEBUG
}

function PureArrayUniform(id, activeInfo, addr) {

	this.id = id;
	this.addr = addr;
	this.size = activeInfo.size;
	this.setValue = getPureArraySetter(activeInfo.type);

	// this.path = activeInfo.name; // DEBUG
}

function StructuredUniform(id) {

	this.id = id;

	UniformContainer.call(this); // mix-in
}

StructuredUniform.prototype.setValue = function (gl, value) {

	// Note: Don't need an extra 'renderer' parameter, since samplers
	// are not allowed in structured uniforms.

	var seq = this.seq;

	for (var i = 0, n = seq.length; i !== n; ++i) {

		var u = seq[i];
		u.setValue(gl, value[u.id]);
	}
};

// --- Top-level ---

// Parser - builds up the property tree from the path strings

var RePathPart = /([\w\d_]+)(\])?(\[|\.)?/g;

// extracts
// 	- the identifier (member name or array index)
//  - followed by an optional right bracket (found when array index)
//  - followed by an optional left bracket or dot (type of subscript)
//
// Note: These portions can be read in a non-overlapping fashion and
// allow straightforward parsing of the hierarchy that WebGL encodes
// in the uniform names.

function addUniform(container, uniformObject) {

	container.seq.push(uniformObject);
	container.map[uniformObject.id] = uniformObject;
}

function parseUniform(activeInfo, addr, container) {

	var path = activeInfo.name,
	    pathLength = path.length;

	// reset RegExp object, because of the early exit of a previous run
	RePathPart.lastIndex = 0;

	for (;;) {

		var match = RePathPart.exec(path),
		    matchEnd = RePathPart.lastIndex,
		    id = match[1],
		    idIsIndex = match[2] === ']',
		    subscript = match[3];

		if (idIsIndex) id = id | 0; // convert to integer

		if (subscript === undefined || subscript === '[' && matchEnd + 2 === pathLength) {
			// bare name or "pure" bottom-level array "[0]" suffix

			addUniform(container, subscript === undefined ? new SingleUniform(id, activeInfo, addr) : new PureArrayUniform(id, activeInfo, addr));

			break;
		} else {
			// step into inner node / create it in case it doesn't exist

			var map = container.map,
			    next = map[id];

			if (next === undefined) {

				next = new StructuredUniform(id);
				addUniform(container, next);
			}

			container = next;
		}
	}
}

// Root Container

function WebGLUniforms(gl, program, renderer) {

	UniformContainer.call(this);

	this.renderer = renderer;

	var n = gl.getProgramParameter(program, gl.ACTIVE_UNIFORMS);

	for (var i = 0; i !== n; ++i) {

		var info = gl.getActiveUniform(program, i),
		    path = info.name,
		    addr = gl.getUniformLocation(program, path);

		parseUniform(info, addr, this);
	}
}

WebGLUniforms.prototype.setValue = function (gl, name, value) {

	var u = this.map[name];

	if (u !== undefined) u.setValue(gl, value, this.renderer);
};

WebGLUniforms.prototype.set = function (gl, object, name) {

	var u = this.map[name];

	if (u !== undefined) u.setValue(gl, object[name], this.renderer);
};

WebGLUniforms.prototype.setOptional = function (gl, object, name) {

	var v = object[name];

	if (v !== undefined) this.setValue(gl, name, v);
};

// Static interface

WebGLUniforms.upload = function (gl, seq, values, renderer) {

	for (var i = 0, n = seq.length; i !== n; ++i) {

		var u = seq[i],
		    v = values[u.id];

		if (v.needsUpdate !== false) {
			// note: always updating when .needsUpdate is undefined

			u.setValue(gl, v.value, renderer);
		}
	}
};

WebGLUniforms.seqWithValue = function (seq, values) {

	var r = [];

	for (var i = 0, n = seq.length; i !== n; ++i) {

		var u = seq[i];
		if (u.id in values) r.push(u);
	}

	return r;
};

/**
 * @author alteredq / http://alteredqualia.com/
 * @author mrdoob / http://mrdoob.com/
 * @author mikael emtinger / http://gomo.se/
 */

var ShaderLib = {

	basic: {

		uniforms: UniformsUtils.merge([UniformsLib.common, UniformsLib.aomap, UniformsLib.fog]),

		vertexShader: ShaderChunk.meshbasic_vert,
		fragmentShader: ShaderChunk.meshbasic_frag

	},

	lambert: {

		uniforms: UniformsUtils.merge([UniformsLib.common, UniformsLib.aomap, UniformsLib.lightmap, UniformsLib.emissivemap, UniformsLib.fog, UniformsLib.lights, {
			emissive: { value: new Color(0x000000) }
		}]),

		vertexShader: ShaderChunk.meshlambert_vert,
		fragmentShader: ShaderChunk.meshlambert_frag

	},

	phong: {

		uniforms: UniformsUtils.merge([UniformsLib.common, UniformsLib.aomap, UniformsLib.lightmap, UniformsLib.emissivemap, UniformsLib.bumpmap, UniformsLib.normalmap, UniformsLib.displacementmap, UniformsLib.fog, UniformsLib.lights, {
			emissive: { value: new Color(0x000000) },
			specular: { value: new Color(0x111111) },
			shininess: { value: 30 }
		}]),

		vertexShader: ShaderChunk.meshphong_vert,
		fragmentShader: ShaderChunk.meshphong_frag

	},

	standard: {

		uniforms: UniformsUtils.merge([UniformsLib.common, UniformsLib.aomap, UniformsLib.lightmap, UniformsLib.emissivemap, UniformsLib.bumpmap, UniformsLib.normalmap, UniformsLib.displacementmap, UniformsLib.roughnessmap, UniformsLib.metalnessmap, UniformsLib.fog, UniformsLib.lights, {
			emissive: { value: new Color(0x000000) },
			roughness: { value: 0.5 },
			metalness: { value: 0 },
			envMapIntensity: { value: 1 } }]),

		vertexShader: ShaderChunk.meshphysical_vert,
		fragmentShader: ShaderChunk.meshphysical_frag

	},

	points: {

		uniforms: UniformsUtils.merge([UniformsLib.points, UniformsLib.fog]),

		vertexShader: ShaderChunk.points_vert,
		fragmentShader: ShaderChunk.points_frag

	},

	dashed: {

		uniforms: UniformsUtils.merge([UniformsLib.common, UniformsLib.fog, {
			scale: { value: 1 },
			dashSize: { value: 1 },
			totalSize: { value: 2 }
		}]),

		vertexShader: ShaderChunk.linedashed_vert,
		fragmentShader: ShaderChunk.linedashed_frag

	},

	depth: {

		uniforms: UniformsUtils.merge([UniformsLib.common, UniformsLib.displacementmap]),

		vertexShader: ShaderChunk.depth_vert,
		fragmentShader: ShaderChunk.depth_frag

	},

	normal: {

		uniforms: {

			opacity: { value: 1.0 }

		},

		vertexShader: ShaderChunk.normal_vert,
		fragmentShader: ShaderChunk.normal_frag

	},

	/* -------------------------------------------------------------------------
 //	Cube map shader
  ------------------------------------------------------------------------- */

	cube: {

		uniforms: {
			tCube: { value: null },
			tFlip: { value: -1 },
			opacity: { value: 1.0 }
		},

		vertexShader: ShaderChunk.cube_vert,
		fragmentShader: ShaderChunk.cube_frag

	},

	/* -------------------------------------------------------------------------
 //	Cube map shader
  ------------------------------------------------------------------------- */

	equirect: {

		uniforms: {
			tEquirect: { value: null },
			tFlip: { value: -1 }
		},

		vertexShader: ShaderChunk.equirect_vert,
		fragmentShader: ShaderChunk.equirect_frag

	},

	distanceRGBA: {

		uniforms: {

			lightPos: { value: new Vector3() }

		},

		vertexShader: ShaderChunk.distanceRGBA_vert,
		fragmentShader: ShaderChunk.distanceRGBA_frag

	}

};

ShaderLib.physical = {

	uniforms: UniformsUtils.merge([ShaderLib.standard.uniforms, {
		clearCoat: { value: 0 },
		clearCoatRoughness: { value: 0 }
	}]),

	vertexShader: ShaderChunk.meshphysical_vert,
	fragmentShader: ShaderChunk.meshphysical_frag

};

/**
 * @author bhouston / http://clara.io
 */

function Box2(min, max) {

	this.min = min !== undefined ? min : new Vector2(+Infinity, +Infinity);
	this.max = max !== undefined ? max : new Vector2(-Infinity, -Infinity);
}

Box2.prototype = {

	constructor: Box2,

	set: function set(min, max) {

		this.min.copy(min);
		this.max.copy(max);

		return this;
	},

	setFromPoints: function setFromPoints(points) {

		this.makeEmpty();

		for (var i = 0, il = points.length; i < il; i++) {

			this.expandByPoint(points[i]);
		}

		return this;
	},

	setFromCenterAndSize: function () {

		var v1 = new Vector2();

		return function setFromCenterAndSize(center, size) {

			var halfSize = v1.copy(size).multiplyScalar(0.5);
			this.min.copy(center).sub(halfSize);
			this.max.copy(center).add(halfSize);

			return this;
		};
	}(),

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(box) {

		this.min.copy(box.min);
		this.max.copy(box.max);

		return this;
	},

	makeEmpty: function makeEmpty() {

		this.min.x = this.min.y = +Infinity;
		this.max.x = this.max.y = -Infinity;

		return this;
	},

	isEmpty: function isEmpty() {

		// this is a more robust check for empty than ( volume <= 0 ) because volume can get positive with two negative axes

		return this.max.x < this.min.x || this.max.y < this.min.y;
	},

	getCenter: function getCenter(optionalTarget) {

		var result = optionalTarget || new Vector2();
		return this.isEmpty() ? result.set(0, 0) : result.addVectors(this.min, this.max).multiplyScalar(0.5);
	},

	getSize: function getSize(optionalTarget) {

		var result = optionalTarget || new Vector2();
		return this.isEmpty() ? result.set(0, 0) : result.subVectors(this.max, this.min);
	},

	expandByPoint: function expandByPoint(point) {

		this.min.min(point);
		this.max.max(point);

		return this;
	},

	expandByVector: function expandByVector(vector) {

		this.min.sub(vector);
		this.max.add(vector);

		return this;
	},

	expandByScalar: function expandByScalar(scalar) {

		this.min.addScalar(-scalar);
		this.max.addScalar(scalar);

		return this;
	},

	containsPoint: function containsPoint(point) {

		if (point.x < this.min.x || point.x > this.max.x || point.y < this.min.y || point.y > this.max.y) {

			return false;
		}

		return true;
	},

	containsBox: function containsBox(box) {

		if (this.min.x <= box.min.x && box.max.x <= this.max.x && this.min.y <= box.min.y && box.max.y <= this.max.y) {

			return true;
		}

		return false;
	},

	getParameter: function getParameter(point, optionalTarget) {

		// This can potentially have a divide by zero if the box
		// has a size dimension of 0.

		var result = optionalTarget || new Vector2();

		return result.set((point.x - this.min.x) / (this.max.x - this.min.x), (point.y - this.min.y) / (this.max.y - this.min.y));
	},

	intersectsBox: function intersectsBox(box) {

		// using 6 splitting planes to rule out intersections.

		if (box.max.x < this.min.x || box.min.x > this.max.x || box.max.y < this.min.y || box.min.y > this.max.y) {

			return false;
		}

		return true;
	},

	clampPoint: function clampPoint(point, optionalTarget) {

		var result = optionalTarget || new Vector2();
		return result.copy(point).clamp(this.min, this.max);
	},

	distanceToPoint: function () {

		var v1 = new Vector2();

		return function distanceToPoint(point) {

			var clampedPoint = v1.copy(point).clamp(this.min, this.max);
			return clampedPoint.sub(point).length();
		};
	}(),

	intersect: function intersect(box) {

		this.min.max(box.min);
		this.max.min(box.max);

		return this;
	},

	union: function union(box) {

		this.min.min(box.min);
		this.max.max(box.max);

		return this;
	},

	translate: function translate(offset) {

		this.min.add(offset);
		this.max.add(offset);

		return this;
	},

	equals: function equals(box) {

		return box.min.equals(this.min) && box.max.equals(this.max);
	}

};

/**
 * @author mikael emtinger / http://gomo.se/
 * @author alteredq / http://alteredqualia.com/
 */

function LensFlarePlugin(renderer, flares) {

	var gl = renderer.context;
	var state = renderer.state;

	var vertexBuffer, elementBuffer;
	var shader, program, attributes, uniforms;

	var tempTexture, occlusionTexture;

	function init() {

		var vertices = new Float32Array([-1, -1, 0, 0, 1, -1, 1, 0, 1, 1, 1, 1, -1, 1, 0, 1]);

		var faces = new Uint16Array([0, 1, 2, 0, 2, 3]);

		// buffers

		vertexBuffer = gl.createBuffer();
		elementBuffer = gl.createBuffer();

		gl.bindBuffer(gl.ARRAY_BUFFER, vertexBuffer);
		gl.bufferData(gl.ARRAY_BUFFER, vertices, gl.STATIC_DRAW);

		gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, elementBuffer);
		gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, faces, gl.STATIC_DRAW);

		// textures

		tempTexture = gl.createTexture();
		occlusionTexture = gl.createTexture();

		state.bindTexture(gl.TEXTURE_2D, tempTexture);
		gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGB, 16, 16, 0, gl.RGB, gl.UNSIGNED_BYTE, null);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);

		state.bindTexture(gl.TEXTURE_2D, occlusionTexture);
		gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, 16, 16, 0, gl.RGBA, gl.UNSIGNED_BYTE, null);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);

		shader = {

			vertexShader: ["uniform lowp int renderType;", "uniform vec3 screenPosition;", "uniform vec2 scale;", "uniform float rotation;", "uniform sampler2D occlusionMap;", "attribute vec2 position;", "attribute vec2 uv;", "varying vec2 vUV;", "varying float vVisibility;", "void main() {", "vUV = uv;", "vec2 pos = position;", "if ( renderType == 2 ) {", "vec4 visibility = texture2D( occlusionMap, vec2( 0.1, 0.1 ) );", "visibility += texture2D( occlusionMap, vec2( 0.5, 0.1 ) );", "visibility += texture2D( occlusionMap, vec2( 0.9, 0.1 ) );", "visibility += texture2D( occlusionMap, vec2( 0.9, 0.5 ) );", "visibility += texture2D( occlusionMap, vec2( 0.9, 0.9 ) );", "visibility += texture2D( occlusionMap, vec2( 0.5, 0.9 ) );", "visibility += texture2D( occlusionMap, vec2( 0.1, 0.9 ) );", "visibility += texture2D( occlusionMap, vec2( 0.1, 0.5 ) );", "visibility += texture2D( occlusionMap, vec2( 0.5, 0.5 ) );", "vVisibility =        visibility.r / 9.0;", "vVisibility *= 1.0 - visibility.g / 9.0;", "vVisibility *=       visibility.b / 9.0;", "vVisibility *= 1.0 - visibility.a / 9.0;", "pos.x = cos( rotation ) * position.x - sin( rotation ) * position.y;", "pos.y = sin( rotation ) * position.x + cos( rotation ) * position.y;", "}", "gl_Position = vec4( ( pos * scale + screenPosition.xy ).xy, screenPosition.z, 1.0 );", "}"].join("\n"),

			fragmentShader: ["uniform lowp int renderType;", "uniform sampler2D map;", "uniform float opacity;", "uniform vec3 color;", "varying vec2 vUV;", "varying float vVisibility;", "void main() {",

			// pink square

			"if ( renderType == 0 ) {", "gl_FragColor = vec4( 1.0, 0.0, 1.0, 0.0 );",

			// restore

			"} else if ( renderType == 1 ) {", "gl_FragColor = texture2D( map, vUV );",

			// flare

			"} else {", "vec4 texture = texture2D( map, vUV );", "texture.a *= opacity * vVisibility;", "gl_FragColor = texture;", "gl_FragColor.rgb *= color;", "}", "}"].join("\n")

		};

		program = createProgram(shader);

		attributes = {
			vertex: gl.getAttribLocation(program, "position"),
			uv: gl.getAttribLocation(program, "uv")
		};

		uniforms = {
			renderType: gl.getUniformLocation(program, "renderType"),
			map: gl.getUniformLocation(program, "map"),
			occlusionMap: gl.getUniformLocation(program, "occlusionMap"),
			opacity: gl.getUniformLocation(program, "opacity"),
			color: gl.getUniformLocation(program, "color"),
			scale: gl.getUniformLocation(program, "scale"),
			rotation: gl.getUniformLocation(program, "rotation"),
			screenPosition: gl.getUniformLocation(program, "screenPosition")
		};
	}

	/*
  * Render lens flares
  * Method: renders 16x16 0xff00ff-colored points scattered over the light source area,
  *         reads these back and calculates occlusion.
  */

	this.render = function (scene, camera, viewport) {

		if (flares.length === 0) return;

		var tempPosition = new Vector3();

		var invAspect = viewport.w / viewport.z,
		    halfViewportWidth = viewport.z * 0.5,
		    halfViewportHeight = viewport.w * 0.5;

		var size = 16 / viewport.w,
		    scale = new Vector2(size * invAspect, size);

		var screenPosition = new Vector3(1, 1, 0),
		    screenPositionPixels = new Vector2(1, 1);

		var validArea = new Box2();

		validArea.min.set(viewport.x, viewport.y);
		validArea.max.set(viewport.x + (viewport.z - 16), viewport.y + (viewport.w - 16));

		if (program === undefined) {

			init();
		}

		gl.useProgram(program);

		state.initAttributes();
		state.enableAttribute(attributes.vertex);
		state.enableAttribute(attributes.uv);
		state.disableUnusedAttributes();

		// loop through all lens flares to update their occlusion and positions
		// setup gl and common used attribs/uniforms

		gl.uniform1i(uniforms.occlusionMap, 0);
		gl.uniform1i(uniforms.map, 1);

		gl.bindBuffer(gl.ARRAY_BUFFER, vertexBuffer);
		gl.vertexAttribPointer(attributes.vertex, 2, gl.FLOAT, false, 2 * 8, 0);
		gl.vertexAttribPointer(attributes.uv, 2, gl.FLOAT, false, 2 * 8, 8);

		gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, elementBuffer);

		state.disable(gl.CULL_FACE);
		state.setDepthWrite(false);

		for (var i = 0, l = flares.length; i < l; i++) {

			size = 16 / viewport.w;
			scale.set(size * invAspect, size);

			// calc object screen position

			var flare = flares[i];

			tempPosition.set(flare.matrixWorld.elements[12], flare.matrixWorld.elements[13], flare.matrixWorld.elements[14]);

			tempPosition.applyMatrix4(camera.matrixWorldInverse);
			tempPosition.applyProjection(camera.projectionMatrix);

			// setup arrays for gl programs

			screenPosition.copy(tempPosition);

			// horizontal and vertical coordinate of the lower left corner of the pixels to copy

			screenPositionPixels.x = viewport.x + screenPosition.x * halfViewportWidth + halfViewportWidth - 8;
			screenPositionPixels.y = viewport.y + screenPosition.y * halfViewportHeight + halfViewportHeight - 8;

			// screen cull

			if (validArea.containsPoint(screenPositionPixels) === true) {

				// save current RGB to temp texture

				state.activeTexture(gl.TEXTURE0);
				state.bindTexture(gl.TEXTURE_2D, null);
				state.activeTexture(gl.TEXTURE1);
				state.bindTexture(gl.TEXTURE_2D, tempTexture);
				gl.copyTexImage2D(gl.TEXTURE_2D, 0, gl.RGB, screenPositionPixels.x, screenPositionPixels.y, 16, 16, 0);

				// render pink quad

				gl.uniform1i(uniforms.renderType, 0);
				gl.uniform2f(uniforms.scale, scale.x, scale.y);
				gl.uniform3f(uniforms.screenPosition, screenPosition.x, screenPosition.y, screenPosition.z);

				state.disable(gl.BLEND);
				state.enable(gl.DEPTH_TEST);

				gl.drawElements(gl.TRIANGLES, 6, gl.UNSIGNED_SHORT, 0);

				// copy result to occlusionMap

				state.activeTexture(gl.TEXTURE0);
				state.bindTexture(gl.TEXTURE_2D, occlusionTexture);
				gl.copyTexImage2D(gl.TEXTURE_2D, 0, gl.RGBA, screenPositionPixels.x, screenPositionPixels.y, 16, 16, 0);

				// restore graphics

				gl.uniform1i(uniforms.renderType, 1);
				state.disable(gl.DEPTH_TEST);

				state.activeTexture(gl.TEXTURE1);
				state.bindTexture(gl.TEXTURE_2D, tempTexture);
				gl.drawElements(gl.TRIANGLES, 6, gl.UNSIGNED_SHORT, 0);

				// update object positions

				flare.positionScreen.copy(screenPosition);

				if (flare.customUpdateCallback) {

					flare.customUpdateCallback(flare);
				} else {

					flare.updateLensFlares();
				}

				// render flares

				gl.uniform1i(uniforms.renderType, 2);
				state.enable(gl.BLEND);

				for (var j = 0, jl = flare.lensFlares.length; j < jl; j++) {

					var sprite = flare.lensFlares[j];

					if (sprite.opacity > 0.001 && sprite.scale > 0.001) {

						screenPosition.x = sprite.x;
						screenPosition.y = sprite.y;
						screenPosition.z = sprite.z;

						size = sprite.size * sprite.scale / viewport.w;

						scale.x = size * invAspect;
						scale.y = size;

						gl.uniform3f(uniforms.screenPosition, screenPosition.x, screenPosition.y, screenPosition.z);
						gl.uniform2f(uniforms.scale, scale.x, scale.y);
						gl.uniform1f(uniforms.rotation, sprite.rotation);

						gl.uniform1f(uniforms.opacity, sprite.opacity);
						gl.uniform3f(uniforms.color, sprite.color.r, sprite.color.g, sprite.color.b);

						state.setBlending(sprite.blending, sprite.blendEquation, sprite.blendSrc, sprite.blendDst);
						renderer.setTexture2D(sprite.texture, 1);

						gl.drawElements(gl.TRIANGLES, 6, gl.UNSIGNED_SHORT, 0);
					}
				}
			}
		}

		// restore gl

		state.enable(gl.CULL_FACE);
		state.enable(gl.DEPTH_TEST);
		state.setDepthWrite(true);

		renderer.resetGLState();
	};

	function createProgram(shader) {

		var program = gl.createProgram();

		var fragmentShader = gl.createShader(gl.FRAGMENT_SHADER);
		var vertexShader = gl.createShader(gl.VERTEX_SHADER);

		var prefix = "precision " + renderer.getPrecision() + " float;\n";

		gl.shaderSource(fragmentShader, prefix + shader.fragmentShader);
		gl.shaderSource(vertexShader, prefix + shader.vertexShader);

		gl.compileShader(fragmentShader);
		gl.compileShader(vertexShader);

		gl.attachShader(program, fragmentShader);
		gl.attachShader(program, vertexShader);

		gl.linkProgram(program);

		return program;
	}
}

/**
 * @author mikael emtinger / http://gomo.se/
 * @author alteredq / http://alteredqualia.com/
 */

function SpritePlugin(renderer, sprites) {

	var gl = renderer.context;
	var state = renderer.state;

	var vertexBuffer, elementBuffer;
	var program, attributes, uniforms;

	var texture;

	// decompose matrixWorld

	var spritePosition = new Vector3();
	var spriteRotation = new Quaternion();
	var spriteScale = new Vector3();

	function init() {

		var vertices = new Float32Array([-0.5, -0.5, 0, 0, 0.5, -0.5, 1, 0, 0.5, 0.5, 1, 1, -0.5, 0.5, 0, 1]);

		var faces = new Uint16Array([0, 1, 2, 0, 2, 3]);

		vertexBuffer = gl.createBuffer();
		elementBuffer = gl.createBuffer();

		gl.bindBuffer(gl.ARRAY_BUFFER, vertexBuffer);
		gl.bufferData(gl.ARRAY_BUFFER, vertices, gl.STATIC_DRAW);

		gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, elementBuffer);
		gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, faces, gl.STATIC_DRAW);

		program = createProgram();

		attributes = {
			position: gl.getAttribLocation(program, 'position'),
			uv: gl.getAttribLocation(program, 'uv')
		};

		uniforms = {
			uvOffset: gl.getUniformLocation(program, 'uvOffset'),
			uvScale: gl.getUniformLocation(program, 'uvScale'),

			rotation: gl.getUniformLocation(program, 'rotation'),
			scale: gl.getUniformLocation(program, 'scale'),

			color: gl.getUniformLocation(program, 'color'),
			map: gl.getUniformLocation(program, 'map'),
			opacity: gl.getUniformLocation(program, 'opacity'),

			modelViewMatrix: gl.getUniformLocation(program, 'modelViewMatrix'),
			projectionMatrix: gl.getUniformLocation(program, 'projectionMatrix'),

			fogType: gl.getUniformLocation(program, 'fogType'),
			fogDensity: gl.getUniformLocation(program, 'fogDensity'),
			fogNear: gl.getUniformLocation(program, 'fogNear'),
			fogFar: gl.getUniformLocation(program, 'fogFar'),
			fogColor: gl.getUniformLocation(program, 'fogColor'),

			alphaTest: gl.getUniformLocation(program, 'alphaTest')
		};

		var canvas = document.createElementNS('http://www.w3.org/1999/xhtml', 'canvas');
		canvas.width = 8;
		canvas.height = 8;

		var context = canvas.getContext('2d');
		context.fillStyle = 'white';
		context.fillRect(0, 0, 8, 8);

		texture = new Texture(canvas);
		texture.needsUpdate = true;
	}

	this.render = function (scene, camera) {

		if (sprites.length === 0) return;

		// setup gl

		if (program === undefined) {

			init();
		}

		gl.useProgram(program);

		state.initAttributes();
		state.enableAttribute(attributes.position);
		state.enableAttribute(attributes.uv);
		state.disableUnusedAttributes();

		state.disable(gl.CULL_FACE);
		state.enable(gl.BLEND);

		gl.bindBuffer(gl.ARRAY_BUFFER, vertexBuffer);
		gl.vertexAttribPointer(attributes.position, 2, gl.FLOAT, false, 2 * 8, 0);
		gl.vertexAttribPointer(attributes.uv, 2, gl.FLOAT, false, 2 * 8, 8);

		gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, elementBuffer);

		gl.uniformMatrix4fv(uniforms.projectionMatrix, false, camera.projectionMatrix.elements);

		state.activeTexture(gl.TEXTURE0);
		gl.uniform1i(uniforms.map, 0);

		var oldFogType = 0;
		var sceneFogType = 0;
		var fog = scene.fog;

		if (fog) {

			gl.uniform3f(uniforms.fogColor, fog.color.r, fog.color.g, fog.color.b);

			if (fog && fog.isFog) {

				gl.uniform1f(uniforms.fogNear, fog.near);
				gl.uniform1f(uniforms.fogFar, fog.far);

				gl.uniform1i(uniforms.fogType, 1);
				oldFogType = 1;
				sceneFogType = 1;
			} else if (fog && fog.isFogExp2) {

				gl.uniform1f(uniforms.fogDensity, fog.density);

				gl.uniform1i(uniforms.fogType, 2);
				oldFogType = 2;
				sceneFogType = 2;
			}
		} else {

			gl.uniform1i(uniforms.fogType, 0);
			oldFogType = 0;
			sceneFogType = 0;
		}

		// update positions and sort

		for (var i = 0, l = sprites.length; i < l; i++) {

			var sprite = sprites[i];

			sprite.modelViewMatrix.multiplyMatrices(camera.matrixWorldInverse, sprite.matrixWorld);
			sprite.z = -sprite.modelViewMatrix.elements[14];
		}

		sprites.sort(painterSortStable);

		// render all sprites

		var scale = [];

		for (var i = 0, l = sprites.length; i < l; i++) {

			var sprite = sprites[i];
			var material = sprite.material;

			if (material.visible === false) continue;

			gl.uniform1f(uniforms.alphaTest, material.alphaTest);
			gl.uniformMatrix4fv(uniforms.modelViewMatrix, false, sprite.modelViewMatrix.elements);

			sprite.matrixWorld.decompose(spritePosition, spriteRotation, spriteScale);

			scale[0] = spriteScale.x;
			scale[1] = spriteScale.y;

			var fogType = 0;

			if (scene.fog && material.fog) {

				fogType = sceneFogType;
			}

			if (oldFogType !== fogType) {

				gl.uniform1i(uniforms.fogType, fogType);
				oldFogType = fogType;
			}

			if (material.map !== null) {

				gl.uniform2f(uniforms.uvOffset, material.map.offset.x, material.map.offset.y);
				gl.uniform2f(uniforms.uvScale, material.map.repeat.x, material.map.repeat.y);
			} else {

				gl.uniform2f(uniforms.uvOffset, 0, 0);
				gl.uniform2f(uniforms.uvScale, 1, 1);
			}

			gl.uniform1f(uniforms.opacity, material.opacity);
			gl.uniform3f(uniforms.color, material.color.r, material.color.g, material.color.b);

			gl.uniform1f(uniforms.rotation, material.rotation);
			gl.uniform2fv(uniforms.scale, scale);

			state.setBlending(material.blending, material.blendEquation, material.blendSrc, material.blendDst);
			state.setDepthTest(material.depthTest);
			state.setDepthWrite(material.depthWrite);

			if (material.map) {

				renderer.setTexture2D(material.map, 0);
			} else {

				renderer.setTexture2D(texture, 0);
			}

			gl.drawElements(gl.TRIANGLES, 6, gl.UNSIGNED_SHORT, 0);
		}

		// restore gl

		state.enable(gl.CULL_FACE);

		renderer.resetGLState();
	};

	function createProgram() {

		var program = gl.createProgram();

		var vertexShader = gl.createShader(gl.VERTEX_SHADER);
		var fragmentShader = gl.createShader(gl.FRAGMENT_SHADER);

		gl.shaderSource(vertexShader, ['precision ' + renderer.getPrecision() + ' float;', 'uniform mat4 modelViewMatrix;', 'uniform mat4 projectionMatrix;', 'uniform float rotation;', 'uniform vec2 scale;', 'uniform vec2 uvOffset;', 'uniform vec2 uvScale;', 'attribute vec2 position;', 'attribute vec2 uv;', 'varying vec2 vUV;', 'void main() {', 'vUV = uvOffset + uv * uvScale;', 'vec2 alignedPosition = position * scale;', 'vec2 rotatedPosition;', 'rotatedPosition.x = cos( rotation ) * alignedPosition.x - sin( rotation ) * alignedPosition.y;', 'rotatedPosition.y = sin( rotation ) * alignedPosition.x + cos( rotation ) * alignedPosition.y;', 'vec4 finalPosition;', 'finalPosition = modelViewMatrix * vec4( 0.0, 0.0, 0.0, 1.0 );', 'finalPosition.xy += rotatedPosition;', 'finalPosition = projectionMatrix * finalPosition;', 'gl_Position = finalPosition;', '}'].join('\n'));

		gl.shaderSource(fragmentShader, ['precision ' + renderer.getPrecision() + ' float;', 'uniform vec3 color;', 'uniform sampler2D map;', 'uniform float opacity;', 'uniform int fogType;', 'uniform vec3 fogColor;', 'uniform float fogDensity;', 'uniform float fogNear;', 'uniform float fogFar;', 'uniform float alphaTest;', 'varying vec2 vUV;', 'void main() {', 'vec4 texture = texture2D( map, vUV );', 'if ( texture.a < alphaTest ) discard;', 'gl_FragColor = vec4( color * texture.xyz, texture.a * opacity );', 'if ( fogType > 0 ) {', 'float depth = gl_FragCoord.z / gl_FragCoord.w;', 'float fogFactor = 0.0;', 'if ( fogType == 1 ) {', 'fogFactor = smoothstep( fogNear, fogFar, depth );', '} else {', 'const float LOG2 = 1.442695;', 'fogFactor = exp2( - fogDensity * fogDensity * depth * depth * LOG2 );', 'fogFactor = 1.0 - clamp( fogFactor, 0.0, 1.0 );', '}', 'gl_FragColor = mix( gl_FragColor, vec4( fogColor, gl_FragColor.w ), fogFactor );', '}', '}'].join('\n'));

		gl.compileShader(vertexShader);
		gl.compileShader(fragmentShader);

		gl.attachShader(program, vertexShader);
		gl.attachShader(program, fragmentShader);

		gl.linkProgram(program);

		return program;
	}

	function painterSortStable(a, b) {

		if (a.renderOrder !== b.renderOrder) {

			return a.renderOrder - b.renderOrder;
		} else if (a.z !== b.z) {

			return b.z - a.z;
		} else {

			return b.id - a.id;
		}
	}
}

/**
 * @author szimek / https://github.com/szimek/
 * @author alteredq / http://alteredqualia.com/
 * @author Marius Kintel / https://github.com/kintel
 */

/*
 In options, we can specify:
 * Texture parameters for an auto-generated target texture
 * depthBuffer/stencilBuffer: Booleans to indicate if we should generate these buffers
*/
function WebGLRenderTarget(width, height, options) {

	this.uuid = _Math.generateUUID();

	this.width = width;
	this.height = height;

	this.scissor = new Vector4(0, 0, width, height);
	this.scissorTest = false;

	this.viewport = new Vector4(0, 0, width, height);

	options = options || {};

	if (options.minFilter === undefined) options.minFilter = LinearFilter;

	this.texture = new Texture(undefined, undefined, options.wrapS, options.wrapT, options.magFilter, options.minFilter, options.format, options.type, options.anisotropy, options.encoding);

	this.depthBuffer = options.depthBuffer !== undefined ? options.depthBuffer : true;
	this.stencilBuffer = options.stencilBuffer !== undefined ? options.stencilBuffer : true;
	this.depthTexture = options.depthTexture !== undefined ? options.depthTexture : null;
}

Object.assign(WebGLRenderTarget.prototype, EventDispatcher.prototype, {

	isWebGLRenderTarget: true,

	setSize: function setSize(width, height) {

		if (this.width !== width || this.height !== height) {

			this.width = width;
			this.height = height;

			this.dispose();
		}

		this.viewport.set(0, 0, width, height);
		this.scissor.set(0, 0, width, height);
	},

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(source) {

		this.width = source.width;
		this.height = source.height;

		this.viewport.copy(source.viewport);

		this.texture = source.texture.clone();

		this.depthBuffer = source.depthBuffer;
		this.stencilBuffer = source.stencilBuffer;
		this.depthTexture = source.depthTexture;

		return this;
	},

	dispose: function dispose() {

		this.dispatchEvent({ type: 'dispose' });
	}

});

/**
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 * @author bhouston / http://clara.io
 */

function Frustum(p0, p1, p2, p3, p4, p5) {

	this.planes = [p0 !== undefined ? p0 : new Plane(), p1 !== undefined ? p1 : new Plane(), p2 !== undefined ? p2 : new Plane(), p3 !== undefined ? p3 : new Plane(), p4 !== undefined ? p4 : new Plane(), p5 !== undefined ? p5 : new Plane()];
}

Frustum.prototype = {

	constructor: Frustum,

	set: function set(p0, p1, p2, p3, p4, p5) {

		var planes = this.planes;

		planes[0].copy(p0);
		planes[1].copy(p1);
		planes[2].copy(p2);
		planes[3].copy(p3);
		planes[4].copy(p4);
		planes[5].copy(p5);

		return this;
	},

	clone: function clone() {

		return new this.constructor().copy(this);
	},

	copy: function copy(frustum) {

		var planes = this.planes;

		for (var i = 0; i < 6; i++) {

			planes[i].copy(frustum.planes[i]);
		}

		return this;
	},

	setFromMatrix: function setFromMatrix(m) {

		var planes = this.planes;
		var me = m.elements;
		var me0 = me[0],
		    me1 = me[1],
		    me2 = me[2],
		    me3 = me[3];
		var me4 = me[4],
		    me5 = me[5],
		    me6 = me[6],
		    me7 = me[7];
		var me8 = me[8],
		    me9 = me[9],
		    me10 = me[10],
		    me11 = me[11];
		var me12 = me[12],
		    me13 = me[13],
		    me14 = me[14],
		    me15 = me[15];

		planes[0].setComponents(me3 - me0, me7 - me4, me11 - me8, me15 - me12).normalize();
		planes[1].setComponents(me3 + me0, me7 + me4, me11 + me8, me15 + me12).normalize();
		planes[2].setComponents(me3 + me1, me7 + me5, me11 + me9, me15 + me13).normalize();
		planes[3].setComponents(me3 - me1, me7 - me5, me11 - me9, me15 - me13).normalize();
		planes[4].setComponents(me3 - me2, me7 - me6, me11 - me10, me15 - me14).normalize();
		planes[5].setComponents(me3 + me2, me7 + me6, me11 + me10, me15 + me14).normalize();

		return this;
	},

	intersectsObject: function () {

		var sphere = new Sphere();

		return function intersectsObject(object) {

			var geometry = object.geometry;

			if (geometry.boundingSphere === null) geometry.computeBoundingSphere();

			sphere.copy(geometry.boundingSphere).applyMatrix4(object.matrixWorld);

			return this.intersectsSphere(sphere);
		};
	}(),

	intersectsSprite: function () {

		var sphere = new Sphere();

		return function intersectsSprite(sprite) {

			sphere.center.set(0, 0, 0);
			sphere.radius = 0.7071067811865476;
			sphere.applyMatrix4(sprite.matrixWorld);

			return this.intersectsSphere(sphere);
		};
	}(),

	intersectsSphere: function intersectsSphere(sphere) {

		var planes = this.planes;
		var center = sphere.center;
		var negRadius = -sphere.radius;

		for (var i = 0; i < 6; i++) {

			var distance = planes[i].distanceToPoint(center);

			if (distance < negRadius) {

				return false;
			}
		}

		return true;
	},

	intersectsBox: function () {

		var p1 = new Vector3(),
		    p2 = new Vector3();

		return function intersectsBox(box) {

			var planes = this.planes;

			for (var i = 0; i < 6; i++) {

				var plane = planes[i];

				p1.x = plane.normal.x > 0 ? box.min.x : box.max.x;
				p2.x = plane.normal.x > 0 ? box.max.x : box.min.x;
				p1.y = plane.normal.y > 0 ? box.min.y : box.max.y;
				p2.y = plane.normal.y > 0 ? box.max.y : box.min.y;
				p1.z = plane.normal.z > 0 ? box.min.z : box.max.z;
				p2.z = plane.normal.z > 0 ? box.max.z : box.min.z;

				var d1 = plane.distanceToPoint(p1);
				var d2 = plane.distanceToPoint(p2);

				// if both outside plane, no intersection

				if (d1 < 0 && d2 < 0) {

					return false;
				}
			}

			return true;
		};
	}(),

	containsPoint: function containsPoint(point) {

		var planes = this.planes;

		for (var i = 0; i < 6; i++) {

			if (planes[i].distanceToPoint(point) < 0) {

				return false;
			}
		}

		return true;
	}

};

/**
 * @author alteredq / http://alteredqualia.com/
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLShadowMap(_renderer, _lights, _objects, capabilities) {

	var _gl = _renderer.context,
	    _state = _renderer.state,
	    _frustum = new Frustum(),
	    _projScreenMatrix = new Matrix4(),
	    _lightShadows = _lights.shadows,
	    _shadowMapSize = new Vector2(),
	    _maxShadowMapSize = new Vector2(capabilities.maxTextureSize, capabilities.maxTextureSize),
	    _lookTarget = new Vector3(),
	    _lightPositionWorld = new Vector3(),
	    _renderList = [],
	    _MorphingFlag = 1,
	    _SkinningFlag = 2,
	    _NumberOfMaterialVariants = (_MorphingFlag | _SkinningFlag) + 1,
	    _depthMaterials = new Array(_NumberOfMaterialVariants),
	    _distanceMaterials = new Array(_NumberOfMaterialVariants),
	    _materialCache = {};

	var cubeDirections = [new Vector3(1, 0, 0), new Vector3(-1, 0, 0), new Vector3(0, 0, 1), new Vector3(0, 0, -1), new Vector3(0, 1, 0), new Vector3(0, -1, 0)];

	var cubeUps = [new Vector3(0, 1, 0), new Vector3(0, 1, 0), new Vector3(0, 1, 0), new Vector3(0, 1, 0), new Vector3(0, 0, 1), new Vector3(0, 0, -1)];

	var cube2DViewPorts = [new Vector4(), new Vector4(), new Vector4(), new Vector4(), new Vector4(), new Vector4()];

	// init

	var depthMaterialTemplate = new MeshDepthMaterial();
	depthMaterialTemplate.depthPacking = RGBADepthPacking;
	depthMaterialTemplate.clipping = true;

	var distanceShader = ShaderLib["distanceRGBA"];
	var distanceUniforms = UniformsUtils.clone(distanceShader.uniforms);

	for (var i = 0; i !== _NumberOfMaterialVariants; ++i) {

		var useMorphing = (i & _MorphingFlag) !== 0;
		var useSkinning = (i & _SkinningFlag) !== 0;

		var depthMaterial = depthMaterialTemplate.clone();
		depthMaterial.morphTargets = useMorphing;
		depthMaterial.skinning = useSkinning;

		_depthMaterials[i] = depthMaterial;

		var distanceMaterial = new ShaderMaterial({
			defines: {
				'USE_SHADOWMAP': ''
			},
			uniforms: distanceUniforms,
			vertexShader: distanceShader.vertexShader,
			fragmentShader: distanceShader.fragmentShader,
			morphTargets: useMorphing,
			skinning: useSkinning,
			clipping: true
		});

		_distanceMaterials[i] = distanceMaterial;
	}

	//

	var scope = this;

	this.enabled = false;

	this.autoUpdate = true;
	this.needsUpdate = false;

	this.type = PCFShadowMap;

	this.renderReverseSided = true;
	this.renderSingleSided = true;

	this.render = function (scene, camera) {

		if (scope.enabled === false) return;
		if (scope.autoUpdate === false && scope.needsUpdate === false) return;

		if (_lightShadows.length === 0) return;

		// Set GL state for depth map.
		_state.clearColor(1, 1, 1, 1);
		_state.disable(_gl.BLEND);
		_state.setDepthTest(true);
		_state.setScissorTest(false);

		// render depth map

		var faceCount, isPointLight;

		for (var i = 0, il = _lightShadows.length; i < il; i++) {

			var light = _lightShadows[i];
			var shadow = light.shadow;

			if (shadow === undefined) {

				console.warn('THREE.WebGLShadowMap:', light, 'has no shadow.');
				continue;
			}

			var shadowCamera = shadow.camera;

			_shadowMapSize.copy(shadow.mapSize);
			_shadowMapSize.min(_maxShadowMapSize);

			if (light && light.isPointLight) {

				faceCount = 6;
				isPointLight = true;

				var vpWidth = _shadowMapSize.x;
				var vpHeight = _shadowMapSize.y;

				// These viewports map a cube-map onto a 2D texture with the
				// following orientation:
				//
				//  xzXZ
				//   y Y
				//
				// X - Positive x direction
				// x - Negative x direction
				// Y - Positive y direction
				// y - Negative y direction
				// Z - Positive z direction
				// z - Negative z direction

				// positive X
				cube2DViewPorts[0].set(vpWidth * 2, vpHeight, vpWidth, vpHeight);
				// negative X
				cube2DViewPorts[1].set(0, vpHeight, vpWidth, vpHeight);
				// positive Z
				cube2DViewPorts[2].set(vpWidth * 3, vpHeight, vpWidth, vpHeight);
				// negative Z
				cube2DViewPorts[3].set(vpWidth, vpHeight, vpWidth, vpHeight);
				// positive Y
				cube2DViewPorts[4].set(vpWidth * 3, 0, vpWidth, vpHeight);
				// negative Y
				cube2DViewPorts[5].set(vpWidth, 0, vpWidth, vpHeight);

				_shadowMapSize.x *= 4.0;
				_shadowMapSize.y *= 2.0;
			} else {

				faceCount = 1;
				isPointLight = false;
			}

			if (shadow.map === null) {

				var pars = { minFilter: NearestFilter, magFilter: NearestFilter, format: RGBAFormat };

				shadow.map = new WebGLRenderTarget(_shadowMapSize.x, _shadowMapSize.y, pars);

				shadowCamera.updateProjectionMatrix();
			}

			if (shadow && shadow.isSpotLightShadow) {

				shadow.update(light);
			}

			var shadowMap = shadow.map;
			var shadowMatrix = shadow.matrix;

			_lightPositionWorld.setFromMatrixPosition(light.matrixWorld);
			shadowCamera.position.copy(_lightPositionWorld);

			_renderer.setRenderTarget(shadowMap);
			_renderer.clear();

			// render shadow map for each cube face (if omni-directional) or
			// run a single pass if not

			for (var face = 0; face < faceCount; face++) {

				if (isPointLight) {

					_lookTarget.copy(shadowCamera.position);
					_lookTarget.add(cubeDirections[face]);
					shadowCamera.up.copy(cubeUps[face]);
					shadowCamera.lookAt(_lookTarget);

					var vpDimensions = cube2DViewPorts[face];
					_state.viewport(vpDimensions);
				} else {

					_lookTarget.setFromMatrixPosition(light.target.matrixWorld);
					shadowCamera.lookAt(_lookTarget);
				}

				shadowCamera.updateMatrixWorld();
				shadowCamera.matrixWorldInverse.getInverse(shadowCamera.matrixWorld);

				// compute shadow matrix

				shadowMatrix.set(0.5, 0.0, 0.0, 0.5, 0.0, 0.5, 0.0, 0.5, 0.0, 0.0, 0.5, 0.5, 0.0, 0.0, 0.0, 1.0);

				shadowMatrix.multiply(shadowCamera.projectionMatrix);
				shadowMatrix.multiply(shadowCamera.matrixWorldInverse);

				// update camera matrices and frustum

				_projScreenMatrix.multiplyMatrices(shadowCamera.projectionMatrix, shadowCamera.matrixWorldInverse);
				_frustum.setFromMatrix(_projScreenMatrix);

				// set object matrices & frustum culling

				_renderList.length = 0;

				projectObject(scene, camera, shadowCamera);

				// render shadow map
				// render regular objects

				for (var j = 0, jl = _renderList.length; j < jl; j++) {

					var object = _renderList[j];
					var geometry = _objects.update(object);
					var material = object.material;

					if (material && material.isMultiMaterial) {

						var groups = geometry.groups;
						var materials = material.materials;

						for (var k = 0, kl = groups.length; k < kl; k++) {

							var group = groups[k];
							var groupMaterial = materials[group.materialIndex];

							if (groupMaterial.visible === true) {

								var depthMaterial = getDepthMaterial(object, groupMaterial, isPointLight, _lightPositionWorld);
								_renderer.renderBufferDirect(shadowCamera, null, geometry, depthMaterial, object, group);
							}
						}
					} else {

						var depthMaterial = getDepthMaterial(object, material, isPointLight, _lightPositionWorld);
						_renderer.renderBufferDirect(shadowCamera, null, geometry, depthMaterial, object, null);
					}
				}
			}
		}

		// Restore GL state.
		var clearColor = _renderer.getClearColor(),
		    clearAlpha = _renderer.getClearAlpha();
		_renderer.setClearColor(clearColor, clearAlpha);

		scope.needsUpdate = false;
	};

	function getDepthMaterial(object, material, isPointLight, lightPositionWorld) {

		var geometry = object.geometry;

		var result = null;

		var materialVariants = _depthMaterials;
		var customMaterial = object.customDepthMaterial;

		if (isPointLight) {

			materialVariants = _distanceMaterials;
			customMaterial = object.customDistanceMaterial;
		}

		if (!customMaterial) {

			var useMorphing = false;

			if (material.morphTargets) {

				if (geometry && geometry.isBufferGeometry) {

					useMorphing = geometry.morphAttributes && geometry.morphAttributes.position && geometry.morphAttributes.position.length > 0;
				} else if (geometry && geometry.isGeometry) {

					useMorphing = geometry.morphTargets && geometry.morphTargets.length > 0;
				}
			}

			var useSkinning = object.isSkinnedMesh && material.skinning;

			var variantIndex = 0;

			if (useMorphing) variantIndex |= _MorphingFlag;
			if (useSkinning) variantIndex |= _SkinningFlag;

			result = materialVariants[variantIndex];
		} else {

			result = customMaterial;
		}

		if (_renderer.localClippingEnabled && material.clipShadows === true && material.clippingPlanes.length !== 0) {

			// in this case we need a unique material instance reflecting the
			// appropriate state

			var keyA = result.uuid,
			    keyB = material.uuid;

			var materialsForVariant = _materialCache[keyA];

			if (materialsForVariant === undefined) {

				materialsForVariant = {};
				_materialCache[keyA] = materialsForVariant;
			}

			var cachedMaterial = materialsForVariant[keyB];

			if (cachedMaterial === undefined) {

				cachedMaterial = result.clone();
				materialsForVariant[keyB] = cachedMaterial;
			}

			result = cachedMaterial;
		}

		result.visible = material.visible;
		result.wireframe = material.wireframe;

		var side = material.side;

		if (scope.renderSingleSided && side == DoubleSide) {

			side = FrontSide;
		}

		if (scope.renderReverseSided) {

			if (side === FrontSide) side = BackSide;else if (side === BackSide) side = FrontSide;
		}

		result.side = side;

		result.clipShadows = material.clipShadows;
		result.clippingPlanes = material.clippingPlanes;

		result.wireframeLinewidth = material.wireframeLinewidth;
		result.linewidth = material.linewidth;

		if (isPointLight && result.uniforms.lightPos !== undefined) {

			result.uniforms.lightPos.value.copy(lightPositionWorld);
		}

		return result;
	}

	function projectObject(object, camera, shadowCamera) {

		if (object.visible === false) return;

		var visible = (object.layers.mask & camera.layers.mask) !== 0;

		if (visible && (object.isMesh || object.isLine || object.isPoints)) {

			if (object.castShadow && (object.frustumCulled === false || _frustum.intersectsObject(object) === true)) {

				var material = object.material;

				if (material.visible === true) {

					object.modelViewMatrix.multiplyMatrices(shadowCamera.matrixWorldInverse, object.matrixWorld);
					_renderList.push(object);
				}
			}
		}

		var children = object.children;

		for (var i = 0, l = children.length; i < l; i++) {

			projectObject(children[i], camera, shadowCamera);
		}
	}
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLIndexedBufferRenderer(gl, extensions, infoRender) {

	var mode;

	function setMode(value) {

		mode = value;
	}

	var type, size;

	function setIndex(index) {

		if (index.array instanceof Uint32Array && extensions.get('OES_element_index_uint')) {

			type = gl.UNSIGNED_INT;
			size = 4;
		} else {

			type = gl.UNSIGNED_SHORT;
			size = 2;
		}
	}

	function render(start, count) {

		gl.drawElements(mode, count, type, start * size);

		infoRender.calls++;
		infoRender.vertices += count;

		if (mode === gl.TRIANGLES) infoRender.faces += count / 3;
	}

	function renderInstances(geometry, start, count) {

		var extension = extensions.get('ANGLE_instanced_arrays');

		if (extension === null) {

			console.error('THREE.WebGLBufferRenderer: using THREE.InstancedBufferGeometry but hardware does not support extension ANGLE_instanced_arrays.');
			return;
		}

		extension.drawElementsInstancedANGLE(mode, count, type, start * size, geometry.maxInstancedCount);

		infoRender.calls++;
		infoRender.vertices += count * geometry.maxInstancedCount;

		if (mode === gl.TRIANGLES) infoRender.faces += geometry.maxInstancedCount * count / 3;
	}

	return {

		setMode: setMode,
		setIndex: setIndex,
		render: render,
		renderInstances: renderInstances

	};
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLBufferRenderer(gl, extensions, infoRender) {

	var mode;

	function setMode(value) {

		mode = value;
	}

	function render(start, count) {

		gl.drawArrays(mode, start, count);

		infoRender.calls++;
		infoRender.vertices += count;

		if (mode === gl.TRIANGLES) infoRender.faces += count / 3;
	}

	function renderInstances(geometry) {

		var extension = extensions.get('ANGLE_instanced_arrays');

		if (extension === null) {

			console.error('THREE.WebGLBufferRenderer: using THREE.InstancedBufferGeometry but hardware does not support extension ANGLE_instanced_arrays.');
			return;
		}

		var position = geometry.attributes.position;

		var count = 0;

		if (position && position.isInterleavedBufferAttribute) {

			count = position.data.count;

			extension.drawArraysInstancedANGLE(mode, 0, count, geometry.maxInstancedCount);
		} else {

			count = position.count;

			extension.drawArraysInstancedANGLE(mode, 0, count, geometry.maxInstancedCount);
		}

		infoRender.calls++;
		infoRender.vertices += count * geometry.maxInstancedCount;

		if (mode === gl.TRIANGLES) infoRender.faces += geometry.maxInstancedCount * count / 3;
	}

	return {
		setMode: setMode,
		render: render,
		renderInstances: renderInstances
	};
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLLights() {

	var lights = {};

	return {

		get: function get(light) {

			if (lights[light.id] !== undefined) {

				return lights[light.id];
			}

			var uniforms;

			switch (light.type) {

				case 'DirectionalLight':
					uniforms = {
						direction: new Vector3(),
						color: new Color(),

						shadow: false,
						shadowBias: 0,
						shadowRadius: 1,
						shadowMapSize: new Vector2()
					};
					break;

				case 'SpotLight':
					uniforms = {
						position: new Vector3(),
						direction: new Vector3(),
						color: new Color(),
						distance: 0,
						coneCos: 0,
						penumbraCos: 0,
						decay: 0,

						shadow: false,
						shadowBias: 0,
						shadowRadius: 1,
						shadowMapSize: new Vector2()
					};
					break;

				case 'PointLight':
					uniforms = {
						position: new Vector3(),
						color: new Color(),
						distance: 0,
						decay: 0,

						shadow: false,
						shadowBias: 0,
						shadowRadius: 1,
						shadowMapSize: new Vector2()
					};
					break;

				case 'HemisphereLight':
					uniforms = {
						direction: new Vector3(),
						skyColor: new Color(),
						groundColor: new Color()
					};
					break;

			}

			lights[light.id] = uniforms;

			return uniforms;
		}

	};
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function addLineNumbers(string) {

	var lines = string.split('\n');

	for (var i = 0; i < lines.length; i++) {

		lines[i] = i + 1 + ': ' + lines[i];
	}

	return lines.join('\n');
}

function WebGLShader(gl, type, string) {

	var shader = gl.createShader(type);

	gl.shaderSource(shader, string);
	gl.compileShader(shader);

	if (gl.getShaderParameter(shader, gl.COMPILE_STATUS) === false) {

		console.error('THREE.WebGLShader: Shader couldn\'t compile.');
	}

	if (gl.getShaderInfoLog(shader) !== '') {

		console.warn('THREE.WebGLShader: gl.getShaderInfoLog()', type === gl.VERTEX_SHADER ? 'vertex' : 'fragment', gl.getShaderInfoLog(shader), addLineNumbers(string));
	}

	// --enable-privileged-webgl-extension
	// console.log( type, gl.getExtension( 'WEBGL_debug_shaders' ).getTranslatedShaderSource( shader ) );

	return shader;
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

var programIdCount = 0;

function getEncodingComponents(encoding) {

	switch (encoding) {

		case LinearEncoding:
			return ['Linear', '( value )'];
		case sRGBEncoding:
			return ['sRGB', '( value )'];
		case RGBEEncoding:
			return ['RGBE', '( value )'];
		case RGBM7Encoding:
			return ['RGBM', '( value, 7.0 )'];
		case RGBM16Encoding:
			return ['RGBM', '( value, 16.0 )'];
		case RGBDEncoding:
			return ['RGBD', '( value, 256.0 )'];
		case GammaEncoding:
			return ['Gamma', '( value, float( GAMMA_FACTOR ) )'];
		default:
			throw new Error('unsupported encoding: ' + encoding);

	}
}

function getTexelDecodingFunction(functionName, encoding) {

	var components = getEncodingComponents(encoding);
	return "vec4 " + functionName + "( vec4 value ) { return " + components[0] + "ToLinear" + components[1] + "; }";
}

function getTexelEncodingFunction(functionName, encoding) {

	var components = getEncodingComponents(encoding);
	return "vec4 " + functionName + "( vec4 value ) { return LinearTo" + components[0] + components[1] + "; }";
}

function getToneMappingFunction(functionName, toneMapping) {

	var toneMappingName;

	switch (toneMapping) {

		case LinearToneMapping:
			toneMappingName = "Linear";
			break;

		case ReinhardToneMapping:
			toneMappingName = "Reinhard";
			break;

		case Uncharted2ToneMapping:
			toneMappingName = "Uncharted2";
			break;

		case CineonToneMapping:
			toneMappingName = "OptimizedCineon";
			break;

		default:
			throw new Error('unsupported toneMapping: ' + toneMapping);

	}

	return "vec3 " + functionName + "( vec3 color ) { return " + toneMappingName + "ToneMapping( color ); }";
}

function generateExtensions(extensions, parameters, rendererExtensions) {

	extensions = extensions || {};

	var chunks = [extensions.derivatives || parameters.envMapCubeUV || parameters.bumpMap || parameters.normalMap || parameters.flatShading ? '#extension GL_OES_standard_derivatives : enable' : '', (extensions.fragDepth || parameters.logarithmicDepthBuffer) && rendererExtensions.get('EXT_frag_depth') ? '#extension GL_EXT_frag_depth : enable' : '', extensions.drawBuffers && rendererExtensions.get('WEBGL_draw_buffers') ? '#extension GL_EXT_draw_buffers : require' : '', (extensions.shaderTextureLOD || parameters.envMap) && rendererExtensions.get('EXT_shader_texture_lod') ? '#extension GL_EXT_shader_texture_lod : enable' : ''];

	return chunks.filter(filterEmptyLine).join('\n');
}

function generateDefines(defines) {

	var chunks = [];

	for (var name in defines) {

		var value = defines[name];

		if (value === false) continue;

		chunks.push('#define ' + name + ' ' + value);
	}

	return chunks.join('\n');
}

function fetchAttributeLocations(gl, program, identifiers) {

	var attributes = {};

	var n = gl.getProgramParameter(program, gl.ACTIVE_ATTRIBUTES);

	for (var i = 0; i < n; i++) {

		var info = gl.getActiveAttrib(program, i);
		var name = info.name;

		// console.log("THREE.WebGLProgram: ACTIVE VERTEX ATTRIBUTE:", name, i );

		attributes[name] = gl.getAttribLocation(program, name);
	}

	return attributes;
}

function filterEmptyLine(string) {

	return string !== '';
}

function replaceLightNums(string, parameters) {

	return string.replace(/NUM_DIR_LIGHTS/g, parameters.numDirLights).replace(/NUM_SPOT_LIGHTS/g, parameters.numSpotLights).replace(/NUM_POINT_LIGHTS/g, parameters.numPointLights).replace(/NUM_HEMI_LIGHTS/g, parameters.numHemiLights);
}

function parseIncludes(string) {

	var pattern = /#include +<([\w\d.]+)>/g;

	function replace(match, include) {

		var replace = ShaderChunk[include];

		if (replace === undefined) {

			throw new Error('Can not resolve #include <' + include + '>');
		}

		return parseIncludes(replace);
	}

	return string.replace(pattern, replace);
}

function unrollLoops(string) {

	var pattern = /for \( int i \= (\d+)\; i < (\d+)\; i \+\+ \) \{([\s\S]+?)(?=\})\}/g;

	function replace(match, start, end, snippet) {

		var unroll = '';

		for (var i = parseInt(start); i < parseInt(end); i++) {

			unroll += snippet.replace(/\[ i \]/g, '[ ' + i + ' ]');
		}

		return unroll;
	}

	return string.replace(pattern, replace);
}

function WebGLProgram(renderer, code, material, parameters) {

	var gl = renderer.context;

	var extensions = material.extensions;
	var defines = material.defines;

	var vertexShader = material.__webglShader.vertexShader;
	var fragmentShader = material.__webglShader.fragmentShader;

	var shadowMapTypeDefine = 'SHADOWMAP_TYPE_BASIC';

	if (parameters.shadowMapType === PCFShadowMap) {

		shadowMapTypeDefine = 'SHADOWMAP_TYPE_PCF';
	} else if (parameters.shadowMapType === PCFSoftShadowMap) {

		shadowMapTypeDefine = 'SHADOWMAP_TYPE_PCF_SOFT';
	}

	var envMapTypeDefine = 'ENVMAP_TYPE_CUBE';
	var envMapModeDefine = 'ENVMAP_MODE_REFLECTION';
	var envMapBlendingDefine = 'ENVMAP_BLENDING_MULTIPLY';

	if (parameters.envMap) {

		switch (material.envMap.mapping) {

			case CubeReflectionMapping:
			case CubeRefractionMapping:
				envMapTypeDefine = 'ENVMAP_TYPE_CUBE';
				break;

			case CubeUVReflectionMapping:
			case CubeUVRefractionMapping:
				envMapTypeDefine = 'ENVMAP_TYPE_CUBE_UV';
				break;

			case EquirectangularReflectionMapping:
			case EquirectangularRefractionMapping:
				envMapTypeDefine = 'ENVMAP_TYPE_EQUIREC';
				break;

			case SphericalReflectionMapping:
				envMapTypeDefine = 'ENVMAP_TYPE_SPHERE';
				break;

		}

		switch (material.envMap.mapping) {

			case CubeRefractionMapping:
			case EquirectangularRefractionMapping:
				envMapModeDefine = 'ENVMAP_MODE_REFRACTION';
				break;

		}

		switch (material.combine) {

			case MultiplyOperation:
				envMapBlendingDefine = 'ENVMAP_BLENDING_MULTIPLY';
				break;

			case MixOperation:
				envMapBlendingDefine = 'ENVMAP_BLENDING_MIX';
				break;

			case AddOperation:
				envMapBlendingDefine = 'ENVMAP_BLENDING_ADD';
				break;

		}
	}

	var gammaFactorDefine = renderer.gammaFactor > 0 ? renderer.gammaFactor : 1.0;

	// console.log( 'building new program ' );

	//

	var customExtensions = generateExtensions(extensions, parameters, renderer.extensions);

	var customDefines = generateDefines(defines);

	//

	var program = gl.createProgram();

	var prefixVertex, prefixFragment;

	if (material.isRawShaderMaterial) {

		prefixVertex = [customDefines, '\n'].filter(filterEmptyLine).join('\n');

		prefixFragment = [customExtensions, customDefines, '\n'].filter(filterEmptyLine).join('\n');
	} else {

		prefixVertex = ['precision ' + parameters.precision + ' float;', 'precision ' + parameters.precision + ' int;', '#define SHADER_NAME ' + material.__webglShader.name, customDefines, parameters.supportsVertexTextures ? '#define VERTEX_TEXTURES' : '', '#define GAMMA_FACTOR ' + gammaFactorDefine, '#define MAX_BONES ' + parameters.maxBones, parameters.map ? '#define USE_MAP' : '', parameters.envMap ? '#define USE_ENVMAP' : '', parameters.envMap ? '#define ' + envMapModeDefine : '', parameters.lightMap ? '#define USE_LIGHTMAP' : '', parameters.aoMap ? '#define USE_AOMAP' : '', parameters.emissiveMap ? '#define USE_EMISSIVEMAP' : '', parameters.bumpMap ? '#define USE_BUMPMAP' : '', parameters.normalMap ? '#define USE_NORMALMAP' : '', parameters.displacementMap && parameters.supportsVertexTextures ? '#define USE_DISPLACEMENTMAP' : '', parameters.specularMap ? '#define USE_SPECULARMAP' : '', parameters.roughnessMap ? '#define USE_ROUGHNESSMAP' : '', parameters.metalnessMap ? '#define USE_METALNESSMAP' : '', parameters.alphaMap ? '#define USE_ALPHAMAP' : '', parameters.vertexColors ? '#define USE_COLOR' : '', parameters.flatShading ? '#define FLAT_SHADED' : '', parameters.skinning ? '#define USE_SKINNING' : '', parameters.useVertexTexture ? '#define BONE_TEXTURE' : '', parameters.morphTargets ? '#define USE_MORPHTARGETS' : '', parameters.morphNormals && parameters.flatShading === false ? '#define USE_MORPHNORMALS' : '', parameters.doubleSided ? '#define DOUBLE_SIDED' : '', parameters.flipSided ? '#define FLIP_SIDED' : '', '#define NUM_CLIPPING_PLANES ' + parameters.numClippingPlanes, parameters.shadowMapEnabled ? '#define USE_SHADOWMAP' : '', parameters.shadowMapEnabled ? '#define ' + shadowMapTypeDefine : '', parameters.sizeAttenuation ? '#define USE_SIZEATTENUATION' : '', parameters.logarithmicDepthBuffer ? '#define USE_LOGDEPTHBUF' : '', parameters.logarithmicDepthBuffer && renderer.extensions.get('EXT_frag_depth') ? '#define USE_LOGDEPTHBUF_EXT' : '', 'uniform mat4 modelMatrix;', 'uniform mat4 modelViewMatrix;', 'uniform mat4 projectionMatrix;', 'uniform mat4 viewMatrix;', 'uniform mat3 normalMatrix;', 'uniform vec3 cameraPosition;', 'attribute vec3 position;', 'attribute vec3 normal;', 'attribute vec2 uv;', '#ifdef USE_COLOR', '	attribute vec3 color;', '#endif', '#ifdef USE_MORPHTARGETS', '	attribute vec3 morphTarget0;', '	attribute vec3 morphTarget1;', '	attribute vec3 morphTarget2;', '	attribute vec3 morphTarget3;', '	#ifdef USE_MORPHNORMALS', '		attribute vec3 morphNormal0;', '		attribute vec3 morphNormal1;', '		attribute vec3 morphNormal2;', '		attribute vec3 morphNormal3;', '	#else', '		attribute vec3 morphTarget4;', '		attribute vec3 morphTarget5;', '		attribute vec3 morphTarget6;', '		attribute vec3 morphTarget7;', '	#endif', '#endif', '#ifdef USE_SKINNING', '	attribute vec4 skinIndex;', '	attribute vec4 skinWeight;', '#endif', '\n'].filter(filterEmptyLine).join('\n');

		prefixFragment = [customExtensions, 'precision ' + parameters.precision + ' float;', 'precision ' + parameters.precision + ' int;', '#define SHADER_NAME ' + material.__webglShader.name, customDefines, parameters.alphaTest ? '#define ALPHATEST ' + parameters.alphaTest : '', '#define GAMMA_FACTOR ' + gammaFactorDefine, parameters.useFog && parameters.fog ? '#define USE_FOG' : '', parameters.useFog && parameters.fogExp ? '#define FOG_EXP2' : '', parameters.map ? '#define USE_MAP' : '', parameters.envMap ? '#define USE_ENVMAP' : '', parameters.envMap ? '#define ' + envMapTypeDefine : '', parameters.envMap ? '#define ' + envMapModeDefine : '', parameters.envMap ? '#define ' + envMapBlendingDefine : '', parameters.lightMap ? '#define USE_LIGHTMAP' : '', parameters.aoMap ? '#define USE_AOMAP' : '', parameters.emissiveMap ? '#define USE_EMISSIVEMAP' : '', parameters.bumpMap ? '#define USE_BUMPMAP' : '', parameters.normalMap ? '#define USE_NORMALMAP' : '', parameters.specularMap ? '#define USE_SPECULARMAP' : '', parameters.roughnessMap ? '#define USE_ROUGHNESSMAP' : '', parameters.metalnessMap ? '#define USE_METALNESSMAP' : '', parameters.alphaMap ? '#define USE_ALPHAMAP' : '', parameters.vertexColors ? '#define USE_COLOR' : '', parameters.flatShading ? '#define FLAT_SHADED' : '', parameters.doubleSided ? '#define DOUBLE_SIDED' : '', parameters.flipSided ? '#define FLIP_SIDED' : '', '#define NUM_CLIPPING_PLANES ' + parameters.numClippingPlanes, '#define UNION_CLIPPING_PLANES ' + (parameters.numClippingPlanes - parameters.numClipIntersection), parameters.shadowMapEnabled ? '#define USE_SHADOWMAP' : '', parameters.shadowMapEnabled ? '#define ' + shadowMapTypeDefine : '', parameters.premultipliedAlpha ? "#define PREMULTIPLIED_ALPHA" : '', parameters.physicallyCorrectLights ? "#define PHYSICALLY_CORRECT_LIGHTS" : '', parameters.logarithmicDepthBuffer ? '#define USE_LOGDEPTHBUF' : '', parameters.logarithmicDepthBuffer && renderer.extensions.get('EXT_frag_depth') ? '#define USE_LOGDEPTHBUF_EXT' : '', parameters.envMap && renderer.extensions.get('EXT_shader_texture_lod') ? '#define TEXTURE_LOD_EXT' : '', 'uniform mat4 viewMatrix;', 'uniform vec3 cameraPosition;', parameters.toneMapping !== NoToneMapping ? "#define TONE_MAPPING" : '', parameters.toneMapping !== NoToneMapping ? ShaderChunk['tonemapping_pars_fragment'] : '', // this code is required here because it is used by the toneMapping() function defined below
		parameters.toneMapping !== NoToneMapping ? getToneMappingFunction("toneMapping", parameters.toneMapping) : '', parameters.outputEncoding || parameters.mapEncoding || parameters.envMapEncoding || parameters.emissiveMapEncoding ? ShaderChunk['encodings_pars_fragment'] : '', // this code is required here because it is used by the various encoding/decoding function defined below
		parameters.mapEncoding ? getTexelDecodingFunction('mapTexelToLinear', parameters.mapEncoding) : '', parameters.envMapEncoding ? getTexelDecodingFunction('envMapTexelToLinear', parameters.envMapEncoding) : '', parameters.emissiveMapEncoding ? getTexelDecodingFunction('emissiveMapTexelToLinear', parameters.emissiveMapEncoding) : '', parameters.outputEncoding ? getTexelEncodingFunction("linearToOutputTexel", parameters.outputEncoding) : '', parameters.depthPacking ? "#define DEPTH_PACKING " + material.depthPacking : '', '\n'].filter(filterEmptyLine).join('\n');
	}

	vertexShader = parseIncludes(vertexShader, parameters);
	vertexShader = replaceLightNums(vertexShader, parameters);

	fragmentShader = parseIncludes(fragmentShader, parameters);
	fragmentShader = replaceLightNums(fragmentShader, parameters);

	if (!material.isShaderMaterial) {

		vertexShader = unrollLoops(vertexShader);
		fragmentShader = unrollLoops(fragmentShader);
	}

	var vertexGlsl = prefixVertex + vertexShader;
	var fragmentGlsl = prefixFragment + fragmentShader;

	// console.log( '*VERTEX*', vertexGlsl );
	// console.log( '*FRAGMENT*', fragmentGlsl );

	var glVertexShader = WebGLShader(gl, gl.VERTEX_SHADER, vertexGlsl);
	var glFragmentShader = WebGLShader(gl, gl.FRAGMENT_SHADER, fragmentGlsl);

	gl.attachShader(program, glVertexShader);
	gl.attachShader(program, glFragmentShader);

	// Force a particular attribute to index 0.

	if (material.index0AttributeName !== undefined) {

		gl.bindAttribLocation(program, 0, material.index0AttributeName);
	} else if (parameters.morphTargets === true) {

		// programs with morphTargets displace position out of attribute 0
		gl.bindAttribLocation(program, 0, 'position');
	}

	gl.linkProgram(program);

	var programLog = gl.getProgramInfoLog(program);
	var vertexLog = gl.getShaderInfoLog(glVertexShader);
	var fragmentLog = gl.getShaderInfoLog(glFragmentShader);

	var runnable = true;
	var haveDiagnostics = true;

	// console.log( '**VERTEX**', gl.getExtension( 'WEBGL_debug_shaders' ).getTranslatedShaderSource( glVertexShader ) );
	// console.log( '**FRAGMENT**', gl.getExtension( 'WEBGL_debug_shaders' ).getTranslatedShaderSource( glFragmentShader ) );

	if (gl.getProgramParameter(program, gl.LINK_STATUS) === false) {

		runnable = false;

		console.error('THREE.WebGLProgram: shader error: ', gl.getError(), 'gl.VALIDATE_STATUS', gl.getProgramParameter(program, gl.VALIDATE_STATUS), 'gl.getProgramInfoLog', programLog, vertexLog, fragmentLog);
	} else if (programLog !== '') {

		console.warn('THREE.WebGLProgram: gl.getProgramInfoLog()', programLog);
	} else if (vertexLog === '' || fragmentLog === '') {

		haveDiagnostics = false;
	}

	if (haveDiagnostics) {

		this.diagnostics = {

			runnable: runnable,
			material: material,

			programLog: programLog,

			vertexShader: {

				log: vertexLog,
				prefix: prefixVertex

			},

			fragmentShader: {

				log: fragmentLog,
				prefix: prefixFragment

			}

		};
	}

	// clean up

	gl.deleteShader(glVertexShader);
	gl.deleteShader(glFragmentShader);

	// set up caching for uniform locations

	var cachedUniforms;

	this.getUniforms = function () {

		if (cachedUniforms === undefined) {

			cachedUniforms = new WebGLUniforms(gl, program, renderer);
		}

		return cachedUniforms;
	};

	// set up caching for attribute locations

	var cachedAttributes;

	this.getAttributes = function () {

		if (cachedAttributes === undefined) {

			cachedAttributes = fetchAttributeLocations(gl, program);
		}

		return cachedAttributes;
	};

	// free resource

	this.destroy = function () {

		gl.deleteProgram(program);
		this.program = undefined;
	};

	// DEPRECATED

	Object.defineProperties(this, {

		uniforms: {
			get: function get() {

				console.warn('THREE.WebGLProgram: .uniforms is now .getUniforms().');
				return this.getUniforms();
			}
		},

		attributes: {
			get: function get() {

				console.warn('THREE.WebGLProgram: .attributes is now .getAttributes().');
				return this.getAttributes();
			}
		}

	});

	//

	this.id = programIdCount++;
	this.code = code;
	this.usedTimes = 1;
	this.program = program;
	this.vertexShader = glVertexShader;
	this.fragmentShader = glFragmentShader;

	return this;
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLPrograms(renderer, capabilities) {

	var programs = [];

	var shaderIDs = {
		MeshDepthMaterial: 'depth',
		MeshNormalMaterial: 'normal',
		MeshBasicMaterial: 'basic',
		MeshLambertMaterial: 'lambert',
		MeshPhongMaterial: 'phong',
		MeshStandardMaterial: 'physical',
		MeshPhysicalMaterial: 'physical',
		LineBasicMaterial: 'basic',
		LineDashedMaterial: 'dashed',
		PointsMaterial: 'points'
	};

	var parameterNames = ["precision", "supportsVertexTextures", "map", "mapEncoding", "envMap", "envMapMode", "envMapEncoding", "lightMap", "aoMap", "emissiveMap", "emissiveMapEncoding", "bumpMap", "normalMap", "displacementMap", "specularMap", "roughnessMap", "metalnessMap", "alphaMap", "combine", "vertexColors", "fog", "useFog", "fogExp", "flatShading", "sizeAttenuation", "logarithmicDepthBuffer", "skinning", "maxBones", "useVertexTexture", "morphTargets", "morphNormals", "maxMorphTargets", "maxMorphNormals", "premultipliedAlpha", "numDirLights", "numPointLights", "numSpotLights", "numHemiLights", "shadowMapEnabled", "shadowMapType", "toneMapping", 'physicallyCorrectLights', "alphaTest", "doubleSided", "flipSided", "numClippingPlanes", "numClipIntersection", "depthPacking"];

	function allocateBones(object) {

		if (capabilities.floatVertexTextures && object && object.skeleton && object.skeleton.useVertexTexture) {

			return 1024;
		} else {

			// default for when object is not specified
			// ( for example when prebuilding shader to be used with multiple objects )
			//
			//  - leave some extra space for other uniforms
			//  - limit here is ANGLE's 254 max uniform vectors
			//    (up to 54 should be safe)

			var nVertexUniforms = capabilities.maxVertexUniforms;
			var nVertexMatrices = Math.floor((nVertexUniforms - 20) / 4);

			var maxBones = nVertexMatrices;

			if (object !== undefined && object && object.isSkinnedMesh) {

				maxBones = Math.min(object.skeleton.bones.length, maxBones);

				if (maxBones < object.skeleton.bones.length) {

					console.warn('WebGLRenderer: too many bones - ' + object.skeleton.bones.length + ', this GPU supports just ' + maxBones + ' (try OpenGL instead of ANGLE)');
				}
			}

			return maxBones;
		}
	}

	function getTextureEncodingFromMap(map, gammaOverrideLinear) {

		var encoding;

		if (!map) {

			encoding = LinearEncoding;
		} else if (map && map.isTexture) {

			encoding = map.encoding;
		} else if (map && map.isWebGLRenderTarget) {

			console.warn("THREE.WebGLPrograms.getTextureEncodingFromMap: don't use render targets as textures. Use their .texture property instead.");
			encoding = map.texture.encoding;
		}

		// add backwards compatibility for WebGLRenderer.gammaInput/gammaOutput parameter, should probably be removed at some point.
		if (encoding === LinearEncoding && gammaOverrideLinear) {

			encoding = GammaEncoding;
		}

		return encoding;
	}

	this.getParameters = function (material, lights, fog, nClipPlanes, nClipIntersection, object) {

		var shaderID = shaderIDs[material.type];

		// heuristics to create shader parameters according to lights in the scene
		// (not to blow over maxLights budget)

		var maxBones = allocateBones(object);
		var precision = renderer.getPrecision();

		if (material.precision !== null) {

			precision = capabilities.getMaxPrecision(material.precision);

			if (precision !== material.precision) {

				console.warn('THREE.WebGLProgram.getParameters:', material.precision, 'not supported, using', precision, 'instead.');
			}
		}

		var currentRenderTarget = renderer.getCurrentRenderTarget();

		var parameters = {

			shaderID: shaderID,

			precision: precision,
			supportsVertexTextures: capabilities.vertexTextures,
			outputEncoding: getTextureEncodingFromMap(!currentRenderTarget ? null : currentRenderTarget.texture, renderer.gammaOutput),
			map: !!material.map,
			mapEncoding: getTextureEncodingFromMap(material.map, renderer.gammaInput),
			envMap: !!material.envMap,
			envMapMode: material.envMap && material.envMap.mapping,
			envMapEncoding: getTextureEncodingFromMap(material.envMap, renderer.gammaInput),
			envMapCubeUV: !!material.envMap && (material.envMap.mapping === CubeUVReflectionMapping || material.envMap.mapping === CubeUVRefractionMapping),
			lightMap: !!material.lightMap,
			aoMap: !!material.aoMap,
			emissiveMap: !!material.emissiveMap,
			emissiveMapEncoding: getTextureEncodingFromMap(material.emissiveMap, renderer.gammaInput),
			bumpMap: !!material.bumpMap,
			normalMap: !!material.normalMap,
			displacementMap: !!material.displacementMap,
			roughnessMap: !!material.roughnessMap,
			metalnessMap: !!material.metalnessMap,
			specularMap: !!material.specularMap,
			alphaMap: !!material.alphaMap,

			combine: material.combine,

			vertexColors: material.vertexColors,

			fog: !!fog,
			useFog: material.fog,
			fogExp: fog && fog.isFogExp2,

			flatShading: material.shading === FlatShading,

			sizeAttenuation: material.sizeAttenuation,
			logarithmicDepthBuffer: capabilities.logarithmicDepthBuffer,

			skinning: material.skinning,
			maxBones: maxBones,
			useVertexTexture: capabilities.floatVertexTextures && object && object.skeleton && object.skeleton.useVertexTexture,

			morphTargets: material.morphTargets,
			morphNormals: material.morphNormals,
			maxMorphTargets: renderer.maxMorphTargets,
			maxMorphNormals: renderer.maxMorphNormals,

			numDirLights: lights.directional.length,
			numPointLights: lights.point.length,
			numSpotLights: lights.spot.length,
			numHemiLights: lights.hemi.length,

			numClippingPlanes: nClipPlanes,
			numClipIntersection: nClipIntersection,

			shadowMapEnabled: renderer.shadowMap.enabled && object.receiveShadow && lights.shadows.length > 0,
			shadowMapType: renderer.shadowMap.type,

			toneMapping: renderer.toneMapping,
			physicallyCorrectLights: renderer.physicallyCorrectLights,

			premultipliedAlpha: material.premultipliedAlpha,

			alphaTest: material.alphaTest,
			doubleSided: material.side === DoubleSide,
			flipSided: material.side === BackSide,

			depthPacking: material.depthPacking !== undefined ? material.depthPacking : false

		};

		return parameters;
	};

	this.getProgramCode = function (material, parameters) {

		var array = [];

		if (parameters.shaderID) {

			array.push(parameters.shaderID);
		} else {

			array.push(material.fragmentShader);
			array.push(material.vertexShader);
		}

		if (material.defines !== undefined) {

			for (var name in material.defines) {

				array.push(name);
				array.push(material.defines[name]);
			}
		}

		for (var i = 0; i < parameterNames.length; i++) {

			array.push(parameters[parameterNames[i]]);
		}

		return array.join();
	};

	this.acquireProgram = function (material, parameters, code) {

		var program;

		// Check if code has been already compiled
		for (var p = 0, pl = programs.length; p < pl; p++) {

			var programInfo = programs[p];

			if (programInfo.code === code) {

				program = programInfo;
				++program.usedTimes;

				break;
			}
		}

		if (program === undefined) {

			program = new WebGLProgram(renderer, code, material, parameters);
			programs.push(program);
		}

		return program;
	};

	this.releaseProgram = function (program) {

		if (--program.usedTimes === 0) {

			// Remove from unordered set
			var i = programs.indexOf(program);
			programs[i] = programs[programs.length - 1];
			programs.pop();

			// Free WebGL resources
			program.destroy();
		}
	};

	// Exposed for resource monitoring & error feedback via renderer.info:
	this.programs = programs;
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLGeometries(gl, properties, info) {

	var geometries = {};

	function onGeometryDispose(event) {

		var geometry = event.target;
		var buffergeometry = geometries[geometry.id];

		if (buffergeometry.index !== null) {

			deleteAttribute(buffergeometry.index);
		}

		deleteAttributes(buffergeometry.attributes);

		geometry.removeEventListener('dispose', onGeometryDispose);

		delete geometries[geometry.id];

		// TODO

		var property = properties.get(geometry);

		if (property.wireframe) {

			deleteAttribute(property.wireframe);
		}

		properties.delete(geometry);

		var bufferproperty = properties.get(buffergeometry);

		if (bufferproperty.wireframe) {

			deleteAttribute(bufferproperty.wireframe);
		}

		properties.delete(buffergeometry);

		//

		info.memory.geometries--;
	}

	function getAttributeBuffer(attribute) {

		if (attribute.isInterleavedBufferAttribute) {

			return properties.get(attribute.data).__webglBuffer;
		}

		return properties.get(attribute).__webglBuffer;
	}

	function deleteAttribute(attribute) {

		var buffer = getAttributeBuffer(attribute);

		if (buffer !== undefined) {

			gl.deleteBuffer(buffer);
			removeAttributeBuffer(attribute);
		}
	}

	function deleteAttributes(attributes) {

		for (var name in attributes) {

			deleteAttribute(attributes[name]);
		}
	}

	function removeAttributeBuffer(attribute) {

		if (attribute.isInterleavedBufferAttribute) {

			properties.delete(attribute.data);
		} else {

			properties.delete(attribute);
		}
	}

	return {

		get: function get(object) {

			var geometry = object.geometry;

			if (geometries[geometry.id] !== undefined) {

				return geometries[geometry.id];
			}

			geometry.addEventListener('dispose', onGeometryDispose);

			var buffergeometry;

			if (geometry.isBufferGeometry) {

				buffergeometry = geometry;
			} else if (geometry.isGeometry) {

				if (geometry._bufferGeometry === undefined) {

					geometry._bufferGeometry = new BufferGeometry().setFromObject(object);
				}

				buffergeometry = geometry._bufferGeometry;
			}

			geometries[geometry.id] = buffergeometry;

			info.memory.geometries++;

			return buffergeometry;
		}

	};
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLObjects(gl, properties, info) {

	var geometries = new WebGLGeometries(gl, properties, info);

	//

	function update(object) {

		// TODO: Avoid updating twice (when using shadowMap). Maybe add frame counter.

		var geometry = geometries.get(object);

		if (object.geometry.isGeometry) {

			geometry.updateFromObject(object);
		}

		var index = geometry.index;
		var attributes = geometry.attributes;

		if (index !== null) {

			updateAttribute(index, gl.ELEMENT_ARRAY_BUFFER);
		}

		for (var name in attributes) {

			updateAttribute(attributes[name], gl.ARRAY_BUFFER);
		}

		// morph targets

		var morphAttributes = geometry.morphAttributes;

		for (var name in morphAttributes) {

			var array = morphAttributes[name];

			for (var i = 0, l = array.length; i < l; i++) {

				updateAttribute(array[i], gl.ARRAY_BUFFER);
			}
		}

		return geometry;
	}

	function updateAttribute(attribute, bufferType) {

		var data = attribute.isInterleavedBufferAttribute ? attribute.data : attribute;

		var attributeProperties = properties.get(data);

		if (attributeProperties.__webglBuffer === undefined) {

			createBuffer(attributeProperties, data, bufferType);
		} else if (attributeProperties.version !== data.version) {

			updateBuffer(attributeProperties, data, bufferType);
		}
	}

	function createBuffer(attributeProperties, data, bufferType) {

		attributeProperties.__webglBuffer = gl.createBuffer();
		gl.bindBuffer(bufferType, attributeProperties.__webglBuffer);

		var usage = data.dynamic ? gl.DYNAMIC_DRAW : gl.STATIC_DRAW;

		gl.bufferData(bufferType, data.array, usage);

		attributeProperties.version = data.version;
	}

	function updateBuffer(attributeProperties, data, bufferType) {

		gl.bindBuffer(bufferType, attributeProperties.__webglBuffer);

		if (data.dynamic === false) {

			gl.bufferData(bufferType, data.array, gl.STATIC_DRAW);
		} else if (data.updateRange.count === -1) {

			// Not using update ranges

			gl.bufferSubData(bufferType, 0, data.array);
		} else if (data.updateRange.count === 0) {

			console.error('THREE.WebGLObjects.updateBuffer: dynamic THREE.BufferAttribute marked as needsUpdate but updateRange.count is 0, ensure you are using set methods or updating manually.');
		} else {

			gl.bufferSubData(bufferType, data.updateRange.offset * data.array.BYTES_PER_ELEMENT, data.array.subarray(data.updateRange.offset, data.updateRange.offset + data.updateRange.count));

			data.updateRange.count = 0; // reset range
		}

		attributeProperties.version = data.version;
	}

	function getAttributeBuffer(attribute) {

		if (attribute.isInterleavedBufferAttribute) {

			return properties.get(attribute.data).__webglBuffer;
		}

		return properties.get(attribute).__webglBuffer;
	}

	function getWireframeAttribute(geometry) {

		var property = properties.get(geometry);

		if (property.wireframe !== undefined) {

			return property.wireframe;
		}

		var indices = [];

		var index = geometry.index;
		var attributes = geometry.attributes;
		var position = attributes.position;

		// console.time( 'wireframe' );

		if (index !== null) {

			var edges = {};
			var array = index.array;

			for (var i = 0, l = array.length; i < l; i += 3) {

				var a = array[i + 0];
				var b = array[i + 1];
				var c = array[i + 2];

				indices.push(a, b, b, c, c, a);
			}
		} else {

			var array = attributes.position.array;

			for (var i = 0, l = array.length / 3 - 1; i < l; i += 3) {

				var a = i + 0;
				var b = i + 1;
				var c = i + 2;

				indices.push(a, b, b, c, c, a);
			}
		}

		// console.timeEnd( 'wireframe' );

		var TypeArray = position.count > 65535 ? Uint32Array : Uint16Array;
		var attribute = new BufferAttribute(new TypeArray(indices), 1);

		updateAttribute(attribute, gl.ELEMENT_ARRAY_BUFFER);

		property.wireframe = attribute;

		return attribute;
	}

	return {

		getAttributeBuffer: getAttributeBuffer,
		getWireframeAttribute: getWireframeAttribute,

		update: update

	};
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLTextures(_gl, extensions, state, properties, capabilities, paramThreeToGL, info) {

	var _infoMemory = info.memory;
	var _isWebGL2 = typeof WebGL2RenderingContext !== 'undefined' && _gl instanceof WebGL2RenderingContext;

	//

	function clampToMaxSize(image, maxSize) {

		if (image.width > maxSize || image.height > maxSize) {

			// Warning: Scaling through the canvas will only work with images that use
			// premultiplied alpha.

			var scale = maxSize / Math.max(image.width, image.height);

			var canvas = document.createElementNS('http://www.w3.org/1999/xhtml', 'canvas');
			canvas.width = Math.floor(image.width * scale);
			canvas.height = Math.floor(image.height * scale);

			var context = canvas.getContext('2d');
			context.drawImage(image, 0, 0, image.width, image.height, 0, 0, canvas.width, canvas.height);

			console.warn('THREE.WebGLRenderer: image is too big (' + image.width + 'x' + image.height + '). Resized to ' + canvas.width + 'x' + canvas.height, image);

			return canvas;
		}

		return image;
	}

	function isPowerOfTwo(image) {

		return _Math.isPowerOfTwo(image.width) && _Math.isPowerOfTwo(image.height);
	}

	function makePowerOfTwo(image) {

		if (image instanceof HTMLImageElement || image instanceof HTMLCanvasElement) {

			var canvas = document.createElementNS('http://www.w3.org/1999/xhtml', 'canvas');
			canvas.width = _Math.nearestPowerOfTwo(image.width);
			canvas.height = _Math.nearestPowerOfTwo(image.height);

			var context = canvas.getContext('2d');
			context.drawImage(image, 0, 0, canvas.width, canvas.height);

			console.warn('THREE.WebGLRenderer: image is not power of two (' + image.width + 'x' + image.height + '). Resized to ' + canvas.width + 'x' + canvas.height, image);

			return canvas;
		}

		return image;
	}

	function textureNeedsPowerOfTwo(texture) {

		if (texture.wrapS !== ClampToEdgeWrapping || texture.wrapT !== ClampToEdgeWrapping) return true;
		if (texture.minFilter !== NearestFilter && texture.minFilter !== LinearFilter) return true;

		return false;
	}

	// Fallback filters for non-power-of-2 textures

	function filterFallback(f) {

		if (f === NearestFilter || f === NearestMipMapNearestFilter || f === NearestMipMapLinearFilter) {

			return _gl.NEAREST;
		}

		return _gl.LINEAR;
	}

	//

	function onTextureDispose(event) {

		var texture = event.target;

		texture.removeEventListener('dispose', onTextureDispose);

		deallocateTexture(texture);

		_infoMemory.textures--;
	}

	function onRenderTargetDispose(event) {

		var renderTarget = event.target;

		renderTarget.removeEventListener('dispose', onRenderTargetDispose);

		deallocateRenderTarget(renderTarget);

		_infoMemory.textures--;
	}

	//

	function deallocateTexture(texture) {

		var textureProperties = properties.get(texture);

		if (texture.image && textureProperties.__image__webglTextureCube) {

			// cube texture

			_gl.deleteTexture(textureProperties.__image__webglTextureCube);
		} else {

			// 2D texture

			if (textureProperties.__webglInit === undefined) return;

			_gl.deleteTexture(textureProperties.__webglTexture);
		}

		// remove all webgl properties
		properties.delete(texture);
	}

	function deallocateRenderTarget(renderTarget) {

		var renderTargetProperties = properties.get(renderTarget);
		var textureProperties = properties.get(renderTarget.texture);

		if (!renderTarget) return;

		if (textureProperties.__webglTexture !== undefined) {

			_gl.deleteTexture(textureProperties.__webglTexture);
		}

		if (renderTarget.depthTexture) {

			renderTarget.depthTexture.dispose();
		}

		if (renderTarget && renderTarget.isWebGLRenderTargetCube) {

			for (var i = 0; i < 6; i++) {

				_gl.deleteFramebuffer(renderTargetProperties.__webglFramebuffer[i]);
				if (renderTargetProperties.__webglDepthbuffer) _gl.deleteRenderbuffer(renderTargetProperties.__webglDepthbuffer[i]);
			}
		} else {

			_gl.deleteFramebuffer(renderTargetProperties.__webglFramebuffer);
			if (renderTargetProperties.__webglDepthbuffer) _gl.deleteRenderbuffer(renderTargetProperties.__webglDepthbuffer);
		}

		properties.delete(renderTarget.texture);
		properties.delete(renderTarget);
	}

	//


	function setTexture2D(texture, slot) {

		var textureProperties = properties.get(texture);

		if (texture.version > 0 && textureProperties.__version !== texture.version) {

			var image = texture.image;

			if (image === undefined) {

				console.warn('THREE.WebGLRenderer: Texture marked for update but image is undefined', texture);
			} else if (image.complete === false) {

				console.warn('THREE.WebGLRenderer: Texture marked for update but image is incomplete', texture);
			} else {

				uploadTexture(textureProperties, texture, slot);
				return;
			}
		}

		state.activeTexture(_gl.TEXTURE0 + slot);
		state.bindTexture(_gl.TEXTURE_2D, textureProperties.__webglTexture);
	}

	function setTextureCube(texture, slot) {

		var textureProperties = properties.get(texture);

		if (texture.image.length === 6) {

			if (texture.version > 0 && textureProperties.__version !== texture.version) {

				if (!textureProperties.__image__webglTextureCube) {

					texture.addEventListener('dispose', onTextureDispose);

					textureProperties.__image__webglTextureCube = _gl.createTexture();

					_infoMemory.textures++;
				}

				state.activeTexture(_gl.TEXTURE0 + slot);
				state.bindTexture(_gl.TEXTURE_CUBE_MAP, textureProperties.__image__webglTextureCube);

				_gl.pixelStorei(_gl.UNPACK_FLIP_Y_WEBGL, texture.flipY);

				var isCompressed = texture && texture.isCompressedTexture;
				var isDataTexture = texture.image[0] && texture.image[0].isDataTexture;

				var cubeImage = [];

				for (var i = 0; i < 6; i++) {

					if (!isCompressed && !isDataTexture) {

						cubeImage[i] = clampToMaxSize(texture.image[i], capabilities.maxCubemapSize);
					} else {

						cubeImage[i] = isDataTexture ? texture.image[i].image : texture.image[i];
					}
				}

				var image = cubeImage[0],
				    isPowerOfTwoImage = isPowerOfTwo(image),
				    glFormat = paramThreeToGL(texture.format),
				    glType = paramThreeToGL(texture.type);

				setTextureParameters(_gl.TEXTURE_CUBE_MAP, texture, isPowerOfTwoImage);

				for (var i = 0; i < 6; i++) {

					if (!isCompressed) {

						if (isDataTexture) {

							state.texImage2D(_gl.TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, glFormat, cubeImage[i].width, cubeImage[i].height, 0, glFormat, glType, cubeImage[i].data);
						} else {

							state.texImage2D(_gl.TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, glFormat, glFormat, glType, cubeImage[i]);
						}
					} else {

						var mipmap,
						    mipmaps = cubeImage[i].mipmaps;

						for (var j = 0, jl = mipmaps.length; j < jl; j++) {

							mipmap = mipmaps[j];

							if (texture.format !== RGBAFormat && texture.format !== RGBFormat) {

								if (state.getCompressedTextureFormats().indexOf(glFormat) > -1) {

									state.compressedTexImage2D(_gl.TEXTURE_CUBE_MAP_POSITIVE_X + i, j, glFormat, mipmap.width, mipmap.height, 0, mipmap.data);
								} else {

									console.warn("THREE.WebGLRenderer: Attempt to load unsupported compressed texture format in .setTextureCube()");
								}
							} else {

								state.texImage2D(_gl.TEXTURE_CUBE_MAP_POSITIVE_X + i, j, glFormat, mipmap.width, mipmap.height, 0, glFormat, glType, mipmap.data);
							}
						}
					}
				}

				if (texture.generateMipmaps && isPowerOfTwoImage) {

					_gl.generateMipmap(_gl.TEXTURE_CUBE_MAP);
				}

				textureProperties.__version = texture.version;

				if (texture.onUpdate) texture.onUpdate(texture);
			} else {

				state.activeTexture(_gl.TEXTURE0 + slot);
				state.bindTexture(_gl.TEXTURE_CUBE_MAP, textureProperties.__image__webglTextureCube);
			}
		}
	}

	function setTextureCubeDynamic(texture, slot) {

		state.activeTexture(_gl.TEXTURE0 + slot);
		state.bindTexture(_gl.TEXTURE_CUBE_MAP, properties.get(texture).__webglTexture);
	}

	function setTextureParameters(textureType, texture, isPowerOfTwoImage) {

		var extension;

		if (isPowerOfTwoImage) {

			_gl.texParameteri(textureType, _gl.TEXTURE_WRAP_S, paramThreeToGL(texture.wrapS));
			_gl.texParameteri(textureType, _gl.TEXTURE_WRAP_T, paramThreeToGL(texture.wrapT));

			_gl.texParameteri(textureType, _gl.TEXTURE_MAG_FILTER, paramThreeToGL(texture.magFilter));
			_gl.texParameteri(textureType, _gl.TEXTURE_MIN_FILTER, paramThreeToGL(texture.minFilter));
		} else {

			_gl.texParameteri(textureType, _gl.TEXTURE_WRAP_S, _gl.CLAMP_TO_EDGE);
			_gl.texParameteri(textureType, _gl.TEXTURE_WRAP_T, _gl.CLAMP_TO_EDGE);

			if (texture.wrapS !== ClampToEdgeWrapping || texture.wrapT !== ClampToEdgeWrapping) {

				console.warn('THREE.WebGLRenderer: Texture is not power of two. Texture.wrapS and Texture.wrapT should be set to THREE.ClampToEdgeWrapping.', texture);
			}

			_gl.texParameteri(textureType, _gl.TEXTURE_MAG_FILTER, filterFallback(texture.magFilter));
			_gl.texParameteri(textureType, _gl.TEXTURE_MIN_FILTER, filterFallback(texture.minFilter));

			if (texture.minFilter !== NearestFilter && texture.minFilter !== LinearFilter) {

				console.warn('THREE.WebGLRenderer: Texture is not power of two. Texture.minFilter should be set to THREE.NearestFilter or THREE.LinearFilter.', texture);
			}
		}

		extension = extensions.get('EXT_texture_filter_anisotropic');

		if (extension) {

			if (texture.type === FloatType && extensions.get('OES_texture_float_linear') === null) return;
			if (texture.type === HalfFloatType && extensions.get('OES_texture_half_float_linear') === null) return;

			if (texture.anisotropy > 1 || properties.get(texture).__currentAnisotropy) {

				_gl.texParameterf(textureType, extension.TEXTURE_MAX_ANISOTROPY_EXT, Math.min(texture.anisotropy, capabilities.getMaxAnisotropy()));
				properties.get(texture).__currentAnisotropy = texture.anisotropy;
			}
		}
	}

	function uploadTexture(textureProperties, texture, slot) {

		if (textureProperties.__webglInit === undefined) {

			textureProperties.__webglInit = true;

			texture.addEventListener('dispose', onTextureDispose);

			textureProperties.__webglTexture = _gl.createTexture();

			_infoMemory.textures++;
		}

		state.activeTexture(_gl.TEXTURE0 + slot);
		state.bindTexture(_gl.TEXTURE_2D, textureProperties.__webglTexture);

		_gl.pixelStorei(_gl.UNPACK_FLIP_Y_WEBGL, texture.flipY);
		_gl.pixelStorei(_gl.UNPACK_PREMULTIPLY_ALPHA_WEBGL, texture.premultiplyAlpha);
		_gl.pixelStorei(_gl.UNPACK_ALIGNMENT, texture.unpackAlignment);

		var image = clampToMaxSize(texture.image, capabilities.maxTextureSize);

		if (textureNeedsPowerOfTwo(texture) && isPowerOfTwo(image) === false) {

			image = makePowerOfTwo(image);
		}

		var isPowerOfTwoImage = isPowerOfTwo(image),
		    glFormat = paramThreeToGL(texture.format),
		    glType = paramThreeToGL(texture.type);

		setTextureParameters(_gl.TEXTURE_2D, texture, isPowerOfTwoImage);

		var mipmap,
		    mipmaps = texture.mipmaps;

		if (texture && texture.isDepthTexture) {

			// populate depth texture with dummy data

			var internalFormat = _gl.DEPTH_COMPONENT;

			if (texture.type === FloatType) {

				if (!_isWebGL2) throw new Error('Float Depth Texture only supported in WebGL2.0');
				internalFormat = _gl.DEPTH_COMPONENT32F;
			} else if (_isWebGL2) {

				// WebGL 2.0 requires signed internalformat for glTexImage2D
				internalFormat = _gl.DEPTH_COMPONENT16;
			}

			// Depth stencil textures need the DEPTH_STENCIL internal format
			// (https://www.khronos.org/registry/webgl/extensions/WEBGL_depth_texture/)
			if (texture.format === DepthStencilFormat) {

				internalFormat = _gl.DEPTH_STENCIL;
			}

			state.texImage2D(_gl.TEXTURE_2D, 0, internalFormat, image.width, image.height, 0, glFormat, glType, null);
		} else if (texture && texture.isDataTexture) {

			// use manually created mipmaps if available
			// if there are no manual mipmaps
			// set 0 level mipmap and then use GL to generate other mipmap levels

			if (mipmaps.length > 0 && isPowerOfTwoImage) {

				for (var i = 0, il = mipmaps.length; i < il; i++) {

					mipmap = mipmaps[i];
					state.texImage2D(_gl.TEXTURE_2D, i, glFormat, mipmap.width, mipmap.height, 0, glFormat, glType, mipmap.data);
				}

				texture.generateMipmaps = false;
			} else {

				state.texImage2D(_gl.TEXTURE_2D, 0, glFormat, image.width, image.height, 0, glFormat, glType, image.data);
			}
		} else if (texture && texture.isCompressedTexture) {

			for (var i = 0, il = mipmaps.length; i < il; i++) {

				mipmap = mipmaps[i];

				if (texture.format !== RGBAFormat && texture.format !== RGBFormat) {

					if (state.getCompressedTextureFormats().indexOf(glFormat) > -1) {

						state.compressedTexImage2D(_gl.TEXTURE_2D, i, glFormat, mipmap.width, mipmap.height, 0, mipmap.data);
					} else {

						console.warn("THREE.WebGLRenderer: Attempt to load unsupported compressed texture format in .uploadTexture()");
					}
				} else {

					state.texImage2D(_gl.TEXTURE_2D, i, glFormat, mipmap.width, mipmap.height, 0, glFormat, glType, mipmap.data);
				}
			}
		} else {

			// regular Texture (image, video, canvas)

			// use manually created mipmaps if available
			// if there are no manual mipmaps
			// set 0 level mipmap and then use GL to generate other mipmap levels

			if (mipmaps.length > 0 && isPowerOfTwoImage) {

				for (var i = 0, il = mipmaps.length; i < il; i++) {

					mipmap = mipmaps[i];
					state.texImage2D(_gl.TEXTURE_2D, i, glFormat, glFormat, glType, mipmap);
				}

				texture.generateMipmaps = false;
			} else {

				state.texImage2D(_gl.TEXTURE_2D, 0, glFormat, glFormat, glType, image);
			}
		}

		if (texture.generateMipmaps && isPowerOfTwoImage) _gl.generateMipmap(_gl.TEXTURE_2D);

		textureProperties.__version = texture.version;

		if (texture.onUpdate) texture.onUpdate(texture);
	}

	// Render targets

	// Setup storage for target texture and bind it to correct framebuffer
	function setupFrameBufferTexture(framebuffer, renderTarget, attachment, textureTarget) {

		var glFormat = paramThreeToGL(renderTarget.texture.format);
		var glType = paramThreeToGL(renderTarget.texture.type);
		state.texImage2D(textureTarget, 0, glFormat, renderTarget.width, renderTarget.height, 0, glFormat, glType, null);
		_gl.bindFramebuffer(_gl.FRAMEBUFFER, framebuffer);
		_gl.framebufferTexture2D(_gl.FRAMEBUFFER, attachment, textureTarget, properties.get(renderTarget.texture).__webglTexture, 0);
		_gl.bindFramebuffer(_gl.FRAMEBUFFER, null);
	}

	// Setup storage for internal depth/stencil buffers and bind to correct framebuffer
	function setupRenderBufferStorage(renderbuffer, renderTarget) {

		_gl.bindRenderbuffer(_gl.RENDERBUFFER, renderbuffer);

		if (renderTarget.depthBuffer && !renderTarget.stencilBuffer) {

			_gl.renderbufferStorage(_gl.RENDERBUFFER, _gl.DEPTH_COMPONENT16, renderTarget.width, renderTarget.height);
			_gl.framebufferRenderbuffer(_gl.FRAMEBUFFER, _gl.DEPTH_ATTACHMENT, _gl.RENDERBUFFER, renderbuffer);
		} else if (renderTarget.depthBuffer && renderTarget.stencilBuffer) {

			_gl.renderbufferStorage(_gl.RENDERBUFFER, _gl.DEPTH_STENCIL, renderTarget.width, renderTarget.height);
			_gl.framebufferRenderbuffer(_gl.FRAMEBUFFER, _gl.DEPTH_STENCIL_ATTACHMENT, _gl.RENDERBUFFER, renderbuffer);
		} else {

			// FIXME: We don't support !depth !stencil
			_gl.renderbufferStorage(_gl.RENDERBUFFER, _gl.RGBA4, renderTarget.width, renderTarget.height);
		}

		_gl.bindRenderbuffer(_gl.RENDERBUFFER, null);
	}

	// Setup resources for a Depth Texture for a FBO (needs an extension)
	function setupDepthTexture(framebuffer, renderTarget) {

		var isCube = renderTarget && renderTarget.isWebGLRenderTargetCube;
		if (isCube) throw new Error('Depth Texture with cube render targets is not supported!');

		_gl.bindFramebuffer(_gl.FRAMEBUFFER, framebuffer);

		if (!(renderTarget.depthTexture && renderTarget.depthTexture.isDepthTexture)) {

			throw new Error('renderTarget.depthTexture must be an instance of THREE.DepthTexture');
		}

		// upload an empty depth texture with framebuffer size
		if (!properties.get(renderTarget.depthTexture).__webglTexture || renderTarget.depthTexture.image.width !== renderTarget.width || renderTarget.depthTexture.image.height !== renderTarget.height) {
			renderTarget.depthTexture.image.width = renderTarget.width;
			renderTarget.depthTexture.image.height = renderTarget.height;
			renderTarget.depthTexture.needsUpdate = true;
		}

		setTexture2D(renderTarget.depthTexture, 0);

		var webglDepthTexture = properties.get(renderTarget.depthTexture).__webglTexture;

		if (renderTarget.depthTexture.format === DepthFormat) {

			_gl.framebufferTexture2D(_gl.FRAMEBUFFER, _gl.DEPTH_ATTACHMENT, _gl.TEXTURE_2D, webglDepthTexture, 0);
		} else if (renderTarget.depthTexture.format === DepthStencilFormat) {

			_gl.framebufferTexture2D(_gl.FRAMEBUFFER, _gl.DEPTH_STENCIL_ATTACHMENT, _gl.TEXTURE_2D, webglDepthTexture, 0);
		} else {

			throw new Error('Unknown depthTexture format');
		}
	}

	// Setup GL resources for a non-texture depth buffer
	function setupDepthRenderbuffer(renderTarget) {

		var renderTargetProperties = properties.get(renderTarget);

		var isCube = renderTarget && renderTarget.isWebGLRenderTargetCube;

		if (renderTarget.depthTexture) {

			if (isCube) throw new Error('target.depthTexture not supported in Cube render targets');

			setupDepthTexture(renderTargetProperties.__webglFramebuffer, renderTarget);
		} else {

			if (isCube) {

				renderTargetProperties.__webglDepthbuffer = [];

				for (var i = 0; i < 6; i++) {

					_gl.bindFramebuffer(_gl.FRAMEBUFFER, renderTargetProperties.__webglFramebuffer[i]);
					renderTargetProperties.__webglDepthbuffer[i] = _gl.createRenderbuffer();
					setupRenderBufferStorage(renderTargetProperties.__webglDepthbuffer[i], renderTarget);
				}
			} else {

				_gl.bindFramebuffer(_gl.FRAMEBUFFER, renderTargetProperties.__webglFramebuffer);
				renderTargetProperties.__webglDepthbuffer = _gl.createRenderbuffer();
				setupRenderBufferStorage(renderTargetProperties.__webglDepthbuffer, renderTarget);
			}
		}

		_gl.bindFramebuffer(_gl.FRAMEBUFFER, null);
	}

	// Set up GL resources for the render target
	function setupRenderTarget(renderTarget) {

		var renderTargetProperties = properties.get(renderTarget);
		var textureProperties = properties.get(renderTarget.texture);

		renderTarget.addEventListener('dispose', onRenderTargetDispose);

		textureProperties.__webglTexture = _gl.createTexture();

		_infoMemory.textures++;

		var isCube = renderTarget && renderTarget.isWebGLRenderTargetCube;
		var isTargetPowerOfTwo = isPowerOfTwo(renderTarget);

		// Setup framebuffer

		if (isCube) {

			renderTargetProperties.__webglFramebuffer = [];

			for (var i = 0; i < 6; i++) {

				renderTargetProperties.__webglFramebuffer[i] = _gl.createFramebuffer();
			}
		} else {

			renderTargetProperties.__webglFramebuffer = _gl.createFramebuffer();
		}

		// Setup color buffer

		if (isCube) {

			state.bindTexture(_gl.TEXTURE_CUBE_MAP, textureProperties.__webglTexture);
			setTextureParameters(_gl.TEXTURE_CUBE_MAP, renderTarget.texture, isTargetPowerOfTwo);

			for (var i = 0; i < 6; i++) {

				setupFrameBufferTexture(renderTargetProperties.__webglFramebuffer[i], renderTarget, _gl.COLOR_ATTACHMENT0, _gl.TEXTURE_CUBE_MAP_POSITIVE_X + i);
			}

			if (renderTarget.texture.generateMipmaps && isTargetPowerOfTwo) _gl.generateMipmap(_gl.TEXTURE_CUBE_MAP);
			state.bindTexture(_gl.TEXTURE_CUBE_MAP, null);
		} else {

			state.bindTexture(_gl.TEXTURE_2D, textureProperties.__webglTexture);
			setTextureParameters(_gl.TEXTURE_2D, renderTarget.texture, isTargetPowerOfTwo);
			setupFrameBufferTexture(renderTargetProperties.__webglFramebuffer, renderTarget, _gl.COLOR_ATTACHMENT0, _gl.TEXTURE_2D);

			if (renderTarget.texture.generateMipmaps && isTargetPowerOfTwo) _gl.generateMipmap(_gl.TEXTURE_2D);
			state.bindTexture(_gl.TEXTURE_2D, null);
		}

		// Setup depth and stencil buffers

		if (renderTarget.depthBuffer) {

			setupDepthRenderbuffer(renderTarget);
		}
	}

	function updateRenderTargetMipmap(renderTarget) {

		var texture = renderTarget.texture;

		if (texture.generateMipmaps && isPowerOfTwo(renderTarget) && texture.minFilter !== NearestFilter && texture.minFilter !== LinearFilter) {

			var target = renderTarget && renderTarget.isWebGLRenderTargetCube ? _gl.TEXTURE_CUBE_MAP : _gl.TEXTURE_2D;
			var webglTexture = properties.get(texture).__webglTexture;

			state.bindTexture(target, webglTexture);
			_gl.generateMipmap(target);
			state.bindTexture(target, null);
		}
	}

	this.setTexture2D = setTexture2D;
	this.setTextureCube = setTextureCube;
	this.setTextureCubeDynamic = setTextureCubeDynamic;
	this.setupRenderTarget = setupRenderTarget;
	this.updateRenderTargetMipmap = updateRenderTargetMipmap;
}

/**
 * @author fordacious / fordacious.github.io
 */

function WebGLProperties() {

	var properties = {};

	return {

		get: function get(object) {

			var uuid = object.uuid;
			var map = properties[uuid];

			if (map === undefined) {

				map = {};
				properties[uuid] = map;
			}

			return map;
		},

		delete: function _delete(object) {

			delete properties[object.uuid];
		},

		clear: function clear() {

			properties = {};
		}

	};
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLState(gl, extensions, paramThreeToGL) {

	function ColorBuffer() {

		var locked = false;

		var color = new Vector4();
		var currentColorMask = null;
		var currentColorClear = new Vector4();

		return {

			setMask: function setMask(colorMask) {

				if (currentColorMask !== colorMask && !locked) {

					gl.colorMask(colorMask, colorMask, colorMask, colorMask);
					currentColorMask = colorMask;
				}
			},

			setLocked: function setLocked(lock) {

				locked = lock;
			},

			setClear: function setClear(r, g, b, a) {

				color.set(r, g, b, a);

				if (currentColorClear.equals(color) === false) {

					gl.clearColor(r, g, b, a);
					currentColorClear.copy(color);
				}
			},

			reset: function reset() {

				locked = false;

				currentColorMask = null;
				currentColorClear.set(0, 0, 0, 1);
			}

		};
	}

	function DepthBuffer() {

		var locked = false;

		var currentDepthMask = null;
		var currentDepthFunc = null;
		var currentDepthClear = null;

		return {

			setTest: function setTest(depthTest) {

				if (depthTest) {

					enable(gl.DEPTH_TEST);
				} else {

					disable(gl.DEPTH_TEST);
				}
			},

			setMask: function setMask(depthMask) {

				if (currentDepthMask !== depthMask && !locked) {

					gl.depthMask(depthMask);
					currentDepthMask = depthMask;
				}
			},

			setFunc: function setFunc(depthFunc) {

				if (currentDepthFunc !== depthFunc) {

					if (depthFunc) {

						switch (depthFunc) {

							case NeverDepth:

								gl.depthFunc(gl.NEVER);
								break;

							case AlwaysDepth:

								gl.depthFunc(gl.ALWAYS);
								break;

							case LessDepth:

								gl.depthFunc(gl.LESS);
								break;

							case LessEqualDepth:

								gl.depthFunc(gl.LEQUAL);
								break;

							case EqualDepth:

								gl.depthFunc(gl.EQUAL);
								break;

							case GreaterEqualDepth:

								gl.depthFunc(gl.GEQUAL);
								break;

							case GreaterDepth:

								gl.depthFunc(gl.GREATER);
								break;

							case NotEqualDepth:

								gl.depthFunc(gl.NOTEQUAL);
								break;

							default:

								gl.depthFunc(gl.LEQUAL);

						}
					} else {

						gl.depthFunc(gl.LEQUAL);
					}

					currentDepthFunc = depthFunc;
				}
			},

			setLocked: function setLocked(lock) {

				locked = lock;
			},

			setClear: function setClear(depth) {

				if (currentDepthClear !== depth) {

					gl.clearDepth(depth);
					currentDepthClear = depth;
				}
			},

			reset: function reset() {

				locked = false;

				currentDepthMask = null;
				currentDepthFunc = null;
				currentDepthClear = null;
			}

		};
	}

	function StencilBuffer() {

		var locked = false;

		var currentStencilMask = null;
		var currentStencilFunc = null;
		var currentStencilRef = null;
		var currentStencilFuncMask = null;
		var currentStencilFail = null;
		var currentStencilZFail = null;
		var currentStencilZPass = null;
		var currentStencilClear = null;

		return {

			setTest: function setTest(stencilTest) {

				if (stencilTest) {

					enable(gl.STENCIL_TEST);
				} else {

					disable(gl.STENCIL_TEST);
				}
			},

			setMask: function setMask(stencilMask) {

				if (currentStencilMask !== stencilMask && !locked) {

					gl.stencilMask(stencilMask);
					currentStencilMask = stencilMask;
				}
			},

			setFunc: function setFunc(stencilFunc, stencilRef, stencilMask) {

				if (currentStencilFunc !== stencilFunc || currentStencilRef !== stencilRef || currentStencilFuncMask !== stencilMask) {

					gl.stencilFunc(stencilFunc, stencilRef, stencilMask);

					currentStencilFunc = stencilFunc;
					currentStencilRef = stencilRef;
					currentStencilFuncMask = stencilMask;
				}
			},

			setOp: function setOp(stencilFail, stencilZFail, stencilZPass) {

				if (currentStencilFail !== stencilFail || currentStencilZFail !== stencilZFail || currentStencilZPass !== stencilZPass) {

					gl.stencilOp(stencilFail, stencilZFail, stencilZPass);

					currentStencilFail = stencilFail;
					currentStencilZFail = stencilZFail;
					currentStencilZPass = stencilZPass;
				}
			},

			setLocked: function setLocked(lock) {

				locked = lock;
			},

			setClear: function setClear(stencil) {

				if (currentStencilClear !== stencil) {

					gl.clearStencil(stencil);
					currentStencilClear = stencil;
				}
			},

			reset: function reset() {

				locked = false;

				currentStencilMask = null;
				currentStencilFunc = null;
				currentStencilRef = null;
				currentStencilFuncMask = null;
				currentStencilFail = null;
				currentStencilZFail = null;
				currentStencilZPass = null;
				currentStencilClear = null;
			}

		};
	}

	//

	var colorBuffer = new ColorBuffer();
	var depthBuffer = new DepthBuffer();
	var stencilBuffer = new StencilBuffer();

	var maxVertexAttributes = gl.getParameter(gl.MAX_VERTEX_ATTRIBS);
	var newAttributes = new Uint8Array(maxVertexAttributes);
	var enabledAttributes = new Uint8Array(maxVertexAttributes);
	var attributeDivisors = new Uint8Array(maxVertexAttributes);

	var capabilities = {};

	var compressedTextureFormats = null;

	var currentBlending = null;
	var currentBlendEquation = null;
	var currentBlendSrc = null;
	var currentBlendDst = null;
	var currentBlendEquationAlpha = null;
	var currentBlendSrcAlpha = null;
	var currentBlendDstAlpha = null;
	var currentPremultipledAlpha = false;

	var currentFlipSided = null;
	var currentCullFace = null;

	var currentLineWidth = null;

	var currentPolygonOffsetFactor = null;
	var currentPolygonOffsetUnits = null;

	var currentScissorTest = null;

	var maxTextures = gl.getParameter(gl.MAX_TEXTURE_IMAGE_UNITS);

	var currentTextureSlot = null;
	var currentBoundTextures = {};

	var currentScissor = new Vector4();
	var currentViewport = new Vector4();

	function createTexture(type, target, count) {

		var data = new Uint8Array(4); // 4 is required to match default unpack alignment of 4.
		var texture = gl.createTexture();

		gl.bindTexture(type, texture);
		gl.texParameteri(type, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
		gl.texParameteri(type, gl.TEXTURE_MAG_FILTER, gl.NEAREST);

		for (var i = 0; i < count; i++) {

			gl.texImage2D(target + i, 0, gl.RGBA, 1, 1, 0, gl.RGBA, gl.UNSIGNED_BYTE, data);
		}

		return texture;
	}

	var emptyTextures = {};
	emptyTextures[gl.TEXTURE_2D] = createTexture(gl.TEXTURE_2D, gl.TEXTURE_2D, 1);
	emptyTextures[gl.TEXTURE_CUBE_MAP] = createTexture(gl.TEXTURE_CUBE_MAP, gl.TEXTURE_CUBE_MAP_POSITIVE_X, 6);

	//

	function init() {

		clearColor(0, 0, 0, 1);
		clearDepth(1);
		clearStencil(0);

		enable(gl.DEPTH_TEST);
		setDepthFunc(LessEqualDepth);

		setFlipSided(false);
		setCullFace(CullFaceBack);
		enable(gl.CULL_FACE);

		enable(gl.BLEND);
		setBlending(NormalBlending);
	}

	function initAttributes() {

		for (var i = 0, l = newAttributes.length; i < l; i++) {

			newAttributes[i] = 0;
		}
	}

	function enableAttribute(attribute) {

		newAttributes[attribute] = 1;

		if (enabledAttributes[attribute] === 0) {

			gl.enableVertexAttribArray(attribute);
			enabledAttributes[attribute] = 1;
		}

		if (attributeDivisors[attribute] !== 0) {

			var extension = extensions.get('ANGLE_instanced_arrays');

			extension.vertexAttribDivisorANGLE(attribute, 0);
			attributeDivisors[attribute] = 0;
		}
	}

	function enableAttributeAndDivisor(attribute, meshPerAttribute, extension) {

		newAttributes[attribute] = 1;

		if (enabledAttributes[attribute] === 0) {

			gl.enableVertexAttribArray(attribute);
			enabledAttributes[attribute] = 1;
		}

		if (attributeDivisors[attribute] !== meshPerAttribute) {

			extension.vertexAttribDivisorANGLE(attribute, meshPerAttribute);
			attributeDivisors[attribute] = meshPerAttribute;
		}
	}

	function disableUnusedAttributes() {

		for (var i = 0, l = enabledAttributes.length; i !== l; ++i) {

			if (enabledAttributes[i] !== newAttributes[i]) {

				gl.disableVertexAttribArray(i);
				enabledAttributes[i] = 0;
			}
		}
	}

	function enable(id) {

		if (capabilities[id] !== true) {

			gl.enable(id);
			capabilities[id] = true;
		}
	}

	function disable(id) {

		if (capabilities[id] !== false) {

			gl.disable(id);
			capabilities[id] = false;
		}
	}

	function getCompressedTextureFormats() {

		if (compressedTextureFormats === null) {

			compressedTextureFormats = [];

			if (extensions.get('WEBGL_compressed_texture_pvrtc') || extensions.get('WEBGL_compressed_texture_s3tc') || extensions.get('WEBGL_compressed_texture_etc1')) {

				var formats = gl.getParameter(gl.COMPRESSED_TEXTURE_FORMATS);

				for (var i = 0; i < formats.length; i++) {

					compressedTextureFormats.push(formats[i]);
				}
			}
		}

		return compressedTextureFormats;
	}

	function setBlending(blending, blendEquation, blendSrc, blendDst, blendEquationAlpha, blendSrcAlpha, blendDstAlpha, premultipliedAlpha) {

		if (blending !== NoBlending) {

			enable(gl.BLEND);
		} else {

			disable(gl.BLEND);
		}

		if (blending !== currentBlending || premultipliedAlpha !== currentPremultipledAlpha) {

			if (blending === AdditiveBlending) {

				if (premultipliedAlpha) {

					gl.blendEquationSeparate(gl.FUNC_ADD, gl.FUNC_ADD);
					gl.blendFuncSeparate(gl.ONE, gl.ONE, gl.ONE, gl.ONE);
				} else {

					gl.blendEquation(gl.FUNC_ADD);
					gl.blendFunc(gl.SRC_ALPHA, gl.ONE);
				}
			} else if (blending === SubtractiveBlending) {

				if (premultipliedAlpha) {

					gl.blendEquationSeparate(gl.FUNC_ADD, gl.FUNC_ADD);
					gl.blendFuncSeparate(gl.ZERO, gl.ZERO, gl.ONE_MINUS_SRC_COLOR, gl.ONE_MINUS_SRC_ALPHA);
				} else {

					gl.blendEquation(gl.FUNC_ADD);
					gl.blendFunc(gl.ZERO, gl.ONE_MINUS_SRC_COLOR);
				}
			} else if (blending === MultiplyBlending) {

				if (premultipliedAlpha) {

					gl.blendEquationSeparate(gl.FUNC_ADD, gl.FUNC_ADD);
					gl.blendFuncSeparate(gl.ZERO, gl.SRC_COLOR, gl.ZERO, gl.SRC_ALPHA);
				} else {

					gl.blendEquation(gl.FUNC_ADD);
					gl.blendFunc(gl.ZERO, gl.SRC_COLOR);
				}
			} else {

				if (premultipliedAlpha) {

					gl.blendEquationSeparate(gl.FUNC_ADD, gl.FUNC_ADD);
					gl.blendFuncSeparate(gl.ONE, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
				} else {

					gl.blendEquationSeparate(gl.FUNC_ADD, gl.FUNC_ADD);
					gl.blendFuncSeparate(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
				}
			}

			currentBlending = blending;
			currentPremultipledAlpha = premultipliedAlpha;
		}

		if (blending === CustomBlending) {

			blendEquationAlpha = blendEquationAlpha || blendEquation;
			blendSrcAlpha = blendSrcAlpha || blendSrc;
			blendDstAlpha = blendDstAlpha || blendDst;

			if (blendEquation !== currentBlendEquation || blendEquationAlpha !== currentBlendEquationAlpha) {

				gl.blendEquationSeparate(paramThreeToGL(blendEquation), paramThreeToGL(blendEquationAlpha));

				currentBlendEquation = blendEquation;
				currentBlendEquationAlpha = blendEquationAlpha;
			}

			if (blendSrc !== currentBlendSrc || blendDst !== currentBlendDst || blendSrcAlpha !== currentBlendSrcAlpha || blendDstAlpha !== currentBlendDstAlpha) {

				gl.blendFuncSeparate(paramThreeToGL(blendSrc), paramThreeToGL(blendDst), paramThreeToGL(blendSrcAlpha), paramThreeToGL(blendDstAlpha));

				currentBlendSrc = blendSrc;
				currentBlendDst = blendDst;
				currentBlendSrcAlpha = blendSrcAlpha;
				currentBlendDstAlpha = blendDstAlpha;
			}
		} else {

			currentBlendEquation = null;
			currentBlendSrc = null;
			currentBlendDst = null;
			currentBlendEquationAlpha = null;
			currentBlendSrcAlpha = null;
			currentBlendDstAlpha = null;
		}
	}

	// TODO Deprecate

	function setColorWrite(colorWrite) {

		colorBuffer.setMask(colorWrite);
	}

	function setDepthTest(depthTest) {

		depthBuffer.setTest(depthTest);
	}

	function setDepthWrite(depthWrite) {

		depthBuffer.setMask(depthWrite);
	}

	function setDepthFunc(depthFunc) {

		depthBuffer.setFunc(depthFunc);
	}

	function setStencilTest(stencilTest) {

		stencilBuffer.setTest(stencilTest);
	}

	function setStencilWrite(stencilWrite) {

		stencilBuffer.setMask(stencilWrite);
	}

	function setStencilFunc(stencilFunc, stencilRef, stencilMask) {

		stencilBuffer.setFunc(stencilFunc, stencilRef, stencilMask);
	}

	function setStencilOp(stencilFail, stencilZFail, stencilZPass) {

		stencilBuffer.setOp(stencilFail, stencilZFail, stencilZPass);
	}

	//

	function setFlipSided(flipSided) {

		if (currentFlipSided !== flipSided) {

			if (flipSided) {

				gl.frontFace(gl.CW);
			} else {

				gl.frontFace(gl.CCW);
			}

			currentFlipSided = flipSided;
		}
	}

	function setCullFace(cullFace) {

		if (cullFace !== CullFaceNone) {

			enable(gl.CULL_FACE);

			if (cullFace !== currentCullFace) {

				if (cullFace === CullFaceBack) {

					gl.cullFace(gl.BACK);
				} else if (cullFace === CullFaceFront) {

					gl.cullFace(gl.FRONT);
				} else {

					gl.cullFace(gl.FRONT_AND_BACK);
				}
			}
		} else {

			disable(gl.CULL_FACE);
		}

		currentCullFace = cullFace;
	}

	function setLineWidth(width) {

		if (width !== currentLineWidth) {

			gl.lineWidth(width);

			currentLineWidth = width;
		}
	}

	function setPolygonOffset(polygonOffset, factor, units) {

		if (polygonOffset) {

			enable(gl.POLYGON_OFFSET_FILL);

			if (currentPolygonOffsetFactor !== factor || currentPolygonOffsetUnits !== units) {

				gl.polygonOffset(factor, units);

				currentPolygonOffsetFactor = factor;
				currentPolygonOffsetUnits = units;
			}
		} else {

			disable(gl.POLYGON_OFFSET_FILL);
		}
	}

	function getScissorTest() {

		return currentScissorTest;
	}

	function setScissorTest(scissorTest) {

		currentScissorTest = scissorTest;

		if (scissorTest) {

			enable(gl.SCISSOR_TEST);
		} else {

			disable(gl.SCISSOR_TEST);
		}
	}

	// texture

	function activeTexture(webglSlot) {

		if (webglSlot === undefined) webglSlot = gl.TEXTURE0 + maxTextures - 1;

		if (currentTextureSlot !== webglSlot) {

			gl.activeTexture(webglSlot);
			currentTextureSlot = webglSlot;
		}
	}

	function bindTexture(webglType, webglTexture) {

		if (currentTextureSlot === null) {

			activeTexture();
		}

		var boundTexture = currentBoundTextures[currentTextureSlot];

		if (boundTexture === undefined) {

			boundTexture = { type: undefined, texture: undefined };
			currentBoundTextures[currentTextureSlot] = boundTexture;
		}

		if (boundTexture.type !== webglType || boundTexture.texture !== webglTexture) {

			gl.bindTexture(webglType, webglTexture || emptyTextures[webglType]);

			boundTexture.type = webglType;
			boundTexture.texture = webglTexture;
		}
	}

	function compressedTexImage2D() {

		try {

			gl.compressedTexImage2D.apply(gl, arguments);
		} catch (error) {

			console.error(error);
		}
	}

	function texImage2D() {

		try {

			gl.texImage2D.apply(gl, arguments);
		} catch (error) {

			console.error(error);
		}
	}

	// TODO Deprecate

	function clearColor(r, g, b, a) {

		colorBuffer.setClear(r, g, b, a);
	}

	function clearDepth(depth) {

		depthBuffer.setClear(depth);
	}

	function clearStencil(stencil) {

		stencilBuffer.setClear(stencil);
	}

	//

	function scissor(scissor) {

		if (currentScissor.equals(scissor) === false) {

			gl.scissor(scissor.x, scissor.y, scissor.z, scissor.w);
			currentScissor.copy(scissor);
		}
	}

	function viewport(viewport) {

		if (currentViewport.equals(viewport) === false) {

			gl.viewport(viewport.x, viewport.y, viewport.z, viewport.w);
			currentViewport.copy(viewport);
		}
	}

	//

	function reset() {

		for (var i = 0; i < enabledAttributes.length; i++) {

			if (enabledAttributes[i] === 1) {

				gl.disableVertexAttribArray(i);
				enabledAttributes[i] = 0;
			}
		}

		capabilities = {};

		compressedTextureFormats = null;

		currentTextureSlot = null;
		currentBoundTextures = {};

		currentBlending = null;

		currentFlipSided = null;
		currentCullFace = null;

		colorBuffer.reset();
		depthBuffer.reset();
		stencilBuffer.reset();
	}

	return {

		buffers: {
			color: colorBuffer,
			depth: depthBuffer,
			stencil: stencilBuffer
		},

		init: init,
		initAttributes: initAttributes,
		enableAttribute: enableAttribute,
		enableAttributeAndDivisor: enableAttributeAndDivisor,
		disableUnusedAttributes: disableUnusedAttributes,
		enable: enable,
		disable: disable,
		getCompressedTextureFormats: getCompressedTextureFormats,

		setBlending: setBlending,

		setColorWrite: setColorWrite,
		setDepthTest: setDepthTest,
		setDepthWrite: setDepthWrite,
		setDepthFunc: setDepthFunc,
		setStencilTest: setStencilTest,
		setStencilWrite: setStencilWrite,
		setStencilFunc: setStencilFunc,
		setStencilOp: setStencilOp,

		setFlipSided: setFlipSided,
		setCullFace: setCullFace,

		setLineWidth: setLineWidth,
		setPolygonOffset: setPolygonOffset,

		getScissorTest: getScissorTest,
		setScissorTest: setScissorTest,

		activeTexture: activeTexture,
		bindTexture: bindTexture,
		compressedTexImage2D: compressedTexImage2D,
		texImage2D: texImage2D,

		clearColor: clearColor,
		clearDepth: clearDepth,
		clearStencil: clearStencil,

		scissor: scissor,
		viewport: viewport,

		reset: reset

	};
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLCapabilities(gl, extensions, parameters) {

	var maxAnisotropy;

	function getMaxAnisotropy() {

		if (maxAnisotropy !== undefined) return maxAnisotropy;

		var extension = extensions.get('EXT_texture_filter_anisotropic');

		if (extension !== null) {

			maxAnisotropy = gl.getParameter(extension.MAX_TEXTURE_MAX_ANISOTROPY_EXT);
		} else {

			maxAnisotropy = 0;
		}

		return maxAnisotropy;
	}

	function getMaxPrecision(precision) {

		if (precision === 'highp') {

			if (gl.getShaderPrecisionFormat(gl.VERTEX_SHADER, gl.HIGH_FLOAT).precision > 0 && gl.getShaderPrecisionFormat(gl.FRAGMENT_SHADER, gl.HIGH_FLOAT).precision > 0) {

				return 'highp';
			}

			precision = 'mediump';
		}

		if (precision === 'mediump') {

			if (gl.getShaderPrecisionFormat(gl.VERTEX_SHADER, gl.MEDIUM_FLOAT).precision > 0 && gl.getShaderPrecisionFormat(gl.FRAGMENT_SHADER, gl.MEDIUM_FLOAT).precision > 0) {

				return 'mediump';
			}
		}

		return 'lowp';
	}

	var precision = parameters.precision !== undefined ? parameters.precision : 'highp';
	var maxPrecision = getMaxPrecision(precision);

	if (maxPrecision !== precision) {

		console.warn('THREE.WebGLRenderer:', precision, 'not supported, using', maxPrecision, 'instead.');
		precision = maxPrecision;
	}

	var logarithmicDepthBuffer = parameters.logarithmicDepthBuffer === true && !!extensions.get('EXT_frag_depth');

	var maxTextures = gl.getParameter(gl.MAX_TEXTURE_IMAGE_UNITS);
	var maxVertexTextures = gl.getParameter(gl.MAX_VERTEX_TEXTURE_IMAGE_UNITS);
	var maxTextureSize = gl.getParameter(gl.MAX_TEXTURE_SIZE);
	var maxCubemapSize = gl.getParameter(gl.MAX_CUBE_MAP_TEXTURE_SIZE);

	var maxAttributes = gl.getParameter(gl.MAX_VERTEX_ATTRIBS);
	var maxVertexUniforms = gl.getParameter(gl.MAX_VERTEX_UNIFORM_VECTORS);
	var maxVaryings = gl.getParameter(gl.MAX_VARYING_VECTORS);
	var maxFragmentUniforms = gl.getParameter(gl.MAX_FRAGMENT_UNIFORM_VECTORS);

	var vertexTextures = maxVertexTextures > 0;
	var floatFragmentTextures = !!extensions.get('OES_texture_float');
	var floatVertexTextures = vertexTextures && floatFragmentTextures;

	return {

		getMaxAnisotropy: getMaxAnisotropy,
		getMaxPrecision: getMaxPrecision,

		precision: precision,
		logarithmicDepthBuffer: logarithmicDepthBuffer,

		maxTextures: maxTextures,
		maxVertexTextures: maxVertexTextures,
		maxTextureSize: maxTextureSize,
		maxCubemapSize: maxCubemapSize,

		maxAttributes: maxAttributes,
		maxVertexUniforms: maxVertexUniforms,
		maxVaryings: maxVaryings,
		maxFragmentUniforms: maxFragmentUniforms,

		vertexTextures: vertexTextures,
		floatFragmentTextures: floatFragmentTextures,
		floatVertexTextures: floatVertexTextures

	};
}

/**
 * @author mrdoob / http://mrdoob.com/
 */

function WebGLExtensions(gl) {

	var extensions = {};

	return {

		get: function get(name) {

			if (extensions[name] !== undefined) {

				return extensions[name];
			}

			var extension;

			switch (name) {

				case 'WEBGL_depth_texture':
					extension = gl.getExtension('WEBGL_depth_texture') || gl.getExtension('MOZ_WEBGL_depth_texture') || gl.getExtension('WEBKIT_WEBGL_depth_texture');
					break;

				case 'EXT_texture_filter_anisotropic':
					extension = gl.getExtension('EXT_texture_filter_anisotropic') || gl.getExtension('MOZ_EXT_texture_filter_anisotropic') || gl.getExtension('WEBKIT_EXT_texture_filter_anisotropic');
					break;

				case 'WEBGL_compressed_texture_s3tc':
					extension = gl.getExtension('WEBGL_compressed_texture_s3tc') || gl.getExtension('MOZ_WEBGL_compressed_texture_s3tc') || gl.getExtension('WEBKIT_WEBGL_compressed_texture_s3tc');
					break;

				case 'WEBGL_compressed_texture_pvrtc':
					extension = gl.getExtension('WEBGL_compressed_texture_pvrtc') || gl.getExtension('WEBKIT_WEBGL_compressed_texture_pvrtc');
					break;

				case 'WEBGL_compressed_texture_etc1':
					extension = gl.getExtension('WEBGL_compressed_texture_etc1');
					break;

				default:
					extension = gl.getExtension(name);

			}

			if (extension === null) {

				console.warn('THREE.WebGLRenderer: ' + name + ' extension not supported.');
			}

			extensions[name] = extension;

			return extension;
		}

	};
}

/**
 * @author tschw
 */

function WebGLClipping() {

	var scope = this,
	    globalState = null,
	    numGlobalPlanes = 0,
	    localClippingEnabled = false,
	    renderingShadows = false,
	    plane = new Plane(),
	    viewNormalMatrix = new Matrix3(),
	    uniform = { value: null, needsUpdate: false };

	this.uniform = uniform;
	this.numPlanes = 0;
	this.numIntersection = 0;

	this.init = function (planes, enableLocalClipping, camera) {

		var enabled = planes.length !== 0 || enableLocalClipping ||
		// enable state of previous frame - the clipping code has to
		// run another frame in order to reset the state:
		numGlobalPlanes !== 0 || localClippingEnabled;

		localClippingEnabled = enableLocalClipping;

		globalState = projectPlanes(planes, camera, 0);
		numGlobalPlanes = planes.length;

		return enabled;
	};

	this.beginShadows = function () {

		renderingShadows = true;
		projectPlanes(null);
	};

	this.endShadows = function () {

		renderingShadows = false;
		resetGlobalState();
	};

	this.setState = function (planes, clipIntersection, clipShadows, camera, cache, fromCache) {

		if (!localClippingEnabled || planes === null || planes.length === 0 || renderingShadows && !clipShadows) {
			// there's no local clipping

			if (renderingShadows) {
				// there's no global clipping

				projectPlanes(null);
			} else {

				resetGlobalState();
			}
		} else {

			var nGlobal = renderingShadows ? 0 : numGlobalPlanes,
			    lGlobal = nGlobal * 4,
			    dstArray = cache.clippingState || null;

			uniform.value = dstArray; // ensure unique state

			dstArray = projectPlanes(planes, camera, lGlobal, fromCache);

			for (var i = 0; i !== lGlobal; ++i) {

				dstArray[i] = globalState[i];
			}

			cache.clippingState = dstArray;
			this.numIntersection = clipIntersection ? this.numPlanes : 0;
			this.numPlanes += nGlobal;
		}
	};

	function resetGlobalState() {

		if (uniform.value !== globalState) {

			uniform.value = globalState;
			uniform.needsUpdate = numGlobalPlanes > 0;
		}

		scope.numPlanes = numGlobalPlanes;
		scope.numIntersection = 0;
	}

	function projectPlanes(planes, camera, dstOffset, skipTransform) {

		var nPlanes = planes !== null ? planes.length : 0,
		    dstArray = null;

		if (nPlanes !== 0) {

			dstArray = uniform.value;

			if (skipTransform !== true || dstArray === null) {

				var flatSize = dstOffset + nPlanes * 4,
				    viewMatrix = camera.matrixWorldInverse;

				viewNormalMatrix.getNormalMatrix(viewMatrix);

				if (dstArray === null || dstArray.length < flatSize) {

					dstArray = new Float32Array(flatSize);
				}

				for (var i = 0, i4 = dstOffset; i !== nPlanes; ++i, i4 += 4) {

					plane.copy(planes[i]).applyMatrix4(viewMatrix, viewNormalMatrix);

					plane.normal.toArray(dstArray, i4);
					dstArray[i4 + 3] = plane.constant;
				}
			}

			uniform.value = dstArray;
			uniform.needsUpdate = true;
		}

		scope.numPlanes = nPlanes;

		return dstArray;
	}
}

/**
 * @author supereggbert / http://www.paulbrunt.co.uk/
 * @author mrdoob / http://mrdoob.com/
 * @author alteredq / http://alteredqualia.com/
 * @author szimek / https://github.com/szimek/
 * @author tschw
 */

function WebGLRenderer(parameters) {

	console.log('THREE.WebGLRenderer', REVISION);

	parameters = parameters || {};

	var _canvas = parameters.canvas !== undefined ? parameters.canvas : document.createElementNS('http://www.w3.org/1999/xhtml', 'canvas'),
	    _context = parameters.context !== undefined ? parameters.context : null,
	    _alpha = parameters.alpha !== undefined ? parameters.alpha : false,
	    _depth = parameters.depth !== undefined ? parameters.depth : true,
	    _stencil = parameters.stencil !== undefined ? parameters.stencil : true,
	    _antialias = parameters.antialias !== undefined ? parameters.antialias : false,
	    _premultipliedAlpha = parameters.premultipliedAlpha !== undefined ? parameters.premultipliedAlpha : true,
	    _preserveDrawingBuffer = parameters.preserveDrawingBuffer !== undefined ? parameters.preserveDrawingBuffer : false;

	var lights = [];

	var opaqueObjects = [];
	var opaqueObjectsLastIndex = -1;
	var transparentObjects = [];
	var transparentObjectsLastIndex = -1;

	var morphInfluences = new Float32Array(8);

	var sprites = [];
	var lensFlares = [];

	// public properties

	this.domElement = _canvas;
	this.context = null;

	// clearing

	this.autoClear = true;
	this.autoClearColor = true;
	this.autoClearDepth = true;
	this.autoClearStencil = true;

	// scene graph

	this.sortObjects = true;

	// user-defined clipping

	this.clippingPlanes = [];
	this.localClippingEnabled = false;

	// physically based shading

	this.gammaFactor = 2.0; // for backwards compatibility
	this.gammaInput = false;
	this.gammaOutput = false;

	// physical lights

	this.physicallyCorrectLights = false;

	// tone mapping

	this.toneMapping = LinearToneMapping;
	this.toneMappingExposure = 1.0;
	this.toneMappingWhitePoint = 1.0;

	// morphs

	this.maxMorphTargets = 8;
	this.maxMorphNormals = 4;

	// internal properties

	var _this = this,


	// internal state cache

	_currentProgram = null,
	    _currentRenderTarget = null,
	    _currentFramebuffer = null,
	    _currentMaterialId = -1,
	    _currentGeometryProgram = '',
	    _currentCamera = null,
	    _currentScissor = new Vector4(),
	    _currentScissorTest = null,
	    _currentViewport = new Vector4(),


	//

	_usedTextureUnits = 0,


	//

	_clearColor = new Color(0x000000),
	    _clearAlpha = 0,
	    _width = _canvas.width,
	    _height = _canvas.height,
	    _pixelRatio = 1,
	    _scissor = new Vector4(0, 0, _width, _height),
	    _scissorTest = false,
	    _viewport = new Vector4(0, 0, _width, _height),


	// frustum

	_frustum = new Frustum(),


	// clipping

	_clipping = new WebGLClipping(),
	    _clippingEnabled = false,
	    _localClippingEnabled = false,
	    _sphere = new Sphere(),


	// camera matrices cache

	_projScreenMatrix = new Matrix4(),
	    _vector3 = new Vector3(),


	// light arrays cache

	_lights = {

		hash: '',

		ambient: [0, 0, 0],
		directional: [],
		directionalShadowMap: [],
		directionalShadowMatrix: [],
		spot: [],
		spotShadowMap: [],
		spotShadowMatrix: [],
		point: [],
		pointShadowMap: [],
		pointShadowMatrix: [],
		hemi: [],

		shadows: []

	},


	// info

	_infoRender = {

		calls: 0,
		vertices: 0,
		faces: 0,
		points: 0

	};

	this.info = {

		render: _infoRender,
		memory: {

			geometries: 0,
			textures: 0

		},
		programs: null

	};

	// initialize

	var _gl;

	try {

		var attributes = {
			alpha: _alpha,
			depth: _depth,
			stencil: _stencil,
			antialias: _antialias,
			premultipliedAlpha: _premultipliedAlpha,
			preserveDrawingBuffer: _preserveDrawingBuffer
		};

		_gl = _context || _canvas.getContext('webgl', attributes) || _canvas.getContext('experimental-webgl', attributes);

		if (_gl === null) {

			if (_canvas.getContext('webgl') !== null) {

				throw 'Error creating WebGL context with your selected attributes.';
			} else {

				throw 'Error creating WebGL context.';
			}
		}

		// Some experimental-webgl implementations do not have getShaderPrecisionFormat

		if (_gl.getShaderPrecisionFormat === undefined) {

			_gl.getShaderPrecisionFormat = function () {

				return { 'rangeMin': 1, 'rangeMax': 1, 'precision': 1 };
			};
		}

		_canvas.addEventListener('webglcontextlost', onContextLost, false);
	} catch (error) {

		console.error('THREE.WebGLRenderer: ' + error);
	}

	var extensions = new WebGLExtensions(_gl);

	extensions.get('WEBGL_depth_texture');
	extensions.get('OES_texture_float');
	extensions.get('OES_texture_float_linear');
	extensions.get('OES_texture_half_float');
	extensions.get('OES_texture_half_float_linear');
	extensions.get('OES_standard_derivatives');
	extensions.get('ANGLE_instanced_arrays');

	if (extensions.get('OES_element_index_uint')) {

		BufferGeometry.MaxIndex = 4294967296;
	}

	var capabilities = new WebGLCapabilities(_gl, extensions, parameters);

	var state = new WebGLState(_gl, extensions, paramThreeToGL);
	var properties = new WebGLProperties();
	var textures = new WebGLTextures(_gl, extensions, state, properties, capabilities, paramThreeToGL, this.info);
	var objects = new WebGLObjects(_gl, properties, this.info);
	var programCache = new WebGLPrograms(this, capabilities);
	var lightCache = new WebGLLights();

	this.info.programs = programCache.programs;

	var bufferRenderer = new WebGLBufferRenderer(_gl, extensions, _infoRender);
	var indexedBufferRenderer = new WebGLIndexedBufferRenderer(_gl, extensions, _infoRender);

	//

	var backgroundCamera = new OrthographicCamera(-1, 1, 1, -1, 0, 1);
	var backgroundCamera2 = new PerspectiveCamera();
	var backgroundPlaneMesh = new Mesh(new PlaneBufferGeometry(2, 2), new MeshBasicMaterial({ depthTest: false, depthWrite: false, fog: false }));
	var backgroundBoxShader = ShaderLib['cube'];
	var backgroundBoxMesh = new Mesh(new BoxBufferGeometry(5, 5, 5), new ShaderMaterial({
		uniforms: backgroundBoxShader.uniforms,
		vertexShader: backgroundBoxShader.vertexShader,
		fragmentShader: backgroundBoxShader.fragmentShader,
		side: BackSide,
		depthTest: false,
		depthWrite: false,
		fog: false
	}));

	//

	function getTargetPixelRatio() {

		return _currentRenderTarget === null ? _pixelRatio : 1;
	}

	function glClearColor(r, g, b, a) {

		if (_premultipliedAlpha === true) {

			r *= a;g *= a;b *= a;
		}

		state.clearColor(r, g, b, a);
	}

	function setDefaultGLState() {

		state.init();

		state.scissor(_currentScissor.copy(_scissor).multiplyScalar(_pixelRatio));
		state.viewport(_currentViewport.copy(_viewport).multiplyScalar(_pixelRatio));

		glClearColor(_clearColor.r, _clearColor.g, _clearColor.b, _clearAlpha);
	}

	function resetGLState() {

		_currentProgram = null;
		_currentCamera = null;

		_currentGeometryProgram = '';
		_currentMaterialId = -1;

		state.reset();
	}

	setDefaultGLState();

	this.context = _gl;
	this.capabilities = capabilities;
	this.extensions = extensions;
	this.properties = properties;
	this.state = state;

	// shadow map

	var shadowMap = new WebGLShadowMap(this, _lights, objects, capabilities);

	this.shadowMap = shadowMap;

	// Plugins

	var spritePlugin = new SpritePlugin(this, sprites);
	var lensFlarePlugin = new LensFlarePlugin(this, lensFlares);

	// API

	this.getContext = function () {

		return _gl;
	};

	this.getContextAttributes = function () {

		return _gl.getContextAttributes();
	};

	this.forceContextLoss = function () {

		extensions.get('WEBGL_lose_context').loseContext();
	};

	this.getMaxAnisotropy = function () {

		return capabilities.getMaxAnisotropy();
	};

	this.getPrecision = function () {

		return capabilities.precision;
	};

	this.getPixelRatio = function () {

		return _pixelRatio;
	};

	this.setPixelRatio = function (value) {

		if (value === undefined) return;

		_pixelRatio = value;

		this.setSize(_viewport.z, _viewport.w, false);
	};

	this.getSize = function () {

		return {
			width: _width,
			height: _height
		};
	};

	this.setSize = function (width, height, updateStyle) {

		_width = width;
		_height = height;

		_canvas.width = width * _pixelRatio;
		_canvas.height = height * _pixelRatio;

		if (updateStyle !== false) {

			_canvas.style.width = width + 'px';
			_canvas.style.height = height + 'px';
		}

		this.setViewport(0, 0, width, height);
	};

	this.setViewport = function (x, y, width, height) {

		state.viewport(_viewport.set(x, y, width, height));
	};

	this.setScissor = function (x, y, width, height) {

		state.scissor(_scissor.set(x, y, width, height));
	};

	this.setScissorTest = function (boolean) {

		state.setScissorTest(_scissorTest = boolean);
	};

	// Clearing

	this.getClearColor = function () {

		return _clearColor;
	};

	this.setClearColor = function (color, alpha) {

		_clearColor.set(color);

		_clearAlpha = alpha !== undefined ? alpha : 1;

		glClearColor(_clearColor.r, _clearColor.g, _clearColor.b, _clearAlpha);
	};

	this.getClearAlpha = function () {

		return _clearAlpha;
	};

	this.setClearAlpha = function (alpha) {

		_clearAlpha = alpha;

		glClearColor(_clearColor.r, _clearColor.g, _clearColor.b, _clearAlpha);
	};

	this.clear = function (color, depth, stencil) {

		var bits = 0;

		if (color === undefined || color) bits |= _gl.COLOR_BUFFER_BIT;
		if (depth === undefined || depth) bits |= _gl.DEPTH_BUFFER_BIT;
		if (stencil === undefined || stencil) bits |= _gl.STENCIL_BUFFER_BIT;

		_gl.clear(bits);
	};

	this.clearColor = function () {

		this.clear(true, false, false);
	};

	this.clearDepth = function () {

		this.clear(false, true, false);
	};

	this.clearStencil = function () {

		this.clear(false, false, true);
	};

	this.clearTarget = function (renderTarget, color, depth, stencil) {

		this.setRenderTarget(renderTarget);
		this.clear(color, depth, stencil);
	};

	// Reset

	this.resetGLState = resetGLState;

	this.dispose = function () {

		transparentObjects = [];
		transparentObjectsLastIndex = -1;
		opaqueObjects = [];
		opaqueObjectsLastIndex = -1;

		_canvas.removeEventListener('webglcontextlost', onContextLost, false);
	};

	// Events

	function onContextLost(event) {

		event.preventDefault();

		resetGLState();
		setDefaultGLState();

		properties.clear();
	}

	function onMaterialDispose(event) {

		var material = event.target;

		material.removeEventListener('dispose', onMaterialDispose);

		deallocateMaterial(material);
	}

	// Buffer deallocation

	function deallocateMaterial(material) {

		releaseMaterialProgramReference(material);

		properties.delete(material);
	}

	function releaseMaterialProgramReference(material) {

		var programInfo = properties.get(material).program;

		material.program = undefined;

		if (programInfo !== undefined) {

			programCache.releaseProgram(programInfo);
		}
	}

	// Buffer rendering

	this.renderBufferImmediate = function (object, program, material) {

		state.initAttributes();

		var buffers = properties.get(object);

		if (object.hasPositions && !buffers.position) buffers.position = _gl.createBuffer();
		if (object.hasNormals && !buffers.normal) buffers.normal = _gl.createBuffer();
		if (object.hasUvs && !buffers.uv) buffers.uv = _gl.createBuffer();
		if (object.hasColors && !buffers.color) buffers.color = _gl.createBuffer();

		var attributes = program.getAttributes();

		if (object.hasPositions) {

			_gl.bindBuffer(_gl.ARRAY_BUFFER, buffers.position);
			_gl.bufferData(_gl.ARRAY_BUFFER, object.positionArray, _gl.DYNAMIC_DRAW);

			state.enableAttribute(attributes.position);
			_gl.vertexAttribPointer(attributes.position, 3, _gl.FLOAT, false, 0, 0);
		}

		if (object.hasNormals) {

			_gl.bindBuffer(_gl.ARRAY_BUFFER, buffers.normal);

			if (!material.isMeshPhongMaterial && !material.isMeshStandardMaterial && material.shading === FlatShading) {

				for (var i = 0, l = object.count * 3; i < l; i += 9) {

					var array = object.normalArray;

					var nx = (array[i + 0] + array[i + 3] + array[i + 6]) / 3;
					var ny = (array[i + 1] + array[i + 4] + array[i + 7]) / 3;
					var nz = (array[i + 2] + array[i + 5] + array[i + 8]) / 3;

					array[i + 0] = nx;
					array[i + 1] = ny;
					array[i + 2] = nz;

					array[i + 3] = nx;
					array[i + 4] = ny;
					array[i + 5] = nz;

					array[i + 6] = nx;
					array[i + 7] = ny;
					array[i + 8] = nz;
				}
			}

			_gl.bufferData(_gl.ARRAY_BUFFER, object.normalArray, _gl.DYNAMIC_DRAW);

			state.enableAttribute(attributes.normal);

			_gl.vertexAttribPointer(attributes.normal, 3, _gl.FLOAT, false, 0, 0);
		}

		if (object.hasUvs && material.map) {

			_gl.bindBuffer(_gl.ARRAY_BUFFER, buffers.uv);
			_gl.bufferData(_gl.ARRAY_BUFFER, object.uvArray, _gl.DYNAMIC_DRAW);

			state.enableAttribute(attributes.uv);

			_gl.vertexAttribPointer(attributes.uv, 2, _gl.FLOAT, false, 0, 0);
		}

		if (object.hasColors && material.vertexColors !== NoColors) {

			_gl.bindBuffer(_gl.ARRAY_BUFFER, buffers.color);
			_gl.bufferData(_gl.ARRAY_BUFFER, object.colorArray, _gl.DYNAMIC_DRAW);

			state.enableAttribute(attributes.color);

			_gl.vertexAttribPointer(attributes.color, 3, _gl.FLOAT, false, 0, 0);
		}

		state.disableUnusedAttributes();

		_gl.drawArrays(_gl.TRIANGLES, 0, object.count);

		object.count = 0;
	};

	this.renderBufferDirect = function (camera, fog, geometry, material, object, group) {

		setMaterial(material);

		var program = setProgram(camera, fog, material, object);

		var updateBuffers = false;
		var geometryProgram = geometry.id + '_' + program.id + '_' + material.wireframe;

		if (geometryProgram !== _currentGeometryProgram) {

			_currentGeometryProgram = geometryProgram;
			updateBuffers = true;
		}

		// morph targets

		var morphTargetInfluences = object.morphTargetInfluences;

		if (morphTargetInfluences !== undefined) {

			var activeInfluences = [];

			for (var i = 0, l = morphTargetInfluences.length; i < l; i++) {

				var influence = morphTargetInfluences[i];
				activeInfluences.push([influence, i]);
			}

			activeInfluences.sort(absNumericalSort);

			if (activeInfluences.length > 8) {

				activeInfluences.length = 8;
			}

			var morphAttributes = geometry.morphAttributes;

			for (var i = 0, l = activeInfluences.length; i < l; i++) {

				var influence = activeInfluences[i];
				morphInfluences[i] = influence[0];

				if (influence[0] !== 0) {

					var index = influence[1];

					if (material.morphTargets === true && morphAttributes.position) geometry.addAttribute('morphTarget' + i, morphAttributes.position[index]);
					if (material.morphNormals === true && morphAttributes.normal) geometry.addAttribute('morphNormal' + i, morphAttributes.normal[index]);
				} else {

					if (material.morphTargets === true) geometry.removeAttribute('morphTarget' + i);
					if (material.morphNormals === true) geometry.removeAttribute('morphNormal' + i);
				}
			}

			for (var i = activeInfluences.length, il = morphInfluences.length; i < il; i++) {

				morphInfluences[i] = 0.0;
			}

			program.getUniforms().setValue(_gl, 'morphTargetInfluences', morphInfluences);

			updateBuffers = true;
		}

		//

		var index = geometry.index;
		var position = geometry.attributes.position;
		var rangeFactor = 1;

		if (material.wireframe === true) {

			index = objects.getWireframeAttribute(geometry);
			rangeFactor = 2;
		}

		var renderer;

		if (index !== null) {

			renderer = indexedBufferRenderer;
			renderer.setIndex(index);
		} else {

			renderer = bufferRenderer;
		}

		if (updateBuffers) {

			setupVertexAttributes(material, program, geometry);

			if (index !== null) {

				_gl.bindBuffer(_gl.ELEMENT_ARRAY_BUFFER, objects.getAttributeBuffer(index));
			}
		}

		//

		var dataCount = 0;

		if (index !== null) {

			dataCount = index.count;
		} else if (position !== undefined) {

			dataCount = position.count;
		}

		var rangeStart = geometry.drawRange.start * rangeFactor;
		var rangeCount = geometry.drawRange.count * rangeFactor;

		var groupStart = group !== null ? group.start * rangeFactor : 0;
		var groupCount = group !== null ? group.count * rangeFactor : Infinity;

		var drawStart = Math.max(rangeStart, groupStart);
		var drawEnd = Math.min(dataCount, rangeStart + rangeCount, groupStart + groupCount) - 1;

		var drawCount = Math.max(0, drawEnd - drawStart + 1);

		if (drawCount === 0) return;

		//

		if (object.isMesh) {

			if (material.wireframe === true) {

				state.setLineWidth(material.wireframeLinewidth * getTargetPixelRatio());
				renderer.setMode(_gl.LINES);
			} else {

				switch (object.drawMode) {

					case TrianglesDrawMode:
						renderer.setMode(_gl.TRIANGLES);
						break;

					case TriangleStripDrawMode:
						renderer.setMode(_gl.TRIANGLE_STRIP);
						break;

					case TriangleFanDrawMode:
						renderer.setMode(_gl.TRIANGLE_FAN);
						break;

				}
			}
		} else if (object.isLine) {

			var lineWidth = material.linewidth;

			if (lineWidth === undefined) lineWidth = 1; // Not using Line*Material

			state.setLineWidth(lineWidth * getTargetPixelRatio());

			if (object.isLineSegments) {

				renderer.setMode(_gl.LINES);
			} else {

				renderer.setMode(_gl.LINE_STRIP);
			}
		} else if (object.isPoints) {

			renderer.setMode(_gl.POINTS);
		}

		if (geometry && geometry.isInstancedBufferGeometry) {

			if (geometry.maxInstancedCount > 0) {

				renderer.renderInstances(geometry, drawStart, drawCount);
			}
		} else {

			renderer.render(drawStart, drawCount);
		}
	};

	function setupVertexAttributes(material, program, geometry, startIndex) {

		var extension;

		if (geometry && geometry.isInstancedBufferGeometry) {

			extension = extensions.get('ANGLE_instanced_arrays');

			if (extension === null) {

				console.error('THREE.WebGLRenderer.setupVertexAttributes: using THREE.InstancedBufferGeometry but hardware does not support extension ANGLE_instanced_arrays.');
				return;
			}
		}

		if (startIndex === undefined) startIndex = 0;

		state.initAttributes();

		var geometryAttributes = geometry.attributes;

		var programAttributes = program.getAttributes();

		var materialDefaultAttributeValues = material.defaultAttributeValues;

		for (var name in programAttributes) {

			var programAttribute = programAttributes[name];

			if (programAttribute >= 0) {

				var geometryAttribute = geometryAttributes[name];

				if (geometryAttribute !== undefined) {

					var type = _gl.FLOAT;
					var array = geometryAttribute.array;
					var normalized = geometryAttribute.normalized;

					if (array instanceof Float32Array) {

						type = _gl.FLOAT;
					} else if (array instanceof Float64Array) {

						console.warn("Unsupported data buffer format: Float64Array");
					} else if (array instanceof Uint16Array) {

						type = _gl.UNSIGNED_SHORT;
					} else if (array instanceof Int16Array) {

						type = _gl.SHORT;
					} else if (array instanceof Uint32Array) {

						type = _gl.UNSIGNED_INT;
					} else if (array instanceof Int32Array) {

						type = _gl.INT;
					} else if (array instanceof Int8Array) {

						type = _gl.BYTE;
					} else if (array instanceof Uint8Array) {

						type = _gl.UNSIGNED_BYTE;
					}

					var size = geometryAttribute.itemSize;
					var buffer = objects.getAttributeBuffer(geometryAttribute);

					if (geometryAttribute.isInterleavedBufferAttribute) {

						var data = geometryAttribute.data;
						var stride = data.stride;
						var offset = geometryAttribute.offset;

						if (data && data.isInstancedInterleavedBuffer) {

							state.enableAttributeAndDivisor(programAttribute, data.meshPerAttribute, extension);

							if (geometry.maxInstancedCount === undefined) {

								geometry.maxInstancedCount = data.meshPerAttribute * data.count;
							}
						} else {

							state.enableAttribute(programAttribute);
						}

						_gl.bindBuffer(_gl.ARRAY_BUFFER, buffer);
						_gl.vertexAttribPointer(programAttribute, size, type, normalized, stride * data.array.BYTES_PER_ELEMENT, (startIndex * stride + offset) * data.array.BYTES_PER_ELEMENT);
					} else {

						if (geometryAttribute.isInstancedBufferAttribute) {

							state.enableAttributeAndDivisor(programAttribute, geometryAttribute.meshPerAttribute, extension);

							if (geometry.maxInstancedCount === undefined) {

								geometry.maxInstancedCount = geometryAttribute.meshPerAttribute * geometryAttribute.count;
							}
						} else {

							state.enableAttribute(programAttribute);
						}

						_gl.bindBuffer(_gl.ARRAY_BUFFER, buffer);
						_gl.vertexAttribPointer(programAttribute, size, type, normalized, 0, startIndex * size * geometryAttribute.array.BYTES_PER_ELEMENT);
					}
				} else if (materialDefaultAttributeValues !== undefined) {

					var value = materialDefaultAttributeValues[name];

					if (value !== undefined) {

						switch (value.length) {

							case 2:
								_gl.vertexAttrib2fv(programAttribute, value);
								break;

							case 3:
								_gl.vertexAttrib3fv(programAttribute, value);
								break;

							case 4:
								_gl.vertexAttrib4fv(programAttribute, value);
								break;

							default:
								_gl.vertexAttrib1fv(programAttribute, value);

						}
					}
				}
			}
		}

		state.disableUnusedAttributes();
	}

	// Sorting

	function absNumericalSort(a, b) {

		return Math.abs(b[0]) - Math.abs(a[0]);
	}

	function painterSortStable(a, b) {

		if (a.object.renderOrder !== b.object.renderOrder) {

			return a.object.renderOrder - b.object.renderOrder;
		} else if (a.material.program && b.material.program && a.material.program !== b.material.program) {

			return a.material.program.id - b.material.program.id;
		} else if (a.material.id !== b.material.id) {

			return a.material.id - b.material.id;
		} else if (a.z !== b.z) {

			return a.z - b.z;
		} else {

			return a.id - b.id;
		}
	}

	function reversePainterSortStable(a, b) {

		if (a.object.renderOrder !== b.object.renderOrder) {

			return a.object.renderOrder - b.object.renderOrder;
		}if (a.z !== b.z) {

			return b.z - a.z;
		} else {

			return a.id - b.id;
		}
	}

	// Rendering

	this.render = function (scene, camera, renderTarget, forceClear) {

		if (camera !== undefined && camera.isCamera !== true) {

			console.error('THREE.WebGLRenderer.render: camera is not an instance of THREE.Camera.');
			return;
		}

		// reset caching for this frame

		_currentGeometryProgram = '';
		_currentMaterialId = -1;
		_currentCamera = null;

		// update scene graph

		if (scene.autoUpdate === true) scene.updateMatrixWorld();

		// update camera matrices and frustum

		if (camera.parent === null) camera.updateMatrixWorld();

		camera.matrixWorldInverse.getInverse(camera.matrixWorld);

		_projScreenMatrix.multiplyMatrices(camera.projectionMatrix, camera.matrixWorldInverse);
		_frustum.setFromMatrix(_projScreenMatrix);

		lights.length = 0;

		opaqueObjectsLastIndex = -1;
		transparentObjectsLastIndex = -1;

		sprites.length = 0;
		lensFlares.length = 0;

		_localClippingEnabled = this.localClippingEnabled;
		_clippingEnabled = _clipping.init(this.clippingPlanes, _localClippingEnabled, camera);

		projectObject(scene, camera);

		opaqueObjects.length = opaqueObjectsLastIndex + 1;
		transparentObjects.length = transparentObjectsLastIndex + 1;

		if (_this.sortObjects === true) {

			opaqueObjects.sort(painterSortStable);
			transparentObjects.sort(reversePainterSortStable);
		}

		//

		if (_clippingEnabled) _clipping.beginShadows();

		setupShadows(lights);

		shadowMap.render(scene, camera);

		setupLights(lights, camera);

		if (_clippingEnabled) _clipping.endShadows();

		//

		_infoRender.calls = 0;
		_infoRender.vertices = 0;
		_infoRender.faces = 0;
		_infoRender.points = 0;

		if (renderTarget === undefined) {

			renderTarget = null;
		}

		this.setRenderTarget(renderTarget);

		//

		var background = scene.background;

		if (background === null) {

			glClearColor(_clearColor.r, _clearColor.g, _clearColor.b, _clearAlpha);
		} else if (background && background.isColor) {

			glClearColor(background.r, background.g, background.b, 1);
			forceClear = true;
		}

		if (this.autoClear || forceClear) {

			this.clear(this.autoClearColor, this.autoClearDepth, this.autoClearStencil);
		}

		if (background && background.isCubeTexture) {

			backgroundCamera2.projectionMatrix.copy(camera.projectionMatrix);

			backgroundCamera2.matrixWorld.extractRotation(camera.matrixWorld);
			backgroundCamera2.matrixWorldInverse.getInverse(backgroundCamera2.matrixWorld);

			backgroundBoxMesh.material.uniforms["tCube"].value = background;
			backgroundBoxMesh.modelViewMatrix.multiplyMatrices(backgroundCamera2.matrixWorldInverse, backgroundBoxMesh.matrixWorld);

			objects.update(backgroundBoxMesh);

			_this.renderBufferDirect(backgroundCamera2, null, backgroundBoxMesh.geometry, backgroundBoxMesh.material, backgroundBoxMesh, null);
		} else if (background && background.isTexture) {

			backgroundPlaneMesh.material.map = background;

			objects.update(backgroundPlaneMesh);

			_this.renderBufferDirect(backgroundCamera, null, backgroundPlaneMesh.geometry, backgroundPlaneMesh.material, backgroundPlaneMesh, null);
		}

		//

		if (scene.overrideMaterial) {

			var overrideMaterial = scene.overrideMaterial;

			renderObjects(opaqueObjects, scene, camera, overrideMaterial);
			renderObjects(transparentObjects, scene, camera, overrideMaterial);
		} else {

			// opaque pass (front-to-back order)

			state.setBlending(NoBlending);
			renderObjects(opaqueObjects, scene, camera);

			// transparent pass (back-to-front order)

			renderObjects(transparentObjects, scene, camera);
		}

		// custom render plugins (post pass)

		spritePlugin.render(scene, camera);
		lensFlarePlugin.render(scene, camera, _currentViewport);

		// Generate mipmap if we're using any kind of mipmap filtering

		if (renderTarget) {

			textures.updateRenderTargetMipmap(renderTarget);
		}

		// Ensure depth buffer writing is enabled so it can be cleared on next render

		state.setDepthTest(true);
		state.setDepthWrite(true);
		state.setColorWrite(true);

		// _gl.finish();
	};

	function pushRenderItem(object, geometry, material, z, group) {

		var array, index;

		// allocate the next position in the appropriate array

		if (material.transparent) {

			array = transparentObjects;
			index = ++transparentObjectsLastIndex;
		} else {

			array = opaqueObjects;
			index = ++opaqueObjectsLastIndex;
		}

		// recycle existing render item or grow the array

		var renderItem = array[index];

		if (renderItem !== undefined) {

			renderItem.id = object.id;
			renderItem.object = object;
			renderItem.geometry = geometry;
			renderItem.material = material;
			renderItem.z = _vector3.z;
			renderItem.group = group;
		} else {

			renderItem = {
				id: object.id,
				object: object,
				geometry: geometry,
				material: material,
				z: _vector3.z,
				group: group
			};

			// assert( index === array.length );
			array.push(renderItem);
		}
	}

	// TODO Duplicated code (Frustum)

	function isObjectViewable(object) {

		var geometry = object.geometry;

		if (geometry.boundingSphere === null) geometry.computeBoundingSphere();

		_sphere.copy(geometry.boundingSphere).applyMatrix4(object.matrixWorld);

		return isSphereViewable(_sphere);
	}

	function isSpriteViewable(sprite) {

		_sphere.center.set(0, 0, 0);
		_sphere.radius = 0.7071067811865476;
		_sphere.applyMatrix4(sprite.matrixWorld);

		return isSphereViewable(_sphere);
	}

	function isSphereViewable(sphere) {

		if (!_frustum.intersectsSphere(sphere)) return false;

		var numPlanes = _clipping.numPlanes;

		if (numPlanes === 0) return true;

		var planes = _this.clippingPlanes,
		    center = sphere.center,
		    negRad = -sphere.radius,
		    i = 0;

		do {

			// out when deeper than radius in the negative halfspace
			if (planes[i].distanceToPoint(center) < negRad) return false;
		} while (++i !== numPlanes);

		return true;
	}

	function projectObject(object, camera) {

		if (object.visible === false) return;

		var visible = (object.layers.mask & camera.layers.mask) !== 0;

		if (visible) {

			if (object.isLight) {

				lights.push(object);
			} else if (object.isSprite) {

				if (object.frustumCulled === false || isSpriteViewable(object) === true) {

					sprites.push(object);
				}
			} else if (object.isLensFlare) {

				lensFlares.push(object);
			} else if (object.isImmediateRenderObject) {

				if (_this.sortObjects === true) {

					_vector3.setFromMatrixPosition(object.matrixWorld);
					_vector3.applyProjection(_projScreenMatrix);
				}

				pushRenderItem(object, null, object.material, _vector3.z, null);
			} else if (object.isMesh || object.isLine || object.isPoints) {

				if (object.isSkinnedMesh) {

					object.skeleton.update();
				}

				if (object.frustumCulled === false || isObjectViewable(object) === true) {

					var material = object.material;

					if (material.visible === true) {

						if (_this.sortObjects === true) {

							_vector3.setFromMatrixPosition(object.matrixWorld);
							_vector3.applyProjection(_projScreenMatrix);
						}

						var geometry = objects.update(object);

						if (material.isMultiMaterial) {

							var groups = geometry.groups;
							var materials = material.materials;

							for (var i = 0, l = groups.length; i < l; i++) {

								var group = groups[i];
								var groupMaterial = materials[group.materialIndex];

								if (groupMaterial.visible === true) {

									pushRenderItem(object, geometry, groupMaterial, _vector3.z, group);
								}
							}
						} else {

							pushRenderItem(object, geometry, material, _vector3.z, null);
						}
					}
				}
			}
		}

		var children = object.children;

		for (var i = 0, l = children.length; i < l; i++) {

			projectObject(children[i], camera);
		}
	}

	function renderObjects(renderList, scene, camera, overrideMaterial) {

		for (var i = 0, l = renderList.length; i < l; i++) {

			var renderItem = renderList[i];

			var object = renderItem.object;
			var geometry = renderItem.geometry;
			var material = overrideMaterial === undefined ? renderItem.material : overrideMaterial;
			var group = renderItem.group;

			object.modelViewMatrix.multiplyMatrices(camera.matrixWorldInverse, object.matrixWorld);
			object.normalMatrix.getNormalMatrix(object.modelViewMatrix);

			object.onBeforeRender(_this, scene, camera, geometry, material, group);

			if (object.isImmediateRenderObject) {

				setMaterial(material);

				var program = setProgram(camera, scene.fog, material, object);

				_currentGeometryProgram = '';

				object.render(function (object) {

					_this.renderBufferImmediate(object, program, material);
				});
			} else {

				_this.renderBufferDirect(camera, scene.fog, geometry, material, object, group);
			}

			object.onAfterRender(_this, scene, camera, geometry, material, group);
		}
	}

	function initMaterial(material, fog, object) {

		var materialProperties = properties.get(material);

		var parameters = programCache.getParameters(material, _lights, fog, _clipping.numPlanes, _clipping.numIntersection, object);

		var code = programCache.getProgramCode(material, parameters);

		var program = materialProperties.program;
		var programChange = true;

		if (program === undefined) {

			// new material
			material.addEventListener('dispose', onMaterialDispose);
		} else if (program.code !== code) {

			// changed glsl or parameters
			releaseMaterialProgramReference(material);
		} else if (parameters.shaderID !== undefined) {

			// same glsl and uniform list
			return;
		} else {

			// only rebuild uniform list
			programChange = false;
		}

		if (programChange) {

			if (parameters.shaderID) {

				var shader = ShaderLib[parameters.shaderID];

				materialProperties.__webglShader = {
					name: material.type,
					uniforms: UniformsUtils.clone(shader.uniforms),
					vertexShader: shader.vertexShader,
					fragmentShader: shader.fragmentShader
				};
			} else {

				materialProperties.__webglShader = {
					name: material.type,
					uniforms: material.uniforms,
					vertexShader: material.vertexShader,
					fragmentShader: material.fragmentShader
				};
			}

			material.__webglShader = materialProperties.__webglShader;

			program = programCache.acquireProgram(material, parameters, code);

			materialProperties.program = program;
			material.program = program;
		}

		var attributes = program.getAttributes();

		if (material.morphTargets) {

			material.numSupportedMorphTargets = 0;

			for (var i = 0; i < _this.maxMorphTargets; i++) {

				if (attributes['morphTarget' + i] >= 0) {

					material.numSupportedMorphTargets++;
				}
			}
		}

		if (material.morphNormals) {

			material.numSupportedMorphNormals = 0;

			for (var i = 0; i < _this.maxMorphNormals; i++) {

				if (attributes['morphNormal' + i] >= 0) {

					material.numSupportedMorphNormals++;
				}
			}
		}

		var uniforms = materialProperties.__webglShader.uniforms;

		if (!material.isShaderMaterial && !material.isRawShaderMaterial || material.clipping === true) {

			materialProperties.numClippingPlanes = _clipping.numPlanes;
			materialProperties.numIntersection = _clipping.numIntersection;
			uniforms.clippingPlanes = _clipping.uniform;
		}

		materialProperties.fog = fog;

		// store the light setup it was created for

		materialProperties.lightsHash = _lights.hash;

		if (material.lights) {

			// wire up the material to this renderer's lighting state

			uniforms.ambientLightColor.value = _lights.ambient;
			uniforms.directionalLights.value = _lights.directional;
			uniforms.spotLights.value = _lights.spot;
			uniforms.pointLights.value = _lights.point;
			uniforms.hemisphereLights.value = _lights.hemi;

			uniforms.directionalShadowMap.value = _lights.directionalShadowMap;
			uniforms.directionalShadowMatrix.value = _lights.directionalShadowMatrix;
			uniforms.spotShadowMap.value = _lights.spotShadowMap;
			uniforms.spotShadowMatrix.value = _lights.spotShadowMatrix;
			uniforms.pointShadowMap.value = _lights.pointShadowMap;
			uniforms.pointShadowMatrix.value = _lights.pointShadowMatrix;
		}

		var progUniforms = materialProperties.program.getUniforms(),
		    uniformsList = WebGLUniforms.seqWithValue(progUniforms.seq, uniforms);

		materialProperties.uniformsList = uniformsList;
	}

	function setMaterial(material) {

		material.side === DoubleSide ? state.disable(_gl.CULL_FACE) : state.enable(_gl.CULL_FACE);

		state.setFlipSided(material.side === BackSide);

		material.transparent === true ? state.setBlending(material.blending, material.blendEquation, material.blendSrc, material.blendDst, material.blendEquationAlpha, material.blendSrcAlpha, material.blendDstAlpha, material.premultipliedAlpha) : state.setBlending(NoBlending);

		state.setDepthFunc(material.depthFunc);
		state.setDepthTest(material.depthTest);
		state.setDepthWrite(material.depthWrite);
		state.setColorWrite(material.colorWrite);
		state.setPolygonOffset(material.polygonOffset, material.polygonOffsetFactor, material.polygonOffsetUnits);
	}

	function setProgram(camera, fog, material, object) {

		_usedTextureUnits = 0;

		var materialProperties = properties.get(material);

		if (_clippingEnabled) {

			if (_localClippingEnabled || camera !== _currentCamera) {

				var useCache = camera === _currentCamera && material.id === _currentMaterialId;

				// we might want to call this function with some ClippingGroup
				// object instead of the material, once it becomes feasible
				// (#8465, #8379)
				_clipping.setState(material.clippingPlanes, material.clipIntersection, material.clipShadows, camera, materialProperties, useCache);
			}
		}

		if (material.needsUpdate === false) {

			if (materialProperties.program === undefined) {

				material.needsUpdate = true;
			} else if (material.fog && materialProperties.fog !== fog) {

				material.needsUpdate = true;
			} else if (material.lights && materialProperties.lightsHash !== _lights.hash) {

				material.needsUpdate = true;
			} else if (materialProperties.numClippingPlanes !== undefined && (materialProperties.numClippingPlanes !== _clipping.numPlanes || materialProperties.numIntersection !== _clipping.numIntersection)) {

				material.needsUpdate = true;
			}
		}

		if (material.needsUpdate) {

			initMaterial(material, fog, object);
			material.needsUpdate = false;
		}

		var refreshProgram = false;
		var refreshMaterial = false;
		var refreshLights = false;

		var program = materialProperties.program,
		    p_uniforms = program.getUniforms(),
		    m_uniforms = materialProperties.__webglShader.uniforms;

		if (program.id !== _currentProgram) {

			_gl.useProgram(program.program);
			_currentProgram = program.id;

			refreshProgram = true;
			refreshMaterial = true;
			refreshLights = true;
		}

		if (material.id !== _currentMaterialId) {

			_currentMaterialId = material.id;

			refreshMaterial = true;
		}

		if (refreshProgram || camera !== _currentCamera) {

			p_uniforms.set(_gl, camera, 'projectionMatrix');

			if (capabilities.logarithmicDepthBuffer) {

				p_uniforms.setValue(_gl, 'logDepthBufFC', 2.0 / (Math.log(camera.far + 1.0) / Math.LN2));
			}

			if (camera !== _currentCamera) {

				_currentCamera = camera;

				// lighting uniforms depend on the camera so enforce an update
				// now, in case this material supports lights - or later, when
				// the next material that does gets activated:

				refreshMaterial = true; // set to true on material change
				refreshLights = true; // remains set until update done
			}

			// load material specific uniforms
			// (shader material also gets them for the sake of genericity)

			if (material.isShaderMaterial || material.isMeshPhongMaterial || material.isMeshStandardMaterial || material.envMap) {

				var uCamPos = p_uniforms.map.cameraPosition;

				if (uCamPos !== undefined) {

					uCamPos.setValue(_gl, _vector3.setFromMatrixPosition(camera.matrixWorld));
				}
			}

			if (material.isMeshPhongMaterial || material.isMeshLambertMaterial || material.isMeshBasicMaterial || material.isMeshStandardMaterial || material.isShaderMaterial || material.skinning) {

				p_uniforms.setValue(_gl, 'viewMatrix', camera.matrixWorldInverse);
			}

			p_uniforms.set(_gl, _this, 'toneMappingExposure');
			p_uniforms.set(_gl, _this, 'toneMappingWhitePoint');
		}

		// skinning uniforms must be set even if material didn't change
		// auto-setting of texture unit for bone texture must go before other textures
		// not sure why, but otherwise weird things happen

		if (material.skinning) {

			p_uniforms.setOptional(_gl, object, 'bindMatrix');
			p_uniforms.setOptional(_gl, object, 'bindMatrixInverse');

			var skeleton = object.skeleton;

			if (skeleton) {

				if (capabilities.floatVertexTextures && skeleton.useVertexTexture) {

					p_uniforms.set(_gl, skeleton, 'boneTexture');
					p_uniforms.set(_gl, skeleton, 'boneTextureWidth');
					p_uniforms.set(_gl, skeleton, 'boneTextureHeight');
				} else {

					p_uniforms.setOptional(_gl, skeleton, 'boneMatrices');
				}
			}
		}

		if (refreshMaterial) {

			if (material.lights) {

				// the current material requires lighting info

				// note: all lighting uniforms are always set correctly
				// they simply reference the renderer's state for their
				// values
				//
				// use the current material's .needsUpdate flags to set
				// the GL state when required

				markUniformsLightsNeedsUpdate(m_uniforms, refreshLights);
			}

			// refresh uniforms common to several materials

			if (fog && material.fog) {

				refreshUniformsFog(m_uniforms, fog);
			}

			if (material.isMeshBasicMaterial || material.isMeshLambertMaterial || material.isMeshPhongMaterial || material.isMeshStandardMaterial || material.isMeshDepthMaterial) {

				refreshUniformsCommon(m_uniforms, material);
			}

			// refresh single material specific uniforms

			if (material.isLineBasicMaterial) {

				refreshUniformsLine(m_uniforms, material);
			} else if (material.isLineDashedMaterial) {

				refreshUniformsLine(m_uniforms, material);
				refreshUniformsDash(m_uniforms, material);
			} else if (material.isPointsMaterial) {

				refreshUniformsPoints(m_uniforms, material);
			} else if (material.isMeshLambertMaterial) {

				refreshUniformsLambert(m_uniforms, material);
			} else if (material.isMeshPhongMaterial) {

				refreshUniformsPhong(m_uniforms, material);
			} else if (material.isMeshPhysicalMaterial) {

				refreshUniformsPhysical(m_uniforms, material);
			} else if (material.isMeshStandardMaterial) {

				refreshUniformsStandard(m_uniforms, material);
			} else if (material.isMeshDepthMaterial) {

				if (material.displacementMap) {

					m_uniforms.displacementMap.value = material.displacementMap;
					m_uniforms.displacementScale.value = material.displacementScale;
					m_uniforms.displacementBias.value = material.displacementBias;
				}
			} else if (material.isMeshNormalMaterial) {

				m_uniforms.opacity.value = material.opacity;
			}

			WebGLUniforms.upload(_gl, materialProperties.uniformsList, m_uniforms, _this);
		}

		// common matrices

		p_uniforms.set(_gl, object, 'modelViewMatrix');
		p_uniforms.set(_gl, object, 'normalMatrix');
		p_uniforms.setValue(_gl, 'modelMatrix', object.matrixWorld);

		return program;
	}

	// Uniforms (refresh uniforms objects)

	function refreshUniformsCommon(uniforms, material) {

		uniforms.opacity.value = material.opacity;

		uniforms.diffuse.value = material.color;

		if (material.emissive) {

			uniforms.emissive.value.copy(material.emissive).multiplyScalar(material.emissiveIntensity);
		}

		uniforms.map.value = material.map;
		uniforms.specularMap.value = material.specularMap;
		uniforms.alphaMap.value = material.alphaMap;

		if (material.aoMap) {

			uniforms.aoMap.value = material.aoMap;
			uniforms.aoMapIntensity.value = material.aoMapIntensity;
		}

		// uv repeat and offset setting priorities
		// 1. color map
		// 2. specular map
		// 3. normal map
		// 4. bump map
		// 5. alpha map
		// 6. emissive map

		var uvScaleMap;

		if (material.map) {

			uvScaleMap = material.map;
		} else if (material.specularMap) {

			uvScaleMap = material.specularMap;
		} else if (material.displacementMap) {

			uvScaleMap = material.displacementMap;
		} else if (material.normalMap) {

			uvScaleMap = material.normalMap;
		} else if (material.bumpMap) {

			uvScaleMap = material.bumpMap;
		} else if (material.roughnessMap) {

			uvScaleMap = material.roughnessMap;
		} else if (material.metalnessMap) {

			uvScaleMap = material.metalnessMap;
		} else if (material.alphaMap) {

			uvScaleMap = material.alphaMap;
		} else if (material.emissiveMap) {

			uvScaleMap = material.emissiveMap;
		}

		if (uvScaleMap !== undefined) {

			// backwards compatibility
			if (uvScaleMap.isWebGLRenderTarget) {

				uvScaleMap = uvScaleMap.texture;
			}

			var offset = uvScaleMap.offset;
			var repeat = uvScaleMap.repeat;

			uniforms.offsetRepeat.value.set(offset.x, offset.y, repeat.x, repeat.y);
		}

		uniforms.envMap.value = material.envMap;

		// don't flip CubeTexture envMaps, flip everything else:
		//  WebGLRenderTargetCube will be flipped for backwards compatibility
		//  WebGLRenderTargetCube.texture will be flipped because it's a Texture and NOT a CubeTexture
		// this check must be handled differently, or removed entirely, if WebGLRenderTargetCube uses a CubeTexture in the future
		uniforms.flipEnvMap.value = !(material.envMap && material.envMap.isCubeTexture) ? 1 : -1;

		uniforms.reflectivity.value = material.reflectivity;
		uniforms.refractionRatio.value = material.refractionRatio;
	}

	function refreshUniformsLine(uniforms, material) {

		uniforms.diffuse.value = material.color;
		uniforms.opacity.value = material.opacity;
	}

	function refreshUniformsDash(uniforms, material) {

		uniforms.dashSize.value = material.dashSize;
		uniforms.totalSize.value = material.dashSize + material.gapSize;
		uniforms.scale.value = material.scale;
	}

	function refreshUniformsPoints(uniforms, material) {

		uniforms.diffuse.value = material.color;
		uniforms.opacity.value = material.opacity;
		uniforms.size.value = material.size * _pixelRatio;
		uniforms.scale.value = _height * 0.5;

		uniforms.map.value = material.map;

		if (material.map !== null) {

			var offset = material.map.offset;
			var repeat = material.map.repeat;

			uniforms.offsetRepeat.value.set(offset.x, offset.y, repeat.x, repeat.y);
		}
	}

	function refreshUniformsFog(uniforms, fog) {

		uniforms.fogColor.value = fog.color;

		if (fog.isFog) {

			uniforms.fogNear.value = fog.near;
			uniforms.fogFar.value = fog.far;
		} else if (fog.isFogExp2) {

			uniforms.fogDensity.value = fog.density;
		}
	}

	function refreshUniformsLambert(uniforms, material) {

		if (material.lightMap) {

			uniforms.lightMap.value = material.lightMap;
			uniforms.lightMapIntensity.value = material.lightMapIntensity;
		}

		if (material.emissiveMap) {

			uniforms.emissiveMap.value = material.emissiveMap;
		}
	}

	function refreshUniformsPhong(uniforms, material) {

		uniforms.specular.value = material.specular;
		uniforms.shininess.value = Math.max(material.shininess, 1e-4); // to prevent pow( 0.0, 0.0 )

		if (material.lightMap) {

			uniforms.lightMap.value = material.lightMap;
			uniforms.lightMapIntensity.value = material.lightMapIntensity;
		}

		if (material.emissiveMap) {

			uniforms.emissiveMap.value = material.emissiveMap;
		}

		if (material.bumpMap) {

			uniforms.bumpMap.value = material.bumpMap;
			uniforms.bumpScale.value = material.bumpScale;
		}

		if (material.normalMap) {

			uniforms.normalMap.value = material.normalMap;
			uniforms.normalScale.value.copy(material.normalScale);
		}

		if (material.displacementMap) {

			uniforms.displacementMap.value = material.displacementMap;
			uniforms.displacementScale.value = material.displacementScale;
			uniforms.displacementBias.value = material.displacementBias;
		}
	}

	function refreshUniformsStandard(uniforms, material) {

		uniforms.roughness.value = material.roughness;
		uniforms.metalness.value = material.metalness;

		if (material.roughnessMap) {

			uniforms.roughnessMap.value = material.roughnessMap;
		}

		if (material.metalnessMap) {

			uniforms.metalnessMap.value = material.metalnessMap;
		}

		if (material.lightMap) {

			uniforms.lightMap.value = material.lightMap;
			uniforms.lightMapIntensity.value = material.lightMapIntensity;
		}

		if (material.emissiveMap) {

			uniforms.emissiveMap.value = material.emissiveMap;
		}

		if (material.bumpMap) {

			uniforms.bumpMap.value = material.bumpMap;
			uniforms.bumpScale.value = material.bumpScale;
		}

		if (material.normalMap) {

			uniforms.normalMap.value = material.normalMap;
			uniforms.normalScale.value.copy(material.normalScale);
		}

		if (material.displacementMap) {

			uniforms.displacementMap.value = material.displacementMap;
			uniforms.displacementScale.value = material.displacementScale;
			uniforms.displacementBias.value = material.displacementBias;
		}

		if (material.envMap) {

			//uniforms.envMap.value = material.envMap; // part of uniforms common
			uniforms.envMapIntensity.value = material.envMapIntensity;
		}
	}

	function refreshUniformsPhysical(uniforms, material) {

		uniforms.clearCoat.value = material.clearCoat;
		uniforms.clearCoatRoughness.value = material.clearCoatRoughness;

		refreshUniformsStandard(uniforms, material);
	}

	// If uniforms are marked as clean, they don't need to be loaded to the GPU.

	function markUniformsLightsNeedsUpdate(uniforms, value) {

		uniforms.ambientLightColor.needsUpdate = value;

		uniforms.directionalLights.needsUpdate = value;
		uniforms.pointLights.needsUpdate = value;
		uniforms.spotLights.needsUpdate = value;
		uniforms.hemisphereLights.needsUpdate = value;
	}

	// Lighting

	function setupShadows(lights) {

		var lightShadowsLength = 0;

		for (var i = 0, l = lights.length; i < l; i++) {

			var light = lights[i];

			if (light.castShadow) {

				_lights.shadows[lightShadowsLength++] = light;
			}
		}

		_lights.shadows.length = lightShadowsLength;
	}

	function setupLights(lights, camera) {

		var l,
		    ll,
		    light,
		    r = 0,
		    g = 0,
		    b = 0,
		    color,
		    intensity,
		    distance,
		    shadowMap,
		    viewMatrix = camera.matrixWorldInverse,
		    directionalLength = 0,
		    pointLength = 0,
		    spotLength = 0,
		    hemiLength = 0;

		for (l = 0, ll = lights.length; l < ll; l++) {

			light = lights[l];

			color = light.color;
			intensity = light.intensity;
			distance = light.distance;

			shadowMap = light.shadow && light.shadow.map ? light.shadow.map.texture : null;

			if (light.isAmbientLight) {

				r += color.r * intensity;
				g += color.g * intensity;
				b += color.b * intensity;
			} else if (light.isDirectionalLight) {

				var uniforms = lightCache.get(light);

				uniforms.color.copy(light.color).multiplyScalar(light.intensity);
				uniforms.direction.setFromMatrixPosition(light.matrixWorld);
				_vector3.setFromMatrixPosition(light.target.matrixWorld);
				uniforms.direction.sub(_vector3);
				uniforms.direction.transformDirection(viewMatrix);

				uniforms.shadow = light.castShadow;

				if (light.castShadow) {

					uniforms.shadowBias = light.shadow.bias;
					uniforms.shadowRadius = light.shadow.radius;
					uniforms.shadowMapSize = light.shadow.mapSize;
				}

				_lights.directionalShadowMap[directionalLength] = shadowMap;
				_lights.directionalShadowMatrix[directionalLength] = light.shadow.matrix;
				_lights.directional[directionalLength++] = uniforms;
			} else if (light.isSpotLight) {

				var uniforms = lightCache.get(light);

				uniforms.position.setFromMatrixPosition(light.matrixWorld);
				uniforms.position.applyMatrix4(viewMatrix);

				uniforms.color.copy(color).multiplyScalar(intensity);
				uniforms.distance = distance;

				uniforms.direction.setFromMatrixPosition(light.matrixWorld);
				_vector3.setFromMatrixPosition(light.target.matrixWorld);
				uniforms.direction.sub(_vector3);
				uniforms.direction.transformDirection(viewMatrix);

				uniforms.coneCos = Math.cos(light.angle);
				uniforms.penumbraCos = Math.cos(light.angle * (1 - light.penumbra));
				uniforms.decay = light.distance === 0 ? 0.0 : light.decay;

				uniforms.shadow = light.castShadow;

				if (light.castShadow) {

					uniforms.shadowBias = light.shadow.bias;
					uniforms.shadowRadius = light.shadow.radius;
					uniforms.shadowMapSize = light.shadow.mapSize;
				}

				_lights.spotShadowMap[spotLength] = shadowMap;
				_lights.spotShadowMatrix[spotLength] = light.shadow.matrix;
				_lights.spot[spotLength++] = uniforms;
			} else if (light.isPointLight) {

				var uniforms = lightCache.get(light);

				uniforms.position.setFromMatrixPosition(light.matrixWorld);
				uniforms.position.applyMatrix4(viewMatrix);

				uniforms.color.copy(light.color).multiplyScalar(light.intensity);
				uniforms.distance = light.distance;
				uniforms.decay = light.distance === 0 ? 0.0 : light.decay;

				uniforms.shadow = light.castShadow;

				if (light.castShadow) {

					uniforms.shadowBias = light.shadow.bias;
					uniforms.shadowRadius = light.shadow.radius;
					uniforms.shadowMapSize = light.shadow.mapSize;
				}

				_lights.pointShadowMap[pointLength] = shadowMap;

				if (_lights.pointShadowMatrix[pointLength] === undefined) {

					_lights.pointShadowMatrix[pointLength] = new Matrix4();
				}

				// for point lights we set the shadow matrix to be a translation-only matrix
				// equal to inverse of the light's position
				_vector3.setFromMatrixPosition(light.matrixWorld).negate();
				_lights.pointShadowMatrix[pointLength].identity().setPosition(_vector3);

				_lights.point[pointLength++] = uniforms;
			} else if (light.isHemisphereLight) {

				var uniforms = lightCache.get(light);

				uniforms.direction.setFromMatrixPosition(light.matrixWorld);
				uniforms.direction.transformDirection(viewMatrix);
				uniforms.direction.normalize();

				uniforms.skyColor.copy(light.color).multiplyScalar(intensity);
				uniforms.groundColor.copy(light.groundColor).multiplyScalar(intensity);

				_lights.hemi[hemiLength++] = uniforms;
			}
		}

		_lights.ambient[0] = r;
		_lights.ambient[1] = g;
		_lights.ambient[2] = b;

		_lights.directional.length = directionalLength;
		_lights.spot.length = spotLength;
		_lights.point.length = pointLength;
		_lights.hemi.length = hemiLength;

		_lights.hash = directionalLength + ',' + pointLength + ',' + spotLength + ',' + hemiLength + ',' + _lights.shadows.length;
	}

	// GL state setting

	this.setFaceCulling = function (cullFace, frontFaceDirection) {

		state.setCullFace(cullFace);
		state.setFlipSided(frontFaceDirection === FrontFaceDirectionCW);
	};

	// Textures

	function allocTextureUnit() {

		var textureUnit = _usedTextureUnits;

		if (textureUnit >= capabilities.maxTextures) {

			console.warn('WebGLRenderer: trying to use ' + textureUnit + ' texture units while this GPU supports only ' + capabilities.maxTextures);
		}

		_usedTextureUnits += 1;

		return textureUnit;
	}

	this.allocTextureUnit = allocTextureUnit;

	// this.setTexture2D = setTexture2D;
	this.setTexture2D = function () {

		var warned = false;

		// backwards compatibility: peel texture.texture
		return function setTexture2D(texture, slot) {

			if (texture && texture.isWebGLRenderTarget) {

				if (!warned) {

					console.warn("THREE.WebGLRenderer.setTexture2D: don't use render targets as textures. Use their .texture property instead.");
					warned = true;
				}

				texture = texture.texture;
			}

			textures.setTexture2D(texture, slot);
		};
	}();

	this.setTexture = function () {

		var warned = false;

		return function setTexture(texture, slot) {

			if (!warned) {

				console.warn("THREE.WebGLRenderer: .setTexture is deprecated, use setTexture2D instead.");
				warned = true;
			}

			textures.setTexture2D(texture, slot);
		};
	}();

	this.setTextureCube = function () {

		var warned = false;

		return function setTextureCube(texture, slot) {

			// backwards compatibility: peel texture.texture
			if (texture && texture.isWebGLRenderTargetCube) {

				if (!warned) {

					console.warn("THREE.WebGLRenderer.setTextureCube: don't use cube render targets as textures. Use their .texture property instead.");
					warned = true;
				}

				texture = texture.texture;
			}

			// currently relying on the fact that WebGLRenderTargetCube.texture is a Texture and NOT a CubeTexture
			// TODO: unify these code paths
			if (texture && texture.isCubeTexture || Array.isArray(texture.image) && texture.image.length === 6) {

				// CompressedTexture can have Array in image :/

				// this function alone should take care of cube textures
				textures.setTextureCube(texture, slot);
			} else {

				// assumed: texture property of THREE.WebGLRenderTargetCube

				textures.setTextureCubeDynamic(texture, slot);
			}
		};
	}();

	this.getCurrentRenderTarget = function () {

		return _currentRenderTarget;
	};

	this.setRenderTarget = function (renderTarget) {

		_currentRenderTarget = renderTarget;

		if (renderTarget && properties.get(renderTarget).__webglFramebuffer === undefined) {

			textures.setupRenderTarget(renderTarget);
		}

		var isCube = renderTarget && renderTarget.isWebGLRenderTargetCube;
		var framebuffer;

		if (renderTarget) {

			var renderTargetProperties = properties.get(renderTarget);

			if (isCube) {

				framebuffer = renderTargetProperties.__webglFramebuffer[renderTarget.activeCubeFace];
			} else {

				framebuffer = renderTargetProperties.__webglFramebuffer;
			}

			_currentScissor.copy(renderTarget.scissor);
			_currentScissorTest = renderTarget.scissorTest;

			_currentViewport.copy(renderTarget.viewport);
		} else {

			framebuffer = null;

			_currentScissor.copy(_scissor).multiplyScalar(_pixelRatio);
			_currentScissorTest = _scissorTest;

			_currentViewport.copy(_viewport).multiplyScalar(_pixelRatio);
		}

		if (_currentFramebuffer !== framebuffer) {

			_gl.bindFramebuffer(_gl.FRAMEBUFFER, framebuffer);
			_currentFramebuffer = framebuffer;
		}

		state.scissor(_currentScissor);
		state.setScissorTest(_currentScissorTest);

		state.viewport(_currentViewport);

		if (isCube) {

			var textureProperties = properties.get(renderTarget.texture);
			_gl.framebufferTexture2D(_gl.FRAMEBUFFER, _gl.COLOR_ATTACHMENT0, _gl.TEXTURE_CUBE_MAP_POSITIVE_X + renderTarget.activeCubeFace, textureProperties.__webglTexture, renderTarget.activeMipMapLevel);
		}
	};

	this.readRenderTargetPixels = function (renderTarget, x, y, width, height, buffer) {

		if ((renderTarget && renderTarget.isWebGLRenderTarget) === false) {

			console.error('THREE.WebGLRenderer.readRenderTargetPixels: renderTarget is not THREE.WebGLRenderTarget.');
			return;
		}

		var framebuffer = properties.get(renderTarget).__webglFramebuffer;

		if (framebuffer) {

			var restore = false;

			if (framebuffer !== _currentFramebuffer) {

				_gl.bindFramebuffer(_gl.FRAMEBUFFER, framebuffer);

				restore = true;
			}

			try {

				var texture = renderTarget.texture;
				var textureFormat = texture.format;
				var textureType = texture.type;

				if (textureFormat !== RGBAFormat && paramThreeToGL(textureFormat) !== _gl.getParameter(_gl.IMPLEMENTATION_COLOR_READ_FORMAT)) {

					console.error('THREE.WebGLRenderer.readRenderTargetPixels: renderTarget is not in RGBA or implementation defined format.');
					return;
				}

				if (textureType !== UnsignedByteType && paramThreeToGL(textureType) !== _gl.getParameter(_gl.IMPLEMENTATION_COLOR_READ_TYPE) && // IE11, Edge and Chrome Mac < 52 (#9513)
				!(textureType === FloatType && (extensions.get('OES_texture_float') || extensions.get('WEBGL_color_buffer_float'))) && // Chrome Mac >= 52 and Firefox
				!(textureType === HalfFloatType && extensions.get('EXT_color_buffer_half_float'))) {

					console.error('THREE.WebGLRenderer.readRenderTargetPixels: renderTarget is not in UnsignedByteType or implementation defined type.');
					return;
				}

				if (_gl.checkFramebufferStatus(_gl.FRAMEBUFFER) === _gl.FRAMEBUFFER_COMPLETE) {

					// the following if statement ensures valid read requests (no out-of-bounds pixels, see #8604)

					if (x >= 0 && x <= renderTarget.width - width && y >= 0 && y <= renderTarget.height - height) {

						_gl.readPixels(x, y, width, height, paramThreeToGL(textureFormat), paramThreeToGL(textureType), buffer);
					}
				} else {

					console.error('THREE.WebGLRenderer.readRenderTargetPixels: readPixels from renderTarget failed. Framebuffer not complete.');
				}
			} finally {

				if (restore) {

					_gl.bindFramebuffer(_gl.FRAMEBUFFER, _currentFramebuffer);
				}
			}
		}
	};

	// Map three.js constants to WebGL constants

	function paramThreeToGL(p) {

		var extension;

		if (p === RepeatWrapping) return _gl.REPEAT;
		if (p === ClampToEdgeWrapping) return _gl.CLAMP_TO_EDGE;
		if (p === MirroredRepeatWrapping) return _gl.MIRRORED_REPEAT;

		if (p === NearestFilter) return _gl.NEAREST;
		if (p === NearestMipMapNearestFilter) return _gl.NEAREST_MIPMAP_NEAREST;
		if (p === NearestMipMapLinearFilter) return _gl.NEAREST_MIPMAP_LINEAR;

		if (p === LinearFilter) return _gl.LINEAR;
		if (p === LinearMipMapNearestFilter) return _gl.LINEAR_MIPMAP_NEAREST;
		if (p === LinearMipMapLinearFilter) return _gl.LINEAR_MIPMAP_LINEAR;

		if (p === UnsignedByteType) return _gl.UNSIGNED_BYTE;
		if (p === UnsignedShort4444Type) return _gl.UNSIGNED_SHORT_4_4_4_4;
		if (p === UnsignedShort5551Type) return _gl.UNSIGNED_SHORT_5_5_5_1;
		if (p === UnsignedShort565Type) return _gl.UNSIGNED_SHORT_5_6_5;

		if (p === ByteType) return _gl.BYTE;
		if (p === ShortType) return _gl.SHORT;
		if (p === UnsignedShortType) return _gl.UNSIGNED_SHORT;
		if (p === IntType) return _gl.INT;
		if (p === UnsignedIntType) return _gl.UNSIGNED_INT;
		if (p === FloatType) return _gl.FLOAT;

		if (p === HalfFloatType) {

			extension = extensions.get('OES_texture_half_float');

			if (extension !== null) return extension.HALF_FLOAT_OES;
		}

		if (p === AlphaFormat) return _gl.ALPHA;
		if (p === RGBFormat) return _gl.RGB;
		if (p === RGBAFormat) return _gl.RGBA;
		if (p === LuminanceFormat) return _gl.LUMINANCE;
		if (p === LuminanceAlphaFormat) return _gl.LUMINANCE_ALPHA;
		if (p === DepthFormat) return _gl.DEPTH_COMPONENT;
		if (p === DepthStencilFormat) return _gl.DEPTH_STENCIL;

		if (p === AddEquation) return _gl.FUNC_ADD;
		if (p === SubtractEquation) return _gl.FUNC_SUBTRACT;
		if (p === ReverseSubtractEquation) return _gl.FUNC_REVERSE_SUBTRACT;

		if (p === ZeroFactor) return _gl.ZERO;
		if (p === OneFactor) return _gl.ONE;
		if (p === SrcColorFactor) return _gl.SRC_COLOR;
		if (p === OneMinusSrcColorFactor) return _gl.ONE_MINUS_SRC_COLOR;
		if (p === SrcAlphaFactor) return _gl.SRC_ALPHA;
		if (p === OneMinusSrcAlphaFactor) return _gl.ONE_MINUS_SRC_ALPHA;
		if (p === DstAlphaFactor) return _gl.DST_ALPHA;
		if (p === OneMinusDstAlphaFactor) return _gl.ONE_MINUS_DST_ALPHA;

		if (p === DstColorFactor) return _gl.DST_COLOR;
		if (p === OneMinusDstColorFactor) return _gl.ONE_MINUS_DST_COLOR;
		if (p === SrcAlphaSaturateFactor) return _gl.SRC_ALPHA_SATURATE;

		if (p === RGB_S3TC_DXT1_Format || p === RGBA_S3TC_DXT1_Format || p === RGBA_S3TC_DXT3_Format || p === RGBA_S3TC_DXT5_Format) {

			extension = extensions.get('WEBGL_compressed_texture_s3tc');

			if (extension !== null) {

				if (p === RGB_S3TC_DXT1_Format) return extension.COMPRESSED_RGB_S3TC_DXT1_EXT;
				if (p === RGBA_S3TC_DXT1_Format) return extension.COMPRESSED_RGBA_S3TC_DXT1_EXT;
				if (p === RGBA_S3TC_DXT3_Format) return extension.COMPRESSED_RGBA_S3TC_DXT3_EXT;
				if (p === RGBA_S3TC_DXT5_Format) return extension.COMPRESSED_RGBA_S3TC_DXT5_EXT;
			}
		}

		if (p === RGB_PVRTC_4BPPV1_Format || p === RGB_PVRTC_2BPPV1_Format || p === RGBA_PVRTC_4BPPV1_Format || p === RGBA_PVRTC_2BPPV1_Format) {

			extension = extensions.get('WEBGL_compressed_texture_pvrtc');

			if (extension !== null) {

				if (p === RGB_PVRTC_4BPPV1_Format) return extension.COMPRESSED_RGB_PVRTC_4BPPV1_IMG;
				if (p === RGB_PVRTC_2BPPV1_Format) return extension.COMPRESSED_RGB_PVRTC_2BPPV1_IMG;
				if (p === RGBA_PVRTC_4BPPV1_Format) return extension.COMPRESSED_RGBA_PVRTC_4BPPV1_IMG;
				if (p === RGBA_PVRTC_2BPPV1_Format) return extension.COMPRESSED_RGBA_PVRTC_2BPPV1_IMG;
			}
		}

		if (p === RGB_ETC1_Format) {

			extension = extensions.get('WEBGL_compressed_texture_etc1');

			if (extension !== null) return extension.COMPRESSED_RGB_ETC1_WEBGL;
		}

		if (p === MinEquation || p === MaxEquation) {

			extension = extensions.get('EXT_blend_minmax');

			if (extension !== null) {

				if (p === MinEquation) return extension.MIN_EXT;
				if (p === MaxEquation) return extension.MAX_EXT;
			}
		}

		if (p === UnsignedInt248Type) {

			extension = extensions.get('WEBGL_depth_texture');

			if (extension !== null) return extension.UNSIGNED_INT_24_8_WEBGL;
		}

		return 0;
	}
}

var MILLISECONDS_TO_SECONDS = 0.001;
var MAX_MOVE_DISTANCE = 5;
var MAX_MOVE_DISTANCE_SQ = MAX_MOVE_DISTANCE * MAX_MOVE_DISTANCE;
var TELEPORT_COOLDOWN = 250;
var TELEPORT_DISPLACEMENT = new Vector3();
var GROUND_HEIGHT = -0.07;

var BrowserEnvironment = function (_AbstractEventEmitter) {
  inherits(BrowserEnvironment, _AbstractEventEmitter);

  function BrowserEnvironment(options) {
    var _arguments = arguments;
    classCallCheck(this, BrowserEnvironment);

    var _this = possibleConstructorReturn(this, (BrowserEnvironment.__proto__ || Object.getPrototypeOf(BrowserEnvironment)).call(this));

    _this.options = Object.assign({}, BrowserEnvironment.DEFAULTS, options);
    _this.options.foregroundColor = _this.options.foregroundColor || complementColor(new Color(_this.options.backgroundColor)).getHex();

    _this.network = null;

    if (_this.options.nonstandardIPD !== null) {
      _this.options.nonstandardIPD *= 0.5;
    }

    _this.audioQueue = [];

    _this.zero = function () {
      if (!_this.lockMovement) {
        _this.input.zero();
        if (_this.quality === Quality.NONE) {
          _this.quality = Quality.HIGH;
        }
      }
    };

    var update = function update(dt) {
      dt *= MILLISECONDS_TO_SECONDS;
      movePlayer(dt);
      moveUI();
      doPicking();
      moveGround();
      if (_this.network) {
        _this.network.update(dt);
      }

      _this.emit("update", dt);
    };

    var movePlayer = function movePlayer(dt) {
      _this.input.update(dt);
    };

    _this.turns = new Angle(0);
    var followEuler = new Euler(),
        maxX = -Math.PI / 4,
        maxY = Math.PI / 6;

    var moveUI = function moveUI(dt) {
      _this.ui.position.copy(_this.input.stage.position);
      followEuler.setFromQuaternion(_this.input.head.quaternion);
      _this.turns.radians = followEuler.y;
      followEuler.set(maxX, _this.turns.radians, 0, "YXZ");
      _this.ui.quaternion.setFromEuler(followEuler);
    };

    var doPicking = function doPicking() {
      for (var i = _this.pickableObjects.length - 1; i >= 0; --i) {
        var inScene = false;
        for (var head = _this.pickableObjects[i].parent; head !== null; head = head.parent) {
          if (head === _this.scene) {
            inScene = true;
          }
        }
        if (!inScene) {
          _this.pickableObjects.splice(i, 1);
        }
      }
      _this.input.resolvePicking(_this.pickableObjects.filter(function (obj) {
        return !obj.disabled;
      }));
    };

    var moveGround = function moveGround() {
      if (_this.sky) {
        _this.sky.position.set(_this.input.head.position.x, 0, _this.input.head.position.z);
      }

      if (_this.ground) {
        _this.ground.position.set(Math.floor(_this.input.head.position.x), GROUND_HEIGHT, Math.floor(_this.input.head.position.z));
        if (_this.ground.material) {
          _this.ground.material.needsUpdate = true;
        }
      }
    };

    var animate = function animate(t) {
      var dt = t - lt,
          i,
          j;
      lt = t;
      update(dt);
      render();
      RAF(animate);
    };

    var render = function render() {
      _this.camera.position.set(0, 0, 0);
      _this.camera.quaternion.set(0, 0, 0, 1);
      _this.audio.setPlayer(_this.input.head.mesh);
      if (_this.input.VR.isPresenting) {
        _this.renderer.clear(true, true, true);

        var trans = _this.input.VR.getTransforms(_this.options.nearPlane, _this.options.nearPlane + _this.options.drawDistance);
        for (var i = 0; trans && i < trans.length; ++i) {
          var st = trans[i],
              v = st.viewport,
              side = 2 * i - 1;
          if (_this.options.nonstandardIPD !== null) {
            st.translation.x = Math.sign(st.translation.x) * _this.options.nonstandardIPD;
          }
          if (_this.options.nonstandardNeckLength !== null) {
            st.translation.y = _this.options.nonstandardNeckLength;
          }
          if (_this.options.nonstandardNeckDepth !== null) {
            st.translation.z = _this.options.nonstandardNeckDepth;
          }
          Entity.eyeBlankAll(i);
          _this.camera.projectionMatrix.copy(st.projection);
          _this.camera.translateOnAxis(st.translation, 1);
          _this.renderer.setViewport(v.left * resolutionScale, v.top * resolutionScale, v.width * resolutionScale, v.height * resolutionScale);
          _this.renderer.render(_this.scene, _this.camera);
          _this.camera.translateOnAxis(st.translation, -1);
        }
        _this.input.submitFrame();
      }

      if (!_this.input.VR.isPresenting || _this.input.VR.canMirror && !_this.options.disableMirroring) {
        _this.camera.fov = _this.options.defaultFOV;
        _this.camera.aspect = _this.renderer.domElement.width / _this.renderer.domElement.height;
        _this.camera.updateProjectionMatrix();
        _this.renderer.clear(true, true, true);
        if (_this.input.mousePointer.unproject) {
          _this.input.mousePointer.unproject.getInverse(_this.camera.projectionMatrix);
        }
        _this.renderer.setViewport(0, 0, _this.renderer.domElement.width, _this.renderer.domElement.height);
        _this.renderer.render(_this.scene, _this.camera);
      }
    };

    var modifyScreen = function modifyScreen() {
      var near = _this.options.nearPlane,
          far = near + _this.options.drawDistance,
          p = _this.input && _this.input.VR && _this.input.VR.getTransforms(near, far);

      if (p) {
        var canvasWidth = 0,
            canvasHeight = 0;

        for (var i = 0; i < p.length; ++i) {
          canvasWidth += p[i].viewport.width;
          canvasHeight = Math.max(canvasHeight, p[i].viewport.height);
        }

        _this.input.Mouse.commands.U.scale = 2 / canvasWidth;
        _this.input.Mouse.commands.V.scale = 2 / canvasHeight;

        canvasWidth = Math.floor(canvasWidth * resolutionScale);
        canvasHeight = Math.floor(canvasHeight * resolutionScale);

        _this.renderer.domElement.width = canvasWidth;
        _this.renderer.domElement.height = canvasHeight;
        if (!_this.timer) {
          render();
        }
      }
    };

    //
    // Initialize local variables
    //

    var lt = 0,
        currentHeading = 0,
        qPitch = new Quaternion(),
        vEye = new Vector3(),
        vBody = new Vector3(),
        modelFiles = {
      scene: _this.options.sceneModel,
      avatar: _this.options.avatarModel,
      button: _this.options.button && typeof _this.options.button.model === "string" && _this.options.button.model,
      font: _this.options.font
    },
        resolutionScale = 1;

    _this.factories = {
      button: Button2D,
      img: Image,
      section: Surface,
      textarea: TextBox,
      avatar: null,
      pre: {
        create: function create() {
          return new TextBox({
            tokenizer: PlainText$2,
            hideLineNumbers: true,
            readOnly: true
          });
        }
      }
    };

    _this.createElement = function (type) {
      if (_this.factories[type]) {
        return _this.factories[type].create();
      }
    };

    _this.appendChild = function (elem) {
      if (elem.isMesh) {
        _this.scene.add(elem);
      } else {
        return elem.addToBrowserEnvironment(_this, _this.scene);
      }
    };

    function setColor(model, color) {
      return model.children[0].material.color.set(color);
    }

    function complementColor(color) {
      var rgb = color.clone();
      var hsl = rgb.getHSL();
      hsl.h = hsl.h + 0.5;
      hsl.l = 1 - hsl.l;
      while (hsl.h > 1) {
        hsl.h -= 1;
      }rgb.setHSL(hsl.h, hsl.s, hsl.l);
      return rgb;
    }

    var modelsReady = ModelLoader.loadObjects(modelFiles).then(function (models) {
      window.text3D = function (font, size, text) {
        var geom = new TextGeometry(text, {
          font: font,
          size: size,
          height: size / 5,
          curveSegments: 2
        });
        geom.computeBoundingSphere();
        geom.computeBoundingBox();
        return geom;
      }.bind(window, models.font);

      if (models.scene) {
        buildScene(models.scene);
      }

      if (models.avatar) {
        _this.factories.avatar = new ModelLoader(models.avatar);
      }

      if (models.button) {
        _this.buttonFactory = new ButtonFactory(models.button, _this.options.button.options);
      } else {
        _this.buttonFactory = new ButtonFactory(colored(box$1(1, 1, 1), 0xff0000), {
          maxThrow: 0.1,
          minDeflection: 10,
          colorUnpressed: 0x7f0000,
          colorPressed: 0x007f00,
          toggle: true
        });
      }
    }).catch(function (err) {
      console.error(err);
      if (!_this.buttonFactory) {
        _this.buttonFactory = new ButtonFactory(colored(box$1(1, 1, 1), 0xff0000), {
          maxThrow: 0.1,
          minDeflection: 10,
          colorUnpressed: 0x7f0000,
          colorPressed: 0x007f00,
          toggle: true
        });
      }
    });

    //
    // Initialize public properties
    //

    _this.speech = new Speech(_this.options.speech);

    _this.audio = new Audio3D();

    if (_this.options.ambientSound) {
      _this.audio.load3DSound(_this.options.ambientSound, true, -1, 1, -1).then(function (aud) {
        if (!(aud.source instanceof MediaElementAudioSourceNode)) {
          aud.volume.gain.value = 0.1;
          aud.source.start();
        }
      }).catch(console.error.bind(console, "Audio3D loadSource"));
    }

    var documentReady = null;
    if (document.readyState === "complete") {
      documentReady = Promise.resolve("already");
    } else {
      documentReady = new Promise(function (resolve, reject) {
        document.addEventListener("readystatechange", function (evt) {
          if (document.readyState === "complete") {
            resolve("had to wait for it");
          }
        }, false);
      });
    }

    _this.music = new Music(_this.audio);

    _this.pickableObjects = [];

    _this.registerPickableObject = _this.pickableObjects.push.bind(_this.pickableObjects);

    _this.currentControl = null;

    var FADE_SPEED = 0.1;
    _this.fadeOut = function () {
      return new Promise(function (resolve, reject) {
        var timer = setInterval(function () {
          _this.fader.material.opacity += FADE_SPEED;
          if (_this.fader.material.opacity >= 1) {
            clearInterval(timer);
            resolve();
          }
        }, 10);
      });
    };

    _this.fadeIn = function () {
      return new Promise(function (resolve, reject) {
        var timer = setInterval(function () {
          _this.fader.material.opacity -= FADE_SPEED;
          if (_this.fader.material.opacity <= 0) {
            clearInterval(timer);
            resolve();
          }
        }, 10);
      });
    };

    _this.teleportAvailable = true;

    _this.transition = function (thunk, check, immediate) {
      if (immediate) {
        thunk();
        return Promise.resolve();
      } else if (!check || check()) {
        _this.fader.visible = true;
        return _this.fadeOut().then(thunk).then(function () {
          return _this.fadeIn();
        }).catch(console.warn.bind(console, "Error while transitioning")).then(function () {
          return _this.fader.visible = false;
        });
      }
    };

    _this.teleport = function (pos, immediate) {
      return _this.transition(function () {
        return _this.input.moveStage(pos);
      }, function () {
        return _this.teleportAvailable && TELEPORT_DISPLACEMENT.copy(pos).sub(_this.input.head.position).length() > 0.2;
      }, immediate);
    };

    var POSITION = new Vector3(),
        START_POINT = new Vector3();

    _this.selectControl = function (evt) {
      var hit = evt.hit,
          obj = hit && hit.object;

      if (_this.ground && obj === _this.ground) {
        if (evt.type === "exit") {
          evt.pointer.disk.visible = false;
        } else if (evt.type !== "exit") {
          POSITION.copy(evt.hit.point).sub(_this.input.head.position);

          var distSq = POSITION.x * POSITION.x + POSITION.z * POSITION.z;
          if (distSq > MAX_MOVE_DISTANCE_SQ) {
            var dist = Math.sqrt(distSq),
                factor = MAX_MOVE_DISTANCE / dist,
                y = POSITION.y;
            POSITION.y = 0;
            POSITION.multiplyScalar(factor);
            POSITION.y = y;
          }

          POSITION.add(_this.input.head.position);

          if (evt.type === "enter") {
            evt.pointer.disk.visible = true;
          } else if (evt.type === "pointerstart" || evt.type === "gazestart") {
            START_POINT.copy(POSITION);
          } else if (evt.type === "pointermove" || evt.type === "gazemove") {
            evt.pointer.moveTeleportPad(POSITION);
          } else if (evt.type === "pointerend" || evt.type === "gazecomplete") {
            START_POINT.sub(POSITION);
            var len = START_POINT.lengthSq();
            if (len < 0.01) {
              _this.teleport(POSITION);
            }
          }
        }
      }

      if (evt.type === "pointerstart" || evt.type === "gazecomplete") {
        var ctrl = obj && (obj.surface || obj.button);
        if (ctrl !== _this.currentControl) {
          if (_this.currentControl) {
            _this.currentControl.blur();
            _this.input.Mouse.commands.pitch.disabled = _this.input.Mouse.commands.heading.disabled = false;
          }
          _this.currentControl = ctrl;
          if (_this.currentControl) {
            _this.currentControl.focus();
            if (obj.surface) {
              _this.input.Mouse.commands.pitch.disabled = _this.input.Mouse.commands.heading.disabled = !_this.input.VR.isPresenting;
            }
          }
        }
      }

      if (_this.currentControl) {
        if (_this.currentControl.dispatchEvent) {
          _this.currentControl.dispatchEvent(evt);
        } else {
          console.log(_this.currentControl);
        }
      } else if (obj) {
        var handler = obj["on" + evt.type];
        if (handler) {
          handler(evt);
        }
      }
    };

    _this.options.scene = _this.scene = _this.options.scene || new Scene();

    if (_this.options.useFog) {
      _this.scene.fog = new FogExp2(_this.options.backgroundColor, 1 / Math.sqrt(_this.options.drawDistance));
    }

    _this.camera = new PerspectiveCamera(75, 1, _this.options.nearPlane, _this.options.nearPlane + _this.options.drawDistance);

    var skyReady = null;
    if (_this.options.skyTexture !== null) {
      skyReady = new Promise(function (resolve, reject) {
        var skyFunc = typeof _this.options.skyTexture === "number" ? colored : textured;

        var skyGeom = null;
        if (typeof _this.options.skyTexture === "string") {
          skyGeom = sphere(_this.options.drawDistance, 18, 9);
        } else {
          var skyDim = _this.options.drawDistance / Math.sqrt(2);
          skyGeom = box$1(skyDim, skyDim, skyDim);
        }

        _this.sky = skyFunc(skyGeom, _this.options.skyTexture, {
          side: BackSide,
          fog: false,
          unshaded: true,
          transparent: true,
          opacity: 1,
          resolve: resolve,
          progress: _this.options.progress
        });
      });
    } else {
      _this.sky = hub$1();
      skyReady = Promise.resolve();
    }

    if (_this.options.disableDefaultLighting) {
      _this.ambient = null;
      _this.sun = null;
    } else {

      _this.ambient = new AmbientLight(0xffffff, 0.5);

      _this.sun = light(0xffffff, 1, 50);
    }

    skyReady = skyReady.then(function () {
      _this.sky.name = "Sky";
      _this.scene.add(_this.sky);

      if (!_this.options.disableDefaultLighting) {
        _this.sky.add(_this.ambient);
        put(_this.sun).on(_this.sky).at(0, 10, 10);
      }
    });

    var groundReady = null;
    if (_this.options.groundTexture !== null) {
      groundReady = new Promise(function (resolve, reject) {
        var dim = 2 * _this.options.drawDistance;
        _this.ground = brick(_this.options.groundTexture, dim * 5, 0.1, dim * 5, {
          txtRepeatX: dim * 5,
          txtRepeatY: dim * 5,
          anisotropy: 8,
          resolve: resolve,
          progress: _this.options.progress
        });
        _this.registerPickableObject(_this.ground);
      });
    } else {
      _this.ground = hub$1();
      groundReady = Promise.resolve();
    }

    groundReady = groundReady.then(function () {
      _this.ground.name = "Ground";
      _this.scene.add(_this.ground);
    });

    _this.ui = hub$1().named("UI");
    _this.scene.add(_this.ui);

    var buildScene = function buildScene(sceneGraph) {
      sceneGraph.buttons = [];
      sceneGraph.traverse(function (child) {
        if (child.isButton) {
          sceneGraph.buttons.push(new Button3D(child.parent, child.name));
        }
        if (child.name) {
          sceneGraph[child.name] = child;
        }
      });
      _this.scene.add.apply(_this.scene, sceneGraph.children);
      _this.scene.traverse(function (obj) {
        if (_this.options.disableDefaultLighting && obj.material) {
          if (obj.material.map) {
            textured(obj, obj.material.map, {
              unshaded: true
            });
          } else {
            colored(obj, obj.material.color.getHex(), {
              unshaded: true
            });
          }
        }
        if (obj.name) {
          _this.scene[obj.name] = obj;
        }
      });
      if (sceneGraph.Camera) {
        _this.camera.position.copy(sceneGraph.Camera.position);
        _this.camera.quaternion.copy(sceneGraph.Camera.quaternion);
      }
      return sceneGraph;
    };

    var currentTimerObject = null;
    _this.timer = 0;
    var RAF = function RAF(callback) {
      currentTimerObject = _this.input.VR.currentDevice || window;
      if (_this.timer !== null) {
        _this.timer = currentTimerObject.requestAnimationFrame(callback);
      }
    };

    _this.goFullScreen = function (index, evt) {
      if (evt !== "Gaze") {
        var _ret = function () {
          var elem = !_this.input.VR.isStereo || isMobile && !_this.input.VR.isNativeMobileWebVR ? _this.options.fullScreenElement : _this.renderer.domElement;
          _this.input.VR.connect(index);
          return {
            v: _this.input.VR.requestPresent([{
              source: elem
            }]).catch(function (exp) {
              return console.error("whaaat", exp);
            }).then(function () {
              return elem.focus();
            })
          };
        }();

        if ((typeof _ret === "undefined" ? "undefined" : _typeof(_ret)) === "object") return _ret.v;
      }
    };

    _this.addAvatar = function (user) {
      console.log(user);
      _this.scene.add(user.stage);
      _this.scene.add(user.head);
    };

    _this.removeAvatar = function (user) {
      _this.scene.remove(user.stage);
      _this.scene.remove(user.head);
    };

    PointerLock.addChangeListener(function (evt) {
      if (_this.input.VR.isPresenting && !PointerLock.isActive) {
        _this.input.cancelVR();
      }
    });

    var fullScreenChange = function fullScreenChange(evt) {
      var presenting = !!_this.input.VR.isPresenting,
          cmd = (presenting ? "remove" : "add") + "Button";
      _this.input.Mouse[cmd]("dx", 0);
      _this.input.Mouse[cmd]("dy", 0);
      _this.input.Mouse.commands.U.disabled = _this.input.Mouse.commands.V.disabled = presenting && !_this.input.VR.isStereo;
      _this.input.Mouse.commands.heading.scale = presenting ? -1 : 1;
      _this.input.Mouse.commands.pitch.scale = presenting ? -1 : 1;
      if (!presenting) {
        _this.input.cancelVR();
      }
      modifyScreen();
    };

    window.addEventListener("vrdisplaypresentchange", fullScreenChange, false);
    window.addEventListener("resize", modifyScreen, false);
    window.addEventListener("blur", _this.stop, false);
    window.addEventListener("focus", _this.start, false);

    documentReady = documentReady.then(function () {
      if (_this.options.renderer) {
        _this.renderer = _this.options.renderer;
      } else {
        _this.renderer = new WebGLRenderer({
          canvas: cascadeElement(_this.options.canvasElement, "canvas", HTMLCanvasElement),
          context: _this.options.context,
          antialias: _this.options.antialias,
          alpha: true,
          logarithmicDepthBuffer: false
        });
        _this.renderer.autoClear = false;
        _this.renderer.sortObjects = true;
        _this.renderer.setClearColor(_this.options.backgroundColor);
        if (!_this.renderer.domElement.parentElement) {
          document.body.appendChild(_this.renderer.domElement);
        }
      }

      _this.options.fullScreenElement = document.querySelector(_this.options.fullScreenElement) || _this.renderer.domElement;
      var maxTabIndex = 0;
      var elementsWithTabIndex = document.querySelectorAll("[tabIndex]");
      for (var i = 0; i < elementsWithTabIndex.length; ++i) {
        maxTabIndex = Math.max(maxTabIndex, elementsWithTabIndex[i].tabIndex);
      }

      _this.renderer.domElement.tabIndex = maxTabIndex + 1;
      _this.renderer.domElement.addEventListener('webglcontextlost', _this.stop, false);
      _this.renderer.domElement.addEventListener('webglcontextrestored', _this.start, false);

      _this.input = new FPSInput(_this.options.fullScreenElement, _this.options);
      _this.input.addEventListener("zero", _this.zero, false);
      _this.input.VR.ready.then(function (displays) {
        return displays.forEach(function (display, i) {
          window.addEventListener("vrdisplayactivate", function (evt) {
            if (evt.display === display) {
              (function () {
                var exitVR = function exitVR() {
                  window.removeEventListener("vrdisplaydeactivate", exitVR);
                  _this.input.cancelVR();
                };
                window.addEventListener("vrdisplaydeactivate", exitVR, false);
                _this.goFullScreen(i);
              })();
            }
          }, false);
        });
      });

      _this.fader = colored(box$1(1, 1, 1), _this.options.backgroundColor, {
        opacity: 0,
        fog: false,
        transparent: true,
        unshaded: true,
        side: BackSide
      });
      _this.fader.visible = false;
      _this.input.head.root.add(_this.fader);

      Pointer.EVENTS.forEach(function (evt) {
        return _this.input.addEventListener(evt, _this.selectControl.bind(_this), false);
      });

      _this.input.forward(_this, Pointer.EVENTS);

      if (!_this.options.disableKeyboard) {
        var setFalse = function setFalse(evt) {
          evt.returnValue = false;
        };

        var keyDown = function keyDown(evt) {
          if (_this.input.VR.isPresenting) {
            if (evt.keyCode === Keys.ESCAPE && !_this.input.VR.isPolyfilled) {
              _this.input.cancelVR();
            }
          }

          if (!_this.lockMovement) {
            _this.input.Keyboard.dispatchEvent(evt);
          } else if (_this.currentControl) {
            _this.currentControl.keyDown(evt);
          }

          _this.emit("keydown", evt);
        },
            keyUp = function keyUp(evt) {
          if (!_this.lockMovement) {
            _this.input.Keyboard.dispatchEvent(evt);
          } else if (_this.currentControl) {
            _this.currentControl.keyUp(evt);
          }

          _this.emit("keyup", evt);
        },
            withCurrentControl = function withCurrentControl(name) {
          return function (evt) {
            if (_this.currentControl) {
              if (_this.currentControl[name]) {
                _this.currentControl[name](evt);
              } else {
                console.warn("Couldn't find %s on %o", name, _this.currentControl);
              }
            }
          };
        };

        window.addEventListener("keydown", keyDown, false);

        window.addEventListener("keyup", keyUp, false);

        window.addEventListener("paste", withCurrentControl("readClipboard"), false);
        window.addEventListener("wheel", withCurrentControl("readWheel"), false);

        var focusClipboard = function focusClipboard(evt) {
          if (_this.lockMovement) {
            var cmdName = _this.input.Keyboard.operatingSystem.makeCommandName(evt, _this.input.Keyboard.codePage);
            if (cmdName === "CUT" || cmdName === "COPY") {
              surrogate.style.display = "block";
              surrogate.focus();
            }
          }
        };

        var clipboardOperation = function clipboardOperation(evt) {
          if (_this.currentControl) {
            _this.currentControl[evt.type + "SelectedText"](evt);
            if (!evt.returnValue) {
              evt.preventDefault();
            }
            surrogate.style.display = "none";
            _this.currentControl.focus();
          }
        };

        // the `surrogate` textarea makes clipboard events possible
        var surrogate = cascadeElement("primrose-surrogate-textarea", "textarea", HTMLTextAreaElement),
            surrogateContainer = makeHidingContainer("primrose-surrogate-textarea-container", surrogate);

        surrogateContainer.style.position = "absolute";
        surrogateContainer.style.overflow = "hidden";
        surrogateContainer.style.width = 0;
        surrogateContainer.style.height = 0;

        surrogate.addEventListener("beforecopy", setFalse, false);
        surrogate.addEventListener("copy", clipboardOperation, false);
        surrogate.addEventListener("beforecut", setFalse, false);
        surrogate.addEventListener("cut", clipboardOperation, false);
        document.body.insertBefore(surrogateContainer, document.body.children[0]);

        window.addEventListener("beforepaste", setFalse, false);
        window.addEventListener("keydown", focusClipboard, true);
      }

      _this.input.head.add(_this.camera);

      return _this.input.ready.then(function () {
        if (_this.options.fullScreenButtonContainer) {
          _this.insertFullScreenButtons(_this.options.fullScreenButtonContainer);
        }
      });
    });

    var allReady = Promise.all([skyReady, groundReady, modelsReady, documentReady]).then(function () {
      _this.renderer.domElement.style.cursor = "default";
      if (_this.options.enableShadows && _this.sun) {
        _this.renderer.shadowMap.enabled = true;
        _this.renderer.shadowMap.type = PCFSoftShadowMap;
        _this.sun.castShadow = true;
        _this.sun.shadow.mapSize.width = _this.sun.shadow.mapSize.height = _this.options.shadowMapSize;
        if (_this.ground.material) {
          _this.ground.receiveShadow = true;
          _this.ground.castShadow = true;
        }
      }
      _this.input.VR.displays.forEach(function (display) {
        if (display.DOMElement !== undefined) {
          display.DOMElement = _this.renderer.domElement;
        }
      });
      _this.input.VR.connect(0);

      _this.emit("ready");
      window.dispatchEvent(new CustomEvent("vrbrowserenvironmentready", {
        detail: _this
      }));
    });

    _this.start = function () {
      allReady.then(function () {
        _this.audio.start();
        lt = performance.now() * MILLISECONDS_TO_SECONDS;
        RAF(animate);
      });
    };

    _this.stop = function () {
      if (currentTimerObject) {
        currentTimerObject.cancelAnimationFrame(_this.timer);
        _this.audio.stop();
        _this.timer = null;
      }
    };

    Object.defineProperties(_this, {
      quality: {
        get: function get() {
          return _this.options.quality;
        },
        set: function set(v) {
          if (0 <= v && v < PIXEL_SCALES.length) {
            _this.options.quality = v;
            resolutionScale = PIXEL_SCALES[v];
          }
          allReady.then(modifyScreen);
        }
      }
    });

    _this.quality = _this.options.quality;

    if (window.alert.toString().indexOf("native code") > -1) {
      // overwrite the native alert functions so they can't be called while in
      // full screen VR mode.

      var rerouteDialog = function rerouteDialog(oldFunction, newFunction) {
        if (!newFunction) {
          newFunction = function newFunction() {};
        }
        return function () {
          if (_this.input.VR.isPresenting) {
            newFunction();
          } else {
            oldFunction.apply(window, _arguments);
          }
        };
      };

      window.alert = rerouteDialog(window.alert);
      window.confirm = rerouteDialog(window.confirm);
      window.prompt = rerouteDialog(window.prompt);
    }

    _this.start();
    return _this;
  }

  createClass(BrowserEnvironment, [{
    key: "connect",
    value: function connect(socket, userName) {

      if (!this.network) {
        this.network = new Manager(this.input, this.audio, this.factories, this.options);
        this.network.addEventListener("addavatar", this.addAvatar);
        this.network.addEventListener("removeavatar", this.removeAvatar);
      }
      return this.network && this.network.connect(socket, userName);
    }
  }, {
    key: "disconnect",
    value: function disconnect() {

      return this.network && this.network.disconnect();
    }
  }, {
    key: "setAudioFromUser",
    value: function setAudioFromUser(userName, audioElement) {

      this.audioQueue.push([userName, audioElement]);
      if (this.network) {
        while (this.audioQueue.length > 0) {
          this.network.setAudioFromUser.apply(this.network, this.audioQueue.shift());
        }
      }
    }
  }, {
    key: "insertFullScreenButtons",
    value: function insertFullScreenButtons(containerSpec) {
      var _this2 = this;

      var container = document.querySelector(containerSpec);
      var newButton = function newButton(title, text, thunk) {
        var btn = document.createElement("button");
        btn.type = "button";
        btn.title = title;
        btn.appendChild(document.createTextNode(text));
        btn.addEventListener("click", thunk, false);
        container.appendChild(btn);
        return btn;
      };

      var buttons = this.displays.map(function (display, i) {
        var enterVR = _this2.goFullScreen.bind(_this2, i),
            btn = newButton(display.displayName, display.displayName, enterVR),
            isStereo = VR.isStereoDisplay(display);
        btn.className = isStereo ? "stereo" : "mono";
        return btn;
      });

      buttons.push(newButton("Primrose", "Primrose", function () {
        return document.location = "https://www.primrosevr.com";
      }));
      return buttons;
    }
  }, {
    key: "lockMovement",
    get: function get() {

      return this.currentControl && this.currentControl.lockMovement;
    }
  }, {
    key: "displays",
    get: function get() {

      return this.input.VR.displays;
    }
  }]);
  return BrowserEnvironment;
}(AbstractEventEmitter);

BrowserEnvironment.DEFAULTS = {
  antialias: true,
  quality: Quality.MAXIMUM,
  useGaze: false,
  useFog: false,
  avatarHeight: 1.65,
  walkSpeed: 2,
  disableKeyboard: false,
  enableShadows: false,
  shadowMapSize: 1024,
  // The acceleration applied to falling objects.
  gravity: 9.8,
  // The amount of time in seconds to require gazes on objects before triggering the gaze event.
  gazeLength: 1.5,
  // By default, what we see in the VR view will get mirrored to a regular view on the primary screen. Set to true to improve performance.
  disableMirroring: false,
  // By default, a single light is added to the scene,
  disableDefaultLighting: false,
  // The color that WebGL clears the background with before drawing.
  backgroundColor: 0xafbfff,
  // the textures to use for the sky and ground
  skyTexture: null,
  groundTexture: null,
  // the near plane of the camera.
  nearPlane: 0.01,
  // the far plane of the camera.
  drawDistance: 100,
  // the field of view to use in non-VR settings.
  defaultFOV: 75,
  // The sound to play on loop in the background.
  ambientSound: null,
  // HTML5 canvas element, if one had already been created.
  canvasElement: "frontBuffer",
  // Three.js renderer, if one had already been created.
  renderer: null,
  // A WebGL context to use, if one had already been created.
  context: null,
  // Three.js scene, if one had already been created.
  scene: null,
  // I highly suggest you don't go down the road that requires setting this. I will not help you understand what it does, because I would rather you just not use it.
  nonstandardIPD: null,
  // This is an experimental feature for setting the height of a user's "neck" on orientation-only systems (such as Google Cardboard and Samsung Gear VR) to create a more realistic feel.
  nonstandardNeckLength: null,
  nonstandardNeckDepth: null,
  showHeadPointer: true
};

Object.assign(window, liveAPI);
var monthNames = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
var dayNames = ["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
var primaryColor = 0xb09000;
var secondaryColor = 0x303030;
var todayColor = 0x306080;
var angleThreshold = 15;
var primaryMaterial = material({ color: primaryColor });
var secondaryMaterial = material({ color: secondaryColor });
var todayMaterial = material({ color: todayColor });
var objs = [];
var dim = 50;
var boxSize = dim / 300;
var boxSpacing = boxSize * 50;
var textSize = boxSize * 0.333;
var hDim = dim / 2;
var deg2rad = function deg2rad(a) {
  return Math.PI * a / 180;
};
var rad2deg = function rad2deg(r) {
  return 180 * r / Math.PI;
};
var rand = function rand() {
  return number(0, dim);
};
var randDegree = function randDegree() {
  return number(-Math.PI, Math.PI);
};
var env = new BrowserEnvironment({
  groundTexture: "../images/deck.png",
  font: "../fonts/helvetiker_regular.typeface.json",
  backgroundColor: 0x000000,
  useFog: true,
  drawDistance: hDim / 2,
  enableShadows: true,
  fullScreenButtonContainer: "#fullScreenButtonContainer"
});

var months = null;
var currentMonth = new Date().getMonth();

function setColor() {
  var _this = this;

  var date = new Date(),
      month = date.getMonth(),
      d = date.getDate() - 1;
  objs.forEach(function (o, i) {
    if (o.box === _this) {
      o.box.material = primaryMaterial;
      o.lbl.material = secondaryMaterial;
      o.bev.material = secondaryMaterial;
    } else {
      o.box.material = month === currentMonth && i === d ? todayMaterial : secondaryMaterial;
      o.lbl.material = primaryMaterial;
      o.bev.material = primaryMaterial;
    }
  });
}

function text(text, size) {
  size = size || 1;
  return text3D(size * textSize, text).center();
}

function hitDate(month, date) {
  var d = new Date();
  d.setMonth(month);
  d.setDate(date);
  env.speech.speak(dayNames[d.getDay()] + ", " + monthNames[month] + " " + date + ", " + d.getFullYear());
}

function showMonth(month) {
  env.ui.children.splice(0);

  var date = new Date(),
      today = null;

  if (month == date.getMonth()) {
    today = date.getDate();
  } else {
    date.setMonth(month);
  }

  date.setDate(1);
  var x = 0,
      y = 0,
      lastDay = 0;
  env.ui.add(months[month].latLon(-72, 0));
  while (date.getMonth() === month) {
    x = date.getDay();
    if (x < lastDay) {
      ++y;
    }

    var d = date.getDate(),
        o = objs[d - 1];
    if (today == d) {
      o.box.material = todayMaterial;
    } else {
      o.box.material = secondaryMaterial;
    }
    o.box.onenter = setColor.bind(o.box);
    env.registerPickableObject(o.box);
    o.box.onselect = hitDate.bind(null, month, d);
    env.ui.add(o.hub.latLon(boxSpacing * (y - 2) - 45, boxSpacing * (3 - x)));
    date.setDate(date.getDate() + 1);
    lastDay = x;
  }
  env.ground.onenter = setColor.bind(env.ground);
}

env.addEventListener("ready", function () {
  env.insertFullScreenButtons("body");
  months = monthNames.map(function (month) {
    return text(month, 3.5).colored(primaryColor).named(month);
  });

  for (var d = 0; d < 31; ++d) {
    var o = hub(),
        b = box(boxSize).colored(secondaryColor).named("box" + (d + 1)),
        bev = box(boxSize * 1.1, boxSize * 1.1, boxSize * 0.9).colored(primaryColor).named("bev" + (d + 1)),
        t = text((d + 1).toString()).colored(primaryColor).named("txt" + (d + 1));

    t.castShadow = true;
    b.receiveShadow = true;
    t.position.z = 0.08;
    o.add(b);
    b.add(bev);
    b.add(t);
    objs.push({
      hub: o,
      box: b,
      lbl: t,
      bev: bev
    });
  }

  showMonth(currentMonth);

  env.ui.add(text("Turn left or right to change month", 0.5).colored(primaryColor).named("inst1").latLon(-20, 0, 1));
  env.ui.add(text("Click on dates.", 0.5).colored(primaryColor).named("inst2").latLon(-17, 0, 1));
});

var lt = 0;
env.addEventListener("update", function () {
  var dt = env.turns.degrees - lt;
  if (Math.abs(dt) > angleThreshold) {
    currentMonth = currentMonth + (dt < 0 ? 1 : -1);
    if (currentMonth < 0) {
      currentMonth += 12;
    } else if (currentMonth >= 12) {
      currentMonth -= 12;
    }
    showMonth(currentMonth);
    lt = env.turns.degrees;
  }
});

})));
